/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     actions.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/18/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version
 04/03/97    KMK      modified for release with text book

 DESCRIPTION:

 This module implements the Translate::UserActions function which is called
 by the SGFramework parser.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#if   defined(UNIX)
#include <unistd.h>
#else
#include <io.h>
#endif

#include <fcntl.h>
#include "types.h"
#include "tokens.h"
#include "states.h"
#include "actions.h"
#include "error.h"
#include "source.h"
#include "symbol.h"
#include "mathnode.h"
#include "index.h"
#include "statemnt.h"
#include "xlatinfo.h"
#include "fileio.h"

#define LTA_LEN		  7
#define INDEX_ERROR	 -1u
#define INDEX_NONE	 -2u
#define EXPRMODE_NORMAL	  1u
#define EXPRMODE_PRECOMP  2u


///////////////////////////////////////////////////////////////////////////////
// Function:	 AdjustTokenStack
// Scope:	 Private
// Description:	 This function pops c tokens from the token state and stores
//		 them in reverse order in the Token array.
void AdjustTokenStack ( UINT c, TOKEN *atoken, TokenStack *stackToken )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp AdjustTokenStack\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  for (--c; c != UINT (-1); --c) atoken[c] = stackToken->pop();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareIndex
// Scope:	 Private
// Description:	 This function attempts to add an index to the symbol table.
//		 The name of the index is stored in the ptok->data.sz buffer.
UINT DeclareIndex ( TOKEN *ptok )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp DeclareIndex\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok->data.sz;

  // report an error if the symbol is already defined
  if (symTable->Retrieve(psz))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok->iPtr);
    return INDEX_ERROR;
  }

  UINT c = symTable->cSymbols(TYPE_INDEX);
  SymIndex *psym = new SymIndex(psz, c);
  symTable->Store(psym);
  return c;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareVariable
// Scope:	 Private
// Description:	 This function attempts to add a variable to the symbol table.
//		 The name of the variable is stored in the ptok->data.sz buffer.
void DeclareVariable ( TOKEN *ptok, UINT &cVariables )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp DeclareVariable\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok->data.sz;

  // report an error if the symbol is already defined
  if (symTable->Retrieve(psz))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok->iPtr);
    return;
  }

#if defined(P16BIT)
  // report an error if the number of variables exceeds the maximum
  if (cVariables >= MAX_VARIABLES)
  { error->AddMessage(EHM_MAXVAR, psz, ptok->iPtr);
    return;
  }
#endif

  // construct and store the variable symbol
  UINT c = symTable->cSymbols(TYPE_VAR);
  UINT u = cVariables++;
  SymVar *psym = new SymVar(psz, c, u);
  symTable->Store(psym);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareArray
// Scope:	 Private
// Description:	 This function attempts to add an array to the symbol table.
//		 The name of the array is stored in the ptok1->data.sz field.
//		 The array dimensions are stored in the math node structure
//		 pointed by the ptok2->data.p field.
void DeclareArray ( TOKEN *ptok1, TOKEN *ptok2, UINT &cVariables )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp DeclareArray\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  char *psz = ptok1->data.sz;
  MathNode *pmn = (MathNode *) ptok2->data.p;

  // report error if the symbol is already defined
  if (symTable->Retrieve(psz))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok1->iPtr);
    delete pmn;
    return;
  }

  // check the number of array dimensions
  UINT cDim = pmn->c;
  if (cDim > 3)
  { error->AddMessage(EHM_ARRAYDIM1, psz, ptok1->iPtr);
    error->AddMessage(EHM_NOTDEC1, "array", psz, ptok1->iPtr);
    delete pmn;
    return;
  }

  // check if each dimension expression is defined
  for(i = 0; i < cDim; ++i)
    if (pmn->apmn[i] == NULL)
      break;

  if (i < cDim)
  { error->AddMessage(EHM_NOTDEC1, "array", psz, ptok1->iPtr);
    delete pmn;
    return;
  }

  // check if each dimension expression evaluated to a positive integer
  BOOL fError = FALSE;
  for(i = 0; i < cDim; ++i)
  { if (pmn->apmn[i]->uNode != TYPE_ICONST)
    { sprintf(pszGlobalBuf, "%d", i + 1);
      error->AddMessage(EHM_ARRAYDIM2, pszGlobalBuf, psz, pmn->apmn[i]->iPtrExpr);
      fError = TRUE;
    }
    else if (pmn->apmn[i]->data.n < 1)
    { sprintf(pszGlobalBuf, "%d", i + 1);
      error->AddMessage(EHM_ARRAYDIM4, pszGlobalBuf, psz, pmn->apmn[i]->iPtrExpr);
      fError = TRUE;
    }
  }
  if (fError)
  { error->AddMessage(EHM_NOTDEC1, "array", psz, ptok1->iPtr);
    delete pmn;
    return;
  }

  // compute the total number of array elements
  unsigned long ul = 1;
  UINT ac[3] = {0, 0, 0};
  for(i = 0; i < cDim; ++i)
  { ac[i] = (UINT) pmn->apmn[i]->data.n;
    ul *= ac[i];
#if defined(P16BIT)
    if (ul > MAX_VARIABLES) { ul = MAX_VARIABLES + 1; break; }
#endif
  }

#if defined(P16BIT)
  // report an error if the number of variables exceeds the maximum
  if (cVariables + (UINT) ul > MAX_VARIABLES)
  { error->AddMessage(EHM_MAXVAR, psz, ptok1->iPtr);
    delete pmn;
    return;
  }
#endif

  // construct and store the array symbol
  UINT c = symTable->cSymbols(TYPE_ARRAY);
  UINT u = cVariables;
  cVariables += (UINT) ul;
  SymArray *psym = new SymArray(psz, c, cDim, ac[0], ac[1], ac[2], u);
  symTable->Store(psym);
  delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareConstant
// Scope:	 Private
// Description:	 This function attempts to add a constant to the symbol table.
//		 The name of the constant is stored in the ptok1->data.sz
//		 field.	 The constant's value is stored in the math node
//		 structure pointed by the ptok2->data.p field.
void DeclareConstant ( TOKEN *ptok1, TOKEN *ptok2 )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp DeclareConstant\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok1->data.sz;
  MathNode *pmn = (MathNode *) ptok2->data.p;

  // report an error if the symbol is already defined
  if (symTable->Retrieve(psz))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok1->iPtr);
    if (pmn) delete pmn;
    return;
  }

  // report a warning if the constant expression is not present
  if (pmn == NULL)
  { error->AddMessage(EHM_NOTDEC1, "constant", psz, ptok1->iPtr);
    return;
  }

  // report an error is the constant expression is not constant
  if (pmn->uExpr & FLAG_VAR)
  { error->AddMessage(EHM_CONST, psz, pmn->iPtrExpr);
    error->AddMessage(EHM_NOTDEC1, "constant", psz, ptok1->iPtr);
    delete pmn;
    return;
  }

  // construct and store the constant symbol
  if (pmn->uExpr & FLAG_REAL)
  { SymRConst *psymR = new SymRConst(psz, pmn->data.r);
    symTable->Store(psymR);
  }
  else
  { SymIConst *psymI = new SymIConst(psz, pmn->data.n);
    symTable->Store(psymI);
  }
  delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnIdentifier
// Scope:	 Private
// Description:	 This function makes either a variable or constant mathnode
//		 depending upon the symbol's type.  The name of the symbol
//		 is stored in the ptok->data.sz buffer.
MathNode * pmnIdentifier ( TOKEN *ptok, BOOL fFuncBody )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnIdentifier\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok->data.sz;
  HashNode *psym = symTable->Retrieve(psz);

  // report an error if the symbol is not defined
  if (psym == NULL)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok->iPtr);
    return NULL;
  }

  // if the identifier is part of a user-defined function body, then
  // global variables and array elements are not allowed
  if (fFuncBody && (psym->uType == TYPE_VAR))
  { error->AddMessage(EHM_SYMNOTALLOWED, psz, ptok->iPtr);
    return NULL;
  }

  // construct and return the appropriate math node
  SymIndex    *psymI;
  SymVar      *psymV;
  SymArg      *psymA;
  SymRConst   *psymRC;
  SymIConst   *psymIC;
  SymLocalVar *psymLV;
  SymLabel    *psymL;
  MathNode    *pmn;
  switch (psym->uType)
  { case TYPE_INDEX:
      psymI = TYPESAFE_DOWNCAST(psym, SymIndex);
      pmn = new MathNode(psymI->i, ptok->iPtr);
      break;

    case TYPE_VAR:
      psymV = TYPESAFE_DOWNCAST(psym, SymVar);
      pmn = new MathNode(psymV->i, psymV->u, ptok->iPtr);
      break;

    case TYPE_RCONST:
    case TYPE_LRCONST:
      psymRC = TYPESAFE_DOWNCAST(psym, SymRConst);
      pmn = new MathNode(psymRC->r, ptok->iPtr);
      break;

    case TYPE_ICONST:
    case TYPE_LICONST:
      psymIC = TYPESAFE_DOWNCAST(psym, SymIConst);
      pmn = new MathNode(psymIC->n, ptok->iPtr);
      break;

    case TYPE_LVAR:
      psymLV = TYPESAFE_DOWNCAST(psym, SymLocalVar);
      pmn = new MathNode(psymLV->i, 0, ptok->iPtr, TRUE);
      break;

    case TYPE_ARG:
      psymA = TYPESAFE_DOWNCAST(psym, SymArg);
      pmn = new MathNode(psymA->i, psymA->f, ptok->iPtr);
      break;

    case TYPE_LABEL:
      psymL = TYPESAFE_DOWNCAST(psym, SymLabel);
      pmn = new MathNode((int) psymL->i, ptok->iPtr);
      break;

    default:
      error->AddMessage(EHM_SYMEXP, psz,
	"index, argument, constant, label or variable", ptok->iPtr);
      pmn = NULL;
      break;
  }
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnArray
// Scope:	 Private
// Description:	 This function makes an array math node expression.  The name
//		 of the symbol is stored in the ptok->data.sz field. The array
//		 dimensions are stored in the math node expression pointed by
//		 the ptok2->data.p field.
MathNode * pmnArray ( TOKEN *ptok1, TOKEN *ptok2, BOOL fFuncBody )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnArray\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  char *psz = ptok1->data.sz;
  MathNode *pmn = (MathNode *) ptok2->data.p;
  HashNode *psym = symTable->Retrieve(psz);

  // report an error if the symbol is not defined
  if (psym == NULL)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok1->iPtr);
    delete pmn;
    return NULL;
  }

  // report an error if the symbol is not an array
  if (psym->uType != TYPE_ARRAY)
  { error->AddMessage(EHM_SYMEXP, psz, "array", ptok1->iPtr);
    delete pmn;
    return NULL;
  }

  // determine if an error has occurred in the specification of the array
  for(i = 0; i < pmn->c; ++i)
    if (pmn->apmn[i] == NULL)
    { delete pmn;
      return NULL;
    }

  // report an error if the array has the wrong number of dimensions
  SymArray *psymArray = TYPESAFE_DOWNCAST(psym, SymArray);
  if (psymArray->c != pmn->c)
  { error->AddMessage(EHM_ARRAYDIM6, psz, ptok1->iPtr);
    delete pmn;
    return NULL;
  }

  // check each array dimensions
  BOOL fError = FALSE;
  for(i = 0; i < pmn->c; ++i)
  { if (pmn->apmn[i]->uExpr & FLAG_REAL)
    { sprintf(pszGlobalBuf, "%d", i + 1);
      error->AddMessage(EHM_ARRAYDIM3, pszGlobalBuf, psz, pmn->apmn[i]->iPtrNode);
      fError = TRUE;
    }
    if (pmn->apmn[i]->uNode != TYPE_ICONST) continue;
    int n = pmn->apmn[i]->data.n;
    if ((n < 0) || ((UINT) n > psymArray->ac[i]))
    { sprintf(pszGlobalBuf, "%d", i + 1);
      error->AddMessage(EHM_ARRAYDIM5, pszGlobalBuf, psz, pmn->apmn[i]->iPtrNode);
      fError = TRUE;
    }
  }

  // if the identifier is part of a user-defined function body, then
  // global variables and array elements are not allowed
  if (fFuncBody)
  { error->AddMessage(EHM_SYMNOTALLOWED, psz, ptok1->iPtr);
    fError = TRUE;
  }

  if (fError)
  { delete pmn;
    return NULL;
  }

  // everything seems o.k.
  pmn->data.array.i  = psymArray->i;
  pmn->data.array.uA = psymArray->u;
  pmn->data.array.iT = 0;
  pmn->data.array.uE = -1u;
  pmn->data.array.ac[0] = psymArray->ac[0];
  pmn->data.array.ac[1] = psymArray->ac[1];
  pmn->data.array.ac[2] = psymArray->ac[2];
  pmn->data.array.pmnNSUM = NULL;
  pmn->data.array.pmnLSUM = NULL;
  pmn->uNode = TYPE_ARRAY;
  pmn->uExpr = FLAG_VAR | FLAG_REAL;
  pmn->iPtrNode = pmn->iPtrExpr = ptok1->iPtr;
  return (pmn);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnFunction
// Scope:	 Private
// Description:	 This function makes a function math node expression.  The name
//		 of the symbol is stored in the ptok1->data.sz field.  The func-
//		 tion arguments are stored in the math node expression pointed
//		 by the ptok2->data.p field.
MathNode * pmnFunction ( TOKEN *ptok1, TOKEN *ptok2 )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnFunction\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  char *psz = ptok1->data.sz;
  MathNode *pmn = (MathNode *) ptok2->data.p;
  HashNode *psym = symTable->Retrieve(psz);

  // report an error if the symbol is not defined
  if (psym == NULL)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok1->iPtr);
    delete pmn;
    return NULL;
  }

  // report an error if the symbol is not a function
  if ((psym->uType != TYPE_FUNC) && (psym->uType != TYPE_UDFUNC))
  { error->AddMessage(EHM_SYMEXP, psz, "function", ptok1->iPtr);
    delete pmn;
    return NULL;
  }

  // determine if an error has occurred in the specification of the function
  for(i = 0; i < pmn->c; ++i)
    if (pmn->apmn[i] == NULL)
    { delete pmn;
      return NULL;
    }

  // typecast the symbol and determine the number of arguments
  UINT c;
  SymFunc *psymFunc;
  SymUDFunc *psymUDFunc;
  if (psym->uType == TYPE_FUNC)
  { psymFunc   = TYPESAFE_DOWNCAST(psym, SymFunc);
    c = psymFunc->c;
  }
  else
  { psymUDFunc = TYPESAFE_DOWNCAST(psym, SymUDFunc);
    c = psymUDFunc->c;

    // report an error if the user-defined function is called from its body
    if (!psymUDFunc->f)
    { error->AddMessage(EHM_RECURSION, psymUDFunc->szName, ptok1->iPtr);
      delete pmn;
      return NULL;
    }
  }

  // report an error if the array has the wrong number of arguments
  if (c != pmn->c)
  { error->AddMessage(EHM_NUMARGS, psz, ptok1->iPtr);
    delete pmn;
    return NULL;
  }

  // everything seems o.k. ... convert arguments to reals if necessary
  for(i = 0; i < c; ++i)
    if (pmn->apmn[i]->uExpr & FLAG_INT)
      pmn->apmn[i] = new MathNode(TYPE_REAL, pmn->apmn[i], pmn->apmn[i]->iPtrNode);

  if (psym->uType == TYPE_FUNC)
  { pmn->uNode = TYPE_FUNC;
    pmn->data.func.iFunc   = psymFunc->i;
    pmn->data.func.iPD	   = 0;
    pmn->data.func.aprf[0] = psymFunc->aprf[0];
    pmn->data.func.aprf[1] = psymFunc->aprf[1];
    pmn->data.func.aprf[2] = psymFunc->aprf[2];
  }
  else
  { pmn->uNode = TYPE_UDFUNC;
    pmn->data.udfunc.iFunc = psymUDFunc->i;
    pmn->data.udfunc.iPD   = 0;
    pmn->data.udfunc.af	   = psymUDFunc->af;
  }

  BOOL fConst	= TRUE;
  BOOL fNodeSum = FALSE;
  for(i = 0; i < c; ++i)
  { UINT uExpr = pmn->apmn[i]->uExpr;
    if (uExpr & FLAG_VAR)  fConst   = FALSE;
    if (uExpr & FLAG_NSUM) fNodeSum = TRUE;
  }
  pmn->uExpr = ((fConst) ? FLAG_CONST : FLAG_VAR) | FLAG_REAL;
  if (fNodeSum) pmn->uExpr |= FLAG_NSUM;
  pmn->iPtrNode = pmn->iPtrExpr = ptok1->iPtr;
  pmn->SimplifyFunc();
  return (pmn);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnUnaryOperator1
// Scope:	 Private
// Description:	 This function makes a unary operator.	The operator's type
//		 is stored in the ptok1->uTokenType field and a pointer to
//		 the operand is stored in the ptok1->data.p field.
MathNode * pmnUnaryOperator1 (TOKEN *ptok1, TOKEN *ptok2)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnUnaryOperator1\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT uOp = ptok1->uTokenType;
  MathNode *pmn1 = (MathNode *) ptok2->data.p;

  // if the operand is not defined, return NULL
  if (pmn1 == NULL) return NULL;

  // form the unary operator expression
  MathNode *pmn;
  switch (uOp)
  { case '+':	   pmn = pmn1;					      break;
    case '-':	   pmn = new MathNode(TYPE_NEG,	 pmn1, ptok1->iPtr);  break;
    case TOK_NOT:  pmn = new MathNode(TYPE_NOT,	 pmn1, ptok1->iPtr);  break;
    case TOK_REAL: pmn = new MathNode(TYPE_REAL, pmn1, ptok1->iPtr);  break;
  }
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnUnaryOperator2
// Scope:	 Private
// Description:	 This function makes a unary operator.	The operator's type
//		 is stored in the ptok1->uTokenType field and a pointer to
//		 the operand is stored in the ptok1->data.p field.
MathNode * pmnUnaryOperator2 (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnUnaryOperator2\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT uOp = ptok1->uTokenType;
  MathNode *pmn1 = (MathNode *) ptok2->data.p;

  // report an error if the triangular mesh has not been declared
  if (!aauNeighbors)
  { error->AddMessage(EHM_TRIMESH, ptok1->iPtr);
    if (pmn1) delete pmn1;
    return NULL;
  }

  // if the operand is not defined, return NULL
  if (pmn1 == NULL) return NULL;

  // if the operand is not an integer expression, report an error and
  // return NULL
  if (pmn1->uExpr & FLAG_REAL)
  { error->AddMessage(EHM_INTEXPR, "operand", ptok2->iPtr);
    delete pmn1;
    return NULL;
  }

  // if an edge label is specified report an error
  UINT iLabel = (UINT) NO_REGION_LABEL;
  BOOL fError = FALSE;
  if (ptok3)
  { if (ptok3->uTokenType == TOK_ALL) iLabel = -1u;
    else
    { HashNode *psym = symTable->Retrieve(ptok3->data.sz);
      if (psym->uType != TYPE_LABEL) fError = TRUE;
      else
      { SymLabel *psymL = TYPESAFE_DOWNCAST(psym, SymLabel);
	iLabel = psymL->i;
	if (iLabel >= cRegLabel) fError = TRUE;
      }
    }
  }

  // report the label error
  if (fError)
  { error->AddMessage(EHM_SYMEXP, ptok3->data.sz, "region label", ptok3->iPtr);
    delete pmn1;
    return NULL;
  }

  // form the unary operator expression
  MathNode *pmn;
  switch (uOp)
  { case TOK_AREA:  pmn = new MathNode(TYPE_AREA, pmn1, ptok1->iPtr);  break;
  }
  pmn->data.n = (int) iLabel;
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnBinaryOperator1
// Scope:	 Private
// Description:	 This function makes a binnary operator.  The operator's type
//		 is stored in the ptok1->uTokenType field and pointers to the
//		 operands are stored in the ptok2->data.p and ptok3->data.p
//		 fields.
MathNode * pmnBinaryOperator1 (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnBinaryOperator1\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT uOp = ptok1->uTokenType;
  MathNode *pmn1 = (MathNode *) ptok2->data.p;
  MathNode *pmn2 = (MathNode *) ptok3->data.p;

  // if the operand is not defined, return NULL
  if ((pmn1 == NULL) || (pmn2 == NULL))
  { if (pmn1) delete pmn1;
    if (pmn2) delete pmn2;
    return NULL;
  }

  // form the binary operator expression
  UINT uType;
  switch (uOp)
  { case '+':	   uType = TYPE_ADD;  break;
    case '-':	   uType = TYPE_SUB;  break;
    case '*':	   uType = TYPE_MUL;  break;
    case '/':	   uType = TYPE_DIV;  break;
    case '>':	   uType = TYPE_GT;   break;
    case '<':	   uType = TYPE_LT;   break;
    case TOK_AND:  uType = TYPE_AND;  break;
    case TOK_OR:   uType = TYPE_OR;   break;
    case TOK_XOR:  uType = TYPE_XOR;  break;
    case TOK_GE:   uType = TYPE_GE;   break;
    case TOK_LE:   uType = TYPE_LE;   break;
    case TOK_EQ:   uType = TYPE_EQ;   break;
    case TOK_NE:   uType = TYPE_NE;   break;
  }
  MathNode *pmn = new MathNode(uType, pmn1, pmn2, ptok1->iPtr);
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnBinaryOperator2
// Scope:	 Private
// Description:	 This function makes a binnary operator.  The operator's type
//		 is stored in the ptok1->uTokenType field and pointers to the
//		 operands are stored in the ptok2->data.p and ptok3->data.p
//		 fields.
MathNode * pmnBinaryOperator2 (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3, TOKEN *ptok4)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnBinaryOperator2\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT uOp = ptok1->uTokenType;
  MathNode *pmn1 = (MathNode *) ptok2->data.p;
  MathNode *pmn2 = (MathNode *) ptok3->data.p;

  // report an error if the triangular mesh has not been declared
  if (!aauNeighbors)
  { error->AddMessage(EHM_TRIMESH, ptok1->iPtr);
    if (pmn1) delete pmn1;
    if (pmn2) delete pmn2;
    return NULL;
  }

  // if the operand is not defined, return NULL
  if ((pmn1 == NULL) || (pmn2 == NULL))
  { if (pmn1) delete pmn1;
    if (pmn2) delete pmn2;
    return NULL;
  }

  // if either operand is not an integer expression, report an error and
  // return NULL
  BOOL fReal1 = pmn1->uExpr & FLAG_REAL;
  BOOL fReal2 = pmn2->uExpr & FLAG_REAL;
  if (fReal1 || fReal2)
  { if (fReal1) error->AddMessage(EHM_INTEXPR, "operand", ptok2->iPtr);
    if (fReal2) error->AddMessage(EHM_INTEXPR, "operand", ptok3->iPtr);
    delete pmn1;
    delete pmn2;
    return NULL;
  }

  // if an edge label is specified report an error
  UINT iLabel = (UINT) NO_REGION_LABEL;
  BOOL fError = FALSE;
  if (ptok4)
  { if (ptok4->uTokenType == TOK_ALL) iLabel = -1u;
    else
    { HashNode *psym = symTable->Retrieve(ptok4->data.sz);
      if (psym->uType != TYPE_LABEL) fError = TRUE;
      else
      { SymLabel *psymL = TYPESAFE_DOWNCAST(psym, SymLabel);
	iLabel = psymL->i;
	if (iLabel >= cRegLabel) fError = TRUE;
      }
    }
  }

  // report the label error
  if (fError)
  { error->AddMessage(EHM_SYMEXP, ptok4->data.sz, "region label", ptok3->iPtr);
    delete pmn1;
    delete pmn2;
    return NULL;
  }

  // form the binary operator expression
  UINT uType;
  switch (uOp)
  { case TOK_AREA1: uType = TYPE_AREA1;	 break;
    case TOK_AREA2: uType = TYPE_AREA2;	 break;
    case TOK_NODE:  uType = TYPE_NODE;	 break;
    case TOK_EDGE:  uType = TYPE_EDGE;	 break;
    case TOK_ELEN:  uType = TYPE_ELEN;	 break;
    case TOK_ILEN:  uType = TYPE_ILEN;	 break;
  }
  MathNode *pmn = new MathNode(uType, pmn1, pmn2, ptok1->iPtr);
  pmn->data.n = (int) iLabel;
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnNodeSum
// Scope:	 Private
// Description:	 This function makes a node summation.	ptok1 is the nsum
//		 token, ptok2 is an integer expression which evaluates to
//		 a node index, ptok3 is the summation index, ptok4 is the
//		 summation label and ptok5 is the summation expression.
MathNode * pmnNodeSum (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3, TOKEN *ptok4, TOKEN *ptok5)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnNodeSum\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT iIndex	    = ptok3->data.u;
  char *pszLabel    = ptok4->data.sz;
  MathNode *pmnNode = (MathNode *) ptok2->data.p;
  MathNode *pmnExpr = (MathNode *) ptok5->data.p;

  // report an error if the triangular mesh has not been declared
  if (!aauNeighbors) error->AddMessage(EHM_TRIMESH, ptok1->iPtr);

  // note an error if the node expression or summation expression are not
  // present or the index is not defined
  BOOL fError = !pmnNode || !pmnExpr || (iIndex == INDEX_ERROR);

  // note an error if the summation label is not defined or is not a label
  UINT iLabel;
  if (ptok4->uTokenType == TOK_ALL) iLabel = -1u;
  else
  { HashNode *psym = symTable->Retrieve(pszLabel);
    if (!psym)
    { error->AddMessage(EHM_SYMNOTDEF, pszLabel, ptok4->iPtr);
      fError = TRUE;
    }
    else if (psym->uType != TYPE_LABEL)
    { error->AddMessage(EHM_SYMEXP, pszLabel, "label", ptok4->iPtr);
      fError = TRUE;
    }
    else
    { SymLabel *psymLabel = TYPESAFE_DOWNCAST(psym, SymLabel);
      iLabel = psymLabel->i;
    }
  }

  // note an error if the index expression is not an integer expression
  if (pmnNode && (pmnNode->uExpr & FLAG_REAL))
  { error->AddMessage(EHM_INTEXPR, "node index", ptok2->iPtr);
    fError = TRUE;
  }

  // note an error if an nsum expression is embedded in the summation
  // expression
  if (pmnExpr && (pmnExpr->uExpr & FLAG_NSUM))
  { error->AddMessage(EHM_SUMEMBED, "nsum", "nsum", ptok5->iPtr);
    fError = TRUE;
  }

  // note an error if an lsum expression is embedded in the summation
  // expression
  if (pmnExpr && (pmnExpr->uExpr & FLAG_LSUM))
  { error->AddMessage(EHM_SUMEMBED, "lsum", "nsum", ptok5->iPtr);
    fError = TRUE;
  }

  // report that the nsum statement is not defined if an error has occured
  if (fError)
  { error->AddMessage(EHM_NOTDEC2, "node summation (nsum)", ptok1->iPtr);
    if (pmnNode) delete pmnNode;
    if (pmnExpr) delete pmnExpr;
    return NULL;
  }

  // create and return the mathnode
  BOOL fTriSum = ptok1->uTokenType == TOK_TSUM;
  MathNode *pmn = new MathNode(pmnNode, iIndex, iLabel, pmnExpr, fTriSum, ptok1->iPtr);
  UINT iExprIndex = symTable->cSymbols(TYPE_NSUM);
  pmn->data.nsum.iExpr = iExprIndex;
  sprintf(pszGlobalBuf, "__NS%u__", iExprIndex);
  symTable->Store(new SymNodeSum(pszGlobalBuf, iExprIndex, pmnExpr));
  symTable->RemoveLast(TYPE_INDEX);
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnLabelSum
// Scope:	 Private
// Description:	 This function makes a label summation.	 ptok1 is the lsum
//		 token, ptok2 is the summation index, ptok4 is the
//		 summation label and ptok4 is the summation expression.
MathNode * pmnLabelSum (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3, TOKEN *ptok4)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnLabelSum\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT iIndex	    = ptok2->data.u;
  char *pszLabel    = ptok3->data.sz;
  MathNode *pmnExpr = (MathNode *) ptok4->data.p;

  // report an error if the triangular mesh has not been declared
  if (!aauNeighbors) error->AddMessage(EHM_TRIMESH, ptok1->iPtr);

  // note an error if the summation expression is not present or the index
  // is not defined
  BOOL fError = !pmnExpr || (iIndex == INDEX_ERROR);

  // note an error if the summation label is not defined or is not a label
  UINT iLabel;
  HashNode *psym = symTable->Retrieve(pszLabel);
  if (!psym)
  { error->AddMessage(EHM_SYMNOTDEF, pszLabel, ptok4->iPtr);
    fError = TRUE;
  }
  else if (psym->uType != TYPE_LABEL)
  { error->AddMessage(EHM_SYMEXP, pszLabel, "label", ptok3->iPtr);
    fError = TRUE;
  }
  else
  { SymLabel *psymLabel = TYPESAFE_DOWNCAST(psym, SymLabel);
    iLabel = psymLabel->i;
  }

  // note an error if an lsum expression is embedded in the summation
  // expression
  if (pmnExpr && (pmnExpr->uExpr & FLAG_LSUM))
  { error->AddMessage(EHM_SUMEMBED, "lsum", "lsum", ptok4->iPtr);
    fError = TRUE;
  }

  // report that the lsum statement is not defined if an error has occured
  if (fError)
  { error->AddMessage(EHM_NOTDEC2, "label summation (lsum)", ptok1->iPtr);
    if (pmnExpr) delete pmnExpr;
    return NULL;
  }

  // create and return the mathnode
  MathNode *pmn = new MathNode(iIndex, iLabel, pmnExpr, ptok1->iPtr);
  UINT iExprIndex = symTable->cSymbols(TYPE_LSUM);
  pmn->data.lsum.iExpr = iExprIndex;
  sprintf(pszGlobalBuf, "__LS%u__", iExprIndex);
  symTable->Store(new SymLabelSum(pszGlobalBuf, iExprIndex, pmnExpr));
  symTable->RemoveLast(TYPE_INDEX);
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pindexMake
// Scope:	 Private
// Description:	 This function makes an instance of the index class.
Index * pindexMake(UINT c, TOKEN *atoken)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pindexMake\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  int  n[3];
  UINT i;
  UINT iIndex = atoken->data.u;
  BOOL fError = iIndex == INDEX_ERROR;

  if (!fError)
    for(i = 1; i < c; ++i)
    { MathNode *pmn = (MathNode *) atoken[i].data.p;

      // if index value expression is not defined, then set the error flag
      if (pmn == NULL) fError = TRUE;

      // if the index value expression does not evaluate to an integer, then
      // report the error and set the error flag.
      else if (pmn->uNode != TYPE_ICONST)
      { strcpy(pszGlobalBuf, (i == 1) ? "start" : ((i == 2) ? "end" : "step"));
	char *psz = symTable->Retrieve(TYPE_INDEX, iIndex)->szName;
	error->AddMessage(EHM_INDEX1, pszGlobalBuf, psz, atoken[i].iPtr);
	fError = TRUE;
      }

      // index value expression is o.k.
      else n[i - 1] = pmn->data.n;
    }

    // if there an error was not detected, construct the index
    Index *pindex = NULL;
    if (!fError)
    { switch (c)
      { case 1:	 pindex = new Index(iIndex);			break;
	case 2:	 pindex = new Index(iIndex, n[0]);		break;
	case 3:	 pindex = new Index(iIndex, n[0], n[1]);	break;
	case 4:	 pindex = new Index(iIndex, n[0], n[1], n[2]);	break;
      }
      pindex->SetBufPtr(atoken->iPtr);
    }

    // deallocate the math node expressions
    for (i = 1; i < c; ++i)
      if (atoken[i].data.p) delete (MathNode *) atoken[i].data.p;

    return pindex;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pindexLabelMake
// Scope:	 Private
// Description:	 This function makes an instance of the index class.
Index * pindexLabelMake(TOKEN *ptok1, TOKEN *ptok2)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pindexLabelMake\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  Index *pindex = NULL;
  UINT iIndex = ptok1->data.u;
  if (iIndex != INDEX_ERROR)
  { SymLabel *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(ptok2->data.sz), SymLabel);
    pindex = new Index(iIndex, psym->i, psym->c, (int *) psym->au);
    pindex->SetBufPtr(ptok1->iPtr);
  }
  return pindex;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 fValidateIndices
// Scope:	 Private
// Description:	 This function validates the passed instance of the indices
//		 class.	 If an error is found, TRUE is returned; otherwise,
//		 FALSE is returned.
BOOL fValidateIndices (TOKEN *ptok, SymArray *psym, Indices *pindices)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp fValidateIndices\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // check the number of dimensions
  UINT c = psym->c;
  if (c != pindices->GetDim())
  { error->AddMessage(EHM_ARRAYDIM6, ptok->data.sz, ptok->iPtr);
    return TRUE;
  }

  // check if all of the indices are defined
  for(i = 0; i < c; ++i)
    if (!pindices->IsDefined(i))
      return TRUE;


  // initialze the array fields of the indices instance
  BOOL fError = FALSE;
  UINT iPtr   = ptok->iPtr;
  pindices->Array(psym->i, psym->u, psym->ac);
  pindices->SetBufPtr(iPtr);

  // check each dimensions
  for(i = 0; i < c; ++i)
  { // if the ALL flag was specified in the index definition, then set the
    // index's values so that the index spans the entire range
    if (pindices->IsAll(i))
    { pindices->SetAll(i);
      continue;
    }

    // check the begin value
    if (!pindices->BeginOK(i))
    { char *psz = symTable->Retrieve(TYPE_INDEX, i)->szName;
      error->AddMessage(EHM_INDEX2, "begin", psz, pindices->GetBufPtr(i));
      fError = TRUE;
    }

    // check the end value
    if (!pindices->EndOK(i))
    { char *psz = symTable->Retrieve(TYPE_INDEX, i)->szName;
      error->AddMessage(EHM_INDEX2, "end", psz, pindices->GetBufPtr(i));
      fError = TRUE;
    }

    // check the step value
    if (!pindices->StepOK(i))
    { char *psz = symTable->Retrieve(TYPE_INDEX, i)->szName;
      error->AddMessage(EHM_INDEX2, "step", psz, pindices->GetBufPtr(i));
      fError = TRUE;
    }
  }

  // return the result
  return(fError);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ValidateVarOrArray
// Scope:	 Private
// Description:	 This function validates the symbol ptok1->data.sz.  If the
//		 symbol is valid, then a pointer to it is returned; otherwise,
//		 NULL is returned.
HashNode * ValidateVarOrArray (TOKEN *ptok1, TOKEN *ptok2)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ValidateVarOrArray\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok1->data.sz;
  HashNode *psym = symTable->Retrieve(psz);
  BOOL fError = FALSE;

  // report error if the symbol is undefined
  if (psym == NULL)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok1->iPtr);
    fError = TRUE;
  }

  if (!fError)
  { // symbol should be an array
    if (ptok2)
    { if (psym->uType != TYPE_ARRAY)
      { error->AddMessage(EHM_SYMEXP, psz, "array", ptok1->iPtr);
	fError = TRUE;
      }
      else
      { SymArray *psymArray = TYPESAFE_DOWNCAST(psym, SymArray);
	Indices	 *pindices  = (Indices *) ptok2->data.p;
	fError = fValidateIndices(ptok1, psymArray, pindices);
      }
    }

    // symbol should be a variable
    else
    { if (psym->uType != TYPE_VAR)
      { error->AddMessage(EHM_SYMEXP, psz, "variable", ptok1->iPtr);
	fError = TRUE;
      }
    }
  }

  // if an error has occurred, then return NULL; otherwise, return the
  // symbol
  return (fError) ? NULL : psym;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessKnown
// Scope:	 Private
// Description:	 This function processes the known and unknown statements
void ProcessKnown (TOKEN *ptok1, TOKEN *ptok2, BOOL fKnown)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessKnown\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  pstatements->AllocBitFlags();
  HashNode *psym = ValidateVarOrArray(ptok1, ptok2);

  // set/clear the appropriate bit flags
  if (psym)
  { if (ptok2)
    { Indices *pindices = (Indices *) ptok2->data.p;
      pindices->Init();
      while (!pindices->IsLoopOver())
      { UINT u = pindices->Element();
	if (fKnown) pstatements->ClearBitFlag(u);
	else	    pstatements->SetBitFlag(u);
	pindices->Next();
      }
    }
    else
    { UINT u = TYPESAFE_DOWNCAST(psym, SymVar)->u;
      if (fKnown) pstatements->ClearBitFlag(u);
      else	  pstatements->SetBitFlag(u);
    }
  }

  // error ... report warning
  else error->AddMessage(EHM_NOTDEC2, (fKnown) ? "known" : "unknown", ptok1->iPtr);

  // deallocate the indices structure
  if (ptok2) delete (Indices *) ptok2->data.p;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessEquation
// Scope:	 Private
// Description:	 This function processes the equation statement.
void ProcessEquation (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3, TOKEN *ptok4)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessEquation\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  HashNode *psym = ValidateVarOrArray(ptok1, ptok2);
  Indices  *pindices = (ptok2) ? (Indices *) ptok2->data.p : NULL;
  MathNode *pmnL = (MathNode *) ptok3->data.p;
  MathNode *pmnR = (MathNode *) ptok4->data.p;
  BOOL fError = (psym) ? FALSE : TRUE;

  // check the left and right parts of the equations
  MathNode *pmn;
  if (!fError)
  { if (pmnL && pmnR) pmn = new MathNode(TYPE_SUB, pmnL, pmnR, -1u);
    else fError = TRUE;
  }

  // deallocate and report equation not declared warning
  if (fError)
  { error->AddMessage(EHM_NOTDEC2, "equation", ptok1->iPtr);
    if (pindices) delete pindices;
    if (pmnL)	  delete pmnL;
    if (pmnR)	  delete pmnR;
  }

  // store equation statement
  else
  { if (pindices)
    { SymArray *psymA = TYPESAFE_DOWNCAST(psym, SymArray);
      ArrayEquation *pae = new ArrayEquation(psymA->i, psymA->u, pindices, pmn, ptok1->iPtr);
      pstatements->AddEqu(pae);
      psymA->f = TRUE;
    }
    else
    { SymVar *psymV = TYPESAFE_DOWNCAST(psym, SymVar);
      VarEquation *pve = new VarEquation(psymV->i, psymV->u, pmn, ptok1->iPtr);
      pstatements->AddEqu(pve);
      psymV->f = TRUE;
    }
  }

  // remove the index symbols from the symbol table
  symTable->RemoveSymbols(TYPE_INDEX);
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessAssignment
// Scope:	 Private
// Description:	 This function processes the assignment statement.
void ProcessAssignment (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessAssignment\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  HashNode *psym = ValidateVarOrArray(ptok1, ptok2);
  Indices  *pindices = (ptok2) ? (Indices *) ptok2->data.p : NULL;
  MathNode *pmn = (MathNode *) ptok3->data.p;
  BOOL fError = (psym) ? FALSE : TRUE;

  // check the equation
  if (pmn == NULL) fError = TRUE;

  // deallocate and report equation not declared warning
  if (fError)
  { error->AddMessage(EHM_NOTDEC2, "assignment", ptok1->iPtr);
    if (pindices) delete pindices;
    if (pmn)	  delete pmn;
  }

  // store equation statement
  else
  { if (pindices)
    { SymArray *psymA = TYPESAFE_DOWNCAST(psym, SymArray);
      ArrayAssign *paa = new ArrayAssign(psymA->i, psymA->u, pindices, pmn, ptok1->iPtr);
      pstatements->AddMain(paa);
    }
    else
    { SymVar *psymV = TYPESAFE_DOWNCAST(psym, SymVar);
      VarAssign *pva = new VarAssign(psymV->i, psymV->u, pmn, ptok1->iPtr);
      pstatements->AddMain(pva);
    }
  }

  // remove the index symbols from the symbol table
  symTable->RemoveSymbols(TYPE_INDEX);
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessSolve
// Scope:	 Private
// Description:	 This function processes the solve statement.
void ProcessSolve (TOKEN *ptok)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessSolve\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  Solve *psolve = new Solve(ptok->iPtr);
  pstatements->AddMain(psolve);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessWrite
// Scope:	 Private
// Description:	 This function processes the solve statement.
void ProcessWrite (TOKEN *ptok)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessWrite\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  Write *pwrite = new Write(ptok->iPtr);
  pstatements->AddMain(pwrite);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessWhile1
// Scope:	 Private
// Description:	 This function processes the while statement.
void ProcessWhile1 (TOKEN *ptok1, TOKEN *ptok2, TokenStack *stackToken)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessWhile1\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (ptok2->data.p)
  { MathNode *pmn = (MathNode *) ptok2->data.p;
    pmn = new MathNode(TYPE_NOT, pmn, ptok2->iPtr);
    Goto *pgoto = new Goto(pmn, -1u, ptok1->iPtr);
    pstatements->AddMain(pgoto);
  }
  else error->AddMessage(EHM_NOTDEC2, "while", ptok1->iPtr);

  TOKEN token;
  token.uTokenType = -1u;
  token.iPtr = -1u;
  token.data.u = (ptok1->data.p) ? pstatements->GetNumOfStatements() - 1 : -1u;
  stackToken->push(token);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessWhile2
// Scope:	 Private
// Description:	 This function processes the while statement.
void ProcessWhile2 (TOKEN *ptok)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessWhile2\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT iStatement = ptok->data.u;
  if (iStatement != -1u)
  { Goto *pgoto = new Goto(NULL, iStatement, -1u);
    pstatements->AddMain(pgoto);
    UINT iIndex = pstatements->GetNumOfStatements();
    pstatements->SetGotoIndex(iStatement, iIndex);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessDo1
// Scope:	 Private
// Description:	 This function processes the do statement.
void ProcessDo1 (TokenStack *stackToken)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessDo1\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  TOKEN token;
  token.uTokenType = -1u;
  token.iPtr = -1u;
  token.data.u = pstatements->GetNumOfStatements();
  stackToken->push(token);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessDo2
// Scope:	 Private
// Description:	 This function processes the do statement.
void ProcessDo2 (TOKEN *ptok1, TOKEN *ptok2)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessDo2\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (ptok2->data.p)
  { MathNode *pmn = (MathNode *) ptok2->data.p;
    Goto *pgoto = new Goto(pmn, ptok1->data.u, ptok2->iPtr);
    pstatements->AddMain(pgoto);
  }
  else error->AddMessage(EHM_NOTDEC2, "do", ptok2->iPtr);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessIf1
// Scope:	 Private
// Description:	 This function processes the if statement.
void ProcessIf1 (TOKEN *ptok1, TOKEN *ptok2, TokenStack *stackToken)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessIf1\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (ptok2->data.p)
  { MathNode *pmn = (MathNode *) ptok2->data.p;
    pmn = new MathNode(TYPE_NOT, pmn, ptok2->iPtr);
    Goto *pgoto = new Goto(pmn, -1u, ptok1->iPtr);
    pstatements->AddMain(pgoto);
  }
  else error->AddMessage(EHM_NOTDEC2, "if", ptok1->iPtr);

  TOKEN token;
  token.uTokenType = -1u;
  token.iPtr = -1u;
  token.data.u = (ptok1->data.p) ? pstatements->GetNumOfStatements() - 1 : -1u;
  stackToken->push(token);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessIf2
// Scope:	 Private
// Description:	 This function processes the if statement.
void ProcessIf2 (TOKEN *ptok, BOOL fElse = FALSE, TokenStack *stackToken = NULL)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessIf2\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // if an error occurred, exit from this routine
  UINT iStatement = ptok->data.u;
  if (iStatement == -1u)
  { if (fElse) stackToken->push(*ptok);
    return;
  }

  // handle the if-else statement
  UINT cStatements = pstatements->GetNumOfStatements();
  if (fElse)
  { Goto *pgoto = new Goto(NULL, -1u, -1u);
    pstatements->AddMain(pgoto);
    ptok->data.u = cStatements++;
    stackToken->push(*ptok);
  }

  // set the index of the conditional goto statement
  pstatements->SetGotoIndex(iStatement, cStatements);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessSet
// Scope:	 Private
// Description:	 This function processes the set statement.
void ProcessSet(TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessSet\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // unsigned integer and real number buffers
  MathNode *pmn;
  UINT u;
  REAL r;

  switch (ptok3->uTokenType)
  { case TOK_INT:
      u = (UINT) ptok3->data.n;
      r = (REAL) u;
      break;

    case TOK_NUM:
      r = ptok3->data.r;
      break;

    case TOK_EXPR:
      pmn = (MathNode *) ptok3->data.p;
      if (pmn == NULL) return;
      if (pmn->uExpr & FLAG_VAR)
      { error->AddMessage(EHM_CONST, "DISTANCE SCALE", pmn->iPtrExpr);
	delete pmn;
	return;
      }
      r = (pmn->uExpr & FLAG_REAL) ? pmn->data.r : (REAL) pmn->data.n;
      break;
  }

  switch (ptok1->uTokenType)
  { // internal variable for the Newton algorithm
    case TOK_NEWTON:
      switch (ptok2->uTokenType)
      { case TOK_ITERS:
	  pNumAlgXferBuf->SetNewtonIterations(u);
	  break;

	case TOK_ACC:
	  pNumAlgXferBuf->SetNewtonAccuracy(r);
	  break;

	case TOK_DAMP:
	  pNumAlgXferBuf->SetNewtonDamping(u);
	  break;
      }
      break;


    // internal variable for the linear solution algorithm
    case TOK_LINSOL:
      switch (ptok2->uTokenType)
      { case TOK_ALGOR:
	  switch (ptok3->uTokenType)
	  { case TOK_SOR:    u = LINSOLVE_SOR;	break;
	    case TOK_PCG:    u = LINSOLVE_PCG;	break;
	    case TOK_GAUSS:  u = LINSOLVE_GE;	break;
	  }
	  pNumAlgXferBuf->SetLinearSolution(u);
	  break;

	case TOK_SORT:
	  switch (ptok3->uTokenType)
	  { case TOK_NONE:  u = SORT_NONE;  break;
	  }
	  pNumAlgXferBuf->SetLinearSort(u);
	  break;

	case TOK_FILL:
	  if ((ptok3->uTokenType == TOK_INFIN) || (u > 5)) u = FILL_INFINITY;
	  pNumAlgXferBuf->SetLinearFill(u);
	  break;

	case TOK_PREC:
	  switch (ptok3->uTokenType)
	  { case TOK_NONE:  u = PRECOND_NONE;  break;
	    case TOK_DIAG:  u = PRECOND_DIAG;  break;
	    case TOK_ILU:   u = PRECOND_ILU;   break;
	  }
	  pNumAlgXferBuf->SetLinearPrecond(u);
	  break;

	case TOK_ITERS:
	  pNumAlgXferBuf->SetLinearIterations(u);
	  break;

	case TOK_ACC:
	  pNumAlgXferBuf->SetLinearAccuracy(r);
	  break;

	case TOK_OVER:
	  if ((r < 1.0) || (r > 2.0))
	  { error->AddMessage(EHM_OVERRELAX, ptok3->iPtr);
	    if (r < 1.0) r = 1.0;
	    if (r > 2.0) r = 2.0;
	  }
	  pNumAlgXferBuf->SetLinearOver(r);
	  break;

	case TOK_ZERO:
	  u = (ptok3->uTokenType == TOK_YES) ? TRUE : FALSE;
	  pNumAlgXferBuf->SetLinearZero(u);
	  break;
      }
      break;

    // distance scale
    case TOK_DIST:
      pNumAlgXferBuf->SetDistScale(r);
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareArgument
// Scope:	 Private
// Description:	 This function validates and adds the user-defined function
//		 argument to the symbol table.
void DeclareArgument(TOKEN *ptok, BOOL fMayBeUnknown, BOOL &fFuncError)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp DeclareArgument\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok->data.sz;

  // since function arguments are of local scope, their names may override
  // global identifiers but not local identifiers.
  HashNode *phnode = symTable->Retrieve(psz);
  if (phnode && (phnode->uType == TYPE_ARG))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok->iPtr);
    fFuncError = TRUE;
    return;
  }

  // @@@ if we want to limit the number of function arguments, then we need
  // to add some more code after the next line
  UINT i = symTable->cSymbols(TYPE_ARG);

  // construct and store the argument symbol
  SymArg *psym = new SymArg(psz, i, fMayBeUnknown);
  symTable->Store(psym);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareLocalConstant
// Scope:	 Private
// Description:	 This function attempts to add a local constant to the symbol
//		 table.	 The name of the constant is stored in the
//		 ptok1->data.sz field.	The constant's value is stored in the
//		 math node structure pointed by the ptok2->data.p field.
void DeclareLocalConstant (TOKEN *ptok1, TOKEN *ptok2, BOOL &fFuncError)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp DeclareLocalConstant\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok1->data.sz;
  MathNode *pmn = (MathNode *) ptok2->data.p;

  // since local constants are of local scope, their names may override
  // global identifiers but not local identifiers.
  HashNode *phnode = symTable->Retrieve(psz);
  if (phnode)
  { UINT u = phnode->uType;
    if ((u == TYPE_ARG) || (u == TYPE_LVAR) || (u == TYPE_LRCONST) || (u == TYPE_LICONST))
    { error->AddMessage(EHM_SYMREDEF, psz, ptok1->iPtr);
      fFuncError = TRUE;
      if (pmn) delete pmn;
      return;
    }
  }

  // report a warning if the constant expression is not present
  if (pmn == NULL)
  { error->AddMessage(EHM_NOTDEC1, "local constant", psz, ptok1->iPtr);
    fFuncError = TRUE;
    return;
  }

  // report an error is the constant expression is not constant
  if (pmn->uExpr & FLAG_VAR)
  { error->AddMessage(EHM_CONST, psz, pmn->iPtrExpr);
    error->AddMessage(EHM_NOTDEC1, "local constant", psz, ptok1->iPtr);
    fFuncError = TRUE;
    delete pmn;
    return;
  }

  // construct and store the constant symbol
  if (pmn->uExpr & FLAG_REAL)
  { SymRConst *psymR = new SymRConst(psz, pmn->data.r);
    psymR->uType = TYPE_LRCONST;
    symTable->Store(psymR);
  }
  else
  { SymIConst *psymI = new SymIConst(psz, pmn->data.n);
    psymI->uType = TYPE_LICONST;
    symTable->Store(psymI);
  }
  delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareUserDefinedFunction
// Scope:	 Private
// Description:	 This function attempts to add a user-defined function to the
//		 symbol table.	The name of the function is stored in the
//		 ptok->data.sz field.
void DeclareUserDefinedFunction (TOKEN *ptok, BOOL &fFuncError, BOOL fExtFunc = FALSE)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp DeclareUserDefinedFunction\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  char *psz = ptok->data.sz;

  // report an error if the symbol is already defined
  if (symTable->Retrieve(psz))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok->iPtr);
    fFuncError = TRUE;
  }

  // determine the function index and the number of arguments
  UINT iFunc = symTable->cSymbols(TYPE_UDFUNC);
  UINT cArgs = symTable->cSymbols(TYPE_ARG);

  // allocate and initialize the may refer to unknown array
  BOOL *afMayBeUnknown = NULL;
  if (cArgs)
  { afMayBeUnknown = new BOOL[cArgs];
    for(i = 0; i < cArgs; ++i)
    { SymArg *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARG, i), SymArg);
      afMayBeUnknown[i] = psym->f;
    }
  }

  // allocate and store user-defined function in the symbol table
  char szError[] = "_error";
  if (fFuncError) psz = szError;
  SymUDFunc *psym = new SymUDFunc(psz, iFunc, cArgs, afMayBeUnknown, ptok->iPtr);
  symTable->Store(psym);

  if (fExtFunc)
  { psym->f = TRUE;
    fFuncError = FALSE;
    symTable->RemoveSymbols(TYPE_ARG);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 AddLocalVariable
// Scope:	 Private
// Description:	 This function attempts to add a local assignment statement to
//		 the user-defined function.  The name of the local variable is
//		 stored in the ptok1->data.sz field and the variables value is
//		 stored in the mathnode expression pointed by the ptok2->data.p
//		 field.
void AddLocalVariable(TOKEN *ptok1, TOKEN *ptok2, BOOL &fFuncError)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp AddLocalVariable\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok1->data.sz;
  MathNode *pmn = (MathNode *) ptok2->data.p;
  BOOL fError = FALSE;

  // since local variables are of local scope, their names may override
  // global identifiers but not local identifiers.
  HashNode *psym = symTable->Retrieve(psz);
  if (psym)
  { UINT u = psym->uType;
    if ((u == TYPE_ARG) || (u == TYPE_LVAR) || (u == TYPE_LRCONST) || (u == TYPE_LICONST))
    { error->AddMessage(EHM_SYMREDEF, psz, ptok1->iPtr);
      fError = TRUE;
    }
  }

  // check the equation
  if (pmn == NULL) fError = TRUE;

  // if an error has occurred, report that the assignment statement is
  // not declared
  if (fError)
  { fFuncError = TRUE;
    if (pmn) delete pmn;
    error->AddMessage(EHM_NOTDEC2, "local assignment", ptok1->iPtr);
  }

  // add the local variable to the symbol table
  else
  { UINT iLocalVar = symTable->cSymbols(TYPE_LVAR);
    SymLocalVar *psymLV = new SymLocalVar(psz, iLocalVar, pmn);
    symTable->Store(psymLV);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 AddLocalReturn
// Scope:	 Private
// Description:	 This function attempts to add a local assignment statement to
//		 the user-defined function.  The return value is stored in the
//		 mathnode expression pointed by the ptok2->data.p field.
void AddLocalReturn (TOKEN *ptok, BOOL &fFuncError)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp AddLocalReturn\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // check the equation
  MathNode *pmn = (MathNode *) ptok->data.p;
  if (pmn == NULL) fFuncError = TRUE;

  // store the return as a local variable
  else
  { UINT iLocalVar = symTable->cSymbols(TYPE_LVAR);
    SymLocalVar *psymLV = new SymLocalVar("return", iLocalVar, pmn);
    symTable->Store(psymLV);
  }

  // fetch the user-defined function
  UINT iFunc = symTable->cSymbols(TYPE_UDFUNC) - 1;
  SymUDFunc *psymUDF = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_UDFUNC, iFunc), SymUDFunc);

  // if an error occurred during the declaration of the function, report
  // that the function was not declared and remove it from the symbol table
  if (fFuncError)
  { char *pszFuncName = psymUDF->szName;
    error->AddMessage(EHM_NOTDEC3, pszFuncName, psymUDF->iPtr);
    symTable->Remove(pszFuncName);
  }

  // otherwise, add the assignment statements to the user-defined function
  else
  { UINT c = symTable->cSymbols(TYPE_LVAR);
    SymLocalVar **apsym = new SymLocalVar * [c];
    for(i = 0; i < c; ++i)
    { SymLocalVar *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_LVAR, i), SymLocalVar);
      apsym[i] = new SymLocalVar(psym->szName, psym->i, psym->pmn);
      psym->pmn = NULL;
    }
    psymUDF->apsym = apsym;
    psymUDF->cAssign = c;
    psymUDF->f = TRUE;
  }

  // remove local symbols from the symbol table
  symTable->RemoveSymbols(TYPE_LVAR);
  symTable->RemoveSymbols(TYPE_LRCONST);
  symTable->RemoveSymbols(TYPE_LICONST);
  symTable->RemoveSymbols(TYPE_ARG);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessOpen
// Scope:	 Private
// Description:	 This function processes the open file statement.
void ProcessOpen (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3, UINT &uIOMode)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessOpen\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uIOMode != IOMODE_NONE) error->AddMessage(EHM_FIOSTATE2, ptok1->iPtr);

  UINT uMode;
  switch (ptok3->uTokenType)
  { case TOK_READ:    uMode = IOMODE_READ;    uIOMode = IOMODE_READ;   break;
    case TOK_WRITE:   uMode = IOMODE_WRITE;   uIOMode = IOMODE_WRITE;  break;
    case TOK_APPEND:  uMode = IOMODE_APPEND;  uIOMode = IOMODE_WRITE;  break;
  }

  FileOpen *pfopen = new FileOpen((char *) ptok2->data.p, uMode, ptok1->iPtr);
  pstatements->AddMain(pfopen);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessClose
// Scope:	 Private
// Description:	 This function processes the close file statement.
void ProcessClose (TOKEN *ptok, UINT &uIOMode)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessClose\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uIOMode = IOMODE_NONE;
  FileClose *pfclose = new FileClose(ptok->iPtr);
  pstatements->AddMain(pfclose);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessGoto
// Scope:	 Private
// Description:	 This function processes the file goto statement.
void ProcessGoto (TOKEN *ptok1, TOKEN *ptok2, UINT &uIOMode)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessGoto\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uIOMode != IOMODE_READ)
    error->AddMessage(EHM_FIOSTATE1, "reading", ptok1->iPtr);

  else
  { FileGoto *pfgoto = new FileGoto((char *) ptok2->data.p, ptok1->iPtr);
    pstatements->AddMain(pfgoto);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessLabel
// Scope:	 Private
// Description:	 This function processes the file label statement.
void ProcessLabel (TOKEN *ptok1, TOKEN *ptok2, UINT &uIOMode)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessLabel\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uIOMode != IOMODE_WRITE)
    error->AddMessage(EHM_FIOSTATE1, "writing or appending", ptok1->iPtr);

  else
  { FileLabel *pflabel = new FileLabel((char *) ptok2->data.p, ptok1->iPtr);
    pstatements->AddMain(pflabel);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessComment
// Scope:	 Private
// Description:	 This function processes the file comment statement.
void ProcessComment (TOKEN *ptok1, TOKEN *ptok2, UINT &uIOMode)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessComment\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uIOMode != IOMODE_WRITE)
    error->AddMessage(EHM_FIOSTATE1, "writing or appending", ptok1->iPtr);

  else
  { FileComment *pfcomment = new FileComment((char *) ptok2->data.p, ptok1->iPtr);
    pstatements->AddMain(pfcomment);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessFileIO
// Scope:	 Private
// Description:	 This function processes the file i/o statements.
void ProcessFileIO (TOKEN *ptok, UINT uIOMode)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessFileIO\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok->data.sz;
  HashNode *psym = symTable->Retrieve(psz);

  // if the symbol has not been defined, then report an error
  if (!psym)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok->iPtr);
    return;
  }

  // if the symbol is not a variable or an array, then report an error
  if ((psym->uType != TYPE_VAR) && (psym->uType != TYPE_ARRAY))
  { error->AddMessage(EHM_SYMEXP, psz, "variable or array", ptok->iPtr);
    return;
  }

  // process variable symbol
  if (psym->uType == TYPE_VAR)
  { SymVar *psymV = TYPESAFE_DOWNCAST(psym, SymVar);
    UINT i = psymV->i;
    UINT u = psymV->u;
    if (uIOMode == IOMODE_READ)
    { FileRead *pfread = new FileRead(i, u, ptok->iPtr);
      pstatements->AddMain(pfread);
    }
    else
    { FileWrite *pfwrite = new FileWrite(i, u, ptok->iPtr);
      pstatements->AddMain(pfwrite);
    }
  }

  // process array symbol
  else
  { SymArray *psymA = TYPESAFE_DOWNCAST(psym, SymArray);
    UINT i = psymA->i;
    UINT u = psymA->u;
    UINT c = psymA->c;
    UINT *ac = psymA->ac;
    if (uIOMode == IOMODE_READ)
    { FileRead *pfread = new FileRead(i, u, c, ac, ptok->iPtr);
      pstatements->AddMain(pfread);
    }
    else
    { FileWrite *pfwrite = new FileWrite(i, u, c, ac, ptok->iPtr);
      pstatements->AddMain(pfwrite);
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareProcedure
// Scope:	 Private
// Description:	 This function processes the beginning of a procedure.
void DeclareProcedure (TOKEN *ptok)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp DeclareProcedure\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok->data.sz;

  // report an error if the symbol is already defined
  if (symTable->Retrieve(psz))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok->iPtr);
    return;
  }

  // construct and store the procedure symbol
  UINT cStatement = pstatements->GetNumOfStatements();
  UINT iProc = symTable->cSymbols(TYPE_PROC);
  SymProc *psym = new SymProc(psz, iProc, cStatement);
  symTable->Store(psym);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessCall
// Scope:	 Private
// Description:	 This function processes the call statement.
void ProcessCall (TOKEN *ptok1, TOKEN *ptok2)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessCall\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz = ptok2->data.sz;
  HashNode *psym = symTable->Retrieve(psz);
  BOOL fError = FALSE;

  // report an error if the symbol is not defined
  if (psym == NULL)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok2->iPtr);
    fError = TRUE;
  }

  // report an error if the symbol is not a procedure
  else if (psym->uType != TYPE_PROC)
  { error->AddMessage(EHM_SYMEXP, psz, "procedure", ptok2->iPtr);
    fError = TRUE;
  }

  // if an error occurs, then report that the statement is not declared
  if (fError) error->AddMessage(EHM_NOTDEC2, "call", ptok1->iPtr);

  // everything seems o.k.
  else
  { SymProc *psymProc = TYPESAFE_DOWNCAST(psym, SymProc);
    Call *ps = new Call(psymProc->iProc, psymProc->i, ptok1->iPtr);
    pstatements->AddMain(ps);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessMesh
// Scope:	 Private
// Description:	 This function processes the mesh statement.
void ProcessMesh (TOKEN *ptok1, TOKEN *ptok2, UINT &cVariables)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessMesh\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // retrieve the mesh file name
  pszMeshFileName = (char *) ptok2->data.p;
  if (!pszMeshFileName)
  { error->AddMessage(EHM_NOTDEC2, "mesh", ptok1->iPtr);
    return;
  }

  // open the mesh file
  int nFile = open(pszMeshFileName, O_RDONLY | O_BINARY);
  if (nFile == -1)
  { error->AddMessage(EHM_FILEOPEN, pszMeshFileName);
    return;
  }

  // read the header
  MESHHEAD meshhead;
  read(nFile, &meshhead, sizeof(MESHHEAD));
  if (strcmp(meshhead.szSign, SIGNATURE_MESH))
  { error->AddMessage(EHM_BADMESHFILE);
    return;
  }

  // initialize global variables
  cRegLabel  = meshhead.cRegLabel;
  cBndryNode = meshhead.cBndryNode;
  cEdge	     = meshhead.cEdge;

  // read constants and add them to the symbol table
  UINT cConstant;
  read(nFile, &cConstant, sizeof(UINT));
  if (meshhead.cConstant != cConstant)
  { error->AddMessage(EHM_BADMESHFILE);
    return;
  }
  for(i = 0; i < cConstant; ++i)
  { XLATCONST xlatconst;
    read(nFile, &xlatconst, sizeof(XLATCONST));
    if (symTable->Retrieve(xlatconst.szName))
      error->AddMessage(EHM_SYMREDEF, xlatconst.szName, ptok2->iPtr);
    else
    { if (xlatconst.uType == 2) // @@@ need to make a global header file
	   symTable->Store(new SymIConst(xlatconst.szName, xlatconst.data.n));
      else symTable->Store(new SymRConst(xlatconst.szName, xlatconst.data.r));
    }
  }

  // read the labels
  UINT cLabel;
  read(nFile, &cLabel, sizeof(UINT));
  if (meshhead.cLabel != cLabel)
  { error->AddMessage(EHM_BADMESHFILE);
    return;
  }
  for(i = 0; i < cLabel; ++i)
  { XLATLABEL xlatlabel;
    read(nFile, &xlatlabel, sizeof(XLATLABEL));
    if (symTable->Retrieve(xlatlabel.szName))
    { error->AddMessage(EHM_SYMREDEF, xlatlabel.szName, ptok2->iPtr);
      lseek(nFile, xlatlabel.cNode * sizeof(UINT), SEEK_CUR);
    }
    else
    { UINT *au = new UINT[xlatlabel.cNode];
      read(nFile, au, xlatlabel.cNode * sizeof(UINT));
      symTable->Store(new SymLabel(xlatlabel.szName, xlatlabel.iIndex, xlatlabel.cNode, au));
    }
  }

  // read the 1D arrays
  UINT cArray;
  read(nFile, &cArray, sizeof(UINT));
  if (meshhead.cArray != cArray)
  { error->AddMessage(EHM_BADMESHFILE);
    return;
  }
  UINT cArraySymbol = symTable->cSymbols(TYPE_ARRAY);
  for(i = 0; i < cArray; ++i)
  { char szArrayName[LEN_IDENT+1];
    read(nFile, szArrayName, LEN_IDENT+1);
    if (symTable->Retrieve(szArrayName))
      error->AddMessage(EHM_SYMREDEF, szArrayName, ptok2->iPtr);
    else
    {
#if defined(P16BIT)
      if (cVariables + meshhead.cNode > MAX_VARIABLES)
	error->AddMessage(EHM_MAXVAR, szArrayName, ptok2->iPtr);
      else
#endif
      symTable->Store(new SymArray(szArrayName, cArraySymbol++, 1, meshhead.cNode, 0, 0, cVariables, TRUE));
      cVariables += meshhead.cNode;
      lseek(nFile, meshhead.cNode * sizeof(REAL), SEEK_CUR);
    }
  }

  // add the NODES constant to the symbol table and advance the file pointer
  cNode = meshhead.cNode;
  lseek(nFile, meshhead.cNode * sizeof(NODE) + sizeof(UINT), SEEK_CUR);
  if (symTable->Retrieve("NODES"))
    error->AddMessage(EHM_SYMREDEF, "NODES", ptok2->iPtr);
  else symTable->Store(new SymIConst("NODES", meshhead.cNode));

  // read the edge array and add the EDGES constant to the symbol table
  UINT cEdge;
  read(nFile, &cEdge, sizeof(UINT));
  if (cEdge != meshhead.cEdge)
  { error->AddMessage(EHM_BADMESHFILE);
    return;
  }
  EDGE *aedge = new EDGE[cEdge];
  read(nFile, aedge, cEdge * sizeof(EDGE));
  if (symTable->Retrieve("EDGES"))
    error->AddMessage(EHM_SYMREDEF, "EDGES", ptok2->iPtr);
  else symTable->Store(new SymIConst("EDGES", cEdge));

  // add the TRIANGLES constant to the symbol table and advance the file pointer
  lseek(nFile, meshhead.cTriangle * sizeof(TRI) + sizeof(UINT), SEEK_CUR);
  if (symTable->Retrieve("TRIANGLES"))
    error->AddMessage(EHM_SYMREDEF, "TRIANGLES", ptok2->iPtr);
  else symTable->Store(new SymIConst("TRIANGLES", meshhead.cTriangle));

  // read adjacency information
  read(nFile, &cAdjInfo, sizeof(UINT));
  if (cAdjInfo != meshhead.cNode + meshhead.cBndryNode * meshhead.cRegLabel)
  { error->AddMessage(EHM_BADMESHFILE);
    return;
  }
  NODEINFO nodeinfo;
  EDGEINFO edgeinfo;
  aauNeighbors = new UINT *[cAdjInfo];
  for(i = 0; i < cAdjInfo; ++i) aauNeighbors[i] = NULL;
  for(i = 0; i < cAdjInfo; ++i)
  { INODE inode;
    read(nFile, &inode, sizeof(INODE));
    read(nFile, &nodeinfo, sizeof(NODEINFO) - sizeof(EDGEINFO *));
    UINT *pu = aauNeighbors[inode] = new UINT[2 * nodeinfo.cNeighbor + 2];
    *(pu++) = nodeinfo.cNeighbor;
    *(pu++) = nodeinfo.cTriangle;
    for(UINT j = 0; j < nodeinfo.cNeighbor; ++j)
    { read(nFile, &edgeinfo, sizeof(EDGEINFO));
      *(pu++) = edgeinfo.inode;
      *(pu++) = aedge[edgeinfo.iedge].uLabel;
    }
  }

  // delete the array of edges and close the file
  delete [] aedge;
  close(nFile);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 fIsLabel
// Scope:	 Private
// Description:	 This function returns TRUE if the identifier is a label
//		 otherwise FALSE is returned.
BOOL fIsLabel (char *psz)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp fIsLabel\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  HashNode *psym = symTable->Retrieve(psz);
  return psym && (psym->uType == TYPE_LABEL);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessPrecompute
// Scope:	 Private
// Description:	 This function processes the precompute statements.
//
//		 ptok1 is the keyword precompute
//		 ptok2 is the node or edge index
//		 ptok3 is the edge node A index
//		 ptok4 is the edge node B index
//		 ptok5 is the keyword odd or even
//		 ptok6 is the user-defined function name
//		 ptok7 is the user-defined function arguments
void ProcessPrecompute (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3, TOKEN *ptok4,
			TOKEN *ptok5, TOKEN *ptok6, TOKEN *ptok7)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp ProcessPrecompute\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  BOOL fError = FALSE;

  // report an error if the triangular mesh has not been declared
  if (!aauNeighbors)
  { error->AddMessage(EHM_TRIMESH, ptok1->iPtr);
    fError = TRUE;
  }

  // initialize local variables
  BOOL fEdge = ptok3 != NULL;
  BOOL fOdd  = fEdge && (ptok5->uTokenType == TOK_ODD);
  UINT i1 = ptok2->data.u;
  UINT i2 = (fEdge) ? ptok3->data.u : 0;
  UINT i3 = (fEdge) ? ptok4->data.u : 0;
  char *psz = ptok6->data.sz;

  // are the indices valid?
  fError = fError || (i1 == INDEX_ERROR) || (i2 == INDEX_ERROR) || (i3 == INDEX_ERROR);

  // is the identifier a user-defined function
  HashNode *psym = symTable->Retrieve(psz);
  if	  (psym == NULL)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok6->iPtr);
    fError = TRUE;
  }
  else if (psym->uType != TYPE_UDFUNC)
  { error->AddMessage(EHM_SYMEXP, psz, "user-defined function", ptok6->iPtr);
    fError = TRUE;
  }

  // deallocate the argument list if an error occurs
  if (fError)
  { error->AddMessage(EHM_NOTDEC2, "precompute", ptok1->iPtr);
    if (ptok7->data.p) delete (MathNode *) ptok7->data.p;
    symTable->RemoveSymbols(TYPE_INDEX);
    return;
  }

  // create a function mathnode
  MathNode *pmn = (fError) ? NULL : pmnFunction(ptok6, ptok7);
  fError = pmn == NULL;

  // report error if one occurs
  if (fError)
  { error->AddMessage(EHM_NOTDEC2, "precompute", ptok1->iPtr);
    symTable->RemoveSymbols(TYPE_INDEX);
    return;
  }

  // initialize the precomputation members of the function
  SymUDFunc *psymUDFunc = TYPESAFE_DOWNCAST(psym, SymUDFunc);
  psymUDFunc->InitPrecompute(!fEdge, fOdd, pmn);
  symTable->RemoveSymbols(TYPE_INDEX);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 NormalExprMode
// Scope:	 Private
// Description:	 This function returns TRUE if the expression mode is NORMAL
//		 otherwise the mathnode expression in ptok2 and ptok3 are
//		 deallocated and FALSE is returned.
BOOL NormalExprMode (UINT uExprMode, TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp NormalExprMode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExprMode == EXPRMODE_NORMAL) return TRUE;
  if (ptok2 && ptok2->data.p) delete (MathNode *) ptok2->data.p;
  if (ptok3 && ptok3->data.p) delete (MathNode *) ptok3->data.p;
  char szOperator[16];
  switch (ptok1->uTokenType)
  { case TYPE_NSUM:   strcpy(szOperator, "nsum");   break;
    case TYPE_LSUM:   strcpy(szOperator, "lsum");   break;
    case TYPE_AREA:   strcpy(szOperator, "area");   break;
    case TYPE_AREA1:  strcpy(szOperator, "area1");  break;
    case TYPE_AREA2:  strcpy(szOperator, "area2");  break;
    case TYPE_NODE:   strcpy(szOperator, "node");   break;
    case TYPE_EDGE:   strcpy(szOperator, "edge");   break;
    case TYPE_ELEN:   strcpy(szOperator, "elen");   break;
    case TYPE_ILEN:   strcpy(szOperator, "ilen");   break;
  }
  error->AddMessage(EHM_PRECOMP, szOperator, "cannot", ptok1->iPtr);
  return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnNoArgELENOperator
// Scope:	 Private
// Description:	 This function returns NULL if the expression mode is NORMAL
//		 otherwise an elen operator with no arguments is created.
MathNode * pmnNoArgELENOperator (UINT uExprMode, BOOL fNode, TOKEN *ptok)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnNoArgELENOperator\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if ((uExprMode == EXPRMODE_NORMAL) || fNode)
  { error->AddMessage(EHM_PRECOMP, "zero argument elen", "only", ptok->iPtr);
    return NULL;
  }
  MathNode *pmn = new MathNode();
  pmn->uNode = TYPE_ELENABS;
  pmn->uExpr = FLAG_VAR | FLAG_REAL;
  pmn->iPtrNode = ptok->iPtr;
  pmn->iPtrExpr = ptok->iPtr;
  pmn->data.n = NO_REGION_LABEL;
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnFuncRef
// Scope:	 Private
// Description:	 This function makes a precomputed function reference mathode.
//		 The '@' symbol is ptok1, the function name is ptok2 and the
//		 the function indices are ptok3.
MathNode * pmnFuncRef (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp pmnFuncRef\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  char *psz = ptok2->data.sz;
  MathNode *pmn = (MathNode *) ptok3->data.p;
  HashNode *psym = symTable->Retrieve(psz);
  MathNode *pmnTemplate;
  SymUDFunc *psymUDF;
  BOOL fNode;
  BOOL fOdd;

  // report an error if the symbol is not defined
  BOOL fError = psym == NULL;
  if (fError)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok2->iPtr);
  }

  // report an error if the symbol is not a precomputed user-defined function
  if (!fError) fError = psym->uType != TYPE_UDFUNC;
  if (!fError)
  { psymUDF = TYPESAFE_DOWNCAST(psym, SymUDFunc);
    pmnTemplate = psymUDF->pmnTemplate;
    fNode = psymUDF->fNode;
    fOdd = psymUDF->fOdd;
    fError = pmnTemplate == NULL;
  }
  if (fError)
  { error->AddMessage(EHM_SYMEXP, psz, "precomputed user-defined function", ptok2->iPtr);
  }

  // determine if an error has occurred in the specification of the function
  // and verify the arguments are integer expressions
  if (!fError)
  { char szBuf[16];
    for(i = 0; i < pmn->c; ++i)
      if (pmn->apmn[i] == NULL)
      { fError = TRUE;
      }
      else if (pmn->apmn[i]->uExpr & FLAG_REAL)
      { sprintf(szBuf, "argument %u", i);
	error->AddMessage(EHM_INTEXPR, szBuf, pmn->apmn[i]->iPtrExpr);
	fError = TRUE;
      }
  }

  // check the number of arguments
  if (!fError)
  { if ((fNode && (pmn->c != 1)) || (!fNode && (pmn->c != 3)))
    { error->AddMessage(EHM_NUMARGS, psz, ptok1->iPtr);
      fError = TRUE;
    }
  }

  // return if an error has occurred
  if (fError)
  { delete pmn;
    return NULL;
  }

  // make function reference
  MathNode *pmnCopy = pmnTemplate->CopyTemplate(pmn);
  pmnCopy->iPtrNode = ptok1->iPtr;
  pmnCopy->iPtrExpr = ptok1->iPtr;
  pmn->uNode = TYPE_PRECOMP;
  pmn->uExpr = pmnCopy->uExpr;
  pmn->iPtrNode = ptok1->iPtr;
  pmn->iPtrExpr = ptok1->iPtr;
  pmn->AddChild(NULL);
  for(i = pmn->c - 2; i != -1u; --i) pmn->apmn[i+1] = pmn->apmn[i];
  pmn->apmn[0] = pmnCopy;
  pmn->data.precomp.fNode = fNode;
  pmn->data.precomp.fOdd = fOdd;
  pmn->data.precomp.au1 = psymUDF->auArg1;
  pmn->data.precomp.au2 = psymUDF->auArg2;
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Source::UserActions
// Scope:	 Private
// Description:	 This function is the user action handler.  Zero is returned
//		 if no error occurs; otherwise, a nonzero interger is returned.
void Source::UserActions (UINT UserAction)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("actions.cpp Source::UserActions\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  static BOOL fPrecomputeAtNodes;
  static UINT uExprMode;
  TOKEN	 atoken[LTA_LEN];

  switch (UserAction)
  { // initialize the user actions function
    case USR_INIT:
      uSyntaxState = ST_PROG;
      uIOMode = IOMODE_NONE;
      cVariables = 0;
      fFuncBody = FALSE;
      fFuncError = FALSE;
      break;

    // parsing program statements
    case USR_PROG:
      uSyntaxState = ST_PROG;
      if (uIOMode != IOMODE_NONE) error->AddMessage(EHM_FILECLOSE, token.iPtr);
      break;

    // parsing main procedure statements
    case USR_MAIN:
      *atoken = stackToken->pop();
      if (iMainProc != -1u) error->AddMessage(EHM_MAINPROC, atoken->iPtr);
      uSyntaxState = ST_ERROR;
      iMainProc = pstatements->GetNumOfStatements();
      break;

    // declare an index
    case USR_DINDEX:
      atoken[0] = stackToken->pop();
      atoken[0].data.u = DeclareIndex(atoken);
      stackToken->push(*atoken);
      break;

    // declare function argument ... may refer to an unknown
    case USR_FARG1:
      DeclareArgument(&token, TRUE, fFuncError);
      break;

    // declare function argument ... may not refer to an unknown
    case USR_FARG2:
      *atoken = stackToken->pop();
      DeclareArgument(atoken, FALSE, fFuncError);
      break;

    // declare a local constant
    case USR_DLCONST:
      AdjustTokenStack(2, atoken, stackToken);
      DeclareLocalConstant(atoken, atoken + 1, fFuncError);
      break;

    // declare a constant
    case USR_DCONST:
      AdjustTokenStack(2, atoken, stackToken);
      DeclareConstant(atoken, atoken + 1);
      break;

    // declare a local variable / assignment
    case USR_LASS:
      AdjustTokenStack(2, atoken, stackToken);
      AddLocalVariable(atoken, atoken + 1, fFuncError);
      break;

    // declare a variable
    case USR_DVAR:
      *atoken = stackToken->pop();
      DeclareVariable(atoken, cVariables);
      break;

    // declare an array
    case USR_DARRAY:
      AdjustTokenStack(2, atoken, stackToken);
      DeclareArray(atoken, atoken + 1, cVariables);
      break;

    // declare a user-defined function
    case USR_DECFUNC:
      *atoken = stackToken->pop();
      DeclareUserDefinedFunction(atoken, fFuncError);
      fFuncBody = TRUE;
      break;

    // declare a external function
    case USR_DECEXT:
      *atoken = stackToken->pop();
      DeclareUserDefinedFunction(atoken, fFuncError, TRUE);
      break;

    // declare a return statement
    case USR_RETURN:
      *atoken = stackToken->pop();
      AddLocalReturn(atoken, fFuncError);
      fFuncBody = fFuncError = FALSE;
      break;

    // declare a range
    case USR_DRANGE:
      atoken[0].data.u = INDEX_NONE;
      stackToken->push(*atoken);
      break;

    // make a real constant math node
    case USR_RCONST:
      token.data.p = (void *) new MathNode(token.data.r, token.iPtr);
      break;

    // make an integer constant math node
    case USR_ICONST:
      token.data.p = (void *) new MathNode(token.data.n, token.iPtr);
      break;

    // make an identifier (i.e. variable or constant) math node
    case USR_ID:
      token.data.p = (void *) pmnIdentifier(&token, fFuncBody);
      break;

    // make an array math node expression
    case USR_ARRAY:
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (void *) pmnArray(atoken, atoken + 1, fFuncBody);
      break;

    // make a function math node expression
    case USR_FUNC:
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (void *) pmnFunction(atoken, atoken + 1);
      break;

    // make a unary operator math node expression
    case USR_UNOP1:
      atoken[0] = stackToken->pop();
      token.data.p = (void *) pmnUnaryOperator1(atoken, &token);
      break;

    // make a unary operator with an integer argument
    case USR_UNOP2:
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (NormalExprMode(uExprMode, atoken, atoken + 1, NULL))
		   ? (void *) pmnUnaryOperator2(atoken, atoken + 1, NULL)
		   : NULL;
      break;

    // make a unary operator with an integer argument
    case USR_UNOP2A:
      AdjustTokenStack(3, atoken, stackToken);
      token.data.p = (NormalExprMode(uExprMode, atoken, atoken + 1, NULL))
		   ? (void *) pmnUnaryOperator2(atoken, atoken + 1, atoken + 2)
		   : NULL;
      break;

    // make a real operator math node expression
    case USR_REAL:
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (void *) pmnUnaryOperator1(atoken, atoken + 1);
      break;

    // make a binary operator math node expression
    case USR_BINOP1:
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (void *) pmnBinaryOperator1(atoken + 1, atoken, &token);
      break;

    // make a binary operator with integer arguments
    case USR_BINOP2:
      AdjustTokenStack(3, atoken, stackToken);
      token.data.p = (NormalExprMode(uExprMode, atoken, atoken + 1, atoken + 2))
		   ? (void *) pmnBinaryOperator2(atoken, atoken + 1, atoken + 2, NULL)
		   : NULL;
      break;

    case USR_BINOP2A:
      AdjustTokenStack(4, atoken, stackToken);
      token.data.p = (NormalExprMode(uExprMode, atoken, atoken + 1, atoken + 2))
		   ? (void *) pmnBinaryOperator2(atoken, atoken + 1, atoken + 2, atoken + 3)
		   : NULL;
      break;

    // node summation statement
    case USR_NSUM:
      AdjustTokenStack(5, atoken, stackToken);
      token.data.p = (NormalExprMode(uExprMode, atoken, atoken + 1, atoken + 4))
		   ? (void *) pmnNodeSum(atoken, atoken + 1, atoken + 2, atoken + 3, atoken + 4)
		   : NULL;
      break;

    // label summation statement
    case USR_LSUM:
      AdjustTokenStack(4, atoken, stackToken);
      token.data.p = (NormalExprMode(uExprMode, atoken, atoken + 3, NULL))
		   ? (void *) pmnLabelSum(atoken, atoken + 1, atoken + 2, atoken + 3)
		   : NULL;
      break;

    // process '(' expr ')', '[' expr ']' and '{' expr '}' expressions
    case USR_PAREN:
      token = stackToken->pop();
      break;

    // create a parent math node
    case USR_EXPRS:
      atoken->data.p = (void *) new MathNode();
      stackToken->push(*atoken);
      break;

    // add a child math node expression to a parent math node
    case USR_ADDEXPR:
      atoken[1] = stackToken->pop();
      atoken[0] = stackToken->pop();
      ((MathNode *) atoken[0].data.p)->AddChild((MathNode *) atoken[1].data.p);
      stackToken->push(atoken[0]);
      break;

    // create an index with start, end and step values specified
    case USR_INDEX4:
      stackToken->push(token);
      AdjustTokenStack(4, atoken, stackToken);
      token.data.p = (void *) pindexMake(4, atoken);
      break;

    // create an index with start and end values
    case USR_INDEX3:
      stackToken->push(token);
      AdjustTokenStack(3, atoken, stackToken);
      token.data.p = (void *) pindexMake(3, atoken);
      break;

    // create an index with a start value
    case USR_INDEX2:
      stackToken->push(token);
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (void *) pindexMake(2, atoken);
      break;

    // create an index which loops through the entire range
    case USR_INDEX1:
      *atoken = stackToken->pop();
      token.data.p = (void *) pindexMake(1, atoken);
      break;

    // create an index which loops over the nodes specified by a label
    case USR_INDEXL:
      stackToken->push(token);
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (void *) pindexLabelMake(atoken, atoken + 1);
      break;

    // create an instance of the indices class
    case USR_INDICES:
      atoken->data.p = (void *) new Indices();
      stackToken->push(*atoken);
      break;

    // add an index to the indices class
    case USR_ADDINDX:
      atoken[1] = stackToken->pop();
      atoken[0] = stackToken->pop();
      ((Indices *) atoken[0].data.p)->AddIndex((Index *) atoken[1].data.p);
      stackToken->push(atoken[0]);
      break;

    // tag variable as a known
    case USR_KNOWN1:
      atoken[0] = stackToken->pop();
      ProcessKnown(atoken, NULL, TRUE);
      break;

    // tag variable as a known
    case USR_KNOWN2:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessKnown(atoken, atoken + 1, TRUE);
      break;

    // tag variable as a known
    case USR_UNKNO1:
      atoken[0] = stackToken->pop();
      ProcessKnown(atoken, NULL, FALSE);
      break;

    // tag variable as a known
    case USR_UNKNO2:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessKnown(atoken, atoken + 1, FALSE);
      break;

    // variable equation declaration
    case USR_EQN1:
      AdjustTokenStack(3, atoken, stackToken);
      ProcessEquation(atoken, NULL, atoken + 1, atoken + 2);
      break;

    // array equation declaration
    case USR_EQN2:
      AdjustTokenStack(4, atoken, stackToken);
      ProcessEquation(atoken, atoken + 1, atoken + 2, atoken + 3);
      break;

    // solve equations statement
    case USR_SOLVE:
      *atoken = stackToken->pop();
      ProcessSolve(atoken);
      break;

    // solve equations statement
    case USR_WRITE:
      *atoken = stackToken->pop();
      ProcessWrite(atoken);
      break;

    // variable assigment statement
    case USR_ASS1:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessAssignment(atoken, NULL, atoken + 1);
      break;

    // array assignment statement
    case USR_ASS2:
      AdjustTokenStack(3, atoken, stackToken);
      ProcessAssignment(atoken, atoken + 1, atoken + 2);
      break;

    // first part of the while statement
    case USR_WHILE1:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessWhile1(atoken, atoken + 1, stackToken);
      break;

    // second part of the while statement
    case USR_WHILE2:
      *atoken = stackToken->pop();
      ProcessWhile2(atoken);
      break;

    // first part of the do statement
    case USR_DO1:
      ProcessDo1(stackToken);
      break;

    // second part of the do statement
    case USR_DO2:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessDo2(atoken, atoken + 1);
      break;

    // first part of the if statement
    case USR_IF1:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessIf1(atoken, atoken + 1, stackToken);
      break;

    // second part of the if or if-else statement
    case USR_IF2:
    case USR_IF3:
      *atoken = stackToken->pop();
      ProcessIf2(atoken, UserAction == USR_IF3, stackToken);
      break;

    // last part of the if-else statement
    case USR_IF4:
      *atoken = stackToken->pop();
      ProcessIf2(atoken);
      break;

    // internal variable assignment
    case USR_SET:
      AdjustTokenStack(3, atoken, stackToken);
      ProcessSet(atoken, atoken + 1, atoken + 2);
      break;

    // file open statement
    case USR_OPEN:
      AdjustTokenStack(3, atoken, stackToken);
      ProcessOpen(atoken, atoken + 1, atoken + 2, uIOMode);
      break;

    // file close statement
    case USR_CLOSE:
      *atoken = stackToken->pop();
      ProcessClose(atoken, uIOMode);
      break;

    // verify read statement
    case USR_VREAD:
      *atoken = stackToken->pop();
      if (uIOMode != IOMODE_READ)
	error->AddMessage(EHM_FIOSTATE1, "reading", atoken->iPtr);
      break;

    // verify read statement
    case USR_VWRITE:
      *atoken = stackToken->pop();
      if (uIOMode != IOMODE_WRITE)
	error->AddMessage(EHM_FIOSTATE1, "writing or appending", atoken->iPtr);
      break;

    // process file goto statement
    case USR_GOTO:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessGoto(atoken, atoken + 1, uIOMode);
      break;

    // process file label statement
    case USR_LABEL:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessLabel(atoken, atoken + 1, uIOMode);
      break;

    // process file goto statement
    case USR_COMMENT:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessComment(atoken, atoken + 1, uIOMode);
      break;

    // process file input/output statement
    case USR_IO:
      *atoken = stackToken->pop();
      ProcessFileIO(atoken, uIOMode);
      break;

    // process exit statement
    case USR_EXIT:
      *atoken = stackToken->pop();
      pstatements->AddMain(new Statement(TYPE_END, atoken->iPtr));
      break;

    // process beginning of procedure
    case USR_PROC1:
      *atoken = stackToken->pop();
      DeclareProcedure(atoken);
      uSyntaxState = ST_ERROR;
      break;

    // process end of procedure
    case USR_PROC2:
      uSyntaxState = ST_PROG;
      if (uIOMode != IOMODE_NONE) error->AddMessage(EHM_FILECLOSE, token.iPtr);
      pstatements->AddMain(new Statement(TYPE_RETURN, -1u));
      pstatements->TestForUserDefProc();
      break;

    // process call statement
    case USR_CALL:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessCall(atoken, atoken + 1);
      break;

    // process the mesh statement
    case USR_MESH:
      AdjustTokenStack(2, atoken, stackToken);
      ProcessMesh(atoken, atoken + 1, cVariables);
      break;

    // test the identifier to see if it is a label
    case USR_LABTEST:
      if (fIsLabel(token.data.sz)) token.uTokenType = TOK_LABEL;
      break;

    // make a reference to a precomputed function
    case USR_REF:
      AdjustTokenStack(3, atoken, stackToken);
      token.data.p = (void *) pmnFuncRef(atoken, atoken + 1, atoken + 2);
      break;

    // set the precompute at nodes flag
    case USR_PCNODE:
      fPrecomputeAtNodes = TRUE;
      break;

    // clear the precompute at nodes flag
    case USR_PCEDGE:
      fPrecomputeAtNodes = FALSE;
      break;

    // process the precompute declaration
    case USR_PRECOMP:
      if (fPrecomputeAtNodes)
      { AdjustTokenStack(4, atoken, stackToken);
	ProcessPrecompute(atoken, atoken+1, NULL, NULL, NULL, atoken+2, atoken+3);
      }
      else
      { AdjustTokenStack(7, atoken, stackToken);
	ProcessPrecompute(atoken, atoken+1, atoken+2, atoken+3, atoken+4, atoken+5, atoken+6);
      }
      break;

    // process the elen operator without arguments
    case USR_ELEN:
      token.data.p = (void *) pmnNoArgELENOperator(uExprMode, fPrecomputeAtNodes, &token);
      break;

    // set the expression mode to NORMAL ... these expressions cannot use
    // the elen operator w/o arguments
    case USR_EM1:
      uExprMode = EXPRMODE_NORMAL;
      break;

    // set the expression mode to PRECOMP ... these expressions cannot use
    // the node, edge, area1, area2, nsum and ilen operators in addition to
    // the elen operator w/ arguments
    case USR_EM2:
      uExprMode = EXPRMODE_PRECOMP;
      break;
  }
}


