/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     codebuf.cpp
 AUTHOR:     Kevin M. Kramer
             J. Joshua Feng

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 09/10/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version
 07/06/00    JJF      modify AddText

 DESCRIPTION:

 This module implements the code buffer class.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#if defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#else
#include <io.h>
#include <sys\stat.h>
#endif

#include <fcntl.h>

#include "codebuf.h"
#include "error.h"
#include "symbol.h"
#include "mathnode.h"


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::CodeBuffer
// Scope:	 Public, Constructor
CodeBuffer::CodeBuffer (char *pszName, char *pszPrefix, UINT uPrec, UINT iLeft, UINT iRight)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::CodeBuffer\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  strcpy(szFileName, pszName);
  strcpy(szUCPrefix, pszPrefix);
  strcpy(szLCPrefix, pszPrefix);
  strupr(szUCPrefix);
  /* strlwr(szLCPrefix); */ /* Switch to case-sensitive by Josh */
  cPrecision = uPrec;
  iLeftMargin = iLeft;
  iRightMargin = iRight;
  char *psz = szBuffer;
  for(iBuffer = 0; iBuffer < iLeftMargin; ++iBuffer, ++psz) *psz = ' ';
  nFile = open(pszName, O_CREAT | O_TRUNC | O_BINARY | O_WRONLY, S_IREAD | S_IWRITE);
  if (nFile == -1) error->AddMessage(EHM_FILEOPEN, pszName);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::~CodeBuffer
// Scope:	 Public, Destructor
CodeBuffer::~CodeBuffer ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::~CodeBuffer\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  Flush();
  close(nFile);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::Flush
// Scope:	 Private
// Description:	 This function flushes the code buffer.
void CodeBuffer::Flush ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::Flush\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

#if defined(UNIX)
  szBuffer[iBuffer++] = '\n';
#else
  szBuffer[iBuffer++] = '\x0d';
  szBuffer[iBuffer++] = '\x0a';
#endif
  write(nFile, szBuffer, iBuffer);
  char *psz = szBuffer;
  for(iBuffer = 0; iBuffer < iLeftMargin; ++iBuffer, ++psz) *psz = ' ';
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddToken
// Scope:	 Private
// Description:	 This function adds a token to the code buffer.
void CodeBuffer::AddToken (char *psz)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddToken\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT c = strlen(psz);
  if (iBuffer + c >= iRightMargin) Flush();
  strcpy(szBuffer + iBuffer, psz);
  iBuffer += c;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::SetLeftMargin
// Scope:	 Public
// Description:	 This function sets the left margin
void CodeBuffer::SetLeftMargin (UINT u)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::SetLeftMargin\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  iLeftMargin = u;

  for(i = 0; i < iBuffer; ++i)
    if (szBuffer[i] != ' ') break;

  if (i == iBuffer)
    for(iBuffer = 0; iBuffer < iLeftMargin; ++iBuffer)
      szBuffer[iBuffer] = ' ';
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddChar
// Scope:	 Public
// Description:	 This function adds a character to the code buffer.
void CodeBuffer::AddChar (char ch)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddChar\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  switch (ch)
  { case '\n':
      Flush();
      break;

    case ' ':
      if (iBuffer + 1 >= iRightMargin) Flush();
      else if (iBuffer > iLeftMargin) szBuffer[iBuffer++] = ch;
      break;

    case '\0':
      break;

    default:
      if (iBuffer + 1 >= iRightMargin) Flush();
      szBuffer[iBuffer++] = ch;
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddIndex
// Scope:	 Public
// Description:	 This function adds an index to the code buffer.
void CodeBuffer::AddIndex (UINT i, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddIndex\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "_i%u", i + 1);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddArgument
// Scope:	 Public
// Description:	 This function adds an argument to the code buffer.
void CodeBuffer::AddArgument (UINT i, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddArgument\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "_x%u", i + 1);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddInteger
// Scope:	 Public
// Description:	 This function adds an integer to the code buffer.
void CodeBuffer::AddInteger (int n, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddInteger\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, " %d", n);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddUInteger
// Scope:	 Public
// Description:	 This function adds an unsigned integer to the code buffer.
void CodeBuffer::AddUInteger (UINT u, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddUInteger\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, " %u", u);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddReal
// Scope:	 Public
// Description:	 This function adds a real number to the code buffer.
void CodeBuffer::AddReal (REAL r, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddReal\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, " %*.*le", cPrecision + 7, cPrecision, r);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddVariable
// Scope:	 Public
// Description:	 This function adds a variable to the code buffer.
void CodeBuffer::AddVariable (char *psz, BOOL fTag, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddVariable\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "%s%s._%s", (fTag) ? "tag" : "var", szLCPrefix, psz);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddVariable
// Scope:	 Public
// Description:	 This function adds a variable to the code buffer.
void CodeBuffer::AddVariable (UINT i, BOOL fTag, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddVariable\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  SymVar *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_VAR, i), SymVar);
  AddVariable(psym->szName, fTag, fAddSpace);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddArray
// Scope:	 Public
// Description:	 This function adds an array to the code buffer.
void CodeBuffer::AddArray (char *psz, BOOL fTag, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddArray\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "%s%s._%s", (fTag) ? "tag" : "var", szLCPrefix, psz);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddArray
// Scope:	 Public
// Description:	 This function adds an array to the code buffer.
void CodeBuffer::AddArray (UINT i, BOOL fTag, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddArray\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  SymArray *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARRAY, i), SymArray);
  AddArray(psym->szName, fTag, fAddSpace);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddLocalVariable
// Scope:	 Public
// Description:	 This function adds a local variable to the code buffer.
void CodeBuffer::AddLocalVariable (UINT i, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddLocalVariable\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "_lvar%u", i + 1);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddLocalDerivative
// Scope:	 Public
// Description:	 This function adds a local derivative to the code buffer.
void CodeBuffer::AddLocalDerivative (UINT i, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddLocalDerivative\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "_lder%u", i + 1);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddEquation
// Scope:	 Public
// Description:	 This function adds an equation to the code buffer.
void CodeBuffer::AddEquation (UINT i, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddEquation\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "EQU_%s_%u", szUCPrefix, i + 1);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddPartial
// Scope:	 Public
// Description:	 This function adds a partial derivative to the code buffer.
void CodeBuffer::AddPartial (UINT i, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddPartial\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "PD_%s_%u", szUCPrefix, i + 1);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddFunction
// Scope:	 Public
// Description:	 This function adds a function to the code buffer.
void CodeBuffer::AddFunction (char *psz, UINT iPD, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddFunction\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // de-alias function names
  char szFuncName[32];
  if	  (!strcmp(psz, "ln"))	strcpy(szFuncName, "log");
  else if (!strcmp(psz, "log")) strcpy(szFuncName, "log10");
  else if (!strcmp(psz, "min")) strcpy(szFuncName, "fmin");
  else if (!strcmp(psz, "max")) strcpy(szFuncName, "fmax");
  else				strcpy(szFuncName, psz);

  char sz[32];
  if (iPD) sprintf(sz, "pd%u%s", iPD, szFuncName);
  else strcpy(sz, szFuncName);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddFunction
// Scope:	 Public
// Description:	 This function adds a function to the code buffer.
void CodeBuffer::AddFunction (UINT i, UINT iPD, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddFunction\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  SymFunc *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_FUNC, i), SymFunc);
  AddFunction(psym->szName, iPD, fAddSpace);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddUDFunction
// Scope:	 Public
// Description:	 This function adds a user-defined function to the code buffer.
void CodeBuffer::AddUDFunction (char *psz, UINT iPD, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddUDFunction\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "UDF%s_%u_%s", szUCPrefix, iPD, psz);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddUDFunction
// Scope:	 Public
// Description:	 This function adds a user-defined function to the code buffer.
void CodeBuffer::AddUDFunction (UINT i, UINT iPD, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddUDFunction\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  SymUDFunc *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_UDFUNC, i), SymUDFunc);
  AddUDFunction(psym->szName, iPD, fAddSpace);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddPrecUDFunc
// Scope:	 Public
// Description:	 This function adds a precomputed user-defined function to the
//		 code buffer.
void CodeBuffer::AddPrecUDFunc (char *psz, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddPrecUDFunc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "pre%s._%s", szLCPrefix, psz);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddPrecUDFunc
// Scope:	 Public
// Description:	 This function adds a precomputed user-defined function to the
//		 code buffer.
void CodeBuffer::AddPrecUDFunc (UINT i, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddPrecUDFunc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  SymUDFunc *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_UDFUNC, i), SymUDFunc);
  AddPrecUDFunc(psym->szName, fAddSpace);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddProcedure
// Scope:	 Public
// Description:	 This function adds a procedure to the code buffer.
void CodeBuffer::AddProcedure (char *psz, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddProcedure\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "PROC%s_%s()", szUCPrefix, psz);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddProcedure
// Scope:	 Public
// Description:	 This function adds a procedure to the code buffer.
void CodeBuffer::AddProcedure (UINT i, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddProcedure\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  SymProc *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_PROC, i), SymProc);
  AddProcedure(psym->szName, fAddSpace);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddLabel
// Scope:	 Public
// Description:	 This function adds a label to the code buffer.
void CodeBuffer::AddLabel (UINT i, BOOL fColon, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddLabel\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  if (fColon) sprintf(sz, "LABEL%d:", i);
  else	      sprintf(sz, "LABEL%d",  i);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddText
// Scope:	 Public
// Description:	 This function adds text to the code buffer.  This function
//		 replaces the substrings @U@ and @L@ with the uppercase and
//		 lowercase prefix.
void CodeBuffer::AddText ( const char *psz )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddText\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  strcpy(pszGlobalBuf, psz);
  char *psz_n = pszGlobalBuf;
  while ((psz_n = strstr(psz_n, "@U@")) != NULL)
  { psz_n[0] = szUCPrefix[0];
    psz_n[1] = szUCPrefix[1];
    psz_n[2] = szUCPrefix[2];
  }

  psz_n = pszGlobalBuf;
  while ((psz_n = strstr(psz_n, "@L@")) != NULL)
  { psz_n[0] = szLCPrefix[0];
    psz_n[1] = szLCPrefix[1];
    psz_n[2] = szLCPrefix[2];
  }

  char *pszH, *pszT;
  for(pszT = pszH = pszGlobalBuf; *pszT; ++pszT)
  { char ch = *pszT;
    if ((ch == ' ') || (ch == '\n'))
    { *pszT = '\0';
      AddToken(pszH);
      AddChar(ch);
      pszH = pszT + 1;
    }
  }
  if (*pszH) AddToken(pszH);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddString
// Scope:	 Public
// Description:	 This function adds a string to the code buffer.
void CodeBuffer::AddString (char *psz)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddString\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT cLen = strlen(psz);
  UINT cCopied = 0;
  BOOL fDone = FALSE;
  while (!fDone)
  { UINT c = iRightMargin - iBuffer - 2;
    if (c > cLen - cCopied) c = cLen - cCopied;
    if (c > 0)
    { szBuffer[iBuffer++] = '\"';
      strncpy(szBuffer + iBuffer, psz + cCopied, c);
      iBuffer += c;
      cCopied += c;
      szBuffer[iBuffer++] = '\"';
    }
    if (cCopied < cLen) AddChar('\n');
    else fDone = TRUE;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddSum
// Scope:	 Public
// Description:	 This function adds a node summation to the code buffer.
void CodeBuffer::AddSum (UINT i, BOOL fAddSpace)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("codebuf.cpp CodeBuffer::AddSum\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char sz[32];
  sprintf(sz, "_sum_%u", i);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}



