/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     grammar.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/14/94    KMK      initial implementation
 04/12/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version
 ??/??/??    JJF      implement integer data type

 DESCRIPTION:

 This module contains the SGFramework grammar table and the error tables.
 CFG (Context-Free Grammer) table, better be LL(1)

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "tokens.h"
#include "states.h"
#include "actions.h"
#include "grammar.h"


///////////////////////////////////////////////////////////////////////////////
/////////////////////////////  SGFRAMEWORK GRAMMAR  ///////////////////////////
///////////////////////////////////////////////////////////////////////////////


/*******************************************************************************

  prog	  :   // nothing
	  |   prog			';'
	  |   prog VAR	    vdecs	';'
	  |   prog CONST    cdecs	';'
	  |   prog EQU	    equ		';'
	  |   prog UNKNOWN  unknos	';'
	  |   prog KNOWN    knowns	';'
	  |   prog SET	    sets	';'
	  |   prog MESH	    string	';'
	  |   prog BEGIN    MAIN  main	END
	  |   prog EOF
	  ;

********************************************************************************

  expr	  :   lexpr
	  |   expr AND lexpr
	  |   expr OR  lexpr
	  |   expr XOR lexpr
	  ;

  lexpr	  :   aexpr
	  |   lexpr '>'	 aexpr
	  |   lexpr '<'	 aexpr
	  |   lexpr '>=' aexpr
	  |   lexpr '<=' aexpr
	  |   lexpr '==' aexpr
	  |   lexpr '<>' aexpr
	  ;

  aexpr	  :   mexpr
	  |   aexpr '+' mexpr
	  |   aexpr '-' mexpr
	  ;

  mexpr	  :   uexpr
	  |   mexpr '*' uexpr
	  |   mexpr '/' uexpr
	  ;

  uexpr	  :   NUM
	  |   INT
	  |   ID
	  |   ID '[' exprs ']'
	  |   ID '(' ')'
	  |   ID '(' exprs ')'
	  |   '+' uexpr
	  |   '-' uexpr
	  |   NOT uexpr
	  |   '(' expr ')'
	  |   '[' expr ']'
	  |   '{' expr '}'
	  |   REAL '(' expr ')'
	  |   AREA '(' expr ')'
	  |   AREA '(' expr ',' ID ')'
	  |   AREA '(' expr ',' expr ')'
	  |   AREA '(' expr ',' expr  ',' ID ')'
	  |   NODE '(' expr ',' expr ')'
	  |   NODE '(' expr ',' expr  ',' ID ')'
	  |   EDGE '(' expr ',' expr ')'
	  |   EDGE '(' expr ',' expr  ',' ID ')'
	  |   ELEN
	  |   ELEN '(' expr ',' expr ')'
	  |   ELEN '(' expr ',' expr  ',' ID ')'
	  |   ILEN '(' expr ',' expr ')'
	  |   ILEN '(' expr ',' expr  ',' ID ')'
	  |   NSUM '(' ID ',' ID ',' ID ',' expr ')'
	  ;

  exprs	  :   expr
	  |   exprs ',' expr
	  ;

********************************************************************************

  indices :   index
	  |   indices ',' index
	  ;

  index	  :   ID '=' ALL
	  |   ID '=' LRANGE '(' ID ')'
	  |   ID '=' expr
	  |   ID '=' expr '..' expr
	  |   ID '=' expr '..' expr ':' expr
	  ;

********************************************************************************

  vdecs	  :   vdec
	  |   vdecs ',' vdec
	  ;

  vdec	  :   ID
	  |   ID '[' exprs ']'
	  ;

  cdecs	  :   cdec
	  |   cdecs ',' cdec
	  ;

  cdec	  :   ID '=' expr
	  ;

********************************************************************************

  knowns  :   kitem
	  |   knowns ',' kitem
	  ;

  unknos  :   kitem
	  |   unknos ',' kitem
	  ;

  kitem	  :   ID
	  |   ID '[' ranges ']'
	  ;

********************************************************************************

  equ	  :   ID '->' expr '=' expr
	  |   ID '[' indices ']' '->' expr '=' expr
	  ;

********************************************************************************

  set	  :   NEWTON ITERS '=' INT
	  |   NEWTON ACC   '=' NUM
	  |   NEWTON DAMP  '=' INT
	  |   LINSOL ALGOR '=' SOR
	  |   LINSOL ALGOR '=' PCG
	  |   LINSOL ALGOR '=' GE
	  |   LINSOL SORT  '=' NONE
	  |   LINSOL FILL  '=' INT
	  |   LINSOL FILL  '=' INFIN
	  |   LINSOL PREC  '=' NONE
	  |   LINSOL PREC  '=' DIAG
	  |   LINSOL PREC  '=' ILU
	  |   LINSOL ITERS '=' INT
	  |   LINSOL ACC   '=' NUM
	  |   LINSOL OVER  '=' NUM
	  |   LINSOL ZERO  '=' YES
	  |   LINSOL ZERO  '=' NO
	  |   DIST   SCALE '=' NUM
	  ;

  sets	  :   set
	  |   sets ',' set
	  ;

********************************************************************************

  fmode	  :   READ
	  |   WRITE
	  |   APPEND
	  ;

  vlist	  :   ID
	  |   vlist ',' ID
	  ;

********************************************************************************

  arg	  :   ID
	  |   '<' ID '>'
	  ;

  args	  :   arg
	  |   args ',' arg
	  ;

  fstat	  :   CONST ID '=' expr ';'
	  |   ASSIGN ID '=' expr ';'
	  |   RETURN expr ';'

  fbody	  :   fstat
	  |   fbody ',' fstat
	  ;

  func	  :   FUNC ID '(' args ')' fbody ;

  precomp :   PRECOMP '@' NODE ID '->' ID '(' exprs ')'
	  |   PRECOMP '@' EDGE ID '(' ID ',' ID ')' ODD	 '->' ID '(' exprs ')'
	  |   PRECOMP '@' EDGE ID '(' ID ',' ID ')' EVEN '->' ID '(' exprs ')'
	  ;

********************************************************************************

  mstat	  :   ';'
	  |   WRITE ';'
	  |   SOLVE ';'
	  |   ASSIGN ID '=' expr ';'
	  |   ASSIGN ID '[' indices ']' '=' expr ';'
	  |   OPEN '"' string '"' fmode ';'
	  |   CLOSE ';'
	  |   FILE GOTO '"' string '"' ';'
	  |   FILE LABEL '"' string '"' ';'
	  |   FILE COMMENT '"' string '"' ';'
	  |   FILE READ vlist ';'
	  |   FILE WRITE vlist ';'
	  |   BEGIN mstats END
	  |   WHILE '(' expr ')' mstat
	  |   DO mstat WHILE '(' expr ')' ';'
	  |   IF '(' expr ')' mstat
	  |   IF '(' expr ')' mstat ELSE mstat
	  ;

  mstats  :   mstat
	  |   mstats mstat
	  ;

  mproc	  :   BEGIN MAIN mstats END

*******************************************************************************/


///////////////////////////////////////////////////////////////////////////////
//////////////////////////  SGFRAMEWORK GRAMMAR TABLE  ////////////////////////
///////////////////////////////////////////////////////////////////////////////


FSM afsmTable[] = {
//	    CURRENT	       INPUT	    NEXT	STACK	      USER
// INDEX     STATE	       TOKEN	    STATE	ACTION	     ACTION
// ------  ----------	    -----------	 -----------  -----------  -----------
/* 0x0000  ST_EXPR     */ { TOK_LEXPR  , ST_EXPR    , ACT_EXPR	 , USR_	      },
/* 0x0001  ST_EXPR     */ { TOK_EXPR   , ST_EXPR1   , ACT_PUSH	 , USR_	      },
/* 0x0002  ST_EXPR     */ { DEFAULT    , ST_LEXPR   , ACT_CALL	 , USR_	      },
/* 0x0003  ST_EXPR1    */ { TOK_AND    , ST_EXPR2   , ACT_PUSH	 , USR_	      },
/* 0x0004  ST_EXPR1    */ { TOK_OR     , ST_EXPR2   , ACT_PUSH	 , USR_	      },
/* 0x0005  ST_EXPR1    */ { TOK_XOR    , ST_EXPR2   , ACT_PUSH	 , USR_	      },
/* 0x0006  ST_EXPR1    */ { DEFAULT    , ST_RETURN  , ACT_POP	 , USR_	      },
/* 0x0007  ST_EXPR2    */ { TOK_LEXPR  , ST_EXPR    , ACT_EXPR	 , USR_BINOP1 },
/* 0x0008  ST_EXPR2    */ { DEFAULT    , ST_LEXPR   , ACT_CALL	 , USR_	      },
/* 0x0009  ST_LEXPR    */ { TOK_AEXPR  , ST_LEXPR   , ACT_LEXPR	 , USR_	      },
/* 0x000a  ST_LEXPR    */ { TOK_LEXPR  , ST_LEXPR1  , ACT_PUSH	 , USR_	      },
/* 0x000b  ST_LEXPR    */ { DEFAULT    , ST_AEXPR   , ACT_CALL	 , USR_	      },
/* 0x000c  ST_LEXPR1   */ { TOK_EQ     , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x000d  ST_LEXPR1   */ { TOK_NE     , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x000e  ST_LEXPR1   */ { '>'	       , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x000f  ST_LEXPR1   */ { '<'	       , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0010  ST_LEXPR1   */ { TOK_GE     , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0011  ST_LEXPR1   */ { TOK_LE     , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0012  ST_LEXPR1   */ { DEFAULT    , ST_RETURN  , ACT_POP	 , USR_	      },
/* 0x0013  ST_LEXPR2   */ { TOK_AEXPR  , ST_LEXPR   , ACT_LEXPR	 , USR_BINOP1 },
/* 0x0014  ST_LEXPR2   */ { DEFAULT    , ST_AEXPR   , ACT_CALL	 , USR_	      },
/* 0x0015  ST_AEXPR    */ { TOK_MEXPR  , ST_AEXPR   , ACT_AEXPR	 , USR_	      },
/* 0x0016  ST_AEXPR    */ { TOK_AEXPR  , ST_AEXPR1  , ACT_PUSH	 , USR_	      },
/* 0x0017  ST_AEXPR    */ { DEFAULT    , ST_MEXPR   , ACT_CALL	 , USR_	      },
/* 0x0018  ST_AEXPR1   */ { '+'	       , ST_AEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0019  ST_AEXPR1   */ { '-'	       , ST_AEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x001a  ST_AEXPR1   */ { DEFAULT    , ST_RETURN  , ACT_POP	 , USR_	      },
/* 0x001b  ST_AEXPR2   */ { TOK_MEXPR  , ST_AEXPR   , ACT_AEXPR	 , USR_BINOP1 },
/* 0x001c  ST_AEXPR2   */ { DEFAULT    , ST_MEXPR   , ACT_CALL	 , USR_	      },
/* 0x001d  ST_MEXPR    */ { TOK_UEXPR  , ST_MEXPR   , ACT_MEXPR	 , USR_	      },
/* 0x001e  ST_MEXPR    */ { TOK_MEXPR  , ST_MEXPR1  , ACT_PUSH	 , USR_	      },
/* 0x001f  ST_MEXPR    */ { DEFAULT    , ST_UEXPR   , ACT_CALL	 , USR_	      },
/* 0x0020  ST_MEXPR1   */ { '*'	       , ST_MEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0021  ST_MEXPR1   */ { '/'	       , ST_MEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0022  ST_MEXPR1   */ { DEFAULT    , ST_RETURN  , ACT_POP	 , USR_	      },
/* 0x0023  ST_MEXPR2   */ { TOK_UEXPR  , ST_MEXPR   , ACT_MEXPR	 , USR_BINOP1 },
/* 0x0024  ST_MEXPR2   */ { DEFAULT    , ST_UEXPR   , ACT_CALL	 , USR_	      },
/* 0x0025  ST_UEXPR    */ { TOK_NUM    , ST_RETURN  , ACT_UEXPR	 , USR_RCONST },
/* 0x0026  ST_UEXPR    */ { TOK_INT    , ST_RETURN  , ACT_UEXPR	 , USR_ICONST },
/* 0x0027  ST_UEXPR    */ { TOK_ID     , ST_UEXPR1  , ACT_PUSH	 , USR_	      },
/* 0x0028  ST_UEXPR    */ { '+'	       , ST_UEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0029  ST_UEXPR    */ { '-'	       , ST_UEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x002a  ST_UEXPR    */ { TOK_NOT    , ST_UEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x002b  ST_UEXPR    */ { '('	       , ST_UEXPR3  , ACT_DROP	 , USR_	      },
/* 0x002c  ST_UEXPR    */ { '['	       , ST_UEXPR4  , ACT_DROP	 , USR_	      },
/* 0x002d  ST_UEXPR    */ { '{'	       , ST_UEXPR5  , ACT_DROP	 , USR_	      },
/* 0x002e  ST_UEXPR    */ { TOK_REAL   , ST_UEXPR6  , ACT_PUSH	 , USR_	      },
/* 0x002f  ST_UEXPR    */ { TOK_AREA   , ST_UEXPR7  , ACT_PUSH	 , USR_	      },
/* 0x0030  ST_UEXPR    */ { TOK_AREA1  , ST_UEXPR8  , ACT_PUSH	 , USR_	      },
/* 0x0031  ST_UEXPR    */ { TOK_AREA2  , ST_UEXPR8  , ACT_PUSH	 , USR_	      },
/* 0x0032  ST_UEXPR    */ { TOK_NODE   , ST_UEXPR8  , ACT_PUSH	 , USR_	      },
/* 0x0033  ST_UEXPR    */ { TOK_EDGE   , ST_UEXPR8  , ACT_PUSH	 , USR_	      },
/* 0x0034  ST_UEXPR    */ { TOK_ILEN   , ST_UEXPR8  , ACT_PUSH	 , USR_	      },
/* 0x0035  ST_UEXPR    */ { TOK_ELEN   , ST_UEXPR9  , ACT_PUSH	 , USR_	      },
/* 0x0036  ST_UEXPR    */ { TOK_NSUM   , ST_UEXPR10 , ACT_PUSH	 , USR_	      },
/* 0x0037  ST_UEXPR    */ { TOK_TSUM   , ST_UEXPR10 , ACT_PUSH	 , USR_	      },
/* 0x0038  ST_UEXPR    */ { TOK_LSUM   , ST_UEXPR11 , ACT_PUSH	 , USR_	      },
/* 0x0039  ST_UEXPR    */ { '@'	       , ST_UEXPR12 , ACT_PUSH	 , USR_	      },
/* 0x003a  ST_UEXPR1   */ { '('	       , ST_UEXPR1B , ACT_DROP	 , USR_	      },
/* 0x003b  ST_UEXPR1   */ { '['	       , ST_UEXPR1D , ACT_DROP	 , USR_	      },
/* 0x003c  ST_UEXPR1   */ { DEFAULT    , ST_UEXPR1A , ACT_POP	 , USR_	      },
/* 0x003d  ST_UEXPR1A  */ { DEFAULT    , ST_RETURN  , ACT_UEXPR	 , USR_ID     },
/* 0x003e  ST_UEXPR1B  */ { TOK_EXPRS  , ST_UEXPR1C , ACT_PUSH	 , USR_	      },
/* 0x003f  ST_UEXPR1B  */ { DEFAULT    , ST_EXPRS   , ACT_CALL	 , USR_	      },
/* 0x0040  ST_UEXPR1C  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_FUNC   },
/* 0x0041  ST_UEXPR1D  */ { TOK_EXPRS  , ST_UEXPR1E , ACT_PUSH	 , USR_	      },
/* 0x0042  ST_UEXPR1D  */ { DEFAULT    , ST_EXPRS   , ACT_CALL	 , USR_	      },
/* 0x0043  ST_UEXPR1E  */ { ']'	       , ST_RETURN  , ACT_UEXPR	 , USR_ARRAY  },
/* 0x0044  ST_UEXPR2   */ { TOK_UEXPR  , ST_RETURN  , ACT_UEXPR	 , USR_UNOP1  },
/* 0x0045  ST_UEXPR2   */ { DEFAULT    , ST_UEXPR   , ACT_CALL	 , USR_	      },
/* 0x0046  ST_UEXPR3   */ { TOK_EXPR   , ST_UEXPR3A , ACT_PUSH	 , USR_	      },
/* 0x0047  ST_UEXPR3   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0048  ST_UEXPR3A  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_PAREN  },
/* 0x0049  ST_UEXPR4   */ { TOK_EXPR   , ST_UEXPR4A , ACT_PUSH	 , USR_	      },
/* 0x004a  ST_UEXPR4   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x004b  ST_UEXPR4A  */ { ']'	       , ST_RETURN  , ACT_UEXPR	 , USR_PAREN  },
/* 0x004c  ST_UEXPR5   */ { TOK_EXPR   , ST_UEXPR5A , ACT_PUSH	 , USR_	      },
/* 0x004d  ST_UEXPR5   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x004e  ST_UEXPR5A  */ { '}'	       , ST_RETURN  , ACT_UEXPR	 , USR_PAREN  },
/* 0x004f  ST_UEXPR6   */ { '('	       , ST_UEXPR6A , ACT_DROP	 , USR_	      },
/* 0x0050  ST_UEXPR6A  */ { TOK_EXPR   , ST_UEXPR6B , ACT_PUSH	 , USR_	      },
/* 0x0051  ST_UEXPR6A  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0052  ST_UEXPR6B  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_REAL   },
/* 0x0053  ST_UEXPR7   */ { '('	       , ST_UEXPR7A , ACT_DROP	 , USR_	      },
/* 0x0054  ST_UEXPR7A  */ { TOK_EXPR   , ST_UEXPR7B , ACT_PUSH	 , USR_	      },
/* 0x0055  ST_UEXPR7A  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0056  ST_UEXPR7B  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_UNOP2  },
/* 0x0057  ST_UEXPR7B  */ { ','	       , ST_UEXPR7C , ACT_DROP	 , USR_	      },
/* 0x0058  ST_UEXPR7C  */ { TOK_ID     , ST_UEXPR7D , ACT_PUSH	 , USR_	      },
/* 0x0059  ST_UEXPR7C  */ { TOK_ALL    , ST_UEXPR7D , ACT_PUSH	 , USR_	      },
/* 0x005a  ST_UEXPR7D  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_UNOP2A },
/* 0x005b  ST_UEXPR8   */ { '('	       , ST_UEXPR8A , ACT_DROP	 , USR_	      },
/* 0x005c  ST_UEXPR8A  */ { TOK_EXPR   , ST_UEXPR8B , ACT_PUSH	 , USR_	      },
/* 0x005d  ST_UEXPR8A  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x005e  ST_UEXPR8B  */ { ','	       , ST_UEXPR8C , ACT_DROP	 , USR_	      },
/* 0x005f  ST_UEXPR8C  */ { TOK_EXPR   , ST_UEXPR8D , ACT_PUSH	 , USR_	      },
/* 0x0060  ST_UEXPR8C  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0061  ST_UEXPR8D  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_BINOP2 },
/* 0x0062  ST_UEXPR8D  */ { ','	       , ST_UEXPR8E , ACT_DROP	 , USR_	      },
/* 0x0063  ST_UEXPR8E  */ { TOK_ID     , ST_UEXPR8F , ACT_PUSH	 , USR_	      },
/* 0x0064  ST_UEXPR8E  */ { TOK_ALL    , ST_UEXPR8F , ACT_PUSH	 , USR_	      },
/* 0x0065  ST_UEXPR8F  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_BINOP2A},
/* 0x0066  ST_UEXPR9   */ { '('	       , ST_UEXPR9A , ACT_DROP	 , USR_	      },
/* 0x0067  ST_UEXPR9   */ { DEFAULT    , ST_UEXPR9G , ACT_POP	 , USR_	      },
/* 0x0068  ST_UEXPR9A  */ { TOK_EXPR   , ST_UEXPR9B , ACT_PUSH	 , USR_	      },
/* 0x0069  ST_UEXPR9A  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x006a  ST_UEXPR9B  */ { ','	       , ST_UEXPR9C , ACT_DROP	 , USR_	      },
/* 0x006b  ST_UEXPR9C  */ { TOK_EXPR   , ST_UEXPR9D , ACT_PUSH	 , USR_	      },
/* 0x006c  ST_UEXPR9C  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x006d  ST_UEXPR9D  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_BINOP2 },
/* 0x006e  ST_UEXPR9D  */ { ','	       , ST_UEXPR9E , ACT_DROP	 , USR_	      },
/* 0x006f  ST_UEXPR9E  */ { TOK_ID     , ST_UEXPR9F , ACT_PUSH	 , USR_	      },
/* 0x0070  ST_UEXPR9E  */ { TOK_ALL    , ST_UEXPR9F , ACT_PUSH	 , USR_	      },
/* 0x0071  ST_UEXPR9F  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_BINOP2A},
/* 0x0072  ST_UEXPR9G  */ { DEFAULT    , ST_RETURN  , ACT_UEXPR	 , USR_ELEN   },
/* 0x0073  ST_UEXPR10  */ { '('	       , ST_UEXPR10A, ACT_DROP	 , USR_	      },
/* 0x0074  ST_UEXPR10A */ { TOK_EXPR   , ST_UEXPR10B, ACT_PUSH	 , USR_	      },
/* 0x0075  ST_UEXPR10A */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0076  ST_UEXPR10B */ { ','	       , ST_UEXPR10C, ACT_DROP	 , USR_	      },
/* 0x0077  ST_UEXPR10C */ { TOK_ID     , ST_UEXPR10D, ACT_PUSH	 , USR_DINDEX },
/* 0x0078  ST_UEXPR10D */ { ','	       , ST_UEXPR10E, ACT_DROP	 , USR_	      },
/* 0x0079  ST_UEXPR10E */ { TOK_ID     , ST_UEXPR10F, ACT_PUSH	 , USR_	      },
/* 0x007a  ST_UEXPR10E */ { TOK_ALL    , ST_UEXPR10F, ACT_PUSH	 , USR_	      },
/* 0x007b  ST_UEXPR10F */ { ','	       , ST_UEXPR10G, ACT_DROP	 , USR_	      },
/* 0x007c  ST_UEXPR10G */ { TOK_EXPR   , ST_UEXPR10H, ACT_PUSH	 , USR_	      },
/* 0x007d  ST_UEXPR10G */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x007e  ST_UEXPR10H */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_NSUM   },
/* 0x007f  ST_UEXPR11  */ { '('	       , ST_UEXPR11A, ACT_DROP	 , USR_	      },
/* 0x0080  ST_UEXPR11A */ { TOK_ID     , ST_UEXPR11B, ACT_PUSH	 , USR_DINDEX },
/* 0x0081  ST_UEXPR11B */ { ','	       , ST_UEXPR11C, ACT_DROP	 , USR_	      },
/* 0x0082  ST_UEXPR11C */ { TOK_ID     , ST_UEXPR11D, ACT_PUSH	 , USR_	      },
/* 0x0083  ST_UEXPR11D */ { ','	       , ST_UEXPR11E, ACT_DROP	 , USR_	      },
/* 0x0084  ST_UEXPR11E */ { TOK_EXPR   , ST_UEXPR11F, ACT_PUSH	 , USR_	      },
/* 0x0085  ST_UEXPR11E */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0086  ST_UEXPR11F */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_LSUM   },
/* 0x0087  ST_UEXPR12  */ { TOK_ID     , ST_UEXPR12A, ACT_PUSH	 , USR_	      },
/* 0x0088  ST_UEXPR12A */ { '('	       , ST_UEXPR12B, ACT_DROP	 , USR_	      },
/* 0x0089  ST_UEXPR12B */ { TOK_EXPRS  , ST_UEXPR12C, ACT_PUSH	 , USR_	      },
/* 0x008a  ST_UEXPR12B */ { DEFAULT    , ST_EXPRS   , ACT_CALL	 , USR_	      },
/* 0x008b  ST_UEXPR12C */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_REF    },
/* 0x008c  ST_EXPRS    */ { DEFAULT    , ST_EXPRS1  , ACT_	 , USR_EXPRS  },
/* 0x008d  ST_EXPRS1   */ { TOK_EXPR   , ST_EXPRS2  , ACT_PUSH	 , USR_ADDEXPR},
/* 0x008e  ST_EXPRS1   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x008f  ST_EXPRS2   */ { ','	       , ST_EXPRS1  , ACT_DROP	 , USR_	      },
/* 0x0090  ST_EXPRS2   */ { DEFAULT    , ST_EXPRS3  , ACT_POP	 , USR_	      },
/* 0x0091  ST_EXPRS3   */ { DEFAULT    , ST_RETURN  , ACT_EXPRS	 , USR_	      },
/* 0x0092  ST_VDECS    */ { TOK_ID     , ST_VDECS1  , ACT_PUSH	 , USR_	      },
/* 0x0093  ST_VDECS1   */ { ','	       , ST_VDECS   , ACT_DROP	 , USR_DVAR   },
/* 0x0094  ST_VDECS1   */ { ';'	       , ST_VDECS5  , ACT_DROP	 , USR_DVAR   },
/* 0x0095  ST_VDECS1   */ { '['	       , ST_VDECS2  , ACT_DROP	 , USR_	      },
/* 0x0096  ST_VDECS2   */ { TOK_EXPRS  , ST_VDECS3  , ACT_PUSH	 , USR_	      },
/* 0x0097  ST_VDECS2   */ { DEFAULT    , ST_EXPRS   , ACT_CALL	 , USR_EM1    },
/* 0x0098  ST_VDECS3   */ { ']'	       , ST_VDECS4  , ACT_DROP	 , USR_	      },
/* 0x0099  ST_VDECS4   */ { ','	       , ST_VDECS   , ACT_DROP	 , USR_DARRAY },
/* 0x009a  ST_VDECS4   */ { ';'	       , ST_VDECS5  , ACT_DROP	 , USR_DARRAY },
/* 0x009b  ST_VDECS5   */ { DEFAULT    , ST_PROG    , ACT_	 , USR_	      },
/* 0x009c  ST_CDECS    */ { TOK_ID     , ST_CDECS1  , ACT_PUSH	 , USR_	      },
/* 0x009d  ST_CDECS1   */ { '='	       , ST_CDECS2  , ACT_DROP	 , USR_	      },
/* 0x009e  ST_CDECS2   */ { TOK_EXPR   , ST_CDECS3  , ACT_PUSH	 , USR_DCONST },
/* 0x009f  ST_CDECS2   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x00a0  ST_CDECS3   */ { ','	       , ST_CDECS   , ACT_DROP	 , USR_	      },
/* 0x00a1  ST_CDECS3   */ { ';'	       , ST_CDECS4  , ACT_DROP	 , USR_	      },
/* 0x00a2  ST_CDECS4   */ { DEFAULT    , ST_PROG    , ACT_	 , USR_	      },
/* 0x00a3  ST_INDEX    */ { TOK_ID     , ST_INDEX1  , ACT_PUSH	 , USR_DINDEX },
/* 0x00a4  ST_INDEX1   */ { '='	       , ST_INDEX2  , ACT_DROP	 , USR_	      },
/* 0x00a5  ST_INDEX2   */ { TOK_ID     , ST_INDEX2A , ACT_	 , USR_LABTEST},
/* 0x00a6  ST_INDEX2   */ { DEFAULT    , ST_INDEX2B , ACT_	 , USR_	      },
/* 0x00a7  ST_INDEX2A  */ { TOK_LABEL  , ST_RETURN  , ACT_INDEX	 , USR_INDEXL },
/* 0x00a8  ST_INDEX2A  */ { DEFAULT    , ST_INDEX2B , ACT_	 , USR_	      },
/* 0x00a9  ST_INDEX2B  */ { TOK_ALL    , ST_RETURN  , ACT_INDEX	 , USR_INDEX1 },
/* 0x00aa  ST_INDEX2B  */ { TOK_EXPR   , ST_INDEX3  , ACT_PUSH	 , USR_	      },
/* 0x00ab  ST_INDEX2B  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x00ac  ST_INDEX3   */ { TOK_RANGE  , ST_INDEX4  , ACT_DROP	 , USR_	      },
/* 0x00ad  ST_INDEX3   */ { DEFAULT    , ST_INDEX7  , ACT_POP	 , USR_INDEX2 },
/* 0x00ae  ST_INDEX4   */ { TOK_EXPR   , ST_INDEX5  , ACT_PUSH	 , USR_	      },
/* 0x00af  ST_INDEX4   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x00b0  ST_INDEX5   */ { ':'	       , ST_INDEX6  , ACT_DROP	 , USR_	      },
/* 0x00b1  ST_INDEX5   */ { DEFAULT    , ST_INDEX7  , ACT_POP	 , USR_INDEX3 },
/* 0x00b2  ST_INDEX6   */ { TOK_EXPR   , ST_INDEX7  , ACT_	 , USR_INDEX4 },
/* 0x00b3  ST_INDEX6   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x00b4  ST_INDEX7   */ { DEFAULT    , ST_RETURN  , ACT_INDEX	 , USR_	      },
/* 0x00b5  ST_INDICES  */ { DEFAULT    , ST_INDICES1, ACT_	 , USR_INDICES},
/* 0x00b6  ST_INDICES1 */ { TOK_INDEX  , ST_INDICES2, ACT_PUSH	 , USR_ADDINDX},
/* 0x00b7  ST_INDICES1 */ { DEFAULT    , ST_INDEX   , ACT_CALL	 , USR_	      },
/* 0x00b8  ST_INDICES2 */ { ','	       , ST_INDICES1, ACT_DROP	 , USR_	      },
/* 0x00b9  ST_INDICES2 */ { DEFAULT    , ST_INDICES3, ACT_POP	 , USR_	      },
/* 0x00ba  ST_INDICES3 */ { DEFAULT    , ST_RETURN  , ACT_INDICES, USR_	      },
/* 0x00bb  ST_RANGE    */ { DEFAULT    , ST_INDEX2  , ACT_	 , USR_DRANGE },
/* 0x00bc  ST_RANGES   */ { DEFAULT    , ST_RANGES1 , ACT_	 , USR_INDICES},
/* 0x00bd  ST_RANGES1  */ { TOK_INDEX  , ST_RANGES2 , ACT_PUSH	 , USR_ADDINDX},
/* 0x00be  ST_RANGES1  */ { DEFAULT    , ST_RANGE   , ACT_CALL	 , USR_	      },
/* 0x00bf  ST_RANGES2  */ { ','	       , ST_RANGES1 , ACT_DROP	 , USR_	      },
/* 0x00c0  ST_RANGES2  */ { DEFAULT    , ST_RANGES3 , ACT_POP	 , USR_	      },
/* 0x00c1  ST_RANGES3  */ { DEFAULT    , ST_RETURN  , ACT_RANGES , USR_	      },
/* 0x00c2  ST_KNOWN    */ { TOK_ID     , ST_KNOWN1  , ACT_PUSH	 , USR_	      },
/* 0x00c3  ST_KNOWN1   */ { '['	       , ST_KNOWN2  , ACT_DROP	 , USR_	      },
/* 0x00c4  ST_KNOWN1   */ { ','	       , ST_KNOWN   , ACT_DROP	 , USR_KNOWN1 },
/* 0x00c5  ST_KNOWN1   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_KNOWN1 },
/* 0x00c6  ST_KNOWN2   */ { TOK_RANGES , ST_KNOWN3  , ACT_PUSH	 , USR_	      },
/* 0x00c7  ST_KNOWN2   */ { DEFAULT    , ST_RANGES  , ACT_CALL	 , USR_	      },
/* 0x00c8  ST_KNOWN3   */ { ']'	       , ST_KNOWN4  , ACT_DROP	 , USR_	      },
/* 0x00c9  ST_KNOWN4   */ { ','	       , ST_KNOWN   , ACT_DROP	 , USR_KNOWN2 },
/* 0x00ca  ST_KNOWN4   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_KNOWN2 },
/* 0x00cb  ST_UNKNO    */ { TOK_ID     , ST_UNKNO1  , ACT_PUSH	 , USR_	      },
/* 0x00cc  ST_UNKNO1   */ { '['	       , ST_UNKNO2  , ACT_DROP	 , USR_	      },
/* 0x00cd  ST_UNKNO1   */ { ','	       , ST_UNKNO   , ACT_DROP	 , USR_UNKNO1 },
/* 0x00ce  ST_UNKNO1   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_UNKNO1 },
/* 0x00cf  ST_UNKNO2   */ { TOK_RANGES , ST_UNKNO3  , ACT_PUSH	 , USR_	      },
/* 0x00d0  ST_UNKNO2   */ { DEFAULT    , ST_RANGES  , ACT_CALL	 , USR_	      },
/* 0x00d1  ST_UNKNO3   */ { ']'	       , ST_UNKNO4  , ACT_DROP	 , USR_	      },
/* 0x00d2  ST_UNKNO4   */ { ','	       , ST_UNKNO   , ACT_DROP	 , USR_UNKNO2 },
/* 0x00d3  ST_UNKNO4   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_UNKNO2 },
/* 0x00d4  ST_EQU      */ { TOK_ID     , ST_EQU1    , ACT_PUSH	 , USR_	      },
/* 0x00d5  ST_EQU1     */ { TOK_ARROW  , ST_EQU2    , ACT_DROP	 , USR_	      },
/* 0x00d6  ST_EQU1     */ { '['	       , ST_EQU5    , ACT_DROP	 , USR_	      },
/* 0x00d7  ST_EQU2     */ { TOK_EXPR   , ST_EQU3    , ACT_PUSH	 , USR_	      },
/* 0x00d8  ST_EQU2     */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x00d9  ST_EQU3     */ { '='	       , ST_EQU4    , ACT_DROP	 , USR_	      },
/* 0x00da  ST_EQU4     */ { TOK_EXPR   , ST_EQU11   , ACT_PUSH	 , USR_EQN1   },
/* 0x00db  ST_EQU4     */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x00dc  ST_EQU5     */ { TOK_INDICES, ST_EQU6    , ACT_PUSH	 , USR_	      },
/* 0x00dd  ST_EQU5     */ { DEFAULT    , ST_INDICES , ACT_CALL	 , USR_	      },
/* 0x00de  ST_EQU6     */ { ']'	       , ST_EQU7    , ACT_DROP	 , USR_	      },
/* 0x00df  ST_EQU7     */ { TOK_ARROW  , ST_EQU8    , ACT_DROP	 , USR_	      },
/* 0x00e0  ST_EQU8     */ { TOK_EXPR   , ST_EQU9    , ACT_PUSH	 , USR_	      },
/* 0x00e1  ST_EQU8     */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x00e2  ST_EQU9     */ { '='	       , ST_EQU10   , ACT_DROP	 , USR_	      },
/* 0x00e3  ST_EQU10    */ { TOK_EXPR   , ST_EQU11   , ACT_PUSH	 , USR_EQN2   },
/* 0x00e4  ST_EQU10    */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x00e5  ST_EQU11    */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_	      },
/* 0x00e6  ST_FARG     */ { '<'	       , ST_FARG1   , ACT_DROP	 , USR_	      },
/* 0x00e7  ST_FARG     */ { TOK_ID     , ST_RETURN  , ACT_FARG	 , USR_FARG1  },
/* 0x00e8  ST_FARG1    */ { TOK_ID     , ST_FARG2   , ACT_PUSH	 , USR_	      },
/* 0x00e9  ST_FARG2    */ { '>'	       , ST_RETURN  , ACT_FARG	 , USR_FARG2  },
/* 0x00ea  ST_FARGS    */ { TOK_FARG   , ST_FARGS1  , ACT_DROP	 , USR_	      },
/* 0x00eb  ST_FARGS    */ { DEFAULT    , ST_FARG    , ACT_CALL	 , USR_	      },
/* 0x00ec  ST_FARGS1   */ { ','	       , ST_FARGS   , ACT_DROP	 , USR_	      },
/* 0x00ed  ST_FARGS1   */ { DEFAULT    , ST_FARGS2  , ACT_NEW	 , USR_	      },
/* 0x00ee  ST_FARGS2   */ { DEFAULT    , ST_RETURN  , ACT_FARGS	 , USR_	      },
/* 0x00ef  ST_FUNC     */ { TOK_ID     , ST_FUNC1   , ACT_PUSH	 , USR_	      },
/* 0x00f0  ST_FUNC1    */ { '('	       , ST_FUNC2   , ACT_DROP	 , USR_	      },
/* 0x00f1  ST_FUNC2    */ { TOK_FARGS  , ST_FUNC3   , ACT_DROP	 , USR_	      },
/* 0x00f2  ST_FUNC2    */ { DEFAULT    , ST_FARGS   , ACT_CALL	 , USR_	      },
/* 0x00f3  ST_FUNC3    */ { ')'	       , ST_FBODY   , ACT_DROP	 , USR_DECFUNC},
/* 0x00f4  ST_FBODY    */ { TOK_CONST  , ST_FBODY1  , ACT_DROP	 , USR_	      },
/* 0x00f5  ST_FBODY    */ { TOK_ASSIGN , ST_FBODY2  , ACT_DROP	 , USR_	      },
/* 0x00f6  ST_FBODY    */ { TOK_RETURN , ST_FBODY3  , ACT_DROP	 , USR_	      },
/* 0x00f7  ST_FBODY1   */ { TOK_ID     , ST_FBODY1A , ACT_PUSH	 , USR_	      },
/* 0x00f8  ST_FBODY1A  */ { '='	       , ST_FBODY1B , ACT_DROP	 , USR_	      },
/* 0x00f9  ST_FBODY1B  */ { TOK_EXPR   , ST_FBODY1C , ACT_PUSH	 , USR_DLCONST},
/* 0x00fa  ST_FBODY1B  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x00fb  ST_FBODY1C  */ { ','	       , ST_FBODY1  , ACT_DROP	 , USR_	      },
/* 0x00fc  ST_FBODY1C  */ { ';'	       , ST_FBODY   , ACT_DROP	 , USR_	      },
/* 0x00fd  ST_FBODY2   */ { TOK_ID     , ST_FBODY2A , ACT_PUSH	 , USR_	      },
/* 0x00fe  ST_FBODY2A  */ { '='	       , ST_FBODY2B , ACT_DROP	 , USR_	      },
/* 0x00ff  ST_FBODY2B  */ { TOK_EXPR   , ST_FBODY2C , ACT_PUSH	 , USR_LASS   },
/* 0x0100  ST_FBODY2B  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x0101  ST_FBODY2C  */ { ';'	       , ST_FBODY   , ACT_DROP	 , USR_	      },
/* 0x0102  ST_FBODY3   */ { TOK_EXPR   , ST_FBODY3A , ACT_PUSH	 , USR_RETURN },
/* 0x0103  ST_FBODY3   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x0104  ST_FBODY3A  */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_	      },
/* 0x0105  ST_EXT      */ { TOK_ID     , ST_EXT1    , ACT_PUSH	 , USR_	      },
/* 0x0106  ST_EXT1     */ { '('	       , ST_EXT2    , ACT_DROP	 , USR_	      },
/* 0x0107  ST_EXT2     */ { TOK_FARGS  , ST_EXT3    , ACT_DROP	 , USR_	      },
/* 0x0108  ST_EXT2     */ { DEFAULT    , ST_FARGS   , ACT_CALL	 , USR_	      },
/* 0x0109  ST_EXT3     */ { ')'	       , ST_EXT4    , ACT_DROP	 , USR_	      },
/* 0x010a  ST_EXT4     */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_DECEXT },
/* 0x010b  ST_MSTAT    */ { ';'	       , ST_RETURN  , ACT_MSTAT	 , USR_	      },
/* 0x010c  ST_MSTAT    */ { TOK_SOLVE  , ST_MSTAT1  , ACT_PUSH	 , USR_SOLVE  },
/* 0x010d  ST_MSTAT    */ { TOK_WRITE  , ST_MSTAT1  , ACT_PUSH	 , USR_WRITE  },
/* 0x010e  ST_MSTAT    */ { TOK_EXIT   , ST_MSTAT1  , ACT_PUSH	 , USR_EXIT   },
/* 0x010f  ST_MSTAT    */ { TOK_ASSIGN , ST_MSTAT2  , ACT_DROP	 , USR_	      },
/* 0x0110  ST_MSTAT    */ { TOK_BEGIN  , ST_MSTAT3  , ACT_DROP	 , USR_	      },
/* 0x0111  ST_MSTAT    */ { TOK_WHILE  , ST_MSTAT4  , ACT_PUSH	 , USR_	      },
/* 0x0112  ST_MSTAT    */ { TOK_DO     , ST_MSTAT5  , ACT_DROP	 , USR_	      },
/* 0x0113  ST_MSTAT    */ { TOK_IF     , ST_MSTAT6  , ACT_PUSH	 , USR_	      },
/* 0x0114  ST_MSTAT    */ { TOK_OPEN   , ST_MSTAT7  , ACT_PUSH	 , USR_	      },
/* 0x0115  ST_MSTAT    */ { TOK_CLOSE  , ST_MSTAT1  , ACT_PUSH	 , USR_CLOSE  },
/* 0x0116  ST_MSTAT    */ { TOK_FILE   , ST_MSTAT8  , ACT_PUSH	 , USR_	      },
/* 0x0117  ST_MSTAT    */ { TOK_CALL   , ST_MSTAT9  , ACT_PUSH	 , USR_	      },
/* 0x0118  ST_MSTAT1   */ { ';'	       , ST_RETURN  , ACT_MSTAT	 , USR_	      },
/* 0x0119  ST_MSTAT2   */ { TOK_ID     , ST_MSTAT2A , ACT_PUSH	 , USR_	      },
/* 0x011a  ST_MSTAT2A  */ { '='	       , ST_MSTAT2B , ACT_DROP	 , USR_	      },
/* 0x011b  ST_MSTAT2A  */ { '['	       , ST_MSTAT2C , ACT_DROP	 , USR_	      },
/* 0x011c  ST_MSTAT2B  */ { TOK_EXPR   , ST_MSTAT2G , ACT_PUSH	 , USR_ASS1   },
/* 0x011d  ST_MSTAT2B  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x011e  ST_MSTAT2C  */ { TOK_INDICES, ST_MSTAT2D , ACT_PUSH	 , USR_	      },
/* 0x011f  ST_MSTAT2C  */ { DEFAULT    , ST_INDICES , ACT_CALL	 , USR_	      },
/* 0x0120  ST_MSTAT2D  */ { ']'	       , ST_MSTAT2E , ACT_DROP	 , USR_	      },
/* 0x0121  ST_MSTAT2E  */ { '='	       , ST_MSTAT2F , ACT_DROP	 , USR_	      },
/* 0x0122  ST_MSTAT2F  */ { TOK_EXPR   , ST_MSTAT2G , ACT_PUSH	 , USR_ASS2   },
/* 0x0123  ST_MSTAT2F  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x0124  ST_MSTAT2G  */ { ';'	       , ST_RETURN  , ACT_MSTAT	 , USR_	      },
/* 0x0125  ST_MSTAT3   */ { TOK_MSTATS , ST_RETURN  , ACT_MSTAT	 , USR_	      },
/* 0x0126  ST_MSTAT3   */ { DEFAULT    , ST_MSTATS  , ACT_CALL	 , USR_	      },
/* 0x0127  ST_MSTAT4   */ { '('	       , ST_MSTAT4A , ACT_DROP	 , USR_	      },
/* 0x0128  ST_MSTAT4A  */ { TOK_EXPR   , ST_MSTAT4B , ACT_PUSH	 , USR_	      },
/* 0x0129  ST_MSTAT4A  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x012a  ST_MSTAT4B  */ { ')'	       , ST_MSTAT4C , ACT_DROP	 , USR_	      },
/* 0x012b  ST_MSTAT4C  */ { TOK_MSTAT  , ST_RETURN  , ACT_MSTAT	 , USR_WHILE2 },
/* 0x012c  ST_MSTAT4C  */ { DEFAULT    , ST_MSTAT   , ACT_CALL	 , USR_WHILE1 },
/* 0x012d  ST_MSTAT5   */ { TOK_MSTAT  , ST_MSTAT5A , ACT_DROP	 , USR_	      },
/* 0x012e  ST_MSTAT5   */ { DEFAULT    , ST_MSTAT   , ACT_CALL	 , USR_DO1    },
/* 0x012f  ST_MSTAT5A  */ { TOK_WHILE  , ST_MSTAT5B , ACT_DROP	 , USR_	      },
/* 0x0130  ST_MSTAT5B  */ { '('	       , ST_MSTAT5C , ACT_DROP	 , USR_	      },
/* 0x0131  ST_MSTAT5C  */ { TOK_EXPR   , ST_MSTAT5D , ACT_PUSH	 , USR_	      },
/* 0x0132  ST_MSTAT5C  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x0133  ST_MSTAT5D  */ { ')'	       , ST_MSTAT5E , ACT_DROP	 , USR_	      },
/* 0x0134  ST_MSTAT5E  */ { ';'	       , ST_RETURN  , ACT_MSTAT	 , USR_DO2    },
/* 0x0135  ST_MSTAT6   */ { '('	       , ST_MSTAT6A , ACT_DROP	 , USR_	      },
/* 0x0136  ST_MSTAT6A  */ { TOK_EXPR   , ST_MSTAT6B , ACT_PUSH	 , USR_	      },
/* 0x0137  ST_MSTAT6A  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x0138  ST_MSTAT6B  */ { ')'	       , ST_MSTAT6C , ACT_DROP	 , USR_	      },
/* 0x0139  ST_MSTAT6C  */ { TOK_MSTAT  , ST_MSTAT6D , ACT_PUSH	 , USR_	      },
/* 0x013a  ST_MSTAT6C  */ { DEFAULT    , ST_MSTAT   , ACT_CALL	 , USR_IF1    },
/* 0x013b  ST_MSTAT6D  */ { TOK_ELSE   , ST_MSTAT6F , ACT_DROP	 , USR_	      },
/* 0x013c  ST_MSTAT6D  */ { DEFAULT    , ST_MSTAT6E , ACT_POP	 , USR_	      },
/* 0x013d  ST_MSTAT6E  */ { DEFAULT    , ST_RETURN  , ACT_MSTAT	 , USR_IF2    },
/* 0x013e  ST_MSTAT6F  */ { DEFAULT    , ST_MSTAT6G , ACT_REM1	 , USR_	      },
/* 0x013f  ST_MSTAT6G  */ { TOK_MSTAT  , ST_RETURN  , ACT_MSTAT	 , USR_IF4    },
/* 0x0140  ST_MSTAT6G  */ { DEFAULT    , ST_MSTAT   , ACT_CALL	 , USR_IF3    },
/* 0x0141  ST_MSTAT7   */ { TOK_STRING , ST_MSTAT7A , ACT_PUSH	 , USR_	      },
/* 0x0142  ST_MSTAT7A  */ { TOK_READ   , ST_MSTAT1  , ACT_PUSH	 , USR_OPEN   },
/* 0x0143  ST_MSTAT7A  */ { TOK_WRITE  , ST_MSTAT1  , ACT_PUSH	 , USR_OPEN   },
/* 0x0144  ST_MSTAT7A  */ { TOK_APPEND , ST_MSTAT1  , ACT_PUSH	 , USR_OPEN   },
/* 0x0145  ST_MSTAT8   */ { TOK_GOTO   , ST_MSTAT8A , ACT_DROP	 , USR_	      },
/* 0x0146  ST_MSTAT8   */ { TOK_LABEL  , ST_MSTAT8B , ACT_DROP	 , USR_	      },
/* 0x0147  ST_MSTAT8   */ { TOK_COMMENT, ST_MSTAT8C , ACT_DROP	 , USR_	      },
/* 0x0148  ST_MSTAT8   */ { TOK_READ   , ST_MSTAT8D , ACT_DROP	 , USR_VREAD  },
/* 0x0149  ST_MSTAT8   */ { TOK_WRITE  , ST_MSTAT8D , ACT_DROP	 , USR_VWRITE },
/* 0x014a  ST_MSTAT8A  */ { TOK_STRING , ST_MSTAT1  , ACT_PUSH	 , USR_GOTO   },
/* 0x014b  ST_MSTAT8B  */ { TOK_STRING , ST_MSTAT1  , ACT_PUSH	 , USR_LABEL  },
/* 0x014c  ST_MSTAT8C  */ { TOK_STRING , ST_MSTAT1  , ACT_PUSH	 , USR_COMMENT},
/* 0x014d  ST_MSTAT8D  */ { TOK_ID     , ST_MSTAT8E , ACT_PUSH	 , USR_IO     },
/* 0x014e  ST_MSTAT8E  */ { ','	       , ST_MSTAT8D , ACT_DROP	 , USR_	      },
/* 0x014f  ST_MSTAT8E  */ { ';'	       , ST_RETURN  , ACT_MSTAT	 , USR_	      },
/* 0x0150  ST_MSTAT9   */ { TOK_ID     , ST_MSTAT9A , ACT_PUSH	 , USR_CALL   },
/* 0x0151  ST_MSTAT9A  */ { ';'	       , ST_RETURN  , ACT_MSTAT	 , USR_	      },
/* 0x0152  ST_MSTATS   */ { TOK_MSTAT  , ST_MSTATS  , ACT_DROP	 , USR_	      },
/* 0x0153  ST_MSTATS   */ { TOK_END    , ST_RETURN  , ACT_MSTATS , USR_	      },
/* 0x0154  ST_MSTATS   */ { DEFAULT    , ST_MSTAT   , ACT_CALL	 , USR_	      },
/* 0x0155  ST_SETS     */ { TOK_NEWTON , ST_SETS1   , ACT_PUSH	 , USR_	      },
/* 0x0156  ST_SETS     */ { TOK_LINSOL , ST_SETS2   , ACT_PUSH	 , USR_	      },
/* 0x0157  ST_SETS     */ { TOK_DIST   , ST_SETS10  , ACT_PUSH	 , USR_	      },
/* 0x0158  ST_SETS1    */ { TOK_ITERS  , ST_SETS3   , ACT_PUSH	 , USR_	      },
/* 0x0159  ST_SETS1    */ { TOK_ACC    , ST_SETS4   , ACT_PUSH	 , USR_	      },
/* 0x015a  ST_SETS1    */ { TOK_DAMP   , ST_SETS3   , ACT_PUSH	 , USR_	      },
/* 0x015b  ST_SETS2    */ { TOK_ALGOR  , ST_SETS5   , ACT_PUSH	 , USR_	      },
/* 0x015c  ST_SETS2    */ { TOK_SORT   , ST_SETS6   , ACT_PUSH	 , USR_	      },
/* 0x015d  ST_SETS2    */ { TOK_FILL   , ST_SETS7   , ACT_PUSH	 , USR_	      },
/* 0x015e  ST_SETS2    */ { TOK_PREC   , ST_SETS8   , ACT_PUSH	 , USR_	      },
/* 0x015f  ST_SETS2    */ { TOK_ITERS  , ST_SETS3   , ACT_PUSH	 , USR_	      },
/* 0x0160  ST_SETS2    */ { TOK_ACC    , ST_SETS4   , ACT_PUSH	 , USR_	      },
/* 0x0161  ST_SETS2    */ { TOK_OVER   , ST_SETS4   , ACT_PUSH	 , USR_	      },
/* 0x0162  ST_SETS2    */ { TOK_ZERO   , ST_SETS9   , ACT_PUSH	 , USR_	      },
/* 0x0163  ST_SETS3    */ { '='	       , ST_SETS3A  , ACT_DROP	 , USR_	      },
/* 0x0164  ST_SETS3A   */ { TOK_INT    , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0165  ST_SETS4    */ { '='	       , ST_SETS4A  , ACT_DROP	 , USR_	      },
/* 0x0166  ST_SETS4A   */ { TOK_INT    , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0167  ST_SETS4A   */ { TOK_NUM    , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0168  ST_SETS5    */ { '='	       , ST_SETS5A  , ACT_DROP	 , USR_	      },
/* 0x0169  ST_SETS5A   */ { TOK_SOR    , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x016a  ST_SETS5A   */ { TOK_PCG    , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x016b  ST_SETS5A   */ { TOK_GAUSS  , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x016c  ST_SETS6    */ { '='	       , ST_SETS6A  , ACT_DROP	 , USR_	      },
/* 0x016d  ST_SETS6A   */ { TOK_NONE   , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x016e  ST_SETS7    */ { '='	       , ST_SETS7A  , ACT_DROP	 , USR_	      },
/* 0x016f  ST_SETS7A   */ { TOK_INT    , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0170  ST_SETS7A   */ { TOK_INFIN  , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0171  ST_SETS8    */ { '='	       , ST_SETS8A  , ACT_DROP	 , USR_	      },
/* 0x0172  ST_SETS8A   */ { TOK_NONE   , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0173  ST_SETS8A   */ { TOK_DIAG   , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0174  ST_SETS8A   */ { TOK_ILU    , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0175  ST_SETS9    */ { '='	       , ST_SETS9A  , ACT_DROP	 , USR_	      },
/* 0x0176  ST_SETS9A   */ { TOK_YES    , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0177  ST_SETS9A   */ { TOK_NO     , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x0178  ST_SETS10   */ { TOK_SCALE  , ST_SETS10A , ACT_PUSH	 , USR_	      },
/* 0x0179  ST_SETS10A  */ { '='	       , ST_SETS10B , ACT_DROP	 , USR_	      },
/* 0x017a  ST_SETS10B  */ { TOK_EXPR   , ST_SETS11  , ACT_PUSH	 , USR_SET    },
/* 0x017b  ST_SETS10B  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_EM1    },
/* 0x017c  ST_SETS11   */ { ','	       , ST_SETS    , ACT_DROP	 , USR_	      },
/* 0x017d  ST_SETS11   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_	      },
/* 0x017e  ST_MESH     */ { TOK_STRING , ST_MESH1   , ACT_PUSH	 , USR_	      },
/* 0x017f  ST_MESH1    */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_MESH   },
/* 0x0180  ST_PREC     */ { '@'	       , ST_PREC1   , ACT_DROP	 , USR_	      },
/* 0x0181  ST_PREC1    */ { TOK_NODE   , ST_PREC2   , ACT_DROP	 , USR_PCNODE },
/* 0x0182  ST_PREC1    */ { TOK_EDGE   , ST_PREC3   , ACT_DROP	 , USR_PCEDGE },
/* 0x0183  ST_PREC2    */ { TOK_ID     , ST_PREC4   , ACT_PUSH	 , USR_DINDEX },
/* 0x0184  ST_PREC3    */ { TOK_ID     , ST_PREC3A  , ACT_PUSH	 , USR_DINDEX },
/* 0x0185  ST_PREC3A   */ { '('	       , ST_PREC3B  , ACT_DROP	 , USR_	      },
/* 0x0186  ST_PREC3B   */ { TOK_ID     , ST_PREC3C  , ACT_PUSH	 , USR_DINDEX },
/* 0x0187  ST_PREC3C   */ { ','	       , ST_PREC3D  , ACT_DROP	 , USR_	      },
/* 0x0188  ST_PREC3D   */ { TOK_ID     , ST_PREC3E  , ACT_PUSH	 , USR_DINDEX },
/* 0x0189  ST_PREC3E   */ { ')'	       , ST_PREC3F  , ACT_DROP	 , USR_	      },
/* 0x018a  ST_PREC3F   */ { TOK_ODD    , ST_PREC4   , ACT_PUSH	 , USR_	      },
/* 0x018b  ST_PREC3F   */ { TOK_EVEN   , ST_PREC4   , ACT_PUSH	 , USR_	      },
/* 0x018c  ST_PREC4    */ { TOK_ARROW  , ST_PREC5   , ACT_DROP	 , USR_	      },
/* 0x018d  ST_PREC5    */ { TOK_ID     , ST_PREC6   , ACT_PUSH	 , USR_	      },
/* 0x018e  ST_PREC6    */ { '('	       , ST_PREC7   , ACT_DROP	 , USR_	      },
/* 0x018f  ST_PREC7    */ { TOK_EXPRS  , ST_PREC8   , ACT_PUSH	 , USR_	      },
/* 0x0190  ST_PREC7    */ { DEFAULT    , ST_EXPRS   , ACT_CALL	 , USR_EM2    },
/* 0x0191  ST_PREC8    */ { ')'	       , ST_PREC9   , ACT_DROP	 , USR_	      },
/* 0x0192  ST_PREC9    */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_PRECOMP},
/* 0x0193  ST_BEGIN    */ { TOK_MAIN   , ST_MPROC   , ACT_PUSH	 , USR_MAIN   },
/* 0x0194  ST_BEGIN    */ { TOK_ID     , ST_PROC    , ACT_PUSH	 , USR_PROC1  },
/* 0x0195  ST_MPROC    */ { TOK_MSTATS , ST_PROG    , ACT_DROP	 , USR_PROG   },
/* 0x0196  ST_MPROC    */ { DEFAULT    , ST_MSTATS  , ACT_CALL	 , USR_	      },
/* 0x0197  ST_PROC     */ { TOK_MSTATS , ST_PROG    , ACT_DROP	 , USR_PROC2  },
/* 0x0198  ST_PROC     */ { DEFAULT    , ST_MSTATS  , ACT_CALL	 , USR_	      },
/* 0x0199  ST_PROG     */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_	      },
/* 0x019a  ST_PROG     */ { TOK_VAR    , ST_VDECS   , ACT_DROP	 , USR_	      },
/* 0x019b  ST_PROG     */ { TOK_CONST  , ST_CDECS   , ACT_DROP	 , USR_	      },
/* 0x019c  ST_PROG     */ { TOK_KNOWN  , ST_KNOWN   , ACT_DROP	 , USR_	      },
/* 0x019d  ST_PROG     */ { TOK_UNKNOWN, ST_UNKNO   , ACT_DROP	 , USR_	      },
/* 0x019e  ST_PROG     */ { TOK_BEGIN  , ST_BEGIN   , ACT_DROP	 , USR_	      },
/* 0x019f  ST_PROG     */ { TOK_EQU    , ST_EQU	    , ACT_DROP	 , USR_	      },
/* 0x01a0  ST_PROG     */ { TOK_SET    , ST_SETS    , ACT_DROP	 , USR_	      },
/* 0x01a1  ST_PROG     */ { TOK_FUNC   , ST_FUNC    , ACT_DROP	 , USR_	      },
/* 0x01a2  ST_PROG     */ { TOK_EXT    , ST_EXT	    , ACT_DROP	 , USR_	      },
/* 0x01a3  ST_PROG     */ { TOK_MESH   , ST_MESH    , ACT_PUSH	 , USR_	      },
/* 0x01a4  ST_PROG     */ { TOK_PRECOMP, ST_PREC    , ACT_PUSH	 , USR_	      },
/* 0x01a5  ST_PROG     */ { TOK_EOF    , ST_RETURN  , ACT_	 , USR_	      },
/* 0x01a6  ST_ERROR    */ { TOK_EOF    , ST_RETURN  , ACT_	 , USR_	      },
/* 0x01a7  ST_START    */ { DEFAULT    , ST_PROG    , ACT_	 , USR_INIT   }
};

#if 0
/* 0x0000  ST_	       */ {	       , ST_	    , ACT_	 , USR_	      },
/* 0x0000  ST_	       */ {	       , ST_	    , ACT_	 , USR_	      },
/* 0x0000  ST_	       */ {	       , ST_	    , ACT_	 , USR_	      },
/* 0x0000  ST_	       */ {	       , ST_	    , ACT_	 , USR_	      },
#endif

///////////////////////////////////////////////////////////////////////////////
///////////////////////	 SGFRAMEWORK GRAMMAR TABLE INDEX  /////////////////////
///////////////////////////////////////////////////////////////////////////////


UINT ifsmTable[] = {
/* 0x0000 */  0x0000, 0x0003, 0x0007, 0x0009, 0x000c, 0x0013, 0x0015, 0x0018,
/* 0x0008 */  0x001b, 0x001d, 0x0020, 0x0023, 0x0025, 0x003a, 0x003d, 0x003e,
/* 0x0010 */  0x0040, 0x0041, 0x0043, 0x0044, 0x0046, 0x0048, 0x0049, 0x004b,
/* 0x0018 */  0x004c, 0x004e, 0x004f, 0x0050, 0x0052, 0x0053, 0x0054, 0x0056,
/* 0x0020 */  0x0058, 0x005a, 0x005b, 0x005c, 0x005e, 0x005f, 0x0061, 0x0063,
/* 0x0028 */  0x0065, 0x0066, 0x0068, 0x006a, 0x006b, 0x006d, 0x006f, 0x0071,
/* 0x0030 */  0x0072, 0x0073, 0x0074, 0x0076, 0x0077, 0x0078, 0x0079, 0x007b,
/* 0x0038 */  0x007c, 0x007e, 0x007f, 0x0080, 0x0081, 0x0082, 0x0083, 0x0084,
/* 0x0040 */  0x0086, 0x0087, 0x0088, 0x0089, 0x008b, 0x008c, 0x008d, 0x008f,
/* 0x0048 */  0x0091, 0x0092, 0x0093, 0x0096, 0x0098, 0x0099, 0x009b, 0x009c,
/* 0x0050 */  0x009d, 0x009e, 0x00a0, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a7,
/* 0x0058 */  0x00a9, 0x00ac, 0x00ae, 0x00b0, 0x00b2, 0x00b4, 0x00b5, 0x00b6,
/* 0x0060 */  0x00b8, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00bf, 0x00c1, 0x00c2,
/* 0x0068 */  0x00c3, 0x00c6, 0x00c8, 0x00c9, 0x00cb, 0x00cc, 0x00cf, 0x00d1,
/* 0x0070 */  0x00d2, 0x00d4, 0x00d5, 0x00d7, 0x00d9, 0x00da, 0x00dc, 0x00de,
/* 0x0078 */  0x00df, 0x00e0, 0x00e2, 0x00e3, 0x00e5, 0x00e6, 0x00e8, 0x00e9,
/* 0x0080 */  0x00ea, 0x00ec, 0x00ee, 0x00ef, 0x00f0, 0x00f1, 0x00f3, 0x00f4,
/* 0x0088 */  0x00f7, 0x00f8, 0x00f9, 0x00fb, 0x00fd, 0x00fe, 0x00ff, 0x0101,
/* 0x0090 */  0x0102, 0x0104, 0x0105, 0x0106, 0x0107, 0x0109, 0x010a, 0x010b,
/* 0x0098 */  0x0118, 0x0119, 0x011a, 0x011c, 0x011e, 0x0120, 0x0121, 0x0122,
/* 0x00a0 */  0x0124, 0x0125, 0x0127, 0x0128, 0x012a, 0x012b, 0x012d, 0x012f,
/* 0x00a8 */  0x0130, 0x0131, 0x0133, 0x0134, 0x0135, 0x0136, 0x0138, 0x0139,
/* 0x00b0 */  0x013b, 0x013d, 0x013e, 0x013f, 0x0141, 0x0142, 0x0145, 0x014a,
/* 0x00b8 */  0x014b, 0x014c, 0x014d, 0x014e, 0x0150, 0x0151, 0x0152, 0x0155,
/* 0x00c0 */  0x0158, 0x015b, 0x0163, 0x0164, 0x0165, 0x0166, 0x0168, 0x0169,
/* 0x00c8 */  0x016c, 0x016d, 0x016e, 0x016f, 0x0171, 0x0172, 0x0175, 0x0176,
/* 0x00d0 */  0x0178, 0x0179, 0x017a, 0x017c, 0x017e, 0x017f, 0x0180, 0x0181,
/* 0x00d8 */  0x0183, 0x0184, 0x0185, 0x0186, 0x0187, 0x0188, 0x0189, 0x018a,
/* 0x00e0 */  0x018c, 0x018d, 0x018e, 0x018f, 0x0191, 0x0192, 0x0193, 0x0195,
/* 0x00e8 */  0x0197, 0x0199, 0x01a6, 0x01a7, 0x01a8
};


///////////////////////////////////////////////////////////////////////////////
//////////  SGFRAMEWORK SYNTAX ERROR -- EXPECTED TOKEN(S) TABLE INDEX  ////////
///////////////////////////////////////////////////////////////////////////////


char TOKEXP_LPARENTH[] = "'('";
char TOKEXP_RPARENTH[] = "')'";
char TOKEXP_RBRACKET[] = "']'";
char TOKEXP_RBRACE[]   = "'}'";
char TOKEXP_EQUAL[]    = "'='";
char TOKEXP_AT[]       = "'@'";
char TOKEXP_SEMI[]     = "';'";
char TOKEXP_GT[]       = "'>'";
char TOKEXP_COMMA[]    = "','";
char TOKEXP_ARROW[]    = "\"->\"";
char TOKEXP_TERM[]     = "',' or ';'";
char TOKEXP_TERMORBR[] = "'[' or ',' or ';'";
char TOKEXP_COMORRP[]  = "',' or ')'";
char TOKEXP_EQORLB[]   = "'[' or '='";
char TOKEXP_EQU1[]     = "'[' or \"->\"";
char TOKEXP_ID[]       = "identifier";
char TOKEXP_FARG[]     = "identifier or '<'";
char TOKEXP_LABEL[]    = "identifier or keyword \"all\"";
char TOKEXP_NODEEDGE[] = "keyword \"node\" or \"edge\"";
char TOKEXP_ODDEVEN[]  = "keyword \"odd\" or \"even\"";
char TOKEXP_WHILE[]    = "keyword \"while\"";
char TOKEXP_BEGIN[]    = "keyword \"main\"";
char TOKEXP_SETTYPE[]  = "keyword \"newton\", \"linsol\" or \"distance\"";
char TOKEXP_SCALE[]    = "keyword \"scale\"";
char TOKEXP_NEWTVAR[]  = "name of Newton algorithm variable";
char TOKEXP_LINVAR[]   = "name of linear solution algorithm variable";
char TOKEXP_INT[]      = "integer";
char TOKEXP_NUM[]      = "integer or real number";
char TOKEXP_LINSOL[]   = "name of linear solution algorithm";
char TOKEXP_SORT[]     = "name of sorting algorithm";
char TOKEXP_PRECOND[]  = "name of preconditioner";
char TOKEXP_FILL[]     = "integer or keyword \"infinity\"";
char TOKEXP_YESORNO[]  = "keyword \"yes\" or \"no\"";
char TOKEXP_FILEMODE[] = "keyword \"read\", \"write\" or \"append\"";
char TOKEXP_UEXPR[]    = "unary expression";
char TOKEXP_MAIN[]     = "procedure statement";
char TOKEXP_FILE[]     = "file I/O command";
char TOKEXP_FBODY[]    = "function body statement";
char TOKEXP_PROG[]     = "program statement";
char TOKEXP_STRING[]   = "string";


SEE aseeTable[] =
{ { ST_UEXPR,	  TOKEXP_UEXPR	  },
  { ST_UEXPR1C,	  TOKEXP_RPARENTH },
  { ST_UEXPR1E,	  TOKEXP_RBRACKET },
  { ST_UEXPR3A,	  TOKEXP_RPARENTH },
  { ST_UEXPR4A,	  TOKEXP_RBRACKET },
  { ST_UEXPR5A,	  TOKEXP_RBRACE	  },
  { ST_UEXPR6,	  TOKEXP_LPARENTH },
  { ST_UEXPR6B,	  TOKEXP_RPARENTH },
  { ST_UEXPR7,	  TOKEXP_LPARENTH },
  { ST_UEXPR7B,	  TOKEXP_COMORRP  },
  { ST_UEXPR7C,	  TOKEXP_LABEL	  },
  { ST_UEXPR7D,	  TOKEXP_RPARENTH },
  { ST_UEXPR8,	  TOKEXP_LPARENTH },
  { ST_UEXPR8B,	  TOKEXP_COMMA	  },
  { ST_UEXPR8D,	  TOKEXP_COMORRP  },
  { ST_UEXPR8E,	  TOKEXP_LABEL	  },
  { ST_UEXPR8F,	  TOKEXP_RPARENTH },
  { ST_UEXPR9B,	  TOKEXP_COMMA	  },
  { ST_UEXPR9D,	  TOKEXP_COMORRP  },
  { ST_UEXPR9E,	  TOKEXP_LABEL	  },
  { ST_UEXPR9F,	  TOKEXP_RPARENTH },
  { ST_UEXPR10,	  TOKEXP_LPARENTH },
  { ST_UEXPR10B,  TOKEXP_COMMA	  },
  { ST_UEXPR10C,  TOKEXP_ID	  },
  { ST_UEXPR10D,  TOKEXP_COMMA	  },
  { ST_UEXPR10E,  TOKEXP_LABEL	  },
  { ST_UEXPR10F,  TOKEXP_COMMA	  },
  { ST_UEXPR10H,  TOKEXP_RPARENTH },
  { ST_UEXPR11,	  TOKEXP_LPARENTH },
  { ST_UEXPR11A,  TOKEXP_ID	  },
  { ST_UEXPR11B,  TOKEXP_COMMA	  },
  { ST_UEXPR11C,  TOKEXP_ID	  },
  { ST_UEXPR11D,  TOKEXP_COMMA	  },
  { ST_UEXPR11F,  TOKEXP_RPARENTH },
  { ST_UEXPR12,	  TOKEXP_ID	  },
  { ST_UEXPR12A,  TOKEXP_LPARENTH },
  { ST_UEXPR12C,  TOKEXP_RPARENTH },
  { ST_VDECS,	  TOKEXP_ID	  },
  { ST_VDECS1,	  TOKEXP_TERMORBR },
  { ST_VDECS3,	  TOKEXP_RBRACKET },
  { ST_VDECS4,	  TOKEXP_TERM	  },
  { ST_CDECS,	  TOKEXP_ID	  },
  { ST_CDECS1,	  TOKEXP_EQUAL	  },
  { ST_CDECS3,	  TOKEXP_TERM	  },
  { ST_INDEX,	  TOKEXP_ID	  },
  { ST_INDEX1,	  TOKEXP_EQUAL	  },
  { ST_KNOWN,	  TOKEXP_ID	  },
  { ST_KNOWN1,	  TOKEXP_TERMORBR },
  { ST_KNOWN3,	  TOKEXP_RBRACKET },
  { ST_KNOWN4,	  TOKEXP_TERM	  },
  { ST_UNKNO,	  TOKEXP_ID	  },
  { ST_UNKNO1,	  TOKEXP_TERMORBR },
  { ST_UNKNO3,	  TOKEXP_RBRACKET },
  { ST_UNKNO4,	  TOKEXP_TERM	  },
  { ST_EQU,	  TOKEXP_ID	  },
  { ST_EQU1,	  TOKEXP_EQU1	  },
  { ST_EQU3,	  TOKEXP_EQUAL	  },
  { ST_EQU6,	  TOKEXP_RBRACKET },
  { ST_EQU7,	  TOKEXP_ARROW	  },
  { ST_EQU9,	  TOKEXP_EQUAL	  },
  { ST_EQU11,	  TOKEXP_SEMI	  },
  { ST_FARG,	  TOKEXP_FARG	  },
  { ST_FARG1,	  TOKEXP_ID	  },
  { ST_FARG2,	  TOKEXP_GT	  },
  { ST_FUNC,	  TOKEXP_ID	  },
  { ST_FUNC1,	  TOKEXP_LPARENTH },
  { ST_FUNC3,	  TOKEXP_RPARENTH },
  { ST_FBODY,	  TOKEXP_FBODY	  },
  { ST_FBODY1,	  TOKEXP_ID	  },
  { ST_FBODY1A,	  TOKEXP_EQUAL	  },
  { ST_FBODY1C,	  TOKEXP_TERM	  },
  { ST_FBODY2,	  TOKEXP_ID	  },
  { ST_FBODY2A,	  TOKEXP_EQUAL	  },
  { ST_FBODY2C,	  TOKEXP_SEMI	  },
  { ST_FBODY3A,	  TOKEXP_SEMI	  },
  { ST_EXT,	  TOKEXP_ID	  },
  { ST_EXT1,	  TOKEXP_LPARENTH },
  { ST_EXT3,	  TOKEXP_RPARENTH },
  { ST_EXT4,	  TOKEXP_SEMI	  },
  { ST_MSTAT,	  TOKEXP_MAIN	  },
  { ST_MSTAT1,	  TOKEXP_SEMI	  },
  { ST_MSTAT2,	  TOKEXP_ID	  },
  { ST_MSTAT2A,	  TOKEXP_EQORLB	  },
  { ST_MSTAT2D,	  TOKEXP_RBRACKET },
  { ST_MSTAT2E,	  TOKEXP_EQUAL	  },
  { ST_MSTAT2G,	  TOKEXP_SEMI	  },
  { ST_MSTAT4,	  TOKEXP_LPARENTH },
  { ST_MSTAT4B,	  TOKEXP_RPARENTH },
  { ST_MSTAT5A,	  TOKEXP_WHILE	  },
  { ST_MSTAT5B,	  TOKEXP_LPARENTH },
  { ST_MSTAT5D,	  TOKEXP_RPARENTH },
  { ST_MSTAT5E,	  TOKEXP_SEMI	  },
  { ST_MSTAT6,	  TOKEXP_LPARENTH },
  { ST_MSTAT6B,	  TOKEXP_RPARENTH },
  { ST_MSTAT7,	  TOKEXP_STRING	  },
  { ST_MSTAT7A,	  TOKEXP_FILEMODE },
  { ST_MSTAT8,	  TOKEXP_FILE	  },
  { ST_MSTAT8A,	  TOKEXP_STRING	  },
  { ST_MSTAT8B,	  TOKEXP_STRING	  },
  { ST_MSTAT8C,	  TOKEXP_STRING	  },
  { ST_MSTAT8D,	  TOKEXP_ID	  },
  { ST_MSTAT8E,	  TOKEXP_TERM	  },
  { ST_MSTAT9,	  TOKEXP_ID	  },
  { ST_MSTAT9A,	  TOKEXP_SEMI	  },
  { ST_SETS,	  TOKEXP_SETTYPE  },
  { ST_SETS1,	  TOKEXP_NEWTVAR  },
  { ST_SETS2,	  TOKEXP_LINVAR	  },
  { ST_SETS3,	  TOKEXP_EQUAL	  },
  { ST_SETS3A,	  TOKEXP_INT	  },
  { ST_SETS4,	  TOKEXP_EQUAL	  },
  { ST_SETS4A,	  TOKEXP_NUM	  },
  { ST_SETS5,	  TOKEXP_EQUAL	  },
  { ST_SETS5A,	  TOKEXP_LINSOL	  },
  { ST_SETS6,	  TOKEXP_EQUAL	  },
  { ST_SETS6A,	  TOKEXP_SORT	  },
  { ST_SETS7,	  TOKEXP_EQUAL	  },
  { ST_SETS7A,	  TOKEXP_FILL	  },
  { ST_SETS8,	  TOKEXP_EQUAL	  },
  { ST_SETS8A,	  TOKEXP_PRECOND  },
  { ST_SETS9,	  TOKEXP_EQUAL	  },
  { ST_SETS9A,	  TOKEXP_YESORNO  },
  { ST_SETS10,	  TOKEXP_SCALE	  },
  { ST_SETS10A,	  TOKEXP_EQUAL	  },
  { ST_SETS11,	  TOKEXP_TERM	  },
  { ST_MESH,	  TOKEXP_STRING	  },
  { ST_MESH1,	  TOKEXP_SEMI	  },
  { ST_PREC,	  TOKEXP_AT	  },
  { ST_PREC1,	  TOKEXP_NODEEDGE },
  { ST_PREC2,	  TOKEXP_ID	  },
  { ST_PREC3,	  TOKEXP_ID	  },
  { ST_PREC3A,	  TOKEXP_LPARENTH },
  { ST_PREC3B,	  TOKEXP_ID	  },
  { ST_PREC3C,	  TOKEXP_COMMA	  },
  { ST_PREC3D,	  TOKEXP_ID	  },
  { ST_PREC3E,	  TOKEXP_RPARENTH },
  { ST_PREC3F,	  TOKEXP_ODDEVEN  },
  { ST_PREC4,	  TOKEXP_ARROW	  },
  { ST_PREC5,	  TOKEXP_ID	  },
  { ST_PREC6,	  TOKEXP_LPARENTH },
  { ST_PREC8,	  TOKEXP_RPARENTH },
  { ST_PREC9,	  TOKEXP_SEMI	  },
  { ST_BEGIN,	  TOKEXP_BEGIN	  },
  { ST_PROG,	  TOKEXP_PROG	  },
  { -1u,	  NULL		  }
};


