/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     mathnode.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/19/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the SGFramework expression tree consisting of a n-ary
 tree of math nodes.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#if !defined(UNIX)
#include <mem.h>
#endif

#include "types.h"
#include "error.h"
#include "symbol.h"
#include "mathnode.h"
#include "statemnt.h"


///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE MATH NODE CLASS  //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Description:	 Default math node constructor
MathNode::MathNode ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uNode	   = -1u;
  uExpr	   =  0u;
  iPtrNode = -1u;
  iPtrExpr = -1u;
  c	   =  0u;
  apmn	   = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates an integer constant math node
MathNode::MathNode ( int n, UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uNode = TYPE_ICONST;
  uExpr = FLAG_CONST | FLAG_INT;
  iPtrNode = i;
  iPtrExpr = i;
  c = 0;
  apmn = NULL;
  data.n = n;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a real constant math node
MathNode::MathNode ( REAL r, UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uNode = TYPE_RCONST;
  uExpr = FLAG_CONST | FLAG_REAL;
  iPtrNode = i;
  iPtrExpr = i;
  c = 0;
  apmn = NULL;
  data.r = r;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates an index math node
MathNode::MathNode ( UINT iIndex, UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uNode = TYPE_INDEX;
  uExpr	 = FLAG_VAR | FLAG_INT;
  iPtrNode = i;
  iPtrExpr = i;
  c = 0;
  apmn = NULL;
  data.index.i = iIndex;
  data.index.n = -1;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a variable math node
MathNode::MathNode ( UINT iVar, UINT u, UINT i, BOOL f )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uNode	 = (f) ? TYPE_LVAR : TYPE_VAR;
  uExpr	 = FLAG_VAR | FLAG_REAL;
  iPtrNode = i;
  iPtrExpr = i;
  c = 0;
  apmn = NULL;
  if (f)
  { data.lvar.i = iVar;
    data.lvar.iDiff = u;
  }
  else
  { data.var.i = iVar;
    data.var.u = u;
    data.var.iT = 0;
    data.var.pmnNSUM = NULL;
    data.var.pmnLSUM = NULL;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates an argument math node
MathNode::MathNode ( UINT iArg, BOOL fMayBeUnknown, UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uNode = TYPE_ARG;
  uExpr = FLAG_VAR | FLAG_REAL;
  iPtrNode = i;
  iPtrExpr = i;
  c = 0;
  apmn = NULL;
  data.arg.i = iArg;
  data.arg.f = fMayBeUnknown;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a unary operator math node expression
MathNode::MathNode ( UINT uType, MathNode *pmn, UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // initialize the generic fields
  uExpr = pmn->uExpr;
  iPtrNode = i;
  iPtrExpr = i;
  c = 1;
  apmn = (MathNode **) calloc(1, sizeof(MathNode*));
  apmn[0] = pmn;

  switch (uNode = uType)
  { case TYPE_NEG:
      SimplifyNeg();
      break;

    case TYPE_NOT:
      SimplifyNot();
      break;

    case TYPE_REAL:
      uExpr &= ~FLAG_INT;
      uExpr |=	FLAG_REAL;
      SimplifyReal();
      break;

    case TYPE_AREA:
      uExpr &= ~FLAG_INT;
      uExpr |=	FLAG_REAL;
      uExpr &= ~FLAG_CONST;
      uExpr |=	FLAG_VAR;
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a binary operator math node expression
MathNode::MathNode ( UINT uType, MathNode *pmn1, MathNode *pmn2, UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // initialize the generic fields
  uNode = uType;
  UINT uExpr1 = pmn1->uExpr;
  UINT uExpr2 = pmn2->uExpr;
  BOOL fInt1  = uExpr1 & FLAG_INT;
  BOOL fInt2  = uExpr2 & FLAG_INT;
  uExpr	 = ((uExpr1 & FLAG_VAR) || (uExpr2 & FLAG_VAR)) ? FLAG_VAR : FLAG_CONST;
  uExpr |= (fInt1 && fInt2) ? FLAG_INT : FLAG_REAL;
  uExpr |= (uExpr1 & FLAG_NSUM) | (uExpr2 & FLAG_NSUM);
  uExpr |= (uExpr1 & FLAG_LSUM) | (uExpr2 & FLAG_LSUM);
  iPtrNode = i;
  iPtrExpr = pmn1->iPtrNode;
  c = 2;
  apmn = (MathNode**) calloc(2, sizeof(MathNode*));
  BOOL fReal = uExpr & FLAG_REAL;
  apmn[0] = (fReal && fInt1) ? new MathNode(TYPE_REAL, pmn1, i) : pmn1;
  apmn[1] = (fReal && fInt2) ? new MathNode(TYPE_REAL, pmn2, i) : pmn2;

  // operator specific code
  switch (uNode = uType)
  { case TYPE_ADD:
      SimplifyAdd();
      break;

    case TYPE_SUB:
      SimplifySub();
      break;

    case TYPE_MUL:
      SimplifyMul();
      break;

    case TYPE_DIV:
      SimplifyDiv();
      break;

    case TYPE_AND:
      SimplifyAnd();
      break;

    case TYPE_OR:
      SimplifyOr();
      break;

    case TYPE_XOR:
      SimplifyXor();
      break;

    case TYPE_GT:
      SimplifyGT();
      break;

    case TYPE_GE:
      SimplifyGE();
      break;

    case TYPE_LT:
      SimplifyLT();
      break;

    case TYPE_LE:
      SimplifyLE();
      break;

    case TYPE_EQ:
      SimplifyEQ();
      break;

    case TYPE_NE:
      SimplifyNE();
      break;

    case TYPE_AREA1:
    case TYPE_AREA2:
    case TYPE_ELEN:
    case TYPE_ILEN:
      uExpr &= ~(FLAG_INT  | FLAG_CONST);
      uExpr |=	(FLAG_REAL | FLAG_VAR  );
      break;

    case TYPE_NODE:
    case TYPE_EDGE:
      uExpr &= ~(FLAG_REAL | FLAG_CONST);
      uExpr |=	(FLAG_INT  | FLAG_VAR  );
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a node summation math node expression
MathNode::MathNode ( MathNode *pmnNode, UINT iIndex, UINT iLabel,
                     MathNode *pmnExpr, BOOL f, UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // initialize the generic fields
  uNode = TYPE_NSUM;
  uExpr = FLAG_VAR | FLAG_REAL | FLAG_NSUM;
  iPtrNode = i;
  iPtrExpr = i;
  c = 2;
  apmn = (MathNode**) calloc(2, sizeof(MathNode*));
  apmn[0] = pmnNode;
  apmn[1] = pmnExpr;
  data.nsum.fTriSum = f;
  data.nsum.iIndex  = iIndex;
  data.nsum.iLabel  = iLabel;
  data.nsum.iExpr   = -1u;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a label summation math node expression
MathNode::MathNode ( UINT iIndex, UINT iLabel, MathNode *pmnExpr, UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // initialize the generic fields
  uNode = TYPE_LSUM;
  uExpr = FLAG_VAR | FLAG_REAL | FLAG_LSUM;
  uExpr |= (pmnExpr->uExpr & FLAG_NSUM);
  iPtrNode = i;
  iPtrExpr = i;
  c = 1;
  apmn = (MathNode**) calloc(1, sizeof(MathNode*));
  apmn[0] = pmnExpr;
  data.lsum.iIndex  = iIndex;
  data.lsum.iLabel  = iLabel;
  data.lsum.iExpr   = -1u;
}


///////////////////////////////////////////////////////////////////////////////
MathNode::~MathNode ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::~MathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  for(i = 0; i < c; ++i) if (apmn[i]) delete apmn[i];
  if (apmn) free((void *) apmn);
}

///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds a child to the math node.
void MathNode::AddChild ( MathNode *pmn )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::AddChild\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i = c++;
  apmn = (MathNode **) realloc(apmn, c * sizeof(MathNode *));
  apmn[i] = pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function kills (deallocates) the children expression
//		 trees.
void MathNode::KillChildren ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::KillChildren\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  for(i = 0; i < c; ++i) if (apmn[i]) delete apmn[i];
  if (apmn) { free((void *) apmn); apmn = NULL; }
  c = 0;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function promotes the iHeir child to the parent node and
//		 deallocates the heir's siblings.
void MathNode::NameHeir ( UINT iHeir )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::NameHeir\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // delete every child except the heir
  UINT i;
  for(i = 0; i < c; ++i)
    if (i != iHeir)
      delete apmn[i];

  // replace the parent w/ the heir
  MathNode *pmn = apmn[iHeir];
  free((void *) apmn);
  *this = *pmn;

  // in order to delete the memory allocated for the heir, we need to change
  // the number of children it has so that the delete operator does not delete
  // the heir's information which is now stored in the parent node.
  pmn->c = 0;
  pmn->apmn = NULL;
  delete pmn;
}



///////////////////////////////////////////////////////////////////////////////
// Description:	 This function applies the following transformation to math
//		 expression trees.  If fLeft is TRUE then the left cycle is
//		 performed; otherwise, the right cycle is performed.
//
//	    A		     B		       A		B
//	   / \	   left	     |		      / \     right	|
//	  B   Y	   ---->     A	     OR	     X	 B    ---->	A
//	  |	   cycle    / \			 |    cycle    / \
//	  X		   X   Y		 Y	      X	  Y
//
// NOTE:  A and B are either operators or function math nodes.
//	  X and Y are arbitrary math trees.				      
void MathNode::Cycle ( BOOL fLeft )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::Cycle\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  MathNode *pmn1, *pmn2;
  MathNode *pmnX, *pmnY;
  char buf[sizeof(MathNode)];

  if (fLeft)
  { pmn1 = this;
    pmn2 = apmn[0];
    pmnX = apmn[0]->apmn[0];
    pmnY = apmn[1];
  }
  else
  { pmn1 = this;
    pmn2 = apmn[1];
    pmnX = apmn[0];
    pmnY = apmn[1]->apmn[0];
  }

  memcpy(buf,  pmn1, sizeof(MathNode)); // note, before this swap
  memcpy(pmn1, pmn2, sizeof(MathNode)); // pmn1 points to mnA and
  memcpy(pmn2, buf,  sizeof(MathNode)); // pmn2 points to mnB ...

  pmn1->apmn[0] = pmn2;			// now pmn1 points to mnB
  pmn2->apmn[0] = pmnX;			// and pmn2 points to mnA
  pmn2->apmn[1] = pmnY;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function swaps the i-th and j-th children
void MathNode::SwapChildren ( UINT i, UINT j )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::SwapChildren\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  MathNode *pmn;
  pmn = apmn[i];
  apmn[i] = apmn[j];
  apmn[j] = pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function compares the expression trees.
BOOL MathNode::Compare ( MathNode *pmn )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::Compare\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  if (uNode != pmn->uNode) return FALSE;
  switch (uNode)
  { case TYPE_INDEX:
	return data.n == pmn->data.n;

    case TYPE_RCONST:
      return data.r == pmn->data.r;

    case TYPE_ICONST:
      return data.n == pmn->data.n;

    case TYPE_ARG:
      return data.arg.i == pmn->data.arg.i;

    case TYPE_LVAR:
      return data.lvar.i == pmn->data.lvar.i;

    case TYPE_VAR:
      return data.var.i == pmn->data.var.i;

    case TYPE_ARRAY:
      if ((data.array.i != pmn->data.array.i) || (c != pmn->c)) return FALSE;
      if (!c) return data.array.uE == pmn->data.array.uE;
      for(i = 0; i < c; ++i)
	if (!apmn[i]->Compare(pmn->apmn[i]))
	  return FALSE;
      return TRUE;

    case TYPE_FUNC:
      if (data.func.iFunc != pmn->data.func.iFunc) return FALSE;
      if (data.func.iPD	  != pmn->data.func.iPD)   return FALSE;
      for(i = 0; i < c; ++i)
	if (!apmn[i]->Compare(pmn->apmn[i]))
	  return FALSE;
      return TRUE;

    case TYPE_UDFUNC:
      if (data.udfunc.iFunc != pmn->data.udfunc.iFunc) return FALSE;
      if (data.udfunc.iPD   != pmn->data.udfunc.iPD)   return FALSE;
      for(i = 0; i < c; ++i)
	if (!apmn[i]->Compare(pmn->apmn[i]))
	  return FALSE;
      return TRUE;

    case TYPE_NEG:
    case TYPE_NOT:
    case TYPE_REAL:
      return apmn[0]->Compare(pmn->apmn[0]);

    case TYPE_PRECOMP:
      return apmn[0]->Compare(pmn->apmn[0]);

    case TYPE_AREA:
      return apmn[0]->Compare(pmn->apmn[0]) && (data.n == pmn->data.n);

    case TYPE_ADD:
    case TYPE_MUL:
    case TYPE_AND:
    case TYPE_OR:
    case TYPE_XOR:
    case TYPE_EQ:
    case TYPE_NE:
      return
	(apmn[0]->Compare(pmn->apmn[0]) && apmn[1]->Compare(pmn->apmn[1])) ||
	(apmn[0]->Compare(pmn->apmn[1]) && apmn[1]->Compare(pmn->apmn[0]));

    case TYPE_SUB:
    case TYPE_DIV:
    case TYPE_GT:
    case TYPE_GE:
    case TYPE_LT:
    case TYPE_LE:
      return apmn[0]->Compare(pmn->apmn[0]) && apmn[1]->Compare(pmn->apmn[1]);

    case TYPE_AREA1:
    case TYPE_AREA2:
    case TYPE_NODE:
    case TYPE_EDGE:
    case TYPE_ELEN:
    case TYPE_ILEN:
      return apmn[0]->Compare(pmn->apmn[0]) &&
	     apmn[1]->Compare(pmn->apmn[1]) &&
	     (data.n == pmn->data.n);

    case TYPE_NSUM:
      return (data.nsum.iIndex == pmn->data.nsum.iIndex) &&
	     (data.nsum.iLabel == pmn->data.nsum.iLabel) &&
	      apmn[0]->Compare(pmn->apmn[0])		 &&
	      apmn[1]->Compare(pmn->apmn[1]);

    case TYPE_LSUM:
      return (data.lsum.iIndex == pmn->data.lsum.iIndex) &&
	     (data.lsum.iLabel == pmn->data.lsum.iLabel) &&
	      apmn[0]->Compare(pmn->apmn[0]);
  }
  return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function copies the expression tree.
MathNode *MathNode::Copy ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::Copy\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // make a new math node and copy the parent node
  MathNode *pmn = new MathNode();
  *pmn = *this;

  // if the parent node has children, then copy the children
  if (c)
  { pmn->apmn = new MathNode*[c];
    for(i = 0; i < c; ++i)
      pmn->apmn[i] = apmn[i]->Copy();
  }

  // return the copy
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function differentiates the expression tree.
MathNode *MathNode::Diff ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::Diff\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  MathNode *pmn, *pmn1, *pmn2;
  UINT i;

  switch (uNode)
  { case TYPE_INDEX:
    case TYPE_RCONST:
    case TYPE_ICONST:
      pmn = new MathNode(0.0, iPtrNode);
      break;

    case TYPE_VAR:
      if (TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_VAR, data.var.i), SymVar)->f)
      { pmn = Copy();
	pmn->uExpr |= FLAG_DIFF;
      }
      else pmn = new MathNode(0.0, iPtrNode);
      break;

    case TYPE_ARG:
    case TYPE_LVAR:
      pmn = Copy();
      pmn->uExpr |= FLAG_DIFF;
      break;

    case TYPE_ARRAY:
      if (TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARRAY, data.array.i), SymArray)->f)
      { pmn = Copy();
	pmn->uExpr |= FLAG_DIFF;
      }
      else pmn = new MathNode(0.0, iPtrNode);
      break;

    case TYPE_FUNC:
      pmn1 = apmn[0]->Diff();
      pmn2 = Copy();
      pmn2->data.func.iPD = 1;
      pmn = new MathNode(TYPE_MUL, pmn1, pmn2, iPtrNode);
      pmn->uExpr |= FLAG_ZEROCHECK;
      for(i = 1; i < c; ++i)
      { pmn1 = apmn[i]->Diff();
	pmn2 = Copy();
	pmn2->data.func.iPD = i + 1;
	pmn1 = new MathNode(TYPE_MUL, pmn1, pmn2, iPtrNode);
	pmn1->uExpr |= FLAG_ZEROCHECK;
	pmn  = new MathNode(TYPE_ADD, pmn,  pmn1, iPtrNode);
      }
      break;

    case TYPE_UDFUNC:
      pmn1 = apmn[0]->Diff();
      pmn2 = Copy();
      pmn2->data.udfunc.iPD = 1;
      pmn2->SimplifyFunc();
      pmn = new MathNode(TYPE_MUL, pmn1, pmn2, iPtrNode);
      pmn->uExpr |= FLAG_ZEROCHECK;
      for(i = 1; i < c; ++i)
      { pmn1 = apmn[i]->Diff();
	pmn2 = Copy();
	pmn2->data.udfunc.iPD = i + 1;
	pmn2->SimplifyFunc();
	pmn1 = new MathNode(TYPE_MUL, pmn1, pmn2, iPtrNode);
	pmn1->uExpr |= FLAG_ZEROCHECK;
	pmn  = new MathNode(TYPE_ADD, pmn,  pmn1, iPtrNode);
      }
      break;

    case TYPE_PRECOMP:
      pmn1 = apmn[0]->apmn[0]->Diff();
      pmn2 = Copy();
      pmn2->apmn[0]->data.func.iPD = 1;
      pmn2->SimplifyFunc();
      pmn = new MathNode(TYPE_MUL, pmn1, pmn2, iPtrNode);
      pmn->uExpr |= FLAG_ZEROCHECK;
      for(i = 1; i < apmn[0]->c; ++i)
      { pmn1 = apmn[0]->apmn[i]->Diff();
	pmn2 = Copy();
	pmn2->apmn[0]->data.func.iPD = i + 1;
	pmn2->SimplifyFunc();
	pmn1 = new MathNode(TYPE_MUL, pmn1, pmn2, iPtrNode);
	pmn1->uExpr |= FLAG_ZEROCHECK;
	pmn  = new MathNode(TYPE_ADD, pmn,  pmn1, iPtrNode);
      }
      break;

    case TYPE_ADD:
    case TYPE_SUB:
      pmn1 = apmn[0]->Diff();
      pmn2 = apmn[1]->Diff();
      pmn = new MathNode(uNode, pmn1, pmn2, iPtrNode);
      break;

    case TYPE_MUL:
      pmn1 = new MathNode(TYPE_MUL, apmn[0]->Diff(), apmn[1]->Copy(), iPtrNode);
      pmn2 = new MathNode(TYPE_MUL, apmn[0]->Copy(), apmn[1]->Diff(), iPtrNode);
      pmn  = new MathNode(TYPE_ADD, pmn1, pmn2, iPtrNode);
      break;

    case TYPE_DIV:
      pmn2 = new MathNode();
      pmn2->AddChild(apmn[1]->Copy());
      pmn2->MakeSq();
      pmn2->iPtrNode = pmn2->iPtrExpr = iPtrNode;
      pmn1 = new MathNode(TYPE_DIV, apmn[1]->Diff(), pmn2, iPtrNode);
      pmn2 = new MathNode(TYPE_MUL, apmn[0]->Copy(), pmn1, iPtrNode);
      pmn1 = new MathNode(TYPE_DIV, apmn[0]->Diff(), apmn[1]->Copy(), iPtrNode);
      pmn  = new MathNode(TYPE_SUB, pmn1, pmn2, iPtrNode);
      break;

    case TYPE_NEG:
    case TYPE_REAL:
      pmn = new MathNode(uNode, apmn[0]->Diff(), iPtrNode);
      break;

    case TYPE_NSUM:
//    pmn = new MathNode(apmn[0], data.nsum.iIndex, data.nsum.iLabel, apmn[1]->Diff(), iPtrNode);
      pmn = apmn[1]->Diff();
      break;

    case TYPE_LSUM:
      pmn = apmn[0]->Diff();
      break;

    // the derivatives for the logical and comparison operators evaluate to
    // zero except with a possible singularity ... therefore, the user should
    // exercise care in using these operators in equation statements.
    // furthermore, the derivatives of the AREA, AREA1, AREA2, NODE, EDGE,
    // ELEN and ILEN operators evaluate to zero.
    default:
      if (uExpr & FLAG_REAL) pmn = new MathNode(0.0, iPtrNode);
      else		     pmn = new MathNode(0,   iPtrNode);
  }
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function evaluates integer expressions.  The result is
//		 stored in the integer stack of the program class.
void MathNode::Evaluate ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::Evaluate\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  switch (uNode)
  { case TYPE_ICONST:
      pstatements->PushNum(data.n);
      break;

    case TYPE_INDEX:
      pstatements->PushIndex(data.index.i);
      break;

    case TYPE_ADD:
    case TYPE_SUB:
    case TYPE_MUL:
    case TYPE_DIV:
    case TYPE_AND:
    case TYPE_OR:
    case TYPE_XOR:
    case TYPE_GT:
    case TYPE_GE:
    case TYPE_LT:
    case TYPE_LE:
    case TYPE_EQ:
    case TYPE_NE:
      apmn[0]->Evaluate();
      apmn[1]->Evaluate();
      pstatements->Operator(uNode, iPtrNode);
      break;

    case TYPE_NODE:
    case TYPE_EDGE:
      apmn[0]->Evaluate();
      apmn[1]->Evaluate();
      pstatements->PushNum(data.n);
      pstatements->Operator(uNode, iPtrNode);
      break;

    case TYPE_NEG:
    case TYPE_NOT:
      apmn[0]->Evaluate();
      pstatements->Operator(uNode);
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function searches the math node expression for variables
//		 and arrays and stores them in the program class.
void MathNode::Variables ( BOOL fNoDiffExpr )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::Variables\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // is this node a variable?
  if	  (uNode == TYPE_VAR)
  { if (fNoDiffExpr) data.var.iT = -1u;
    pstatements->AddVar(this);
    return;
  }

  // is this node an array?
  else if (uNode == TYPE_ARRAY)
  { if (fNoDiffExpr) data.array.iT = -1u;
    pstatements->AddVar(this);
    return;
  }

  // search children for variables or arrays in user-defined functions ...
  // this routine will also marked those variables which are in argument
  // expressions that will not be differentiated.
  if (uNode == TYPE_UDFUNC)
  { for(i = 0; i < c; ++i)
      apmn[i]->Variables(fNoDiffExpr || !data.udfunc.af[i]);
    return;
  }

  // search children for variables or arrays
  for(i = 0; i < c; ++i) apmn[i]->Variables(fNoDiffExpr);
}


static UINT iDiffArg;
static UINT cLocal;
static UINT iLVarFlag;
static BOOL *afLocal;
static REAL *arLocal;


#define CHECK1	if (apmn[0]->uExpr & FLAG_CONST)	\
		{ uExpr &= ~FLAG_VAR;			\
		  uExpr |=  FLAG_CONST;			\
		}

#define CHECK2	if ((apmn[0]->uExpr & FLAG_CONST) &&	\
		    (apmn[1]->uExpr & FLAG_CONST))	\
		{ uExpr &= ~FLAG_VAR;			\
		  uExpr |=  FLAG_CONST;			\
		}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function simplifies derivatives of local assignment
//		 statements by evaluating the delta functions and
//		 calling the appropriate simplifying functions.
BOOL MathNode::SimplifyDiff ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::SimplifyDiff\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  BOOL fResult = FALSE;
  switch (uNode)
  { case TYPE_ARG:
      if (uExpr & FLAG_DIFF)
      { if (data.arg.i == iDiffArg) MakeOne();
	else MakeZero();
	fResult = TRUE;
      }
      break;

    case TYPE_LVAR:
      iLVarFlag = data.lvar.i;
      if (uExpr & FLAG_DIFF) iLVarFlag += (iDiffArg + 1) * cLocal;
      if (afLocal[iLVarFlag])
      { uNode = TYPE_RCONST;
	uExpr = FLAG_CONST | FLAG_REAL;
	data.r = arLocal[iLVarFlag];
	fResult = TRUE;
      }
      break;

    case TYPE_VAR:
      if (uExpr & FLAG_DIFF)
      { if (data.var.iT == iDiffArg) MakeOne();
	else MakeZero();
	fResult = TRUE;
      }
      break;

    case TYPE_ARRAY:
      if (uExpr & FLAG_DIFF)
      { if (data.array.iT == iDiffArg) MakeOne();
	else MakeZero();
	fResult = TRUE;
      }
      break;

    case TYPE_ADD:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2;  SimplifyAdd(); }
      break;

    case TYPE_SUB:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifySub(); }
      break;

    case TYPE_MUL:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyMul(); }
      break;

    case TYPE_DIV:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyDiv(); }
      break;

    case TYPE_AND:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyAnd(); }
      break;

    case TYPE_OR:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyOr(); }
      break;

    case TYPE_XOR:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyXor(); }
      break;

    case TYPE_GT:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyGT(); }
      break;

    case TYPE_GE:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyGE(); }
      break;

    case TYPE_LT:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyLT(); }
      break;

    case TYPE_LE:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyLE(); }
      break;

    case TYPE_EQ:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyEQ(); }
      break;

    case TYPE_NE:
      fResult |= apmn[0]->SimplifyDiff();
      fResult |= apmn[1]->SimplifyDiff();
      if (fResult) { CHECK2; SimplifyNE(); }
      break;

    case TYPE_NEG:
      fResult |= apmn[0]->SimplifyDiff();
      if (fResult) { CHECK1; SimplifyNeg(); }
      break;

    case TYPE_NOT:
      fResult |= apmn[0]->SimplifyDiff();
      if (fResult) { CHECK1; SimplifyNot(); }
      break;

    case TYPE_REAL:
      fResult |= apmn[0]->SimplifyDiff();
      if (fResult) { CHECK1; SimplifyReal(); }
      break;
  }

  return fResult;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function simplifies derivatives of local assignment
//		 statements by evaluating the delta functions and
//		 calling the appropriate simplifying functions.
void MathNode::SimplifyDiff ( UINT iArg, UINT c, BOOL *af, REAL *ar )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::SimplifyDiff\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  iDiffArg = iArg;
  cLocal = c;
  afLocal = af;
  arLocal = ar;
  SimplifyDiff();
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function simplifies derivatives of local assignment
//		 statements by evaluating the delta functions and
//		 calling the appropriate simplifying functions.
void MathNode::SimplifyDiff ( UINT iDiff )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::SimplifyDiff\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  iDiffArg = iDiff;
  SimplifyDiff();
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function searches the mathnode expression for local
//		 variables.
static UINT *afIsPresent;
void MathNode::FindLocalVars ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::FindLocalVars\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // is this node a local variable?
  if (uNode == TYPE_LVAR)
  { afIsPresent[data.lvar.i] |= (uExpr & FLAG_DIFF) ? LVAR_DER : LVAR_VAR;
    return;
  }

  // search children for variables or arrays
  for(i = 0; i < c; ++i) apmn[i]->FindLocalVars();
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function searches the mathnode expression for local
//		 variables.
void MathNode::FindLocalVars ( UINT *af )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::FindLocalVars\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  afIsPresent = af;
  FindLocalVars();
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function searches the mathnode expression for variables
//		 and array elements in node and label summation expressions and
//		 tags them.
static void *pmnLSUM = NULL;
static void *pmnNSUM = NULL;
static int iRegLabel;
void MathNode::TagSumVariables ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::TagNodeSumVariables\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  if (uExpr & FLAG_SUM)
  { if	    (uNode == TYPE_LSUM)
    { pmnLSUM = (void *) this;
      apmn[0]->TagLabelSum();
    }

    else if (uNode == TYPE_NSUM)
    { pmnNSUM = (void *) this;
      iRegLabel = (data.nsum.iLabel < cRegLabel) ? (int) data.nsum.iLabel : -1;
      apmn[1]->TagNodeSum();
      apmn[1]->InitRegLabel();
      return;
    }

    // search the expressions children
    for(i = 0; i < c; ++i) apmn[i]->TagSumVariables();
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function searches the mathnode expression for node,
//		 edge, and area operators and tags their labels if they
//		 were not defined.
void MathNode::InitRegLabel ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::InitRegLabel\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  switch (uNode)
  { case TYPE_AREA:
    case TYPE_AREA1:
    case TYPE_AREA2:
    case TYPE_NODE:
    case TYPE_EDGE:
    case TYPE_ELEN:
    case TYPE_ILEN:
      if (data.n == -2) data.n = iRegLabel;
      return;
  }

  // search the expressions children
  for(i = 0; i < c; ++i) apmn[i]->InitRegLabel();
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function searches the mathnode expression for arrays
//		 and variables and tags
void MathNode::TagNodeSum ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::TagNodeSum\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // tag the arrays
  if	  (uNode == TYPE_ARRAY) { data.array.pmnNSUM = pmnNSUM;	 return; }
  else if (uNode == TYPE_VAR)	{ data.var.pmnNSUM   = pmnNSUM;	 return; }

  // search the expressions children
  for(i = 0; i < c; ++i) apmn[i]->TagNodeSum();
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function searches the mathnode expression for arrays
//		 and variables and tags
void MathNode::TagLabelSum ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::TagLabelSum\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // tag the arrays
  if	  (uNode == TYPE_ARRAY) { data.array.pmnLSUM = pmnLSUM;	 return; }
  else if (uNode == TYPE_VAR)	{ data.var.pmnLSUM   = pmnLSUM;	 return; }

  // search the expressions children
  for(i = 0; i < c; ++i) apmn[i]->TagLabelSum();
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function copies the template and inserts the appropriate
//		 fields for the indices.
static MathNode *pmnTemplateIndices;
MathNode * MathNode::CopyTemplate ( MathNode *pmnIndices )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::CopyTemplate\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  pmnTemplateIndices = pmnIndices;
  return CopyTemplate();
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function copies the template and inserts the appropriate
//		 fields for the indices.
MathNode * MathNode::CopyTemplate ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::CopyTemplate\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // make a new math node and copy the parent node
  MathNode *pmn;
  if (uNode == TYPE_INDEX)
  { pmn = pmnTemplateIndices->apmn[data.index.i]->Copy();
  }
  else if ((uNode == TYPE_ELENABS) && (c == 0))
  { pmn = new MathNode();
    *pmn = *this;
//  pmn->c = 2;
//  pmn->apmn = new MathNode*[2];
//  pmn->apmn[0] = pmnTemplateIndices->apmn[1]->Copy();
//  pmn->apmn[1] = pmnTemplateIndices->apmn[2]->Copy();
    pmn->c = 1;
    pmn->apmn = new MathNode*[1];
    pmn->apmn[0] = pmnTemplateIndices->apmn[0]->Copy();
  }
  else
  { pmn = new MathNode();
    *pmn = *this;
  }

  // if the parent node has children, then copy the children
  if (c)
  { pmn->apmn = new MathNode*[c];
    for(i = 0; i < c; ++i)
      pmn->apmn[i] = apmn[i]->CopyTemplate();
  }

  // return the copy
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function searches the template mathnode expression for
//		 node indices and flips them.
void MathNode::FlipTemplateIndices ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::FlipTemplateIndices\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // tag the arrays
  if (uNode == TYPE_INDEX)
  { if	    (data.index.i == 1) data.index.i = 2;
    else if (data.index.i == 2) data.index.i = 1;
  }

  // search the expressions children
  for(i = 0; i < c; ++i) apmn[i]->FlipTemplateIndices();
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function determines if an expression is odd or even
//		 with respect to its arguments.
void MathNode::OddOrEven ( MathNode *pmnFlip )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("mathnode.cpp MathNode::OddOrEven\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  switch (uNode)
  { case TYPE_INDEX:
      uExpr |= FLAG_TEST;
      break;

    case TYPE_RCONST:
    case TYPE_ICONST:
    case TYPE_VAR:
    case TYPE_ELENABS:
      uExpr |= FLAG_EVEN;
      break;

    case TYPE_ARRAY:
      uExpr |= (((c > 0) && (apmn[0]->uExpr & FLAG_CONST)) &&
		((c > 1) && (apmn[1]->uExpr & FLAG_CONST)) &&
		((c > 2) && (apmn[2]->uExpr & FLAG_CONST)))
	    ? FLAG_EVEN : FLAG_TEST;
      break;

    case TYPE_ADD:
      apmn[0]->OddOrEven(pmnFlip->apmn[0]);
      apmn[1]->OddOrEven(pmnFlip->apmn[1]);
      if      ((apmn[0]->uExpr & FLAG_ODD ) && (apmn[1]->uExpr & FLAG_ODD ))
	uExpr |= FLAG_ODD;
      else if ((apmn[0]->uExpr & FLAG_EVEN) && (apmn[1]->uExpr & FLAG_EVEN))
	uExpr |= FLAG_EVEN;
      else if ((apmn[0]->uExpr & FLAG_TEST) && (apmn[1]->uExpr & FLAG_TEST))
      { if (Compare(pmnFlip)) uExpr |= FLAG_EVEN;
      }
      break;

    case TYPE_SUB:
      apmn[0]->OddOrEven(pmnFlip->apmn[0]);
      apmn[1]->OddOrEven(pmnFlip->apmn[1]);
      if      ((apmn[0]->uExpr & FLAG_ODD ) && (apmn[1]->uExpr & FLAG_ODD ))
	uExpr |= FLAG_ODD;
      else if ((apmn[0]->uExpr & FLAG_EVEN) && (apmn[1]->uExpr & FLAG_EVEN))
      { pmnFlip = new MathNode(TYPE_NEG, pmnFlip, pmnFlip->iPtrExpr);
	if (Compare(pmnFlip)) uExpr |= FLAG_ODD;
      }
      else if ((apmn[0]->uExpr & FLAG_TEST) && (apmn[1]->uExpr & FLAG_TEST))
      { pmnFlip = new MathNode(TYPE_NEG, pmnFlip, pmnFlip->iPtrExpr);
	if (Compare(pmnFlip)) uExpr |= FLAG_ODD;
      }
      break;

    case TYPE_MUL:
    case TYPE_DIV:
      apmn[0]->OddOrEven(pmnFlip->apmn[0]);
      apmn[1]->OddOrEven(pmnFlip->apmn[1]);
      if (!(apmn[0]->uExpr & MASK_UNKNOWN)) return;
      if (!(apmn[1]->uExpr & MASK_UNKNOWN)) return;
      if      ((apmn[0]->uExpr & FLAG_ODD ) && (apmn[1]->uExpr & FLAG_ODD ))
	uExpr |= FLAG_EVEN;
      else if ((apmn[0]->uExpr & FLAG_EVEN) && (apmn[1]->uExpr & FLAG_EVEN))
	uExpr |= FLAG_EVEN;
      if      ((apmn[0]->uExpr & FLAG_ODD ) && (apmn[1]->uExpr & FLAG_EVEN))
	uExpr |= FLAG_ODD;
      else if ((apmn[0]->uExpr & FLAG_EVEN) && (apmn[1]->uExpr & FLAG_ODD ))
	uExpr |= FLAG_ODD;
      else uExpr |= FLAG_TEST;
      break;

    case TYPE_NEG:
    case TYPE_REAL:
      uExpr |= apmn[0]->uExpr & MASK_UNKNOWN;
      break;

    case TYPE_PRECOMP:
      uExpr |= (data.precomp.fOdd) ? FLAG_ODD : FLAG_EVEN;
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
////////////  DYNAMIC ARRAY OF POINTERS TO INSTANCES OF MATHNODES  ////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
DAPMathNode::DAPMathNode ( UINT c )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPMathNode::DAPMathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  cElements = 0;
  fManaged = FALSE;
  dynarray = (MathNode **) calloc((cSize = uInc = c), sizeof(MathNode*));
}


///////////////////////////////////////////////////////////////////////////////
DAPMathNode::~DAPMathNode ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPMathNode::~DAPMathNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // if the array is managed then deallocate the stored objects
  if (fManaged)
    for(i = 0; i < cElements; ++i)
      delete dynarray[i];

  free((void *) dynarray);
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
UINT DAPMathNode::Add (MathNode *t)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPMathNode::Add\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // increase the array size if necessary
  if (cElements == cSize)
  { cSize += uInc;
    dynarray = (MathNode **) realloc(dynarray, cSize * sizeof(MathNode *));
  }

  dynarray[cElements] = t;
  return cElements++; // should it be ++cElements ?fengj
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function removes the last element from the dynarmic array
void DAPMathNode::RemoveLast ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPMathNode::RemoveLast\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  --cElements;

  // if the array is managed then deallocate the stored object
  if (fManaged) delete dynarray[cElements];
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function removes an element from the dynarmic array
void DAPMathNode::Remove (UINT iRemove)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPMathNode::Remove\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  --cElements;
  if (fManaged) delete dynarray[iRemove];
  for(i = iRemove; i < cElements; ++i)
    dynarray[i] = dynarray[i+1];
  dynarray[cElements] = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
void DAPMathNode::Flush ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPMathNode::Flush\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // if the array is managed then deallocate the stored objects
  if (fManaged)
    for(i = 0; i < cElements; ++i)
      delete dynarray[i];

  // reset the element count
  cElements = 0;
}


