/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     mathnode.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/19/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the SGFramework expression tree consisting of a n-ary
 tree of math nodes.

-----------------------------------------------------------------------------*/


#ifndef __mathnode_h
#define __mathnode_h

#ifndef __types_h
#include "types.h"
#endif

#ifndef __mathfunc_h
#include "mathfunc.h"
#endif


// Hate, like acid, damages the vessel in which it is stored and the object
// on which it is poured -- Richard W. De Haan


typedef struct tagARG			// arg
{ UINT i;				// argument index
  BOOL f;				// may refer to an unknown
} ARG;

typedef struct tagINDEX			// index
{ UINT i;				// index index
  int  n;				// index value
} INDEX;

typedef struct tagICONST		// iconst
{ int  i;				// integer constant's value
} ICONST;

typedef struct tagRCONST		// rconst
{ REAL r;				// real constant's value
} SCONST;

typedef struct tagNSUM			// nsum
{ BOOL fTriSum;				// set if triangle summation
  UINT iIndex;				// summation index
  UINT iLabel;				// label index
  UINT iExpr;				// nsum math node expression index
} NSUM;

typedef struct tagLSUM			// lsum
{ UINT iIndex;				// summation index
  UINT iLabel;				// label index
  UINT iExpr;				// lsum math node expression index
} LSUM;

typedef struct tagLOCALVAR		// lvar
{ UINT i;				// local variable index
  UINT iDiff;				// differentiated w/ respect to
} LOCALVAR;

typedef struct tagVAR			// var
{ UINT i;				// variable index
  UINT u;				// variable offset
  UINT iT;				// temporary index
  void *pmnNSUM;			// pointer to node sum
  void *pmnLSUM;			// pointer to label sum
} VAR;

typedef struct tagARRAY			// array
{ UINT i;				// array index
  UINT ac[3];				// elements per dimension
  UINT uA;				// array offset
  UINT uE;				// element offset
  UINT iT;				// temporary index
  void *pmnNSUM;			// pointer to node sum
  void *pmnLSUM;			// pointer to label sum
} ARRAY;

typedef struct tagFUNC			// func
{ UINT iFunc;				// function index
  UINT iPD;				// partial derivitive index
  PRF  aprf[3];				// pointer to function and derivatives
} FUNC;

typedef struct tagUDFUNC		// udfunc
{ UINT iFunc;				// function index
  UINT iPD;				// partial derivative index
  BOOL *af;				// pointer to "may be unknown" flags
} UDFUNC;

typedef struct tagPRECOMP		// precomp
{ BOOL fOdd;				// odd function
  BOOL fNode;				// precomputed at node
  UINT *au1;				// argument to index array 1
  UINT *au2;				// argument to index array 2
} PRECOMP;

typedef union tagMNDATA
{ int	    n;
  ARG	    arg;
  VAR	    var;
  BOOL	    f;
  REAL	    r;
  FUNC	    func;
  NSUM	    nsum;
  LSUM	    lsum;
  INDEX	    index;
  ARRAY	    array;
  UDFUNC    udfunc;
  PRECOMP   precomp;
  LOCALVAR  lvar;
} MNDATA;


class MathNode
{ protected:
    void KillChildren ();		// deallocate children trees
    void NameHeir (UINT);		// name heir
    void Cycle (BOOL);			// cycle math node operators
    void SwapChildren (UINT, UINT);	// swap children trees

    inline BOOL IsZero	   () { return ((uNode == TYPE_RCONST) && (data.r ==  0.0)) || ((uNode == TYPE_ICONST) && (data.n ==  0)); }
    inline BOOL IsOne	   () { return ((uNode == TYPE_RCONST) && (data.r ==  1.0)) || ((uNode == TYPE_ICONST) && (data.n ==  1)); }
    inline BOOL IsMinusOne () { return ((uNode == TYPE_RCONST) && (data.r == -1.0)) || ((uNode == TYPE_ICONST) && (data.n == -1)); }
    inline BOOL IsNeg	   () { return	 uNode == TYPE_NEG; }
    inline BOOL IsInv	   () { return	(uNode == TYPE_FUNC) && (data.func.iFunc == FUNC_INV) && (data.func.iPD == 0); }
    inline BOOL Is   (UINT u) { return	 uNode == u; }

    void MakeZero ();			// change math node to 0.0
    void MakeOne  ();			// change math node to 1.0
    void MakeInv  ();			// make inverse function
    void MakeSq	  ();			// make square function

    void ValidateInteger (long int);	// determines if long will fit in int
    void SimplifyAdd ();		// simplify addition expressions
    void SimplifySub ();		// simplify subtration expressions
    void SimplifyMul ();		// simplify multiplication expressions
    void SimplifyDiv ();		// simplify division expressions
    void SimplifyNeg ();		// simplify negation expressions
    void SimplifyNot ();		// simplify not expressions
    void SimplifyReal ();		// simplify convert to real expressions
    void SimplifyPow ();		// simplify pow function expressions
    void SimplifyAnd ();		// simplify and expressions
    void SimplifyOr  ();		// simplify or expressions
    void SimplifyXor ();		// simplify exclusive or expressions
    void SimplifyGT  ();		// simplify > expressions
    void SimplifyGE  ();		// simplify >= expressions
    void SimplifyLT  ();		// simplify < expressions
    void SimplifyLE  ();		// simplify <= expressions
    void SimplifyEQ  ();		// simplify == expressions
    void SimplifyNE  ();		// simplify <> expressions
    BOOL SimplifyDiff ();		// simplify derivative expressions

    void FindLocalVars ();		// find local variables
    void WriteCode ();			// convert the expression to C++ code
    void TagNodeSum ();			// tag node sum arrays and variables
    void TagLabelSum ();		// tag label sum arrays and variables
    void InitRegLabel ();		// initialize region labels
    MathNode * CopyTemplate ();		// copy the template

  public:
    void SimplifyFunc  ();		// simplify functions
    void SimplifyArray ();		// simplify arrays
    void SimplifyDiff (UINT);		// simplify derivative expression
    void SimplifyDiff (UINT, UINT, BOOL *, REAL *);

  public:
    UINT uNode;				// node type
    UINT uExpr;				// expression type
    UINT iPtrNode;			// buffer index of math node
    UINT iPtrExpr;			// buffer index of expression
    UINT c;				// number of children
    MathNode **apmn;			// array of pointers to children
    MNDATA data;			// type specific information

  public:
    MathNode *Copy ();			// copy math node expression
    MathNode *CopyTemplate (MathNode *);// copy the template
    MathNode *Diff ();			// differentiate the math node expr.
    BOOL Compare (MathNode *);		// compare two expressions
    void FlipTemplateIndices();		// flip template indices (i,j) -> (j,i)
    void AddChild(MathNode *);		// add a child to the parent
    void Evaluate ();			// evaluate an integer expression
    void Variables (BOOL = FALSE);	// searches and stores vars and arrays
    void FindLocalVars (UINT *);	// search for local variables
    void WriteCode (void *, BOOL, BOOL);// convert the expression to C++ code
    void MarkSums ();			// mark the node and label sums
    void WriteSums ();			// write node and label summations
    void WriteSums (void *, BOOL);	// write node and label summations
    void TagSumVariables ();		// tag node sum arrays and variables
    void OddOrEven (MathNode *);	// determine if function is odd or even


  // MathNode Constructors
  // 1.	  unknown math node
  // 2.	  integer constant math node
  // 3.	  real constant math node
  // 4.	  index math node
  // 5.	  variable or local variable math node
  // 6.	  argument math node
  // 7.	  unary operator math node expression
  // 8.	  binary operator math node epression
  // 9.	  node summation expression
  // 10.  label summation expression
  public:
  /* 1 */   MathNode ();
  /* 2 */   MathNode (int, UINT);
  /* 3 */   MathNode (double, UINT);
  /* 4 */   MathNode (UINT, UINT);
  /* 5 */   MathNode (UINT, UINT, UINT, BOOL = FALSE);
  /* 6 */   MathNode (UINT, BOOL, UINT);
  /* 7 */   MathNode (UINT, MathNode *, UINT);
  /* 8 */   MathNode (UINT, MathNode *, MathNode *, UINT);
  /* 9 */   MathNode (MathNode *, UINT, UINT, MathNode *, BOOL, UINT);
  /* 10 */  MathNode (UINT, UINT, MathNode *, UINT);
	   ~MathNode ();
};


class DAPMathNode
{ private:
    UINT cSize;				// total number of elements
    UINT uInc;				// increment size
    UINT cElements;			// number of elements in array
    BOOL fManaged;			// managed array
    MathNode **dynarray;		// dynamic array

  public:
    void SetManaged (BOOL f)		{ fManaged = f; }
    UINT GetElements ()			{ return cElements; }
    UINT Add (MathNode *);		// add an item to the array
    void RemoveLast ();			// remove last element
    void Remove(UINT);			// remove an element
    void Flush ();			// remove all pointers from the array
    MathNode * operator[](UINT i)	{ return dynarray[i]; }

  public:
     DAPMathNode (UINT = 16);
    ~DAPMathNode ();
};

#endif

