/*-----------------------------------------------------------------------------
 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  SGFramework Minimum Degree Program
 FILE:	     order.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 06/16/96    KMK      initial implementation
 03/25/97    JJF      modification

 DESCRIPTION:

 This module contains the functions needed to implement a minimum ordering
 algorithm.

-----------------------------------------------------------------------------*/


#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>

#if defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#define	 O_BINARY 0x0000
#else
#include <values.h>
#include <io.h>
#include <sys\stat.h>
#endif


///////////////////////////////////////////////////////////////////////////////
////////////////////////////  TYPE DEFINITIONS	///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


#define TRUE  1
#define FALSE 0

#define STATUS_FAILURE	  1
#define STATUS_SUCCESS	  0
#define SIGNATURE_PERM	  "@~!__PERM__!~@"

typedef int	     BOOL;
typedef unsigned int UINT;
typedef double	     REAL;
typedef int	     I3ARRAY[3];


///////////////////////////////////////////////////////////////////////////////
///////////////	 MISCELLANEOUS CONSTANTS, MACROS AND VARIABLES	///////////////
///////////////////////////////////////////////////////////////////////////////


#define LEN_IDENT	    15		// identifier length
#define LEN_FILENAME	    128		// maximum file name length
#define TYPE_RCONST	    3		// real constant
#define TYPE_ICONST	    4		// integer constant

typedef struct tagTOPHEAD		// tophead
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cElement;			// number of variables & array elements
  UINT cTag;				// number of unknown elements
  UINT cTopology;			// number of nonzeroes excluding pivots
  UINT cEquation;			// number of equations
} TOPHEAD;

typedef struct tagPERMHEAD		// permhead
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  char szPerm[64];			// ordering algorithm
  UINT cTag;				// number of unknown elements
} PERMHEAD;

typedef struct tagXLATCONST
{ char szName[LEN_IDENT+1];		// constant name
  UINT uType;				// constant type
  union					//
  { REAL r;				// value of real constant
    int	 n;				// value of integer constant
  } data;
} XLATCONST;

typedef struct tagXLATVAR
{ char szName[LEN_IDENT+1];		// variable name
  UINT uOffset;				// variable offset
} XLATVAR;

typedef struct tagXLATARRAY
{ char szName[LEN_IDENT+1];		// array name
  UINT uOffset;				// array offset
  UINT cDim;				// number of dimensions
  UINT acElements[3];			// elements per dimensions
} XLATARRAY;

typedef struct tagTAG			// tag
{ UINT iVar;				// element array index (i.e. offset)
  UINT iEqu;				// equation index
  UINT iArray;				// array index ... -1u == variable
} TAG;

typedef struct tagTOP			// top
{ UINT i;				// index in topology array
  UINT c;				// number of nonzeroes in equation
} TOP;

///////////////////////////////////////////////////////////////////////////////
/////////////////////////  NONZERO ELEMENT STRUCTURE  /////////////////////////
///////////////////////////////////////////////////////////////////////////////


typedef struct tagNONZERO		// nz
{ UINT r;				// row index
  UINT c;				// column index
  struct tagNONZERO *pnzRow;		// link to next element
  struct tagNONZERO *pnzCol;		// link to next element
} NONZERO;


///////////////////////////////////////////////////////////////////////////////
///////////  DYNAMIC STACK CLASS OF POINTERS TO NONZERO STRUCTURES  ///////////
///////////////////////////////////////////////////////////////////////////////


class PNZStack
{ private:
    UINT u;		    // increment
    UINT c;		    // stack size
    UINT i;		    // stack index
    NONZERO **stack;	    // stack for ptr to NZ blocks alloc in NonZeroHeap
    void grow ()	    { c += u;
			      stack = (NONZERO **) realloc(stack,
				c * sizeof(NONZERO *)); }

  public:
    void push (NONZERO *T)  { if (i == c) grow(); stack[i++] = T; }
    NONZERO *pop ()	    { return stack[--i]; }
    BOOL empty ()	    { return !i; }
    void flush ()	    { i = 0; }

  public:
     PNZStack (UINT n = 64) { i = 0; stack = (NONZERO**) calloc((c = u = n),
                                             sizeof(NONZERO*)); }
    ~PNZStack ()	    { free((void *) stack); }
};


///////////////////////////////////////////////////////////////////////////////
/////////////////////////  NONZERO ELEMENT HEAP CLASS  ////////////////////////
///////////////////////////////////////////////////////////////////////////////


class NonZeroHeap
{ private:
    PNZStack stack;			// nonzero element stack
    NONZERO *pnzFree;			// free element list

  public:
    NONZERO * Alloc ();			// allocate a nonzero element
    void Dealloc (NONZERO *);		// deallocate a nonzero element

  public:
     NonZeroHeap ();
    ~NonZeroHeap ();
};

extern NonZeroHeap *pnzHeap;


#if defined(P16BIT)
#define NONZEROES_PER_ALLOC 1024
#else
#define NONZEROES_PER_ALLOC 16384
#endif


///////////////////////////////////////////////////////////////////////////////
////////////////////////////  SPARSE MATRIX CLASS  ////////////////////////////
///////////////////////////////////////////////////////////////////////////////


class SparseMatrix			// mtrx
{ public:
    UINT c;				// dimension of square matrix
    BOOL fSorted;			// set if matrix is sorted.
    UINT *ac;				// number of neighbors
    NONZERO **apnzRow;			// head of row link lists
    NONZERO **apnzCol;			// head of column link lists

  public:
    void Deallocate ();			// deallocate the matrix elements
    UINT Order (UINT *);		// order the matrix

    // allocate the nonzero elements of the Jacobian matrix
    UINT AllocateJacobian (UINT *, TOP *);

  public:
     SparseMatrix (UINT);		// constructor
    ~SparseMatrix ();			// destructor
};


///////////////////////////////////////////////////////////////////////////////
////////////  COMMAND LINE PARSER CLASS, VARIABLES AND FUNCTIONS  /////////////
///////////////////////////////////////////////////////////////////////////////

#include "cmdparse.h"

typedef struct tagCMDARGS
{ BOOL fQuiet;
  char *pszTopFile;
  char *pszPermFile;
  char *pszSMMSFile;
} CMDARGS;

CMDARGS cmdargs;

CmdParse *pcmd = NULL;

char *apszHelp[] = {
  "usage:  order [options] topfile permfile",
  "  -h	   display this help screen",
  "  -q	   run quietly",
  "  -Pxxx generate a SMMS permutation vector named xxx",
  NULL
};


///////////////////////////////////////////////////////////////////////////////
// Function:	 nParseCommandLine
// Scope:	 Public
// Description:	 This function parses the command line arguments.
int nParseCommandLine (CmdParse *pcmdparse)
{
  if ((cmdargs.fQuiet = pcmdparse->fArgExist("q")) == FALSE)
    fprintf(stdout, "SG MMD Order        Version 1.0  (c) 1997  K. Kramer\n");

  if (pcmdparse->fArgExist("h") || (pcmdparse->GetNumOfPosArgs() != 2))
  { for(char **ppsz = apszHelp; *ppsz; ++ppsz)
      fprintf(stdout, "%s\n", *ppsz);
    return STATUS_FAILURE;
  }


  cmdargs.pszTopFile  = pcmdparse->pszGetPosArg(0, "sim.top");
  cmdargs.pszPermFile = pcmdparse->pszGetPosArg(1, "sim.prm");
  cmdargs.pszSMMSFile = pcmdparse->pszGetSwitchArg("P", NULL);
  return STATUS_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
//////////////////  NONZERO ELEMENT STRUCTURE AND HEAP CLASS  /////////////////
///////////////////////////////////////////////////////////////////////////////


#if defined(P16BIT)
#define NONZEROES_PER_ALLOC 1024
#else
#define NONZEROES_PER_ALLOC 16384
#endif

// global declaration
NonZeroHeap *pnzHeap = NULL;


///////////////////////////////////////////////////////////////////////////////
// Function:	 NonZeroHeap::NonZeroHeap
// Scope:	 Public, Constructor
NonZeroHeap::NonZeroHeap ()
{
  pnzFree = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 NonZeroHeap::~NonZeroHeap
// Scope:	 Public, Destructor
NonZeroHeap::~NonZeroHeap ()
{
  // release each NonZero blocks allocated (recorded in PNZStack)
  while (!stack.empty())
  { NONZERO *pnz = stack.pop();
    delete [] pnz;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 NonZeroHeap::Alloc
// Scope:	 Public
// Description:	 This function allocates a nonzero element from the nonzero
//		 heap.
NONZERO * NonZeroHeap::Alloc ()
{
  // if no nonzero elements are in the free list then allocate another block
  // of nonzero elements
  if (!pnzFree)
  { NONZERO *anz = new NONZERO[NONZEROES_PER_ALLOC];
    NONZERO *pnz = anz;
    for(UINT i = 1; i < NONZEROES_PER_ALLOC; ++i)
    { pnz->pnzRow = pnz + 1;
      ++pnz;
    }
    pnz->pnzRow = NULL;
    pnzFree = anz;
    stack.push(anz);
  }

  // return an element from the free list
  NONZERO *pnz = pnzFree;
  pnzFree = pnzFree->pnzRow;
  return pnz;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 NonZeroHeap::Dealloc
// Scope:	 Public
// Description:	 This function deallocates a nonzero element and returns it
//		 to the nonzero heap.
void NonZeroHeap::Dealloc (NONZERO *pnz)
{
  pnz->pnzRow = pnzFree;
  pnzFree = pnz;
}


///////////////////////////////////////////////////////////////////////////////
////////////////////////////  SPARSE MATRIX CLASS  ////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 SparseMatrix::SparseMatrix
// Scope:	 Public, Constructor
SparseMatrix::SparseMatrix (UINT cDim)
{
  c	   = cDim;
  ac	   = new UINT [cDim];
  apnzRow  = new NONZERO* [cDim];
  apnzCol  = new NONZERO* [cDim];
  fSorted  = FALSE;

  for(UINT i = 0; i < cDim; ++i)
  { apnzRow[i] = NULL;
    apnzCol[i] = NULL;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 SparseMatrix::~SparseMatrix
// Scope:	 Public, Destructor
SparseMatrix::~SparseMatrix ()
{
  Deallocate();
  delete [] ac;
  delete [] apnzRow;
  delete [] apnzCol;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 SparseMatrix::Deallocate
// Scope:	 Public
// Description:	 This function deallocates the nonzero elements of the matrix.
void SparseMatrix::Deallocate ()
{
  // for each row
  UINT i;
  NONZERO *pnz, *pnzT, **ppnz;
  for(i = 0, ppnz = apnzRow; i < c; ++i, ++ppnz)
  { for(pnz = *ppnz; pnz; pnz = pnzT)
    { pnzT = pnz->pnzRow;
      pnzHeap->Dealloc(pnz);
    }
    apnzRow[i] = NULL;
    apnzCol[i] = NULL;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 SparseMatrix::AllocateJacobian
// Scope:	 Public
// Description:	 This function allocates the Jacobian matrix by adding the
//		 appropriate nonzero elements to the sparse matrix pmtrx.
//		 This function only constructs the topology of the Jacobian
//		 matrix.  The SparseMatrix::ComputeJacobian function must be
//		 called to set the values of the Jacobian matrix elements.
//		 SparseMatrix::AddFill may to called to add fill to the
//		 matrix topology.
UINT SparseMatrix::AllocateJacobian (UINT *aiTop, TOP  *atop)
{
  UINT cNonZero = 0;
  UINT *pc = ac;

  // for each row
  for(UINT r = 0; r < c; ++r, ++pc)
  {
    // initialize variables and pointers
    TOP	 *ptop = atop + r;

    // compute and link the pivot
    NONZERO *pnz = pnzHeap->Alloc();
    pnz->r = r;
    pnz->c = r;
    pnz->pnzRow = apnzRow[r];
    pnz->pnzCol = apnzCol[r];
    apnzRow[r]	= pnz;
    apnzCol[r]	= pnz;

    // link the other nonzero elements in this row
    UINT cDensity = ptop->c;
    UINT *pu = aiTop + ptop->i;
    for(UINT i = 0; i < cDensity; ++i, ++pu)
    { UINT c = *pu;
      pnz = pnzHeap->Alloc();
      pnz->r = r;
      pnz->c = c;

      // link the nonzero element
      pnz->pnzRow = apnzRow[r];
      pnz->pnzCol = apnzCol[c];
      apnzRow[r]  = pnz;
      apnzCol[c]  = pnz;
      ++cNonZero;
    }

    cNonZero += cDensity + 1;
    *pc = cDensity;
  }

  return cNonZero;
}


#if defined(DEBUG)
void TestSort (UINT iPiv, UINT iRow, NONZERO **apnzSort, NONZERO *pnzHead, NONZERO *pnzTail)
{
  if (pnzHead && (pnzHead->pnzCol))
  { fprintf(stderr, "iPiv = %u, iRow = %u, pnzHead->pnzCol != NULL\n", iPiv, iRow);
    exit(2);
  }

  if (pnzTail && (pnzTail->pnzRow))
  { fprintf(stderr, "iPiv = %u, iRow = %u, pnzTail->pnzRow != NULL\n", iPiv, iRow);
    exit(2);
  }

  UINT cLast = -1u;
  UINT cList = 0;
  for(NONZERO *pnz = pnzHead; pnz; pnz = pnz->pnzRow, ++cList)
  { if ((pnz != pnzHead) && (pnz->pnzCol->pnzRow != pnz))
    { fprintf(stderr, "iPiv = %u, iRow = %u, pnz->pnzCol->pnzRow != pnz\n", iPiv, iRow);
      exit(2);
    }
    if ((pnz != pnzTail) && (pnz->pnzRow->pnzCol != pnz))
    { fprintf(stderr, "iPiv = %u, iRow = %u, pnz->pnzRow->pnzCol != pnz\n", iPiv, iRow);
      exit(2);
    }
    if (pnz->c != cLast)
    { if (apnzSort[pnz->c] != pnz)
      { fprintf(stderr, "iPiv = %u, iRow = %u, apnzSort[pnz->c] != pnz\n", iPiv, iRow);
	exit(2);
      }
    }
    cLast = pnz->c;
    fprintf(stdout, "%u ", pnz->r);
  }
  fprintf(stdout, "[%u]\n", cList);
}
#endif


UINT *auInfo;


///////////////////////////////////////////////////////////////////////////////
// Function:	 SparseMatrix::Order
// Scope:	 Public
// Description:
UINT SparseMatrix::Order (UINT *aiPerm)
{
  NONZERO *pnz, *pnz1, *pnz2, *pnzP, *pnzN, *pnzPivRow, *pnzCol;

  // determine the strongly connected blocks
  UINT i;
  UINT iStack = 0;
  UINT iBlock = 0;
  UINT *auBlock = new UINT [c];
  UINT *aiStack = new UINT [c];
  BOOL *afVisit = new BOOL [c];
  UINT *pu;
  BOOL *pf;
  for(i = 0, pu = auBlock, pf = afVisit; i < c; ++i, ++pu, ++pf)
  { *pu = 0;
    *pf = FALSE;
  }
  for(i = 0, pu = auBlock; i < c; ++i, ++pu)
  { if (*pu == 0)
    { ++iBlock;
      aiStack[iStack++] = i;
      afVisit[i] = TRUE;
      while (iStack)
      { UINT iRow = aiStack[--iStack];
	auBlock[iRow] = iBlock;
	for(pnz = apnzRow[iRow]; pnz; pnz = pnz->pnzRow)
	{ UINT iCol = pnz->c;
	  if (!afVisit[iCol])
	  { aiStack[iStack++] = iCol;
	    afVisit[iCol] = TRUE;
	  }
	}
	for(pnz = apnzCol[iRow]; pnz; pnz = pnz->pnzCol)
	{ UINT iRow = pnz->r;
	  if (!afVisit[iRow])
	  { aiStack[iStack++] = iRow;
	    afVisit[iRow] = TRUE;
	  }
	}
      }
    }
  }


  // initialize local variables
  UINT cFill  = 0;
  UINT iOrder = 0;
  UINT cBlock = iBlock;
  UINT iPivot;

  // allocate the information structure
  auInfo = new UINT[cBlock+2];
  auInfo[0] = cBlock;
  auInfo[cBlock+1] = c;

  // create and initialize the sort list and the logical working row
  NONZERO **apnzSort = new NONZERO* [c];
  NONZERO **ppnz = apnzSort;
  BOOL *af = new BOOL[c];
  for(i = 0, pf = af; i < c; ++i, ++pf, ++ppnz)
  { *ppnz = NULL;
    *pf = FALSE;
  }

  // order each stronly connect block
  for(iBlock = 1; iBlock <= cBlock; ++iBlock)
  { auInfo[iBlock] = iOrder;

    // form doubly linked lists of row with the same number of neighbors
    UINT *pc = ac;
    UINT *pu = auBlock;
    ppnz = apnzRow;
    for(i = 0; i < c; ++i, ++pc, ++pu, ++ppnz)
    { if (*pu == iBlock)
      { UINT cNeighbor = *pc;
	ppnz = apnzSort + cNeighbor;
	pnz1 = *ppnz;
	pnz = pnzHeap->Alloc();
	pnz->r = i;
	pnz->c = cNeighbor;
	if (pnz1)
	{ pnz2 = pnz1->pnzCol;
	  pnz->pnzRow = pnz1;
	  pnz->pnzCol = pnz2;
	  pnz1->pnzCol = pnz;
	  pnz2->pnzRow = pnz;
	}
	else
	{ pnz->pnzRow = pnz;
	  pnz->pnzCol = pnz;
	}
	*ppnz = pnz;
      }
    }

    // form a doubly linked list of all the rows such that the rows with
    // the fewest number of neighbors appear first in the list.
    NONZERO *pnzHead = NULL;
    NONZERO *pnzTail = NULL;
    for(i = 0, ppnz = apnzSort; i < c; ++i, ++ppnz)
    { if ((pnz1 = *ppnz) != NULL)
      { pnz2 = pnz1->pnzCol;
	if (!pnzHead)
	{ pnzHead = pnz1;
	  pnzTail = pnz2;
	  pnz1->pnzCol = NULL;
	  pnz2->pnzRow = NULL;
	}
	else
	{ pnzTail->pnzRow = pnz1;
	  pnz1->pnzCol = pnzTail;
	  pnz2->pnzRow = NULL;
	  pnzTail = pnz2;
	}
      }
    }

#if defined(DEBUG)
    TestSort(-1u, -1u, apnzSort, pnzHead, pnzTail);
#endif

    while (pnzHead->pnzRow)
    {
      // select the row with the fewest number of neighbors as the next pivot
      pnz = pnzHead;
      iPivot = pnz->r;
      pnzHead = pnz->pnzRow;
      pnzHead->pnzCol = NULL;
      apnzSort[pnz->c] = NULL;
      apnzSort[pnzHead->c] = pnzHead;
      pnzHeap->Dealloc(pnz);
      pnzPivRow = apnzRow[iPivot];

      // symbolically do the Gaussian elimination
      for(pnzCol = apnzCol[iPivot]; pnzCol; pnzCol = pnzCol->pnzCol)
      { UINT iRow = pnzCol->r;
	if ((iRow != -1u) && (iRow != iPivot))
	{ UINT c1 = ac[iRow];
	  UINT c2 = c1 - 1;

	  // load the logical working row
	  for(pnz = apnzRow[iRow]; pnz; pnz = pnz->pnzRow) af[pnz->c] = TRUE;

	  // symbolically add the pivot row to the current row
	  for(pnz = pnzPivRow; pnz; pnz = pnz->pnzRow)
	  { UINT iCol = pnz->c;
	    if ((pnz->r != -1u) && !af[iCol])
	    { pnz1 = pnzHeap->Alloc();
	      pnz1->r = iRow;
	      pnz1->c = iCol;
	      pnz1->pnzRow = apnzRow[iRow];
	      pnz1->pnzCol = apnzCol[iCol];
	      apnzRow[iRow] = pnz1;
	      apnzCol[iCol] = pnz1;
	      ++cFill;
	      ++c2;
	    }
	  }

	  // unload the working row
	  for(pnz = apnzRow[iRow]; pnz; pnz = pnz->pnzRow) af[pnz->c] = FALSE;

	  // mark the eliminated node
	  pnzCol->r = -1u;

	  // update the neighbor count if changed
	  if (c1 != c2)
	  { ac[iRow] = c2;

	    // remove the previous count
	    for(pnz = apnzSort[c1]; pnz; pnz = pnz->pnzRow)
	      if (pnz->r == iRow) break;

#if defined(DEBUG)
	    if (pnz == NULL)
	    { fprintf(stderr, "error:  could not find element %u!\n", iRow);
	      exit(2);
	    }
#endif

	    pnzN = pnz->pnzRow;
	    pnzP = pnz->pnzCol;

	    if (pnz == apnzSort[c1])
	      apnzSort[c1] = (pnzN && (pnzN->c == c1)) ? pnzN : (NONZERO *) NULL;

	    if ((pnz == pnzHead) && (pnz == pnzTail))
	    { pnzHead = NULL;
	      pnzTail = NULL;
	    }
	    else if (pnz == pnzHead)
	    { pnzHead = pnzN;
	      pnzN->pnzCol = NULL;
	    }
	    else if (pnz == pnzTail)
	    { pnzTail = pnzP;
	      pnzP->pnzRow = NULL;
	    }
	    else
	    { pnzP->pnzRow = pnzN;
	      pnzN->pnzCol = pnzP;
	    }

#if defined(DEBUG)
	    TestSort(iPivot, iRow, apnzSort, pnzHead, pnzTail);
#endif
	    // add the new count
	    pnz->c = c2;
	    pnzN = apnzSort[c2];
	    apnzSort[c2] = pnz;
	    if (pnzN)
	    { if (pnzN == pnzHead)
	      { pnz->pnzRow = pnzN;
		pnz->pnzCol = NULL;
		pnzN->pnzCol = pnz;
		pnzHead = pnz;
	      }
	      else
	      { pnzP = pnzN->pnzCol;
		pnz->pnzRow = pnzN;
		pnz->pnzCol = pnzP;
		pnzN->pnzCol = pnz;
		pnzP->pnzRow = pnz;
	      }
	    }
	    else
	    { if (pnzHead == NULL)
	      { pnz->pnzCol = NULL;
		pnz->pnzRow = NULL;
		pnzHead = pnz;
		pnzTail = pnz;
	      }
	      else if (pnzHead->c >= c2)
	      { pnz->pnzCol = NULL;
		pnz->pnzRow = pnzHead;
		pnzHead->pnzCol = pnz;
		pnzHead = pnz;
	      }
	      else if (pnzTail->c < c2)
	      { pnz->pnzRow = NULL;
		pnz->pnzCol = pnzTail;
		pnzTail->pnzRow = pnz;
		pnzTail = pnz;
	      }
	      else
	      { while ((pnzN = apnzSort[++c2]) == NULL);
		pnzP = pnzN->pnzCol;
		pnz->pnzRow = pnzN;
		pnz->pnzCol = pnzP;
		pnzP->pnzRow = pnz;
		pnzN->pnzCol = pnz;
	      }
	    }
#if defined(DEBUG)
	    TestSort(iPivot, iRow, apnzSort, pnzHead, pnzTail);
#endif
	  }
	}
      }

      // mark the nodes in the pivot row
      for(pnz = pnzPivRow; pnz; pnz = pnz->pnzRow) pnz->r = -1u;
#if defined(DEBUG)
      fprintf(stdout, "Perm[%u] = %u\n", iOrder, iPivot);
#endif
      aiPerm[iOrder++] = iPivot;
    }

    // remove the last entry in the sorting list and add it the perm vector
#if defined(DEBUG)
    fprintf(stdout, "Perm[%u] = %u\n", iOrder, pnzHead->r);
#endif
    aiPerm[iOrder++] = pnzHead->r;
    apnzSort[pnzHead->c] = NULL;
    pnzHeap->Dealloc(pnzHead);
  }

  if (!cmdargs.fQuiet)
  { fprintf(stdout, "Fills:	 %u\n", cFill);
    fprintf(stdout, "Blocks:	 %u\n", cBlock);
  }

  delete [] auBlock;
  delete [] aiStack;
  delete [] apnzSort;
  delete [] af;
  return cFill;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ReadTopFile
// Scope:	 Public
// Description:	 This function reads the topology file
//
// *.top    char szLogo[64];     // text logo
//          char szSign[16];     // signature
//          UINT cElement;       // # of variables & array elements
//          UINT cTag;           // # of unknown elements
//          UINT cTopology;      // # of nonzeroes excluding pivots
//          UINT cEquation;      // # of equations
//          ------------------------------------------------
//              cElement * UINT
//          ------------------------------------------------
//  wanted     cTopology * UINT   (Jacobian matrix topology) topology array
//
//          | row 1 |    row 2    | ...........
//          column indices go backwards and exclude pivots
//          ------------------------------------------------
//                  cTag * TAG
//          ------------------------------------------------
//  wanted          cTag * TOP    info for each row in top array
//          ------------------------------------------------
//                  cTag * I3ARRAY
//          ------------------------------------------------
//
void ReadTopFile ( char *pszTopFile, UINT *pcTag, TOP **patop, UINT **paiTop )
{
  int nHandle = open(pszTopFile, O_BINARY | O_RDONLY);
  if (nHandle == -1)
  { fprintf(stderr, "error:  cannot open file \"%s\"\n", pszTopFile);
    exit(2);
  }

  TOPHEAD head;
  read(nHandle, &head, sizeof(TOPHEAD));

  *pcTag = head.cTag;
  *paiTop = new UINT[head.cTopology];
  *patop = new TOP[head.cTag];
  lseek(nHandle, head.cElement * sizeof(UINT), SEEK_CUR);
  read(nHandle, *paiTop, head.cTopology * sizeof(UINT));
  lseek(nHandle, head.cTag * sizeof(TAG), SEEK_CUR);
  read(nHandle, *patop,	 head.cTag * sizeof(TOP));
  close(nHandle);

  if (!cmdargs.fQuiet)
  { fprintf(stdout, "Unknowns:	 %u\n", head.cTag);
    fprintf(stdout, "Nonzeroes:	 %u\n", head.cTopology + head.cTag);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 WritePermFile
// Scope:	 Public
// Description:	 This function writes the permuation file
void WritePermFile(char *pszPermFile, UINT c, UINT *aiPerm, UINT *auInfo)
{
  int nHandle = open(pszPermFile, O_CREAT | O_TRUNC | O_BINARY | O_WRONLY, S_IWRITE | S_IREAD);
  if (nHandle == -1)
  { fprintf(stderr, "error:  cannot create file \"%s\"\n", pszPermFile);
    exit(2);
  }

  PERMHEAD head;
  strcpy(head.szLogo, "SGFramework Permutation File  Version 1.0\n\x1a");
  strcpy(head.szSign, SIGNATURE_PERM);
  strcpy(head.szPerm, "SGFramework Minimum Degree Ordering Algorithm");
  head.cTag = c;
  write(nHandle, &head, sizeof(PERMHEAD));

  write(nHandle, aiPerm, c * sizeof(UINT));
  write(nHandle, auInfo, (auInfo[0] + 2) * sizeof(UINT));
  close(nHandle);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 WriteSMMSFile
// Scope:	 Public
// Description:	 This function writes a SMMS permuation file
void WriteSMMSFile(char *pszPermFile, UINT c, UINT *aiPerm)
{
  UINT i;

  if (!pszPermFile) return;

  FILE *file = fopen(pszPermFile, "wt");
  fprintf(file, "%u\n", c);
  for(i = 0; i < c; ++i)
    fprintf(file, "%u ", aiPerm[i]+1);
  fprintf(file, "\n");
  fclose(file);
}


int main ( UINT cArg, char **apszArg )
{
  CmdParse *pcmdparse = new CmdParse(cArg, apszArg);
  if (nParseCommandLine(pcmdparse) == STATUS_FAILURE) return 1;

  UINT	cTag;
  UINT *aiTop;
  TOP  *atop;

  // create the sparse matrix
  ReadTopFile(cmdargs.pszTopFile, &cTag, &atop, &aiTop); // allocate atop aiTop
  pnzHeap = new NonZeroHeap();
  SparseMatrix *pmtrx = new SparseMatrix(cTag);
  pmtrx->AllocateJacobian(aiTop, atop);

  // create the permuation vector
  UINT *aiPerm = new UINT[cTag];
  auInfo = NULL;
  pmtrx->Order(aiPerm); // auInfo will be allocated
  WritePermFile(cmdargs.pszPermFile, cTag, aiPerm, auInfo);
  WriteSMMSFile(cmdargs.pszSMMSFile, cTag, aiPerm);

  // clean up
  delete [] auInfo;
  delete [] aiPerm;
  delete pmtrx;
  delete pnzHeap;
  delete [] atop;
  delete [] aiTop;
  delete pcmdparse;
  return 0;
}


