/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     parser.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/14/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the Translate::uParse function which is the
 SGFramework parser.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "tokens.h"
#include "states.h"
#include "actions.h"
#include "error.h"
#include "source.h"
#include "mathnode.h"
#include "index.h"
#include "grammar.h"


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function is the SGFramework language parser.
UINT Source::uParse ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("parser.cpp Source::uParse\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  BOOL	fDone = FALSE;
  GetToken(token);
  uState = uLastState = ST_START;
  uSyntaxState = ST_PROG;

  do
  { // locate the input token in the finite uState machine table
    UINT ifsm;
    UINT ifsm1 = ifsmTable[uState];
    UINT ifsm2 = ifsmTable[uState + 1];
    for(ifsm = ifsm1; ifsm < ifsm2; ++ifsm)
      if ((afsmTable[ifsm].input == token.uTokenType) ||
	  (afsmTable[ifsm].input == DEFAULT)) break;

    // illegal input token ... syntax error
    // (no match in the syntax table, no DEFAULT either)
    if (ifsm == ifsm2)
    { SyntaxError();
      continue;
    }

    // legal token
    UINT uStateAction = afsmTable[ifsm].action;
    UINT UserAction   = afsmTable[ifsm].user;

    // execute uState/token actions (except for the type assignments)
    if (!(uStateAction & 0x8000)) switch (uStateAction)
    { case ACT_:	break;

      case ACT_PUSH:	stackToken->push(token);
      case ACT_DROP:	GetToken(token);
			break;

      case ACT_POP:	PutToken(token);
			token = stackToken->pop();
			break;

      case ACT_CALL:	stackState->push(uState);
			break;

      case ACT_REM2:	stackToken->pop();
      case ACT_REM1:	stackToken->pop();
			break;

      case ACT_NEW:	PutToken(token);
			token.uTokenType = -1u;
			break;
    }

    // execute the user action if present
    if (UserAction != USR_) UserActions(UserAction);

    // perform the type assignment if present
    if (uStateAction & 0x8000) token.uTokenType = uStateAction;

    // get the next uState
    uLastState = uState;
    uState = afsmTable[ifsm].next;
    if (uState == ST_RETURN)
    { if (stackState->empty()) fDone = TRUE;
      else uState = stackState->pop();
    }

  } while (!fDone && !error->fFatal());

  return (error->fFatal() || error->fError()) ? -1 : 0;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function is the syntax error handler.
void Source::SyntaxError ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("parser.cpp Source::SyntaxError\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // report syntax error
  for(UINT i = 0; aseeTable[i].state != -1U; ++i)
    if (aseeTable[i].state == uState)
    { error->AddMessage(EHM_SYNTAX, aseeTable[i].psz, token.iPtr);
      break;
    }

  // report additional error if syntax error is caused by end of file token
  if (token.uTokenType == TOK_EOF) error->AddMessage(EHM_UNEXPEOF, token.iPtr);

  // find the next (legalizing) program statement from the token stream
  for (;;)
  { UINT ifsm;
    UINT ifsm1 = ifsmTable[uSyntaxState];
    UINT ifsm2 = ifsmTable[uSyntaxState + 1];
    for(ifsm = ifsm1; ifsm < ifsm2; ++ifsm)
      if ((token.uTokenType == afsmTable[ifsm].input) ||
	  (token.uTokenType == TOK_EOF)) break;
    if (ifsm == ifsm2) GetToken(token);
    else break;
  }

  // delete any allocated classes in the token stack
  while (!stackToken->empty())
  { token = stackToken->pop();
    switch (token.uTokenType)
    { case TOK_EXPR:
      case TOK_MEXPR:
      case TOK_UEXPR:
      case TOK_EXPRS:
	if (token.data.p) delete (MathNode *) token.data.p;
	break;

      case TOK_INDEX:
	if (token.data.p) delete (Index *)    token.data.p;
	break;

      case TOK_INDICES:
	if (token.data.p) delete (Indices *)  token.data.p;
	break;
    }
  }

  // reset the state stack
  stackState->flush();
  uState = uSyntaxState;
}

