/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     simplify.cpp
 AUTHOR:     Kevin M. Kramer
             J. Joshau Feng

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/19/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version
 06/24/00    JJF      simplify arcsin, ... etc

 DESCRIPTION:

 This module simplifies math node expressions.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include <limits.h>
#include "types.h"
#include "error.h"
#include "mathnode.h"
#include "mathfunc.h"


// For through the law I died to the law so that I might live for God.	I
// have been crucified with Christ and I no longer live, but Christ lives
// in me.  The life I live in the body, I live by faith in the Son of God,
// who loved me and gave himself for me.  I do not set aside the grace of
// God, for if righteousness could be gained through the Law, Christ died
// for nothing.	 (Galatians 2:19-21 NIV)


///////////////////////////////////////////////////////////////////////////////
//////////////////  MATH NODE MISCELLANEOUS MEMBER FUNCTIONS //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::MakeZero
// Scope:	 Protected
// Description:	 If the math expression is a real expression then this
//		 function changes the expression to the real constant 0.0;
//		 otherwise, this function changes the expression to the
//		 integer constant 0.
void MathNode::MakeZero ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::MakeZero\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  BOOL fReal = uExpr & FLAG_REAL;
  KillChildren();

  if (fReal)
  { uNode = TYPE_RCONST;
    uExpr = FLAG_CONST | FLAG_REAL;
    data.r = 0.0;
  }
  else
  { uNode = TYPE_ICONST;
    uExpr = FLAG_CONST | FLAG_INT;
    data.n = 0;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::MakeOne
// Scope:	 Protected
// Description:	 If the math expression is a real expression then this
//		 function changes the expression to the real constant 1.0;
//		 otherwise, this function changes the expression to the
//		 integer constant 1.
void MathNode::MakeOne ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::MakeOne\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  BOOL fReal = uExpr & FLAG_REAL;
  KillChildren();

  if (fReal)
  { uNode = TYPE_RCONST;
    uExpr = FLAG_CONST | FLAG_REAL;
    data.r = 1.0;
  }
  else
  { uNode = TYPE_ICONST;
    uExpr = FLAG_CONST | FLAG_INT;
    data.n = 1;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::MakeInv
// Scope:	 Protected
// Description:	 This function changes the expression to the inverse function
void MathNode::MakeInv ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::MakeInv\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uNode = TYPE_FUNC;
  uExpr = apmn[0]->uExpr;
  data.func.iFunc = FUNC_INV;
  data.func.iPD = 0;
  data.func.aprf[0] = PRF(inv);
  data.func.aprf[1] = PRF(pd1inv);
  data.func.aprf[2] = NULL;
  SimplifyFunc();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::MakeSq
// Scope:	 Protected
// Description:	 This function changes the expression to the inverse function
void MathNode::MakeSq ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::MakeSq\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uNode = TYPE_FUNC;
  uExpr = apmn[0]->uExpr;
  data.func.iFunc = FUNC_SQ;
  data.func.iPD = 0;
  data.func.aprf[0] = PRF(sq);
  data.func.aprf[1] = PRF(pd1sq);
  data.func.aprf[2] = NULL;
  SimplifyFunc();
}


///////////////////////////////////////////////////////////////////////////////
//////////////////  MATH NODE SIMPLIFICATION MEMBER FUNCTIONS /////////////////
///////////////////////////////////////////////////////////////////////////////


#define LEFT  TRUE
#define RIGHT FALSE


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::ValidateInteger
// Scope:	 Protected
// Description:	 This function determines if the long integer l can be stored
//		 as an integer.	 If the integer is out of bounds then a error
//		 message is generated.
void MathNode::ValidateInteger (long int l)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::ValidateInteger\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if ((l < (long) INT_MIN) || (l > (long) INT_MAX))
    error->AddMessage(EHM_INTOVER, iPtrNode);

  data.n = (int) l;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyAdd
// Scope:	 Protected
// Description:	 This function simplifies addition math nodes
void MathNode::SimplifyAdd ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyAdd\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = apmn[0]->data.r + apmn[1]->data.r;
    }
    else
    { uNode = TYPE_ICONST;
      ValidateInteger((long) apmn[0]->data.n + (long) apmn[1]->data.n);
    }
    KillChildren();
    return;
  }

  //	the following addition expressions are simplified
  //
  //	1)   0	+  X  -->  X
  //	2)   X	+  0  -->  X
  //	3)  -X	+  X  -->  0
  //	4)   X	+ -X  -->  0
  //	5)  -X	+  Y  -->  Y - X
  //	6)   X	+ -Y  -->  X - Y
  //	7)  -X	+ -Y  --> -(X + Y)

  //  cases 1 and 2
  if	  (apmn[0]->IsZero()) NameHeir(1);
  else if (apmn[1]->IsZero()) NameHeir(0);

  //  case 7
  else if (apmn[0]->IsNeg() && apmn[1]->IsNeg())
  { apmn[1]->NameHeir(0);
    Cycle(LEFT);
  }

  //  cases 3 and 5
  else if (apmn[0]->IsNeg())
  { if (apmn[1]->Compare(apmn[0]->apmn[0])) MakeZero();
    else
    { apmn[0]->NameHeir(0);
      SwapChildren(0,1);
      uNode = TYPE_SUB;
    }
  }

  //  cases 4 and 6
  else if (apmn[1]->IsNeg())
  { if (apmn[0]->Compare(apmn[1]->apmn[0])) MakeZero();
    else
    { apmn[1]->NameHeir(0);
      uNode = TYPE_SUB;
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifySub
// Scope:	 Protected
// Description:	 This function simplifies subtraction math nodes
void MathNode::SimplifySub ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifySub\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = apmn[0]->data.r - apmn[1]->data.r;
    }
    else
    { uNode = TYPE_ICONST;
      ValidateInteger((long) apmn[0]->data.n - (long) apmn[1]->data.n);
    }
    KillChildren();
    return;
  }

  //	the following subtraction expressions are simplified
  //
  //	1)   X	-  0  -->  X
  //	2)   0	-  X  --> -X
  //	3)   0	- -X  -->  X
  //	4)   X	-  X  -->  0
  //	5)  -X	- -Y  --> -(X - Y)
  //	6)  -X	-  Y  --> -(X + Y)
  //	7)   X	- -Y  -->  X + Y

  //  case 1
  if (apmn[1]->IsZero()) NameHeir(0);

  //  cases 2 and 3
  else if (apmn[0]->IsZero())
  { if (apmn[1]->IsNeg())
    { NameHeir(1);
      NameHeir(0);
    }
    else
    { delete apmn[0];
      uNode = TYPE_NEG;
      c = 1;
      apmn[0] = apmn[1];
    }
  }

  //  case 4
  else if (apmn[0]->Compare(apmn[1])) MakeZero();

  //  case 5
  else if (apmn[0]->IsNeg() && apmn[1]->IsNeg())
  { apmn[1]->NameHeir(0);
    Cycle(LEFT);
  }

  //  case 6
  else if (apmn[0]->IsNeg())
  { Cycle(LEFT);
    apmn[0]->uNode = TYPE_ADD;
  }

  //  case 7
  else if (apmn[1]->IsNeg())
  { apmn[1]->NameHeir(0);
    uNode = TYPE_ADD;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyMul
// Scope:	 Protected
// Description:	 This function simplifies multiplication math nodes
void MathNode::SimplifyMul ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyMul\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = apmn[0]->data.r * apmn[1]->data.r;
    }
    else
    { uNode = TYPE_ICONST;
      ValidateInteger((long) apmn[0]->data.n * (long) apmn[1]->data.n);
    }
    KillChildren();
    return;
  }

  //	the following multiplication expressions are simplified
  //
  //	1)   X	    *  0      -->  0
  //	2)   0	    *  X      -->  0
  //	3)   X	    *  1      -->  X
  //	4)   1	    *  X      -->  X
  //	5)   X	    * -1      --> -X
  //	6)  -1	    *  X      --> -X
  //	7)  -X	    *  Y      --> -(X * Y)
  //	8)   X	    * -Y      --> -(X * Y)
  //	9)  -X	    * -Y      -->  X * Y
  //   10)   X	    *  inv(X) -->  1
  //   11)   inv(X) *  X      -->  1
  //   12)   inv(X) *  inv(Y) -->  inv(X * Y)
  //   13)   X	    *  inv(Y) -->  X / Y
  //   14)   inv(X) *  Y      -->  Y / X

  //  cases 1 and 2
  if (apmn[0]->IsZero() || apmn[1]->IsZero()) MakeZero();

  //  cases 3 and 4
  else if (apmn[0]->IsOne()) NameHeir(1);
  else if (apmn[1]->IsOne()) NameHeir(0);

  //  case 5
  else if (apmn[1]->IsMinusOne())
  { delete apmn[1];
    c = 1;
    uNode = TYPE_NEG;
  }

  //  case 6
  else if (apmn[0]->IsMinusOne())
  { delete apmn[0];
    c = 1;
    apmn[0] = apmn[1];
    uNode = TYPE_NEG;
  }

  //  case 9
  else if (apmn[0]->IsNeg() && apmn[1]->IsNeg())
  { apmn[0]->NameHeir(0);
    apmn[1]->NameHeir(0);
    SimplifyMul();
  }

  //  cases 7 and 8
  else if (apmn[0]->IsNeg()) { Cycle(LEFT);   apmn[0]->SimplifyMul(); }
  else if (apmn[1]->IsNeg()) { Cycle(RIGHT);  apmn[0]->SimplifyMul(); }

  //  case 12
  else if (apmn[0]->IsInv() && apmn[1]->IsInv())
  { apmn[1]->NameHeir(0);
    Cycle(LEFT);
  }

  //  cases 10 and 13
  else if (apmn[1]->IsInv())
  { if (apmn[0]->Compare(apmn[1]->apmn[0])) MakeOne();
    else
    { apmn[1]->NameHeir(0);
      uNode = TYPE_DIV;
    }
  }

  //  cases 11 and 14
  else if (apmn[0]->IsInv())
  { if (apmn[1]->Compare(apmn[0]->apmn[0])) MakeOne();
    else
    { apmn[0]->NameHeir(0);
      uNode = TYPE_DIV;
      SwapChildren(0,1);
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyDiv
// Scope:	 Protected
// Description:	 This function simplifies division math nodes
void MathNode::SimplifyDiv ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyDiv\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  //  case 1 ... see below
  if (apmn[1]->IsZero())
  { error->AddMessage(EHM_DIVBYZERO, "/ operator", iPtrNode);
    return;
  }

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = apmn[0]->data.r / apmn[1]->data.r;
    }
    else
    { uNode = TYPE_ICONST;
      ValidateInteger((long) apmn[0]->data.n / (long) apmn[1]->data.n);
    }
    KillChildren();
    return;
  }

  //	the following division expressions are simplified
  //
  //	1)   X	    /  0      -->  error ... checked above
  //	2)   0	    /  X      -->  0
  //	3)   X	    /  1      -->  X
  //	4)   1	    /  X      -->  inv(X)
  //	5)   X	    / -1      --> -X
  //	6)  -X	    / -1      -->  X
  //	7)  -1	    /  X      --> -inv(X)
  //	8)  -1	    / -X      -->  inv(X)
  //	9)   X	    /  X      -->  1
  //   10)  -X	    / -Y      -->  X / Y
  //   11)  -X	    /  Y      --> -(X / Y)
  //   12)   X	    / -Y      --> -(X / Y)
  //   13)   inv(X) / inv(Y)  -->  Y / X
  //   14)   X	    / inv(Y)  -->  X * Y
  //   15)   inv(X) / Y	      -->  inv(X * Y)

  //  case 2
  if (apmn[0]->IsZero()) MakeZero();

  //  cases 3 and 4
  else if (apmn[1]->IsOne()) NameHeir(0);
  else if (apmn[0]->IsOne())
  { delete apmn[0];
    c = 1;
    apmn[0] = apmn[1];
    MakeInv();
    SimplifyFunc();
  }

  //  cases 5 and 6
  else if (apmn[1]->IsMinusOne())
  { if (apmn[0]->IsNeg())
    { NameHeir(0);
      NameHeir(0);
    }
    else
    { delete apmn[1];
      c = 1;
      uNode = TYPE_NEG;
    }
  }

  //  cases 7 and 8
  else if (apmn[0]->IsMinusOne())
  { if (apmn[1]->IsNeg())
    { delete apmn[0];
      c = 1;
      apmn[1]->NameHeir(0);
      apmn[0] = apmn[1];
      MakeInv();
      SimplifyFunc();
    }
    else
    { delete apmn[0];			    //	    DIV		NEG
      MathNode *pmn = new MathNode();	    //	    / \	  --->	 |
      pmn->AddChild(apmn[1]);		    //	  -1   X	INV
      pmn->MakeInv();			    //			 |
      pmn->iPtrNode = iPtrNode;		    //			 X
      pmn->iPtrExpr = iPtrExpr;		    //
      c = 1;				    // DIV  -->	 NEG
      apmn[0] = pmn;			    // INV created, -1 deleted
      uNode = TYPE_NEG;
      apmn[0]->SimplifyFunc();
    }
  }

  //  case 9
  else if (apmn[0]->Compare(apmn[1])) MakeOne();

  //  case 10
  else if (apmn[0]->IsNeg() && apmn[1]->IsNeg())
  { apmn[0]->NameHeir(0);
    apmn[1]->NameHeir(0);
    apmn[0]->SimplifyDiv();
  }

  //  case 11
  else if (apmn[0]->IsNeg())
  { Cycle(LEFT);
    apmn[0]->SimplifyDiv();
  }

  //  case 12
  else if (apmn[1]->IsNeg())
  { Cycle(RIGHT);
    apmn[0]->SimplifyDiv();
  }

  //  case 13
  else if (apmn[0]->IsInv() && apmn[1]->IsInv())
  { apmn[0]->NameHeir(0);
    apmn[1]->NameHeir(0);
    SwapChildren(0,1);
  }

  //  case 14
  else if (apmn[1]->IsInv())
  { apmn[1]->NameHeir(0);
    uNode = TYPE_MUL;
  }

  //  case 15
  else if (apmn[0]->IsInv())
  { Cycle(LEFT);
    apmn[0]->uNode = TYPE_MUL;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyAnd
// Scope:	 Protected
// Description:	 This function simplifies logical and math nodes
void MathNode::SimplifyAnd ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyAnd\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (apmn[0]->data.r && apmn[1]->data.r) ? 1.0 : 0.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (apmn[0]->data.n && apmn[1]->data.n) ? 1 : 0;
    }
    KillChildren();
    return;
  }

  //	the following and expressions are simplified
  //
  //	1)   F	and  X	-->  F
  //	2)   X	and  F	-->  F

  //  cases 1 and 2
  if (apmn[0]->IsZero() || apmn[1]->IsZero())
    MakeZero();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyOr
// Scope:	 Protected
// Description:	 This function simplifies logical or math nodes
void MathNode::SimplifyOr ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyOr\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (apmn[0]->data.r || apmn[1]->data.r) ? 1.0 : 0.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (apmn[0]->data.n || apmn[1]->data.n) ? 1 : 0;
    }
    KillChildren();
    return;
  }

  //	the following and expressions are simplified
  //
  //	1)   T	or  X  -->  T
  //	2)   X	or  T  -->  T

  //  cases 1 and 2
  if (((apmn[0]->uExpr & FLAG_CONST) && !apmn[0]->IsZero()) ||
      ((apmn[1]->uExpr & FLAG_CONST) && !apmn[0]->IsZero()))
    MakeOne();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyXor
// Scope:	 Protected
// Description:	 This function simplifies logical xor math nodes
void MathNode::SimplifyXor ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyXor\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (( apmn[0]->data.r && !apmn[1]->data.r) ||
		(!apmn[0]->data.r &&  apmn[1]->data.r)) ? 1.0 : 0.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (( apmn[0]->data.n && !apmn[1]->data.n) ||
		(!apmn[0]->data.n &&  apmn[1]->data.n)) ? 1 : 0;
    }
    KillChildren();
    return;
  }

  //	the following and expressions are simplified
  //
  //	1)   T	xor  X	-->  not X
  //	2)   X	xor  T	-->  not X

  //  case 1
  if ((apmn[0]->uExpr & FLAG_CONST) && !apmn[0]->IsZero())
  { delete apmn[0];
    apmn[0] = apmn[1];
    c = 1;
    uNode = TYPE_NOT;
  }

  //  case 2
  else if ((apmn[1]->uExpr & FLAG_CONST) && !apmn[0]->IsZero())
  { delete apmn[1];
    c = 1;
    uNode = TYPE_NOT;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyGT
// Scope:	 Protected
// Description:	 This function simplifies greater than math nodes
void MathNode::SimplifyGT ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyGT\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (apmn[0]->data.r > apmn[1]->data.r) ? 1.0 : 0.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (apmn[0]->data.n > apmn[1]->data.n) ? 1 : 0;
    }
    KillChildren();
  }
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyGE
// Scope:	 Protected
// Description:	 This function simplifies greater than or equal to math nodes
void MathNode::SimplifyGE ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyGE\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (apmn[0]->data.r >= apmn[1]->data.r) ? 1.0 : 0.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (apmn[0]->data.n >= apmn[1]->data.n) ? 1 : 0;
    }
    KillChildren();
  }
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyLT
// Scope:	 Protected
// Description:	 This function simplifies less than math nodes
void MathNode::SimplifyLT ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyLT\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (apmn[0]->data.r < apmn[1]->data.r) ? 1.0 : 0.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (apmn[0]->data.n < apmn[1]->data.n) ? 1 : 0;
    }
    KillChildren();
  }
  return;
}


///////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyLE
// Scope:	 Protected
// Description:	 This function simplifies less than or equal to math nodes
void MathNode::SimplifyLE ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyLE\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (apmn[0]->data.r <= apmn[1]->data.r) ? 1.0 : 0.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (apmn[0]->data.n <= apmn[1]->data.n) ? 1 : 0;
    }
    KillChildren();
  }
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyEQ
// Scope:	 Protected
// Description:	 This function simplifies equal to math nodes
void MathNode::SimplifyEQ ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyEQ\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (apmn[0]->data.r == apmn[1]->data.r) ? 1.0 : 0.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (apmn[0]->data.n == apmn[1]->data.n) ? 1 : 0;
    }
    KillChildren();
  }
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyNE
// Scope:	 Protected
// Description:	 This function simplifies not equal to math nodes
void MathNode::SimplifyNE ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyNE\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (apmn[0]->data.r != apmn[1]->data.r) ? 1.0 : 0.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (apmn[0]->data.n != apmn[1]->data.n) ? 1 : 0;
    }
    KillChildren();
  }
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyNeg
// Scope:	 Protected
// Description:	 This function simplifies negation math nodes
void MathNode::SimplifyNeg ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyNeg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = - apmn[0]->data.r;
    }
    else
    { uNode = TYPE_ICONST;
      ValidateInteger(- (long) apmn[0]->data.n);
    }
    KillChildren();
    return;
  }

  //	the following negation expressions are simplified
  //
  //	1)  -(-X)    --> X
  //	2)  -(X - Y) --> Y - X

  //  case 1
  if (apmn[0]->IsNeg())
  { NameHeir(0);
    NameHeir(0);
  }

  //  case 2
  else if (apmn[0]->Is(TYPE_SUB))
  { NameHeir(0);
    SwapChildren(0,1);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyReal
// Scope:	 Protected
// Description:	 This function simplifies convert to real math nodes
void MathNode::SimplifyReal ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyReal\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (apmn[0]->uExpr & FLAG_INT)
    { uNode = TYPE_RCONST;
      data.r = (REAL) apmn[0]->data.n;
    }
    else
    { uNode = TYPE_RCONST;
      data.r = apmn[0]->data.r;
    }
    KillChildren();
    return;
  }

  //	the following convert to real expressions are simplified
  //
  //	1)  real(X) --> X     (assuming X is a real expression)

  //  case 1
  if (apmn[0]->uExpr & FLAG_REAL)
    NameHeir(0);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyNot
// Scope:	 Protected
// Description:	 This function simplifies not math nodes
void MathNode::SimplifyNot ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyNot\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uExpr & FLAG_CONST)
  { if (uExpr & FLAG_REAL)
    { uNode = TYPE_RCONST;
      data.r = (apmn[0]->data.r) ? 0.0 : 1.0;
    }
    else
    { uNode = TYPE_ICONST;
      data.n = (apmn[0]->data.n) ? 0 : 1;
    }
    KillChildren();
    return;
  }

  //	the following logical not expressions are simplified
  //
  //	1)  not(gt(X))	--> le(X)
  //	2)  not(ge(X))	--> lt(X)
  //	3)  not(lt(X))	--> ge(X)
  //	4)  not(le(X))	--> gt(X)
  //	5)  not(eq(X))	--> ne(X)
  //	6)  not(ne(X))	--> eq(X)

  //  case 1
  if (apmn[0]->Is(TYPE_GT))
  { NameHeir(0);
    uNode = TYPE_LE;
  }

  //  case 2
  else if (apmn[0]->Is(TYPE_GE))
  { NameHeir(0);
    uNode = TYPE_LT;
  }

  //  case 3
  else if (apmn[0]->Is(TYPE_LT))
  { NameHeir(0);
    uNode = TYPE_GE;
  }

  //  case 4
  else if (apmn[0]->Is(TYPE_LE))
  { NameHeir(0);
    uNode = TYPE_GT;
  }

  //  case 5
  else if (apmn[0]->Is(TYPE_EQ))
  { NameHeir(0);
    uNode = TYPE_NE;
  }

  //  case 6
  else if (apmn[0]->Is(TYPE_NE))
  { NameHeir(0);
    uNode = TYPE_EQ;
  }

  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyPow
// Scope:	 Protected
// Description:	 This function simplifies pow function math nodes
void MathNode::SimplifyPow ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyPow\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  //	the following exponentiation expressions are simplified
  //
  //	1)   1	^  X  -->  1
  //	2)   X	^  0  -->  1
  //	3)   X	^  1  -->  X
  //	4)   X	^ -1  -->  inv(X)

  //  cases 1 and 2
  if (apmn[0]->IsOne() || apmn[1]->IsZero())  MakeOne();

  //  case 3
  else if (apmn[1]->IsOne()) NameHeir(0);

  //  case 4
  else if (apmn[1]->IsMinusOne())
  { delete apmn[1];
    c = 1;
    MakeInv();
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyFunc
// Scope:	 Protected
// Description:	 This function simplifies function math nodes
void MathNode::SimplifyFunc ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyFunc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // avoid bogus function by Josh
  if (uNode == TYPE_NEG)
  { apmn[0]->SimplifyFunc();
    return;
  }

  // simplify user-defined functions
  if (uNode == TYPE_UDFUNC)
  { UINT iPD = data.udfunc.iPD;
    if (iPD  && !data.udfunc.af[iPD - 1])
      MakeZero();

    // @@@
    if (uExpr & FLAG_CONST)
    { uExpr &= ~FLAG_CONST;
      uExpr |=	FLAG_VAR;
    }
    return;
  }

  // simplify precomputed user-defined functions
  if (uNode == TYPE_PRECOMP)
  { UINT iPD = apmn[0]->data.udfunc.iPD;
    if (iPD  && !apmn[0]->data.udfunc.af[iPD - 1])
      MakeZero();

    // @@@
    if (uExpr & FLAG_CONST) uExpr &= ~FLAG_CONST;
    return;
  }

  // simplify internal functions w/ constant arguments
  if (uExpr & FLAG_CONST)
  { REAL arg1, arg2, result;
    PRF	 prf  = data.func.aprf[data.func.iPD];

    arg1 = apmn[0]->data.r;
    if (c > 1) arg2 = apmn[1]->data.r;

    switch (c)
    { case 1:  result = PRF1(prf)(arg1);	      break;
      case 2:  result = PRF2(prf)(arg1, arg2);	      break;
    }

    KillChildren();
    uNode = TYPE_RCONST;
    uExpr = FLAG_CONST | FLAG_REAL;
    data.r = result;
    return;
  }

  //	if the function is the pow function, then call the SimplifyPow
  //	member function and return

  if (data.func.iFunc == FUNC_POW)
  { SimplifyPow();
    return;
  }

  //	the following subtraction expressions are simplified
  //
  //	1)  Odd functions:			      func(-X) --> -func(X)
  //	2)  Even functions:			      func(-X) -->  func(X)
  //	3)  Function with inverse function argument:  func1(func2(X)) --> X
  //
  //	4)  min(-x,-y) --> -max(x,y)
  //	5)  max(-x,-y) --> -min(x,y)
  //	6)  ave(-x,-y) --> -ave(x,y)

  #define DERIVATIVE(x)	 ((1 << 8) + x)
  #define FUNCTIONS(x,y) ((x << 8) + y)

  if (apmn[0]->IsNeg())
  { if (c == 1)
    { switch ((data.func.iPD << 8) + data.func.iFunc)
      { case FUNC_INV:			  case FUNC_SINH:
	case FUNC_TANH:			  case FUNC_COTH:
	case FUNC_CSCH:			  case FUNC_SIN:
	case FUNC_TAN:			  case FUNC_COT:
	case FUNC_CSC:			  case FUNC_SIGN:
	case FUNC_ASINH:		  case FUNC_ASIN: // by JJF
	case FUNC_ATANH:		  case FUNC_ATAN:
	case FUNC_ACOTH:		  case FUNC_ACOT:
	case FUNC_ACSCH:		  case FUNC_ACSC:
	case DERIVATIVE(FUNC_LN):	  case DERIVATIVE(FUNC_LOG):
	case DERIVATIVE(FUNC_SQ):	  case DERIVATIVE(FUNC_COSH):
	case DERIVATIVE(FUNC_SECH):	  case DERIVATIVE(FUNC_COS):
	case DERIVATIVE(FUNC_SEC):
	case DERIVATIVE(FUNC_ACOSH):	  case DERIVATIVE(FUNC_ACOS): // by JJF
	case DERIVATIVE(FUNC_ASECH):	  case DERIVATIVE(FUNC_ASEC):
	  uNode = TYPE_NEG;		  /* ODD FUNCTIONS */
	  apmn[0]->uNode = TYPE_FUNC;
	  apmn[0]->data.func = data.func;
	  // it's possible to hide some bogus function here (by Josh).
	  break;

	case FUNC_SQ:			  case FUNC_COSH:
	case FUNC_SECH:			  case FUNC_COS:
	case FUNC_SEC:			  case FUNC_ABS:
	case FUNC_ACOSH:		  case FUNC_ACOS: // by JJF
	case FUNC_ASECH:		  case FUNC_ASEC:
	case DERIVATIVE(FUNC_INV):	  case DERIVATIVE(FUNC_SINH):
	case DERIVATIVE(FUNC_TANH):	  case DERIVATIVE(FUNC_COTH):
	case DERIVATIVE(FUNC_CSCH):	  case DERIVATIVE(FUNC_SIN):
	case DERIVATIVE(FUNC_TAN):	  case DERIVATIVE(FUNC_COT):
	case DERIVATIVE(FUNC_CSC):
	case DERIVATIVE(FUNC_ASINH):	  case DERIVATIVE(FUNC_ASIN): // by JJF
	case DERIVATIVE(FUNC_ATANH):	  case DERIVATIVE(FUNC_ATAN):
	case DERIVATIVE(FUNC_ACOTH):	  case DERIVATIVE(FUNC_ACOT):
	case DERIVATIVE(FUNC_ACSCH):	  case DERIVATIVE(FUNC_ACSC):
	  apmn[0]->NameHeir(0);		  /* EVEN FUNCTIONS */
	  break;
      }
    }

    else if ((c == 2) && apmn[1]->IsNeg())
    { switch (data.func.iFunc)
      { case FUNC_MIN:
	case FUNC_MAX:
	case FUNC_AVE:
	  apmn[1]->NameHeir(0);
	  Cycle(LEFT);
	  break;
      }
    }
    return;
  }

  if (apmn[0]->uNode == TYPE_FUNC)
  { switch FUNCTIONS(data.func.iFunc, apmn[0]->data.func.iFunc)
    { case FUNCTIONS(FUNC_EXP, FUNC_LN):
      case FUNCTIONS(FUNC_LN, FUNC_EXP):
      case FUNCTIONS(FUNC_POW10, FUNC_LOG):
      case FUNCTIONS(FUNC_LOG, FUNC_POW10):
      case FUNCTIONS(FUNC_SQ, FUNC_SQRT):
      case FUNCTIONS(FUNC_SQRT, FUNC_SQ):
      case FUNCTIONS(FUNC_INV, FUNC_INV):
	NameHeir(0);
	NameHeir(0);
	break;

      case FUNCTIONS(FUNC_ABS, FUNC_ABS):
      case FUNCTIONS(FUNC_SIGN, FUNC_SIGN):
	NameHeir(0);
	break;

      case FUNCTIONS(FUNC_STEP, FUNC_STEP):
	MakeOne();
	break;
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SimplifyArray
// Scope:	 Protected
// Description:	 This function simplifies array math nodes
void MathNode::SimplifyArray ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("simplify.cpp MathNode::SimplifyArray\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // are the children of the array math expression constant?
  BOOL fElement = TRUE;
  for(i = 0; i < c; ++i) if (apmn[i]->uExpr & FLAG_VAR) fElement = FALSE;

  // yes, they are constant
  // hence the array expression always refers to the same element
  if (fElement)
  { UINT u = apmn[0]->data.n;
    for(i = 1; i < c; ++i)
    { u *= data.array.ac[i];
      u += apmn[i]->data.n;
    }
    data.array.uE = data.array.uA + u;
    KillChildren();
  }
}
