/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     statemnt.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 04/18/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the classes that store and process the equation
 and main procedure statements.

-----------------------------------------------------------------------------*/


#ifndef __statemnt_h
#define __statemnt_h

#ifndef __mathnode_h
#include "mathnode.h"
#endif

#ifndef __index_h
#include "index.h"
#endif


// statement base class
class Statement
{ public:
    UINT uType;				// statement type
    UINT iPtr;				// buffer pointer

  public:
    Statement (UINT, UINT);		// constructor
    virtual ~Statement ();		// destructor
};


// variable assignment statement
class VarAssign : public Statement
{ public:
    UINT i;				// variable index
    UINT u;				// offset
    MathNode *pmn;			// assignment expression

  public:
     VarAssign (UINT, UINT, MathNode *, UINT);
    ~VarAssign ();
};


// array assignment statement
class ArrayAssign : public Statement
{ public:
    UINT i;				// variable index
    UINT u;				// offset
    Indices *pindices;			// array indices
    MathNode *pmn;			// assignment expression

  public:
    ArrayAssign (UINT, UINT, Indices *, MathNode *, UINT);
   ~ArrayAssign ();
};


// solve statement
class Solve : public Statement
{ public:
     Solve (UINT);
    ~Solve ();
};


// write statement
class Write : public Statement
{ public:
    Write (UINT);
   ~Write ();
};


// goto statement
class Goto : public Statement
{ public:
    MathNode *pmn;
    UINT i;

  public:
    Goto (MathNode *, UINT, UINT);
   ~Goto ();
};


// call statement
class Call : public Statement
{ public:
    UINT iProc;
    UINT i;

  public:
    Call (UINT, UINT, UINT);
   ~Call ();
};


// file open statement
class FileOpen : public Statement
{ public:
    UINT u;
    char *psz;

  public:
    FileOpen (char *, UINT, UINT);
   ~FileOpen ();
};


// file close statement
class FileClose : public Statement
{ public:
    FileClose (UINT);
   ~FileClose ();
};


// file goto statement
class FileGoto : public Statement
{ public:
    char *psz;

  public:
    FileGoto (char *, UINT);
   ~FileGoto ();
};


// file label statement
class FileLabel : public Statement
{ public:
    char *psz;

  public:
    FileLabel (char *, UINT);
   ~FileLabel ();
};


// file label statement
class FileComment : public Statement
{ public:
    char *psz;

  public:
    FileComment (char *, UINT);
   ~FileComment ();
};


// file read statement
class FileRead : public Statement
{ public:
    UINT i;
    UINT u;
    UINT c;
    UINT ac[3];

  public:
    FileRead (UINT, UINT, UINT);
    FileRead (UINT, UINT, UINT, UINT *, UINT);
   ~FileRead ();
};


// file read statement
class FileWrite : public Statement
{ public:
    UINT i;
    UINT u;
    UINT c;
    UINT ac[3];

  public:
    FileWrite (UINT, UINT, UINT);
    FileWrite (UINT, UINT, UINT, UINT *, UINT);
   ~FileWrite ();
};


// variable equation statement
class VarEquation : public Statement
{ public:
    UINT i;				// variable index
    UINT u;				// offset
    MathNode *pmn;			// equation expression

  public:
    UINT cContrib;			// number of contributions to Jacobian
    MathNode **apmnElem;		// contribution elements
    MathNode **apmnDiff;		// contribution expressions

  public:
     VarEquation (UINT, UINT, MathNode *, UINT);
    ~VarEquation ();
};


// array equation statement
class ArrayEquation : public Statement
{ public:
    UINT i;				// array index
    UINT u;				// offset
    Indices *pindices;			// array indices
    MathNode *pmn;			// equation expression

  public:
    UINT cContrib;			// number of contributions to Jacobian
    MathNode **apmnElem;		// contribution elements
    MathNode **apmnDiff;		// contribution expressions

  public:
     ArrayEquation (UINT, UINT, Indices *, MathNode *, UINT);
    ~ArrayEquation ();
};


class DAPStatement
{ private:
    UINT cSize;				// total number of elements
    UINT uInc;				// increment size
    UINT cElements;			// number of elements in array
    BOOL fManaged;			// managed array
    Statement **dynarray;		// dynamic array

  public:
    void SetManaged (BOOL f)		{ fManaged = f; }
    UINT GetElements ()			{ return cElements; }
    UINT Add (Statement *t);		// add an item to the array
    void RemoveLast ();			// remove last element
    void Remove(UINT i);		// remove an element
    void Flush ();			// remove all pointers from the array
    Statement * operator[](UINT i)	{ return dynarray[i]; }

  public:
     DAPStatement (UINT = 16);
    ~DAPStatement ();
};


#if !defined(P16BIT)
#ifndef __da_uint_h
#include "da_uint.h"
#endif
#endif


class IntStack
{ private:
    UINT u;		      // increment
    UINT c;		      // stack size
    UINT i;		      // stack index
    int *stack;		      // stack
    void grow ()	      { c+= u; stack = (int *) realloc(stack, c * sizeof(int)); }

  public:
    void push  (int T)	      { if (i == c) grow(); stack[i++] = T; }
    int pop    ()	      { return stack[--i]; }
    BOOL empty ()	      { return !i; }
    void flush ()	      { i = 0; }

  public:
     IntStack (UINT n = 64)   { i = 0; stack = (int*) calloc((c = u = n),
                                               sizeof(int)); }
    ~IntStack ()	      { free((void *) stack); }
};


#ifndef __xlatinfo_h
#include "xlatinfo.h"
#endif


class Statements
{ private:
    DAPStatement DAPMainProc;		// main procedure statements
    DAPStatement DAPEquation;		// equation statements
    DAPMathNode	 DAPMathExpr;		// variable or array math nodes
    UINT cTag;				// # of unknowns
    UINT cTop;				// # of entries in topology array
    UINT cElem;				// # of variables and array elements

    TAG	 *atag;				// tag information array
    TOP	 *atop;				// topology information array
    UINT *aiTag;			// tag index array
#if defined(P16BIT)
    UINT *aiTop;			// topology index array
    UINT *abfTag;			// bit flags for user tagging
#else
    DAUINT DAbfTag;
    DAUINT DAiTop;
#endif
    I3ARRAY *aanEI;			// equation indices

    BOOL fError;			// error flag
    IntStack stack;			// index evaluation stack
    Indices *pindices;			// current indices
    Indices *pindicesNULL;		// null indices

    int nTag1 ();			// tag unknown variables
    int nTag2 ();			// tag unknown variables
    void Allocate ();			// allocate arrays
    BOOL VarOK (MathNode *);		// verify array bounds
    int nDiffInfo (UINT, BOOL);		// verify/compute differentiation info
    void ComputeContribs (Statement *, UINT);	// compute contributions to the Jacobian
    BOOL AddTopologyIndex (UINT, UINT, UINT &); // add index to topology info

  public:
    inline void	 AddMain      (Statement *p) { DAPMainProc.Add(p); }
    inline void	 AddEqu	      (Statement *p) { DAPEquation.Add(p); }
    inline void	 AddVar	      (MathNode	 *p) { DAPMathExpr.Add(p); }
    inline void	 PushIndex    (UINT i)	     { stack.push(pindices->GetCurrent(i)); }
    inline void	 PushNum      (int  n)	     { stack.push(n); }
    inline int	 PopNum	      ()	     { return stack.pop(); }
    inline void	 ClearFlag    ()	     { fError = FALSE; }
#if defined(P16BIT)
    inline void	 SetBitFlag   (UINT i)	     { abfTag[i>>4] |=	(1 << (i & 0xF)); }
    inline void	 ClearBitFlag (UINT i)	     { abfTag[i>>4] &= ~(1 << (i & 0xF)); }
    inline BOOL	 IsBitFlagSet (UINT i)	     { return abfTag[i>>4] & (1 << (i & 0xF)); }
#else
    void SetBitFlag   (UINT i);
    void ClearBitFlag (UINT i);
    BOOL IsBitFlagSet (UINT i);
#endif
    inline UINT	 GetNumOfElem ()	     { return cElem; }
    inline UINT *GetTagVector ()	     { return aiTag; }
    inline UINT	 GetNumOfEqu  ()	     { return DAPEquation.GetElements(); }
    inline UINT	 GetNumOfStatements ()	     { return DAPMainProc.GetElements(); }

    void Operator (UINT, UINT = -1u);	// perform the specified operation
    void AllocBitFlags ();		// allocate bit flags
    void SetGotoIndex (UINT, UINT);	// set the index of a goto statement
    BOOL ProcessMain ();		// verify assignment statements
    BOOL ProcessEquations (UINT);	// process equation statements

    void WriteTopFile  (int, TOPHEAD *);  // write topology file
    void CodeProcedure (char *, void *, UINT);	// code a procedure
    void DeclareProcedureSums (void *, UINT);	// write node and label sums
    void CodeEquations (void *);	// code equations
    void TestForUserDefProc ();		// test if procedure is user-def?

  public:
     Statements ();			// constructor
    ~Statements ();			// destructor
};


// external definition
extern Statements *pstatements;


#endif
