/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     symbol.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/18/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGFramework syntax
 10/20/94    KMK      modified for command line version

 DESCRIPTION:

 This module implements the SGFramework symbol table.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "types.h"
#include "error.h"
#include "symbol.h"
#include "mathfunc.h"


// global declaration
SymbolTable *symTable = NULL;


///////////////////////////////////////////////////////////////////////////////
///////////////////  MEMBER FUNCTIONS OF THE SYMBOL CLASSES  //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Description:
SymIndex::SymIndex ( char *pszName, UINT iIndex )
	: HashNode ( pszName, TYPE_INDEX )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymIndex::SymIndex\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iIndex;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymIConst::SymIConst ( char *pszName, int nValue )
	 : HashNode  ( pszName, TYPE_ICONST )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymIConst::SymIConst\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  n = nValue;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymRConst::SymRConst ( char *pszName, REAL rValue )
	 : HashNode  ( pszName, TYPE_RCONST )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymRConst::SymRConst\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  r = rValue;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymArg::SymArg	 ( char *psz, UINT iVar, BOOL fMayBeUnknown )
      : HashNode ( psz, TYPE_ARG )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymArg::SymArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iVar;
  f = fMayBeUnknown;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymVar::SymVar	 ( char *psz, UINT iVar, UINT uOff )
      : HashNode ( psz, TYPE_VAR )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymVar::SymVar\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iVar;
  f = FALSE;
  u = uOff;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymLocalVar::SymLocalVar ( char *pszName, UINT iLocalVar, MathNode *pmnExpr )
	   : HashNode	 ( pszName, TYPE_LVAR )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymLocalVar::SymLocalVar\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iLocalVar;
  pmn = pmnExpr;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymLocalVar::~SymLocalVar ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymLocalVar::~SymLocalVar\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (pmn) delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymArray::SymArray ( char *psz, UINT iVar, UINT cDim,
                     UINT a1, UINT a2, UINT a3, UINT uOff, BOOL fMV )
	: HashNode ( psz, TYPE_ARRAY )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymArray::SymArray\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iVar;
  c = cDim;
  f = FALSE;
  u = uOff;
  fMeshVar = fMV;
  ac[0] = a1;
  ac[1] = a2;
  ac[2] = a3;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymLabel::SymLabel ( char *psz, UINT iLabel, UINT cNodes, UINT *auNodeList )
	: HashNode ( psz, TYPE_LABEL )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymLabel::SymLabel\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iLabel;
  c = cNodes;
  au = auNodeList;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymLabel::~SymLabel ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymLabel::~SymLabel\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (au) delete [] au; // ?fengj
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymFunc::SymFunc  ( char *pszName, UINT iFunc, UINT cArgs, PRF *aprfList )
       : HashNode ( pszName, TYPE_FUNC )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymFunc::SymFunc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iFunc;
  c = cArgs;
  aprf[0] = aprfList[0];
  aprf[1] = aprfList[1];
  aprf[2] = aprfList[2];
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymUDFunc::SymUDFunc ( char *pszName, UINT iFunc, UINT cArgs,
                       BOOL *afMayBeUnknown, UINT iSource )
	 : HashNode  ( pszName, TYPE_UDFUNC )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymUDFunc::SymUDFunc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iFunc;
  c = cArgs;
  f = FALSE;
  af = afMayBeUnknown;
  iPtr = iSource;
  apsym = NULL;
  cAssign = 0;
  auArg1 = NULL;
  auArg2 = NULL;
  pmnTemplate = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymUDFunc::~SymUDFunc ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymUDFunc::~SymUDFunc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  if (af) delete [] af;
  if (auArg1) delete [] auArg1;
  if (auArg2) delete [] auArg2;
  if (pmnTemplate) delete pmnTemplate;
  if (apsym)
    for(i = 0; i < cAssign; ++i)
      delete apsym[i];

  // sould we delete [] apsym here? Josh
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function initializes the member data that is used for
//		 precomputation of functions.
void SymUDFunc::InitPrecompute ( BOOL f1, BOOL f2, MathNode *pmn )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymUDFunc::InitPrecompute\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i, j;

  fNode = f1;
  fOdd	= f2;
  pmnTemplate = pmn;
  UINT iIndex = 1;
  auArg1 = new UINT[c];	       // @@@ what if c = 0
  auArg2 = new UINT[c];	       // @@@ what if c = 0
  for (i = 0; i < c; ++i)
  { auArg1[i] = (af[i]) ? iIndex++ : -1u;
    auArg2[i] = (af[i]) ? 0 : -1u;
  }

  if (fNode) return;

  MathNode *pmnFlip = pmn->Copy();
  pmnFlip->FlipTemplateIndices();
  for (i = 0; i < c; ++i)
  { if (auArg2[i]) continue;
    for (j = i+1; j < c; ++j)
    { if (auArg2[j]) continue;
      if (pmn->apmn[i]->Compare(pmnFlip->apmn[j]))
      { auArg2[i] = auArg1[j];
	auArg2[j] = auArg1[i];
	break;
      }
    }
    if (!auArg2[i])
    { pmn->apmn[i]->OddOrEven(pmnFlip->apmn[i]);
      if      (pmn->apmn[i]->uExpr & FLAG_EVEN) auArg2[i] = +auArg1[i];
      else if (pmn->apmn[i]->uExpr & FLAG_ODD)	auArg2[i] = -auArg1[i];
      else
      { error->AddMessage(EHM_ODDEVEN, pmn->apmn[i]->iPtrExpr);
      }
    }
  }
  delete pmnFlip;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymProc::SymProc  ( char *pszName, UINT iProcedure, UINT iIndex, BOOL f )
       : HashNode ( pszName, TYPE_PROC )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymProc::SymProc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  iProc = iProcedure;
  i = iIndex;
  fExt = f;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymNodeSum::SymNodeSum ( char *pszName, UINT iIndex, MathNode *pmnExpr )
	  : HashNode   ( pszName, TYPE_NSUM )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymNodeSum::SymNodeSum\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iIndex;
  pmn = pmnExpr;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymLabelSum::SymLabelSum ( char *pszName, UINT iIndex, MathNode *pmnExpr )
	   : HashNode	 ( pszName, TYPE_LSUM )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymLabelSum::SymLabelSum\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iIndex;
  pmn = pmnExpr;
}


///////////////////////////////////////////////////////////////////////////////
/////////////////  MEMBER FUNCTIONS OF THE SYMBOL TABLE CLASS  ////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Description:
SymbolTable::SymbolTable ( UINT c, BOOL fAddFuncs ) : HashTable ( c )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymbolTable::SymbolTable\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  for(i = 0; i < SYMBOLS; ++i)
    DAP[i]  = new DAPHashNode;
  if (fAddFuncs) AddFunctions();
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymbolTable::~SymbolTable ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymbolTable::~SymbolTable\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  for (UINT i = 0; i < SYMBOLS; ++i)
    delete DAP[i];
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds the built-in functions to the symbol table.
void SymbolTable::AddFunctions ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymbolTable::AddFunctions\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  for (UINT i = 0; i < cFuncs; ++i)
  { SymFunc *psym = new SymFunc(aift[i].sz, i, aift[i].c, aift[i].aprf);
    Store(psym);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function stores a symbol in the symbol table.
void SymbolTable::Store ( HashNode *phnode )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymbolTable::Store\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  DAP[phnode->uType]->Add(phnode);
  HashTable::Store(phnode);
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function removes the ith symbol of type u.
void SymbolTable::RemoveLast ( UINT u )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymbolTable::RemoveLast\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i = (*DAP[u]).GetElements();
  if (i)
  { char *psz = (*DAP[u])[i-1]->szName;
    Remove(psz);
    (*DAP[u]).RemoveLast();
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function removes the symbols of type uType from the
//		 symbol table.
void SymbolTable::RemoveSymbols ( UINT uType )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("symbol.cpp SymbolTable::RemoveSymbols\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT c = DAP[uType]->GetElements();
  for (UINT i = 0; i < c; ++i)
  { char *psz = (*DAP[uType])[i]->szName;
    Remove(psz);
  }
  DAP[uType]->Flush();
}


