/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     symbol.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/18/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the SGFramework symbol table.

-----------------------------------------------------------------------------*/


#ifndef __symbol_h
#define __symbol_h

#ifndef __types_h
#include "types.h"
#endif

#ifndef __hash_h
#include "hash.h"
#endif

#ifndef __statemnt_h
#include "statemnt.h"
#endif


// Index Symbol
class SymIndex : public HashNode
{ public:
    UINT i;				// index's index
    SymIndex (char *, UINT);
};


// Integer Constant Symbol
class SymIConst : public HashNode
{ public:
    int n;				// constant's value
    SymIConst (char *, int);
};


// Real Constant Symbol
class SymRConst : public HashNode
{ public:
    REAL r;				// constant's value
    SymRConst (char *, REAL);
};


// Argument Symbol
class SymArg : public HashNode
{ public:
    UINT i;				// argument index
    BOOL f;				// may refer to an unknown
    SymArg (char *, UINT, BOOL);
};


// Local Variable Symbol
class SymLocalVar : public HashNode
{ public:
    UINT i;				// local variable index
    MathNode *pmn;			// assignment statement
    SymLocalVar (char *, UINT, MathNode *);
   ~SymLocalVar ();
};


// Variable Symbol
class SymVar : public HashNode
{ public:
    UINT i;				// variable index
    BOOL f;				// unknown variable
    UINT u;				// offset
    SymVar (char *, UINT, UINT);
};


// Array Symbol
class SymArray : public HashNode
{ public:
    UINT i;				// variable index
    UINT c;				// number of dimensions
    UINT ac[3];				// elements per dimension
    BOOL f;				// contains unknown array element(s)
    BOOL fMeshVar;			// mesh variable
    UINT u;				// offset
    SymArray (char *, UINT, UINT, UINT, UINT, UINT, UINT, BOOL = FALSE);
};


// Label Symbol
class SymLabel : public HashNode
{ public:
    UINT i;				// label index
    UINT c;				// number of nodes in list
    UINT *au;				// node list
    SymLabel(char *, UINT, UINT, UINT *);
   ~SymLabel();
};


// Function Symbol
class SymFunc : public HashNode
{ public:
    UINT i;				// function's index
    UINT c;				// number of arguments
    PRF	 aprf[3];			// pointer to function and derivatives
    SymFunc (char *, UINT, UINT, PRF *);
};


// User-defined Function Symbol
class SymUDFunc : public HashNode
{ public:
    UINT i;				// user-defined function's index
    UINT c;				// number of arguments
    UINT iPtr;				// source index of function
    UINT cAssign;			// number of assignment statements
    BOOL f;				// declaration complete
    BOOL *af;				// may refer to unknown flag array
    SymLocalVar **apsym;		// local assignment statements

  public:				// the following is for precomputing
    BOOL fNode;				// precompute function at nodes
    BOOL fOdd;				// odd edge function w/ respect to nodes
    UINT *auArg1;			// argument to index array 1
    UINT *auArg2;			// argument to index array 2
    MathNode *pmnTemplate;		// precomputation template
    void InitPrecompute (BOOL, BOOL, MathNode *);

  public:
    SymUDFunc(char *, UINT, UINT, BOOL *, UINT = -1u);
   ~SymUDFunc();
};


// Procedure Symbol
class SymProc : public HashNode
{ public:
    UINT iProc;				// procedure index
    UINT i;				// statement index
    BOOL fExt;				// external procedure
    SymProc (char *, UINT, UINT, BOOL = FALSE);
};


// Node Summation Expression Symbol
class SymNodeSum : public HashNode
{ public:
    UINT i;				// node summation expression index
    MathNode *pmn;			// node summation expression
    SymNodeSum (char *, UINT, MathNode *);
};


// Label Summation Expression Symbol
class SymLabelSum : public HashNode
{ public:
    UINT i;				// label summation expression index
    MathNode *pmn;			// label summation expression
    SymLabelSum (char *, UINT, MathNode *);
};


// This class implements the SGFramework symbol table.
//   Store() stores a symbol in the symbol table
//   Remove() removes a symbol from the symbol table
//   RemoveSymbols() removes every symbol of a given type from the symbol table
//   Retrieve() retrieves a symbol from the symbol table
//   cSymbols() returns the number of specified tokens in the symbol table
class SymbolTable : public HashTable
{ protected:
    DAPHashNode *DAP[SYMBOLS];
    void  AddFunctions ();

  public:
    void Store (HashNode *);
    void Remove (char *psz)  { HashTable::Remove(psz); }
    void RemoveLast (UINT);
    void RemoveSymbols (UINT);
    inline void FlushSymbolList (UINT u)    { DAP[u]->Flush(); }
    inline UINT cSymbols	(UINT u)    { return DAP[u]->GetElements(); }
    inline HashNode * Retrieve  (char *psz) { return HashTable::Retrieve(psz); }
    inline HashNode * Retrieve  (UINT u, UINT i) { return (*DAP[u])[i]; }

  public:
    SymbolTable (UINT = 397, BOOL = FALSE);
   ~SymbolTable ();
};


// global reference to the symbol table instance
extern SymbolTable *symTable;


#endif




