/*-----------------------------------------------------------------------------

 FILE:	   CMDPARSE.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/19/93    KMK      initial implementation

 DESCRIPTION:

 This module contains functions that parse the command line module.  Arguments
 may be accessed by their position or switch.  Case is sensitive.

-----------------------------------------------------------------------------*/


#include <ctype.h>		    /* character type functions */
#include <stdio.h>		    /* standard input and output functions */
#include <stdlib.h>		    /* standard library functions */
#include <string.h>		    /* string manipulation functions */

#include "typedefs.h"		    /* type definitions and global constants */
#include "error.h"
#include "literals.h"

static WORD  cSwitchArgs;	    /* number of switch arguments      */
static PSTR *apszSwitchArg;	    /* switch argument pointer array   */
static WORD  cPosArgs;		    /* number of positional arguments  */
static PSTR *apszPosArg;	    /* positional argument index array */


#if defined(ANSIPROTO)		    /* private function prototypes */
BOOL fSwitchArgument(PSTR);
#else
BOOL fSwitchArgument();
#endif


/*****************************************************************************/
/**************************  PUBLIC FUNCTIONS  *******************************/
/*****************************************************************************/


/* vInitCmdParser:  This function initializes the global variables of the
 * command line parser module.	In addition, the position arguments are
 * noted and ordered.
 */
#if defined(ANSIPROTO)
void
vInitCmdParser(
int   cArgs,			    /* number of arguments including program */
PSTR *apszArg)			    /* array of pointer to command line args */
#else
void
vInitCmdParser(cArgs, apszArg)
int   cArgs;
PSTR *apszArg;
#endif
{
  WORD i;			  /* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.fSwitchArgument\n");
#endif

  /* determine number of switch and position arguments */
  cSwitchArgs = 0;
  cPosArgs    = 0;
  for(i = 1; i < cArgs; ++i)
    if (fSwitchArgument(apszArg[i])) ++cSwitchArgs;
    else			     ++cPosArgs;

  /* allocate pointer arrays for the switch and position arguments */
  apszSwitchArg = apszPosArg = NULL;
  if (cSwitchArgs)
  { apszSwitchArg = (PSTR *) calloc(cSwitchArgs, sizeof(PSTR));
    MEMCHECK(apszSwitchArg, LIT_CMDPARSE, LIT_CMDPARSE1);
  }
  if (cPosArgs)
  { apszPosArg	  = (PSTR *) calloc(cPosArgs,	 sizeof(PSTR));
    MEMCHECK(apszPosArg,    LIT_CMDPARSE, LIT_CMDPARSE1);
  }


  /* store the pointers to the switch and position arguments */
  cSwitchArgs = 0;
  cPosArgs    = 0;
  for(i = 1; i < cArgs; ++i)
    if (fSwitchArgument(apszArg[i]))
      apszSwitchArg[cSwitchArgs++] = apszArg[i];
    else
      apszPosArg[cPosArgs++]	   = apszArg[i];

  return;

} /* vInitCmdParser */


/* ----------------------------------------------------------------------------
 * vDestroyCmdParser:  This function deallocates the data structures used by
 * the command line parser module.
 */
void
vDestroyCmdParser()
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.vDestroyCmdParser\n");
#endif

  cSwitchArgs = 0;
  cPosArgs    = 0;
  free(apszSwitchArg);
  free(apszPosArg);
  return;

}


/* ----------------------------------------------------------------------------
 * wNumOfArgs:	The function returns the number of command line arguments.
 */
WORD
wNumOfArgs()
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.wNumOfArgs\n");
#endif

  return(cSwitchArgs + cPosArgs);

} /* wNumOfArgs */


/* ----------------------------------------------------------------------------
 * wNumOfSwitchArgs:  The function returns the number of switch command line
 * arguments.
 */
WORD
wNumOfSwitchArgs()
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.wNumOfSwitchArgs\n");
#endif

  return(cSwitchArgs);

} /* wNumOfSwitchArgs */


/* ----------------------------------------------------------------------------
 * wNumOfPosArgs:  The function returns the number of positional command line
 * arguments.
 */
WORD
wNumOfPosArgs()
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.wNumOfPosArgs\n");
#endif

  return(cPosArgs);

} /* wNumOfPosArgs */


/* ----------------------------------------------------------------------------
 * fParamExist:	 This function returns TRUE if the specified switch argument
 * exists; otherwise, FALSE is returned.
 */
#if defined(ANSIPROTO)
BOOL
fParamExist(
PSTR pszParam)			    /* specified parameter */
#else
BOOL
fParamExist(pszParam)
PSTR pszParam;
#endif
{
  PSTR *ppsz;			  /* pointer to a string pointer */
  WORD i;			  /* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.fParamExist\n");
#endif

  for(i = 0, ppsz = apszSwitchArg; i < cSwitchArgs; ++i, ++ppsz)
    if (!strncmp(*ppsz + 1, pszParam, strlen(pszParam)))
      return(TRUE);

  return(FALSE);

} /* fParamExist */


/* ----------------------------------------------------------------------------
 * nGetSwitchParam:  This function returns the integer parameter associated
 * with the specified switch parameter.
 */
#if defined(ANSIPROTO)
int
nGetSwitchParam(
PSTR pszParam,			    /* specified parameter */
int  nDefault,			    /* default parameter */
int  nMin,			    /* minimum value */
int  nMax)			    /* maximum value */
#else
int
nGetSwitchParam(pszParam, nDefault, nMin, nMax)
PSTR pszParam;
int  nDefault;
int  nMin;
int  nMax;
#endif
{
  PSTR *ppsz;			  /* specified switch parameter */
  WORD i;			  /* loop index */
  int  nValue;			  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.nGetSwitchParam\n");
#endif

  /* search for the specified switch parameter */
  for(i = 0, ppsz = apszSwitchArg; i < cSwitchArgs; ++i, ++ppsz)
    if (!strncmp(*ppsz + 1, pszParam, strlen(pszParam)))
    {
      nValue = atoi(*ppsz + strlen(pszParam) + 1);
      if (nValue < nMin) nValue = nMin;
      if (nValue > nMax) nValue = nMax;
      return(nValue);
    }

  return(nDefault);   /* return default if parameter does not exist */

} /* nGetSwitchParam */


/* ----------------------------------------------------------------------------
 * wGetSwitchParam:  This function returns the word parameter associated
 * with the specified switch parameter.
 */
#if defined(ANSIPROTO)
WORD
wGetSwitchParam(
PSTR pszParam,			    /* specified parameter */
WORD wDefault,			    /* default parameter */
WORD wMin,			    /* minimum value */
WORD wMax)			    /* maximum value */
#else
WORD
wGetSwitchParam(pszParam, wDefault, wMin, wMax)
PSTR pszParam;
WORD wDefault;
WORD wMin;
WORD wMax;
#endif
{
  PSTR *ppsz;			  /* specified switch parameter */
  WORD i;			  /* loop index */
  WORD wValue;			  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.wGetSwitchParam\n");
#endif

  /* search for the specified switch parameter */
  for(i = 0, ppsz = apszSwitchArg; i < cSwitchArgs; ++i, ++ppsz)
    if (!strncmp(*ppsz + 1, pszParam, strlen(pszParam)))
    {
      wValue = (WORD) atoi(*ppsz + strlen(pszParam) + 1);
      if (wValue < wMin) wValue = wMin;
      if (wValue > wMax) wValue = wMax;
      return(wValue);
    }

  return(wDefault);   /* return default if parameter does not exist */

} /* wGetSwitchParam */


/* ----------------------------------------------------------------------------
 * lGetSwitchParam:  This function returns the long parameter associated
 * with the specified switch parameter.
 */
#if defined(ANSIPROTO)
long
lGetSwitchParam(
PSTR pszParam,			    /* specified parameter */
long lDefault,			    /* default parameter */
long lMin,			    /* minimum value */
long lMax)			    /* maximum value */
#else
long
lGetSwitchParam(pszParam, lDefault, lMin, lMax)
PSTR pszParam;
long lDefault;
long lMin;
long lMax;
#endif
{
  PSTR *ppsz;			  /* specified switch parameter */
  WORD i;			  /* loop index */
  long lValue;			  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.lGetSwitchParam\n");
#endif

  /* search for the specified switch parameter */
  for(i = 0, ppsz = apszSwitchArg; i < cSwitchArgs; ++i, ++ppsz)
    if (!strncmp(*ppsz + 1, pszParam, strlen(pszParam)))
    {
      lValue = atol(*ppsz + strlen(pszParam) + 1);
      if (lValue < lMin) lValue = lMin;
      if (lValue > lMax) lValue = lMax;
      return(lValue);
    }

  return(lDefault);   /* return default if parameter does not exist */

} /* lGetSwitchParam */


/* ----------------------------------------------------------------------------
 * dwGetSwitchParam:  This function returns the double word parameter
 * associated with the specified switch parameter.
 */
#if defined(ANSIPROTO)
DWORD
dwGetSwitchParam(
PSTR  pszParam,			    /* specified parameter */
DWORD dwDefault,		    /* default parameter */
DWORD dwMin,			    /* minimum value */
DWORD dwMax)			    /* maximum value */
#else
DWORD
dwGetSwitchParam(pszParam, dwDefault, dwMin, dwMax)
PSTR  pszParam;
DWORD dwDefault;
DWORD dwMin;
DWORD dwMax;
#endif
{
  PSTR	*ppsz;			  /* specified switch parameter */
  WORD	i;			  /* loop index */
  DWORD dwValue;		  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.dwGetSwitchParam\n");
#endif

  /* search for the specified switch parameter */
  for(i = 0, ppsz = apszSwitchArg; i < cSwitchArgs; ++i, ++ppsz)
    if (!strncmp(*ppsz + 1, pszParam, strlen(pszParam)))
    {
      dwValue = (DWORD) atol(*ppsz + strlen(pszParam) + 1);
      if (dwValue < dwMin) dwValue = dwMin;
      if (dwValue > dwMax) dwValue = dwMax;
      return(dwValue);
    }

  return(dwDefault);  /* return default if parameter does not exist */

} /* dwGetSwitchParam */


/* ----------------------------------------------------------------------------
 * dGetSwitchParam:  This function returns the double parameter associated
 * with the specified switch parameter.
 */
#if defined(ANSIPROTO)
double
dGetSwitchParam(
PSTR   pszParam,		    /* specified parameter */
double dDefault,		    /* default parameter */
double dMin,			    /* minimum value */
double dMax)			    /* maximum value */
#else
double
dGetSwitchParam(pszParam, dDefault, dMin, dMax)
PSTR   pszParam;
double dDefault;
double dMin;
double dMax;
#endif
{
  PSTR	 *ppsz;			  /* specified switch parameter */
  WORD	 i;			  /* loop index */
  double dValue;		  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.dGetSwitchParam\n");
#endif

  /* search for the specified switch parameter */
  for(i = 0, ppsz = apszSwitchArg; i < cSwitchArgs; ++i, ++ppsz)
    if (!strncmp(*ppsz + 1, pszParam, strlen(pszParam)))
    {
      dValue = atof(*ppsz + strlen(pszParam) + 1);
      if (dValue < dMin) dValue = dMin;
      if (dValue > dMax) dValue = dMax;
      return(dValue);
    }

  return(dDefault);   /* return default if parameter does not exist */

} /* dGetSwitchParam */


/* ----------------------------------------------------------------------------
 * pszGetSwitchParam:  This function returns the string parameter associated
 * with the specified switch parameter.
 */
#if defined(ANSIPROTO)
PSTR
pszGetSwitchParam(
PSTR pszParam,			    /* specified parameter */
PSTR pszDefault)		    /* default parameter */
#else
PSTR
pszGetSwitchParam(pszParam, pszDefault)
PSTR pszParam;
PSTR pszDefault;
#endif
{
  PSTR	 *ppsz;			  /* specified switch parameter */
  WORD	 i;			  /* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.pszGetSwitchParam\n");
#endif

  /* search for the specified switch parameter */
  for(i = 0, ppsz = apszSwitchArg; i < cSwitchArgs; ++i, ++ppsz)
    if (!strncmp(*ppsz + 1, pszParam, strlen(pszParam)))
      return(*ppsz + strlen(pszParam) + 1);

  return(pszDefault);	  /* return default if parameter does not exist */

} /* pszGetSwitchParam */


/* ----------------------------------------------------------------------------
 * nGetPosParam:  This function returns the integer parameter in the specified
 * position.
 */
#if defined(ANSIPROTO)
int
nGetPosParam(
WORD wPosition,			    /* specified position */
int  nDefault,			    /* default parameter */
int  nMin,			    /* minimum value */
int  nMax)			    /* maximum value */
#else
int
nGetPosParam(wPosition, nDefault, nMin, nMax)
WORD wPosition;
int  nDefault;
int  nMin;
int  nMax;
#endif
{
  int  nValue;			  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.nGetPosParam\n");
#endif

  /* if the specified position is illegal, then return the default */
  if ((wPosition < 1) || (wPosition > cPosArgs)) return(nDefault);

  /* validate and return the parameter specified in the specified position */
  nValue = atoi(apszPosArg[wPosition - 1]);
  if (nValue < nMin) nValue = nMin;
  if (nValue > nMax) nValue = nMax;
  return(nValue);

} /* nGetPosParam */


/* ----------------------------------------------------------------------------
 * wGetPosParam:  This function returns the word parameter in the specified
 * position.
 */
#if defined(ANSIPROTO)
WORD
wGetPosParam(
WORD wPosition,			    /* specified position */
WORD wDefault,			    /* default parameter */
WORD wMin,			    /* minimum value */
WORD wMax)			    /* maximum value */
#else
WORD
wGetPosParam(wPosition, wDefault, wMin, wMax)
WORD wPosition;
WORD wDefault;
WORD wMin;
WORD wMax;
#endif
{
  WORD wValue;			  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.wGetPosParam\n");
#endif

  /* if the specified position is illegal, then return the default */
  if ((wPosition < 1) || (wPosition > cPosArgs)) return(wDefault);

  /* validate and return the parameter specified in the specified position */
  wValue = atoi(apszPosArg[wPosition - 1]);
  if (wValue < wMin) wValue = wMin;
  if (wValue > wMax) wValue = wMax;
  return(wValue);

} /* wGetPosParam */


/* ----------------------------------------------------------------------------
 * lGetPosParam:  This function returns the long parameter in the specified
 * position.
 */
#if defined(ANSIPROTO)
long
lGetPosParam(
WORD wPosition,			    /* specified position */
long lDefault,			    /* default parameter */
long lMin,			    /* minimum value */
long lMax)			    /* maximum value */
#else
long
lGetPosParam(wPosition, lDefault, lMin, lMax)
WORD wPosition;
long lDefault;
long lMin;
long lMax;
#endif
{
  long lValue;			  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.lGetPosParam\n");
#endif

  /* if the specified position is illegal, then return the default */
  if ((wPosition < 1) || (wPosition > cPosArgs)) return(lDefault);

  /* validate and return the parameter specified in the specified position */
  lValue = atol(apszPosArg[wPosition - 1]);
  if (lValue < lMin) lValue = lMin;
  if (lValue > lMax) lValue = lMax;
  return(lValue);

} /* lGetPosParam */


/* ----------------------------------------------------------------------------
 * dwGetPosParam:  This function returns the double word parameter in the
 * specified position.
 */
#if defined(ANSIPROTO)
DWORD
dwGetPosParam(
WORD  wPosition,		    /* specified position */
DWORD dwDefault,		    /* default parameter */
DWORD dwMin,			    /* minimum value */
DWORD dwMax)			    /* maximum value */
#else
DWORD
dwGetPosParam(wPosition, dwDefault, dwMin, dwMax)
WORD  wPosition;
DWORD dwDefault;
DWORD dwMin;
DWORD dwMax;
#endif
{
  DWORD dwValue;		  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.dwGetPosParam\n");
#endif

  /* if the specified position is illegal, then return the default */
  if ((wPosition < 1) || (wPosition > cPosArgs)) return(dwDefault);

  /* validate and return the parameter specified in the specified position */
  dwValue = (DWORD) atol(apszPosArg[wPosition - 1]);
  if (dwValue < dwMin) dwValue = dwMin;
  if (dwValue > dwMax) dwValue = dwMax;
  return(dwValue);

} /* dwGetPosParam */


/* ----------------------------------------------------------------------------
 * dGetPosParam:  This function returns the double parameter in the specified
 * position.
 */
#if defined(ANSIPROTO)
double
dGetPosParam(
WORD   wPosition,		    /* specified position */
double dDefault,		    /* default parameter */
double dMin,			    /* minimum value */
double dMax)			    /* maximum value */
#else
double
dGetPosParam(wPosition, dDefault, dMin, dMax)
WORD   wPosition;
double dDefault;
double dMin;
double dMax;
#endif
{
  double dValue;		  /* value of specified switch parameter */

#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.dGetPosParam\n");
#endif

  /* if the specified position is illegal, then return the default */
  if ((wPosition < 1) || (wPosition > cPosArgs)) return(dDefault);

  /* validate and return the parameter specified in the specified position */
  dValue = atof(apszPosArg[wPosition - 1]);
  if (dValue < dMin) dValue = dMin;
  if (dValue > dMax) dValue = dMax;
  return(dValue);

} /* dGetPosParam */


/* ----------------------------------------------------------------------------
 * pszGetPosParam:  This function returns the string parameter in the specified
 * position.
 */
#if defined(ANSIPROTO)
PSTR
pszGetPosParam(
WORD wPosition,			    /* specified position */
PSTR pszDefault)		    /* default parameter */
#else
PSTR
pszGetPosParam(wPosition, pszDefault)
WORD wPosition;
PSTR pszDefault;
#endif
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.pszGetPosParam\n");
#endif

  /* if the specified position is illegal, then return the default */
  if ((wPosition < 1) || (wPosition > cPosArgs)) return(pszDefault);

  /* validate and return the parameter specified in the specified position */
  return(apszPosArg[wPosition - 1]);

} /* pszGetPosParam */


/*****************************************************************************/
/**************************  PRIVATE FUNCTIONS	******************************/
/*****************************************************************************/


/* ----------------------------------------------------------------------------
 * fSwitchArgument:  This function returns TRUE if the specified argument is
 * switch argument or FALSE if it is a positional argument.
 */
#if defined(ANSIPROTO)
BOOL
fSwitchArgument(
PSTR pszArgument)		    /* specified argument */
#else
BOOL
fSwitchArgument(pszArgument)
PSTR pszArgument;
#endif
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("cmdparse.fSwitchArgument\n");
#endif

  if ((pszArgument[0] == '-') && !isdigit(pszArgument[1])) return(TRUE);
  else							   return(FALSE);

} /* fSwitchArgument */
