/*-----------------------------------------------------------------------------

 FILE:	   GEOMETRY.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/25/93    KMK      initial implementation
 10/29/93    KMK      second rewrite

 DESCRIPTION:

 This module contains geometric functions that compute the angle between three
 points, computes the local spacing between three points, computes the quality
 factor of a triangle, etc.

-----------------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "typedefs.h"
#include "literals.h"
#include "error.h"
#include "geometry.h"


#if defined(ANSIPROTO)
REAL  ComputeDistance	(REAL, REAL, UINT);
#else
REAL  ComputeDistance	();
#endif


/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * CompAngle:  This function computes the angle between three points given
 * by coordinates (x1,y1), (x2,y2) and (x3,y3).	 The angle is measured from
 * vector 12 to 32.
 */
#if defined(ANSIPROTO)
REAL
CompAngle(
REAL x1, REAL y1,			/* coordinate (x1,y1) */
REAL x2, REAL y2,			/* coordinate (x2,y2) */
REAL x3, REAL y3)			/* coordinate (x3,y3) */
#else
REAL
CompAngle(x1,y1,x2,y2,x3,y3)
REAL x1; REAL y1;
REAL x2; REAL y2;
REAL x3; REAL y3;
#endif
{
  REAL xl,yl;				/* coordinates of vector 12 */
  REAL xr,yr;				/* coordinates of vector 32 */
  REAL r;				/* temporary number */
  REAL angle;				/* angle between vectors 12 & 32 */

#if defined(DEBUG_FUNC)
DEBUG_INFO("geometry.CompAngle\n");
#endif

  xl = x1 - x2;	 yl = y1 - y2;
  xr = x3 - x2;	 yr = y3 - y2;

  r = (xl*xr+yl*yr)/(sqrt(sq(xl)+sq(yl))*sqrt(sq(xr)+sq(yr)));
  if	  (r >	1.0) r =  1.0;
  else if (r < -1.0) r = -1.0;
  angle = acos(r);
  if (xl*yr-xr*yl > 0) angle = 2*PI-angle;
  return(angle);

} /* CompAngle */



/* ----------------------------------------------------------------------------
 * CompHlocal:	This function computes the local spacing between three points
 * given by coordinates (x1,y1), (x2,y2) and (x3,y3).
 */
#if defined(ANSIPROTO)
REAL
CompHlocal(
REAL x1, REAL y1,			/* coordinate (x1,y1) */
REAL x2, REAL y2,			/* coordinate (x2,y2) */
REAL x3, REAL y3)			/* coordinate (x3,y3) */
#else
REAL
CompHlocal(x1,y1,x2,y2,x3,y3)
REAL x1; REAL y1;
REAL x2; REAL y2;
REAL x3; REAL y3;
#endif
{
  REAL xl,yl;				/* coordinates of vector 12 */
  REAL xr,yr;				/* coordinates of vector 32 */

#if defined(DEBUG_FUNC)
DEBUG_INFO("geometry.CompHlocal\n");
#endif

  xl = x1 - x2;	 yl = y1 - y2;
  xr = x3 - x2;	 yr = y3 - y2;

  /* @@@ check if sqrt(sqrt(arg)) is faster than pow(arg,0.25) */
  return(pow((xl*xl+yl*yl)*(xr*xr+yr*yr),0.25));

} /* CompHlocal */


/* ----------------------------------------------------------------------------
 * CompQuality:	 This function computes the quality factor of the triangle
 * given by coordinates (x1,y1), (x2,y2) and (x3,y3).  The quality factor
 * is normalized so that an equilateral triangle has a quality factor of
 * one and a triangle with a very small angle has a quality factor approaching
 * zero.
 */
#if defined(ANSIPROTO)
REAL
CompQuality(
REAL x1, REAL y1,			/* coordinate (x1,y1) */
REAL x2, REAL y2,			/* coordinate (x2,y2) */
REAL x3, REAL y3)			/* coordinate (x3,y3) */
#else
REAL
CompQuality(x1,y1,x2,y2,x3,y3)
REAL x1; REAL y1;
REAL x2; REAL y2;
REAL x3; REAL y3;
#endif
{
  REAL xl,yl;				/* coordinates of vector 12 */
  REAL xr,yr;				/* coordinates of vector 32 */
  REAL det;				/* determinent of coordinate matrix */
  REAL sss;				/* sum of side lengths squared */

#if defined(DEBUG_FUNC)
DEBUG_INFO("geometry.CompQuality\n");
#endif

  xl = x1 - x2;	 yl = y1 - y2;
  xr = x3 - x2;	 yr = y3 - y2;

  det = xr * yl - xl * yr;
  sss = (xl*xl+yl*yl) + (xr*xr+yr*yr) + (sq(xl-xr)+sq(yl-yr));
  return(det * 3.46410161514 / sss);

} /* CompQuality */


/* ----------------------------------------------------------------------------
 * CompRectQual:  This function computes the quality factor of the rectangle
 * given by coordinates (x1,y1), (x2,y2) and (x3,y3).  The quality factor is
 * normalized so that a square has a quality factor of one and the quality
 * factor of a rectangle with a poor aspect ration approaches zero.
 */
#if defined(ANSIPROTO)
REAL
CompRectQual(
REAL x1, REAL y1,			/* coordinate (x1,y1) */
REAL x2, REAL y2,			/* coordinate (x2,y2) */
REAL x3, REAL y3)			/* coordinate (x3,y3) */
#else
REAL
CompRectQual(x1,y1,x2,y2,x3,y3)
REAL x1; REAL y1;
REAL x2; REAL y2;
REAL x3; REAL y3;
#endif
{
  REAL d1,d2;				/* length of edges */
  REAL Qual;				/* quality factor */

#if defined(DEBUG_FUNC)
DEBUG_INFO("geometry.CompQuality\n");
#endif

  d1 = sqrt(sq(x1-x2)+sq(y1-y2));
  d2 = sqrt(sq(x3-x2)+sq(y3-y2));
  Qual = d1/d2;
  return (Qual > 1.0) ? 1.0 / Qual : Qual;

} /* CompQuality */


/* ----------------------------------------------------------------------------
 * nCircleCenter:  This function determines the center of the circle that
 * passes through points (x1,y1), (x2,y2) and (x3,y3).	If the points are
 * coplaner then GEOMETRY_COLINEAR is returned otherwise GEOMETRY_PLANER is
 * returned.
 */
#if defined(ANSIPROTO)
int
nCircleCenter(
REAL  x1, REAL	y1,			/* coordinate (x1,y1) */
REAL  x2, REAL	y2,			/* coordinate (x2,y2) */
REAL  x3, REAL	y3,			/* coordinate (x3,y3) */
PREAL xc, PREAL yc)			/* center coordinates (xc,yc) */
#else
int
nCircleCenter(x1,y1,x2,y2,x3,y3,xc,yc)
REAL  x1; REAL	y1;
REAL  x2; REAL	y2;
REAL  x3; REAL	y3;
PREAL xc; PREAL yc;
#endif
{
  REAL xl,yl;				/* coordinates of vector 12 */
  REAL xr,yr;				/* coordinates of vector 32 */
  REAL rl,rr;				/* ??? */
  REAL det;				/* determinent of coordinate matrix */

#if defined(DEBUG_FUNC)
DEBUG_INFO("geometry.nCircleCenter\n");
#endif

  xl = x1 - x2;	 yl = y1 - y2;
  xr = x3 - x2;	 yr = y3 - y2;

  det = xr * yl - xl * yr;
  if (det == 0.0)
  { *xc = *yc = 0.0;
    return(GEOMETRY_COLINEAR);
  }

  rl = (xl * (x1 + x2) + yl * (y1 + y2)) / det;
  rr = (xr * (x3 + x2) + yr * (y3 + y2)) / det;

  *xc = -0.5 * (rl * yr - rr * yl);
  *yc = -0.5 * (rr * xl - rl * xr);
  return(GEOMETRY_PLANER);

} /* nCircleCenter */


/* ----------------------------------------------------------------------------
 * nParamArc:  This function parametrizes the arc which passes through the
 * points (x1,y1), (x2,y2) and (x3,y3) with center (xc,yc).
 */
#if defined(ANSIPROTO)
int
nParamArc(
REAL  x1, REAL	y1,			/* coordinates (x1,y1) */
REAL  x2, REAL	y2,			/* coordinates (x2,y2) */
REAL  x3, REAL	y3,			/* coordinates (x3,y3) */
REAL  xc, REAL	yc,			/* coordinates (xc,yc) */
PREAL a1, PREAL a3,			/* angle 1 and angle 2 */
PREAL r)				/* radius */
#else
int
nParamArc(x1,y1,x2,y2,x3,y3,xc,yc,a1,a3,r)
REAL  x1; REAL	y1;
REAL  x2; REAL	y2;
REAL  x3; REAL	y3;
REAL  xc; REAL	yc;
PREAL a1; PREAL a3;
PREAL r;
#endif
{
  REAL A1,A2,A3,R;			/* angles and radius */
  REAL t;

#if defined(DEBUG_FUNC)
DEBUG_INFO("geometry.nParamArc\n");
#endif

  /* compute the radius and the angles */
  R = sqrt(sq(x1-xc) + sq(y1-yc));
  if (x1 > y1)
  { t = (x1 - xc) / R;
    if	    (t >  1.0) t =  1.0;
    else if (t < -1.0) t = -1.0;
    A1 = acos(t);
    if (y1 < yc) A1 = 2*PI - A1;
  }
  else
  { t = (y1 - yc) / R;
    if	    (t >  1.0) t =  1.0;
    else if (t < -1.0) t = -1.0;
    A1 = asin(t);
    if (x1 < xc) A1 = PI - A1;
  }
  if (x2 > y2)
  { t = (x2 - xc) / R;
    if	    (t >  1.0) t =  1.0;
    else if (t < -1.0) t = -1.0;
    A2 = acos(t);
    if (y2 < yc) A2 = 2*PI - A2;
  }
  else
  { t = (y2 - yc) / R;
    if	    (t >  1.0) t =  1.0;
    else if (t < -1.0) t = -1.0;
    A2 = asin(t);
    if (x2 < xc) A2 = PI - A2;
  }
  if (x3 > y3)
  { t = (x3 - xc) / R;
    if	    (t >  1.0) t =  1.0;
    else if (t < -1.0) t = -1.0;
    A3 = acos(t);
    if (y3 < yc) A3 = 2*PI - A3;
  }
  else
  { t = (y3 - yc) / R;
    if	    (t >  1.0) t =  1.0;
    else if (t < -1.0) t = -1.0;
    A3 = asin(t);
    if (x3 < xc) A3 = PI - A3;
  }

  /* select the appropriate branch */
  if ((A3 > A1) && ((A2 < A1) || (A2 > A3))) A3 = 2*PI - A3;
  if ((A1 > A3) && ((A2 < A3) || (A2 > A1))) A1 = 2*PI - A1;

  /* return w/ results */
  *r  = R;
  *a1 = A1;
  *a3 = A3;
  return(0);

} /* nParamArc */


/* ----------------------------------------------------------------------------
 * nIntersection:  This function finds the intersection of the line segments
 * whose end points are {A & A+dA} and {B & B+dB}.  The point of intersection
 * is not returned in terms of (x,y) coordinates.  It is specified in terms
 * of the coordinate system set up by the vectors dA and dB.  The (x,y)
 * coordinates of the intersection can be determined by the following
 * equations:
 *
 *	    x = A.x + Amul * dA.x = B.x + Bmul * dB.x
 *	    y = A.y + Amul * dA.y = B.y + Bmul * dB.y
 *
 * If the line segments are parallel then GEOMETRY_PARALLEL is returned
 * otherwise GEOMETRY_INTERSECTION is returned.
 */
#if defined(ANSIPROTO)
int
nIntersection(
PCRDNT pcrdntA,				/* point A */
PCRDNT pcrdntdA,			/* vector from A */
PCRDNT pcrdntB,				/* point B */
PCRDNT pcrdntdB,			/* vector from B */
PREAL  pAmul,				/* intersection point */
PREAL  pBmul)				/* intersection point */
#else
int
nIntersection(pcrdntA,pcrdntdB,pcrdntB,pcrdntdB,pAmul,pBmul)
PCRDNT	pcrdntA;
PCRDNT	pcrdntdB;
PCRDNT	pcrdntB;
PCRDNT	pcrdntdB;
PREAL	pAmul;
PREAL	pBmul;
#endif
{
  CRDNT crdntAB;			/* vector AB */
  REAL	Det;				/* determinent of dA & dB matrix */

#if defined(DEBUG_FUNC)
DEBUG_INFO("geometry.nIntersection\n");
#endif

  crdntAB.x = pcrdntA->x - pcrdntB->x;
  crdntAB.y = pcrdntA->y - pcrdntB->y;
  Det = pcrdntdA->y * pcrdntdB->x - pcrdntdA->x * pcrdntdB->y;

  if (Det == 0.0)
  { *pAmul = *pBmul = 0.0;
    return(GEOMETRY_PARALLEL);
  }

  *pAmul = (crdntAB.x * pcrdntdB->y - crdntAB.y * pcrdntdB->x) / Det;
  *pBmul = (crdntAB.x * pcrdntdA->y - crdntAB.y * pcrdntdA->x) / Det;
  return(GEOMETRY_INTERSECT);

} /* nIntersection */


/* ----------------------------------------------------------------------------
 * cNumNewPoints:  This function determines the number of new nodes to add to
 * an edge of length D with left and right spacing weights of Hl and Hr.
 */
#if defined(ANSIPROTO)
int
cNumNewPoints(
REAL  Hl,				/* left node spacing */
REAL  Hr,				/* right node spacing */
REAL  D,				/* edge length */
WORD  cSeg,				/* number of segments */
PREAL Hinit,				/* initial node spacing */
PREAL Hgrade)				/* node spacing grade */
#else
int
cNumNewPoints(Hl,Hr,D,cSeg,Hinit,Hgrade)
REAL  Hl;
REAL  Hr;
REAL  D;
WORD  cSeg;
PREAL Hinit;
PREAL Hgrade;
#endif
{
  double Grade = cmdargs.Grade;

  double Hmin;				/* normalized minimum spacing */
  double Hmax;				/* normalized maximum spacing */
  double Total;				/* running length total */
  double Ratio;				/* ratio of left & right spacing */
  WORD	 cNewPoints;			/* number of new points */

#if defined(DEBUG_FUNC)
DEBUG_INFO("edge.cNumNewPoints\n");
#endif

  /* initialize new point structure */
  cNewPoints = 0;
  *Hinit     = 1.0;
  *Hgrade    = 1.0;

  /* compute normalized maximum and minimum spacing */
  Hmax = max(Hl,Hr)/D;
  Hmin = min(Hl,Hr)/D;

  /* compute number of new points if not specified */
  if (cSeg) cNewPoints = cSeg - 1;
  else
  { if ((Hmax >= 1.0 - EPS) || (Hmin * Grade >= 1.0 - EPS)) return(0);
    Total = Hmin;
    do
    {
      ++cNewPoints;
      Hmin = min(Hmin*Grade,Hmax);
      Total += Hmin;
    } while (Total < 1.0 - EPS);
    if (Total > 1.0 + 0.5*Hmax) --cNewPoints;
  }

  if (cNewPoints == 0) return(0);

  /* compute the initial node spacing and node spacing grade */
  Ratio = Hr/Hl;
  *Hinit = 1.0/(double) (cNewPoints+1);
  if (fabs(Ratio-1.0) >= 0.001)
  {
    *Hgrade = pow(Ratio,*Hinit);
    *Hgrade = min(Grade,*Hgrade);
    *Hgrade = max(1.0/Grade,*Hgrade);
    *Hinit = (*Hgrade-1.0)/(Ratio-1.0);
  }

  return(cNewPoints);

} /* cNumNewPoints */


/* ----------------------------------------------------------------------------
 * Function:	 cComputeGrade
 * Description:	 This function computes the initial spacing, grade and number
 *		 of subdivisions for an edge.  The grade is chosen to be as
 *		 close to the passed value as possible.
 */
int cComputeGrade(REAL dx, REAL *xs, REAL *xg)
{
  REAL x1 = 0.0;
  REAL x2 = *xg;
  WORD i,c;

  /* if the initial spacing is greater than the edge length, then set the */
  /* initial spacing to the edge length and return;			  */
  if (*xs > dx)
  { *xs = dx;
    return 0;
  }

  /* if the grade is zero, readjust the initial spacing */
  if (x2 == 0.0)
  { c = (WORD) ceil(dx / *xs);
    *xs = 1.0 / (REAL) c;
    *xg = 1.0;
    return c - 1;
  }

  /* compute the number of subdivisions */
  for(i = 1;;++i)
    if (ComputeDistance(*xs, *xg, i) >= dx)
      break;

  /* adjust the grade so that the calculated number of subdivisons */
  /* smoothly increase in size					   */
  while (fabs(x1 - x2) > 1.0e-14)
  { *xg = (x1 + x2) / 2.0;
    if (ComputeDistance(*xs, *xg, i) > dx) x2 = *xg;
    else				   x1 = *xg;
  }

  *xs /= dx;
  *xg += 1.0;
  return i - 1;
}


/******************************************************************************/
/**************************  PRIVATE FUNCTIONS	*******************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * Function:	 ComputeDistance
 * Scope:	 Private
 * Description:	 This function computes the length of an edge which consists
 *		 of c segments of graded length.  The initial segment length
 *		 is xs and the grade is xg.
 */
REAL ComputeDistance (REAL xs, REAL xg, UINT c)
{
  UINT i;
  REAL x = xs;
  for(i = 1; i < c; ++i)
  { xs *= (1.0 + xg);
    x  += xs;
  }
  return x;
}


