/*-----------------------------------------------------------------------------

 FILE:	   MAIN.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/19/93    KMK      initial implementation
 11/02/94    KMK      user interface updated

 DESCRIPTION:

 This module contains the main function.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(UNIX)
#include "xgraph.h"
#else
#include "wgraph.h"
#endif

#include "typedefs.h"
#include "literals.h"
#include "error.h"
#include "cmdparse.h"
#include "mesh.h"
#include "skel.h"
#include "skl2msh.h"
#include "enhance.h"
#include "triang.h"
#include "graphics.h"


/* help information */
static PSTR apszHelp[] =
{ LIT_MAIN2,  LIT_MAIN3,  LIT_MAIN4,  LIT_MAIN5,  LIT_MAIN6,  LIT_MAIN7,
  LIT_MAIN8,  LIT_MAIN9,  LIT_MAIN10, LIT_MAIN11, LIT_MAIN12, LIT_MAIN13,
  LIT_MAIN14, LIT_MAIN15, LIT_MAIN16, LIT_MAIN17, LIT_MAIN18, LIT_MAIN20,
  LIT_MAIN21, LIT_MAIN26, LIT_MAIN22, LIT_MAIN23, LIT_MAIN24, LIT_MAIN25,
  NULL
};

/* declare global data */
CMDARGS cmdargs;
char  szGlobalBuf[GLOBAL_BUFSIZE];

#if defined(ANSIPROTO)
BOOL	fParseCommandLineArgs	(void);
int	nAsciiOutputMesh	(PMESH, char *);
#else
BOOL	fParseCommandLineArgs	();
int	nAsciiOutputMesh	();
#endif

// redraw function
static PMESH pmesh;
void Redraw ()	{ nRateMesh(pmesh); }


/* ----------------------------------------------------------------------------
 * main:  this function display the copyright message, parses the command line
 * arguments, initializes the various modules, calls the parser, and closes
 * the various modules.
 */
int main ( int cArg, PSTR apszArg[] )
{
  PSKEL pskel;
  //  PMESH pmesh;

  #if defined(DEBUG_FUNC)
  DEBUG_INFO("main.main\n");
  #endif

  /* display the copyright notice and initialize the command line parser */
  fprintf(stderr, "%s\n", LIT_MAIN1);
  vInitCmdParser(cArg, apszArg);

  /* parse the command line arguments */
  if (fParseCommandLineArgs() == FALSE) return(RETCODE_CMDLINE);

  /* initialize the error system, file system and global variables */
  vErrorInit(cmdargs.wErrDev, NULL);
  *szGlobalBuf = '\0';

  /* triangulate the mesh */
  pskel = pskelCreate(cmdargs.pszSource);
  pmesh = pmeshCreate();
  nSkel2Mesh(pskel,pmesh);

  #if 0
  if (cmdargs.wVisual == VISUAL_CONSTRUCT)
    nCreateWindow(pskel->xMin,pskel->yMin,pskel->xMax,pskel->yMax);
  #endif

  nInitTriangModule();
  nTriangulate(pmesh);

  GRInitGraphics();

  if (cmdargs.wVisual == VISUAL_RATING)
    nCreateWindow(pskel->xMin,pskel->yMin,pskel->xMax,pskel->yMax, Redraw);

  #if defined(UNIX)
  BOOL fDone = FALSE;
  short int keycode;
  int arg1, arg2;
  while (!fDone)
  { switch(getevent(&keycode, &arg1, &arg2))
    { case EXPOSE:
        Redraw();
        flushdisplay();
        break;

      case CONFIGURENOTIFY:
        break;

      case BUTTONPRESS:
      case KEYPRESS:
        fDone = TRUE;
        break;
    }
  }
  #endif

  GRFreeGraphics();

  if (cmdargs.fGeneratePSFile) nPostScript(pmesh);

  /* write grid */
  nWriteMesh(cmdargs.pszSource, pmesh);
  if (cmdargs.pszAsciiMesh) nAsciiOutputMesh(pmesh, cmdargs.pszAsciiMesh);

  /* prepare the program for termination */
  vExitFunction();
  return(RETCODE_SUCCESS);

} /* main */


/*****************************************************************************/
/************  THE FOLLOWING FUNCTIONS ARE GLOBAL TO THE PROGRAM  ************/
/*****************************************************************************/


/* ----------------------------------------------------------------------------
 * vExitFunction:  This function prepares the program for termination.
 */
void
vExitFunction()
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("main.vExitFunction\n");
#endif

  return;

} /* vExitFunction */


/* ----------------------------------------------------------------------------
 * pszLowerCase:  This function converts the specified string to lower case.
 */
#if defined(ANSIPROTO)
PSTR
pszLowerCase(
PSTR pszOriginal)		    /* specified string */
#else
PSTR
pszLowerCase(pszOriginal)
PSTR pszOriginal;
#endif
{
  PSTR psz = pszOriginal;	  /* "scanning" pointer */

#if defined(DEBUG_FUNC)
DEBUG_INFO("main.pszLowerCase\n");
#endif

  while (*psz)	      /* convert the string to lower case */
  {
    if ((*psz >= 'A') && (*psz <= 'Z')) *psz += (char) ('a' - 'A');
    ++psz;
  }
  return(pszOriginal);

} /* pszLowerCase */


/* ----------------------------------------------------------------------------
 * pszUpperCase:  This function converts the specified string to upper case.
 */
#if defined(ANSIPROTO)
PSTR
pszUpperCase(
PSTR pszOriginal)		    /* specified string */
#else
PSTR
pszUpperCase(pszOriginal)
PSTR pszOriginal;
#endif
{
  PSTR psz = pszOriginal;	  /* "scanning" pointer */

#if defined(DEBUG_FUNC)
DEBUG_INFO("main.pszUpperCase\n");
#endif

  while (*psz)	      /* convert the string to lower case */
  {
    if ((*psz >= 'a') && (*psz <= 'z')) *psz -= (char) ('a' - 'A');
    ++psz;
  }
  return(pszOriginal);

} /* pszUpperCase */


/* ----------------------------------------------------------------------------
 * sq:	square of the argument.
 */
#if defined(ANSIPROTO)
double sq(double x)
#else
double sq(x)
double x;
#endif
{

#if defined(DEBUG_FUNC)
DEBUG_INFO("main.sq\n");
#endif

  return x * x;

} /* sq */


/*****************************************************************************/
/**************************  PRIVATE FUNCTIONS	******************************/
/*****************************************************************************/


/* ----------------------------------------------------------------------------
 * fParseCommandLineArgs:  This function parses the command line arguments and
 * stores the command line information in the command line arguments structure.
 * TRUE is returned if the command line arguments are valid; otherwise, FALSE
 * is returned.
 */
BOOL
fParseCommandLineArgs()
{
  PSTR *ppsz;			  /* pointer to string pointer */

#if defined(DEBUG_FUNC)
DEBUG_INFO("main.fParseCommandLineArgs\n");
#endif

  /* display the help information if requested */
  if (fParamExist("h") || (wNumOfPosArgs() != 1))
  {
    for(ppsz = apszHelp; *ppsz != NULL; ++ppsz)
      fprintf(stdout, "%s\n", *ppsz);
    return(FALSE);
  }

  /* parse the command line arguments */
  cmdargs.pszSource    = pszGetPosParam(1, NULL);
  cmdargs.pszAsciiMesh = pszGetSwitchParam("O", NULL);
  cmdargs.wErrDev     = (fParamExist("eo")) ? ERRDEV_STDOUT : ERRDEV_STDERR;
  cmdargs.fGraded     = fParamExist("i+") && !fParamExist("i-");
  cmdargs.fExhaustive = fParamExist("t+") || !fParamExist("t-");
  cmdargs.fNoCentroid = fParamExist("c-") && !fParamExist("c+");
  cmdargs.fAutoSpace  = !fParamExist("s");
  cmdargs.cQualEnh1   = wGetSwitchParam("qe", 2, 0, 32767);
  cmdargs.cQualEnh2   = wGetSwitchParam("qn", 2, 0, 32767);
  cmdargs.cQualEnh3   = wGetSwitchParam("qb", 2, 0, 32767);
  cmdargs.Grade	      = dGetSwitchParam("g", 1.5, 1.5, 2.5);
  cmdargs.Hmax	      = dGetSwitchParam("s", 0.0, 1.0e-300, 1.0e300);
  cmdargs.Qmin	      = dGetSwitchParam("m", 0.35, 0.0, 1.0);
  cmdargs.cNoPress    = nGetSwitchParam("z", -1, 0, 32767);
  cmdargs.cPresses    = 0;
  cmdargs.fKeepAspect = fParamExist("aspect");
  cmdargs.fGeneratePSFile = fParamExist("ps");
  cmdargs.fNoQualReport = fParamExist("nqr");
  if	  (fParamExist("vn")) cmdargs.wVisual = VISUAL_NONE;
#if 0
  else if (fParamExist("vc")) cmdargs.wVisual = VISUAL_CONSTRUCT;
#endif
  else			      cmdargs.wVisual = VISUAL_RATING;
  cmdargs.nResolution = nGetSwitchParam("r", 2, 1, 3);
  return(TRUE);

} /* fParseCommandLineArgs */


/* ----------------------------------------------------------------------------
 * nAsciiOutputMesh:  This function writes the mesh in an ascii format.
 */
int nAsciiOutputMesh(PMESH pmesh, char *pszFileName)
{
  PNODE anode = pmesh->anode;
  PEDGE aedge = pmesh->aedge;
  PTRI	atri  = pmesh->atri;
  WORD	cnode = pmesh->cnode;
  WORD	cedge = pmesh->cedge;
  WORD	ctri  = pmesh->ctri;

  PNODE pnode;
  PEDGE pedge;
  PTRI	ptri;
  FILE *file;
  int	i;

  file = fopen(pszFileName, "wt");

  fprintf(file, "TRIANGLES\n");
  fprintf(file, "      |     |		EDGES		       TRIANGLES\n");
  fprintf(file, "  i   |  u  |	 AB    BC  AC/CD   AD	  AB	BC  AC/CD   AD\n");
  fprintf(file, "----- | --- | ----- ----- ----- -----	----- ----- ----- -----\n");
  for(i = 0, ptri = atri; i < ctri; ++i, ++ptri)
  { fprintf(file, "%5u | ", i);
    if (ptri->wLabel != -1u) fprintf(file, "%3u | ", ptri->wLabel);  else fprintf(file, "    | ");

    if (ptri->iedgeAB != -1u) fprintf(file, "%5u ", ptri->iedgeAB); else fprintf(file, "      ");
    if (ptri->iedgeBC != -1u) fprintf(file, "%5u ", ptri->iedgeBC); else fprintf(file, "      ");
    if (ptri->iedgeAC != -1u) fprintf(file, "%5u ", ptri->iedgeAC); else fprintf(file, "      ");
    if (ptri->iedgeAD != -1u) fprintf(file, "%5u ", ptri->iedgeAD); else fprintf(file, "  n/a ");
    fprintf(file, " ");
    if (ptri->itriAB != -1u) fprintf(file, "%5u ", ptri->itriAB); else fprintf(file, "	    ");
    if (ptri->itriBC != -1u) fprintf(file, "%5u ", ptri->itriBC); else fprintf(file, "	    ");
    if (ptri->itriAC != -1u) fprintf(file, "%5u ", ptri->itriAC); else fprintf(file, "	    ");
    if (ptri->itriAD != -1u) fprintf(file, "%5u" , ptri->itriAD); else fprintf(file, "	n/a");
    fprintf(file, "\n");
  }

  fprintf(file, "\nEDGES\n");
  fprintf(file, "      |     |	  NODES	  \n");
  fprintf(file, "  i   |  u  |	 A     B  \n");
  fprintf(file, "----- | --- | ----- -----\n");
  for(i = 0, pedge = aedge; i < cedge; ++i, ++pedge)
  { fprintf(file, "%5u | ", i);
    if (pedge->wLabel != -1u) fprintf(file, "%3u | ", pedge->wLabel); else fprintf(file, "    | ");
    fprintf(file, "%5u %5u\n", pedge->inodeA, pedge->inodeB);
  }

  fprintf(file, "\nNODES\n");
  fprintf(file, "  i   |	x		y	\n");
  fprintf(file, "----- | --------------- ---------------\n");
  for(i = 0, pnode = anode; i < cnode; ++i, ++pnode)
    fprintf(file, "%5u | %15.8le %15.8le\n", i, pnode->x, pnode->y);

  fclose(file);
  return(0);

}
