/*-----------------------------------------------------------------------------

 FILE:	   MESH.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/28/93    KMK      initial implementation

 DESCRIPTION:

 This module contains files which manipulate the mesh data structure.  The
 mesh data structure contains a dynamic array of nodes, a dynamic array of
 triangles, a doublely linked-list of vertices and a stack of polygons.

-----------------------------------------------------------------------------*/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if   defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#define O_BINARY 0x0000
#else
#include <io.h>
#include <sys\stat.h>
#endif

#include <fcntl.h>
#include "typedefs.h"
#include "literals.h"
#include "error.h"
#include "mesh.h"
#include "xlatinfo.h"


#define INCREMENT_NODE	    128
#define INCREMENT_EDGE	    128
#define INCREMENT_TRIANGLE  128
#define INCREMENT_VERTEX    128
#define INCREMENT_POLYGON   32


typedef struct tagMESHHEAD	    /* meshhead */
{ char	szLogo[128];		    /* file logo */
  WORD	cnode;			    /* number of nodes */
  WORD	cedge;			    /* number of edges */
  WORD	ctri;			    /* number of triangles */
  BYTE	abReserved[122];	    /* reserved for future use */
} MESHHEAD;
typedef MESHHEAD *PMESHHEAD;	    /* pmeshhead */


/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * pmeshCreate:	 The function creates and initializes a mesh data structure.
 * A pointer to this structure is returned.
 */
PMESH
pmeshCreate()
{
  PMESH pmesh;				/* mesh to allocate */
  PVTX	pvtx;				/* walking pointer */
  int	i;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.pmeshCreate\n");
#endif

  /* allocate mesh structure */
  pmesh = (PMESH) calloc(1, sizeof(MESH));
  MEMCHECK(pmesh, LIT_MESH, LIT_MESH1);

  /* allocate dynamic node array and initialize corresponding fields */
  pmesh->cnode = 0;
  pmesh->canode = INCREMENT_NODE;
  pmesh->anode = (PNODE) calloc(INCREMENT_NODE, sizeof(NODE));
  MEMCHECK(pmesh->anode, LIT_MESH, LIT_MESH1);

  /* allocate dynamic edge array and initialize corresponding fields */
  pmesh->cedge = 0;
  pmesh->caedge = INCREMENT_NODE;
  pmesh->aedge = (PEDGE) calloc(INCREMENT_EDGE, sizeof(EDGE));
  MEMCHECK(pmesh->aedge, LIT_MESH, LIT_MESH1);

  /* allocate dynamic triangle array and initialize corresponding fields */
  pmesh->ctri = 0;
  pmesh->catri = INCREMENT_TRIANGLE;
  pmesh->atri = (PTRI) calloc(INCREMENT_TRIANGLE, sizeof(TRI));
  MEMCHECK(pmesh->atri, LIT_MESH, LIT_MESH1);

  /* allocate dynamic vertex array and initialize corresponding fields */
  pmesh->cvtx = 0;
  pmesh->cavtx = INCREMENT_VERTEX;
  pmesh->avtx = (PVTX) calloc(INCREMENT_VERTEX, sizeof(VTX));
  MEMCHECK(pmesh->avtx, LIT_MESH, LIT_MESH1);

  /* allocate dynamic polygon array and initialize corresponding fields */
  pmesh->cpoly = 0;
  pmesh->capoly = INCREMENT_POLYGON;
  pmesh->apoly = (PPOLY) calloc(INCREMENT_POLYGON, sizeof(POLY));
  MEMCHECK(pmesh->apoly, LIT_MESH, LIT_MESH1);

  /* link the free list of vertices */
  pmesh->ivtxFree = 0;
  for(i = 0, pvtx = pmesh->avtx; i < INCREMENT_VERTEX; ++i, ++pvtx)
    pvtx->ivtxR = (i == INCREMENT_VERTEX - 1) ? (IVTX) -1 : i + 1 ;

  /* return the mesh structure */
  return(pmesh);

} /* pmeshCreate */


/* ----------------------------------------------------------------------------
 * inodeAddNode:  This function adds a node to the dynamic node array.
 */
#if defined(ANSIPROTO)
INODE inodeAddNode(
PMESH pmesh,				/* mesh structure */
REAL  x,				/* x coordinate */
REAL  y,				/* y coordinate */
BOOL  fPermenent)			/* permenent node flag */
#else
INODE inodeAddNode(pmesh,x,y,fPermenent)
PMESH pmesh;
REAL  x;
REAL  y;
BOOL  fPermenent;
#endif
{
  PNODE anode  = pmesh->anode;
  WORD	cnode  = pmesh->cnode;
  WORD	canode = pmesh->canode;

  PNODE pnode;				/* node pointer */

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.inodeAddNode\n");
#endif

  /* array full? yes, then increase its size */
  if (cnode == canode)
  { canode += INCREMENT_NODE;
    pmesh->canode = canode;
    pmesh->anode = anode = (PNODE) realloc(anode, canode * sizeof(NODE));
    MEMCHECK(anode, LIT_MESH, LIT_MESH2);
  }

  /* store coordinates in node array */
  pnode = anode + cnode;
  pnode->x = x;
  pnode->y = y;

  /* increment node count if node is permenent and return */
  if (fPermenent) ++pmesh->cnode;
  return(cnode);

} /* inodeAddNode */


/* ----------------------------------------------------------------------------
 * iedgeAddEdge:  This function adds an edge to the dynamic edge array.
 */
#if defined(ANSIPROTO)
IEDGE iedgeAddEdge(
PMESH pmesh,				/* mesh structure */
INODE inodeA,				/* index of node A */
INODE inodeB)				/* index of node B */
#else
IEDGE iedgeAddEdge(pmesh,inodeA,inodeB)
PMESH pmesh;
INODE inodeA;
INODE inodeB;
#endif
{
  PEDGE aedge  = pmesh->aedge;
  WORD	cedge  = pmesh->cedge;
  WORD	caedge = pmesh->caedge;

  PEDGE pedge;				/* edge pointer */

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.iedgeAddEdge\n");
#endif

  /* array full? yes, then increase its size */
  if (cedge == caedge)
  { caedge += INCREMENT_EDGE;
    pmesh->caedge = caedge;
    pmesh->aedge = aedge = (PEDGE) realloc(aedge, caedge * sizeof(EDGE));
    MEMCHECK(aedge, LIT_MESH, LIT_MESH3);
  }

  /* store coordinates in edge array */
  pedge = aedge + cedge;
  pedge->inodeA	 = inodeA;
  pedge->inodeB	 = inodeB;
  pedge->wLabel	 = 0;
  pedge->itriAdj = (ITRI) -1;

  /* increment edge count and return */
  ++pmesh->cedge;
  return(cedge);

} /* iedgeAddEdge */


/* ----------------------------------------------------------------------------
 * itriAddTriangle:  This function adds a triangle to the dynamic triangle
 * array.
 */
#if defined(ANSIPROTO)
ITRI itriAddTriangle(
PMESH pmesh,				 /* mesh structure */
IEDGE iedgeAB,				 /* index of edge AB */
IEDGE iedgeBC,				 /* index of edge BC */
IEDGE iedgeCD,				 /* index of edge AC or CD */
IEDGE iedgeAD)				 /* index of edge AD */
#else
ITRI itriAddTriangle(pmesh,iedgeAB,iedgeBC,iedgeCD,iedgeAD)
PMESH pmesh;
IEDGE iedgeAB;
IEDGE iedgeBC;
IEDGE iedgeCD;
IEDGE iedgeAD;
#endif
{
  PTRI	atri	= pmesh->atri;
  PEDGE aedge	= pmesh->aedge;
  PEDGE pedgeAB = aedge + iedgeAB;
  PEDGE pedgeBC = aedge + iedgeBC;
  PEDGE pedgeCD = aedge + iedgeCD;
  PEDGE pedgeAD = (iedgeAD == -1u) ? NULL : aedge + iedgeAD;
  WORD	ctri	= pmesh->ctri;
  WORD	catri	= pmesh->catri;

  PTRI ptri;				/* triangle pointer */
  PTRI ptriAdj;				/* adjacent triangle pointer */
  WORD itriAdj;				/* adjacent triangle index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.itriAddTriangle\n");
#endif

  /* array full? yes, then increase its size */
  if (ctri == catri)
  { catri += INCREMENT_TRIANGLE;
    pmesh->catri = catri;
    pmesh->atri = atri = (PTRI) realloc(atri, catri * sizeof(TRI));
    MEMCHECK(atri, LIT_MESH, LIT_MESH4);
  }

  /* store edges in tri array */
  ptri = atri + ctri;
  ptri->iedgeAB = iedgeAB;
  ptri->iedgeBC = iedgeBC;
  ptri->iedgeCD = iedgeCD;
  ptri->iedgeAD = iedgeAD;
  ptri->itriAD	= -1u;
  ptri->wStatus = 0;

  /* make a link to the element adjacent to edge AB */
  itriAdj = pedgeAB->itriAdj;
  if (itriAdj == (ITRI) -1)
  { pedgeAB->itriAdj = ctri;
    ptri->itriAB = (ITRI) -1;
  }
  else
  { ptri->itriAB = itriAdj;
    ptriAdj = atri + itriAdj;
    if	    (iedgeAB == ptriAdj->iedgeAB) ptriAdj->itriAB = ctri;
    else if (iedgeAB == ptriAdj->iedgeBC) ptriAdj->itriBC = ctri;
    else if (iedgeAB == ptriAdj->iedgeCD) ptriAdj->itriCD = ctri;
    else if (iedgeAB == ptriAdj->iedgeAD) ptriAdj->itriAD = ctri;
    else vErrorHandler(ERROR_PROGRAMMER, NULL, NULL, TRUE);
  }

  /* make a link to the element adjacent to edge BC */
  itriAdj = pedgeBC->itriAdj;
  if (itriAdj == (ITRI) -1)
  { pedgeBC->itriAdj = ctri;
    ptri->itriBC = (ITRI) -1;
  }
  else
  { ptri->itriBC = itriAdj;
    ptriAdj = atri + itriAdj;
    if	    (iedgeBC == ptriAdj->iedgeAB) ptriAdj->itriAB = ctri;
    else if (iedgeBC == ptriAdj->iedgeBC) ptriAdj->itriBC = ctri;
    else if (iedgeBC == ptriAdj->iedgeCD) ptriAdj->itriCD = ctri;
    else if (iedgeBC == ptriAdj->iedgeAD) ptriAdj->itriAD = ctri;
    else vErrorHandler(ERROR_PROGRAMMER, NULL, NULL, TRUE);
  }

  /* make a link to the element adjacent to edge AC or CD */
  itriAdj = pedgeCD->itriAdj;
  if (itriAdj == (ITRI) -1)
  { pedgeCD->itriAdj = ctri;
    ptri->itriCD = (ITRI) -1;
  }
  else
  { ptri->itriCD = itriAdj;
    ptriAdj = atri + itriAdj;
    if	    (iedgeCD == ptriAdj->iedgeAB) ptriAdj->itriAB = ctri;
    else if (iedgeCD == ptriAdj->iedgeBC) ptriAdj->itriBC = ctri;
    else if (iedgeCD == ptriAdj->iedgeCD) ptriAdj->itriCD = ctri;
    else if (iedgeCD == ptriAdj->iedgeAD) ptriAdj->itriAD = ctri;
    else vErrorHandler(ERROR_PROGRAMMER, NULL, NULL, TRUE);
  }

  /* make a link to the element adjacent to edge AD if this */
  /* element is a rectangle */
  if (pedgeAD)
  { itriAdj = pedgeAD->itriAdj;
    if (itriAdj == (ITRI) -1)
    { pedgeAD->itriAdj = ctri;
      ptri->itriAD = (ITRI) -1;
    }
    else
    { ptri->itriAD = itriAdj;
      ptriAdj = atri + itriAdj;
      if      (iedgeAD == ptriAdj->iedgeAB) ptriAdj->itriAB = ctri;
      else if (iedgeAD == ptriAdj->iedgeBC) ptriAdj->itriBC = ctri;
      else if (iedgeAD == ptriAdj->iedgeCD) ptriAdj->itriCD = ctri;
      else if (iedgeAD == ptriAdj->iedgeAD) ptriAdj->itriAD = ctri;
      else vErrorHandler(ERROR_PROGRAMMER, NULL, NULL, TRUE);
    }
  }

  /* increment triangle count and return */
  ++pmesh->ctri;
  return(ctri);

} /* itriAddTriangle */


/* ----------------------------------------------------------------------------
 * ivtxAddVertex:  This function adds the vertex inode to the specified polygon.
 */
#if defined(ANSIPROTO)
IVTX
ivtxAddVertex(
PMESH pmesh,				/* mesh structure */
INODE inode,				/* node index */
IEDGE iedge)				/* edge index */
#else
IVTX
ivtxAddVertex(pmesh,inode,iedge)
PMESH pmesh;
INODE inode;
IEDGE iedge;
#endif
{
  PPOLY ppoly	 = pmesh->ppoly;
  PVTX	avtx	 = pmesh->avtx;
  IVTX	ivtxFree = pmesh->ivtxFree;
  IVTX	ivtxHead = ppoly->ivtxHead;

  PVTX	  pvtx;				/* walking pointer */
  WORD	  cavtx;			/* size of vertex array */
  IVTX	  ivtxTail;			/* tail of vertex list */
  int	  i;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.ivtxAddVertex\n");
#endif

  /* if the dynamic vertex array is full then increase its size */
  if (ivtxFree == (WORD) -1)
  { cavtx = ivtxFree = pmesh->cavtx;
    cavtx += INCREMENT_VERTEX;
    pmesh->cavtx = cavtx;
    pmesh->avtx = avtx = (PVTX) realloc(avtx, cavtx * sizeof(VTX));
    MEMCHECK(avtx, LIT_MESH, LIT_MESH5);
    for(i = ivtxFree, pvtx = avtx + ivtxFree; i < cavtx; ++i, ++pvtx)
      pvtx->ivtxR = (i == cavtx - 1) ? (IVTX) -1 : i + 1;
  }


  /* initialize the node index of the vertex structure and update the */
  /* pointer to the list of free vertices */
  i = ivtxFree;
  pvtx = avtx + ivtxFree;
  ivtxFree = pvtx->ivtxR;
  pvtx->inode = inode;
  pvtx->iedge = iedge;

  /* link the vertex into the list */
  if (ppoly->cvtx == 0)
    ivtxHead = pvtx->ivtxL = pvtx->ivtxR = i;
  else
  { ivtxTail = avtx[ivtxHead].ivtxL;
    pvtx->ivtxL = ivtxTail;
    pvtx->ivtxR = ivtxHead;
    avtx[ivtxHead].ivtxL = avtx[ivtxTail].ivtxR = i;
  }

  /* update the mesh structure */
  ++pmesh->cvtx;
  ++ppoly->cvtx;
  pmesh->ivtxFree = ivtxFree;
  ppoly->ivtxHead = ivtxHead;
  return(i);

} /* ivtxAddVertex */



/* ----------------------------------------------------------------------------
 * nDeleteVertex:  This function deletes the specified vertex from the dynamic
 * vertex array.
 */
#if defined(ANSIPROTO)
int
nDeleteVertex(
PMESH pmesh,				/* mesh structure */
IVTX  ivtx)				/* index of vertex in which to delete */
#else
int
nDeleteVertex(pmesh,ivtx)
PMESH pmesh;
IVTX  ivtx;
#endif
{
  PPOLY ppoly	 = pmesh->ppoly;
  PVTX	avtx	 = pmesh->avtx;
  PVTX	pvtx	 = avtx + ivtx;
  IVTX	ivtxFree = pmesh->ivtxFree;
  IVTX	ivtxL	 = pvtx->ivtxL;
  IVTX	ivtxR	 = pvtx->ivtxR;

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.nDeleteVertex\n");
#endif

  if (ivtx == ppoly->ivtxHead) ppoly->ivtxHead = ivtxR;
  avtx[ivtxL].ivtxR = ivtxR;
  avtx[ivtxR].ivtxL = ivtxL;
  pvtx->ivtxR = ivtxFree;
  ivtxFree = ivtx;

  pmesh->cvtx -= 1;
  ppoly->cvtx -= 1;
  pmesh->ivtxFree = ivtxFree;
  return(pmesh->cvtx);

} /* nDeleteVertex */


/* ----------------------------------------------------------------------------
 * nSplitPolygon:  This function splits the TOS polygon into two regions at
 * the specified vertices.  Calling this function will invalidate the Angle
 * and Hlocal fields of specified vertices and their adjacent vertices.
 * After the split ivtxA and ivtxB become the head of the vertex lists.
 */
#if defined(ANSIPROTO)
int
nSplitPolygon(
PMESH pmesh,				/* mesh structure */
IVTX  ivtxA,				/* splitting vertex */
IVTX  ivtxB)				/* splitting vertex */
#else
int
nSplitPolygon(pmesh,ivtxA,ivtxB)
PMESH pmesh;
IVTX  ivtxA;
IVTX  ivtxB;
#endif
{
  PPOLY ppoly  = pmesh->ppoly;
  PVTX	avtx   = pmesh->avtx;
  PVTX	pvtxA  = avtx + ivtxA;
  PVTX	pvtxB  = avtx + ivtxB;
  IVTX	ivtxAL = pvtxA->ivtxL;
  IVTX	ivtxBL = pvtxB->ivtxL;

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.nSplitPolygon\n");
#endif

  avtx[ivtxAL].ivtxR = ivtxB;
  avtx[ivtxBL].ivtxR = ivtxA;
  pvtxA->ivtxL = ivtxBL;
  pvtxB->ivtxL = ivtxAL;
  ppoly->ivtxHead = ivtxB;
  ivtxAddVertex(pmesh,pvtxA->inode,(IEDGE) -1);
  ppoly->ivtxHead = ivtxA;
  ivtxAddVertex(pmesh,pvtxB->inode,(IEDGE) -1);
  return(0);

} /* nSplitPolygon */


/* ----------------------------------------------------------------------------
 * cPushPolygon:  This function pushes a polygon on the polygon stack.	The
 * number of polygons in the stack is returned.
 */
#if defined(ANSIPROTO)
WORD
cPushPolygon(
PMESH pmesh,				/* mesh structure */
PPOLY ppoly)				/* polygon to push on top of stack */
#else
WORD
cPushPolygon(pmesh,ppoly)
PMESH pmesh;
PPOLY ppoly;
#endif
{
  PPOLY apoly  = pmesh->apoly;
  WORD	cpoly  = pmesh->cpoly;
  WORD	capoly = pmesh->capoly;

  PPOLY ppolyTOS;			/* polygon pointer */

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.cPushPolygon\n");
#endif

  /* array full? yes, then increase its size */
  if (cpoly == capoly)
  { capoly += INCREMENT_POLYGON;
    pmesh->capoly = capoly;
    pmesh->apoly = apoly = (PPOLY) realloc(apoly, capoly * sizeof(POLY));
    MEMCHECK(apoly, LIT_MESH, LIT_MESH6);
  }

  /* add polygon to the TOS */
  ppolyTOS = apoly + cpoly;
  memcpy(ppolyTOS,ppoly,sizeof(POLY));

  /* increment polygon count and return */
  pmesh->cpoly = ++cpoly;
  pmesh->ppoly = ppolyTOS;
  return(cpoly);

} /* cPushPolygon */


/* ----------------------------------------------------------------------------
 * cPopPolygon:	 This function pops a polygon from the polygon stack and
 * discards it.	 The number of polygons in the stack is returned.
 */
#if defined(ANSIPROTO)
WORD
cPopPolygon(
PMESH pmesh)				/* mesh structure */
#else
WORD
cPopPolygon(pmesh)
PMESH pmesh;
#endif
{
  WORD cpoly = pmesh->cpoly;

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.cPopPolygon\n");
#endif

  if (cpoly)
  { pmesh->cpoly = --cpoly;
    pmesh->ppoly = (cpoly) ? pmesh->apoly + cpoly - 1 : NULL;
  }
  return(cpoly);

} /* cPopPolygon */


/* ----------------------------------------------------------------------------
 * fPolyStackEmpty:  This function returns TRUE if the polygon stack is empty.
 */
#if defined(ANSIPROTO)
BOOL
fPolyStackEmpty(
PMESH pmesh)				/* mesh structure */
#else
BOOL
fPolyStackEmpty(pmesh)
PMESH pmesh;
#endif
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.fPolyStackEmpty\n");
#endif

  return(pmesh->cpoly == 0);

} /* fPolyStackEmpty */


/* ----------------------------------------------------------------------------
 * nMinMaxAngles:  This function sets the minimum and maximum angle vertex
 * index pointers.  The number of vertices in the polygon is also updated.
 */
#if defined(ANSIPROTO)
int
nMinMaxAngles(
PMESH pmesh)				/* mesh structure */
#else
int
nMinMaxAngles(pmesh)
PMESH pmesh;
#endif
{
  PPOLY ppoly	 = pmesh->ppoly;
  PVTX	avtx	 = pmesh->avtx;
  IVTX	ivtxHead = ppoly->ivtxHead;

  double MinAngle;			/* minimum angle */
  double MaxAngle;			/* maximum angle */
  double Angle;				/* angle */
  IVTX	 ivtxMin;			/* index of minimum angle */
  IVTX	 ivtxMax;			/* index of maximum angle */
  IVTX	 ivtx;				/* vertex index */
  WORD	   cRun;												  /* run count */
  WORD	   cRunMax;												  /* max run count */
  WORD	 cvtx;				/* number of vertices */

#if defined(DEBUG_FUNC)
DEBUG_INFO("polygon.nMinMaxAngles\n");
#endif

  MinAngle = 2*PI;
  MaxAngle = 0;
  cvtx	   = 0;
  cRun	   = 0;
  cRunMax  = 0;
  ivtx	   = ivtxHead;
  do
  { Angle = avtx[ivtx].Angle;
    if (Angle < MinAngle) { ivtxMin = ivtx; MinAngle = Angle; }
    if (fabs(Angle - MaxAngle) < 0.01) ++cRun;
    else
    {	    if (cRun > cRunMax)
      { cRunMax = cRun;
	ivtxMax = avtx[ivtx].ivtxL;
	cRun = 0;
      }
    }
    if (Angle - MaxAngle >= 0.01)
    { ivtxMax = ivtx;
      cRun = 1;
      MaxAngle = Angle;
    }
    ivtx = avtx[ivtx].ivtxR;
    ++cvtx;
  } while (ivtx != ivtxHead);

  cRunMax /= 2;
  for(; cRunMax; --cRunMax) ivtxMax = avtx[ivtxMax].ivtxL;

  ppoly->ivtxMinAng = ivtxMin;
  ppoly->ivtxMaxAng = ivtxMax;
  ppoly->cvtx	    = cvtx;
  return(0);

} /* nMinMaxAngles */


/* ----------------------------------------------------------------------------
 * nWriteMesh:	This file generates the mesh data file.
 */
#if defined(ANSIPROTO)
int
nWriteMesh(
PSTR  pszFileName,		    /* name of translated skeleton file */
PMESH pmesh)			    /* mesh data structure */
#else
int
nWriteMesh(pszFileName, pmesh)
PSTR  pszFileName;
PMESH pmesh;
#endif
{
  SKELHEAD skelhead;		    /* head of skeleton file */
  int	   nFileIn;		    /* translated input file */
  int	   nFileOut;		    /* translated grid file */

  PSTR	   psz;
  UINT	   i;
  UINT	   c;

#if defined(DEBUG_FUNC)
DEBUG_INFO("mesh.nWriteMesh\n");
#endif

  /* form output name */
  strcpy(szGlobalBuf, pszFileName);
  psz = strchr(szGlobalBuf, '.');
  if (psz) *psz = '\0';
  strcat(szGlobalBuf, ".grd");

  /* open the skeleton file */
  nFileIn = open(pszFileName, O_RDONLY | O_BINARY);
  if (nFileIn == -1) vErrorHandler(ERROR_FILENOTFOUND, pszFileName, NULL, TRUE);

  /* create the grid file */
  nFileOut = open(szGlobalBuf, O_WRONLY | O_BINARY | O_TRUNC | O_CREAT, S_IREAD | S_IWRITE);
  if (nFileOut == -1) vErrorHandler(ERROR_FILEOPEN, szGlobalBuf, NULL, TRUE);

  /* create the grid header */
  read(nFileIn, &skelhead, sizeof(SKELHEAD));
  skelhead.cPoint  = pmesh->cnode;
  skelhead.cEdge   = pmesh->cedge;
  skelhead.cRegion = pmesh->ctri;
  write(nFileOut, &skelhead, sizeof(SKELHEAD));

  /* write constants, labels and refinement variables */
  c = skelhead.cConstant * sizeof(XLATCONST)  +
      skelhead.cLabel	 * sizeof(XLATLABEL)  +
      skelhead.cRefine	 * sizeof(XLATREFINE) +
      3			 * sizeof(WORD);
  psz = (char *) malloc(c);
  read (nFileIn,  psz, c);
  write(nFileOut, psz, c);
  free(psz);

  /* write nodes, edges and triangles */
  write(nFileOut, &skelhead.cPoint,  sizeof(UINT));
  write(nFileOut, pmesh->anode, skelhead.cPoint * sizeof(NODE));
  write(nFileOut, &skelhead.cEdge,   sizeof(UINT));
  for(i = 0; i < skelhead.cEdge; ++i)
    write(nFileOut, pmesh->aedge + i, sizeof(EDGE) - sizeof(ITRI));
  write(nFileOut, &skelhead.cRegion, sizeof(UINT));
  for(i = 0; i < skelhead.cRegion; ++i)
    write(nFileOut, pmesh->atri + i,  sizeof(TRI)  - sizeof(WORD));

  /* close the files */
  close(nFileIn);
  close(nFileOut);
  return(0);

} /* nWriteMesh */

