/*-----------------------------------------------------------------------------

 FILE:	   MESH.H

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/28/93    KMK      initial implementation

 DESCRIPTION:

 This module contains files which manipulate the mesh data structure.  The
 mesh data structure contains a dynamic array of nodes, a dynamic array of
 triangles, a doublely linked-list of vertices and a stack of polygons.

-----------------------------------------------------------------------------*/


#ifndef HEADER_MESH
#define HEADER_MESH


/******************************************************************************/
/***********************  PUBLIC TYPE DEFINITION  *****************************/
/******************************************************************************/


typedef WORD IVTX;			/* ivtx */
typedef WORD INODE;			/* inode */
typedef WORD IEDGE;			/* iedge */
typedef WORD ITRI;			/* itri */

typedef struct tagNODE			/* node */
{ REAL	x;				/* x coordinate */
  REAL	y;				/* y coordinate */
} NODE;
typedef NODE *PNODE;			/* pnode */

typedef struct tagEDGE			/* edge */
{ WORD	wLabel;				/* edge label */
  INODE inodeA;				/* index of node A */
  INODE inodeB;				/* index of node B */
  ITRI	itriAdj;			/* adjacency information */
} EDGE;
typedef EDGE *PEDGE;			/* pedge */

#define iedgeCD iedgeAC
#define itriCD	itriAC
typedef struct tagTRI			/* tri */
{ WORD	wLabel;				/* region label */
  IEDGE iedgeAB;			/* index of edge AB */
  IEDGE iedgeBC;			/* index of edge BC */
  IEDGE iedgeAC;			/* index of edge AC or CD */
  IEDGE iedgeAD;			/* index of edge AD */
  ITRI	itriAB;				/* index of tri adj to edge AB */
  ITRI	itriBC;				/* index of tri adj to edge BC */
  ITRI	itriAC;				/* index of tri adj to edge AC or CD */
  ITRI	itriAD;				/* index of tri adj to edge AD */
  WORD	wStatus;			/* status field */
} TRI;
typedef TRI *PTRI;			/* ptri */

typedef struct tagVTX			/* vtx */
{ INODE inode;				/* node index */
  IVTX	ivtxL;				/* index of left vertex */
  IVTX	ivtxR;				/* index of right vertex */
  IEDGE iedge;				/* index of edge (this & right vtx) */
  REAL	Angle;				/* angle */
  REAL	Hlocal;				/* local spacing */
} VTX;
typedef VTX *PVTX;			/* pvtx */

typedef struct tagPOLY			/* poly */
{ REAL	Hmax;				/* maximum edge length */
  WORD	wLabel;				/* region label */
  WORD	iRegion;			/* region index */
  WORD	cvtx;				/* number of vertices in polygon */
  IVTX	ivtxHead;			/* index of first vertex in polygon */
  IVTX	ivtxMaxAng;			/* index of vertex w/ largest angle */
  IVTX	ivtxMinAng;			/* index of vertex w/ smallest angle */
  BOOL	fRect;				/* divide into rectangular elements */
} POLY;
typedef POLY *PPOLY;			/* ppoly */

typedef struct tagMESH			/* mesh */
{ PNODE anode;				/* array of nodes     (dyn. array) */
  PEDGE aedge;				/* array of edges     (dyn. array) */
  PTRI	atri;				/* array of triangles (dyn. array) */
  PVTX	avtx;				/* array of vertices  (link list)  */
  PPOLY apoly;				/* array of polygons  (stack)	   */
  PPOLY ppoly;				/* ptr to top of stack polygon */
  WORD	cnode;				/* number of nodes */
  WORD	cedge;				/* number of edges */
  WORD	ctri;				/* number of triangles */
  WORD	cvtx;				/* number of vertices */
  WORD	cpoly;				/* number of polygons */
  WORD	canode;				/* size of node array */
  WORD	caedge;				/* size of edge array */
  WORD	catri;				/* size of triangle array */
  WORD	cavtx;				/* size of vertex array */
  WORD	capoly;				/* size of polygon array */
  IVTX	ivtxFree;			/* head of free vertex list */
} MESH;
typedef MESH *PMESH;			/* pmesh */


/******************************************************************************/
/*********************	PUBLIC CONSTANTS AND MACROS  **************************/
/******************************************************************************/


#define pnodeGetNode(pmesh,inode)   (pmesh->anode) + inode
#define pedgeGetEdge(pmesh,iedge)   (pmesh->aedge) + iedge
#define ptriGetTriangle(pmesh,itri) (pmesh->itri)  + itri
#define pvtxGetVertex(pmesh,ivtx)   (pmesh->avtx)  + ivtx
#define ppolyGetPolygon()	     pmesh->ppoly


/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


#if defined(ANSIPROTO)
PMESH	pmeshCreate	  (void);
INODE	inodeAddNode	  (PMESH,REAL,REAL,BOOL);
IEDGE	iedgeAddEdge	  (PMESH,INODE,INODE);
ITRI	itriAddTriangle	  (PMESH,IEDGE,IEDGE,IEDGE,IEDGE);
IVTX	ivtxAddVertex	  (PMESH,INODE,IEDGE);
int	nDeleteVertex	  (PMESH,IVTX);
int	nSplitPolygon	  (PMESH,IVTX,IVTX);
WORD	cPushPolygon	  (PMESH,PPOLY);
WORD	cPopPolygon	  (PMESH);
BOOL	fPolyStackEmpty	  (PMESH);
int	nMinMaxAngles	  (PMESH);
int	nWriteMesh	  (PSTR, PMESH);
#else
PMESH	pmeshCreate	  ();
INODE	inodeAddNode	  ();
IEDGE	iedgeAddEdge	  ();
ITRI	itriAddTriangle	  ();
IVTX	ivtxAddVertex	  ();
int	nDeleteVertex	  ();
int	nSplitPolygon	  ();
WORD	cPushPolygon	  ();
WORD	cPopPolygon	  ();
BOOL	fPolyStackEmpty	  ();
int	nMinMaxAngles	  ();
int	nWriteMesh	  ();
#endif


#endif
