/*-----------------------------------------------------------------------------

 FILE:     SKEL.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE     INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/19/93    KMK      initial implementation
 10/28/93    KMK      second rewrite
 11/02/94    KMK      new user interface implemented

 DESCRIPTION:

 This module contains the data structures used to describe the skeleton of
 the simulation domain.

-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#if defined(UNIX)
#include <unistd.h>
#define O_BINARY 0x0000
#else
#include <io.h>
#endif

#include "typedefs.h"
#include "literals.h"
#include "error.h"
#include "skel.h"
#include "xlatinfo.h"


/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * pskelCreate:   This function creates, initializes and returns a domain
 * skeleton structure.
 */
PSKEL pskelCreate ( PSTR pszFileName )
{
  SKELHEAD   skelhead;      /* skeleton header */
  XLATPOINT  xlatpoint;     /* translated point structure */
  XLATEDGE   xlatedge;      /* translated edge structure */
  XLATREGION xlatregion;    /* translated region structure */

  PSKEL pskel;        /* skeleton structure to create */
  PPNT  ppnt;         /* walking pointers */
  PSEG  pseg;         /* walking pointers */
  PREG  preg;         /* walking pointers */
  WORD  c;            /* count */
  long  lOffset;      /* file offset */
  int   nHandle;      /* file handle */
  int   i;            /* loop index */

  #if defined(DEBUG_FUNC)
  DEBUG_INFO("skel.pskelCreate\n");
  #endif

  /* open the skeleton file */
  nHandle = open(pszFileName, O_RDONLY | O_BINARY);
  if (nHandle == -1) vErrorHandler(ERROR_FILENOTFOUND, pszFileName, NULL, TRUE);

  /* read the header */
  read(nHandle, &skelhead, sizeof(SKELHEAD));

  /* move the file pointer to the beginning of the point structures */
  lOffset = (long) (skelhead.cConstant * sizeof(XLATCONST))  +
      (long) (skelhead.cLabel    * sizeof(XLATLABEL))  +
      (long) (skelhead.cRefine   * sizeof(XLATREFINE)) +
      (long) (3                  * sizeof(WORD));
  lseek(nHandle, (off_t) lOffset, SEEK_CUR);

  /* create and initialize the skeleton structure */
  pskel = (PSKEL) calloc(1, sizeof(SKEL));
  MEMCHECK(pskel, LIT_SKEL, LIT_SKEL1);
  pskel->capnt = skelhead.cPoint;
  pskel->caseg = skelhead.cEdge;
  pskel->careg = skelhead.cRegion;
  pskel->xMin = pskel->yMin =  1.0e100;
  pskel->xMax = pskel->yMax = -1.0e100;

  /* allocate the array fields */
  ppnt = pskel->apnt = (PPNT) calloc(skelhead.cPoint, sizeof(PNT));
  MEMCHECK(pskel->apnt, LIT_SKEL, LIT_SKEL1);
  pseg = pskel->aseg = (PSEG) calloc(skelhead.cEdge, sizeof(SEG));
  MEMCHECK(pskel->aseg, LIT_SKEL, LIT_SKEL1);
  preg = pskel->areg = (PREG) calloc(skelhead.cRegion, sizeof(REG));
  MEMCHECK(pskel->areg, LIT_SKEL, LIT_SKEL1);

  /* read the points */
  read(nHandle, &c, sizeof(WORD));
  if (c != skelhead.cPoint)
    vErrorHandler(ERROR_CORRUPT, pszFileName, NULL, TRUE);
  for(i = 0; i < c; ++i, ++ppnt)
  { read(nHandle, &xlatpoint, sizeof(XLATPOINT));
    ppnt->x = xlatpoint.x;
    ppnt->y = xlatpoint.y;
  }

  /* read the edges */
  read(nHandle, &c, sizeof(WORD));
  if (c != skelhead.cEdge)
    vErrorHandler(ERROR_CORRUPT, pszFileName, NULL, TRUE);
  for(i = 0; i < c; ++i, ++pseg)
  { read(nHandle, &xlatedge, sizeof(XLATEDGE));
    pseg->wLabel = xlatedge.iLabel;
    pseg->ipntA = xlatedge.iPointA;
    pseg->ipntB = xlatedge.iPointC;
    pseg->ipntM = xlatedge.iPointB;
    pseg->cipnt = 0;
    pseg->cSegment = xlatedge.cSegment;
    pseg->InitialSpacing = xlatedge.InitialSpacing;
    pseg->SpacingGrade = xlatedge.SpacingGrade;
  }

  /* read the regions */
  read(nHandle, &c, sizeof(WORD));
  if (c != skelhead.cRegion) vErrorHandler(ERROR_CORRUPT, pszFileName, NULL, TRUE);
  for(i = 0; i < c; ++i, ++preg)
  { read(nHandle, &xlatregion, sizeof(XLATREGION));
    preg->Hmax   = xlatregion.Hmax;
    preg->wLabel = xlatregion.iLabel;
    preg->ciseg   = xlatregion.cEdge;
    preg->aiseg   = (ISEG *) calloc(xlatregion.cEdge, sizeof(ISEG));
    preg->fRect   = xlatregion.fRect;
    MEMCHECK(preg->aiseg, LIT_SKEL, LIT_SKEL1);
    read(nHandle, preg->aiseg, xlatregion.cEdge * sizeof(ISEG));
  }

  /* close the file and return the skeleton structure */
  close(nHandle);
  return(pskel);

} /* pskelCreate */



