/*-----------------------------------------------------------------------------

 FILE:	   SKL2MSH.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/29/93    KMK      initial implementation

 DESCRIPTION:

 This module contains functions that convert the skeleton structure to the
 mesh structure.

-----------------------------------------------------------------------------*/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "typedefs.h"
#include "literals.h"
#include "error.h"
#include "skel.h"
#include "mesh.h"
#include "geometry.h"


#if defined(ANSIPROTO)
int    nAddNewPoints   (PSKEL);
PREAL  aAssignWeights  (PSKEL);
int    nFormPolygon    (PMESH,PSKEL,IREG,IEDGE *);
#else
int    nAddNewPoints   ();
PREAL  aAssignWeights  ();
int    nFormPolygon    ();
#endif


/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * nSkel2Mesh:	This function converts the skeleton structure to the mesh
 * structure.
 */
#if defined(ANSIPROTO)
int
nSkel2Mesh(
PSKEL pskel,				/* skeleton structure */
PMESH pmesh)				/* mesh structure */
#else
int
nSkel2Mesh(pskel,pmesh)
PSKEL pskel;
PMESH pmesh;
#endif
{
  PPNT	 apnt  = pskel->apnt;
  PSEG	 aseg  = pskel->aseg;
  WORD	 capnt = pskel->capnt;
  WORD	 caseg = pskel->caseg;
  WORD	 careg = pskel->careg;

  IEDGE *aiedge;			/* array of edge indices */
  IEDGE *piedge;			/* walking pointer */
  IEDGE	 iedge;				/* edge index */
  IPNT	*pipnt;				/* walking pointer */
  PPNT	 ppnt;				/* walking pointer */
  PSEG	 pseg;				/* walking pointer */
  PPNT	 ppntA,ppntM,ppntB;		/* point structures of A, M & B */
  REAL	 xc,yc;				/* coordinates of circle center */
  REAL	 angleA,angleB;			/* arc angles */
  REAL	 r;				/* circle radius */
  REAL	 d;				/* distance */
  REAL	 Spacing;			/* default spacing */
  WORD	 c;				/* count */
  int	 i,j;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("skl2msh.nSkel2Mesh\n");
#endif

  /* calculate curved edge information and default spacing */
  Spacing = -1.0;
  for(i = 0, pseg = aseg; i < caseg; ++i, ++pseg)
  { ppntA = apnt + pseg->ipntA;
    ppntB = apnt + pseg->ipntB;

    if (pseg->ipntM != (IPNT) -1)
    { ppntM = apnt + pseg->ipntM;
      nCircleCenter(ppntA->x, ppntA->y, ppntM->x, ppntM->y,
		    ppntB->x, ppntB->y, &xc, &yc);
      nParamArc	   (ppntA->x, ppntA->y, ppntM->x, ppntM->y,
		    ppntB->x, ppntB->y, xc, yc, &angleA, &angleB, &r);
      pseg->xCenter = xc;
      pseg->yCenter = yc;
      pseg->Radius  = r;
      pseg->ThetaA  = angleA;
      pseg->ThetaB  = angleB;
      d = r * angleA - angleB;
      if (d < 0.0) d = -d;
      if (d > Spacing) Spacing = d;
    }
    else
    { d = sqrt(sq(ppntA->x - ppntB->x) + sq(ppntA->y - ppntB->y));
      if (d > Spacing) Spacing = d;
    }
  }
  if (cmdargs.fAutoSpace) cmdargs.Hmax = Spacing / 10.0;

  /* remove the midpoints of curved segments from the point array */
  for(i = 0; i < caseg; ++i)
  { IPNT ipnt = aseg[i].ipntM;
    if (ipnt == (IPNT) -1) continue;

    /* remove the point */
    --capnt;
    for(j = ipnt; j < capnt; ++j)
    { apnt[j].x = apnt[j+1].x;
      apnt[j].y = apnt[j+1].y;
    }

    /* update the segments */
    for(j = 0, pseg = aseg; j < caseg; ++j, ++pseg)
    { if (pseg->ipntA > ipnt) --pseg->ipntA;
      if (pseg->ipntB > ipnt) --pseg->ipntB;
      if ((pseg->ipntM != (IPNT) -1) && (pseg->ipntM > ipnt)) --pseg->ipntM;
    }
  }
  pskel->capnt = capnt;

  /* add points to each segments */
  nAddNewPoints(pskel);
  apnt	= pskel->apnt;
  capnt = pskel->capnt;

  /* copy the data in the skeleton data structure to the mesh data structure */
  aiedge = piedge = (IEDGE *) calloc(caseg, sizeof(IEDGE));
  MEMCHECK(aiedge, LIT_SKL2MSH, LIT_SKL2MSH1);
  for(i = 0, ppnt = apnt; i < capnt; ++i, ++ppnt)
    inodeAddNode(pmesh, ppnt->x, ppnt->y, TRUE);
  for(i = 0, pseg = aseg; i < caseg; ++i, ++pseg)
  { *(piedge++) = pmesh->cedge;
    c = pseg->cipnt - 1;
    for(j = 0, pipnt = pseg->aipnt; j < c; ++j, ++pipnt)
    { iedge = iedgeAddEdge(pmesh, pipnt[0], pipnt[1]);
      pmesh->aedge[iedge].wLabel = pseg->wLabel;
    }
  }
  for(i = 0; i < careg; ++i) nFormPolygon(pmesh, pskel, i, aiedge);

  /* deallocate the edge index array and return */
  free(aiedge);
  return(0);


} /* nSkel2Mesh */


/******************************************************************************/
/**************************  PRIVATE FUNCTIONS	*******************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * nAddNewPoints:  This function adds nodes to the edges of the skeleton.
 */
#if defined(ANSIPROTO)
int
nAddNewPoints(
PSKEL pskel)				/* skeleton structure */
#else
int
nAddNewPoints(pskel)
PSKEL pskel;
#endif
{
  PPNT	apnt  = pskel->apnt;
  WORD	capnt = pskel->capnt;
  WORD	caseg = pskel->caseg;

  PREAL aH;				/* weight vector */
  PSEG	pseg;				/* walking pointer */
  IPNT *aipnt;				/* array of point indices */
  IPNT *pipnt;				/* walking pnter */
  IPNT	ipntA;				/* index of point A */
  IPNT	ipntB;				/* index of point B */
  IPNT	ipntM;				/* index of point M */
  int	cNP;				/* number of new pnts */
  REAL	H;				/* node spacing */
  REAL	dH;				/* delta node spacing */
  REAL	Hgrade;				/* spacing grade */
  REAL	D;				/* seg length */
  REAL	R;				/* curve radius */
  REAL	dA;				/* delta angle */
  REAL	Arg;				/* angle */
  REAL	x,y;				/* (x,y) coordinates */
  REAL	dx,dy;				/* delta (x,y) coordinates */
  WORD	cSeg;				/* number of segments */
  int	i,j,k;				/* loop indices */

#if defined(DEBUG_FUNC)
DEBUG_INFO("skl2msh.nAddNewPoints\n");
#endif

  /* construct the weight vector */
  aH = aAssignWeights(pskel);

  /* loop through each segment and add points */
  /* do the segments with the specified grade first so that weight vector */
  /* can be updated */
  for(i = 0; i < 2; ++i)
  { for(j = 0, pseg = pskel->aseg; j < caseg; ++j, ++pseg)
    { Hgrade = pseg->SpacingGrade;
      if ((!i && (Hgrade < 0.0)) || (i && pseg->cipnt)) continue;
      ipntA  = pseg->ipntA;
      ipntB  = pseg->ipntB;
      ipntM  = pseg->ipntM;
      cSeg   = pseg->cSegment;
      dH     = pseg->InitialSpacing;
      /* calculate the number of points to add to the segment */
      if (ipntM != (WORD) -1)
      { R  = pseg->Radius;
	dA = pseg->ThetaB - pseg->ThetaA;
	D  = fabs(dA) * R;
      }
      else
      { x  = apnt[ipntA].x;
	y  = apnt[ipntA].y;
	dx = apnt[ipntB].x - x;
	dy = apnt[ipntB].y - y;
	D  = sqrt(dx*dx + dy*dy);
      }

      cNP = (Hgrade < 0.0)
	  ? cNumNewPoints(aH[ipntA], aH[ipntB], D, cSeg, &dH, &Hgrade)
	  : cComputeGrade(D, &dH, &Hgrade);
      pseg->cipnt = cNP + 2;
      pseg->aipnt = pipnt = aipnt = (IPNT *) calloc(cNP + 2, sizeof(IPNT));
      MEMCHECK(aipnt, LIT_SKL2MSH, LIT_SKL2MSH2);
      *(pipnt++) = ipntA;

      if (cNP)
      { if (!i) aH[ipntA] = min(dH,aH[ipntA]);
	H = 0.0;
	apnt = (PPNT) realloc(apnt, (capnt + cNP) * sizeof(PNT));
	MEMCHECK(apnt, LIT_SKL2MSH, LIT_SKL2MSH2);

	for(k = 0; k < cNP; ++k, ++pipnt)
	{ H  += dH;
	  dH *= Hgrade;
	  if (ipntM != (IPNT) -1)
	  { Arg = pseg->ThetaA + H * dA;
	    apnt[capnt].x = pseg->xCenter + R * cos(Arg);
	    apnt[capnt].y = pseg->yCenter + R * sin(Arg);
	  }
	  else
	  { apnt[capnt].x = x + H * dx;
	    apnt[capnt].y = y + H * dy;
	  }
	  *pipnt = capnt++;
	}
	if (!i) aH[ipntB] = min(H,aH[ipntB]);
      }
      *pipnt = ipntB;
    }
  }

  /* update the node array & # of nodes, deallocate the weight vector */
  /* and return */
  pskel->apnt  = apnt;
  pskel->capnt = capnt;
  free(aH);
  return(capnt);

} /* nAddNewPoints */


/* ----------------------------------------------------------------------------
 * aAssignWeights:  This function returns a vector of vertex weights.
 */
#if defined(ANSIPROTO)
PREAL
aAssignWeights(
PSKEL pskel)				/* skeleton structure */
#else
PREAL
aAssignWeights(pskel)
PSKEL pskel;
#endif
{
  PPNT	apnt   = pskel->apnt;
  WORD	capnt  = pskel->capnt;
  WORD	caseg  = pskel->caseg;
  REAL	Hmax   = cmdargs.Hmax;

  PSEG	pseg;				/* walking pnter */
  PREAL aH;				/* weight vector */
  PREAL pd;				/* walking pnter */
  REAL	D;				/* seg length */
  IPNT	ipntA,ipntM,ipntB;		/* point indices of A, M and B */
  int	i;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("skl2msh.aAssignWeights\n");
#endif

  /* allocate the weight vector */
  aH = (PREAL) calloc(capnt, sizeof(REAL));
  MEMCHECK(aH, LIT_SKL2MSH, LIT_SKL2MSH3);
  for(i = 0, pd = aH; i < capnt; ++i, ++pd) *pd = Hmax;

  /* compute the vertex weights */
  for(i = 0, pseg = pskel->aseg; i < caseg; ++i, ++pseg)
  { ipntA = pseg->ipntA;
    ipntB = pseg->ipntB;
    ipntM = pseg->ipntM;
    if (ipntM != (WORD) -1)
	 D = min(fabs(pseg->ThetaA - pseg->ThetaB) , PI/4.0) * pseg->Radius;
    else D = sqrt(sq(apnt[ipntA].x - apnt[ipntB].x) +
		  sq(apnt[ipntA].y - apnt[ipntB].y));
    aH[ipntA] = min(D, aH[ipntA]);
    aH[ipntB] = min(D, aH[ipntB]);
  }

  return(aH);

} /* aAssignWeights */


/* ----------------------------------------------------------------------------
 * nFormPolygon:  This function makes a polygon from the specified region
 */
#if defined(ANSIPROTO)
int
nFormPolygon(
PMESH  pmesh,				/* mesh structure */
PSKEL  pskel,				/* skeleton domain structure */
IREG   ireg,				/* specified region */
IEDGE *aiedge)				/* array of edge indices */
#else
int
nFormPolygon(pmesh,pskel,ireg,aiedge)
PMESH  pmesh;
PSKEL  pskel;
IREG   ireg;
IEDGE *aiedge;
#endif
{
  PSEG	 aseg  = pskel->aseg;
  PREG	 preg  = pskel->areg + ireg;
  ISEG	*aiseg = preg->aiseg;
  WORD	 ciseg = preg->ciseg;

  POLY	 poly;				/* polygon */
  PPOLY	 ppoly;				/* polygon pointer */
  PNODE	 anode;				/* node array */
  PVTX	 avtx;				/* vertex array */
  PVTX	 pvtx;				/* vertex pointer */
  PSEG	 pseg;				/* segment pointer */
  IPNT	*pipnt;				/* walking pointer */
  ISEG	*piseg;				/* walking pointer */
  IEDGE	 iedge;				/* edge index */
  IPNT	 ipntLast;			/* last point index */
  IVTX	 ivtxHead;			/* first vertex in polygon */
  IVTX	 ivtx;				/* vertex index */
  INODE	 inodeL,inodeM,inodeR;		/* node indices */
  WORD	 cvtx;				/* number of vertices in polygon */
  WORD	 c;				/* count */
  BOOL	 fForward;			/* direction flag */
  REAL	 xMin,yMin,xMax,yMax;		/* minimum and maximum coordinates */
  REAL	 x,y;				/* coordinates */
  int	 i,j;				/* loop indices */

#if defined(DEBUG_FUNC)
DEBUG_INFO("skl2msh.nFormPolygon\n");
#endif

  /* initialize and store a polygon structure */
  poly.fRect	= preg->fRect;
  poly.Hmax	= preg->Hmax;
  poly.wLabel	= preg->wLabel;
  poly.iRegion	= ireg;
  poly.cvtx	= 0;
  poly.ivtxHead = 0;
  ipntLast	= 0;
  cPushPolygon(pmesh,&poly);

  for(i = 0, piseg = aiseg; i < ciseg; ++i, ++piseg)
  { pseg  = aseg + *piseg;
    if (i) fForward =  (pseg->ipntA == ipntLast);
    else   fForward = ((pseg->ipntB == aseg[piseg[1]].ipntA) ||
		       (pseg->ipntB == aseg[piseg[1]].ipntB));
    c	  = pseg->cipnt - 1;
    pipnt = pseg->aipnt	   + ((fForward) ? 0 : c);
    iedge = aiedge[*piseg] + ((fForward) ? 0 : c - 1);
    for(j = 0; j < c; ++j)
    { ivtx = ivtxAddVertex(pmesh, *pipnt, iedge);
      if (fForward) { ++pipnt; ++iedge; }
      else	    { --pipnt; --iedge; }
    }
    ipntLast = *pipnt;
  }

  anode	   = pmesh->anode;
  ppoly	   = pmesh->ppoly;
  avtx	   = pmesh->avtx;
  cvtx	   = ppoly->cvtx;
  ivtxHead = ppoly->ivtxHead;
  xMin	   = pskel->xMin;
  yMin	   = pskel->yMin;
  xMax	   = pskel->xMax;
  yMax	   = pskel->yMax;
  for(i = 0, ivtx = ivtxHead; i < cvtx; ++i, ivtx = pvtx->ivtxR)
  { pvtx   = avtx + ivtx;
    inodeL = avtx[pvtx->ivtxL].inode;
    inodeM = pvtx->inode;
    inodeR = avtx[pvtx->ivtxR].inode;
    pvtx->Angle	 = CompAngleI (anode,inodeL,inodeM,inodeR);
    pvtx->Hlocal = CompHlocalI(anode,inodeL,inodeM,inodeR);
    x = anode[inodeM].x;
    y = anode[inodeM].y;
    if (x < xMin) xMin = x;
    if (x > xMax) xMax = x;
    if (y < yMin) yMin = y;
    if (y > yMax) yMax = y;
  }
  nMinMaxAngles(pmesh);
  pskel->xMin	    = xMin;
  pskel->yMin	    = yMin;
  pskel->xMax	    = xMax;
  pskel->yMax	    = yMax;
  return(0);

} /* nFormPolygon */

