/*-----------------------------------------------------------------------------

 FILE:	   TRIANG.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/25/93    KMK      initial implementation
 11/02/93    KMK      second rewrite

 DESCRIPTION:

 This module contains functions which perform the triangulation of a region.

-----------------------------------------------------------------------------*/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if 0
#include <conio.h>
#endif

#include "typedefs.h"
#include "literals.h"
#include "error.h"
#include "mesh.h"
#include "enhance.h"
#include "geometry.h"
#include "graphics.h"


typedef struct tagPARAM			/* param */
{ double Quality;
  double MaxAngle;
  double MinAngle;
  double Grade;
  double Hmax;
  double Acceptable;
  double Best;
} PARAM;
typedef PARAM *PPARAM;			/* pparam */

static PARAM param;


IVTX aivtxList[1024];
WORD civtxList;


#define CompAng(x,y,z) CompAngleI(anode,avtx[x].inode,		  \
				  avtx[y].inode,avtx[z].inode)

#define CompHloc(x,y,z) CompHlocalI(anode,avtx[x].inode,	 \
				    avtx[y].inode,avtx[z].inode)

#define CompQual(x,y,z) CompQualityI(anode,avtx[x].inode,	  \
				     avtx[y].inode,avtx[z].inode)

#define STATUS_NO	 0
#define STATUS_OK	-1
#define STATUS_YES	 1
#define STATUS_SPLIT	 1
#define STATUS_DONE	 0


#if defined(ANSIPROTO)
int   nTriangPolygon	(PMESH);
int   nRectangPolygon	(PMESH);
int   nChopOK		(PMESH,IVTX);
int   nChopVertex	(PMESH,IVTX);
int   nCentroidOK	(PMESH);
int   nAddCentroid	(PMESH);
#else
int   nTriangPolygon	();
int   nRectangPolygon	();
int   nChopOK		();
int   nChopVertex	();
int   nCentroidOK	();
int   nAddCentroid	();
#endif


/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * nInitTriangModule:  This function initializes the triangulation module.
 */
int
nInitTriangModule()
{

#if defined(DEBUG_FUNC)
DEBUG_INFO("triang.nInitTriangModule\n");
#endif

  param.Quality	   = 0.5*sqrt(3.0);
  param.MinAngle   = PI/4.0 - EPS;
  param.MaxAngle   = PI/2.0 + EPS;
  param.Grade	   = cmdargs.Grade;
  param.Hmax	   = cmdargs.Hmax;
  param.Acceptable = 0.98;
  return(0);

} /* nInitTriangModule */


/* ----------------------------------------------------------------------------
 * nTriangulate:  This function recursively triangulates the mesh.
 */
#if defined(ANSIPROTO)
int
nTriangulate(
PMESH pmesh)				/* mesh structure */
#else
int
nTriangulate(pmesh)
PMESH pmesh;
#endif
{
  int i;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("triang.nTriangulate\n");
#endif

  while (!fPolyStackEmpty(pmesh))
  {
#if 0
    if (cmdargs.wVisual == VISUAL_CONSTRUCT)
    { nGraphPolygon(pmesh);
      if (cmdargs.cNoPress != -1)
      { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	else { if (getch() == 'q') exit(20); }
      }
    }
#endif
    if (pmesh->ppoly->fRect) nRectangPolygon(pmesh);
    else		     nTriangPolygon(pmesh);
  }

  for(i = 0; i < cmdargs.cQualEnh1; ++i) nEnhMeshEdge(pmesh);
  for(i = 0; i < cmdargs.cQualEnh2; ++i) nEnhMeshNode(pmesh);

  for(i = 0; i < cmdargs.cQualEnh3; ++i)
  { nEnhMeshEdge(pmesh);
    nEnhMeshNode(pmesh);
  }

  return(0);

} /* nTriangulate */


/******************************************************************************/
/**************************  PRIVATE FUNCTIONS	*******************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * nChopOK:  This function determines if it is ok to chop a triangle from
 * vertex ivtx of the polygon.	The return values are as follows:
 *
 *   STATUS_NO:	    do not chop this vertex
 *   STATUS_OK:	    it is ok to chop this vertex
 *   STATUS_YES:    definitely chop this vertex
 */
#if defined(ANSIPROTO)
int
nChopOK(
PMESH pmesh,				/* mesh structure */
IVTX  ivtx)				/* index of vertex to chop */
#else
int
nChopOK(pmesh,ivtx)
PMESH pmesh;
IVTX  ivtx;
#endif
{
  PPOLY ppoly	= pmesh->ppoly;
  PNODE anode	= pmesh->anode;
  PEDGE aedge	= pmesh->aedge;
  PVTX	avtx	= pmesh->avtx;
  WORD	cvtx	= ppoly->cvtx;
  IVTX	ivtxMax = ppoly->ivtxMaxAng;
  PVTX	pvtxM	= avtx + ivtx;
  IVTX	ivtxL	= pvtxM->ivtxL;
  IVTX	ivtxR	= pvtxM->ivtxR;
  PVTX	pvtxL	= avtx + ivtxL;
  PVTX	pvtxR	= avtx + ivtxR;
  INODE inodeL	= pvtxL->inode;
  INODE inodeM	= pvtxM->inode;
  INODE inodeR	= pvtxR->inode;

  REAL	Angle;				/* angle of vertex ivtx (?) */
  REAL	AngleL;				/* angle of vertex left of ivtx */
  REAL	AngleR;				/* angle of vertex right of ivtx */
  REAL	DistLR;				/* distance */
  REAL	DistL;				/* distance */
  REAL	DistR;				/* distance */
  REAL	Qual;				/* triangle quality factor */
  REAL	Qual1;				/* triangle quality factor */
  REAL	Qual2;				/* triangle quality factor */
  REAL	Rating;				/* feasibility of this chop */
  REAL	Hlocal;				/* local spacing */
  REAL	Amul;				/* length of intersection along dA */
  REAL	Bmul;				/* length of intersection along dB */
  CRDNT crdntA;				/* coordinates of point A */
  CRDNT crdntdA;			/* coordinates of vector dA */
  CRDNT crdntB;				/* coordinates of point B */
  CRDNT crdntdB;			/* coordinates of vector dB */
  IVTX	ivtxW;				/* walking index */
  INODE inodeN;				/* node index */
  int	i;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("traing.nChopOK\n");
#endif

  /* initialize the chop rating to highest setting */
  Rating = 1.0;

  /* special case if polygon is a quadralateral */
  if (cvtx == 4)
  { if (avtx[ivtxMax].Angle >= PI + EPS)
    { if ((ivtxMax == ivtxR) || (ivtxMax == ivtxL))
      { param.Best = Rating;
	return((Rating < 1.0) ? STATUS_OK : STATUS_YES);
      }
      else return(STATUS_NO);
    }
    inodeN = avtx[pvtxR->ivtxR].inode;
    Qual1  = CompQualityI(anode,inodeL,inodeM,inodeR);
    Qual2  = CompQualityI(anode,inodeL,inodeN,inodeR);
    Qual   = min(fabs(Qual1),fabs(Qual2));
    Rating = min(Rating,Qual/param.Quality);
    if (Rating < param.Best) return(STATUS_NO);

    /* @@@ check for two curved edges */
  }
  else
  {
    /* check the angles of the proposed new polygon */
    Angle  = pvtxM->Angle;
    AngleL = CompAngleI(anode,inodeR,inodeL,inodeM);
    AngleR = CompAngleI(anode,inodeM,inodeR,inodeL);
    if ((AngleL < param.MinAngle) || (AngleL > param.MaxAngle) ||
	(AngleR < param.MinAngle) || (AngleR > param.MaxAngle) ||
	(Angle	< param.MinAngle) || (Angle  > param.MaxAngle))
      return(STATUS_NO);

    Angle  = min(pvtxL->Angle - AngleL, pvtxR->Angle - AngleR);
    Angle /= PI / 3.0;
    if (Angle > 1.0) Angle = 1.0 / Angle;
    Rating = min(Rating,Angle);
/*  Rating = min(Rating,Angle/param.MinAngle);	*/
    if (Rating < param.Best) return(STATUS_NO);

    /* check side lengths of the prosed new polygon */
    DistLR = sqrt(sq(anode[inodeL].x - anode[inodeR].x)+
		      sq(anode[inodeL].y - anode[inodeR].y));
    DistL  = pvtxL->Hlocal / DistLR;
    DistR  = pvtxR->Hlocal / DistLR;
    if(DistL < 1.0) DistL = 1.0 / DistL;
    if(DistR < 1.0) DistR = 1.0 / DistR;
    DistLR = max(DistL, DistR);
    Rating = min(Rating, param.Grade / DistLR);
    if (Rating < param.Best) return(STATUS_NO);

    /* check the geometry of the propsed triangle */
    Qual = CompQualityI(anode,inodeL,inodeM,inodeR);
    Rating = min(Rating, fabs(Qual) / param.Quality);
    if (Rating < param.Best) return(STATUS_NO);

    /* derate the triangle if it is composed of two boundaries */
    if (aedge[pvtxL->iedge].wLabel && aedge[pvtxM->iedge].wLabel) Rating /= 2.0;

    /* if polygon is convex, then return OKCHOP or YESCHOP */
    if (avtx[ivtxMax].Angle <= PI + EPS)
    { param.Best = Rating;
      return((Rating < 1.0) ? STATUS_OK : STATUS_YES);
    }

    /* make sure not other vertices are inside the proposed */
    /* triangle or are too close */
    Hlocal    = pvtxM->Hlocal;
    crdntA.x  = anode[inodeR].x;
    crdntA.y  = anode[inodeR].y;
    crdntdA.x = anode[inodeL].x - crdntA.x;
    crdntdA.y = anode[inodeL].y - crdntA.y;
    crdntB.x  = anode[inodeM].x;
    crdntB.y  = anode[inodeM].y;
    ivtxW     = ivtxR;
    for(i = 3; i < cvtx; ++i)
    { ivtxW = avtx[ivtxW].ivtxR;
      if (avtx[ivtxW].Angle <= PI + EPS) continue;
      crdntdB.x = anode[avtx[ivtxW].inode].x - crdntB.x;
      crdntdB.y = anode[avtx[ivtxW].inode].y - crdntB.y;

      if ((nIntersection(&crdntA,&crdntdA,&crdntB,&crdntdB,
			 &Amul,&Bmul) == GEOMETRY_INTERSECT) ||
	  (Amul < -EPS)					     ||
	  (Amul > 1.0 + EPS)) continue;

      if (Bmul > 1.0) return(STATUS_NO);

      if ((avtx[ivtxW].Hlocal + Hlocal) * Bmul < Hlocal) continue;

      Qual1  = CompQualityI(anode,inodeL,inodeR,avtx[ivtxW].inode);
      Rating = min(Rating, fabs(Qual1)/param.Quality);
      if (Rating < param.Best) return(STATUS_NO);
    }
  }

  param.Best = Rating;
  return((Rating < 1.0) ? STATUS_OK : STATUS_YES);

} /* nChopOK */


/* ----------------------------------------------------------------------------
 * nChopVertex:	 This function chops the specified vertex and adds the
 * resulting triangle to the mesh structure.  The index of the triangle is
 * returned.
 */
#if defined(ANSIPROTO)
int
nChopVertex(
PMESH pmesh,				/* mesh structure */
IVTX  ivtx)				/* index of vertex to chop */
#else
int
nChopVertex(pmesh,ivtx)
PMESH pmesh;
IVTX  ivtx;
#endif
{
  PNODE anode  = pmesh->anode;
  PVTX	avtx   = pmesh->avtx;
  PVTX	pvtx   = avtx + ivtx;
  IVTX	ivtxL  = pvtx->ivtxL;
  IVTX	ivtxR  = pvtx->ivtxR;
  PVTX	pvtxL  = avtx + ivtxL;
  PVTX	pvtxR  = avtx + ivtxR;
  INODE inodeL = avtx[ivtxL].inode;
  INODE inodeM = pvtx->inode;
  INODE inodeR = avtx[ivtxR].inode;

  INODE inodeLL,inodeRR;		/* node indices */
  IEDGE iedge;				/* edge index */
  ITRI	itri;				/* triangle index */
  IVTX	ivtxLL,ivtxRR;			/* vertex indices */

#if defined(DEBUG_FUNC)
DEBUG_INFO("triang.nChopVertex\n");
#endif

  if (pmesh->ppoly->cvtx > 3) iedge = iedgeAddEdge(pmesh,inodeL,inodeR);
  else			      iedge = pvtxR->iedge;
  itri	= itriAddTriangle(pmesh,pvtxL->iedge,pvtx->iedge,iedge,-1u);
  pmesh->atri[itri].wLabel = pmesh->ppoly->wLabel;
  if (cmdargs.wVisual == VISUAL_CONSTRUCT)
    nDrawTriangle(anode,inodeL,inodeM,inodeR);	 /* @@@ */
  nDeleteVertex(pmesh,ivtx);

  if (pmesh->ppoly->cvtx < 3)
  {
    nDeleteVertex(pmesh,ivtxL);
    nDeleteVertex(pmesh,ivtxR);
    cPopPolygon(pmesh);
  }
  else
  {
    ivtxLL  = avtx[ivtxL].ivtxL;
    ivtxRR  = avtx[ivtxR].ivtxR;
    inodeLL = avtx[ivtxLL].inode;
    inodeRR = avtx[ivtxRR].inode;

/*  linking is already done by nDeleteVertex
    avtx[ivtxL].ivtxR = ivtxR;
    avtx[ivtxR].ivtxL = ivtxL;		      */
    pvtxL->iedge = iedge;
    pvtxL->Hlocal = CompHlocalI(anode,inodeLL,inodeL,inodeR);
    pvtxR->Hlocal = CompHlocalI(anode,inodeL,inodeR,inodeRR);
    pvtxL->Angle  = CompAngleI(anode,inodeLL,inodeL,inodeR);
    pvtxR->Angle  = CompAngleI(anode,inodeL,inodeR,inodeRR);
    nMinMaxAngles(pmesh);
  }

  /* @@@ a triangle with two curved boundary edges ??? */
#if 0
  if ((cmdargs.wVisual == VISUAL_CONSTRUCT) && (cmdargs.cNoPress != -1))
  { if (cmdargs.cNoPress) --cmdargs.cNoPress;
    else { if (getch() == 'q') exit(20); }
  }
#endif
  return(0);

} /* nChopVertex */


/* ----------------------------------------------------------------------------
 * nCentroidOK:	 This function determines if it is ok to triangulate a polygon
 * by adding its centroid and triangulating the entire region.
 *
 *   STATUS_NO:	    do not chop this vertex
 *   STATUS_OK:	    it is ok to chop this vertex
 *   STATUS_YES:    definitely chop this vertex
 */
#if defined(ANSIPROTO)
int
nCentroidOK(
PMESH pmesh)				/* mesh structure */
#else
int
nCentroidOK(pmesh)
PMESH pmesh;
#endif
{
  PPOLY ppoly	= pmesh->ppoly;
  PNODE anode	= pmesh->anode;
  PVTX	avtx	= pmesh->avtx;
  IVTX	ivtxMax = ppoly->ivtxMaxAng;
  WORD	cvtx	= ppoly->cvtx;
  REAL	Best	= param.Best;
  REAL	Quality = param.Quality;

  INODE inode;				/* node index of centroid */
  PVTX	pvtx;				/* walking pointer */
  PVTX	pvtxR;				/* right neighbor of pvtx */
  REAL	x,y;				/* coordinate of centroid */
  REAL	Qual;				/* triangle quality factor */
  REAL	Rating;				/* feasibility of this operation */
  int	i;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("traing.nCentroidOK\n");
#endif

  /* determine if the polygon should be triangulated by the centroid */
  /* method ... check the number of vertices and the maximum angle   */
  if ((cvtx < 4) || (cvtx > 8)) return(STATUS_NO);
  pvtx = avtx + ivtxMax;
  if (pvtx->Angle > PI) return(STATUS_NO);
  switch (cvtx)
  { case 8:   Rating = 0.8; break;
    case 7:   Rating = 0.9; break;
    default:  Rating = 1.0;
  }
  if (Rating <= Best) return(STATUS_NO);

  /* compute the centroid */
  x = y = 0.0;
  for(i = 0; i < cvtx; ++i)
  { inode = pvtx->inode;
    x += anode[inode].x;
    y += anode[inode].y;
    pvtx = avtx + pvtx->ivtxR;
  }
  x /= (double) cvtx;
  y /= (double) cvtx;

  /* store the centroid in the node array ... if this method is not used */
  /* then the centroid will be overwritten */
  inode = inodeAddNode(pmesh,x,y,FALSE);
  anode = pmesh->anode;

  /* check the quality factor of the resulting triangles */
  for(i = 0; i < cvtx; ++i)
  { pvtxR  = avtx + pvtx->ivtxR;
    Qual   = CompQualityI(anode,pvtxR->inode,pvtx->inode,inode);
    Rating = min(Rating,fabs(Qual)/Quality);
    if (Rating <= Best) return(STATUS_NO);
    pvtx = pvtxR;
  }

  param.Best = Rating;
  return((Rating < 1.0) ? STATUS_OK : STATUS_YES);

} /* nCentroidOK */


/* ----------------------------------------------------------------------------
 * nAddCentroid:  This function traingulates a polygon by introducing its
 * centroid and splitting it into n triangles.
 */
#if defined(ANSIPROTO)
int
nAddCentroid(
PMESH pmesh)				/* mesh structure */
#else
int
nAddCentroid(pmesh)
PMESH pmesh;
#endif
{
  PPOLY	 ppoly	  = pmesh->ppoly;
  PNODE	 anode	  = pmesh->anode;
  PVTX	 avtx	  = pmesh->avtx;
  WORD	 cvtx	  = ppoly->cvtx;

  IEDGE *piedge;			/* edge index pointer */
  IEDGE	 aiedge[9];			/* array of edge indices */
  INODE	 inode;				/* node index */
  ITRI	 itri;				/* triangle index */
  PVTX	 pvtx;				/* walking pointer */
  PVTX	 pvtxR;				/* right neighbor of pvtx */
  REAL	 x,y;				/* centroid */
  int	 i;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("traing.nAddCentroid\n");
#endif

  /* compute the centroid */
  x = y = 0.0;
  pvtx = avtx + ppoly->ivtxHead;
  for(i = 0; i < cvtx; ++i)
  { x += anode[pvtx->inode].x;
    y += anode[pvtx->inode].y;
    pvtx = avtx + pvtx->ivtxR;
  }
  x /= (double) cvtx;
  y /= (double) cvtx;

  /* add centroid to node list */
  inode = inodeAddNode(pmesh,x,y,TRUE);
  anode = pmesh->anode;

  /* add new edges to edge list and make the list circular */
  for(i = 0, piedge = aiedge; i < cvtx; ++i, ++piedge)
  { *piedge = iedgeAddEdge(pmesh,pvtx->inode,inode);
    pvtx = avtx + pvtx->ivtxR;
  }
  *piedge = *aiedge;

  /* triangulate the polygon */
  for(i = 0, piedge = aiedge; i < cvtx; ++i, ++piedge)
  { pvtxR = avtx + pvtx->ivtxR;
    itri = itriAddTriangle(pmesh,pvtx->iedge,piedge[0],piedge[1],-1u);
    pmesh->atri[itri].wLabel = pmesh->ppoly->wLabel;
    if (cmdargs.wVisual == VISUAL_CONSTRUCT)
      nDrawTriangle(anode,pvtx->inode,pvtxR->inode,inode);  /* @@@ */
    pvtx = pvtxR;
  }

  /* remove the polygon's vertices from the vertex list and pop the */
  /* polygon from the polygon stack */
  for(i = 0; i < cvtx; ++i) nDeleteVertex(pmesh, ppoly->ivtxHead);
  cPopPolygon(pmesh);

#if 0
  if ((cmdargs.wVisual == VISUAL_CONSTRUCT) && (cmdargs.cNoPress != -1))
  { if (cmdargs.cNoPress) --cmdargs.cNoPress;
    else { if (getch() == 'q') exit(20); }
  }
#endif
  return(0);

} /* nAddCentroid */


/* ----------------------------------------------------------------------------
 * cVisibleVertices:  This function determines the vertices that are visible
 * from ivtx not including its immediate neighbors.  The length of the list
 * is returned.
 */
#if defined(ANSIPROTO)
WORD
cVisibleVertices(
PMESH pmesh,				/* mesh structure */
IVTX  ivtx)				/* vertex index */
#else
WORD
cVisibleVertices(pmesh,ivtx)
PMESH pmesh;
IVTX  ivtx;
#endif
{
  PNODE anode = pmesh->anode;
  PVTX	avtx  = pmesh->avtx;
  PVTX	pvtx  = avtx + ivtx;
  IVTX	ivtxR = pvtx->ivtxR;

  REAL	AngleVBA;			/* angle */
  REAL	AngleCBA;			/* angle */
  REAL	AngleABW;			/* angle */
  REAL	QualABW;			/* triangle quality factor */
  IVTX *pivtx;				/* walking pointer */
  IVTX	ilist;				/* list index */
  IVTX	ivtxA,ivtxB,ivtxC,ivtxW;	/* vertex indices */
  BOOL	fDone;				/* done flag */
  int	i;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("triang.cVisibleVertices\n");
#endif

  /* add the right neighbor to the list */
  ilist = 0;
  *aivtxList = ivtxW = ivtxR;

  for(fDone = FALSE; !fDone;)
  { /* walk around the polygon and visit each vertex ... assume for now the */
    /* current vertex is visible and add to the list */
    ivtxW = avtx[ivtxW].ivtxR;
    if (ivtxW == ivtx) break;
    aivtxList[++ilist] = ivtxW;

    ivtxA = ivtxW;
    ivtxB = aivtxList[ilist-1];
    AngleVBA = CompAng(ivtx,ivtxB,ivtxA);

    /* if the polygon bends towards ivtx then add another vertex to the list */
    if ((AngleVBA > EPS) && (AngleVBA < PI - EPS)) continue;

    /* check which way the boundary bends */
    AngleCBA = 2*PI;
    if (ilist > 1)
    { ivtxC = aivtxList[ilist-2];
      AngleCBA = CompAng(ivtxC,ivtxB,ivtxA);
    }

    /* boundary bends away from ivtx */
    if (AngleCBA > PI)
    { for(;;)
      { ivtxW = avtx[ivtxW].ivtxR;
	if (ivtxW == ivtx) { fDone = TRUE; break; }
	aivtxList[ilist] = ivtxW;
	AngleABW = CompAng(ivtxA,ivtxB,ivtxW);
	QualABW	 = CompQual(ivtxA,ivtxB,ivtxW);
	if	(QualABW >  EPS) AngleVBA += AngleABW;
	else if (QualABW < -EPS) AngleVBA += AngleABW - 2*PI;
	if ((AngleVBA > EPS) && (AngleVBA < PI - EPS)) break;
	ivtxA = ivtxW;
      }
    }

    /* boundary bends towards ivtx */
    else
    { do
      { aivtxList[--ilist] = ivtxA;
	if (ilist < 1) break;
	ivtxB = aivtxList[ilist-1];
	AngleVBA = CompAng(ivtx,ivtxB,ivtxA);
      } while ((AngleVBA <= EPS) || (AngleVBA >= PI - EPS));
    }
  }
  --ilist;
  if (ilist < 1) return(0);

  for(i = 0, pivtx = aivtxList; i < ilist; ++i, ++pivtx) *pivtx = *(pivtx+1);
  return(ilist);

} /* cVisibleVertices */


/* ----------------------------------------------------------------------------
 * nSplitOK:  This function determines if it possible to split the region into
 * two parts through vertex ivtx.
 */
#if defined(ANSIPROTO)
int
nSplitOK(
PMESH  pmesh,				/* polygonal region to triangulate */
IVTX   ivtx,				/* vertex from which to split */
IVTX  *ivtxJoin)			/* split from ivtx to ivtxJoin */
#else
int
nSplitOK(pmesh,ivtx,ivtxJoin)
PMESH  pmesh;
IVTX   ivtx;
IVTX  *ivtxJoin;
#endif
{
  PNODE anode  = pmesh->anode;
  PVTX	avtx   = pmesh->avtx;
  PVTX	pvtx   = avtx + ivtx;
  IVTX	ivtxL  = pvtx->ivtxL;
  IVTX	ivtxR  = pvtx->ivtxR;

  REAL	dT;				/* buffer */
  REAL	Rating;				/* feasibility of a chop */
  REAL	RatingT;			/* feasibility of this chop */
  REAL	Angle;				/* angle */
  REAL	Angle1;				/* angle */
  REAL	Angle2;				/* angle */
  REAL	Angle3;				/* angle */
  REAL	Angle4;				/* angle */
  REAL	Dist;				/* distance */
  REAL	Dist1;				/* distance */
  REAL	Dist2;				/* distance */
  REAL	Dist3;				/* distance */
  REAL	Dist4;				/* distance */
  REAL	DistVL;				/* distance */
  REAL	DistVR;				/* distance */
  REAL	DistVM;				/* distance */
  REAL	Amul;				/* length of intersection along dA */
  REAL	Bmul;				/* length of intersection along dB */
  REAL	Hinit;				/* initial node spacing */
  REAL	Hgrade;				/* grade of node spacing */
  CRDNT crdnt;				/* coordinate of vertex ivtx */
  CRDNT crdntA;				/* coordinates of point A */
  CRDNT crdntdA;			/* coordinates of vector dA */
  CRDNT crdntB;				/* coordinates of point B */
  CRDNT crdntdB;			/* coordinates of vector dB */
  BOOL	fGood;				/* good place to split flag */
  WORD	civtxList;			/* number of vertices in list */
  WORD	cNP;				/* number of new points */
  IVTX	ilist;				/* list indices */
  IVTX	ivtxA,ivtxB,ivtxM;		/* vertex indices */
  IVTX	ivtxT,ivtxML,ivtxMR;		/* vertex indices */
  INODE inodeML,inodeMR,inodeM;		/* node indices */
  int	i;				/* loop index */


#if defined(DEBUG_FUNC)
DEBUG_INFO("triang.nSplitOK\n");
#endif

  crdnt.x = anode[pvtx->inode].x;
  crdnt.y = anode[pvtx->inode].y;

  /* determine the vertices visible from ivtx */
  civtxList = cVisibleVertices(pmesh,ivtx);

  /* delete points distant from kv or vornoi points whose edge */
  /* is outside the region */
  if (civtxList > 0)
  {
    ilist=0;
    for(i = 1; i < civtxList; ++i)
    { aivtxList[++ilist] = aivtxList[i];

      for(;;)
      { ivtxB = aivtxList[ilist-1];
	ivtxA = aivtxList[ilist];
	crdntA.x  = (crdnt.x + anode[avtx[ivtxA].inode].x)/2.0e0;
	crdntA.y  = (crdnt.y + anode[avtx[ivtxA].inode].y)/2.0e0;
	crdntdA.x = crdnt.y - anode[avtx[ivtxA].inode].y;
	crdntdA.y = anode[avtx[ivtxA].inode].x - crdnt.x;
	crdntB.x  = (crdnt.x + anode[avtx[ivtxB].inode].x)/2.0e0;
	crdntB.y  = (crdnt.y + anode[avtx[ivtxB].inode].y)/2.0e0;
	crdntdB.x = crdnt.y - anode[avtx[ivtxB].inode].y;
	crdntdB.y = anode[avtx[ivtxB].inode].x - crdnt.x;
	nIntersection(&crdntA,&crdntdA,&crdntB,&crdntdB,&Amul,&Bmul);
	if(Amul * Bmul <= 0.0e0) break;
	ivtxT = ivtxR;
	if(Amul > 0.0) ivtxT = ivtxL;
	crdntA.x += Amul * crdntdA.x;
	crdntA.y += Amul * crdntdA.y;
	crdntB.x = crdnt.x;
	crdntB.y = crdnt.y;
	crdntdB.x = anode[avtx[ivtxT].inode].x - crdnt.x;
	crdntdB.y = anode[avtx[ivtxT].inode].y - crdnt.y;
	crdntdA.x = -crdntdB.y;
	crdntdA.y =  crdntdB.x;
	nIntersection(&crdntA,&crdntdA,&crdntB,&crdntdB,&Amul,&Bmul);
	if ((Amul >= 0.0) && (ivtxT == ivtxR))
	{ aivtxList[--ilist] = ivtxA;
	  if  (ilist < 1) break;
	}
	if ((Amul <= 0.0) && (ivtxT == ivtxL)) --ilist;
	break;
      }
    }
    civtxList = ilist + 1;
  }

  /* of the remaining points keep only vornoi points */
  if (civtxList > 1)
  {
    ilist=1;
    for(i = 2; i < civtxList; ++i)
    { aivtxList[++ilist] = aivtxList[i];
      do
      { ivtxB = aivtxList[ilist-2];
	ivtxM = aivtxList[ilist-1];
	ivtxA = aivtxList[ilist];
	crdntA.x  = (crdnt.x + anode[avtx[ivtxM].inode].x)/2.0e0;
	crdntA.y  = (crdnt.y + anode[avtx[ivtxM].inode].y)/2.0e0;
	crdntdA.x = crdnt.y - anode[avtx[ivtxM].inode].y;
	crdntdA.y = anode[avtx[ivtxM].inode].x - crdnt.x;
	crdntB.x  = (crdnt.x + anode[avtx[ivtxA].inode].x)/2.0e0;
	crdntB.y  = (crdnt.y + anode[avtx[ivtxA].inode].y)/2.0e0;
	crdntdB.x = crdnt.y - anode[avtx[ivtxA].inode].y;
	crdntdB.y = anode[avtx[ivtxA].inode].x - crdnt.x;
	nIntersection(&crdntA,&crdntdA,&crdntB,&crdntdB,&Amul,&Bmul);
	dT = Amul;
	crdntB.x  = (crdnt.x + anode[avtx[ivtxB].inode].x)/2.0e0;
	crdntB.y  = (crdnt.y + anode[avtx[ivtxB].inode].y)/2.0e0;
	crdntdB.x = crdnt.y - anode[avtx[ivtxB].inode].y;
	crdntdB.y = anode[avtx[ivtxB].inode].x - crdnt.x;
	nIntersection(&crdntA,&crdntdA,&crdntB,&crdntdB,&Amul,&Bmul);
	if (dT > Amul) break;
	aivtxList[--ilist] = ivtxA;
      } while (ilist > 1);
    }
    civtxList = ilist + 1;
  }

  /* determine the best point ... */
  DistVR = sqrt(sq(anode[avtx[ivtxR].inode].x - crdnt.x) +
		sq(anode[avtx[ivtxR].inode].y - crdnt.y));
  DistVL = sqrt(sq(anode[avtx[ivtxL].inode].x - crdnt.x) +
		sq(anode[avtx[ivtxL].inode].y - crdnt.y));

  *ivtxJoin = ivtxT  = (WORD) -1;
  Rating = 0.0;
  fGood	 = FALSE;
  for(i = 0; i < civtxList; ++i)
  {
    ivtxM  = aivtxList[i];
    ivtxML = avtx[ivtxM].ivtxL;
    ivtxMR = avtx[ivtxM].ivtxR;

    /* check the four angles generated by splitting the polygon */
    /* at vertices ivtx and ivtxM */
    Angle1  = CompAng(ivtxM,ivtx,ivtxR);
    Angle2  = pvtx->Angle - Angle1;
    Angle3  = CompAng(ivtxML,ivtxM,ivtx);
    Angle4  = avtx[ivtxM].Angle - Angle3;
    Angle   = min(Angle1,min(Angle2,min(Angle3,Angle4)));
    Angle  /= PI / 3.0;
    if (Angle > 1.0) Angle = 1.0 / Angle;
    RatingT = Angle;
/*  RatingT = min(1.0,Angle/param.MinAngle);  */
    if (RatingT <= Rating) continue;


    /* check side lengths for the first and last segments on the */
    /* proposed split */
    DistVM = sqrt(sq(anode[avtx[ivtxM].inode].x - crdnt.x) +
		  sq(anode[avtx[ivtxM].inode].y - crdnt.y));
    cNP = cNumNewPoints(pvtx->Hlocal,avtx[ivtxM].Hlocal,DistVM,0,&Hinit,&Hgrade);
    dT	= Hinit * DistVM;
    Dist1 = DistVR / dT;
    Dist2 = DistVL / dT;
    dT *= pow(Hgrade,(REAL) cNP);
    inodeM  = avtx[ivtxM].inode;
    inodeMR = avtx[ivtxMR].inode;
    inodeML = avtx[ivtxML].inode;
    Dist3 = sqrt(sq(anode[inodeM].x - anode[inodeMR].x) +
		 sq(anode[inodeM].y - anode[inodeMR].y)) / dT;
    Dist4 = sqrt(sq(anode[inodeM].x - anode[inodeML].x) +
		 sq(anode[inodeM].y - anode[inodeML].y)) / dT;
    if (Dist1 < 1.0) Dist1 = 1.0 / Dist1;
    if (Dist2 < 1.0) Dist2 = 1.0 / Dist2;
    if (Dist3 < 1.0) Dist3 = 1.0 / Dist3;
    if (Dist4 < 1.0) Dist4 = 1.0 / Dist4;
    Dist = max(Dist1,max(Dist2,max(Dist3,Dist4)));
    RatingT = min(RatingT,param.Grade/Dist);
    if (RatingT <= Rating) continue;

    Rating = RatingT;
    ivtxT  = ivtxM;
    if (Rating == 1.0) { fGood = TRUE; break; }
  }

  if (!fGood && ((ivtxT == (WORD) -1) || (Rating < param.Best)))
    return(STATUS_NO);

  *ivtxJoin = ivtxT;
  param.Best = Rating;
  return((Rating < 1.0) ? STATUS_OK : STATUS_YES);

} /* nSplitOK */


/* ----------------------------------------------------------------------------
 * nSplitRegion:  This function splits the polygon at the specified indices
 * and adds new nodes along the dividing line.
 */
#if defined(ANSIPROTO)
int
nSplitRegion(
PMESH pmesh,				/* polygon */
IVTX  ivtxA,				/* index of vertex to chop */
IVTX  ivtxB)				/* index of vertex to chop */
#else
int
nSplitRegion(pmesh,ivtxA,ivtxB);
PMESH pmesh;
IVTX  ivtxA;
IVTX  ivtxB;
#endif
{
  PPOLY ppoly  = pmesh->ppoly;
  PNODE anode  = pmesh->anode;
  PVTX	avtx   = pmesh->avtx;
  PVTX	pvtxA  = avtx + ivtxA;
  PVTX	pvtxB  = avtx + ivtxB;
  REAL	Grade  = param.Grade;
  REAL	Hmax   = (ppoly->Hmax > 0.0) ? ppoly->Hmax : param.Hmax;
  WORD	cvtx   = ppoly->cvtx;
  IVTX	ivtxAL = pvtxA->ivtxL;
  IVTX	ivtxAR = pvtxA->ivtxR;
  IVTX	ivtxBL = pvtxB->ivtxL;
  IVTX	ivtxBR = pvtxB->ivtxR;

  POLY	poly;
  CRDNT crdntM;
  CRDNT crdntP,crdntdP;
  CRDNT crdntQ,crdntdQ;
  INODE inode;				/* node index */
  IEDGE iedge;				/* edge index */
  REAL	Pscale,PscaleT;
  REAL	Qscale,QscaleT;
  REAL	Z;
  REAL	dT;				/* buffer */
  REAL	H;				/* local node spacing */
  REAL	dH;				/* change in node spacing */
  REAL	Theta;				/* ??? */
  REAL	FA,FB;
  REAL	Hmin;				/* ??? */
  REAL	DistAB;				/* distance between nodes A and B */
  REAL	DistAM;
  REAL	DistBM;
  REAL	HA;				/* local spacing of node A */
  REAL	HB;				/* local spacing of node B */
  REAL	HM;
  REAL	xA,yA,xB,yB;			/* coordinates of nodes A & B */
  REAL	dx,dy;				/* vector BA */
  REAL	Hinit;				/* initial node spacing */
  REAL	Hgrade;				/* node spacing grade */
  BOOL	fGrade;				/* set to grade new node spacing */
  IVTX	ivtxA1,ivtxB1;			/* vertex indices */
  IVTX	ivtxA2,ivtxB2;			/* vertex indices */
  IVTX	ivtxH1,ivtxH2;			/* vertex indices */
  IVTX	ivtxT;				/* vertex index */
  WORD	cNP;				/* number of points to add to edge */
  WORD	cvtx1;				/* number of vertices */
  IVTX	ivtxMin1;			/* index of vertex w/ min angle */
  IVTX	ivtxMax1;			/* index of vertex w/ max angle */
  int	i;				/* loop index */

#if defined(DEBUG_FUNC)
DEBUG_INFO("triang.nSplitRegion\n");
#endif

  /* compute the distance between nodes A and B and compute the local */
  /* spacing of nodes A and B */
  xA = anode[pvtxA->inode].x;
  yA = anode[pvtxA->inode].y;
  xB = anode[pvtxB->inode].x;
  yB = anode[pvtxB->inode].y;
  dx = xB - xA;
  dy = yB - yA;
  DistAB = sqrt(dx*dx + dy*dy);
  HA = pvtxA->Hlocal;
  HB = pvtxB->Hlocal;

  /* determine if the node spacing should increase towards the middle */
  /* of the edge */
  fGrade = cmdargs.fGraded;
  if ((cvtx < 9) || (DistAB <= Grade * (HA + HB))) fGrade = FALSE;
  if (fGrade)
  { dT = DistAB * (Grade - 1.0) / Grade;
    Theta = (HA - HB) / (2.0 * dT);
    if (fabs(Theta) > 0.4) fGrade = FALSE;
  }
  if (fGrade)
  { FA = 0.5 - Theta;
    FB = 0.5 + Theta;
    HM = (dT + HA + HB) / 2.0;
    Hmin = FA * HB + FB * HA;
    if (Hmax > 0.0) HM = min(HM,Hmax);
    if (HM < Hmin) fGrade = FALSE;
  }

  /* if node spacing should be graded, compute grade information */
  if (fGrade)
  { crdntM.x = FA * xB + FB * xA;
    crdntM.y = FA * yB + FB * yA;
    crdntP.x  = crdntM.x;
    crdntP.y  = crdntM.y;
    crdntdP.x = dx;
    crdntdP.y = dy;
    crdntdQ.x = -dy;
    crdntdQ.y = dx;
    ivtxT = ivtxB;
    for(i = 0; i < cvtx; ++i)
    { ivtxT = avtx[ivtxT].ivtxR;
      if ((ivtxT == ivtxAL) || (ivtxT == ivtxA) || (ivtxT == ivtxAR) ||
	  (ivtxT == ivtxBL) || (ivtxT == ivtxB) || (ivtxT == ivtxBR))
	continue;
      crdntQ.x = anode[avtx[ivtxT].inode].x;
      crdntQ.y = anode[avtx[ivtxT].inode].y;
      nIntersection(&crdntP,&crdntdP,&crdntQ,&crdntdQ,&Pscale,&Qscale);
      QscaleT = fabs(Qscale) * dT;
      Z = QscaleT + avtx[ivtxT].Hlocal / Grade;
      if (Pscale < 0.0)
      { PscaleT = -Pscale / FA;
	if ((PscaleT >= 1.0) || (QscaleT > HM * (1.0 - PscaleT))) continue;
	H = HM + (HA - HM) * PscaleT;
	if (H <= Z) continue;
	H = (Z - PscaleT * HA) / (1.0 - PscaleT);
      }
      else
      { PscaleT = Pscale / FB;
	if ((PscaleT >= 1.0) || (QscaleT > HM * (1.0 - PscaleT))) continue;
	H = HM + (HB - HM) * PscaleT;
	if (H <= Z) continue;
	H = (Z - PscaleT * HB) / (1.0 - PscaleT);
      }
      HM = min(HM,H);
      if (HM < Hmin) { fGrade = FALSE; break; }
    }
  }

  /* split the polygon into two regions and update the angles */
  nSplitPolygon(pmesh,ivtxA,ivtxB);
  avtx	 = pmesh->avtx;
  ivtxA1 = ivtxA;
  ivtxB1 = avtx[ivtxA].ivtxL;
  ivtxA2 = avtx[ivtxB].ivtxL;
  ivtxB2 = ivtxB;
  avtx[ivtxA1].Angle = CompAng(avtx[ivtxA1].ivtxL,ivtxA1,avtx[ivtxA1].ivtxR);
  avtx[ivtxA2].Angle = CompAng(avtx[ivtxA2].ivtxL,ivtxA2,avtx[ivtxA2].ivtxR);
  avtx[ivtxB1].Angle = CompAng(avtx[ivtxB1].ivtxL,ivtxB1,avtx[ivtxB1].ivtxR);
  avtx[ivtxB2].Angle = CompAng(avtx[ivtxB2].ivtxL,ivtxB2,avtx[ivtxB2].ivtxR);
  ivtxH1 = ivtxA1;
  ivtxH2 = ivtxB2;

  /* new node spacing should be graded */
  if (fGrade)
  { /* set up the first segment */
    DistAM = FA * DistAB;
    if ((cNP = cNumNewPoints(HA,HM,DistAM,0,&Hinit,&Hgrade)) > 0)
    { H	 = 0.0;
      dH = Hinit;
      dx = crdntM.x - xA;
      dy = crdntM.y - yA;
      for(i = 0; i < cNP; ++i)
      { H    += dH;
	dH   *= Hgrade;
	inode = inodeAddNode(pmesh,xA+H*dx,yA+H*dy,TRUE);
	anode = pmesh->anode;
	if (cmdargs.wVisual == VISUAL_CONSTRUCT) nPlotPoint(anode,inode,fGrade);

	/* @@@ link into the two polygon */
	ppoly->ivtxHead = ivtxH1;
	ivtxH1 = ivtxT = ivtxAddVertex(pmesh,inode,(IEDGE) -1);
	pmesh->avtx[ivtxT].Angle = PI;
	ppoly->ivtxHead = ivtxH2;
	ivtxT = ivtxAddVertex(pmesh,inode,(IEDGE) -1);
	pmesh->avtx[ivtxT].Angle = PI;
      }
    }

    /* set up the second segment */
    DistBM = FB * DistAB;
    if ((cNP = cNumNewPoints(HM,HB,DistBM,0,&Hinit,&Hgrade)) > 0)
    { H	 = 0.0;
      dH = Hinit;
      dx = xB - crdntM.x;
      dy = yB - crdntM.y;
      for(i = 0; i < cNP; ++i)
      { H    += dH;
	dH   *= Hgrade;
	inode = inodeAddNode(pmesh,crdntM.x+H*dx,crdntM.y+H*dy,TRUE);
	anode = pmesh->anode;
	if (cmdargs.wVisual == VISUAL_CONSTRUCT) nPlotPoint(anode,inode,fGrade);

	/* @@@ link into two polygon */
	ppoly->ivtxHead = ivtxH1;
	ivtxH1 = ivtxT = ivtxAddVertex(pmesh,inode,(IEDGE) -1);;
	pmesh->avtx[ivtxT].Angle = PI;
	ppoly->ivtxHead = ivtxH2;
	ivtxT = ivtxAddVertex(pmesh,inode,(IEDGE) -1);;
	pmesh->avtx[ivtxT].Angle = PI;
      }
    }
  }

  /* new node spacing should not be graded */
  else
  { if ((cNP = cNumNewPoints(HA,HB,DistAB,0,&Hinit,&Hgrade)) > 0)
    { H	 = 0.0;
      dH = Hinit;
      for(i = 0; i < cNP; ++i)
      { H += dH;
	dH *= Hgrade;
	inode = inodeAddNode(pmesh,xA+H*dx,yA+H*dy,TRUE);
	anode = pmesh->anode;
	if (cmdargs.wVisual == VISUAL_CONSTRUCT) nPlotPoint(anode,inode,fGrade);

	/* @@@ link into two polygon */
	ppoly->ivtxHead = ivtxH1;
	ivtxH1 = ivtxT = ivtxAddVertex(pmesh,inode,(IEDGE) -1);
	pmesh->avtx[ivtxT].Angle = PI;
	ppoly->ivtxHead = ivtxH2;
	ivtxT = ivtxAddVertex(pmesh,inode,(IEDGE) -1);
	pmesh->avtx[ivtxT].Angle = PI;
      }
    }
  }

  avtx = pmesh->avtx;
  avtx[ivtxA1].Hlocal = CompHloc(avtx[ivtxA1].ivtxL,ivtxA1,avtx[ivtxA1].ivtxR);
  avtx[ivtxA2].Hlocal = CompHloc(avtx[ivtxA2].ivtxL,ivtxA2,avtx[ivtxA2].ivtxR);
  avtx[ivtxB1].Hlocal = CompHloc(avtx[ivtxB1].ivtxL,ivtxB1,avtx[ivtxB1].ivtxR);
  avtx[ivtxB2].Hlocal = CompHloc(avtx[ivtxB2].ivtxL,ivtxB2,avtx[ivtxB2].ivtxR);
  /* @@@ this can be code more efficiently */
  for(ivtxH1 = avtx[ivtxA1].ivtxL, ivtxH2 = avtx[ivtxA2].ivtxR;
      ivtxH1 != ivtxB1;
      ivtxH1 = avtx[ivtxH1].ivtxL, ivtxH2 = avtx[ivtxH2].ivtxR)
      avtx[ivtxH1].Hlocal = avtx[ivtxH2].Hlocal =
    CompHloc(avtx[ivtxH1].ivtxL,ivtxH1,avtx[ivtxH1].ivtxR);

  /* add edges */
  for(ivtxH1 = ivtxB1, ivtxH2 = avtx[ivtxB2].ivtxL;
      ivtxH1 != ivtxA1;
      ivtxH1 = avtx[ivtxH1].ivtxR, ivtxH2 = avtx[ivtxH2].ivtxL)
  { iedge = iedgeAddEdge(pmesh,avtx[ivtxH1].inode,avtx[ivtxH2].inode);
    avtx[ivtxH1].iedge = avtx[ivtxH2].iedge = iedge;
  }

  ppoly->ivtxHead = ivtxA1;
  nMinMaxAngles(pmesh);
  cvtx1 = ppoly->cvtx;
  ivtxMin1 = ppoly->ivtxMinAng;
  ivtxMax1 = ppoly->ivtxMaxAng;
  ppoly->ivtxHead = ivtxA2;
  nMinMaxAngles(pmesh);
  if (ppoly->cvtx >= cvtx1)
  { poly.cvtx	    = cvtx1;
    poly.ivtxMinAng = ivtxMin1;
    poly.ivtxMaxAng = ivtxMax1;
    poly.ivtxHead   = ivtxA1;
  }
  else
  { poly.cvtx	      = ppoly->cvtx;
    poly.ivtxMinAng   = ppoly->ivtxMinAng;
    poly.ivtxMaxAng   = ppoly->ivtxMaxAng;
    poly.ivtxHead     = ivtxA2;
    ppoly->cvtx	      = cvtx1;
    ppoly->ivtxMinAng = ivtxMin1;
    ppoly->ivtxMaxAng = ivtxMax1;
    ppoly->ivtxHead   = ivtxA1;
  }
  poly.wLabel  = ppoly->wLabel;
  poly.iRegion = ppoly->iRegion;
  poly.fRect   = FALSE;
  poly.Hmax    = ppoly->Hmax;
  cPushPolygon(pmesh,&poly);

#if 0
  if ((cmdargs.wVisual == VISUAL_CONSTRUCT) && (cmdargs.cNoPress != -1))
  { if (cmdargs.cNoPress) --cmdargs.cNoPress;
    else { if (getch() == 'q') exit(20); }
  }
#endif
  return(0);

} /* nSplitRegion */


/* ----------------------------------------------------------------------------
 * nTriangPolygon:  This function triangulates the region enclosed by the
 * polygon which is on the top of the polygon stack.  If the region is split
 * into two regions then STATUS_SPLIT is returned otherwise STATUS_DONE is
 * returned.
 */
#if defined(ANSIPROTO)
int
nTriangPolygon(
PMESH pmesh)				/* mesh structure */
#else
int
nTriangPolygon(pmesh)
PMESH pmesh;
#endif
{
  PPOLY ppoly;				/* TOS polygon */
  PVTX	avtx;				/* array of vertices */
  WORD	cvtx;				/* number of vertices */
  WORD	ivtxAccept1;			/* index of an acceptable vertex */
  WORD	ivtxAccept2;			/* index of an acceptable vertex */
  WORD	ivtxAccept3;			/* index of an acceptable vertex */
  WORD	ivtx,ivtxT;			/* vertex indices */
  REAL	Qual;				/* triangle quality */
  int	nChop;				/* chopping result */
  int	nSplit;				/* splitting result */
  int	nCentroid;			/* convexing result */
  int	nResult;			/* result */
  int	i;				/* loop index */
  PVTX	pvtx;
  INODE inodeL,inodeM,inodeR;

#if defined(DEBUG_FUNC)
DEBUG_INFO("triang.nTriangPolygon\n");
#endif

  /* TRIANGULATION ALGORITHM: */
  /* if cvtx == 3 then chop triangle and return */
  /* if smallest angle <= MaxAngle then check for a chop */
  /* if convex polygon w/ cvtx <= 8 then check for convex triangulation */
  /* check if polygon can be divided into two polygons */
  /* make the best of a bad situation */

  for(;;)
  { ppoly = pmesh->ppoly;
    avtx  = pmesh->avtx;
    ivtx  = ppoly->ivtxMinAng;
    cvtx  = ppoly->cvtx;

    if (cvtx <	3)   return(STATUS_DONE);
    if (cvtx == 3)
    { pvtx = avtx + ivtx;
      inodeM = pvtx->inode;
      inodeL = avtx[pvtx->ivtxL].inode;
      inodeR = avtx[pvtx->ivtxR].inode;
      Qual = CompQualityI(pmesh->anode,inodeL,inodeM,inodeR);
      if (Qual < cmdargs.Qmin) break;
      nChopVertex(pmesh,ivtx);
      return(STATUS_DONE);
    }

    nChop	= STATUS_NO;
    nCentroid	= STATUS_NO;
//  nSplit	= STATUS_NO;

    param.Best	= 0.0;
    ivtxAccept1 = (WORD) -1;
    if (avtx[ivtx].Angle <= param.MaxAngle)
    { for(i = 0; i < cvtx; ++i)
      { nResult = nChopOK(pmesh,ivtx);
	if (nResult != STATUS_NO)
	{ nChop	      = nResult;
	  ivtxAccept1 = ivtx;
	  if (!cmdargs.fExhaustive && (nResult == STATUS_YES)) break;
	}
	ivtx = avtx[ivtx].ivtxR;
      }
    }

    if (!cmdargs.fExhaustive && (param.Best >= param.Acceptable))
    { nChopVertex(pmesh,ivtxAccept1);
      continue;
    }

    if (!cmdargs.fNoCentroid && (cvtx < 9))
    { nCentroid = nCentroidOK(pmesh);
      if (!cmdargs.fExhaustive && (param.Best >= param.Acceptable))
      { nAddCentroid(pmesh);
	return(STATUS_DONE);
      }
    }

    nSplit = STATUS_NO;
    ivtxAccept2 = ivtxAccept3 = (WORD) -1;
    ivtx = ppoly->ivtxMaxAng;
    for(i = 0; i < cvtx; ++i)
    { nResult = nSplitOK(pmesh,ivtx,&ivtxT);
      if (nResult != STATUS_NO)
      { nSplit	    = nResult;
	ivtxAccept2 = ivtxT;
	ivtxAccept3 = ivtx;
	if (!cmdargs.fExhaustive && (nResult == STATUS_YES)) break;
      }
      ivtx = avtx[ivtx].ivtxR;
    }

    /* this triangulation really stinks, fail ... */
    if (param.Best < cmdargs.Qmin) break;

    if (nSplit != STATUS_NO)
    { nSplitRegion(pmesh,ivtxAccept2,ivtxAccept3);
      return(STATUS_SPLIT);
    }

    if (nCentroid != STATUS_NO)
    { nAddCentroid(pmesh);
      return(STATUS_DONE);
    }

    if (nChop != STATUS_NO)
    { nChopVertex(pmesh,ivtxAccept1);
      continue;
    }

    break;
  }

#if 0
  if (cmdargs.wVisual == VISUAL_CONSTRUCT) { getch(); nCloseWindow(); }
#endif
  vErrorHandler(ERROR_TRIANGFAIL, NULL, NULL, TRUE);
  return(STATUS_DONE);

} /* nTriangPolygon */


/* ----------------------------------------------------------------------------
 * nRectangPolygon:  This function rectangulates the region enclosed by the
 * polygon which is on the top of the polygon stack.
 */
#if defined(ANSIPROTO)
int
nRectangPolygon(
PMESH pmesh)				/* mesh structure */
#else
int
nRectangPolygon(pmesh)
PMESH pmesh;
#endif
{
  PPOLY ppoly;				/* TOS polygon */
  PNODE anode;				/* array of nodes */
  PVTX	avtx;				/* array of vertices */
  WORD	cvtx;				/* number of vertices */
  IEDGE iedgeL;				/* edge index */
  IEDGE iedgeR;				/* edge index */
  INODE inodeM;				/* node index */
  INODE inodeL;				/* node index */
  INODE inodeR;				/* node index */
  INODE inodeT;				/* node index */
  INODE inodeLL;			/* node index */
  INODE inodeRR;			/* node index */
  PNODE pnodeM;				/* node pointer */
  PNODE pnodeL;				/* node pointer */
  PNODE pnodeR;				/* node pointer */
//PNODE pnodeT;				/* node pointer */
  NODE	nodeT;				/* node */
  ITRI	itri;				/* triangle index */
  IVTX	ivtxM;				/* vertex index */
  IVTX	ivtxL;				/* vertex index */
  IVTX	ivtxR;				/* vertex index */
  IVTX	ivtxT;				/* vertex index */
  IVTX	ivtxLL;				/* vertex index */
  IVTX	ivtxRR;				/* vertex index */
  PVTX	pvtxM;				/* vertex pointer */
  PVTX	pvtxL;				/* vertex pointer */
  PVTX	pvtxR;				/* vertex pointer */
  PVTX	pvtxT;				/* vertex pointer */
  WORD	cQuadL;				/* left vertex angle / (PI/2) */
  WORD	cQuadR;				/* right vertex angle / (PI/2) */
  REAL	dxL,dyL;			/* lengths */
  REAL	dxR,dyR;			/* lengths */

#if defined(DEBUG_FUNC)
DEBUG_INFO("triang.nRectangPolygon\n");
#endif

  /* RECTANGULATION ALGORITHM: */
  /* search for a vertex that has an angle of PI/2 and chop a rectangle */
  /* since the angles can only be 3*PI/2, PI and PI/2 the minimum angle */
  /* will allows be PI/2 */

  for(;;)
  { ppoly  = pmesh->ppoly;
    anode  = pmesh->anode;
    avtx   = pmesh->avtx;
    cvtx   = ppoly->cvtx;
    ivtxM  = ppoly->ivtxMinAng;
    pvtxM  = avtx + ivtxM;
    ivtxL  = pvtxM->ivtxL;
    ivtxR  = pvtxM->ivtxR;
    pvtxL  = avtx + ivtxL;
    pvtxR  = avtx + ivtxR;
    inodeM = pvtxM->inode;
    inodeL = pvtxL->inode;
    inodeR = pvtxR->inode;
    cQuadL = (WORD) ((pvtxL->Angle + 0.5) / (PI / 2.0));
    cQuadR = (WORD) ((pvtxR->Angle + 0.5) / (PI / 2.0));


    /*	  END OF RECTANGULATION			       */
    /*						       */
    /*	  x----o				       */
    /*	  |    |				       */
    /*	  |    |				       */
    /*	  o----o				       */

    if (cvtx == 4)
    {
      ivtxT  = pvtxR->ivtxR;
      pvtxT  = avtx + ivtxT;
//    inodeT = pvtxT->inode;
      itri  = itriAddTriangle(pmesh,pvtxL->iedge,pvtxM->iedge,
				    pvtxR->iedge,pvtxT->iedge);
      pmesh->atri[itri].wLabel = pmesh->ppoly->wLabel;
      nDeleteVertex(pmesh,ivtxL);
      nDeleteVertex(pmesh,ivtxM);
      nDeleteVertex(pmesh,ivtxR);
      nDeleteVertex(pmesh,ivtxT);
      cPopPolygon(pmesh);
    }


    /*	  DO NOT NEED TO ADD A NODE		       */
    /*						       */
    /*	  x----o----o----o	  x----o----o----o     */
    /*	  |		 |	  |    |	 |     */
    /*	  |		 |   TO	  |    |	 |     */
    /*	  o----o----o----o	  o----o----o----o     */


    else if (cQuadR == 1)
    {
      ivtxT  = pvtxR->ivtxR;
      pvtxT  = avtx + ivtxT;
      inodeT = pvtxT->inode;
      iedgeR = iedgeAddEdge(pmesh,inodeL,inodeT);
      itri  = itriAddTriangle(pmesh,pvtxL->iedge,pvtxM->iedge,
				    pvtxR->iedge,iedgeR);
      pmesh->atri[itri].wLabel = pmesh->ppoly->wLabel;
      nDeleteVertex(pmesh,ivtxM);
      nDeleteVertex(pmesh,ivtxR);
      pvtxL->iedge = iedgeR;
      ivtxLL  = avtx[ivtxL].ivtxL;
      ivtxRR  = avtx[ivtxT].ivtxR;
      inodeLL = avtx[ivtxLL].inode;
      inodeRR = avtx[ivtxRR].inode;
      pvtxL->Hlocal = CompHlocalI(anode,inodeLL,inodeL,inodeT);
      pvtxT->Hlocal = CompHlocalI(anode,inodeL,inodeT,inodeRR);
      pvtxL->Angle  = CompAngleI(anode,inodeLL,inodeL,inodeT);
      pvtxT->Angle  = CompAngleI(anode,inodeL,inodeT,inodeRR);
      nMinMaxAngles(pmesh);
    }

    else if (cQuadL == 1)
    {
      ivtxT  = pvtxL->ivtxL;
      pvtxT  = avtx + ivtxT;
      inodeT = pvtxT->inode;
      iedgeL = iedgeAddEdge(pmesh,inodeR,inodeT);
      itri  = itriAddTriangle(pmesh,pvtxT->iedge,pvtxL->iedge,
				    pvtxM->iedge,iedgeL);
      pmesh->atri[itri].wLabel = pmesh->ppoly->wLabel;
      nDeleteVertex(pmesh,ivtxM);
      nDeleteVertex(pmesh,ivtxL);
      pvtxT->iedge = iedgeL;
      ivtxLL  = avtx[ivtxT].ivtxL;
      ivtxRR  = avtx[ivtxR].ivtxR;
      inodeLL = avtx[ivtxLL].inode;
      inodeRR = avtx[ivtxRR].inode;
      pvtxT->Hlocal = CompHlocalI(anode,inodeLL,inodeT,inodeR);
      pvtxR->Hlocal = CompHlocalI(anode,inodeT,inodeR,inodeRR);
      pvtxT->Angle  = CompAngleI(anode,inodeLL,inodeT,inodeR);
      pvtxR->Angle  = CompAngleI(anode,inodeT,inodeR,inodeRR);
      nMinMaxAngles(pmesh);
    }

    /*	  NEED TO ADD A NODE			       */
    /*						       */
    /*	  x----o----o		   x----o----o	       */
    /*	  |			   |	|	       */
    /*	  |			   |	|	       */
    /*	  o		     TO	   o----o	       */
    /*	  |			   |		       */
    /*	  |			   |		       */
    /*	  o			   o		       */

    else
    {
      pnodeL = anode + inodeL;
      pnodeM = anode + inodeM;
      pnodeR = anode + inodeR;
      dxL = fabs(pnodeM->x - pnodeL->x);
      dxR = fabs(pnodeM->x - pnodeR->x);
      dyL = fabs(pnodeM->y - pnodeL->y);
      dyR = fabs(pnodeM->y - pnodeR->y);
      nodeT.x = (dxL < dxR) ? pnodeR->x : pnodeL->x;
      nodeT.y = (dyL < dyR) ? pnodeR->y : pnodeL->y;
      inodeT = inodeAddNode(pmesh,nodeT.x,nodeT.y,TRUE);
      iedgeL = iedgeAddEdge(pmesh,inodeL,inodeT);
      iedgeR = iedgeAddEdge(pmesh,inodeR,inodeT);
      itri  = itriAddTriangle(pmesh,iedgeL,pvtxL->iedge,pvtxM->iedge,iedgeR);
      anode = pmesh->anode;
      pmesh->atri[itri].wLabel = pmesh->ppoly->wLabel;
      pvtxM->inode = inodeT;
      pvtxM->iedge = iedgeR;
      pvtxL->iedge = iedgeL;
      ivtxLL  = avtx[ivtxL].ivtxL;
      ivtxRR  = avtx[ivtxR].ivtxR;
      inodeLL = avtx[ivtxLL].inode;
      inodeRR = avtx[ivtxRR].inode;
      pvtxL->Hlocal = CompHlocalI(anode,inodeLL,inodeL,inodeT);
      pvtxM->Hlocal = CompHlocalI(anode,inodeL,inodeT,inodeR);
      pvtxR->Hlocal = CompHlocalI(anode,inodeT,inodeR,inodeRR);
      pvtxL->Angle  = CompAngleI(anode,inodeLL,inodeL,inodeT);
      pvtxM->Angle  = CompAngleI(anode,inodeL,inodeT,inodeR);
      pvtxR->Angle  = CompAngleI(anode,inodeT,inodeR,inodeRR);
      nMinMaxAngles(pmesh);
    }


    /* draw the rectangle if in visual construction mode */
#if 0
    if (cmdargs.wVisual == VISUAL_CONSTRUCT)
    { nDrawRectangle(anode,inodeL,inodeM,inodeR,inodeT);   /* @@@ */
      if (cmdargs.cNoPress != -1)
      { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	else { if (getch() == 'q') exit(20); }
      }
    }
#endif

    /* terminate loop */
    if (cvtx == 4) break;
  }

  return 0;

} /* nRectangPolygon */

