/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     actions.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/01/94    KMK      initial implementation

 DESCRIPTION:

 This module implements the Translate::UserActions function which is called
 by the SGFramework parser.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "types.h"
#include "tokens.h"
#include "states.h"
#include "actions.h"
#include "error.h"
#include "source.h"
#include "symbol.h"
#include "mathnode.h"

#define LTA_LEN	      6


///////////////////////////////////////////////////////////////////////////////
// Function:	 AdjustTokenStack
// Scope:	 Private
// Description:	 This function pops c tokens from the token state and stores
//		 them in reverse order in the Token array.
void AdjustTokenStack(UINT c, TOKEN *atoken, TokenStack *stackToken)
{
  for(--c; c != UINT (-1); --c) atoken[c] = stackToken->pop();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareConstant
// Scope:	 Private
// Description:	 This function attempts to add a constant to the symbol table.
//		 The name of the constant is stored in the ptok1->data.sz
//		 field.	 The constant's value is stored in the math node
//		 structure pointed by the ptok2->data.p field.
void DeclareConstant (TOKEN *ptok1, TOKEN *ptok2)
{
  char *psz = ptok1->data.sz;
  MathNode *pmn = (MathNode *) ptok2->data.p;

  // report an error if the symbol is already defined
  if (symTable->Retrieve(psz))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok1->iPtr);
    if (pmn) delete pmn;
    return;
  }

  // report a warning if the constant expression is not present
  if (pmn == NULL)
  { error->AddMessage(EHM_NOTDEC1, "constant", psz, ptok1->iPtr);
    return;
  }

  // report an error is the constant expression is not constant
  if (pmn->uExpr & FLAG_VAR)
  { error->AddMessage(EHM_CONST, psz, pmn->iPtrExpr);
    error->AddMessage(EHM_NOTDEC1, "constant", psz, ptok1->iPtr);
    delete pmn;
    return;
  }

  // construct and store the constant symbol
  if (pmn->uExpr & FLAG_REAL)
  { SymRConst *psymR = new SymRConst(psz, pmn->data.r);
    symTable->Store(psymR);
  }
  else
  { SymIConst *psymI = new SymIConst(psz, pmn->data.n);
    symTable->Store(psymI);
  }
  delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnIdentifier
// Scope:	 Private
// Description:	 This function makes either a variable or constant mathnode
//		 depending upon the symbol's type.  The name of the symbol
//		 is stored in the ptok->data.sz buffer.
MathNode * pmnIdentifier (TOKEN *ptok)
{
  char *psz = ptok->data.sz;
  HashNode *psym = symTable->Retrieve(psz);

  // report an error if the symbol is not defined
  if (psym == NULL)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok->iPtr);
    return NULL;
  }

  // construct and return the appropriate math node
  SymCoord    *psymC;
  SymRConst   *psymRC;
  SymIConst   *psymIC;
  SymRefine   *psymRef;
  MathNode    *pmn;
  switch (psym->uType)
  { case TYPE_RCONST:
      psymRC = TYPESAFE_DOWNCAST(psym, SymRConst);
      pmn = new MathNode(psymRC->r, ptok->iPtr);
      break;

    case TYPE_ICONST:
      psymIC = TYPESAFE_DOWNCAST(psym, SymIConst);
      pmn = new MathNode(psymIC->n, ptok->iPtr);
      break;

    case TYPE_COORD:
      psymC = TYPESAFE_DOWNCAST(psym, SymCoord);
      pmn = new MathNode(psymC->i, ptok->iPtr);
      break;

    case TYPE_REFINE:
      psymRef = TYPESAFE_DOWNCAST(psym, SymRefine);
      pmn = new MathNode();
      pmn->uNode = TYPE_REFINE;
      pmn->uExpr = FLAG_REAL | FLAG_VAR;
      pmn->iPtrNode = ptok->iPtr;
      pmn->iPtrExpr = ptok->iPtr;
      pmn->data.n = (int) psymRef->i;
      break;

    default:
      error->AddMessage(EHM_SYMEXP, psz, "constant, refinement or coordinate", ptok->iPtr);
      pmn = NULL;
      break;
  }
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnFunction
// Scope:	 Private
// Description:	 This function makes a function math node expression.  The name
//		 of the symbol is stored in the ptok->data.sz field.  The func-
//		 tion arguments are stored in the math node expression pointed
//		 by the ptok2->data.p field.
MathNode * pmnFunction (TOKEN *ptok1, TOKEN *ptok2)
{
  UINT i;
  char *psz = ptok1->data.sz;
  MathNode *pmn = (MathNode *) ptok2->data.p;
  HashNode *psym = symTable->Retrieve(psz);

  // report an error if the symbol is not defined
  if (psym == NULL)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok1->iPtr);
    delete pmn;
    return NULL;
  }

  // report an error if the symbol is not a function
  if (psym->uType != TYPE_FUNC)
  { error->AddMessage(EHM_SYMEXP, psz, "function", ptok1->iPtr);
    delete pmn;
    return NULL;
  }

  // determine if an error has occurred in the specification of the function
  for(i = 0; i < pmn->c; ++i)
    if (pmn->apmn[i] == NULL)
    { delete pmn;
      return NULL;
    }

  // typecast the symbol and determine the number of arguments
  SymFunc *psymFunc = TYPESAFE_DOWNCAST(psym, SymFunc);
  UINT c = psymFunc->c;

  // report an error if the array has the wrong number of arguments
  if (c != pmn->c)
  { error->AddMessage(EHM_NUMARGS, psz, ptok1->iPtr);
    delete pmn;
    return NULL;
  }

  // everything seems o.k. ... convert arguments to reals if necessary
  for(i = 0; i < c; ++i)
    if (pmn->apmn[i]->uExpr & FLAG_INT)
      pmn->apmn[i] = new MathNode(TYPE_REAL, pmn->apmn[i], pmn->apmn[i]->iPtrNode);

  pmn->uNode = TYPE_FUNC;
  pmn->data.func.iFunc	 = psymFunc->i;
  pmn->data.func.iPD	 = 0;
  pmn->data.func.aprf[0] = psymFunc->aprf[0];
  pmn->data.func.aprf[1] = psymFunc->aprf[1];
  pmn->data.func.aprf[2] = psymFunc->aprf[2];

  BOOL f = TRUE;
  for(i = 0; i < c; ++i)  if (pmn->apmn[i]->uExpr & FLAG_VAR) f = FALSE;
  pmn->uExpr = ((f) ? FLAG_CONST : FLAG_VAR) | FLAG_REAL;
  pmn->iPtrNode = pmn->iPtrExpr = ptok1->iPtr;
  pmn->SimplifyFunc();
  return (pmn);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnUnaryOperator
// Scope:	 Private
// Description:	 This function makes a unary operator.	The operator's type
//		 is stored in the ptok1->uTokenType field and a pointer to
//		 the operand is stored in the ptok1->data.p field.
MathNode * pmnUnaryOperator (TOKEN *ptok1, TOKEN *ptok2)
{
  UINT uOp = ptok1->uTokenType;
  MathNode *pmn1 = (MathNode *) ptok2->data.p;

  // if the operand is not defined, return NULL
  if (pmn1 == NULL) return NULL;

  // form the unary operator expression
  MathNode *pmn;
  switch (uOp)
  { case '+':	   pmn = pmn1;					      break;
    case '-':	   pmn = new MathNode(TYPE_NEG,	 pmn1, ptok1->iPtr);  break;
    case TOK_NOT:  pmn = new MathNode(TYPE_NOT,	 pmn1, ptok1->iPtr);  break;
    case TOK_REAL: pmn = new MathNode(TYPE_REAL, pmn1, ptok1->iPtr);  break;
  }
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pmnBinaryOperator
// Scope:	 Private
// Description:	 This function makes a binnary operator.  The operator's type
//		 is stored in the ptok1->uTokenType field and pointers to the
//		 operands are stored in the ptok2->data.p and ptok3->data.p
//		 fields.
MathNode * pmnBinaryOperator (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
  UINT uOp = ptok1->uTokenType;
  MathNode *pmn1 = (MathNode *) ptok2->data.p;
  MathNode *pmn2 = (MathNode *) ptok3->data.p;

  // if the operand is not defined, return NULL
  if ((pmn1 == NULL) || (pmn2 == NULL))
  { if (pmn1) delete pmn1;
    if (pmn2) delete pmn2;
    return NULL;
  }

  // form the binary operator expression
  UINT uType;
  switch (uOp)
  { case '+':	   uType = TYPE_ADD;  break;
    case '-':	   uType = TYPE_SUB;  break;
    case '*':	   uType = TYPE_MUL;  break;
    case '/':	   uType = TYPE_DIV;  break;
    case '>':	   uType = TYPE_GT;   break;
    case '<':	   uType = TYPE_LT;   break;
    case TOK_AND:  uType = TYPE_AND;  break;
    case TOK_OR:   uType = TYPE_OR;   break;
    case TOK_XOR:  uType = TYPE_XOR;  break;
    case TOK_GE:   uType = TYPE_GE;   break;
    case TOK_LE:   uType = TYPE_LE;   break;
    case TOK_EQ:   uType = TYPE_EQ;   break;
    case TOK_NE:   uType = TYPE_NE;   break;
  }
  MathNode *pmn = new MathNode(uType, pmn1, pmn2, ptok1->iPtr);
  return pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareCoordinates
// Scope:	 Private
// Description:	 This function creates a coordinate symbol.
void DeclareCoordinates (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
  char *psz1 = ptok2->data.sz;
  char *psz2 = ptok3->data.sz;

  // report an error if the symbols are already defined
  BOOL fError = FALSE;
  if (symTable->Retrieve(psz1))
  { error->AddMessage(EHM_SYMREDEF, psz1, ptok2->iPtr);
    fError = TRUE;
  }
  if (symTable->Retrieve(psz2))
  { error->AddMessage(EHM_SYMREDEF, psz2, ptok3->iPtr);
    fError = TRUE;
  }
  if (fError)
  { error->AddMessage(EHM_NOTDEC2, ptok1->iPtr);
    return;
  }

  // create and store coordinate symbols
  symTable->Store(new SymCoord(psz1, 0));
  symTable->Store(new SymCoord(psz2, 1));
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclarePoint
// Scope:	 Private
// Description:	 This function creates a point symbol.
void DeclarePoint (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
  char *psz = ptok1->data.sz;
  MathNode *pmnX = (MathNode *) ptok2->data.p;
  MathNode *pmnY = (MathNode *) ptok3->data.p;

  // report an error if the point symbol is already defined
  if (symTable->Retrieve(psz))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok1->iPtr);
    if (pmnX) delete pmnX;
    if (pmnY) delete pmnY;
    return;
  }

  // verify the point coordinates evaluate to a constant
  // report an error is the constant expression is not constant
  BOOL fError = FALSE;
  if (pmnX && (pmnX->uExpr & FLAG_VAR))
  { error->AddMessage(EHM_CONST, "x coordinate", pmnX->iPtrExpr);
    fError = TRUE;
  }
  if (pmnY && (pmnY->uExpr & FLAG_VAR))
  { error->AddMessage(EHM_CONST, "y coordinate", pmnY->iPtrExpr);
    fError = TRUE;
  }
  if (fError || !pmnX || !pmnY)
  { error->AddMessage(EHM_NOTDEC1, "point", psz, ptok1->iPtr);
    if (pmnX) delete pmnX;
    if (pmnY) delete pmnY;
    return;
  }

  // create and store point symbol
  UINT i = symTable->cSymbols(TYPE_POINT);
  REAL x = (pmnX->uExpr & FLAG_REAL) ? pmnX->data.r : (REAL) pmnX->data.n;
  REAL y = (pmnY->uExpr & FLAG_REAL) ? pmnY->data.r : (REAL) pmnY->data.n;
  symTable->Store(new SymPoint(psz, i, x, y));
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 VerifyLabel
// Scope:	 Private
// Description:	 This function verify the passed label and returns TRUE if the
//		 label is valid and FALSE if it is invalid.
BOOL VerifyLabel (TOKEN *ptok, UINT &iLabel, UINT uType)
{
  // assign default label index
  iLabel = 0;

  // if the label does not exist then return
  if (!ptok) return TRUE;

  // otherwise verify the label
  char *psz = ptok->data.sz;
  HashNode *psym = symTable->Retrieve(psz);

  // if the symbol has not been defined, then create a label symbol
  // and store it in the symbol table
  if (!psym)
  { iLabel = symTable->cSymbols(TYPE_LABEL) + 1;
    symTable->Store(new SymLabel(psz, iLabel, uType));
    return TRUE;
  }

  // symbol exists ... verify the type
  BOOL fError = psym->uType != TYPE_LABEL;
  if (!fError)
  { SymLabel *psymLabel = TYPESAFE_DOWNCAST(psym, SymLabel);
    fError = psymLabel->u != uType;
    if (!fError) iLabel = psymLabel->i;
  }

  // report an error if necessary and return
  if (fError) error->AddMessage(EHM_SYMEXP, psz, (uType == TYPE_EDGE) ? "edge label" : "region label", ptok->iPtr);
  return !fError;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 VerifyPointOrEdge
// Scope:	 Private
// Description:	 This function verify the passed point or edge and returns
//		 TRUE if the label is valid and FALSE if it is invalid.
BOOL VerifyPointOrEdge (TOKEN *ptok, UINT &i, UINT uType)
{
  // assign default index
  i = -1u;

  // if the point or edge does not exist, then return
  if (!ptok) return TRUE;

  char *psz = ptok->data.sz;
  HashNode *psym = symTable->Retrieve(psz);

  // report an error if the identifier does not exist
  if (!psym)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok->iPtr);
    return FALSE;
  }

  // report an error if the identifer is not the right type
  else if (psym->uType != uType)
  { error->AddMessage(EHM_SYMEXP, psz, (uType == TYPE_EDGE) ? "edge" : "point", ptok->iPtr);
    return FALSE;
  }

  // retrieve the edge or region index and return
  i = (uType == TYPE_EDGE) ? TYPESAFE_DOWNCAST(psym, SymEdge)->i
			   : TYPESAFE_DOWNCAST(psym, SymPoint)->i;
  return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareEdge
// Scope:	 Private
// Description:	 This function creates an edge symbol.
void DeclareEdge (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3, TOKEN *ptok4,
		  TOKEN *ptok5, TOKEN *ptok6, TOKEN *ptok7)
{
  char *psz1 = ptok1->data.sz;
  MathNode *pmn1 = (ptok6) ? (MathNode *) ptok6->data.p : (MathNode *) NULL;
  MathNode *pmn2 = (ptok7) ? (MathNode *) ptok7->data.p : (MathNode *) NULL;

  // report an error if the edge symbol is already defined
  if (symTable->Retrieve(psz1))
  { error->AddMessage(EHM_SYMREDEF, psz1, ptok1->iPtr);
    if (pmn1) delete pmn1;
    if (pmn2) delete pmn2;
    return;
  }

  // verify the label
  UINT iLabel;
  BOOL fError = !VerifyLabel(ptok2, iLabel, TYPE_EDGE);

  // verify the first point
  UINT iPnt1;
  fError |= !VerifyPointOrEdge(ptok3, iPnt1, TYPE_POINT);

  // verify the second point
  UINT iPnt2;
  fError |= !VerifyPointOrEdge(ptok4, iPnt2, TYPE_POINT);

  // verify the third point
  UINT iPnt3;
  fError |= !VerifyPointOrEdge(ptok5, iPnt3, TYPE_POINT);

  // check for user defined information
  UINT cSeg = 0;
  REAL InitialSpacing = -1.0;
  REAL SpacingGrade   = -1.0;
  BOOL fError1 = !pmn1;
  BOOL fError2 = !pmn2;
  if (ptok6)
  { if (ptok4)
    { if (!fError1)
      { if ((pmn1->uExpr & FLAG_VAR)  ||
	    (pmn1->uExpr & FLAG_REAL) ||
	    (pmn1->data.n < 0))
	{ error->AddMessage(EHM_POSICONST, "expression", ptok6->iPtr);
	  fError1 = TRUE;
	}
	else cSeg = (UINT) pmn1->data.n;
      }
      fError = fError || fError1;
    }
    else
    { if (!fError1)
      { if (pmn1->uExpr & FLAG_VAR) fError1 = TRUE;
	if (!fError1)
	{ InitialSpacing = (pmn1->uExpr & FLAG_REAL) ? pmn1->data.r : (REAL) pmn1->data.n;
	  fError1 = InitialSpacing < 0.0;
	}
	if (fError1) error->AddMessage(EHM_POSCONST, "expression", ptok6->iPtr);
      }
      if (!fError2)
      { if (pmn2->uExpr & FLAG_VAR) fError2 = TRUE;
	if (!fError2)
	{ SpacingGrade = (pmn2->uExpr & FLAG_REAL) ? pmn2->data.r : (REAL) pmn2->data.n;
	  fError2 = SpacingGrade < 0.0;
	}
	if (fError2) error->AddMessage(EHM_POSCONST, "expression", ptok7->iPtr);
      }
      fError = fError || fError1 || fError2;
    }
  }

  // if an error occurs report that the edge is not declared
  if (fError)
  { error->AddMessage(EHM_NOTDEC1, "edge", psz1, ptok1->iPtr);
    return;
  }

  // create and store an edge symbol
  UINT i = symTable->cSymbols(TYPE_EDGE);
  symTable->Store(new SymEdge(psz1, i, iLabel, iPnt1, iPnt2, iPnt3, cSeg,
			      InitialSpacing, SpacingGrade));
  if (pmn1) delete pmn1;
  if (pmn2) delete pmn2;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareRegion
// Scope:	 Private
// Description:	 This function creates a region symbol.
void DeclareRegion (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3, BOOL fRect,
		    TOKEN *ptok4)
{
  char *psz1 = ptok1->data.sz;
  MathNode *pmn = (ptok4) ? (MathNode *) ptok4->data.p : (MathNode *) NULL;

  // report an error if the region symbol is already defined
  if (symTable->Retrieve(psz1))
  { error->AddMessage(EHM_SYMREDEF, psz1, ptok1->iPtr);
    delete (DAPTOKEN *) ptok3->data.p;
    if (pmn) delete pmn;
    return;
  }

  // verify the label
  UINT iLabel;
  BOOL fError = !VerifyLabel(ptok2, iLabel, TYPE_REGION);

  // verify the edges of the regions
  DAPTOKEN *pta = (DAPTOKEN *) ptok3->data.p;
  UINT c = pta->GetElements();
  UINT *auEdge = new UINT[c];
  for(UINT i = 0; i < c; ++i)
  { TOKEN *ptok = pta->operator[](i);
    fError = fError || !VerifyPointOrEdge(ptok, auEdge[i], TYPE_EDGE);
  }

  // check for user defined information
  REAL r = -1.0;
  BOOL fError1 = !pmn;
  if (ptok4)
  { if (!fError1)
    { if (pmn->uExpr & FLAG_VAR) fError1 = TRUE;
      if (!fError1)
      { r = (pmn->uExpr & FLAG_REAL) ? pmn->data.r : (REAL) pmn->data.n;
	if (r < 0.0) fError1 = TRUE;
	if (fError1) error->AddMessage(EHM_POSCONST, "expression", ptok4->iPtr);
      }
    }
    fError = fError || fError1;
  }

  // determine if the region can be divided into rectangular elements
  if (fRect && !fError)
  { fError = c != 4;
    SymEdge *psymE1, *psymE2, *psymE3, *psymE4;
    if (!fError)
    { psymE1 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_EDGE, auEdge[0]), SymEdge);
      psymE2 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_EDGE, auEdge[1]), SymEdge);
      psymE3 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_EDGE, auEdge[2]), SymEdge);
      psymE4 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_EDGE, auEdge[3]), SymEdge);
      fError = (psymE1->iM != -1u) ||
	       (psymE2->iM != -1u) ||
	       (psymE3->iM != -1u) ||
	       (psymE4->iM != -1u) ||
	       (psymE1->SpacingGrade < 0.0) ||
	       (psymE2->SpacingGrade < 0.0) ||
	       (psymE3->SpacingGrade < 0.0) ||
	       (psymE4->SpacingGrade < 0.0) ||
	       (psymE1->InitialSpacing != psymE3->InitialSpacing) ||
	       (psymE2->InitialSpacing != psymE4->InitialSpacing) ||
	       (psymE1->SpacingGrade != psymE3->SpacingGrade) ||
	       (psymE2->SpacingGrade != psymE4->SpacingGrade);
    }
    SymPoint *psymP1, *psymP2, *psymP3, *psymP4;
    if (!fError)
    { BOOL fHor = FALSE;
      BOOL fVer = FALSE;
      psymP1 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_POINT, psymE1->iB), SymPoint);
      psymP2 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_POINT, psymE1->iE), SymPoint);
      psymP3 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_POINT, psymE3->iB), SymPoint);
      psymP4 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_POINT, psymE3->iE), SymPoint);
      if ((psymP1->x == psymP3->x) && (psymP2->x == psymP4->x)) fHor = !fHor;
      if ((psymP1->y == psymP3->y) && (psymP2->y == psymP4->y)) fVer = !fVer;
      psymP1 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_POINT, psymE2->iB), SymPoint);
      psymP2 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_POINT, psymE2->iE), SymPoint);
      psymP3 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_POINT, psymE4->iB), SymPoint);
      psymP4 = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_POINT, psymE4->iE), SymPoint);
      if ((psymP1->x == psymP3->x) && (psymP2->x == psymP4->x)) fHor = !fHor;
      if ((psymP1->y == psymP3->y) && (psymP2->y == psymP4->y)) fVer = !fVer;
      fError = !fHor || !fVer;
    }
    if (fError)
    { error->AddMessage(EHM_NORECTREG, psz1, ptok1->iPtr);
      return;
    }
  }

  // if an error occurs report that the region is not declared
  if (fError)
  { error->AddMessage(EHM_NOTDEC1, "edge", psz1, ptok1->iPtr);
    delete [] auEdge;
    delete (DAPTOKEN *) ptok3->data.p;
    if (pmn) delete pmn;
    return;
  }

  // otherwise create and store the symbol
  UINT iRegion = symTable->cSymbols(TYPE_REGION);
  symTable->Store(new SymRegion(psz1, iRegion, iLabel, c, auEdge, fRect, r));
  delete (DAPTOKEN *) ptok3->data.p;
  if (pmn) delete pmn;
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DeclareRefine
// Scope:	 Private
// Description:	 This function creates a refine symbol.
void DeclareRefine (TOKEN *ptok1, TOKEN *ptok2, UINT u, REAL Distance)
{
  char *psz = ptok1->data.sz;
  MathNode *pmn = (MathNode *) ptok2->data.p;

  // report an error if the symbol is already defined
  if (symTable->Retrieve(psz))
  { error->AddMessage(EHM_SYMREDEF, psz, ptok1->iPtr);
    if (pmn) delete pmn;
    return;
  }

  // report a warning if the refinement expression is not present
  if ((pmn == NULL) || (Distance < 0.0))
  { error->AddMessage(EHM_NOTDEC1, "refinement", psz, ptok1->iPtr);
    return;
  }


  // construct and store the refinement symbol
  UINT iRefine = symTable->cSymbols(TYPE_REFINE);
  UINT uMeasure = (u == TOK_LOG) ? DISTANCE_LOG : ((u == TOK_SLOG) ? DISTANCE_SLOG : DISTANCE_LINEAR);
  symTable->Store(new SymRefine(psz, iRefine, (void *) pmn, uMeasure, Distance));
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessRefineDistance
// Scope:	 Private
// Description:	 This function sets the global rRefineDistance variable.
void ProcessRefineDistance (TOKEN *ptok)
{
  MathNode *pmn = (MathNode *) ptok->data.p;
  rRefineDistance = -1.0;

  // return in the expression is not present
  if (!pmn) return;

  // report an error if the expression is not constant
  if (( pmn->uExpr & FLAG_VAR) ||
      ((pmn->uExpr & FLAG_REAL) && (pmn->data.r < 0.0)) ||
      ((pmn->uExpr & FLAG_INT ) && (pmn->data.n < 0)))
  { error->AddMessage(EHM_POSCONST, "refinement distance", pmn->iPtrExpr);
    delete pmn;
    return;
  }

  // set the refinement distance
  rRefineDistance = (pmn->uExpr & FLAG_REAL) ? pmn->data.r : (REAL) pmn->data.n;
  delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessSet1
// Scope:	 Private
// Description:	 This function initializes a set variable.
void ProcessSet1 (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
  MathNode *pmn = (MathNode *) ptok3->data.p;
  BOOL fError = pmn == NULL;

  if (!fError)
  { // report an error if the expression is not an integer constant
    if (ptok2->uTokenType == TOK_DIV)
    { if ((pmn->uExpr & FLAG_VAR)  ||
	  (pmn->uExpr & FLAG_REAL) ||
	  (pmn->data.n < 0))
      { error->AddMessage(EHM_POSICONST, "min/max divisions", pmn->iPtrExpr);
	fError = TRUE;
      }
    }
  }

  // set the appropriate variable
  if (!fError)
  { if (ptok1->uTokenType == TOK_MAX)
    { if (ptok2->uTokenType == TOK_DIV)
      { cMaxDivisions = pmn->data.n;
	if (pmn) delete pmn;
      }
      else
      { if (pMaxEdgeLen) delete (MathNode *) pMaxEdgeLen;
	pMaxEdgeLen = (void *) pmn;
      }
    }
    else
    { if (ptok2->uTokenType == TOK_DIV)
      { cMinDivisions = pmn->data.n;
	if (pmn) delete pmn;
      }
      else
      { if (pMinEdgeLen) delete (MathNode *) pMinEdgeLen;
	pMinEdgeLen = (void *) pmn;
      }
    }
  }

  // report that the internal variabl is not declared if an error occurred
  // and return
  if (fError) error->AddMessage(EHM_NOTDEC, "set variable", ptok1->iPtr);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ProcessSet2
// Scope:	 Private
// Description:	 This function initializes a set variable.
void ProcessSet2 (TOKEN *ptok1, TOKEN *ptok2, TOKEN *ptok3)
{
  MathNode *pmn = (MathNode *) ptok3->data.p;
  char *psz = ptok2->data.sz;
  BOOL fError = pmn == NULL;

  // which coordinate was specified
  HashNode *psym = symTable->Retrieve(psz);
  if (psym == NULL)
  { error->AddMessage(EHM_SYMNOTDEF, psz, ptok2->iPtr);
    fError = TRUE;
  }
  else if (psym->uType != TYPE_COORD)
  { error->AddMessage(EHM_SYMEXP, psz, "coordinate", ptok2->iPtr);
    fError = TRUE;
  }

  // set the appropriate variable
  if (!fError)
  { if (TYPESAFE_DOWNCAST(psym, SymCoord)->i == 0)
    { if (pMaxXEdgeLen) delete (MathNode *) pMaxXEdgeLen;
      pMaxXEdgeLen = (void *) pmn;
    }
    else
    { if (pMaxYEdgeLen) delete (MathNode *) pMaxYEdgeLen;
      pMaxYEdgeLen = (void *) pmn;
    }
  }

  // report that the internal variabl is not declared if an error occurred
  // and return
  if (fError)
  { error->AddMessage(EHM_NOTDEC, "set variable", ptok1->iPtr);
    if (pmn) delete pmn;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Source::UserActions
// Scope:	 Private
// Description:	 This function is the user action handler.  Zero is returned
//		 if no error occurs; otherwise, a nonzero interger is returned.
void Source::UserActions (UINT UserAction)
{
  TOKEN	 atoken[LTA_LEN];
  TOKEN *ptok;

  switch (UserAction)
  { // initialize the user actions function
    case USR_INIT:
      uSyntaxState  = ST_PROG;
      fLabelPresent = FALSE;
      pMinEdgeLen   = NULL;
      pMaxEdgeLen   = NULL;
      pMaxXEdgeLen  = NULL;
      pMaxYEdgeLen  = NULL;
      cMinDivisions = 0;
      cMaxDivisions = 0;
      break;

    // declare a constant
    case USR_DCONST:
      AdjustTokenStack(2, atoken, stackToken);
      DeclareConstant(atoken, atoken+1);
      break;

    // make a real constant math node
    case USR_RCONST:
      token.data.p = (void *) new MathNode(token.data.r, token.iPtr);
      break;

    // make an integer constant math node
    case USR_ICONST:
      token.data.p = (void *) new MathNode(token.data.n, token.iPtr);
      break;

    // make an identifier (i.e. variable or constant) math node
    case USR_ID:
      token.data.p = (void *) pmnIdentifier(&token);
      break;

    // make a function math node expression
    case USR_FUNC:
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (void *) pmnFunction(atoken, atoken+1);
      break;

    // make a unary operator math node expression
    case USR_UNARYOP:
      atoken[0] = stackToken->pop();
      token.data.p = (void *) pmnUnaryOperator(atoken, &token);
      break;

    // make a real operator math node expression
    case USR_REAL:
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (void *) pmnUnaryOperator(atoken, atoken+1);
      break;

    // make a binary operator math node expression
    case USR_BINOP:
      AdjustTokenStack(2, atoken, stackToken);
      token.data.p = (void *) pmnBinaryOperator(atoken+1, atoken, &token);
      break;

    // process '(' expr ')', '[' expr ']' and '{' expr '}' expressions
    case USR_PAREN:
      token = stackToken->pop();
      break;

    // create a parent math node
    case USR_EXPRS:
      atoken->data.p = (void *) new MathNode();
      stackToken->push(*atoken);
      break;

    // add a child math node expression to a parent math node
    case USR_ADDEXPR:
      atoken[1] = stackToken->pop();
      atoken[0] = stackToken->pop();
      ((MathNode *) atoken[0].data.p)->AddChild((MathNode *) atoken[1].data.p);
      stackToken->push(atoken[0]);
      break;

    // create a token list
    case USR_RLIST:
      atoken->data.p = (void *) new DAPTOKEN;
      stackToken->push(*atoken);
      break;

    // add a token to the token list
    case USR_ADDID:
      ptok = new TOKEN;
      *ptok = stackToken->pop();
      atoken[0] = stackToken->pop();
      ((DAPTOKEN *) atoken[0].data.p)->Add(ptok);
      stackToken->push(atoken[0]);
      break;

    // create a coordinate symbol
    case USR_DCOORD:
      AdjustTokenStack(3, atoken, stackToken);
      DeclareCoordinates(atoken, atoken+1, atoken+2);
      break;

    // create a point symbol
    case USR_DPOINT:
      AdjustTokenStack(3, atoken, stackToken);
      DeclarePoint(atoken, atoken+1, atoken+2);
      break;

    // set the label present flag
    case USR_LABEL:
      fLabelPresent = TRUE;
      break;

    // clear the label present flag
    case USR_NOLABEL:
      fLabelPresent = FALSE;
      break;

    // declare a straight edge with default spacing
    case USR_DEDGE1:
      if (fLabelPresent)
      { AdjustTokenStack(4, atoken, stackToken);
	DeclareEdge(atoken, atoken+1, atoken+2, NULL, atoken+3, NULL, NULL);
      }
      else
      { AdjustTokenStack(3, atoken, stackToken);
	DeclareEdge(atoken, NULL, atoken+1, NULL, atoken+2, NULL, NULL);
      }
      break;

    // declare a straight edge with user defined spacing
    case USR_DEDGE2:
      if (fLabelPresent)
      { AdjustTokenStack(6, atoken, stackToken);
	DeclareEdge(atoken, atoken+1, atoken+2, NULL, atoken+3, atoken+4,
		    atoken+5);
      }
      else
      { AdjustTokenStack(5, atoken, stackToken);
	DeclareEdge(atoken, NULL, atoken+1, NULL, atoken+2, atoken+3,
		    atoken+4);
      }
      break;

    // declare a curved edge with default spacing
    case USR_DEDGE3:
      if (fLabelPresent)
      { AdjustTokenStack(5, atoken, stackToken);
	DeclareEdge(atoken, atoken+1, atoken+2, atoken+3, atoken+4, NULL, NULL);
      }
      else
      { AdjustTokenStack(4, atoken, stackToken);
	DeclareEdge(atoken, NULL, atoken+1, atoken+2, atoken+3, NULL, NULL);
      }
      break;

    // declare a curved edge with a user defined number of points
    case USR_DEDGE4:
      if (fLabelPresent)
      { AdjustTokenStack(6, atoken, stackToken);
	DeclareEdge(atoken, atoken+1, atoken+2, atoken+3, atoken+4,
		    atoken+5, NULL);
      }
      else
      { AdjustTokenStack(5, atoken, stackToken);
	DeclareEdge(atoken, NULL, atoken+1, atoken+2, atoken+3,
		    atoken+4, NULL);
      }
      break;

    // declare a region
    case USR_DREG1:
      AdjustTokenStack(3, atoken, stackToken);
      DeclareRegion(atoken, atoken+1, atoken+2, FALSE, NULL);
      break;

    // declare a region
    case USR_DREG2:
      AdjustTokenStack(3, atoken, stackToken);
      DeclareRegion(atoken, atoken+1, atoken+2, TRUE, NULL);
      break;

    // declare a region with user defined weight
    case USR_DREG3:
      AdjustTokenStack(4, atoken, stackToken);
      DeclareRegion(atoken, atoken+1, atoken+2, FALSE, atoken+3);
      break;

    // declare a refinement variable
    case USR_REFINE:
      AdjustTokenStack(2, atoken, stackToken);
      DeclareRefine(atoken, atoken+1, uRefineMeasure, rRefineDistance);
      break;

    // default refinement variable parameters
    case USR_REFDEF:
      uRefineMeasure  = TOK_LINEAR;
      rRefineDistance = 1.0;
      break;

    // initialize refinement variable measurement parameter
    case USR_REFMEAS:
      atoken[0] = stackToken->pop();
      uRefineMeasure = atoken->uTokenType;
      break;

    // initialize refinement variable distance parameter
    case USR_REFDIST:
      atoken[0] = stackToken->pop();
      ProcessRefineDistance(atoken);
      break;

    // initialize a set variable
    case USR_SET1:
      AdjustTokenStack(3, atoken, stackToken);
      ProcessSet1(atoken, atoken+1, atoken+2);
      break;

    // initialize a set variable
    case USR_SET2:
      AdjustTokenStack(4, atoken, stackToken);
      ProcessSet2(atoken, atoken+1, atoken+3);
      break;
  }
}

