/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     codebuf.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/02/94    KMK      initial implementation

 DESCRIPTION:

 This module implements the code buffer class.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#if defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#else
#include <io.h>
#include <sys\stat.h>
#endif

#include <fcntl.h>
#include "codebuf.h"
#include "error.h"
#include "symbol.h"
#include "mathnode.h"


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::CodeBuffer
// Scope:	 Public, Constructor
CodeBuffer::CodeBuffer (char *pszName, char *pszPrefix, UINT uPrec, UINT iLeft, UINT iRight)
{
  strcpy(szFileName, pszName);
  strcpy(szUCPrefix, pszPrefix);
  strcpy(szLCPrefix, pszPrefix);
  strupr(szUCPrefix);
  /* strlwr(szLCPrefix); */ /* Switch to case-sensitive by Josh */
  cPrecision = uPrec;
  iLeftMargin = iLeft;
  iRightMargin = iRight;
  char *psz = szBuffer;
  for(iBuffer = 0; iBuffer < iLeftMargin; ++iBuffer, ++psz) *psz = ' ';
  nFile = open(pszName, O_CREAT | O_TRUNC | O_BINARY | O_WRONLY, S_IREAD | S_IWRITE);
  if (nFile == -1) error->AddMessage(EHM_FILEOPEN, pszName);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::~CodeBuffer
// Scope:	 Public, Destructor
CodeBuffer::~CodeBuffer ()
{
  Flush();
  close(nFile);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::Flush
// Scope:	 Private
// Description:	 This function flushes the code buffer.
void CodeBuffer::Flush ()
{
#if defined(UNIX)
  szBuffer[iBuffer++] = '\n';
#else
  szBuffer[iBuffer++] = '\x0d';
  szBuffer[iBuffer++] = '\x0a';
#endif
  write(nFile, szBuffer, iBuffer);
  char *psz = szBuffer;
  for(iBuffer = 0; iBuffer < iLeftMargin; ++iBuffer, ++psz) *psz = ' ';
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddToken
// Scope:	 Private
// Description:	 This function adds a token to the code buffer.
void CodeBuffer::AddToken (char *psz)
{
  UINT c = strlen(psz);
  if (iBuffer + c >= iRightMargin) Flush();
  strcpy(szBuffer + iBuffer, psz);
  iBuffer += c;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::SetLeftMargin
// Scope:	 Public
// Description:	 This function sets the left margin
void CodeBuffer::SetLeftMargin (UINT u)
{
  UINT i;
  iLeftMargin = u;

  for(i = 0; i < iBuffer; ++i)
    if (szBuffer[i] != ' ') break;

  if (i == iBuffer)
    for(iBuffer = 0; iBuffer < iLeftMargin; ++iBuffer)
      szBuffer[iBuffer] = ' ';
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddChar
// Scope:	 Public
// Description:	 This function adds a character to the code buffer.
void CodeBuffer::AddChar (char ch)
{
  switch (ch)
  { case '\n':
      Flush();
      break;

    case ' ':
      if (iBuffer + 1 >= iRightMargin) Flush();
      else if (iBuffer > iLeftMargin) szBuffer[iBuffer++] = ch;
      break;

    case '\0':
      break;

    default:
      if (iBuffer + 1 >= iRightMargin) Flush();
      szBuffer[iBuffer++] = ch;
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddInteger
// Scope:	 Public
// Description:	 This function adds an integer to the code buffer.
void CodeBuffer::AddInteger (int n, BOOL fAddSpace)
{
  char sz[32];
  sprintf(sz, "%d", n);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddUInteger
// Scope:	 Public
// Description:	 This function adds an unsigned integer to the code buffer.
void CodeBuffer::AddUInteger (UINT u, BOOL fAddSpace)
{
  char sz[32];
  sprintf(sz, "%u", u);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddReal
// Scope:	 Public
// Description:	 This function adds a real number to the code buffer.
void CodeBuffer::AddReal (REAL r, BOOL fAddSpace)
{
  char sz[32];
  sprintf(sz, "%*.*le", cPrecision, cPrecision, r);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddCoordinate
// Scope:	 Public
// Description:	 This function adds a coordinate to the code buffer.
void CodeBuffer::AddCoordinate (UINT i, BOOL fAddSpace)
{
  char sz[32];
  sprintf(sz, "_x%u", i);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddEquation
// Scope:	 Public
// Description:	 This function adds an equation to the code buffer.
void CodeBuffer::AddEquation (UINT i, BOOL fAddSpace)
{
  char sz[32];
  sprintf(sz, "EQU_%s_%u", szUCPrefix, i + 1);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddFunction
// Scope:	 Public
// Description:	 This function adds a function to the code buffer.
void CodeBuffer::AddFunction (char *psz, UINT iPD, BOOL fAddSpace)
{
  // de-alias function names
  char szFuncName[32];
  if	  (!strcmp(psz, "ln"))	strcpy(szFuncName, "log");
  else if (!strcmp(psz, "log")) strcpy(szFuncName, "log10");
  else				strcpy(szFuncName, psz);

  char sz[32];
  if (iPD) sprintf(sz, "pd%u%s", iPD, szFuncName);
  else strcpy(sz, szFuncName);
  AddToken(sz);
  if (fAddSpace) AddChar(' ');
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddFunction
// Scope:	 Public
// Description:	 This function adds a function to the code buffer.
void CodeBuffer::AddFunction (UINT i, UINT iPD, BOOL fAddSpace)
{
  SymFunc *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_FUNC, i), SymFunc);
  AddFunction(psym->szName, iPD, fAddSpace);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddText
// Scope:	 Public
// Description:	 This function adds text to the code buffer.  This function
//		 replaces the substrings @U@ and @L@ with the uppercase and
//		 lowercase prefix.
void CodeBuffer::AddText (char *psz)
{
  strcpy(pszGlobalBuf, psz);
  psz = pszGlobalBuf;
  while ((psz = strstr(psz, "@U@")) != NULL)
  { psz[0] = szUCPrefix[0];
    psz[1] = szUCPrefix[1];
    psz[2] = szUCPrefix[2];
  }

  psz = pszGlobalBuf;
  while ((psz = strstr(psz, "@L@")) != NULL)
  { psz[0] = szLCPrefix[0];
    psz[1] = szLCPrefix[1];
    psz[2] = szLCPrefix[2];
  }

  char *pszH, *pszT;
  for(pszT = pszH = pszGlobalBuf; *pszT; ++pszT)
  { char ch = *pszT;
    if ((ch == ' ') || (ch == '\n'))
    { *pszT = '\0';
      AddToken(pszH);
      AddChar(ch);
      pszH = pszT + 1;
    }
  }
  if (*pszH) AddToken(pszH);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 CodeBuffer::AddString
// Scope:	 Public
// Description:	 This function adds a string to the code buffer.
void CodeBuffer::AddString (char *psz)
{
  UINT cLen = strlen(psz);
  UINT cCopied = 0;
  BOOL fDone = FALSE;
  while (!fDone)
  { UINT c = iRightMargin - iBuffer - 2;
    if (c > cLen - cCopied) c = cLen - cCopied;
    if (c > 0)
    { szBuffer[iBuffer++] = '\"';
      strncpy(szBuffer + iBuffer, psz + cCopied, c);
      iBuffer += c;
      cCopied += c;
      szBuffer[iBuffer++] = '\"';
    }
    if (cCopied < cLen) AddChar('\n');
    else fDone = TRUE;
  }
}


