/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     error.cpp
 AUTHOR:     Kevin M. Kramer
             J. Joshua Feng

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/25/94    KMK      initial implementation
 10/20/94    KMK      modified for the command line version
 07/06/00    JJF      modify AddMessage

 DESCRIPTION:

 This module implements the SGFramework error handler.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "error.h"
#include "source.h"


// global declaration
ErrorHandler *error;


#define ERROR_BUFSIZE 256
static char szBuf[ERROR_BUFSIZE];


///////////////////////////////////////////////////////////////////////////////
// Description:
//
ErrorHandler::ErrorHandler ( UINT cMaxErrorMsgs, UINT cMaxWarningMsgs )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("error.cpp ErrorHandler::ErrorHandler\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  cMaxErrors = cMaxErrorMsgs;
  cMaxWarnings = cMaxWarningMsgs;
  cErrors = 0;
  cWarnings = 0;
  cFatal = 0;
  cTotal = 0;
  UINT cMessages = cMaxErrors + cMaxWarnings + 1;
  amsg = new MESSAGE[cMessages];
  for (UINT i = 0; i < cMessages; ++i)
    amsg[i].psz = NULL;

}


///////////////////////////////////////////////////////////////////////////////
// Description:
//
ErrorHandler::~ErrorHandler ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("error.cpp ErrorHandler::~ErrorHandler\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT cMessages = cMaxErrors + cMaxWarnings + 1;
  for (UINT i = 0; i < cMessages; ++i)
    if (amsg[i].psz) delete [] amsg[i].psz;
  delete [] amsg;
}



///////////////////////////////////////////////////////////////////////////////
// Description: This function initializes the error handler.
//
void ErrorHandler::Initialize ( void *psource )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("error.cpp ErrorHandler::Initialize\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // initialize the pointer to the active translate client
  p = psource;

  // initialize the error handler variables
  iBuf = 0;
  cErrors = 0;
  cWarnings = 0;
  cFatal = 0;
  cTotal = 0;
  UINT cMessages = cMaxErrors + cMaxWarnings + 1;
  for (UINT i = 0; i < cMessages; ++i)
  { if (amsg[i].psz) delete [] amsg[i].psz;
    amsg[i].psz = NULL;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function is called by all of the AddMessage functions.
//		 It updates the error handler variables and includes line
//		 number information if appropriate.  The message prefix is
//		 stored in the globabl buffer and its length is returned.
UINT ErrorHandler::uMessagePrefix ( UINT uMsg, UINT iBufPos, UINT &iMsg )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("error.cpp ErrorHandler::uMessagePrefix\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  Source *psource = (Source *) p;
  UINT iLoc = (iBufPos == -1U) ? iBuf : iBufPos;
  UINT iLine = (aehm[uMsg].fLineInfo) ? psource->GetLineFromPos(iLoc) : -1U;
  UINT iCol = (iLine != -1U) ? iLoc - psource->GetLineIndex(iLine) : -1U;
  iMsg = cWarnings + cErrors + cFatal;

  switch (aehm[uMsg].uType)
  { case MSG_WARNING:
      if (cWarnings >= cMaxWarnings) return -1u;
      strcpy(szBuf, "[W] ");
      ++cWarnings;
      break;

    case MSG_ERROR:
      if (cErrors >= cMaxErrors) return -1u;
      strcpy(szBuf, "[E] ");
      ++cErrors;
      break;

    case MSG_FATAL:
      if (cFatal >= 1) return -1u;
      strcpy(szBuf, "[F] ");
      ++cFatal;
      break;
  }

  ++cTotal;
  if (iLine != -1u) sprintf(szBuf + 4, "(%u,%u)", iLine + 1, iCol + 1);

  UINT cLen = strlen(szBuf);
  szBuf[cLen++] = ':';
  szBuf[cLen++] = ' ';
  szBuf[cLen++] = ' ';
  amsg[iMsg].iCol = iCol;
  amsg[iMsg].iLine = iLine;
  return cLen;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds an error message to the error handler
void ErrorHandler::AddMessage ( UINT uMsg, UINT iBufPos )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("error.cpp ErrorHandler::AddMessage\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT iMsg;
  UINT cLen = uMessagePrefix(uMsg, iBufPos, iMsg);
  if (cLen == -1u) return;
  strcpy(szBuf + cLen, aehm[uMsg].pszMsg);
  amsg[iMsg].psz = new char[strlen(szBuf) + 1];
  strcpy(amsg[iMsg].psz, szBuf);
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds an error message to the error handler
void ErrorHandler::AddMessage ( UINT uMsg, const char *psz, UINT iBufPos )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("error.cpp ErrorHandler::AddMessage\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT iMsg;
  UINT cLen = uMessagePrefix(uMsg, iBufPos, iMsg);
  if (cLen == -1u) return;
  sprintf(szBuf + cLen, aehm[uMsg].pszMsg, psz);
  amsg[iMsg].psz = new char[strlen(szBuf) + 1];
  strcpy(amsg[iMsg].psz, szBuf);
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds an error message to the error handler
void ErrorHandler::AddMessage ( UINT uMsg, const char *psz1, const char *psz2,
                                UINT iBufPos )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("error.cpp ErrorHandler::AddMessage\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT iMsg;
  UINT cLen = uMessagePrefix(uMsg, iBufPos, iMsg);
  if (cLen == -1u) return;
  sprintf(szBuf + cLen, aehm[uMsg].pszMsg, psz1, psz2);
  amsg[iMsg].psz = new char[strlen(szBuf) + 1];
  strcpy(amsg[iMsg].psz, szBuf);
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function displays the messages to the standard output
//		 device.
void ErrorHandler::ShowMessages ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("error.cpp ErrorHandler::ShowMessages\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // determine the number of messages
  UINT cMsg = cWarnings + cErrors + cFatal;

  // messages are present ... add them to the list box and select the first one
  for (UINT i = 0; i < cMsg; ++i) fprintf(stdout, "%s\n", amsg[i].psz);
}

EHM aehm[] =
{ { MSG_ERROR,	  TRUE,	  "string cannot span multiple lines" },
  { MSG_FATAL,	  TRUE,	  "unexpected end of file" },
  { MSG_ERROR,	  TRUE,	  "%s expected" },
  { MSG_ERROR,	  TRUE,	  "redefinition of symbol \"%s\"" },
  { MSG_ERROR,	  TRUE,	  "symbol \"%s\" not defined" },
  { MSG_ERROR,	  TRUE,	  "\"%s\" is not a %s symbol" },
  { MSG_ERROR,	  TRUE,	  "%s not declared" },
  { MSG_ERROR,	  TRUE,	  "%s symbol \"%s\" not declared" },
  { MSG_ERROR,	  TRUE,	  "coordinates not declared" },
  { MSG_ERROR,	  TRUE,	  "user-defined function \"%s\" not declared" },
  { MSG_ERROR,	  TRUE,	  "\"%s\" does not evaluate to a constant" },
  { MSG_ERROR,	  TRUE,	  "%s does not evaluate to positive constant" },
  { MSG_ERROR,	  TRUE,	  "%s variable does not evaluate to positive integer constant" },
  { MSG_ERROR,	  TRUE,	  "function \"%s\" has the wrong number of arguments" },
  { MSG_FATAL,	  TRUE,	  "division by zero in %s" },
  { MSG_FATAL,	  TRUE,	  "domain error in %s function" },
  { MSG_FATAL,	  TRUE,	  "overflow in %s function" },
  { MSG_FATAL,	  TRUE,	  "underflow in %s function" },
  { MSG_ERROR,	  TRUE,	  "integer overflow" },
  { MSG_ERROR,	  TRUE,	  "cannot open file \"%s\"" },
  { MSG_ERROR,	  TRUE,	  "cannot subdivide region \"%s\" into rectangular elements" }
};

