/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     fileio.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/01/94    KMK      initial implementation

 DESCRIPTION:

 This module implements the FileInputOutput class.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#if defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#else
#include <io.h>
#include <sys\stat.h>
#endif

#include <fcntl.h>
#include "error.h"
#include "fileio.h"
#include "symbol.h"
#include "types.h"
#include "xlatinfo.h"


///////////////////////////////////////////////////////////////////////////////
// Description: Public, Constructor
FileInputOutput::FileInputOutput ( )
{
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function forms the translate and result file name by
//		 modifying the source file name's extension.  The file name
//		 is formed in the global buffer.
void FileInputOutput::FormFileName ( char *pszSource, UINT uFileType,
                                     BOOL fNoPath )
{
  // copy the source file name into the global buffer.
  char *psz = (fNoPath) ? psz = strrchr(pszSource, '\\') : (char *) NULL;
  if (psz) ++psz;
  else psz = pszSource;
  strcpy(pszGlobalBuf, psz);

  // form the file name in the global buffer
  if (uFileType != FILE_SOURCE)
  { char *psz = strchr(pszGlobalBuf, '.');
    if (psz) *psz = '\0';

    psz = strrchr(pszGlobalBuf, '\\');
    if (!psz) psz = pszGlobalBuf;

    switch (uFileType)
    { case FILE_SKEL:
	strcat(pszGlobalBuf, EXT_SKEL);
	break;

      case FILE_GRID:
	strcat(pszGlobalBuf, EXT_GRID);
	break;

      case FILE_REFCODE:
        #ifndef UNIX
	if (strlen(psz) > 4) psz[4] = '\0';
        #endif
	strcat(pszGlobalBuf, "_ref");
	strcat(pszGlobalBuf, EXT_REFCODE);
	break;

      case FILE_REFHEAD:
        #ifndef UNIX
	if (strlen(psz) > 4) psz[4] = '\0';
        #endif
	strcat(pszGlobalBuf, "_ref");
	strcat(pszGlobalBuf, EXT_REFHEAD);
	break;
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description: This function generates the skeleton file.
int FileInputOutput::nWriteSkeletonFile ( char *pszFileName )
{
  UINT i;

  // create skelete file
  FormFileName(pszFileName, FILE_SKEL);
  int nHandle = open(pszGlobalBuf, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY,
                     S_IREAD | S_IWRITE);
  if (nHandle == -1)
  { error->AddMessage(EHM_FILEOPEN, pszGlobalBuf);
    return FILEIO_BADFILE;
  }


  /////////////////////////////
  //  Determine Label Order  //
  /////////////////////////////

  UINT cLabel = symTable->cSymbols(TYPE_LABEL);
  UINT cRegLabel = 0;
  XLATLABEL xlatlabel;
  for(i = 0; i < cLabel; ++i)
  { SymLabel *psymLabel = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_LABEL, i),
                                            SymLabel);
    if (psymLabel->u == TYPE_REGION) ++cRegLabel;
  }
  UINT *aiOrder = new UINT[cLabel];
  UINT *aiInvOrder = new UINT[cLabel];
  UINT iRegLabel = 0;
  UINT iEdgeLabel = cRegLabel;
  for(i = 0; i < cLabel; ++i)
  { SymLabel *psymLabel = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_LABEL, i),
                                            SymLabel);
    if (psymLabel->u == TYPE_REGION) aiOrder[i] = iRegLabel++;
    else			     aiOrder[i] = iEdgeLabel++;
  }
  for(i = 0; i < cLabel; ++i) aiInvOrder[aiOrder[i]] = i;


  ////////////////////////////////
  //  Write Header Information	//
  ////////////////////////////////

  SKELHEAD head;
  strcpy(head.szLogo, "SGFramework Mesh Skeleton File  Version 1.0\n\x1a");
  strcpy(head.szSign, SIGNATURE_SKEL);
  if (symTable->cSymbols(TYPE_COORD)  != 2)
  { strcpy(head.szCoord1, "x");
    strcpy(head.szCoord2, "y");
  }
  else
  { strcpy(head.szCoord1, symTable->Retrieve(TYPE_COORD, 0)->szName);
    strcpy(head.szCoord2, symTable->Retrieve(TYPE_COORD, 1)->szName);
  }
  head.cConstant = symTable->cSymbols(TYPE_ICONST)
                 + symTable->cSymbols(TYPE_RCONST);
  head.cLabel	 = symTable->cSymbols(TYPE_LABEL);
  head.cRefine	 = symTable->cSymbols(TYPE_REFINE);
  head.cPoint	 = symTable->cSymbols(TYPE_POINT);
  head.cEdge	 = symTable->cSymbols(TYPE_EDGE);
  head.cRegion	 = symTable->cSymbols(TYPE_REGION);
  head.cRegLabel = cRegLabel;
  head.cMinDivisions = cMinDivisions;
  head.cMaxDivisions = cMaxDivisions;
  write(nHandle, &head, sizeof(SKELHEAD));

  //////////////////////////////////
  //  Write Constant Information  //
  //////////////////////////////////

  UINT cIConst = symTable->cSymbols(TYPE_ICONST);
  UINT cRConst = symTable->cSymbols(TYPE_RCONST);
  write(nHandle, &head.cConstant, sizeof(UINT));
  XLATCONST xlatconst;
  for(i = 0; i < cIConst; ++i)
  { SymIConst *psymIConst
      = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ICONST, i), SymIConst);
    strcpy(xlatconst.szName, psymIConst->szName);
    xlatconst.uType  = TYPE_ICONST;
    xlatconst.data.n = psymIConst->n;
    write(nHandle, &xlatconst, sizeof(XLATCONST));
  }
  for(i = 0; i < cRConst; ++i)
  { SymRConst *psymRConst
      = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_RCONST, i), SymRConst);
    strcpy(xlatconst.szName, psymRConst->szName);
    xlatconst.uType  = TYPE_RCONST;
    xlatconst.data.r = psymRConst->r;
    write(nHandle, &xlatconst, sizeof(XLATCONST));
  }

  ///////////////////////////////
  //  Write Label Information  //
  ///////////////////////////////

  write(nHandle, &head.cLabel, sizeof(UINT));
  for(i = 0; i < head.cLabel; ++i)
  { SymLabel *psymLabel
      = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_LABEL, aiInvOrder[i]),
                          SymLabel);
    strcpy(xlatlabel.szName, psymLabel->szName);
    xlatlabel.iIndex = i + 1;
    xlatlabel.uType  = psymLabel->u;
    write(nHandle, &xlatlabel, sizeof(XLATLABEL));
  }

  /////////////////////////////////////////////
  //  Write Refinement Variable Information  //
  /////////////////////////////////////////////

  write(nHandle, &head.cRefine, sizeof(UINT));
  XLATREFINE xlatrefine;
  for(i = 0; i < head.cRefine; ++i)
  { SymRefine *psymRefine
      = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_REFINE, i), SymRefine);
    strcpy(xlatrefine.szName, psymRefine->szName);
    xlatrefine.iIndex = psymRefine->i;
    xlatrefine.uMeasure = psymRefine->uMeasure;
    xlatrefine.Distance = psymRefine->Distance;
    write(nHandle, &xlatrefine, sizeof(XLATREFINE));
  }

  ///////////////////////////////
  //  Write Point Information  //
  ///////////////////////////////

  write(nHandle, &head.cPoint, sizeof(UINT));
  XLATPOINT xlatpoint;
  for(i = 0; i < head.cPoint; ++i)
  { SymPoint *psymPoint
      = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_POINT, i), SymPoint);
    xlatpoint.x = psymPoint->x;
    xlatpoint.y = psymPoint->y;
    write(nHandle, &xlatpoint, sizeof(XLATPOINT));
  }

  //////////////////////////////
  //  Write Edge Information  //
  //////////////////////////////

  write(nHandle, &head.cEdge, sizeof(UINT));
  XLATEDGE xlatedge;
  for(i = 0; i < head.cEdge; ++i)
  { SymEdge *psymEdge
      = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_EDGE, i), SymEdge);
    UINT iLabel = psymEdge->iLabel;
    xlatedge.iLabel  = (iLabel) ? aiOrder[iLabel - 1] + 1 : 0;
    xlatedge.iPointA = psymEdge->iB;
    xlatedge.iPointB = psymEdge->iM;
    xlatedge.iPointC = psymEdge->iE;
    xlatedge.cSegment = psymEdge->cSegment;
    xlatedge.InitialSpacing = psymEdge->InitialSpacing;
    xlatedge.SpacingGrade = psymEdge->SpacingGrade;
    write(nHandle, &xlatedge, sizeof(XLATEDGE));
  }

  ////////////////////////////////
  //  Write Region Information	//
  ////////////////////////////////

  write(nHandle, &head.cRegion, sizeof(UINT));
  XLATREGION xlatregion;
  for(i = 0; i < head.cRegion; ++i)
  { SymRegion *psymRegion
      = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_REGION, i), SymRegion);
    xlatregion.iLabel = aiOrder[psymRegion->iLabel - 1] + 1;
    xlatregion.cEdge  = psymRegion->c;
    xlatregion.fRect  = psymRegion->fRect;
    xlatregion.Hmax   = psymRegion->Hmax;
    write(nHandle, &xlatregion, sizeof(XLATREGION));
    write(nHandle, psymRegion->au, xlatregion.cEdge * sizeof(UINT));
  }

  /////////////////////////////////
  //  Close the File and Return	 //
  /////////////////////////////////

  delete [] aiOrder;
  delete [] aiInvOrder;
  close(nHandle);
  return FILEIO_SUCCESS;
}


