/*-----------------------------------------------------------------------------
 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     gencode.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/02/94    KMK      initial implementation

 DESCRIPTION:

 This module implements the code generation functions of the FileInputOutput
 class.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "symbol.h"
#include "gencode.h"
#include "codebuf.h"
#include "fileio.h"
#include "mathnode.h"


static CodeBuffer *pcb;


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::WriteCode
// Scope:	 Public
// Description:	 This function writes C++ source code for the math node
//		 expression.
void MathNode::WriteCode (void *pv)
{
  pcb = (CodeBuffer *) pv;
  WriteCode();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::WriteCode
// Scope:	 Public
// Description:	 This function writes C++ source code for the math node
//		 expression.
void MathNode::WriteCode ()
{
  UINT i;

  switch (uNode)
  { // buffer real constant
    case TYPE_RCONST:
      pcb->AddReal(data.r, FALSE);
      break;

    // buffer integer constant
    case TYPE_ICONST:
      pcb->AddInteger(data.n, FALSE);
      break;

    // buffer coordinate
    case TYPE_COORD:
      pcb->AddCoordinate(data.coord.i, FALSE);
      break;

    // buffer internal function
    case TYPE_FUNC:
      pcb->AddFunction(data.func.iFunc, data.func.iPD, FALSE);
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      for(i = 1; i < c; ++i)
      { pcb->AddChar(',');
	apmn[i]->WriteCode();
      }
      pcb->AddChar(')');
      break;

    // buffer refinement function
    case TYPE_REFINE:
      pcb->AddEquation((UINT) data.n, FALSE);
      pcb->AddText("(_x0, _x1)");
      break;

    // buffer addition operator
    case TYPE_ADD:
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      pcb->AddChar('+');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer subtraction operator
    case TYPE_SUB:
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      pcb->AddChar('-');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer multiplication operator
    case TYPE_MUL:
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      pcb->AddChar('*');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer division operator
    case TYPE_DIV:
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      pcb->AddChar('/');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer and operator
    case TYPE_AND:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")&&(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0");
      break;

    // buffer or operator
    case TYPE_OR:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")||(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0");
      break;

    // buffer xor operator
    case TYPE_XOR:
      pcb->AddText("((((");
      apmn[0]->WriteCode();
      pcb->AddText(")&&!(");
      apmn[1]->WriteCode();
      pcb->AddText("))||(!(");
      apmn[0]->WriteCode();
      pcb->AddText(")&&(");
      apmn[1]->WriteCode();
      pcb->AddText(")))||(!(");
      if (uExpr & FLAG_REAL) pcb->AddText(")))?1.0:0.0)");
      else		     pcb->AddText(")))?1:0");
      break;

    // buffer greater than operator
    case TYPE_GT:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")>(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0");
      break;

    // buffer greater than or equal to operator
    case TYPE_GE:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")>=(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0");
      break;

    // buffer less than operator
    case TYPE_LT:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")<(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0");
      break;

    // buffer less than or equal to operator
    case TYPE_LE:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")<=(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0");
      break;

    // buffer equal to operator
    case TYPE_EQ:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")==(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0");
      break;

    // buffer not equal to operator
    case TYPE_NE:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")!=(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0");
      break;

    // buffer negation operator
    case TYPE_NEG:
      pcb->AddText("-(");
      apmn[0]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer not operator
    case TYPE_NOT:
      pcb->AddText("((");
      apmn[0]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText(")?1.0:0.0)");
      else		     pcb->AddText(")?1:0");
      break;

    // buffer real operator
    case TYPE_REAL:
      pcb->AddText("((REAL)(");
      apmn[0]->WriteCode();
      pcb->AddText("))");
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function writes C++ source code which when linked to
//		 the mesh refinement code to produce an automatic mesh
//		 refinement program.
int FileInputOutput::nWriteRefineCode ( char *pszSource )
{
  char *psz;
  UINT i;

  // form the name prefix
  if ((psz = strrchr(pszSource, '\\')) == NULL) psz = pszSource;
  else ++psz;
  strcpy(pszGlobalBuf, psz);
  if ((psz = strrchr(pszGlobalBuf, '.')) != NULL) *psz = '\0';
  for(i = strlen(pszGlobalBuf); i < LEN_PREFIX; ++i) pszGlobalBuf[i] = '_';
  pszGlobalBuf[LEN_PREFIX] = '\0';
  char szPrefix[LEN_PREFIX + 1];
  strcpy(szPrefix, pszGlobalBuf);

  // allocate the source instance of the code buffer class
  FormFileName(pszSource, FILE_REFCODE);
  CodeBuffer *pcbSource = new CodeBuffer(pszGlobalBuf, szPrefix,
                                         cmdargs.cPrecision);

  // form the grid file name
  FormFileName(pszSource, FILE_GRID);
  char szGridFileName[LEN_FILENAME];
  strcpy(szGridFileName, pszGlobalBuf);

  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("// This code was created by skeleton parser, mesh.exe, version 1.0\n");
  pcbSource->AddText("// Copyright (c) 1994 Kevin M. Kramer\n\n");
  pcbSource->AddText("#include <stdio.h>\n"
		     "#include <math.h>\n\n\n");

  // add type definitions
  pcbSource->AddText("typedef int BOOL;\n"
		     "typedef unsigned int UINT;\n"
		     "typedef double REAL;\n"
		     "typedef double (*PRF2)(double, double);\n\n");

  // add math function prototypes
  pcbSource->AddText("#ifdef abs\n");
  pcbSource->AddText("#undef abs\n");
  pcbSource->AddText("#endif\n\n");
  pcbSource->AddText("#define arcsin(x) asin(x)\n");
  pcbSource->AddText("#define arccos(x) acos(x)\n");
  pcbSource->AddText("#define arctan(x) atan(x)\n");
  pcbSource->AddText("#define abs(x)	fabs(x)\n\n");
  pcbSource->AddText("REAL sq	(REAL);\n");
  pcbSource->AddText("REAL inv	(REAL);\n");
  pcbSource->AddText("REAL coth (REAL);\n");
  pcbSource->AddText("REAL sech (REAL);\n");
  pcbSource->AddText("REAL csch (REAL);\n");
  pcbSource->AddText("REAL cot	(REAL);\n");
  pcbSource->AddText("REAL sec	(REAL);\n");
  pcbSource->AddText("REAL csc	(REAL);\n");
  pcbSource->AddText("REAL ave	(REAL, REAL);\n");
  pcbSource->AddText("REAL erf	(REAL);\n");
  pcbSource->AddText("REAL sign (REAL);\n");
  pcbSource->AddText("REAL step (REAL);\n");
  pcbSource->AddText("REAL fmin (REAL, REAL);\n");
  pcbSource->AddText("REAL fmax (REAL, REAL);\n");
  pcbSource->AddText("REAL bern (REAL);\n");
  pcbSource->AddText("REAL aux1 (REAL);\n");
  pcbSource->AddText("REAL aux2 (REAL);\n");
  pcbSource->AddText("REAL nsdep (REAL, REAL, REAL);\n");
  pcbSource->AddText("REAL ngdep (REAL, REAL, REAL, REAL, REAL);\n\n\n");

  // add refinement equations
  UINT cRefine = symTable->cSymbols(TYPE_REFINE);
  for(i = 0; i < cRefine; ++i)
  { SymRefine *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_REFINE, i),
                                        SymRefine);
    pcbSource->AddText("REAL ");
    pcbSource->AddEquation(i, TRUE);
    pcbSource->AddText("(REAL _x0, REAL _x1)\n{\n");
    pcbSource->SetLeftMargin(2);
    pcbSource->AddText("return ");
    pcbSource->SetLeftMargin(4);
    pcbSource->AddMathNode((MathNode *) psym->pv);
    pcbSource->SetLeftMargin(0);
    pcbSource->AddText(";\n}\n\n\n");
  }

  // add the minimum and maximum edge length functions
  pcbSource->AddText("REAL _MinEdgeLen (REAL _x0, REAL _x1)\n{\n");
  pcbSource->SetLeftMargin(2);
  pcbSource->AddText("return ");
  pcbSource->SetLeftMargin(4);
  if (pMinEdgeLen)
  { pcbSource->AddMathNode((MathNode *) pMinEdgeLen);
    delete (MathNode *) pMinEdgeLen;
  }
  else pcbSource->AddReal(0.0, FALSE);
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText(";\n}\n\n\n");
  pcbSource->AddText("REAL _MaxEdgeLen (REAL _x0, REAL _x1)\n{\n");
  pcbSource->SetLeftMargin(2);
  pcbSource->AddText("return ");
  pcbSource->SetLeftMargin(4);
  if (pMaxEdgeLen)
  { pcbSource->AddMathNode((MathNode *) pMaxEdgeLen);
    delete (MathNode *) pMaxEdgeLen;
  }
  else pcbSource->AddReal(1.0e300, FALSE);
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText(";\n}\n\n\n");

  pcbSource->AddText("REAL _MaxXEdgeLen (REAL _x0, REAL _x1)\n{\n");
  pcbSource->SetLeftMargin(2);
  pcbSource->AddText("return ");
  pcbSource->SetLeftMargin(4);
  if (pMaxXEdgeLen)
  { pcbSource->AddMathNode((MathNode *) pMaxXEdgeLen);
    delete (MathNode *) pMaxXEdgeLen;
  }
  else pcbSource->AddReal(1.0e300, FALSE);
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText(";\n}\n\n\n");
  pcbSource->AddText("REAL _MaxYEdgeLen (REAL _x0, REAL _x1)\n{\n");
  pcbSource->SetLeftMargin(2);
  pcbSource->AddText("return ");
  pcbSource->SetLeftMargin(4);
  if (pMaxYEdgeLen)
  { pcbSource->AddMathNode((MathNode *) pMaxYEdgeLen);
    delete (MathNode *) pMaxYEdgeLen;
  }
  else pcbSource->AddReal(1.0e300, FALSE);
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText(";\n}\n\n\n");

  // form the function table
  if (cRefine)
  { pcbSource->AddText("PRF2 apfRefine[] =\n{ ");
    pcbSource->SetLeftMargin(2);
    pcbSource->AddEquation(0, FALSE);
    for(i = 1; i < cRefine; ++i)
    { pcbSource->AddText(",\n");
      pcbSource->AddEquation(i, FALSE);
    }
    pcbSource->SetLeftMargin(0);
    pcbSource->AddText("\n};\n\n\n");
  }
  else pcbSource->AddText("PRF2 apfRefine = NULL;\n\n\n");
  pcbSource->AddText("int nRefineMesh(UINT, char *[], char *);\n\n\n");
  pcbSource->AddText("int main (UINT cArg, char *apszArg[])\n{\n");
  pcbSource->SetLeftMargin(2);
  pcbSource->AddText("return nRefineMesh(cArg, apszArg, ");
  pcbSource->AddString(szGridFileName);
  pcbSource->AddText(");\n");
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("}\n\n");

  // deallocate the source instance of the code generation class
  delete pcbSource;
  return 0;
}

