/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     grammar.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/31/94    KMK      initial implementation

 DESCRIPTION:

 This module contains the SGFramework grammar table and the error tables.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "tokens.h"
#include "states.h"
#include "actions.h"
#include "grammar.h"


///////////////////////////////////////////////////////////////////////////////
/////////////////////////////  SGFRAMEWORK GRAMMAR  ////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/*******************************************************************************

  prog	  :   // nothing
	  |   prog		     ';'
	  |   prog COORD  ID ',' ID  ';'
	  |   prog CONST  cdecs	     ';'
	  |   prog POINT  pdecs	     ';'
	  |   prog EDGE	  edecs	     ';'
	  |   prog REGION rdecs	     ';'
	  |   prog REFINE refine     ';'
	  |   prog SET	  set	     ';'
	  |   prog EOF
	  ;

********************************************************************************

  expr	  :   lexpr
	  |   expr AND lexpr
	  |   expr OR  lexpr
	  |   expr XOR lexpr
	  ;

  lexpr	  :   aexpr
	  |   lexpr '>'	 aexpr
	  |   lexpr '<'	 aexpr
	  |   lexpr '>=' aexpr
	  |   lexpr '<=' aexpr
	  |   lexpr '==' aexpr
	  |   lexpr '<>' aexpr
	  ;

  aexpr	  :   mexpr
	  |   aexpr '+' mexpr
	  |   aexpr '-' mexpr
	  ;

  mexpr	  :   uexpr
	  |   mexpr '*' uexpr
	  |   mexpr '/' uexpr
	  ;

  uexpr	  :   NUM
	  |   INT
	  |   ID
	  |   ID '(' exprs ')'
	  |   '+' uexpr
	  |   '-' uexpr
	  |   NOT uexpr
	  |   '(' expr ')'
	  |   '[' expr ']'
	  |   '{' expr '}'
	  |   REAL '(' expr ')'
	  ;

  exprs	  :   expr
	  |   exprs ',' expr
	  ;

********************************************************************************

  cdecs	  :   cdec
	  |   cdecs ',' cdec
	  ;

  cdec	  :   ID '=' expr
	  ;

********************************************************************************

  pdecs	  :   pdec
	  |   pdecs ',' pdec
	  ;

  pdec	  :   ID '=' '(' expr ',' expr ')'
	  ;

********************************************************************************

  edecs	  :   edec
	  |   edecs ',' edec
	  ;

  edec	  :   ID '=' ID '[' ID ',' ID ']'
	  |   ID '=' ID '[' ID ',' ID ']' '(' expr ',' expr ')'
	  |   ID '=' ID '[' ID ',' ID ',' ID ']'
	  |   ID '=' ID '[' ID ',' ID ',' ID ']' '(' expr ')'
	  ;

********************************************************************************

  rdecs	  :   rdec
	  |   rdecs ',' rdec
	  ;

  rlist	  :   ID
	  |   rlist ',' ID
	  ;

  rdec	  :   ID '=' ID '{' rlist '}'
	  |   ID '=' ID '{' rlist '}' '(' expr ')'
	  |   ID '=' ID '{' rlist '}' RECT
	  ;

********************************************************************************

  refine  :   label '=' expr
	  |   label '<' LINEAR ',' expr > = expr
	  |   label '<' LOG    ',' expr > = expr
	  |   label '<' SLOG   ',' expr > = expr
	  ;

********************************************************************************

  set	  :   MIN LEN '=' expr
	  |   MAX LEN '=' expr
	  |   MAX ID LEN '=' expr
	  |   MIN DIV '=' int
	  |   MAX DIV '=' int
	  ;

*******************************************************************************/


/*******************************************************************************

  EXAMPLES:

  coordinates  = label, label;

  const	 label = constant expression;

  point	 label = (constant expression, constant expression);

  edge	 label = label [point, point];

  edge	 label = label [point, point, point];

  region label = label {edge, edge, ... edge};

  refine label = expression(x,y);

*******************************************************************************/


////////////////////////////////////////////////////////////////////////////////
//////////////////////////  SGFRAMEWORK GRAMMAR TABLE  /////////////////////////
////////////////////////////////////////////////////////////////////////////////


FSM afsmTable[] = {
//	    CURRENT	       INPUT	    NEXT	STACK	      USER
// INDEX     STATE	       TOKEN	    STATE	ACTION	     ACTION
// ------  ----------	    -----------	 -----------  -----------  -----------
/* 0x0000  ST_EXPR     */ { TOK_LEXPR  , ST_EXPR    , ACT_EXPR	 , USR_	      },
/* 0x0001  ST_EXPR     */ { TOK_EXPR   , ST_EXPR1   , ACT_PUSH	 , USR_	      },
/* 0x0002  ST_EXPR     */ { DEFAULT    , ST_LEXPR   , ACT_CALL	 , USR_	      },
/* 0x0003  ST_EXPR1    */ { TOK_AND    , ST_EXPR2   , ACT_PUSH	 , USR_	      },
/* 0x0004  ST_EXPR1    */ { TOK_OR     , ST_EXPR2   , ACT_PUSH	 , USR_	      },
/* 0x0005  ST_EXPR1    */ { TOK_XOR    , ST_EXPR2   , ACT_PUSH	 , USR_	      },
/* 0x0006  ST_EXPR1    */ { DEFAULT    , ST_RETURN  , ACT_POP	 , USR_	      },
/* 0x0007  ST_EXPR2    */ { TOK_LEXPR  , ST_EXPR    , ACT_EXPR	 , USR_BINOP  },
/* 0x0008  ST_EXPR2    */ { DEFAULT    , ST_LEXPR   , ACT_CALL	 , USR_	      },
/* 0x0009  ST_LEXPR    */ { TOK_AEXPR  , ST_LEXPR   , ACT_LEXPR	 , USR_	      },
/* 0x000a  ST_LEXPR    */ { TOK_LEXPR  , ST_LEXPR1  , ACT_PUSH	 , USR_	      },
/* 0x000b  ST_LEXPR    */ { DEFAULT    , ST_AEXPR   , ACT_CALL	 , USR_	      },
/* 0x000c  ST_LEXPR1   */ { TOK_EQ     , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x000d  ST_LEXPR1   */ { TOK_NE     , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x000e  ST_LEXPR1   */ { '>'	       , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x000f  ST_LEXPR1   */ { '<'	       , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0010  ST_LEXPR1   */ { TOK_GE     , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0011  ST_LEXPR1   */ { TOK_LE     , ST_LEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0012  ST_LEXPR1   */ { DEFAULT    , ST_RETURN  , ACT_POP	 , USR_	      },
/* 0x0013  ST_LEXPR2   */ { TOK_AEXPR  , ST_LEXPR   , ACT_LEXPR	 , USR_BINOP  },
/* 0x0014  ST_LEXPR2   */ { DEFAULT    , ST_AEXPR   , ACT_CALL	 , USR_	      },
/* 0x0015  ST_AEXPR    */ { TOK_MEXPR  , ST_AEXPR   , ACT_AEXPR	 , USR_	      },
/* 0x0016  ST_AEXPR    */ { TOK_AEXPR  , ST_AEXPR1  , ACT_PUSH	 , USR_	      },
/* 0x0017  ST_AEXPR    */ { DEFAULT    , ST_MEXPR   , ACT_CALL	 , USR_	      },
/* 0x0018  ST_AEXPR1   */ { '+'	       , ST_AEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0019  ST_AEXPR1   */ { '-'	       , ST_AEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x001a  ST_AEXPR1   */ { DEFAULT    , ST_RETURN  , ACT_POP	 , USR_	      },
/* 0x001b  ST_AEXPR2   */ { TOK_MEXPR  , ST_AEXPR   , ACT_AEXPR	 , USR_BINOP  },
/* 0x001c  ST_AEXPR2   */ { DEFAULT    , ST_MEXPR   , ACT_CALL	 , USR_	      },
/* 0x001d  ST_MEXPR    */ { TOK_UEXPR  , ST_MEXPR   , ACT_MEXPR	 , USR_	      },
/* 0x001e  ST_MEXPR    */ { TOK_MEXPR  , ST_MEXPR1  , ACT_PUSH	 , USR_	      },
/* 0x001f  ST_MEXPR    */ { DEFAULT    , ST_UEXPR   , ACT_CALL	 , USR_	      },
/* 0x0020  ST_MEXPR1   */ { '*'	       , ST_MEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0021  ST_MEXPR1   */ { '/'	       , ST_MEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0022  ST_MEXPR1   */ { DEFAULT    , ST_RETURN  , ACT_POP	 , USR_	      },
/* 0x0023  ST_MEXPR2   */ { TOK_UEXPR  , ST_MEXPR   , ACT_MEXPR	 , USR_BINOP  },
/* 0x0024  ST_MEXPR2   */ { DEFAULT    , ST_UEXPR   , ACT_CALL	 , USR_	      },
/* 0x0025  ST_UEXPR    */ { TOK_NUM    , ST_RETURN  , ACT_UEXPR	 , USR_RCONST },
/* 0x0026  ST_UEXPR    */ { TOK_INT    , ST_RETURN  , ACT_UEXPR	 , USR_ICONST },
/* 0x0027  ST_UEXPR    */ { TOK_ID     , ST_UEXPR1  , ACT_PUSH	 , USR_	      },
/* 0x0028  ST_UEXPR    */ { '+'	       , ST_UEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x0029  ST_UEXPR    */ { '-'	       , ST_UEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x002a  ST_UEXPR    */ { TOK_NOT    , ST_UEXPR2  , ACT_PUSH	 , USR_	      },
/* 0x002b  ST_UEXPR    */ { '('	       , ST_UEXPR3  , ACT_DROP	 , USR_	      },
/* 0x002c  ST_UEXPR    */ { '['	       , ST_UEXPR4  , ACT_DROP	 , USR_	      },
/* 0x002d  ST_UEXPR    */ { '{'	       , ST_UEXPR5  , ACT_DROP	 , USR_	      },
/* 0x002e  ST_UEXPR    */ { TOK_REAL   , ST_UEXPR6  , ACT_PUSH	 , USR_	      },
/* 0x002f  ST_UEXPR1   */ { '('	       , ST_UEXPR1B , ACT_DROP	 , USR_	      },
/* 0x0030  ST_UEXPR1   */ { DEFAULT    , ST_UEXPR1A , ACT_POP	 , USR_	      },
/* 0x0031  ST_UEXPR1A  */ { DEFAULT    , ST_RETURN  , ACT_UEXPR	 , USR_ID     },
/* 0x0032  ST_UEXPR1B  */ { TOK_EXPRS  , ST_UEXPR1C , ACT_PUSH	 , USR_	      },
/* 0x0033  ST_UEXPR1B  */ { DEFAULT    , ST_EXPRS   , ACT_CALL	 , USR_	      },
/* 0x0034  ST_UEXPR1C  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_FUNC   },
/* 0x0035  ST_UEXPR2   */ { TOK_UEXPR  , ST_RETURN  , ACT_UEXPR	 , USR_UNARYOP},
/* 0x0036  ST_UEXPR2   */ { DEFAULT    , ST_UEXPR   , ACT_CALL	 , USR_	      },
/* 0x0037  ST_UEXPR3   */ { TOK_EXPR   , ST_UEXPR3A , ACT_PUSH	 , USR_	      },
/* 0x0038  ST_UEXPR3   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0039  ST_UEXPR3A  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_PAREN  },
/* 0x003a  ST_UEXPR4   */ { TOK_EXPR   , ST_UEXPR4A , ACT_PUSH	 , USR_	      },
/* 0x003b  ST_UEXPR4   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x003c  ST_UEXPR4A  */ { ']'	       , ST_RETURN  , ACT_UEXPR	 , USR_PAREN  },
/* 0x003d  ST_UEXPR5   */ { TOK_EXPR   , ST_UEXPR5A , ACT_PUSH	 , USR_	      },
/* 0x003e  ST_UEXPR5   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x003f  ST_UEXPR5A  */ { '}'	       , ST_RETURN  , ACT_UEXPR	 , USR_PAREN  },
/* 0x0040  ST_UEXPR6   */ { '('	       , ST_UEXPR6A , ACT_DROP	 , USR_	      },
/* 0x0041  ST_UEXPR6A  */ { TOK_EXPR   , ST_UEXPR6B , ACT_PUSH	 , USR_	      },
/* 0x0042  ST_UEXPR6A  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0043  ST_UEXPR6B  */ { ')'	       , ST_RETURN  , ACT_UEXPR	 , USR_REAL   },
/* 0x0044  ST_EXPRS    */ { DEFAULT    , ST_EXPRS1  , ACT_	 , USR_EXPRS  },
/* 0x0045  ST_EXPRS1   */ { TOK_EXPR   , ST_EXPRS2  , ACT_PUSH	 , USR_ADDEXPR},
/* 0x0046  ST_EXPRS1   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0047  ST_EXPRS2   */ { ','	       , ST_EXPRS1  , ACT_DROP	 , USR_	      },
/* 0x0048  ST_EXPRS2   */ { DEFAULT    , ST_EXPRS3  , ACT_POP	 , USR_	      },
/* 0x0049  ST_EXPRS3   */ { DEFAULT    , ST_RETURN  , ACT_EXPRS	 , USR_	      },
/* 0x004a  ST_COORD    */ { TOK_ID     , ST_COORD1  , ACT_PUSH	 , USR_	      },
/* 0x004b  ST_COORD1   */ { ','	       , ST_COORD2  , ACT_DROP	 , USR_	      },
/* 0x004c  ST_COORD2   */ { TOK_ID     , ST_COORD3  , ACT_PUSH	 , USR_	      },
/* 0x004d  ST_COORD3   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_DCOORD },
/* 0x004e  ST_CDECS    */ { TOK_ID     , ST_CDECS1  , ACT_PUSH	 , USR_	      },
/* 0x004f  ST_CDECS1   */ { '='	       , ST_CDECS2  , ACT_DROP	 , USR_	      },
/* 0x0050  ST_CDECS2   */ { TOK_EXPR   , ST_CDECS3  , ACT_PUSH	 , USR_DCONST },
/* 0x0051  ST_CDECS2   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0052  ST_CDECS3   */ { ','	       , ST_CDECS   , ACT_DROP	 , USR_	      },
/* 0x0053  ST_CDECS3   */ { ';'	       , ST_CDECS4  , ACT_DROP	 , USR_	      },
/* 0x0054  ST_CDECS4   */ { DEFAULT    , ST_PROG    , ACT_	 , USR_	      },
/* 0x0055  ST_PDECS    */ { TOK_ID     , ST_PDECS1  , ACT_PUSH	 , USR_	      },
/* 0x0056  ST_PDECS1   */ { '='	       , ST_PDECS2  , ACT_DROP	 , USR_	      },
/* 0x0057  ST_PDECS2   */ { '('	       , ST_PDECS3  , ACT_DROP	 , USR_	      },
/* 0x0058  ST_PDECS3   */ { TOK_EXPR   , ST_PDECS4  , ACT_PUSH	 , USR_	      },
/* 0x0059  ST_PDECS3   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x005a  ST_PDECS4   */ { ','	       , ST_PDECS5  , ACT_DROP	 , USR_	      },
/* 0x005b  ST_PDECS5   */ { TOK_EXPR   , ST_PDECS6  , ACT_PUSH	 , USR_	      },
/* 0x005c  ST_PDECS5   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x005d  ST_PDECS6   */ { ')'	       , ST_PDECS7  , ACT_DROP	 , USR_DPOINT },
/* 0x005e  ST_PDECS7   */ { ','	       , ST_PDECS   , ACT_DROP	 , USR_	      },
/* 0x005f  ST_PDECS7   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_	      },
/* 0x0060  ST_EDECS    */ { TOK_ID     , ST_EDECS1  , ACT_PUSH	 , USR_	      },
/* 0x0061  ST_EDECS1   */ { '='	       , ST_EDECS2  , ACT_DROP	 , USR_	      },
/* 0x0062  ST_EDECS2   */ { TOK_ID     , ST_EDECS2A , ACT_PUSH	 , USR_LABEL  },
/* 0x0063  ST_EDECS2   */ { '['	       , ST_EDECS3  , ACT_DROP	 , USR_NOLABEL},
/* 0x0064  ST_EDECS2A  */ { '['	       , ST_EDECS3  , ACT_DROP	 , USR_	      },
/* 0x0065  ST_EDECS3   */ { TOK_ID     , ST_EDECS4  , ACT_PUSH	 , USR_	      },
/* 0x0066  ST_EDECS4   */ { ','	       , ST_EDECS5  , ACT_DROP	 , USR_	      },
/* 0x0067  ST_EDECS5   */ { TOK_ID     , ST_EDECS6  , ACT_PUSH	 , USR_	      },
/* 0x0068  ST_EDECS6   */ { ','	       , ST_EDECS12 , ACT_DROP	 , USR_	      },
/* 0x0069  ST_EDECS6   */ { ']'	       , ST_EDECS7  , ACT_DROP	 , USR_	      },
/* 0x006a  ST_EDECS7   */ { '('	       , ST_EDECS8  , ACT_DROP	 , USR_	      },
/* 0x006b  ST_EDECS7   */ { ','	       , ST_EDECS   , ACT_DROP	 , USR_DEDGE1 },
/* 0x006c  ST_EDECS7   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_DEDGE1 },
/* 0x006d  ST_EDECS8   */ { TOK_EXPR   , ST_EDECS9  , ACT_PUSH	 , USR_	      },
/* 0x006e  ST_EDECS8   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x006f  ST_EDECS9   */ { ','	       , ST_EDECS10 , ACT_DROP	 , USR_	      },
/* 0x0070  ST_EDECS10  */ { TOK_EXPR   , ST_EDECS11 , ACT_PUSH	 , USR_	      },
/* 0x0071  ST_EDECS10  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x0072  ST_EDECS11  */ { ')'	       , ST_EDECS17 , ACT_DROP	 , USR_DEDGE2 },
/* 0x0073  ST_EDECS12  */ { TOK_ID     , ST_EDECS13 , ACT_PUSH	 , USR_	      },
/* 0x0074  ST_EDECS13  */ { ']'	       , ST_EDECS14 , ACT_DROP	 , USR_	      },
/* 0x0075  ST_EDECS14  */ { '('	       , ST_EDECS15 , ACT_DROP	 , USR_	      },
/* 0x0076  ST_EDECS14  */ { ','	       , ST_EDECS   , ACT_DROP	 , USR_DEDGE3 },
/* 0x0077  ST_EDECS14  */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_DEDGE3 },
/* 0x0078  ST_EDECS15  */ { TOK_EXPR   , ST_EDECS16 , ACT_PUSH	 , USR_	      },
/* 0x0079  ST_EDECS15  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x007a  ST_EDECS16  */ { ')'	       , ST_EDECS17 , ACT_DROP	 , USR_DEDGE4 },
/* 0x007b  ST_EDECS17  */ { ','	       , ST_EDECS   , ACT_DROP	 , USR_	      },
/* 0x007c  ST_EDECS17  */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_	      },
/* 0x007d  ST_RLIST    */ { DEFAULT    , ST_RLIST1  , ACT_	 , USR_RLIST  },
/* 0x007e  ST_RLIST1   */ { TOK_ID     , ST_RLIST2  , ACT_PUSH	 , USR_ADDID  },
/* 0x007f  ST_RLIST2   */ { ','	       , ST_RLIST1  , ACT_DROP	 , USR_	      },
/* 0x0080  ST_RLIST2   */ { DEFAULT    , ST_RLIST3  , ACT_POP	 , USR_	      },
/* 0x0081  ST_RLIST3   */ { DEFAULT    , ST_RETURN  , ACT_RLIST	 , USR_	      },
/* 0x0082  ST_RDECS    */ { TOK_ID     , ST_RDECS1  , ACT_PUSH	 , USR_	      },
/* 0x0083  ST_RDECS1   */ { '='	       , ST_RDECS2  , ACT_DROP	 , USR_	      },
/* 0x0084  ST_RDECS2   */ { TOK_ID     , ST_RDECS3  , ACT_PUSH	 , USR_	      },
/* 0x0085  ST_RDECS3   */ { '{'	       , ST_RDECS4  , ACT_DROP	 , USR_	      },
/* 0x0086  ST_RDECS4   */ { TOK_RLIST  , ST_RDECS5  , ACT_PUSH	 , USR_	      },
/* 0x0087  ST_RDECS4   */ { DEFAULT    , ST_RLIST   , ACT_CALL	 , USR_	      },
/* 0x0088  ST_RDECS5   */ { '}'	       , ST_RDECS6  , ACT_DROP	 , USR_	      },
/* 0x0089  ST_RDECS6   */ { TOK_RECT   , ST_RDECS9  , ACT_DROP	 , USR_DREG2  },
/* 0x008a  ST_RDECS6   */ { '('	       , ST_RDECS7  , ACT_PUSH	 , USR_	      },
/* 0x008b  ST_RDECS6   */ { ','	       , ST_RDECS   , ACT_DROP	 , USR_DREG1  },
/* 0x008c  ST_RDECS6   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_DREG1  },
/* 0x008d  ST_RDECS7   */ { TOK_EXPR   , ST_RDECS8  , ACT_PUSH	 , USR_	      },
/* 0x008e  ST_RDECS7   */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x008f  ST_RDECS8   */ { ')'	       , ST_RDECS9  , ACT_DROP	 , USR_DREG3  },
/* 0x0090  ST_RDECS9   */ { ','	       , ST_RDECS   , ACT_DROP	 , USR_	      },
/* 0x0091  ST_RDECS9   */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_	      },
/* 0x0092  ST_REFINE   */ { TOK_ID     , ST_REFINE1 , ACT_PUSH	 , USR_REFDEF },
/* 0x0093  ST_REFINE1  */ { '='	       , ST_REFINE2 , ACT_DROP	 , USR_	      },
/* 0x0094  ST_REFINE1  */ { '('	       , ST_REFINE1A, ACT_DROP	 , USR_	      },
/* 0x0095  ST_REFINE1A */ { TOK_LINEAR , ST_REFINE1B, ACT_PUSH	 , USR_REFMEAS},
/* 0x0096  ST_REFINE1A */ { TOK_LOG    , ST_REFINE1B, ACT_PUSH	 , USR_REFMEAS},
/* 0x0097  ST_REFINE1A */ { TOK_SLOG   , ST_REFINE1B, ACT_PUSH	 , USR_REFMEAS},
/* 0x0098  ST_REFINE1B */ { ','	       , ST_REFINE1C, ACT_DROP	 , USR_	      },
/* 0x0099  ST_REFINE1C */ { TOK_EXPR   , ST_REFINE1D, ACT_PUSH	 , USR_REFDIST},
/* 0x009a  ST_REFINE1C */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x009b  ST_REFINE1D */ { ')'	       , ST_REFINE1E, ACT_DROP	 , USR_	      },
/* 0x009c  ST_REFINE1E */ { '='	       , ST_REFINE2 , ACT_DROP	 , USR_	      },
/* 0x009d  ST_REFINE2  */ { TOK_EXPR   , ST_REFINE3 , ACT_PUSH	 , USR_	      },
/* 0x009e  ST_REFINE2  */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x009f  ST_REFINE3  */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_REFINE },
/* 0x00a0  ST_SET      */ { TOK_MIN    , ST_SET1    , ACT_PUSH	 , USR_	      },
/* 0x00a1  ST_SET      */ { TOK_MAX    , ST_SETA    , ACT_PUSH	 , USR_	      },
/* 0x00a2  ST_SETA     */ { TOK_LEN    , ST_SET2    , ACT_PUSH	 , USR_	      },
/* 0x00a3  ST_SETA     */ { TOK_DIV    , ST_SET2    , ACT_PUSH	 , USR_	      },
/* 0x00a1  ST_SETA     */ { TOK_ID     , ST_SETB    , ACT_PUSH	 , USR_	      },
/* 0x00a2  ST_SETB     */ { TOK_LEN    , ST_SETC    , ACT_PUSH	 , USR_	      },
/* 0x00a4  ST_SETC     */ { '='	       , ST_SETD    , ACT_DROP	 , USR_	      },
/* 0x00a5  ST_SETD     */ { TOK_EXPR   , ST_SET4    , ACT_PUSH	 , USR_SET2   },
/* 0x00a6  ST_SETD     */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x00a2  ST_SET1     */ { TOK_LEN    , ST_SET2    , ACT_PUSH	 , USR_	      },
/* 0x00a3  ST_SET1     */ { TOK_DIV    , ST_SET2    , ACT_PUSH	 , USR_	      },
/* 0x00a4  ST_SET2     */ { '='	       , ST_SET3    , ACT_DROP	 , USR_	      },
/* 0x00a5  ST_SET3     */ { TOK_EXPR   , ST_SET4    , ACT_PUSH	 , USR_SET1   },
/* 0x00a6  ST_SET3     */ { DEFAULT    , ST_EXPR    , ACT_CALL	 , USR_	      },
/* 0x00a7  ST_SET4     */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_	      },
/* 0x00a8  ST_PROG     */ { ';'	       , ST_PROG    , ACT_DROP	 , USR_	      },
/* 0x00a9  ST_PROG     */ { TOK_COORD  , ST_COORD   , ACT_PUSH	 , USR_	      },
/* 0x00aa  ST_PROG     */ { TOK_CONST  , ST_CDECS   , ACT_DROP	 , USR_	      },
/* 0x00ab  ST_PROG     */ { TOK_POINT  , ST_PDECS   , ACT_DROP	 , USR_	      },
/* 0x00ac  ST_PROG     */ { TOK_EDGE   , ST_EDECS   , ACT_DROP	 , USR_	      },
/* 0x00ad  ST_PROG     */ { TOK_REGION , ST_RDECS   , ACT_DROP	 , USR_	      },
/* 0x00ae  ST_PROG     */ { TOK_REFINE , ST_REFINE  , ACT_DROP	 , USR_	      },
/* 0x00af  ST_PROG     */ { TOK_SET    , ST_SET	    , ACT_DROP	 , USR_	      },
/* 0x00b0  ST_PROG     */ { TOK_EOF    , ST_RETURN  , ACT_	 , USR_	      },
/* 0x00b1  ST_START    */ { DEFAULT    , ST_PROG    , ACT_	 , USR_INIT   }
};


#if 0
/* 0x0000  ST_	       */ {	       , ST_	    , ACT_	 , USR_	      },
/* 0x0000  ST_	       */ {	       , ST_	    , ACT_	 , USR_	      },
/* 0x0000  ST_	       */ {	       , ST_	    , ACT_	 , USR_	      },
/* 0x0000  ST_	       */ {	       , ST_	    , ACT_	 , USR_	      },
#endif


////////////////////////////////////////////////////////////////////////////////
///////////////////////	 SGFRAMEWORK GRAMMAR TABLE INDEX  //////////////////////
////////////////////////////////////////////////////////////////////////////////


UINT ifsmTable[] = {
/* 0x0000 */  0x0000, 0x0003, 0x0007, 0x0009, 0x000c, 0x0013, 0x0015, 0x0018,
/* 0x0008 */  0x001b, 0x001d, 0x0020, 0x0023, 0x0025, 0x002f, 0x0031, 0x0032,
/* 0x0010 */  0x0034, 0x0035, 0x0037, 0x0039, 0x003a, 0x003c, 0x003d, 0x003f,
/* 0x0018 */  0x0040, 0x0041, 0x0043, 0x0044, 0x0045, 0x0047, 0x0049, 0x004a,
/* 0x0020 */  0x004b, 0x004c, 0x004d, 0x004e, 0x004f, 0x0050, 0x0052, 0x0054,
/* 0x0028 */  0x0055, 0x0056, 0x0057, 0x0058, 0x005a, 0x005b, 0x005d, 0x005e,
/* 0x0030 */  0x0060, 0x0061, 0x0062, 0x0064, 0x0065, 0x0066, 0x0067, 0x0068,
/* 0x0038 */  0x006a, 0x006d, 0x006f, 0x0070, 0x0072, 0x0073, 0x0074, 0x0075,
/* 0x0040 */  0x0078, 0x007a, 0x007b, 0x007d, 0x007e, 0x007f, 0x0081, 0x0082,
/* 0x0048 */  0x0083, 0x0084, 0x0085, 0x0086, 0x0088, 0x0089, 0x008d, 0x008f,
/* 0x0050 */  0x0090, 0x0092, 0x0093, 0x0095, 0x0098, 0x0099, 0x009b, 0x009c,
/* 0x0058 */  0x009d, 0x009f, 0x00a0, 0x00a2, 0x00a5, 0x00a6, 0x00a7, 0x00a9,
/* 0x0060 */  0x00ab, 0x00ac, 0x00ae, 0x00af, 0x00b8, 0x00b9
};


////////////////////////////////////////////////////////////////////////////////
//////////  SGFRAMEWORK SYNTAX ERROR -- EXPECTED TOKEN(S) TABLE INDEX  /////////
////////////////////////////////////////////////////////////////////////////////


char TOKEXP_LPARENTH[] = "'('";
char TOKEXP_RPARENTH[] = "')'";
char TOKEXP_LBRACKET[] = "'['";
char TOKEXP_RBRACKET[] = "']'";
char TOKEXP_LBRACE[]   = "'{'";
char TOKEXP_RBRACE[]   = "'}'";
char TOKEXP_EQUAL[]    = "'='";
char TOKEXP_SEMI[]     = "';'";
char TOKEXP_COMMA[]    = "','";
char TOKEXP_TERM1[]    = "',' or ';'";
char TOKEXP_TERM2[]    = "',' or ']'";
char TOKEXP_TERM3[]    = "',', ';' or '(' ";
char TOKEXP_TERM4[]    = "',', ';', '(' or keyword \"rectangles\"";
char TOKEXP_MINMAX[]   = "keyword \"minimum\" or \"maximum\"";
char TOKEXP_LENDIV[]   = "keyword \"length\" or \"division\"";
char TOKEXP_LENDIVID[] = "keyword \"length\" or \"division\" or identifier";
char TOKEXP_LEN[]      = "keyword \"length\"";
char TOKEXP_REFINE1[]  = "'(' or '='";
char TOKEXP_REFINE2[]  = "keyword \"linear\", \"logarithmic\" or \"signedlog\"";
char TOKEXP_ELABEL[]   = "'[' or edge label";
char TOKEXP_RLABEL[]   = "'{' or region label";
char TOKEXP_ID[]       = "identifier";
char TOKEXP_LABEL[]    = "identifier or '['";
char TOKEXP_UEXPR[]    = "unary expression";
char TOKEXP_PROG[]     = "program statement";


SEE aseeTable[] =
{ { ST_UEXPR,	  TOKEXP_UEXPR	  },
  { ST_UEXPR1C,	  TOKEXP_RPARENTH },
  { ST_UEXPR3A,	  TOKEXP_RPARENTH },
  { ST_UEXPR4A,	  TOKEXP_RBRACKET },
  { ST_UEXPR5A,	  TOKEXP_RBRACE	  },
  { ST_UEXPR6,	  TOKEXP_LPARENTH },
  { ST_UEXPR6B,	  TOKEXP_RPARENTH },
  { ST_COORD,	  TOKEXP_ID	  },
  { ST_COORD1,	  TOKEXP_COMMA	  },
  { ST_COORD2,	  TOKEXP_ID	  },
  { ST_COORD3,	  TOKEXP_SEMI	  },
  { ST_CDECS,	  TOKEXP_ID	  },
  { ST_CDECS1,	  TOKEXP_EQUAL	  },
  { ST_CDECS3,	  TOKEXP_TERM1	  },
  { ST_PDECS,	  TOKEXP_ID	  },
  { ST_PDECS1,	  TOKEXP_EQUAL	  },
  { ST_PDECS2,	  TOKEXP_LPARENTH },
  { ST_PDECS4,	  TOKEXP_COMMA	  },
  { ST_PDECS6,	  TOKEXP_RPARENTH },
  { ST_PDECS7,	  TOKEXP_TERM1	  },
  { ST_EDECS,	  TOKEXP_ID	  },
  { ST_EDECS1,	  TOKEXP_EQUAL	  },
  { ST_EDECS2,	  TOKEXP_ID	  },
  { ST_EDECS2A,	  TOKEXP_LABEL	  },
  { ST_EDECS3,	  TOKEXP_ID	  },
  { ST_EDECS4,	  TOKEXP_COMMA	  },
  { ST_EDECS5,	  TOKEXP_ID	  },
  { ST_EDECS6,	  TOKEXP_TERM2	  },
  { ST_EDECS7,	  TOKEXP_TERM3	  },
  { ST_EDECS9,	  TOKEXP_COMMA	  },
  { ST_EDECS11,	  TOKEXP_RPARENTH },
  { ST_EDECS12,	  TOKEXP_ID	  },
  { ST_EDECS13,	  TOKEXP_RBRACKET },
  { ST_EDECS14,	  TOKEXP_TERM3	  },
  { ST_EDECS16,	  TOKEXP_RPARENTH },
  { ST_EDECS17,	  TOKEXP_TERM1	  },
  { ST_RLIST1,	  TOKEXP_ID	  },
  { ST_RDECS,	  TOKEXP_ID	  },
  { ST_RDECS1,	  TOKEXP_EQUAL	  },
  { ST_RDECS2,	  TOKEXP_ID	  },
  { ST_RDECS3,	  TOKEXP_LBRACE	  },
  { ST_RDECS5,	  TOKEXP_RBRACE	  },
  { ST_RDECS6,	  TOKEXP_TERM4	  },
  { ST_RDECS8,	  TOKEXP_RPARENTH },
  { ST_RDECS9,	  TOKEXP_TERM1	  },
  { ST_REFINE,	  TOKEXP_ID	  },
  { ST_REFINE1,	  TOKEXP_REFINE1  },
  { ST_REFINE1A,  TOKEXP_REFINE2  },
  { ST_REFINE1B,  TOKEXP_COMMA	  },
  { ST_REFINE1D,  TOKEXP_RPARENTH },
  { ST_REFINE1E,  TOKEXP_EQUAL	  },
  { ST_REFINE3,	  TOKEXP_SEMI	  },
  { ST_SET,	  TOKEXP_MINMAX	  },
  { ST_SETA,	  TOKEXP_LENDIVID },
  { ST_SETB,	  TOKEXP_LEN	  },
  { ST_SETC,	  TOKEXP_EQUAL	  },
  { ST_SET1,	  TOKEXP_LENDIV	  },
  { ST_SET2,	  TOKEXP_EQUAL	  },
  { ST_SET4,	  TOKEXP_SEMI	  },
  { ST_PROG,	  TOKEXP_PROG	  },
  { -1u,	  NULL		  }
};


