/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     hash.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/31/94    KMK      initial implementation

 DESCRIPTION:

 This module implements a hash table.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "hash.h"



///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE HASH NODE CLASS  //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Description:
HashNode::HashNode ( char *pszName, UINT uNodeType )
{
  strcpy(szName, pszName);
  uType = uNodeType;
  phnode = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Description: Virtual, Destructor
HashNode::~HashNode ( )
{
}


///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE HASH TABLE CLASS  /////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Description:
HashTable::HashTable ( UINT c )
{
  UINT i;

  // allocate the symbol table
  cSize = c;
  aphnode = new HashNode*[c];
  for(i = 0; i < c; ++i) aphnode[i] = NULL;

  // initialize the character to key alphabet translation table
  cAlphabet = 0;
  for(i = 0  ; i < 'z' + 1; ++i)	      auXlat[i] = 0;
  for(i = '0'; i <= '9'	  ; ++i, ++cAlphabet) auXlat[i] = cAlphabet;
  for(i = 'A'; i <= 'Z'	  ; ++i, ++cAlphabet) auXlat[i] = cAlphabet;
  for(i = 'a'; i <= 'z'	  ; ++i, ++cAlphabet) auXlat[i] = cAlphabet;
}


///////////////////////////////////////////////////////////////////////////////
// Description: Virtual, Destructor
HashTable::~HashTable ( )
{
  for(UINT i = 0; i < cSize; ++i)
  { HashNode *phnode1, *phnode2;
    phnode1 = aphnode[i];
    while (phnode1 != NULL)
    { phnode2 = phnode1->phnodeNext();
      delete phnode1;
      phnode1 = phnode2;
    }
  }
  delete [] aphnode;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function returns the hash value of the pszName key
UINT HashTable::uHash ( char *pszName )
{
  UINT uHashValue;
  char *psz = pszName;
  for (uHashValue = 0; *psz; ++psz)
    uHashValue = (cAlphabet * uHashValue + auXlat[*psz]) % cSize;
  return uHashValue;
}

///////////////////////////////////////////////////////////////////////////////
// Description:	 This function stores a hash node in the hash table.
void HashTable::Store ( HashNode *phnode )
{
  UINT uHashValue = uHash(phnode->szName);
  phnode->Link(aphnode[uHashValue]);
  aphnode[uHashValue] = phnode;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function removes a hash node from the hash table.
void HashTable::Remove ( char *pszName )
{
  UINT uHashValue = uHash(pszName);
  HashNode *phnode = aphnode[uHashValue];
  if (phnode == NULL) return;
  if (!strcmp(phnode->szName, pszName))
  { aphnode[uHashValue] = phnode->phnodeNext();
    delete phnode;
    return;
  }
  HashNode *phnodeT;
  while ((phnodeT = phnode->phnodeNext()) != NULL)
  { if (!strcmp(phnodeT->szName, pszName))
    { phnode->Link(phnodeT->phnodeNext());
      delete phnodeT;
      return;
    }
    else phnode = phnode->phnodeNext();
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function retrieves a hash node in the hash table.
HashNode * HashTable::Retrieve ( char *pszName )
{
  UINT uHashValue = uHash(pszName);
  HashNode *phnode = aphnode[uHashValue];
  while (phnode)
    if (!strcmp(phnode->szName, pszName)) return phnode;
    else phnode = phnode->phnodeNext();
  return NULL;
}


///////////////////////////////////////////////////////////////////////////////
////////////  DYNAMIC ARRAY OF POINTERS TO INSTANCES OF HASHNODES  ////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Description:
DAPHashNode::DAPHashNode ( UINT c )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPHashNode::DAPHashNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  cElements = 0;
  fManaged = FALSE;
  dynarray = (HashNode**) calloc((cSize = uInc = c), sizeof(HashNode*));
}


///////////////////////////////////////////////////////////////////////////////
// Description:
DAPHashNode::~DAPHashNode ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPHashNode::~DAPHashNode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // if the array is managed then deallocate the stored objects
  if (fManaged)
    for (UINT i = 0; i < cElements; ++i)
      delete dynarray[i];

  free((void *) dynarray);
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
UINT DAPHashNode::Add ( HashNode *t )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPHashNode::Add\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // increase the array size if necessary
  if (cElements == cSize)
  { cSize += uInc;
    dynarray = (HashNode **) realloc(dynarray, cSize * sizeof(HashNode *));
  }

  dynarray[cElements] = t;
  return cElements++;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function removes the last element from the dynarmic array
void DAPHashNode::RemoveLast ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPHashNode::RemoveLast\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  --cElements;

  // if the array is managed then deallocate the stored object
  if (fManaged) delete dynarray[cElements];
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function removes an element from the dynarmic array
void DAPHashNode::Remove ( UINT iRemove )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPHashNode::Remove\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  --cElements;
  if (fManaged) delete dynarray[iRemove];
  for(UINT i = iRemove; i < cElements; ++i)
    dynarray[i] = dynarray[i+1];
  dynarray[cElements] = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
void DAPHashNode::Flush ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPHashNode::Flush\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // if the array is managed then deallocate the stored objects
  if (fManaged)
    for(UINT i = 0; i < cElements; ++i)
      delete dynarray[i];

  // reset the element count
  cElements = 0;
}


