/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     mathnode.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/19/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the SGFramework expression tree consisting of a n-ary
 tree of math nodes.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#if !defined(UNIX)
#include <mem.h>
#endif

#include "types.h"
#include "error.h"
#include "symbol.h"
#include "mathnode.h"


///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE MATH NODE CLASS  //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Description:	 Default math node constructor
MathNode::MathNode ( )
{
  uNode	   = -1u;
  uExpr	   =  0u;
  iPtrNode = -1u;
  iPtrExpr = -1u;
  c	   =  0u;
  apmn	   = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates an integer constant math node
MathNode::MathNode ( int n, UINT i )
{
  uNode = TYPE_ICONST;
  uExpr = FLAG_CONST | FLAG_INT;
  iPtrNode = i;
  iPtrExpr = i;
  c = 0;
  apmn = NULL;
  data.n = n;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a real constant math node
MathNode::MathNode ( REAL r, UINT i )
{
  uNode = TYPE_RCONST;
  uExpr = FLAG_CONST | FLAG_REAL;
  iPtrNode = i;
  iPtrExpr = i;
  c = 0;
  apmn = NULL;
  data.r = r;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a coordinate math node
MathNode::MathNode ( UINT iIndex, UINT i )
{
  uNode = TYPE_COORD;
  uExpr	 = FLAG_VAR | FLAG_REAL;
  iPtrNode = i;
  iPtrExpr = i;
  c = 0;
  apmn = NULL;
  data.coord.i = iIndex;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a unary operator math node expression
MathNode::MathNode ( UINT uType, MathNode *pmn, UINT i )
{
  // initialize the generic fields
  uExpr = pmn->uExpr;
  iPtrNode = i;
  iPtrExpr = i;
  c = 1;
  apmn = (MathNode**) calloc(1, sizeof(MathNode*));
  apmn[0] = pmn;

  switch (uNode = uType)
  { case TYPE_NEG:
      SimplifyNeg();
      break;

    case TYPE_NOT:
      SimplifyNot();
      break;

    case TYPE_REAL:
      uExpr &= ~FLAG_INT;
      uExpr |=	FLAG_REAL;
      SimplifyReal();
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 Creates a binary operator math node expression
MathNode::MathNode ( UINT uType, MathNode *pmn1, MathNode *pmn2, UINT i )
{
  // initialize the generic fields
  uNode = uType;
  UINT uExpr1 = pmn1->uExpr;
  UINT uExpr2 = pmn2->uExpr;
  BOOL fInt1  = uExpr1 & FLAG_INT;
  BOOL fInt2  = uExpr2 & FLAG_INT;
  uExpr	 = ((uExpr1 & FLAG_VAR) || (uExpr2 & FLAG_VAR)) ? FLAG_VAR : FLAG_CONST;
  uExpr |= (fInt1 && fInt2) ? FLAG_INT : FLAG_REAL;
  iPtrNode = i;
  iPtrExpr = pmn1->iPtrNode;
  c = 2;
  apmn = (MathNode**) calloc(2, sizeof(MathNode*));
  BOOL fReal = uExpr & FLAG_REAL;
  apmn[0] = (fReal && fInt1) ? new MathNode(TYPE_REAL, pmn1, i) : pmn1;
  apmn[1] = (fReal && fInt2) ? new MathNode(TYPE_REAL, pmn2, i) : pmn2;

  // operator specific code
  switch (uNode = uType)
  { case TYPE_ADD:  SimplifyAdd();  break;
    case TYPE_SUB:  SimplifySub();  break;
    case TYPE_MUL:  SimplifyMul();  break;
    case TYPE_DIV:  SimplifyDiv();  break;
    case TYPE_AND:  SimplifyAnd();  break;
    case TYPE_OR:   SimplifyOr();   break;
    case TYPE_XOR:  SimplifyXor();  break;
    case TYPE_GT:   SimplifyGT();   break;
    case TYPE_GE:   SimplifyGE();   break;
    case TYPE_LT:   SimplifyLT();   break;
    case TYPE_LE:   SimplifyLE();   break;
    case TYPE_EQ:   SimplifyEQ();   break;
    case TYPE_NE:   SimplifyNE();   break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::~MathNode
// Scope:	 Public, Virtual, Destructor
MathNode::~MathNode ()
{
  for(UINT i = 0; i < c; ++i) if (apmn[i]) delete apmn[i];
  if (apmn) free((void *) apmn);
}

///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::AddChild
// Scope:	 Public
// Description:	 This function adds a child to the math node.
void MathNode::AddChild (MathNode *pmn)
{
  UINT i = c++;
  apmn = (MathNode **) realloc(apmn, c * sizeof(MathNode *));
  apmn[i] = pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::KillChildren
// Scope:	 Protected
// Description:	 This function kills (deallocates) the children expression
//		 trees.
void MathNode::KillChildren ()
{
  for(UINT i = 0; i < c; ++i) if (apmn[i]) delete apmn[i];
  if (apmn) { free((void *) apmn); apmn = NULL; }
  c = 0;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::NameHeir
// Scope:	 Protected
// Description:	 This function promotes the iHeir child to the parent node and
//		 deallocates the heir's siblings.
void MathNode::NameHeir (UINT iHeir)
{
  // delete every child except the heir
  for(UINT i = 0; i < c; ++i)
    if (i != iHeir)
      delete apmn[i];

  // replace the parent w/ the heir
  MathNode *pmn = apmn[iHeir];
  free((void *) apmn);
  *this = *pmn;

  // in order to delete the memory allocated for the heir, we need to change
  // the number of children it has so that the delete operator does not delete
  // the heir's information which is now stored in the parent node.
  pmn->c = 0;
  pmn->apmn = NULL;
  delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::NameHeir
// Scope:	 Protected
// Description:	 This function applies the following transformation to math
//		 expression trees.  If fLeft is TRUE then the left cycle is
//		 performed; otherwise, the right cycle is performed.
//
//	    A		     B		       A		B
//	   / \	   left	     |		      / \     right	|
//	  B   Y	   ---->     A	     OR	     X	 B    ---->	A
//	  |	   cycle    / \			 |    cycle    / \
//	  X		   X   Y		 Y	      X	  Y
//
// NOTE:  A and B are either operators or function math nodes.
//	  X and Y are arbitrary math trees.
void MathNode::Cycle (BOOL fLeft)
{
  MathNode *pmn1, *pmn2;
  MathNode *pmnX, *pmnY;
  char buf[sizeof(MathNode)];

  if (fLeft)
  { pmn1 = this;
    pmn2 = apmn[0];
    pmnX = apmn[0]->apmn[0];
    pmnY = apmn[1];
  }
  else
  { pmn1 = this;
    pmn2 = apmn[1];
    pmnX = apmn[0];
    pmnY = apmn[1]->apmn[0];
  }

  memcpy(buf,  pmn1, sizeof(MathNode)); // note, before this swap
  memcpy(pmn1, pmn2, sizeof(MathNode)); // pmn1 points to mnA and
  memcpy(pmn2, buf,  sizeof(MathNode)); // pmn2 points to mnB ...

  pmn1->apmn[0] = pmn2;			// now pmn1 points to mnB
  pmn2->apmn[0] = pmnX;			// and pmn2 points to mnA
  pmn2->apmn[1] = pmnY;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::SwapChildren
// Scope:	 Protected
// Description:	 This function swaps the i-th and j-th children
void MathNode::SwapChildren (UINT i, UINT j)
{
  MathNode *pmn;
  pmn = apmn[i];
  apmn[i] = apmn[j];
  apmn[j] = pmn;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::Compare
// Scope:	 Protected
// Description:	 This function compares the expression trees.
BOOL MathNode::Compare (MathNode *pmn)
{
  UINT i;

  if (uNode != pmn->uNode) return FALSE;
  switch (uNode)
  { case TYPE_RCONST:
      return data.r == pmn->data.r;

    case TYPE_ICONST:
      return data.n == pmn->data.n;

    case TYPE_COORD:
      return data.coord.i == pmn->data.coord.i;

    case TYPE_FUNC:
      if (data.func.iFunc != pmn->data.func.iFunc) return FALSE;
      if (data.func.iPD	  != pmn->data.func.iPD)   return FALSE;
      for(i = 0; i < c; ++i)
	if (!apmn[i]->Compare(pmn->apmn[i]))
	  return FALSE;
      return TRUE;

    case TYPE_NEG:
    case TYPE_NOT:
    case TYPE_REAL:
      return apmn[0]->Compare(pmn->apmn[0]);

    case TYPE_ADD:
    case TYPE_MUL:
    case TYPE_AND:
    case TYPE_OR:
    case TYPE_XOR:
    case TYPE_EQ:
    case TYPE_NE:
      return
	(apmn[0]->Compare(pmn->apmn[0]) && apmn[1]->Compare(pmn->apmn[1])) ||
	(apmn[0]->Compare(pmn->apmn[1]) && apmn[1]->Compare(pmn->apmn[0]));

    case TYPE_SUB:
    case TYPE_DIV:
    case TYPE_GT:
    case TYPE_GE:
    case TYPE_LT:
    case TYPE_LE:
      return apmn[0]->Compare(pmn->apmn[0]) && apmn[1]->Compare(pmn->apmn[1]);
  }
  return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::Copy
// Scope:	 Public
// Description:	 This function copies the expression tree.
MathNode *MathNode::Copy ()
{
  // make a new math node and copy the parent node
  MathNode *pmn = new MathNode();
  *pmn = *this;

  // if the parent node has children, then copy the children
  if (c)
  { pmn->apmn = new MathNode*[c];
    for(UINT i = 0; i < c; ++i)
      pmn->apmn[i] = apmn[i]->Copy();
  }

  // return the copy
  return pmn;
}



