/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     mathnode.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/01/94    KMK      initial implementation

 DESCRIPTION:

 This module implements the SGFramework expression tree consisting of a n-ary
 tree of math nodes.

-----------------------------------------------------------------------------*/


#ifndef __mathnode_h
#define __mathnode_h

#ifndef __types_h
#include "types.h"
#endif

#ifndef __mathfunc_h
#include "mathfunc.h"
#endif

// Hate, like acid, damages the vessel in which it is stored and the object
// on which it is poured -- Richard W. De Haan


typedef struct tagICONST		// iconst
{ int  i;				// integer constant's value
} ICONST;

typedef struct tagRCONST		// rconst
{ REAL r;				// real constant's value
} SCONST;

typedef struct tagCOORD			// coord
{ UINT i;				// coordinate index
} COORD;

typedef struct tagFUNC			// func
{ UINT iFunc;				// function index
  UINT iPD;				// partial derivitive index
  PRF  aprf[3];				// pointer to function and derivatives
} FUNC;

typedef union tagMNDATA
{ int	    n;
  UINT	    u;
  REAL	    r;
  FUNC	    func;
  COORD	    coord;
} MNDATA;


class MathNode
{ protected:
    void KillChildren ();		// deallocate children trees
    void NameHeir (UINT);		// name heir
    void Cycle (BOOL);			// cycle math node operators
    void SwapChildren (UINT, UINT);	// swap children trees
    BOOL Compare (MathNode *);		// compare two expressions
    void WriteCode ();			// convert the expression to C++ code

    inline BOOL IsZero	   () { return ((uNode == TYPE_RCONST) && (data.r ==  0.0)) || ((uNode == TYPE_ICONST) && (data.n ==  0)); }
    inline BOOL IsOne	   () { return ((uNode == TYPE_RCONST) && (data.r ==  1.0)) || ((uNode == TYPE_ICONST) && (data.n ==  1)); }
    inline BOOL IsMinusOne () { return ((uNode == TYPE_RCONST) && (data.r == -1.0)) || ((uNode == TYPE_ICONST) && (data.n == -1)); }
    inline BOOL IsNeg	   () { return	 uNode == TYPE_NEG; }
    inline BOOL IsInv	   () { return	(uNode == TYPE_FUNC) && (data.func.iFunc == FUNC_INV); }
    inline BOOL Is   (UINT u) { return	 uNode == u; }

    void MakeZero ();			// change math node to 0.0
    void MakeOne  ();			// change math node to 1.0
    void MakeInv  ();			// make inverse function
    void MakeSq	  ();			// make square function

    void ValidateInteger (long int);	// determines if long will fit in int
    void SimplifyAdd ();		// simplify addition expressions
    void SimplifySub ();		// simplify subtration expressions
    void SimplifyMul ();		// simplify multiplication expressions
    void SimplifyDiv ();		// simplify division expressions
    void SimplifyNeg ();		// simplify negation expressions
    void SimplifyNot ();		// simplify not expressions
    void SimplifyReal ();		// simplify convert to real expressions
    void SimplifyPow ();		// simplify pow function expressions
    void SimplifyAnd ();		// simplify and expressions
    void SimplifyOr  ();		// simplify or expressions
    void SimplifyXor ();		// simplify exclusive or expressions
    void SimplifyGT  ();		// simplify > expressions
    void SimplifyGE  ();		// simplify >= expressions
    void SimplifyLT  ();		// simplify < expressions
    void SimplifyLE  ();		// simplify <= expressions
    void SimplifyEQ  ();		// simplify == expressions
    void SimplifyNE  ();		// simplify <> expressions

  public:
    void SimplifyFunc  ();		// simplify functions

  public:
    UINT uNode;				// node type
    UINT uExpr;				// expression type
    UINT iPtrNode;			// buffer index of math node
    UINT iPtrExpr;			// buffer index of expression
    UINT c;				// number of children
    MathNode **apmn;			// array of pointers to children
    MNDATA data;			// type specific information

    MathNode *Copy ();			// copy math node expression
    void AddChild(MathNode *);		// add a child to the parent
    void WriteCode (void *);		// convert the expression to C++ code


  // MathNode Constructors
  // 1.	 unknown math node
  // 2.	 integer constant math node
  // 3.	 real constant math node
  // 4.	 coordinate math node
  // 5.	 unary operator math node expression
  // 6.	 binary operator math node epression
  public:
  /* 1 */  MathNode ();
  /* 2 */  MathNode (int, UINT);
  /* 3 */  MathNode (double, UINT);
  /* 4 */  MathNode (UINT, UINT);
  /* 5 */  MathNode (UINT, MathNode *, UINT);
  /* 6 */  MathNode (UINT, MathNode *, MathNode *, UINT);
	  ~MathNode ();
};


#endif

