/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     scanner.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/31/94    KMK      initial implementation

 DESCRIPTION:

 This module implements the Translate::uNextToken and Translate::uKeyWordSearch
 functions which implement the SGFramework scanner.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include <limits.h>
#include "tokens.h"
#include "error.h"
#include "source.h"


#define STATE_START   0x00
#define STATE_ID      0x10
#define STATE_NUM1    0x20
#define STATE_NUM2    0x21
#define STATE_NUM3    0x22
#define STATE_NUM4    0x23
#define STATE_REAL    0x24
#define STATE_INT     0x25
#define STATE_COMM1   0x30
#define STATE_COMM2   0x31
#define STATE_STRING  0x40
#define STATE_RANGE1  0x50
#define STATE_RANGE2  0x51
#define STATE_RANGE3  0x52
#define STATE_ARROW   0x60
#define STATE_LT      0x61
#define STATE_GT      0x62
#define STATE_EQUAL   0x63
#define STATE_EOF     0x70
#define STATE_ERROR1  0xF0
#define STATE_ERROR2  0xF1
#define STATE_ERROR3  0xF2

#define IsAlpha(x) (((x >= 'A') && (x <= 'Z')) || ((x >= 'a') && (x <= 'z')))
#define IsDigit(x)  ((x >= '0') && (x <= '9'))
#define IsSpace(x)  ((x == ' ') || (x == '\t') || (x == '\r'))
#define IsE(x)	    ((x == 'e') || (x == 'E'))
#define IsLF(x)	    (x == '\n')
#define IsSlash(x)  (x == '/')
#define IsQuote(x)  (x == '\"')
#define IsDash(x)   (x == '-')
#define IsEOF(x)    ((x == '\0') || (x == '\x1a'))
#define IsPoint(x)  (x == '.')
#define IsPlus(x)   (x == '+')
#define IsGT(x)	    (x == '>')
#define IsLT(x)	    (x == '<')
#define IsEqual(x)  (x == '=')


typedef struct tagKEYWORD		// keyword
{   char szIdent[16];			// identifier
    UINT uTokenValue;			// value returned to parser
} KEYWORD;


// Since the SimGewn language is currently under developement, its keywords
// are not stored in a hash table.  Instead the keywords are alphabetically
// stored in the following table.  A binary search is performed to locate
// the keywords.
static KEYWORD akeyword[] =
{ { "accuracy",	      TOK_ACC	    },
  { "algorithm",      TOK_ALGOR	    },
  { "all",	      TOK_ALL	    },
  { "and",	      TOK_AND	    },
  { "append",	      TOK_APPEND    },
  { "assign",	      TOK_ASSIGN    },
  { "begin",	      TOK_BEGIN	    },
  { "call",	      TOK_CALL	    },
  { "close",	      TOK_CLOSE	    },
  { "comment",	      TOK_COMMENT   },
  { "const",	      TOK_CONST	    },
  { "coordinates",    TOK_COORD	    },
  { "damping",	      TOK_DAMP	    },
  { "diagonal",	      TOK_DIAG	    },
  { "divisions",      TOK_DIV	    },
  { "do",	      TOK_DO	    },
  { "edge",	      TOK_EDGE	    },
  { "else",	      TOK_ELSE	    },
  { "end",	      TOK_END	    },
  { "exit",	      TOK_EXIT	    },
  { "equ",	      TOK_EQU	    },
  { "file",	      TOK_FILE	    },
  { "fill",	      TOK_FILL	    },
  { "func",	      TOK_FUNC	    },
  { "gausselim",      TOK_GAUSS	    },
  { "goto",	      TOK_GOTO	    },
  { "if",	      TOK_IF	    },
  { "ilu",	      TOK_ILU	    },
  { "infinity",	      TOK_INFIN	    },
  { "iterations",     TOK_ITERS	    },
  { "known",	      TOK_KNOWN	    },
  { "label",	      TOK_LABEL	    },
  { "length",	      TOK_LEN	    },
  { "linear",	      TOK_LINEAR    },
  { "linsol",	      TOK_LINSOL    },
  { "logarithmic",    TOK_LOG	    },
  { "main",	      TOK_MAIN	    },
  { "maximum",	      TOK_MAX	    },
  { "minimum",	      TOK_MIN	    },
  { "newton",	      TOK_NEWTON    },
  { "no",	      TOK_NO	    },
  { "none",	      TOK_NONE	    },
  { "not",	      TOK_NOT	    },
  { "open",	      TOK_OPEN	    },
  { "or",	      TOK_OR	    },
  { "over",	      TOK_OVER	    },
  { "pcg",	      TOK_PCG	    },
  { "point",	      TOK_POINT	    },
  { "preconditioner", TOK_PREC	    },
  { "read",	      TOK_READ	    },
  { "real",	      TOK_REAL	    },
  { "rectangles",     TOK_RECT	    },
  { "refine",	      TOK_REFINE    },
  { "region",	      TOK_REGION    },
  { "return",	      TOK_RETURN    },
  { "set",	      TOK_SET	    },
  { "signedlog",      TOK_SLOG	    },
  { "solve",	      TOK_SOLVE	    },
  { "sor",	      TOK_SOR	    },
  { "sort",	      TOK_SORT	    },
  { "unknown",	      TOK_UNKNOWN   },
  { "var",	      TOK_VAR	    },
  { "while",	      TOK_WHILE	    },
  { "write",	      TOK_WRITE	    },
  { "xor",	      TOK_XOR	    },
  { "yes",	      TOK_YES	    },
  { "zero",	      TOK_ZERO	    }
};
const UINT TABLE_ENTRIES = sizeof(akeyword) / sizeof(KEYWORD);


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function determines if psz is a keyword.	If psz is a
//               keyword then its token value is returned; otherwise,
//               zero is returned.
UINT Source::uKeyWordSearch ( char *psz )
{
  int nCompare;				// result of stricmp
  int iMiddle;				// binary search indices
  int iFirst = 0;
  int iLast  = TABLE_ENTRIES - 1;

  while (iLast >= iFirst)
  { iMiddle = (iFirst + iLast) / 2;
    nCompare = stricmp(psz, akeyword[iMiddle].szIdent);
    if	    (nCompare > 0) iFirst = iMiddle + 1;
    else if (nCompare < 0) iLast  = iMiddle - 1;
    else break;
  }
  return (iLast >= iFirst) ? akeyword[iMiddle].uTokenValue : 0;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function returns a token to the scanner.
void Source::PutToken ( TOKEN &token )
{
  stackInput->push(token);
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function implements the SGFramework language scanner.
//
UINT Source::GetToken ( TOKEN& token )
{
  if (!stackInput->empty())
  { token = stackInput->pop();
    return token.uTokenType;
  }

  char *psz;
  UINT	uState = STATE_START;
  for(;;)
  { char ch = pszBuf[iPtr];
    switch (uState)
    { case STATE_START:
	iRef = iPtr++; // start to recognize a new token
	if	(IsAlpha(ch))  { uState = STATE_ID;   ptt->Store(ch); }
	else if (IsDigit(ch))  { uState = STATE_NUM1; ptt->Store(ch); }
	else if (IsSpace(ch))  { /* discard */		}
	else if (IsLF(ch))     { /* discard */		}
	else if (IsSlash(ch))  { uState = STATE_COMM1;	}
	else if (IsQuote(ch))  { uState = STATE_STRING; }
	else if (IsDash(ch))   { uState = STATE_ARROW;	}
	else if (IsPoint(ch))  { uState = STATE_RANGE1; }
	else if (IsLT(ch))     { uState = STATE_LT;	}
	else if (IsGT(ch))     { uState = STATE_GT;	}
	else if (IsEqual(ch))  { uState = STATE_EQUAL;	}
	else if (IsEOF(ch))    { uState = STATE_EOF; --iPtr; }
	else
	{ token.iPtr = iRef;
	  return token.uTokenType = ch;
	}
	break;

      case STATE_ID:
	if (IsAlpha(ch) || IsDigit(ch)) { ptt->Store(ch); ++iPtr; }
	else
	{ token.iPtr = iRef;
	  UINT uTokenValue = uKeyWordSearch(ptt->GetStr());
	  if (uTokenValue) token.uTokenType = uTokenValue;
	  else
	  { strncpy(token.data.sz, ptt->GetStr(), LEN_IDENT);
	    token.data.sz[LEN_IDENT] = '\0';
	    token.uTokenType = TOK_ID;
	  }
	  ptt->Flush();
	  return token.uTokenType;
	}
	break;

      case STATE_NUM1:
	if	(IsDigit(ch)) {			       ptt->Store(ch); ++iPtr; }
	else if (IsPoint(ch)) { uState = STATE_RANGE2; ptt->Store(ch); ++iPtr; }
	else if (IsE(ch))     { uState = STATE_NUM3;   ptt->Store(ch); ++iPtr; }
	else		      { uState = STATE_INT; }
	break;

      case STATE_NUM2:
	if	(IsDigit(ch))  {		       ptt->Store(ch); ++iPtr; }
	else if (IsE(ch))      { uState = STATE_NUM3;  ptt->Store(ch); ++iPtr; }
	else		       { uState = STATE_REAL;  }
	break;

      case STATE_NUM3:
	if	(IsDigit(ch))  { uState = STATE_NUM4;  ptt->Store(ch); ++iPtr; }
	else if (IsDash(ch))   { uState = STATE_NUM4;  ptt->Store(ch); ++iPtr; }
	else if (IsPlus(ch))   { uState = STATE_NUM4;		       ++iPtr; }
	else		       { uState = STATE_REAL;  }
	break;

      case STATE_NUM4:
	if (IsDigit(ch))       { ptt->Store(ch); ++iPtr; }
	else		       { uState = STATE_REAL;	 }
	break;

      case STATE_REAL:
	token.iPtr   = iRef;
	token.data.r = atof(ptt->GetStr());
	ptt->Flush();
	return token.uTokenType = TOK_NUM;

      case STATE_INT:
	//  NOTE:  if the string does not "fit" in an integer, atoi does not
	//  return 0 like it says it should ... hence, atof is used to check
	//  the integer
	if (atof(ptt->GetStr()) > (REAL) INT_MAX)
	{ uState = STATE_REAL;
	  break;
	}

	token.iPtr = iRef;
	token.data.n = atoi(ptt->GetStr());
	ptt->Flush();
	return token.uTokenType = TOK_INT;

      case STATE_COMM1:
	if (IsSlash(ch)) { uState = STATE_COMM2; ++iPtr; }
	else
	{ token.iPtr = iRef;
	  return token.uTokenType = '/';
	}
	break;

      case STATE_COMM2:
	if	(IsLF(ch))  { uState = STATE_START; ++iPtr; }
	else if (IsEOF(ch)) { uState = STATE_START; ++iPtr; }
	else		    {			    ++iPtr; }
	break;

      case STATE_STRING:
	if	(IsLF(ch))  { uState = STATE_ERROR1; }
	else if (IsEOF(ch)) { uState = STATE_ERROR2; }
	else if (IsQuote(ch))
	{ ++iPtr;
	  token.iPtr = iRef;
	  psz = new char[ptt->GetLen() + 1];
	  strcpy(psz, ptt->GetStr());
	  token.data.p = (void *) psz;
	  ptt->Flush();
	  return token.uTokenType = TOK_STRING;
	}
	else
	{ ++iPtr;
	  ptt->Store(ch);
	}
	break;

      case STATE_RANGE1:
	if (IsPoint(ch))
	{ ++iPtr;
	  token.iPtr = iRef;
	  return token.uTokenType = TOK_RANGE;
	}
	else
	{ token.iPtr = iRef;
	  return token.uTokenType = '.';
	}

      case STATE_RANGE2:
	if	(IsDigit(ch)) { uState = STATE_NUM2;   ptt->Store(ch); ++iPtr; }
	else if (IsE(ch))     { uState = STATE_NUM3;   ptt->Store(ch); ++iPtr; }
	else if (IsPoint(ch)) { uState = STATE_RANGE3;		       ++iPtr; }
	else		      { uState = STATE_REAL;  }
	break;

      case STATE_RANGE3:
	ptt->Pop();
	token.iPtr = iPtr - 2;
	token.uTokenType = TOK_RANGE;
	stackInput->push(token);
	uState = STATE_INT;
	break;

      case STATE_ARROW:
	if (IsGT(ch))
	{ ++iPtr;
	  token.iPtr = iRef;
	  return token.uTokenType = TOK_ARROW;
	}
	else
	{ token.iPtr = iRef;
	  return token.uTokenType = '-';
	}

      case STATE_LT:
	if (IsEqual(ch))
	{ ++iPtr;
	  token.iPtr = iRef;
	  return token.uTokenType = TOK_LE;
	}
	else if (IsGT(ch))
	{ ++iPtr;
	  token.iPtr = iRef;
	  return token.uTokenType = TOK_NE;
	}
	else
	{ token.iPtr = iRef;
	  return token.uTokenType = '<';
	}

      case STATE_GT:
	if (IsEqual(ch))
	{ ++iPtr;
	  token.iPtr = iRef;
	  return token.uTokenType = TOK_GE;
	}
	else
	{ token.iPtr = iRef;
	  return token.uTokenType = '>';
	}

      case STATE_EQUAL:
	if (IsEqual(ch))
	{ ++iPtr;
	  token.iPtr = iRef;
	  return token.uTokenType = TOK_EQ;
	}
	else
	{ token.iPtr = iRef;
	  return token.uTokenType = '=';
	}

      case STATE_EOF:
	token.iPtr = iPtr;
	return token.uTokenType = TOK_EOF;

      case STATE_ERROR1:
	error->AddMessage(EHM_STRLF, iRef);
	uState = STATE_ERROR3;
	break;

      case STATE_ERROR2:
	error->AddMessage(EHM_UNEXPEOF, iPtr - 1);
	ptt->Flush();
	uState = STATE_EOF;
	break;

      case STATE_ERROR3:
	if	(IsEOF(ch)) { uState = STATE_ERROR2; }
	else if (IsQuote(ch))
	{ token.iPtr = iRef;
	  token.data.p = NULL;
	  ptt->Flush();
	  return token.uTokenType = TOK_STRING;
	}
	else ++iPtr;
    }
  }
}

