/*-----------------------------------------------------------------------------

 PROJECT:    SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:       source.h
 AUTHOR:     Kevin M. Kramer

 REVISION  HISTORY
 DATE      INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/01/94    KMK      initial implementation

 DESCRIPTION:

 This module implements a scanner and a parser.

-----------------------------------------------------------------------------*/


#ifndef __source_h
#define __source_h


typedef struct tagTOKEN     // token ('=' can do byte-by-byte copy)
{ UINT iPtr;                // buffer index of token
  UINT uTokenType;          // token type
  union
  { char  sz[LEN_IDENT+1];  // identifier
    int   n;                // integer
    UINT  u;                // unsigned integer
    REAL  r;                // real number
    void *p;                // pointer
  } data;
} TOKEN;
typedef TOKEN *PTOKEN;      // ptoken


class DAPTOKEN
{ private:
    UINT cSize;         // total number of elements
    UINT uInc;          // increment size
    UINT cElements;     // number of elements in array
    TOKEN **dynarray;   // dynamic array
    void IncSize ()     { if (cElements == cSize)
                          { cSize += uInc;
                            dynarray = (TOKEN **) realloc(dynarray, cSize*sizeof(TOKEN*));
                          }
                        }

  public:
    UINT GetElements ()        { return cElements;    }
    UINT Add (TOKEN *t)        { IncSize(); dynarray[cElements] = t;
                                 return cElements++;  }
    void RemoveLast ()         { --cElements;         }
    void Flush ()              { cElements = 0;       }
    TOKEN *operator[] (UINT i) { return dynarray[i];  }

  public:
     DAPTOKEN (UINT c = 16) { cElements = 0; uInc = cSize = c;
                              dynarray = (TOKEN**) calloc(c, sizeof(TOKEN*)); }
    ~DAPTOKEN ()            { for(UINT i = 0; i < cElements; ++i)
                                delete [] dynarray[i]; /* ?fengj */
                              free((void *) dynarray);
                            }
};


// The TokenTank class implements a character buffer.  Instances of this class
// should not be copied or assigned to each other.
//   Store() adds a character to the end of the buffer
//   Pop() removes a character from the buffer
//   Flush() removes all characters from the buffer
//   GetLen() returns the number of characters in the buffer
//   GetChar() returns the first character in the buffer
//   GetStr() returns the character string in the buffer
class TokenTank
{ private:
    char *pszBuf;
    UINT  i;
    UINT  c;

  public:
    void  Store   (char ch)    { pszBuf[i++] = ch; }    /* @@@ overflow check */
    char  Pop     ()           { return pszBuf[--i]; }  /* @@@ overflow check */
    void  Flush   ()           { i = 0; }
    UINT  GetLen  ()           { return i; }
    char  GetChar (UINT i = 0) { return pszBuf[i]; }
    char *GetStr  ()           { pszBuf[i] = '\0'; return pszBuf; }

  public:
     TokenTank (UINT Size)     { pszBuf = new char[Size]; c = Size; i = 0; }
    ~TokenTank ()              { delete [] pszBuf; }
};


class UINTStack
{ private:
    UINT u;      // increment
    UINT c;      // stack size
    UINT i;      // stack index
    UINT *stack; // stack
    void grow () { c += u; stack = (UINT*) realloc(stack, c * sizeof(UINT)); }

  public:
    inline void push  (UINT T) { if (i == c) grow(); stack[i++] = T; }
    inline UINT pop   ()       { return stack[--i]; }
    inline BOOL empty ()       { return !i; }
    inline void flush ()       { i = 0; }

  public:
     UINTStack (UINT n = 64)   { i = 0; stack = (UINT*) calloc((c = u = n),
                                                sizeof(UINT)); }
    ~UINTStack ()              { free((void *) stack); }
};


class TokenStack
{ private:
    UINT u;       // increment
    UINT c;       // stack size
    UINT i;       // stack index
    TOKEN *stack; // stack
    void grow ()  { c+= u;
                    stack = (TOKEN*) realloc(stack, c * sizeof(TOKEN));
                  }

  public:
    inline void push  (TOKEN T) { if (i == c) grow(); stack[i++] = T; }
    inline TOKEN pop  ()        { return stack[--i]; }
    inline BOOL empty ()        { return !i; }
    inline void flush ()        { i = 0; }

  public:
     TokenStack (UINT n = 64)   { i = 0; stack = (TOKEN*) calloc((c = u = n),
                                                 sizeof(TOKEN)); }
    ~TokenStack ()              { free((void *) stack); }
};


// Dynamic Array of UINT (DAUINT)
#ifndef __da_uint_h
#include "da_uint.h"
#endif


class Source
{ // source file buffer data and member functions
  protected:
    UINT cLines;                        // number of lines in buffer
    UINT cChars;                        // number of characters in buffer
    DAUINT DALineIndex;                 // line index array
    char szFileName[LEN_FILENAME];      // file name
    char *pszBuf;                       // source code buffer
    void ReadSourceFile ();             // read source file into buffer

  public:
    UINT GetLineFromPos (UINT);         // get line number from position
    UINT GetLineIndex  (UINT);          // get starting index of line

  protected:
    // scanner data and member functions
    UINT iPtr;                          // current location in buffer
    UINT iRef;                          // reference pointer (for a new token)
    UINT iLine;                         // current line number
    TokenTank *ptt;                     // token tank ... used by scanner
    TokenStack *stackInput;             // input token stack
    UINT uKeyWordSearch (char *psz);    // determine if ID is a keyword
    UINT GetToken (TOKEN& token);       // get the next token
    void PutToken (TOKEN& token);       // return a token to the scanner

    // parser data and member functions
    UINT uState;                        // current parser state
    UINT uLastState;                    // last parser state
    UINT uSyntaxState;                  // syntax error recovery state
    BOOL fLabelPresent;                 // edge or region has label?
    TOKEN token;                        // current input token
    TokenStack *stackToken;             // token stack
    UINTStack *stackState;              // state stack
    void UserActions (UINT UserAction); // user action handler
    void SyntaxError ();                // syntax error handler
    UINT uParse ();                     // parse the specification file

  public:
    void Translate ();                  // translate specification file

  public:
     Source (char *);
    ~Source ();
};

#endif

