/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     symbol.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/31/94    KMK      initial implementation

 DESCRIPTION:

 This module implements the SGFramework symbol table.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "types.h"
#include "symbol.h"
#include "mathfunc.h"
#include "mathnode.h"


// global declaration
SymbolTable *symTable = NULL;


///////////////////////////////////////////////////////////////////////////////
///////////////////  MEMBER FUNCTIONS OF THE SYMBOL CLASSES  //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Description:
SymIConst::SymIConst ( char *pszName, int nValue )
	 : HashNode  ( pszName, TYPE_ICONST )
{
  n = nValue;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymRConst::SymRConst ( char *pszName, REAL rValue )
	 : HashNode  ( pszName, TYPE_RCONST )
{
  r = rValue;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymCoord::SymCoord ( char *psz, UINT iCoord )
	: HashNode ( psz, TYPE_COORD )
{
  i = iCoord;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymLabel::SymLabel ( char *psz, UINT iLabel, UINT uType )
	: HashNode ( psz, TYPE_LABEL )
{
  i = iLabel;
  u = uType;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymPoint::SymPoint ( char *psz, UINT iPoint, REAL rX, REAL rY )
	: HashNode ( psz, TYPE_POINT )
{
  i = iPoint;
  x = rX;
  y = rY;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymEdge::SymEdge  ( char *psz, UINT iEdge, UINT iL, UINT iPntA, UINT iPntB,
		    UINT iPntC, UINT c, REAL InitSpace, REAL SpaceGrade )
       : HashNode ( psz, TYPE_EDGE )
{
  i = iEdge;
  iLabel = iL;
  iB = iPntA;
  iM = iPntB;
  iE = iPntC;
  cSegment = c;
  InitialSpacing = InitSpace;
  SpacingGrade = SpaceGrade;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymRegion::SymRegion ( char *psz, UINT iRegion, UINT iL, UINT cEdges,
		       UINT *auEdges, BOOL f, REAL r )
	 : HashNode  ( psz, TYPE_REGION )
{
  i = iRegion;
  iLabel = iL;
  c = cEdges;
  au = auEdges;
  fRect = f;
  Hmax = r;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymRegion::~SymRegion ( )
{
  if (au) delete [] au;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymRefine::SymRefine ( char *psz, UINT iRefine, void *pvExpr, UINT u, REAL r)
	 : HashNode  ( psz, TYPE_REFINE )
{
  i = iRefine;
  pv = pvExpr;
  uMeasure = u;
  Distance = r;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymRefine::~SymRefine ( )
{
  if (pv) delete (MathNode *) pv;
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymFunc::SymFunc  ( char *pszName, UINT iFunc, UINT cArgs, PRF *aprfList )
       : HashNode ( pszName, TYPE_FUNC )
{
  i = iFunc;
  c = cArgs;
  aprf[0] = aprfList[0];
  aprf[1] = aprfList[1];
  aprf[2] = aprfList[2];
}


///////////////////////////////////////////////////////////////////////////////
/////////////////  MEMBER FUNCTIONS OF THE SYMBOL TABLE CLASS  ////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Description:
SymbolTable::SymbolTable ( UINT c, BOOL fAddFuncs ) : HashTable ( c )
{
  for (UINT i = 0; i < SYMBOLS; ++i)
    DAP[i]  = new DAPHashNode;
  if (fAddFuncs) AddFunctions();
}


///////////////////////////////////////////////////////////////////////////////
// Description:
SymbolTable::~SymbolTable ( )
{
  for (UINT i = 0; i < SYMBOLS; ++i)
    delete DAP[i];
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function adds the built-in functions to the symbol table.
void SymbolTable::AddFunctions ( )
{
  for(UINT i = 0; i < cFuncs; ++i)
  { SymFunc *psym = new SymFunc(aift[i].sz, i, aift[i].c, aift[i].aprf);
    Store(psym);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function stores a symbol in the symbol table.
void SymbolTable::Store ( HashNode *phnode )
{
  DAP[phnode->uType]->Add(phnode);
  HashTable::Store(phnode);
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function removes the symbols of type uType from the
//		 symbol table.
void SymbolTable::RemoveSymbols ( UINT uType )
{
  UINT c = DAP[uType]->GetElements();
  for(UINT i = 0; i < c; ++i)
  { char *psz = (*DAP[uType])[i]->szName;
    Remove(psz);
  }
  DAP[uType]->Flush();
}


