/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     symbol.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/31/94    KMK      initial implementation

 DESCRIPTION:

 This module implements the SGFramework symbol table.

-----------------------------------------------------------------------------*/


#ifndef __symbol_h
#define __symbol_h

#ifndef __types_h
#include "types.h"
#endif

#ifndef __hash_h
#include "hash.h"
#endif


// Integer Constant Symbol
class SymIConst : public HashNode
{ public:
    int n;				// constant's value
    SymIConst (char *, int);
};


// Real Constant Symbol
class SymRConst : public HashNode
{ public:
    REAL r;				// constant's value
    SymRConst (char *, REAL);
};


// Coordinate Symbol
class SymCoord : public HashNode
{ public:
    UINT i;				// coordinate index
    SymCoord (char *, UINT);
};


// Label Symbol
class SymLabel : public HashNode
{ public:
    UINT i;				// label index
    UINT u;				// edge or region label
    SymLabel (char *, UINT, UINT);
};


// Point Symbol
class SymPoint : public HashNode
{ public:
    UINT i;				// point index
    REAL x;				// x coordinate
    REAL y;				// y coordinate
    SymPoint (char *, UINT, REAL, REAL);
};


// Edge Symbol
class SymEdge : public HashNode
{ public:
    UINT i;				// edge index
    UINT iLabel;			// edge label
    UINT iB;				// beginning point
    UINT iM;				// midpoint of curve
    UINT iE;				// ending point
    UINT cSegment;			// number of segments
    REAL InitialSpacing;		// initial spacing
    REAL SpacingGrade;			// spacing grade
    SymEdge (char *, UINT, UINT, UINT, UINT, UINT, UINT, REAL, REAL);
};


// Region Symbol
class SymRegion : public HashNode
{ public:
    UINT i;				// region index
    UINT iLabel;			// region label
    UINT c;				// number of edges
    UINT *au;				// array of edges
    BOOL fRect;				// rectangular region
    REAL Hmax;				// maximum edge spacing
    SymRegion (char *, UINT, UINT, UINT, UINT *, BOOL, REAL);
   ~SymRegion ();
};


// Refine Symbol
class SymRefine : public HashNode
{ public:
    UINT i;				// refine index
    void *pv;				// refine expression
    UINT uMeasure;			// refine measurement
    REAL Distance;			// refine distance
    SymRefine (char *, UINT, void *, UINT, REAL);
   ~SymRefine ();
};


// Function Symbol
class SymFunc : public HashNode
{ public:
    UINT i;				// function's index
    UINT c;				// number of arguments
    PRF	 aprf[3];			// pointer to function and derivatives
    SymFunc (char *, UINT, UINT, PRF *);
};


// This class implements the SGFramework symbol table.
//   Store() stores a symbol in the symbol table
//   Retrieve() retrieves a symbol from the symbol table
//   cSymbols() returns the number of specified tokens in the symbol table
class SymbolTable : public HashTable
{ protected:
    DAPHashNode *DAP[SYMBOLS];
    void  AddFunctions ();

  public:
    void Store (HashNode *);
    void RemoveSymbols (UINT);
    inline void FlushSymbolList (UINT u)    { DAP[u]->Flush(); }
    inline UINT cSymbols	(UINT u)    { return DAP[u]->GetElements(); }
    inline HashNode * Retrieve	(char *psz) { return HashTable::Retrieve(psz); }
    inline HashNode * Retrieve	(UINT u, UINT i) { return (*DAP[u])[i]; }

  public:
    SymbolTable (UINT = 397, BOOL = FALSE);
   ~SymbolTable ();
};


// global reference to the symbol table instance
extern SymbolTable *symTable;


#endif




