/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  refine.exe
 FILE:	     refine.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/04/94    KMK      initial implementation
 12/27/04    gdiso    add cgns output
 DESCRIPTION:

 This module contains the refine class.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"
#include "cgnslib.h"

#ifndef UNIX
#pragma hdrstop
#endif

#if defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#include "xgraph.h"
#else
#include <io.h>
#include <sys\stat.h>
#include "wgraph.h"
#endif

#include <fcntl.h>
#include "mathfunc.h"
#include "refine.h"
#include "cmdparse.h"


extern PRF2 apfRefine[];

char szGlobalBuf[GLOBAL_BUFSIZE];
CMDARGS cmdargs;

#if defined(DEBUG)
char szDEBUG[80];
BOOL fDEBUGDumpMesh = FALSE;
UINT cDEBUGDivision = 0;
UINT iDEBUGStartDump = -1u;
#endif

static char *apszHelp[] =
{ "syntax:  *_ref [options]					* = default",
  "  -h		display this help screen",
  "  -Oxxx	create ASCII output file named xxx",
  "  -c         dump cgns file", 
  "  -cyl	cylindrical geometry",
  "  -dmin###	minimum number of divisions",
  "  -dmax###	maximum number of divisions",
  "  -dadj###	divide triangles of level < ### if adj. triangle will be split",
  "  -rm?=lin	use linear measurement for refinement variable ?",
  "  -rm?=log	use logarithmic measurement for refinement variable ?",
  "  -rm?=slog	use signed logarithmic measurement for refinement variable ?",
  "  -rd?=###	maximum delta for refinement variable ?",
  "  -qe###	number of edge enhancement iterations (default 10)",
  "  -qn###	number of node enhancement iterations (default 10)",
  "  -qb###	number of edge followed by node enhancement iterations (default 10)",
  "  -m###	minimum rectangular splitting angle in degrees (default 0)",
  "  -g###	maximum grade between rectangular elements (default infinite)",
  "  -s###	maximum ratio between length and width (default infinite)",
  "  -aspect	keep x and y aspect ratio",
#if 0
  "  -vc	visualize mesh construction",
#endif
  "  -vr      * visualize mesh rating",
  "  -vn	no graphics",
  "  -yz	do not apply refinement statements to elements above y = 0",
  "  -ps	generate a postscript file named mesh.ps",
  "  -bw	generate grayscale plot",
  "  -r1	graphics screen resolution 640 x 480",
  "  -r2      * graphics screen resolution 800 x 600",
  "  -r3	graphics screen resolution 1024 x 768",
  NULL
};


#define EDGESTATUS_BOUNDARY 0
#define EDGESTATUS_EXTERIOR 1
#define EDGESTATUS_INTERIOR 2

static Refine *prefine	= NULL;
static UINT *auStatus	= NULL;
static UINT cBndryNode	= 0;

void RedrawMesh () { prefine->RateMesh(); }


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function acts the main function.
//
int nRefineMesh ( UINT cArg, char *apszArg[], char *pszFileName )
{
  int i;

  // allocate an instance of the command line parser
  CmdParse *pcmd = new CmdParse(cArg, apszArg);

  // parse the command line arguments
  fprintf(stdout, "SG Mesh Refiner     Version 1.0  (c) 1997  K. Kramer\n");
  if (pcmd->fArgExist("h") || (pcmd->GetNumOfPosArgs() > 1))
  { for(char **ppsz = apszHelp; *ppsz; ++ppsz)
      fprintf(stdout, "%s\n", *ppsz);
    return 1;
  }

  // read the grid
  prefine = new Refine(pszFileName);
  prefine->ReadGrid();

  // parse command line arguments
  cmdargs.pszAsciiMesh = pcmd->pszGetSwitchArg("O", NULL);
  if (pcmd->fArgExist("dmin")) cmdargs.cMinDivisions = (UINT) pcmd->nGetSwitchArg("dmin", 0, 0, 32767);
  if (pcmd->fArgExist("dmax")) cmdargs.cMaxDivisions = (UINT) pcmd->nGetSwitchArg("dmax", 0, 0, 32767);
#if defined(DEBUG)
  iDEBUGStartDump = (UINT) pcmd->nGetSwitchArg("debug", -1, 0, 32767);
#endif
  cmdargs.fGeneratePSFile = pcmd->fArgExist("ps");
  cmdargs.fKeepAspect = pcmd->fArgExist("aspect");
  cmdargs.fCylGeom = pcmd->fArgExist("cyl");
  cmdargs.fNoQualReport = pcmd->fArgExist("nqr");
  cmdargs.fNoPosRef = pcmd->fArgExist("yz");
  cmdargs.cQualEnh1 = (UINT) pcmd->nGetSwitchArg("qe", 10, 0, 32767);
  cmdargs.cQualEnh2 = (UINT) pcmd->nGetSwitchArg("qn", 10, 0, 32767);
  cmdargs.cQualEnh3 = (UINT) pcmd->nGetSwitchArg("qb", 10, 0, 32767);
  cmdargs.cDivAdj   = (UINT) pcmd->nGetSwitchArg("dadj", 0, 0, 32767);
  cmdargs.MinAngle  = pcmd->rGetSwitchArg("m", 0, 0, 45) * PI / 180;
  cmdargs.MaxRatio  = pcmd->rGetSwitchArg("s", 1e100, 1, 1e100);
  cmdargs.MaxGrade  = pcmd->rGetSwitchArg("g", 1e100, 0, 1e100);
  if	  (pcmd->fArgExist("vn")) cmdargs.uVisual = VISUAL_NONE;
#if 0
  else if (pcmd->fArgExist("vc")) cmdargs.uVisual = VISUAL_CONSTRUCT;
#endif
  else				  cmdargs.uVisual = VISUAL_RATING;
  cmdargs.cNoPress = pcmd->nGetSwitchArg("z", -1, 0, 32767);
  cmdargs.cPresses = 0;
  for(i = 0; i < prefine->cRefine; ++i)
  { sprintf(szGlobalBuf, "rm%d=lin", i+1);
    if (pcmd->fArgExist(szGlobalBuf)) prefine->axlatrefine[i].uMeasure = DISTANCE_LINEAR;
    sprintf(szGlobalBuf, "rm%d=log", i+1);
    if (pcmd->fArgExist(szGlobalBuf)) prefine->axlatrefine[i].uMeasure = DISTANCE_LOG;
    sprintf(szGlobalBuf, "rm%d=slog", i+1);
    if (pcmd->fArgExist(szGlobalBuf)) prefine->axlatrefine[i].uMeasure = DISTANCE_SLOG;
    sprintf(szGlobalBuf, "rd%d=", i+1);
    if (pcmd->fArgExist(szGlobalBuf))
      prefine->axlatrefine[i].Distance = pcmd->rGetSwitchArg(szGlobalBuf, 1.0, 0.0, 1e100);
  }
  if	  (pcmd->fArgExist("r1")) cmdargs.nResolution = RES_640x480;
  else if (pcmd->fArgExist("r3")) cmdargs.nResolution = RES_1024x768;
  else				  cmdargs.nResolution = RES_800x600;

#if 0
  if (cmdargs.uVisual == VISUAL_CONSTRUCT)
  { prefine->OpenWindow();
    prefine->DrawMesh();
  }
#endif

  prefine->RefineMesh();

  for(i = 0; i < cmdargs.cQualEnh1; ++i) prefine->EnhanceMeshEdges();
  for(i = 0; i < cmdargs.cQualEnh2; ++i) prefine->EnhanceMeshNodes();
  for(i = 0; i < cmdargs.cQualEnh3; ++i)
  { prefine->EnhanceMeshEdges();
    prefine->EnhanceMeshNodes();
  }

  prefine->ReorderNodes();

  GRInitGraphics();
  if (cmdargs.uVisual == VISUAL_RATING) prefine->SetupGraphicsWindow();

#if defined(UNIX)
  BOOL fDone = FALSE;
  short int keycode;
  int arg1, arg2;
  while (!fDone)
  { switch(getevent(&keycode, &arg1, &arg2))
    { case EXPOSE:
        RedrawMesh();
        flushdisplay();
        break;

      case CONFIGURENOTIFY:
        break;

      case BUTTONPRESS:
      case KEYPRESS:
        fDone = TRUE;
        break;
    }
  }
#endif

  GRFreeGraphics();

  prefine->WriteMesh();
  if (cmdargs.pszAsciiMesh) prefine->AsciiOutputMesh(cmdargs.pszAsciiMesh);

  if (cmdargs.fGeneratePSFile) prefine->GeneratePostScriptFile();
  if(pcmd->fArgExist("c"))
    prefine->WriteCGNS();
  delete pcmd;
  delete prefine;
  return 0;
}



///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Constructor
//
Refine::Refine ( char *pszFileName )
{
  // store the grid file name and construct the mesh file name
  strcpy(szGridName, pszFileName);
  strcpy(szMeshName, pszFileName);
  char *psz = strchr(szMeshName, '.');
  if (psz) *psz = '\0';
  strcat(szMeshName, ".msh");

  // initialize variables
  cRefine = 0;
  xMin = -1.0;
  yMin = -1.0;
  xMax = +1.0;
  yMax = +1.0;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Destructor
//
Refine::~Refine ( )
{
  if (axlatrefine) delete [] axlatrefine;
  if (axlatconst)  delete [] axlatconst;
  if (axlatlabel)  delete [] axlatlabel;
}


///////////////////////////////////////////////////////////////////////////////
//////////////	FUNCTIONS THAT READ THE GRID AND WRITE THE MESH	 //////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function reads the grid.
//
void Refine::ReadGrid ( )
{
  UINT i;

  // open the grid file
  int nFile = open(szGridName, O_BINARY | O_RDONLY);
  if (nFile == -1)
  { fprintf(stdout, "ERROR:  cannot open file \"%s\"\n", szGridName);
    exit(1);
  }

  // read the header
  read(nFile, &gridhead, sizeof(GRIDHEAD));
  cmdargs.cMinDivisions = gridhead.cMinDivisions;
  cmdargs.cMaxDivisions = gridhead.cMaxDivisions;

  // read the constants
  read(nFile, &cConstant, sizeof(UINT));
  if (cConstant != gridhead.cConstant)
  { fprintf(stdout, "ERROR:  corrupt grid file \"%s\"\n", szGridName);
    exit(1);
  }
  if (cConstant)
  { axlatconst = new XLATCONST[cConstant];
    read(nFile, axlatconst, cConstant * sizeof(XLATCONST));
  }
  else axlatconst = NULL;

  // read the labels
  read(nFile, &cLabel, sizeof(UINT));
  if (cLabel != gridhead.cLabel)
  { fprintf(stdout, "ERROR:  corrupt grid file \"%s\"\n", szGridName);
    exit(1);
  }
  if (cLabel)
  { axlatlabel = new XLATLABEL[cLabel];
    read(nFile, axlatlabel, cLabel * sizeof(XLATLABEL));
  }
  else axlatlabel = NULL;

  // read the refinement variables
  read(nFile, &cRefine, sizeof(UINT));
  if (cRefine != gridhead.cRefine)
  { fprintf(stdout, "ERROR:  corrupt grid file \"%s\"\n", szGridName);
    exit(1);
  }
  if (cRefine)
  { axlatrefine = new XLATREFINE[cRefine];
    read(nFile, axlatrefine, cRefine * sizeof(XLATREFINE));
  }
  else axlatrefine = NULL;

  // read the nodes
  UINT cnode;
  xMin = +1.0e100;
  yMin = +1.0e100;
  xMax = -1.0e100;
  yMax = -1.0e100;
  read(nFile, &cnode, sizeof(UINT));
  if (cnode != gridhead.cNode)
  { fprintf(stdout, "ERROR:  corrupt grid file \"%s\"\n", szGridName);
    exit(1);
  }
  for(i = 0; i < cnode; ++i)
  { NODE_t node;
    read(nFile, &node, sizeof(NODE_t));
    danode.Add(node);
    if (node.x > xMax) xMax = node.x;
    if (node.x < xMin) xMin = node.x;
    if (node.y > yMax) yMax = node.y;
    if (node.y < yMin) yMin = node.y;
  }

  // read the edges
  UINT cedge;
  read(nFile, &cedge, sizeof(UINT));
  if (cedge != gridhead.cEdge)
  { fprintf(stdout, "ERROR:  corrupt grid file \"%s\"\n", szGridName);
    exit(1);
  }
  for(i = 0; i < cedge; ++i)
  { EDGE edge;
    read(nFile, &edge, sizeof(EDGE));
    daedge.Add(edge);
  }

  // read the triangles
  UINT ctri;
  read(nFile, &ctri, sizeof(UINT));
  if (ctri != gridhead.cTriangle)
  { fprintf(stdout, "ERROR:  corrupt grid file \"%s\"\n", szGridName);
    exit(1);
  }
  TRI tri;
  tri.iedgeAB2 = -1u;
  tri.iedgeBC2 = -1u;
  tri.iedgeCD2 = -1u;
  tri.iedgeAD2 = -1u;
  tri.itriAB2  = -1u;
  tri.itriBC2  = -1u;
  tri.itriCD2  = -1u;
  tri.itriAD2  = -1u;
  tri.cLevel   = 0;
  tri.uStatus  = 0;
  for(i = 0; i < ctri; ++i)
  { read(nFile, &tri, sizeof(UINT) + 4 * sizeof(IEDGE) + 4 * sizeof(ITRI));
    datri.Add(tri);
  }

  // close
  close(nFile);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::WriteMesh
// Scope:	 Public
// Description:	 This function writes the mesh.
void Refine::WriteMesh ()
{
  UINT i;

  // create the mesh file
  int nFile = open(szMeshName, O_BINARY | O_WRONLY | O_TRUNC | O_CREAT, S_IREAD | S_IWRITE);
  if (nFile == -1)
  { fprintf(stdout, "ERROR:  cannot create file \"%s\"\n", szMeshName);
    exit(1);
  }

  // write the header
  MESHHEAD meshhead;
  strcpy(meshhead.szLogo, "SGFramework Mesh File  Version 1.0\n\x1a");
  strcpy(meshhead.szSign, SIGNATURE_MESH);
  meshhead.cConstant  = gridhead.cConstant;
  meshhead.cLabel     = gridhead.cLabel;
  meshhead.cArray     = gridhead.cRefine + 2;
  meshhead.cRegLabel  = gridhead.cRegLabel;
  meshhead.cBndryNode = cBndryNode;
  meshhead.cNode      = danode.GetElements();
  meshhead.cEdge      = daedge.GetElements();
  meshhead.cTriangle  = datri.GetElements();
  meshhead.fCylGeom   = cmdargs.fCylGeom;
  write(nFile, &meshhead, sizeof(MESHHEAD));

  // write the constants
  write(nFile, &cConstant, sizeof(UINT));
  write(nFile, axlatconst, cConstant * sizeof(XLATCONST));

  // write the labels
  WriteLabelInfo(nFile, meshhead.cNode, meshhead.cEdge, meshhead.cTriangle);

  // write the 1D arrays
  WriteArrayInfo(nFile, meshhead.cNode);

  // write the nodes
  write(nFile, &meshhead.cNode, sizeof(UINT));
  write(nFile, danode.GetPointer(0), meshhead.cNode * sizeof(NODE_t));

  // adjust the edge label indices to a zero base
  PEDGE pedge = daedge.GetPointer(0);
  for(i = 0; i < meshhead.cEdge; ++i, ++pedge) --pedge->uLabel;

  // write the edges
  write(nFile, &meshhead.cEdge, sizeof(UINT));
  write(nFile, daedge.GetPointer(0), meshhead.cEdge * sizeof(EDGE));

  // write the triangles
  write(nFile, &meshhead.cTriangle, sizeof(UINT));
  PTRI ptri = datri.GetPointer(0);
  for(i = 0; i < meshhead.cTriangle; ++i, ++ptri)
  { --ptri->uLabel;
    write(nFile, ptri, sizeof(UINT) + 4 * sizeof(IEDGE) + 4 * sizeof(ITRI));
  }

  // write the node adjacency table
  UINT cAdjEntry = meshhead.cNode + meshhead.cBndryNode * meshhead.cRegLabel;
  write(nFile, &cAdjEntry, sizeof(UINT));
  WriteAdjacencyInfo(nFile);

  // close the file
  close(nFile);
}

///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::WriteCGNS
// Scope:	 Public
// Description:	 This function outputs the mesh in cgns format.
void Refine::WriteCGNS()
{
  //printf("write file to cgns format\n"); 
  int index_file,index_base,index_zone,index_coord,index_section;
  int index_bc,index_discrete,index_sol,index_field,I;
  int i,icount,isize[3];
  char *lable;
  REAL *x,*y;
  int *ielem;
  int ier;
  // remove old file if exist
  remove("mesh.cgns"); 
  
  // open CGNS file for write 
  cg_open("mesh.cgns",MODE_WRITE,&index_file);

  // create base (user can give any name) 
  cg_base_write(index_file,"SG_To_GSS",2,2,&index_base); //two dimension
  
  //this is the main part of the work. the nodes will be classified by 
  //its region lable and insert to each zone
  int reg = 0;                                           //region index
  int cnode = danode.GetElements();                      //total ndoes
  int *iLabel =   new int [gridhead.cRegLabel];          //region lable index
  int *regnodes = new int [gridhead.cRegLabel];          //the node num in each region
  int *regtris = new int [gridhead.cRegLabel];           //the tricell num in each region
  char **reglable = new char*[gridhead.cRegLabel];     //region lable, set as zone name  
  int **local_index = new int*[gridhead.cRegLabel];    //reg_index = local_index[reg][global_index] 
  int *conn_nodes = new int[gridhead.cRegLabel];         //for zone to zone connectivity node num
  int **conn = new int*[gridhead.cRegLabel];           //for zone to zone connectivity information
 
  for(i =0; i<gridhead.cRegLabel; i ++)
  { 
  	reglable[i] = new char[32];
  	local_index[i] = new int[cnode];
  	conn[i] = new int[cnode];
  }	
 
  BOOL *af = new BOOL[cnode];
 
  for(UINT i = 0; i < cLabel; ++i)
  { 
     // determine which nodes have this region label
      if (axlatlabel[i].uType == TYPE_REGION)
      { 
      	iLabel[reg] = i;                                          //storage lable index
      	for(UINT j = 0; j < cnode; j++) local_index[reg][j] = -1; //-1 means node not in this region
      	BOOL *pf = af;
        for(UINT j = 0; j < cnode; ++j, ++pf) *pf = FALSE;        //visited flag array
        regnodes[reg] = 0;
        regtris[reg] = 0;
        conn_nodes[reg] = 0;
      	strcpy(reglable[reg],axlatlabel[i].szName);               //storage lable name
      	PTRI ptri = datri.GetPointer(0);
	for(UINT j = 0; j < datri.GetElements(); ++j, ++ptri)     //search for all the nodes
	  if (ptri->uLabel == iLabel[reg])
	  { regtris[reg]++;
	    PEDGE pedgeAB = daedge.GetPointer(ptri->iedgeAB1);
	    INODE inodeA = pedgeAB->inodeA;
	    INODE inodeB = pedgeAB->inodeB;
	    if (!af[inodeA]) { af[inodeA] = TRUE;local_index[reg][inodeA] = regnodes[reg]++; }
	    if (!af[inodeB]) { af[inodeB] = TRUE;local_index[reg][inodeB] = regnodes[reg]++; }
	    PEDGE pedgeBC = daedge.GetPointer(ptri->iedgeBC1);
	    inodeA = pedgeBC->inodeA;
	    inodeB = pedgeBC->inodeB;
	    if (!af[inodeA]) { af[inodeA] = TRUE;local_index[reg][inodeA] = regnodes[reg]++; }
	    if (!af[inodeB]) { af[inodeB] = TRUE;local_index[reg][inodeB] = regnodes[reg]++; }
	    PEDGE pedgeAC = daedge.GetPointer(ptri->iedgeAC1);
	    inodeA = pedgeAC->inodeA;
	    inodeB = pedgeAC->inodeB;
	    if (!af[inodeA]) { af[inodeA] = TRUE;local_index[reg][inodeA] = regnodes[reg]++; }
	    if (!af[inodeB]) { af[inodeB] = TRUE;local_index[reg][inodeB] = regnodes[reg]++; }
	  }
	 reg++; 
      }

  }
  delete [] af;
  
  // ok, things are done. now we write to each zone  	 
  for(i = 0; i < reg; i++)
  {    
  	isize[0] = regnodes[i];   //the ndoes in this zone
  	isize[1] = regtris[i];    //the cells in this zone
  	isize[2] = 0;             //always zero
  	
  	cg_zone_write(index_file,index_base,reglable[i],isize,Unstructured,&index_zone);
    
    	cg_goto(index_file,index_base,"Zone_t",index_zone,"end");
  	cg_descriptor_write("RegionType",reglable[i]);
  	
    	x = new REAL[regnodes[i]];
    	y = new REAL[regnodes[i]];
    	ielem = new int[regtris[i]*5];
  
    	// write grid coordinates 
       	for(int j=0;j<cnode;j++)
    	  if(local_index[i][j]>-1)  //process node in this region 
  	{
  		x[local_index[i][j]]=danode.GetPointer(j)->x;
        	y[local_index[i][j]]=danode.GetPointer(j)->y;
       	}      
  
    	cg_coord_write(index_file,index_base,index_zone,RealDouble,"CoordinateX",x,&index_coord);
    	cg_coord_write(index_file,index_base,index_zone,RealDouble,"CoordinateY",y,&index_coord);

  	// set element connectivity here
  	UINT ctri = datri.GetElements();
  	PTRI ptri = datri.GetPointer(0);
  	int  fRect;
  	INODE ainode[4];
  	int *pielem=ielem;   
  	for(int j = 0; j < ctri; ++j, ++ptri)
     	 if (ptri->uLabel == iLabel[i])  //triangle belongs to this region?
   	{
   		ElementNodes(ptri, ainode);
   	 	fRect = ptri->iedgeAD1 != -1u;
     		
   	 	if (fRect)	
   		{
    	 	  *pielem = QUAD_4;   pielem++;
    	 	  if(fCCWorder(ainode[VERTEX_A], ainode[VERTEX_B], ainode[VERTEX_C]))
   	          {
   	          	*pielem = local_index[i][ainode[0]]+1; pielem++;
   	          	*pielem = local_index[i][ainode[1]]+1; pielem++;
  	          	*pielem = local_index[i][ainode[2]]+1; pielem++;
  	          	*pielem = local_index[i][ainode[3]]+1; pielem++;
  	          }
  	          else
  	           {
   	          	*pielem = local_index[i][ainode[3]]+1; pielem++;
   	          	*pielem = local_index[i][ainode[2]]+1; pielem++;
  	          	*pielem = local_index[i][ainode[1]]+1; pielem++;
  	          	*pielem = local_index[i][ainode[0]]+1; pielem++;
  	          }	
   		}  
   	 	else
   	 	{ 
       		  *pielem = TRI_3;    pielem++;
       		  if(fCCWorder(ainode[VERTEX_A], ainode[VERTEX_B], ainode[VERTEX_C]))
           	  {
           	  	*pielem = local_index[i][ainode[0]]+1; pielem++;
           	  	*pielem = local_index[i][ainode[1]]+1; pielem++;
           	  	*pielem = local_index[i][ainode[2]]+1; pielem++;
           	  }
           	  else
           	  {
           	  	*pielem = local_index[i][ainode[2]]+1; pielem++;
           	  	*pielem = local_index[i][ainode[1]]+1; pielem++;
           	  	*pielem = local_index[i][ainode[0]]+1; pielem++;
           	  }	
   		}
     
 	}
  	cg_section_write(index_file,index_base,index_zone,"GridElements",MIXED,
                   1,isize[1],0,ielem,&index_section);
        delete [] x;
 	delete [] y;
  	delete [] ielem; 
  	
  	
   	// convert edge lable to segment, write to cgns file 
   	// for each label
	int *asegment = new int[2*cnode];
	int  segrange[2];
	segrange[0] = isize[1]+1;
    	for(int j = 0; j < gridhead.cLabel; j++)
    	{ 	
    		int *psegment = asegment;
      		int csegment = 0;
      		
      		// determine which nodes have this edge label
      		if (axlatlabel[j].uType == TYPE_EDGE)
      		{ 
      			PEDGE pedge = daedge.GetPointer(0);
			for(int k = 0; k < daedge.GetElements(); ++k, ++pedge)
	 		 if (pedge->uLabel == j)
	 		 { 
	 		   int inodeA = pedge->inodeA;
	  		   int inodeB = pedge->inodeB;
	  		   //only record the nodes that belong to this region
	  		   if(local_index[i][inodeA]==-1||local_index[i][inodeB]==-1 )continue;
	  		 	  		  
	  		   *psegment++ = local_index[i][inodeA]+1;
	  		   *psegment++ = local_index[i][inodeB]+1;
	  		   csegment++;
	  		  } 
			if(csegment>0)
			{
			  segrange[1] = segrange[0] + csegment - 1;
			  cg_section_write(index_file,index_base,index_zone,axlatlabel[j].szName,BAR_2,
                   	  		segrange[0],segrange[1],0,asegment,&index_section);
			  segrange[0] = segrange[1]+1;
	                }
	         }
        } 


        //write zone to zone connectivity data if necessary
  	if(reg>1)  //build connectivity data here
  	{
  		for(int k=0; k<reg; k++)
  		  for(int j=0; j<cnode; j++)
  	  	 {
  	   		if(k==i) continue;
  	   		if(local_index[k][j]!=-1&&local_index[i][j]!=-1)
  	   		{  
  	   		   conn[k][conn_nodes[k]]=local_index[k][j]+1;
  	   		   conn_nodes[k]++;
  	   		   conn[i][conn_nodes[i]]=local_index[i][j]+1;
  	   		   conn_nodes[i]++;
  	   		}
  	   	}	   
   	}
   	char conn_name[32];
        int  *conn_i=conn[i];
        for(int j=0; j<reg; j++)
          if(j!=i)
     	  {	//the conn_name is alpha ordered
     	  	 if(strcmp(reglable[i],reglable[j])>0)
     	  	 	sprintf(conn_name,"IF_%s_to_%s",reglable[j],reglable[i]);
     	  	 else
     	  	 	sprintf(conn_name,"IF_%s_to_%s",reglable[i],reglable[j]);	
     	  	 cg_conn_write(index_file,index_base,index_zone,conn_name,Vertex, Abutting1to1,
			PointList, conn_nodes[j], conn_i,reglable[j],Unstructured, 
			PointListDonor,Integer,conn_nodes[j], conn[j],&I);   
		 conn_i+=conn_nodes[j];
	  } 	 	       

  	for(int j=0; j<reg; j++)
        	conn_nodes[j]=0;

        // search edge lable again, write as BC 
   	// just for fit cgns file format. 	
        segrange[0] = isize[1]+1;
    	for(int j = 0; j < gridhead.cLabel; j++)
    	{ 	
    		int *psegment = asegment;
      		int csegment = 0;
      		
      		// determine which nodes have this edge label
      		if (axlatlabel[j].uType == TYPE_EDGE)
      		{ 
      			PEDGE pedge = daedge.GetPointer(0);
			for(int k = 0; k < daedge.GetElements(); ++k, ++pedge)
	 		 if (pedge->uLabel == j)
	 		 { 
	 		   int inodeA = pedge->inodeA;
	  		   int inodeB = pedge->inodeB;
	  		   //only record the nodes that belong to this region
	  		   if(local_index[i][inodeA]==-1||local_index[i][inodeB]==-1 )continue;
	  		   		   
	  		   *psegment++ = local_index[i][inodeA]+1;
	  		   *psegment++ = local_index[i][inodeB]+1;
	  		   csegment++;
	  		 } 
			if(csegment>0)
			{
			  segrange[1] = segrange[0] + csegment - 1;
			  cg_boco_write(index_file,index_base,index_zone,axlatlabel[j].szName,
	                  	        BCTypeUserDefined,ElementRange,2,segrange,&index_bc); 
			  segrange[0] = segrange[1]+1;
	                }
                 }
        } 
        delete [] asegment;	
        
        
        double Mole_x=1.0,Mole_y=1.0;
        int asize=1;
        int index_udata=0;
        cg_goto(index_file,index_base,"Zone_t",index_zone,"end");
	cg_user_data_write("Mole"); 	index_udata++;
	cg_goto(index_file,index_base,"Zone_t",index_zone,"UserDefinedData_t",index_udata,"end");
	cg_array_write("Mole_x",RealDouble,1,&asize,&Mole_x);
	cg_array_write("Mole_y",RealDouble,1,&asize,&Mole_y);
        
        
      	// write Refine data to cgns as solution data  
  	// compute and write the refinement variables
        // the region labal must be Si,GaAs,Ge or semiconductor 
        if(!strcmp(reglable[i],"Si") ||
           !strcmp(reglable[i],"GaAs") ||
           !strcmp(reglable[i],"Ge") ||
           !strcmp(reglable[i],"semiconductor"))
        {   
  		REAL *Nd=(REAL *)malloc(isize[0]*sizeof(REAL));
  		REAL *Na=(REAL *)malloc(isize[0]*sizeof(REAL));
  		for(int j = 0; j < isize[0]; ++j) Nd[j]=Na[j]=0;
  		PNODE pnode = danode.GetPointer(0);
  		for(int j = 0; j < danode.GetElements(); ++j, ++pnode)
  	  	  if(local_index[i][j]>-1)
  	   		for(int k = 0; k < gridhead.cRefine; ++k)
   	     		  if(apfRefine[k](pnode->x, pnode->y)>0)
   	       			Nd[local_index[i][j]]+=apfRefine[k](pnode->x, pnode->y);
   	     		  else
    	       			Na[local_index[i][j]]+=-apfRefine[k](pnode->x, pnode->y); 
   		
   		cg_sol_write(index_file,index_base,index_zone,"Doping",Vertex,&index_sol);      
  		cg_field_write(index_file,index_base,index_zone,index_sol,RealDouble,
                   "Nd",Nd, &index_field);    
 		cg_field_write(index_file,index_base,index_zone,index_sol,RealDouble,
                   "Na",Na, &index_field); 
        	free(Na); free(Nd);    
        }	     
  
  }    
 		
  // all things are done. close CGNS file,free memory 
  cg_close(index_file);
  delete [] iLabel; 
  delete [] regnodes; 
  delete [] regtris; 
  delete [] conn_nodes; 
  for(i =0; i<gridhead.cRegLabel; i ++) 
  {
  	delete [] reglable[i];
    	delete [] local_index[i];
    	delete [] conn[i];
  }
  delete  [] reglable;
  delete  [] local_index;	 
  delete  [] conn;
 
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::AsciiOutputMesh
// Scope:	 Public
// Description:	 This function outputs the mesh in ASCII format.
void Refine::AsciiOutputMesh (char *pszFileName)
{
  FILE *file = fopen(pszFileName, "wt");

  UINT ctri = datri.GetElements();
  UINT cedge = daedge.GetElements();
  UINT cnode = danode.GetElements();

  fprintf(file, "TRIANGLES\n");
  fprintf(file, "      |     |		EDGES		       TRIANGLES\n");
  fprintf(file, "  i   |  u  |	 AB    BC  AC/CD   AD	  AB	BC  AC/CD   AD\n");
  fprintf(file, "----- | --- | ----- ----- ----- -----	----- ----- ----- -----\n");
  for(UINT i = 0; i < ctri; ++i)
  { PTRI ptri = datri.GetPointer(i);
    fprintf(file, "%5u | ", i);
    if (ptri->uLabel != -1u) fprintf(file, "%3u | ", ptri->uLabel);  else fprintf(file, "    | ");

    if (ptri->iedgeAB1 != -1u) fprintf(file, "%5u ", ptri->iedgeAB1); else fprintf(file, "	");
    if (ptri->iedgeBC1 != -1u) fprintf(file, "%5u ", ptri->iedgeBC1); else fprintf(file, "	");
    if (ptri->iedgeAC1 != -1u) fprintf(file, "%5u ", ptri->iedgeAC1); else fprintf(file, "	");
    if (ptri->iedgeAD1 != -1u) fprintf(file, "%5u ", ptri->iedgeAD1); else fprintf(file, "  n/a ");
    fprintf(file, " ");
    if (ptri->itriAB1 != -1u) fprintf(file, "%5u ", ptri->itriAB1); else fprintf(file, "      ");
    if (ptri->itriBC1 != -1u) fprintf(file, "%5u ", ptri->itriBC1); else fprintf(file, "      ");
    if (ptri->itriAC1 != -1u) fprintf(file, "%5u ", ptri->itriAC1); else fprintf(file, "      ");
    if (ptri->itriAD1 != -1u) fprintf(file, "%5u" , ptri->itriAD1); else fprintf(file, "  n/a");
    fprintf(file, "\n");
  }

  fprintf(file, "\nEDGES\n");
  fprintf(file, "      |     |	  NODES	  \n");
  fprintf(file, "  i   |  u  |	 A     B  \n");
  fprintf(file, "----- | --- | ----- -----\n");
  for(IEDGE iedge = 0; iedge < cedge; ++iedge)
  { PEDGE pedge = daedge.GetPointer(iedge);
    fprintf(file, "%5u | ", iedge);
    if (pedge->uLabel != -1u) fprintf(file, "%3u | ", pedge->uLabel); else fprintf(file, "    | ");
    fprintf(file, "%5u %5u\n", pedge->inodeA, pedge->inodeB);
  }

  fprintf(file, "\nNODES\n");
  fprintf(file, "  i   |	x		y	\n");
  fprintf(file, "----- | --------------- ---------------\n");
  for(INODE inode = 0; inode < cnode; ++inode)
  { PNODE pnode = danode.GetPointer(inode);
    fprintf(file, "%5u | %15.8le %15.8le\n", inode, pnode->x, pnode->y);
  }

  fclose(file);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::WriteLabelInfo
// Scope:	 Private
// Description:	 This function writes the label information.
void Refine::WriteLabelInfo (int nFile, UINT cnode, UINT cedge, UINT ctri)
{
  write(nFile, &cLabel, sizeof(UINT));
  if (cLabel)
  { // allocate and initialize flag array and INODE array
    BOOL *af = new BOOL[cnode];
    INODE *ainode = new INODE[cnode];

    // for each label
    for(UINT i = 0; i < cLabel; ++i)
    { BOOL *pf = af;
      for(UINT j = 0; j < cnode; ++j, ++pf) *pf = FALSE;
      INODE *pinode = ainode;
      UINT cainode = 0;
      UINT iLabel = i + 1;

      // determine which nodes have this edge label
      if (axlatlabel[i].uType == TYPE_EDGE)
      { PEDGE pedge = daedge.GetPointer(0);
	for(UINT j = 0; j < cedge; ++j, ++pedge)
	  if (pedge->uLabel == iLabel)
	  { INODE inodeA = pedge->inodeA;
	    INODE inodeB = pedge->inodeB;
	    if (!af[inodeA]) { af[inodeA] = TRUE; *pinode++ = inodeA; ++cainode; }
	    if (!af[inodeB]) { af[inodeB] = TRUE; *pinode++ = inodeB; ++cainode; }
	  }
      }

      // determine which nodes have this region label
      else
      { PTRI ptri = datri.GetPointer(0);
	for(UINT j = 0; j < ctri; ++j, ++ptri)
	  if (ptri->uLabel == iLabel)
	  { PEDGE pedgeAB = daedge.GetPointer(ptri->iedgeAB1);
	    INODE inodeA = pedgeAB->inodeA;
	    INODE inodeB = pedgeAB->inodeB;
	    if (!af[inodeA]) { af[inodeA] = TRUE; *pinode++ = inodeA; ++cainode; }
	    if (!af[inodeB]) { af[inodeB] = TRUE; *pinode++ = inodeB; ++cainode; }
	    PEDGE pedgeBC = daedge.GetPointer(ptri->iedgeBC1);
	    inodeA = pedgeBC->inodeA;
	    inodeB = pedgeBC->inodeB;
	    if (!af[inodeA]) { af[inodeA] = TRUE; *pinode++ = inodeA; ++cainode; }
	    if (!af[inodeB]) { af[inodeB] = TRUE; *pinode++ = inodeB; ++cainode; }
	    PEDGE pedgeAC = daedge.GetPointer(ptri->iedgeAC1);
	    inodeA = pedgeAC->inodeA;
	    inodeB = pedgeAC->inodeB;
	    if (!af[inodeA]) { af[inodeA] = TRUE; *pinode++ = inodeA; ++cainode; }
	    if (!af[inodeB]) { af[inodeB] = TRUE; *pinode++ = inodeB; ++cainode; }
	  }
      }

      // write the label structure and node list
      axlatlabel[i].cNode = cainode;
      axlatlabel[i].iIndex = i;
      write(nFile, axlatlabel + i, sizeof(XLATLABEL));
      write(nFile, ainode, cainode * sizeof(INODE));
    }

    // delete flag and INODE arrays
    delete [] af;
    delete [] ainode;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::WriteArrayInfo
// Scope:	 Private
// Description:	 This function writes the array information.
void Refine::WriteArrayInfo (int nFile, UINT cnode)
{
  UINT i;

  // write the x and y coordinate arrays
  UINT cArray = cRefine + 2;
  write(nFile, &cArray, sizeof(UINT));
  write(nFile, gridhead.szCoord1, LEN_IDENT + 1);
  PNODE pnode = danode.GetPointer(0);
  for(i = 0; i < cnode; ++i, ++pnode)
    write(nFile, &(pnode->x), sizeof(REAL));
  write(nFile, gridhead.szCoord2, LEN_IDENT + 1);
  pnode = danode.GetPointer(0);
  for(i = 0; i < cnode; ++i, ++pnode)
    write(nFile, &(pnode->y), sizeof(REAL));

  // compute and write the refinement variables
  for(i = 0; i < cRefine; ++i)
  { write(nFile, axlatrefine[i].szName, LEN_IDENT + 1);
    pnode = danode.GetPointer(0);
    for(UINT j = 0; j < cnode; ++j, ++pnode)
    { REAL r = apfRefine[i](pnode->x, pnode->y);
      write(nFile, &r, sizeof(REAL));
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 QuickSortCompare
// Scope:	 Private
// Description:	 This function returns a negative integer if the edge given
//		 by the left index should be before the edge given by the
//		 right index, zero if the order does not matter or a positive
//		 integer if the edge given by the left index should be after
//		 edge given by the right index.
static int QuickSortCompare (const void *p1, const void *p2)
{
  UINT i1 = *((UINT *) p1);
  UINT i2 = *((UINT *) p2);
  return auStatus[i1] - auStatus[i2];
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::EdgeStatus
// Scope:	 Private
// Description:	 This function determines if an edge is an interior, exterior
//		 or boundary edge.
UINT Refine::EdgeStatus(UINT uLabel, ITRI itri)
{
  if (itri == -1u) return EDGESTATUS_EXTERIOR;
  PTRI ptri = datri.GetPointer(itri);
  return (uLabel == ptri->uLabel) ? EDGESTATUS_INTERIOR : EDGESTATUS_BOUNDARY;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::ReorderNodes
// Scope:	 Public
// Description:	 This function reorders the nodes so that the boundary nodes
//		 are first, followed by the exterior nodes and finally the
//		 interior nodes.
void Refine::ReorderNodes ()
{
  UINT i;

  // allocate and initialize the node status and node order arrays
  UINT cnode = danode.GetElements();
  auStatus = new UINT[cnode];
  UINT *aiOrder = new UINT[cnode];
  UINT *pu1 = auStatus;
  UINT *pu2 = aiOrder;
  for(i = 0; i < cnode; ++i, ++pu1, ++pu2)
  { *pu1 = EDGESTATUS_INTERIOR;
    *pu2 = i;
  }

  // determine the status of each node
  UINT ctri = datri.GetElements();
  PTRI ptri = datri.GetPointer(0);
  for(i = 0; i < ctri; ++i, ++ptri)
  { INODE ainode[4];
    ElementNodes(ptri, ainode);
    INODE inodeA = ainode[VERTEX_A];
    INODE inodeB = ainode[VERTEX_B];
    INODE inodeC = ainode[VERTEX_C];
    INODE inodeD = ainode[VERTEX_D];
    UINT uStatusA = auStatus[inodeA];
    UINT uStatusB = auStatus[inodeB];
    UINT uStatusC = auStatus[inodeC];

    UINT uLabel = ptri->uLabel;
    UINT uStatusAB = EdgeStatus(uLabel, ptri->itriAB1);
    UINT uStatusBC = EdgeStatus(uLabel, ptri->itriBC1);

    if (uStatusAB < uStatusA) uStatusA = uStatusAB;
    if (uStatusAB < uStatusB) uStatusB = uStatusAB;
    if (uStatusBC < uStatusB) uStatusB = uStatusBC;
    if (uStatusBC < uStatusC) uStatusC = uStatusBC;
    if (inodeD == -1u)
    { UINT uStatusAC = EdgeStatus(uLabel, ptri->itriAC1);
      if (uStatusAC < uStatusA) uStatusA = uStatusAC;
      if (uStatusAC < uStatusC) uStatusC = uStatusAC;
    }
    else
    { UINT uStatusD = auStatus[inodeD];
      UINT uStatusCD = EdgeStatus(uLabel, ptri->itriCD1);
      UINT uStatusAD = EdgeStatus(uLabel, ptri->itriAD1);
      if (uStatusCD < uStatusC) uStatusC = uStatusCD;
      if (uStatusCD < uStatusD) uStatusD = uStatusCD;
      if (uStatusAD < uStatusA) uStatusA = uStatusAD;
      if (uStatusAD < uStatusD) uStatusD = uStatusAD;
      auStatus[inodeD] = uStatusD;
    }

    auStatus[inodeA] = uStatusA;
    auStatus[inodeB] = uStatusB;
    auStatus[inodeC] = uStatusC;
  }

  // determine the edge order
  qsort(aiOrder, cnode, sizeof(UINT), QuickSortCompare);

  // determine the number of boundary nodes
  UINT *pu = aiOrder;
  cBndryNode = 0;
  for(i = 0; i < cnode; ++i, ++pu)
  { if (auStatus[*pu] != EDGESTATUS_BOUNDARY) break;
    ++cBndryNode;
  }

  // find the inverse ordering
  UINT *aiInvOrder = auStatus;
  auStatus = NULL;
  pu = aiOrder;
  for(i = 0; i < cnode; ++i, ++pu) aiInvOrder[*pu] = i;

  // relabel the node indices in the edge structures
  PEDGE pedge = daedge.GetPointer(0);
  UINT cedge = daedge.GetElements();
  for(i = 0; i < cedge; ++i, ++pedge)
  { pedge->inodeA = aiInvOrder[pedge->inodeA];
    pedge->inodeB = aiInvOrder[pedge->inodeB];
  }

  // reorder the nodes
  PNODE anodeT = new NODE_t[cnode];
  PNODE pnode  = danode.GetPointer(0);
  pu = aiInvOrder;
  for(i = 0; i < cnode; ++i, ++pnode, ++pu) anodeT[*pu] = *pnode;
  memcpy(danode.GetPointer(0), anodeT, cnode * sizeof(NODE_t));
  delete [] anodeT;

  // deallocate order arrays and return
  delete [] aiOrder;
  delete [] aiInvOrder;
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::WriteAdjacencyInfo
// Scope:	 Private
// Description:	 This function writes the node adjacency information.
void Refine::WriteAdjacencyInfo (int nFile)
{
  UINT i, j, k, l;

  // allocate and initialize the visit flag array
  UINT cnode = danode.GetElements();
  PBOOL afVisited = new BOOL[cnode];
  PBOOL pf = afVisited;
  for(i = 0; i < cnode; ++i, ++pf) *pf = FALSE;

  // allocate start and ending region indices
  UINT cRegion = gridhead.cRegLabel;
  UINT *aiBegin = new UINT[cRegion];
  UINT *aiEnd	= new UINT[cRegion];

  // loop through each element
  UINT ctri = datri.GetElements();
  for(i = 0; i < ctri; ++i)
  { PTRI ptri = datri.GetPointer(i);
    INODE ainode[4];
    ElementNodes(ptri, ainode);
    INODE *pinode = ainode;
    for(j = 0; j < 4; ++j, ++pinode)
    { if ((*pinode != -1u)&&!afVisited[*pinode])
      { afVisited[*pinode] = TRUE;

	// store the adjacent nodes in the nadj structure
	NADJ nadj;
	InitNodeAdjStructure(nadj, i, j, -1u, FALSE);

	// hit an edge w/o an adjacent triangle ... need to find the nodes
	// we may have missed
	if (!nadj.fGotAll)
	{ NADJ nadjT;
	  InitNodeAdjStructure(nadjT, i, j, -1u, TRUE);
	  UINT c = nadjT.cnode - 2;
	  for(k = nadj.cnode - 1; k != -1u; --k)
	  { nadj.ainode[k + c] = nadj.ainode[k];
	    nadj.aiedge[k + c] = nadj.aiedge[k];
	    nadj.auLabel[k + c] = nadj.auLabel[k];
	  }
	  for(k = 0; k < c; ++k)
	  { nadj.ainode[k] = nadjT.ainode[nadjT.cnode - k - 1];
	    nadj.aiedge[k] = nadjT.aiedge[nadjT.cnode - k - 1];
	    nadj.auLabel[k] = nadjT.auLabel[nadjT.cnode - k - 2];
	  }
	  nadj.cnode += c;
	}

	// compute the corners of the integration box
	UINT c = nadj.cnode;
	REAL x[32];
	REAL y[32];
	PNODE pnodeA = danode.GetPointer(nadj.inode);
	REAL xA	 = pnodeA->x;
	REAL yA	 = pnodeA->y;
	if (!nadj.fGotAll)
	{ PNODE pnodeB = danode.GetPointer(nadj.ainode[0]);
	  x[0] = (xA + pnodeB->x) / 2.0;
	  y[0] = (yA + pnodeB->y) / 2.0;
	}
	else ComputeIntPB(x[0], y[0], nadj.inode, nadj.ainode[0], nadj.ainode[c-1]);
	for(k = 0; k < c-1; ++k)
	  ComputeIntPB(x[k+1], y[k+1], nadj.inode, nadj.ainode[k], nadj.ainode[k+1]);
	if (!nadj.fGotAll)
	{ PNODE pnodeB = danode.GetPointer(nadj.ainode[c-1]);
	  x[c] = (xA + pnodeB->x) / 2.0;
	  y[c] = (yA + pnodeB->y) / 2.0;
	}
	else
	{ x[c] = x[0];
	  y[c] = y[0];
	}

	// compute the node and edge information
	NODEINFO nodeinfo;
	EDGEINFO aedgeinfo[32];
	UINT cTriangles = (nadj.fGotAll) ? c : c-1;
	nodeinfo.cNeighbors = c;
	nodeinfo.cTriangles = cTriangles;
	nodeinfo.Area = 0.0;
	for(k = 0; k < c; ++k)
	{ PNODE pnodeB = danode.GetPointer(nadj.ainode[k]);
	  REAL	elen   = sqrt(sq(xA-pnodeB->x)+sq(yA-pnodeB->y));
	  REAL ilen, area;
	  if (cmdargs.fCylGeom)
	  { ilen = LengthAdjust(x[k],y[k],x[k+1],y[k+1]);
	    area = AreaAdjust(xA,yA,x[k],y[k],x[k+1],y[k+1]);
	  }
	  else
	  { ilen = sqrt(sq(x[k]-x[k+1])+sq(y[k]-y[k+1]));
	    area = ilen * elen / 4.0;
	  }
	  aedgeinfo[k].iedge = nadj.aiedge[k];
	  aedgeinfo[k].inode = nadj.ainode[k];
	  aedgeinfo[k].ielem = nadj.aielem[k];
	  aedgeinfo[k].ilen  = ilen;
	  aedgeinfo[k].elen  = elen;
	  aedgeinfo[k].Area1 = area;
	  aedgeinfo[k].Area2 = 0.0;
	  nodeinfo.Area += area;
	}

	for(k = 0; k < cTriangles; ++k)
	{ PNODE pnodeB = danode.GetPointer(nadj.ainode[k]);
	  PNODE pnodeC = danode.GetPointer(nadj.ainode[k+1]);
	  REAL xo  = x[k+1];
	  REAL yo  = y[k+1];
	  REAL xAB = (xA+pnodeB->x)/2.0;
	  REAL yAB = (yA+pnodeB->y)/2.0;
	  REAL xAC = (xA+pnodeC->x)/2.0;
	  REAL yAC = (yA+pnodeC->y)/2.0;
	  REAL area;
	  if (cmdargs.fCylGeom)
	  { area = AreaAdjust(xA,yA,xo,yo,xAB,yAB)+
		   AreaAdjust(xA,yA,xo,yo,xAC,yAC);
	  }
	  else
	  { area = 0.5*
	      (sqrt(sq(xAB-xA)+sq(yAB-yA))*sqrt(sq(xAB-xo)+sq(yAB-yo))+
	       sqrt(sq(xAC-xA)+sq(yAC-yA))*sqrt(sq(xAC-xo)+sq(yAC-yo)));
	  }
	  aedgeinfo[k].Area2 = area;
	}

	// write the node and edge information to disk
	write(nFile, &nadj.inode, sizeof(INODE));
	write(nFile, &nodeinfo, sizeof(NODEINFO) - sizeof(EDGEINFO *));
	write(nFile, aedgeinfo, c * sizeof(EDGEINFO));

	// is this node a boundary node?
	if (nadj.inode < cBndryNode)
	{ // compute the start and end of the regions
	  UINT *piBegin = aiBegin;
	  UINT *piEnd	= aiEnd;
	  for(k = 0; k < cRegion; ++k, ++piBegin, ++piEnd)
	    *piBegin = *piEnd = -1u;

	  UINT c = nadj.cnode;
	  if (!nadj.fGotAll) --c;
	  UINT *puLabel = nadj.auLabel;
	  for(k = 0; k < c; ++k, ++puLabel)
	  { UINT uLabel = *puLabel;
	    if (aiBegin[uLabel] == -1u)
	    { aiBegin[uLabel] = k;
	      aiEnd[uLabel] = k + 1;
	    }
	    else if (aiEnd[uLabel] == k) ++aiEnd[uLabel];
	  }
	  UINT uLabel = nadj.auLabel[0];
	  puLabel = nadj.auLabel + c - 1;
	  for(k = c; *puLabel == uLabel; --k, --puLabel);
	  aiBegin[uLabel] = k % c;

	  // write the node and edge information to disk
	  piBegin = aiBegin;
	  piEnd = aiEnd;
	  for(k = 0; k < cRegion; ++k, ++piBegin, ++piEnd)
	  { UINT iBegin = *piBegin;
	    UINT iEnd	= *piEnd;
	    UINT cEdge = 0;
	    if (iBegin != -1u)
	    { cEdge  = (iEnd > iBegin) ? iEnd - iBegin : (c + iEnd) - iBegin;
	      ++cEdge;
	    }
	    iEnd %= nadj.cnode;
	    UINT cTEdge = (cEdge) ? cEdge - 1 : 0;
	    nodeinfo.cNeighbors = cEdge;
	    nodeinfo.cTriangles = cTEdge;
	    nodeinfo.Area = 0.0;
	    for(l = 0; l < cEdge; ++l)
	    { UINT m = (iBegin + l) % nadj.cnode;
	      PNODE pnodeB = danode.GetPointer(nadj.ainode[m]);
	      REAL xmdpt = (xA + pnodeB->x) / 2.0;
	      REAL ymdpt = (yA + pnodeB->y) / 2.0;
	      REAL  elen   = sqrt(sq(xA-pnodeB->x)+sq(yA-pnodeB->y));
	      REAL  ilen, area, x1, y1, x2, y2;
	      if      (m == iBegin)
	      { x1 = xmdpt;  x2 = x[m+1];
		y1 = ymdpt;  y2 = y[m+1];
	      }
	      else if (m == iEnd)
	      { x1 = x[m];   x2 = xmdpt;
		y1 = y[m];   y2 = ymdpt;
	      }
	      else
	      { x1 = x[m];   x2 = x[m+1];
		y1 = y[m];   y2 = y[m+1];
	      }
	      if (cmdargs.fCylGeom)
	      { ilen = LengthAdjust(x1,y1,x2,y2);
		area = AreaAdjust(xA,yA,x1,y1,x2,y2);
	      }
	      else
	      { ilen = sqrt(sq(x1-x2)+sq(y1-y2));
		area = ilen * elen / 4.0;
	      }
	      aedgeinfo[l].iedge = nadj.aiedge[m];
	      aedgeinfo[l].inode = nadj.ainode[m];
	      aedgeinfo[l].ielem = nadj.aielem[m];
	      aedgeinfo[l].ilen	 = ilen;
	      aedgeinfo[l].elen	 = elen;
	      aedgeinfo[l].Area1 = area;
	      aedgeinfo[l].Area2 = 0.0;
	      nodeinfo.Area += area;
	    }

	    for(l = 0; l < cTEdge; ++l)
	    { UINT m = (iBegin + l) % nadj.cnode;
	      UINT n = (m + 1) % nadj.cnode;
	      PNODE pnodeB = danode.GetPointer(nadj.ainode[m]);
	      PNODE pnodeC = danode.GetPointer(nadj.ainode[n]);
	      REAL xo  = x[m+1];
	      REAL yo  = y[m+1];
	      REAL xAB = (xA+pnodeB->x)/2.0;
	      REAL yAB = (yA+pnodeB->y)/2.0;
	      REAL xAC = (xA+pnodeC->x)/2.0;
	      REAL yAC = (yA+pnodeC->y)/2.0;
	      REAL area;
	      if (cmdargs.fCylGeom)
	      { area = AreaAdjust(xA,yA,xo,yo,xAB,yAB)+
		       AreaAdjust(xA,yA,xo,yo,xAC,yAC);
	      }
	      else
	      { area = 0.5*
		  (sqrt(sq(xAB-xA)+sq(yAB-yA))*sqrt(sq(xAB-xo)+sq(yAB-yo))+
		   sqrt(sq(xAC-xA)+sq(yAC-yA))*sqrt(sq(xAC-xo)+sq(yAC-yo)));
	      }
	      aedgeinfo[l].Area2 = area;
	    }

	    // write the node and edge information to disk
	    INODE inode = cnode + nadj.inode * cRegion + k;
	    write(nFile, &inode, sizeof(INODE));
	    write(nFile, &nodeinfo, sizeof(NODEINFO) - sizeof(EDGEINFO *));
	    write(nFile, aedgeinfo, cEdge * sizeof(EDGEINFO));
	  }
	}
      }
    }
  }

  // deallocate arrays
  delete [] afVisited;
  delete [] aiBegin;
  delete [] aiEnd;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::LengthAdjust
// Scope:	 Public
// Description:	 This function adjusts the integration length for cylinderical
//		 geometries.  The integration segment becomes an integration
//		 surface by rotating the integration segment around the y axis.
REAL Refine::LengthAdjust (REAL x1, REAL y1, REAL x2, REAL y2)
{
  REAL dx = x2 - x1;
  REAL dy = y2 - y1;
  REAL h  = sqrt(dx * dx + dy * dy);
  REAL A  = PI * (x1 + x2) * h;
  return A;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::AreaAdjust
// Scope:	 Public
// Description:	 This function adjusts the area and partial areas for
//		 cylinderical geometries.  The surfaces become regions
//		 by rotating the surfaces around the y axis.
REAL Refine::AreaAdjust (REAL x1, REAL y1, REAL x2, REAL y2, REAL x3, REAL y3)
{
  // sort the coordinates so that x1 <= x2 <= x3
  REAL xT, yT;
  if (x1 > x2) { xT = x1; yT = y1; x1 = x2; y1 = y2; x2 = xT; y2 = yT; }
  if (x2 > x3) { xT = x2; yT = y2; x2 = x3; y2 = y3; x3 = xT; y3 = yT; }
  if (x1 > x2) { xT = x1; yT = y1; x1 = x2; y1 = y2; x2 = xT; y2 = yT; }

  // calculate the slope of m13 which is never undefined since x1 != x3
  REAL m13 = (y1 - y3) / (x1 - x3);

  // calculate the square and cube of x2
  REAL x2_2 = x2 * x2;
  REAL x2_3 = x2 * x2_2;

  // calculate the volume from integrating from x1 to x2
  REAL V12 = 0.0;
  if (fabs(x1-x2) > 1.0e-14)  // @@@ need to make this relative
  { REAL m12 = (y1 - y2) / (x1 - x2);
    REAL x1_2 = x1 * x1;
    REAL x1_3 = x1 * x1_2;
    V12 = (m12 - m13) * ((x2_3 - x1_3) / 3.0 - x1 * (x2_2 - x1_2) / 2.0);
  }

  // calculate the volume from integrating from x2 to x3
  REAL V23 = 0.0;
  if (fabs(x2-x3) > 1.0e-14)  // @@@ need to make this relative
  { REAL m23 = (y2 - y3) / (x2 - x3);
    REAL x3_2 = x3 * x3;
    REAL x3_3 = x3 * x3_2;
    V23 = (m23 - m13) * ((x3_3 - x2_3) / 3.0 - x3 * (x3_2 - x2_2) / 2.0);
  }

  // calculate the total volume
  REAL V = 2.0 * PI * (fabs(V12) + fabs(V23));
  return V;
}


#define TAG12 0
#define TAG23 1
#define TAG13 2

///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::ComputeIntPB
// Scope:	 Private
// Description:	 This function computes the intersection of the perpendicular
//		 bisection of the triangle given by nodes A, B and C.
void Refine::ComputeIntPB (REAL &x, REAL &y, INODE inodeA, INODE inodeB, INODE inodeC)
{
  PNODE pnodeA = danode.GetPointer(inodeA);
  PNODE pnodeB = danode.GetPointer(inodeB);
  PNODE pnodeC = danode.GetPointer(inodeC);
  REAL x1 = pnodeA->x;	REAL y1 = pnodeA->y;
  REAL x2 = pnodeB->x;	REAL y2 = pnodeB->y;
  REAL x3 = pnodeC->x;	REAL y3 = pnodeC->y;
  REAL dx12 = x1 - x2;
  REAL dx23 = x2 - x3;
  REAL dx13 = x1 - x3;
  REAL m12 = (dx12) ? (y1 - y2) / dx12 : 0.0;
  REAL m23 = (dx23) ? (y2 - y3) / dx23 : 0.0;
  REAL m13 = (dx13) ? (y1 - y3) / dx13 : 0.0;
  REAL x12 = (x1 + x2) / 2.0;  REAL y12 = (y1 + y2) / 2.0;
  REAL x23 = (x2 + x3) / 2.0;  REAL y23 = (y2 + y3) / 2.0;
  REAL x13 = (x1 + x3) / 2.0;  REAL y13 = (y1 + y3) / 2.0;

  UINT iSmallest;
  if (fabs(dx12) < fabs(dx23))
       iSmallest = (fabs(dx12) < fabs(dx13)) ? TAG12 : TAG13;
  else iSmallest = (fabs(dx23) < fabs(dx13)) ? TAG23 : TAG13;

  switch (iSmallest)
  { case TAG12:
      x = (m13 * m23 * (y13 - y23) + m23 * x13 - m13 * x23) / (m23 - m13);
      break;

    case TAG23:
      x = (m13 * m12 * (y13 - y12) + m12 * x13 - m13 * x12) / (m12 - m13);
      break;

    case TAG13:
      x = (m12 * m23 * (y12 - y23) + m23 * x12 - m12 * x23) / (m23 - m12);
      break;
  }

  UINT iLargest;
  if (fabs(m12) > fabs(m23))
       iLargest = (fabs(m12) > fabs(m13)) ? TAG12 : TAG13;
  else iLargest = (fabs(m23) > fabs(m13)) ? TAG23 : TAG13;

  switch (iLargest)
  { case TAG12:
      y = (x12 - x) / m12 + y12;
      break;

    case TAG23:
      y = (x23 - x) / m23 + y23;
      break;

    case TAG13:
      y = (x13 - x) / m13 + y13;
      break;
  }

#if 0
  fprintf(stdout, "inodeA = %u, inodeB = %u, inodeC = %u\n", inodeA, inodeB, inodeC);
  fprintf(stdout, "x1 = %lg, y1 = %lg\n", x1, y1);
  fprintf(stdout, "x2 = %lg, y2 = %lg\n", x2, y2);
  fprintf(stdout, "x3 = %lg, y3 = %lg\n", x3, y3);
  fprintf(stdout, "iSmallest = %u\n", iSmallest);
  fprintf(stdout, "iLargest  = %u\n", iLargest);
  fprintf(stdout, "m12 = %lg\n", m12);
  fprintf(stdout, "m23 = %lg\n", m23);
  fprintf(stdout, "m13 = %lg\n", m13);
  fprintf(stdout, "x = %lg, y = %lg\n\n", x, y);
#endif
}


///////////////////////////////////////////////////////////////////////////////
//////////////////////	FUNCTIONS THAT REFINE THE MESH	///////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function refines the mesh.
//
void Refine::RefineMesh ( )
{
  // refinement triangles that need division
  BOOL fDone = FALSE;
  for (UINT i = 0; (i < cmdargs.cMaxDivisions) && !fDone; ++i)
  { fDone = TRUE;
    UINT ctri = datri.GetElements();
    for (UINT itri = 0; itri < ctri; ++itri)
    { UINT uDiv = uDivideElement(itri);
      UINT itriT = -1u;
      BOOL fNewLevel = TRUE;
      if (uDiv & DIVIDE_TRIANGLE)
      { fDone = FALSE;
	DivideTriangle(itri);
      }
      if (uDiv & DIVIDE_RECT_VER)
      { fDone = FALSE;
	itriT = DivideRectangle(itri, TRUE, TRUE);
	fNewLevel = FALSE;
      }
      if (uDiv & DIVIDE_RECT_HOR)
      { fDone = FALSE;
	DivideRectangle(itri, FALSE, fNewLevel);
	if (itriT != -1u) DivideRectangle(itriT, FALSE, fNewLevel);
      }
    }
  }

  // refine triangles that have two sides divided and those that have one
  // side divided, their level is less than cmdargs.cDivAdj and they are
  // adjacent to a triangle that will be split.	 Divide those rectangles
  // which has two parallel or adjacent sides divided.
  do
  { fDone = TRUE;
    UINT ctri = datri.GetElements();
    for(ITRI itri = 0; itri < ctri; ++itri)
    { PTRI ptri = datri.GetPointer(itri);
      BOOL fDivAB = ptri->iedgeAB2 != -1u;
      BOOL fDivBC = ptri->iedgeBC2 != -1u;
      BOOL fDivAC = ptri->iedgeAC2 != -1u;
      BOOL fDivAD = ptri->iedgeAD2 != -1u;
      BOOL fRect  = ptri->iedgeAD1 != -1u;
      UINT cDiv = (fDivAB ? 1 : 0) + (fDivBC ? 1 : 0) + (fDivAC ? 1 : 0);
      if (fRect)
      { cDiv += (fDivAD ? 1 : 0);
	BOOL fABVert = fIsEdgeVertical(ptri->iedgeAB1);
	BOOL fDivVer = FALSE;
	BOOL fDivHor = FALSE;
	if (cDiv == 1) SplitOrDivide(itri, fDivHor, fDivVer);
	else
	{ if (fDivAB && fDivAC) { if (fABVert) fDivVer = TRUE; else fDivHor = TRUE; }
	  if (fDivBC && fDivAD) { if (fABVert) fDivHor = TRUE; else fDivVer = TRUE; }
	  if ((fDivAB || fDivAC) && (fDivBC || fDivAD)) fDivVer = fDivHor = TRUE;
	}
	ITRI itriT = -1u;
	BOOL fNewLevel = TRUE;
	if (fDivVer)
	{ itriT = DivideRectangle(itri, TRUE, TRUE);
	  fDone = FALSE;
	  fNewLevel = FALSE;
	}
	if (fDivHor)
	{ DivideRectangle(itri, FALSE, fNewLevel);
	  if (itriT != -1u) DivideRectangle(itriT, FALSE, fNewLevel);
	  fDone = FALSE;
	}
      }
      else
      { BOOL fDiv = FALSE;
	if (cDiv > 1) fDiv = TRUE;
	else if ((ptri->cLevel < cmdargs.cDivAdj) && (cDiv == 1))
	{ if (fDivAB)
	  { ITRI itriAC = ptri->itriAC1;
	    ITRI itriBC = ptri->itriBC1;
	    fDiv = (((itriAC != -1u) && NumDiv(itriAC)) ||
		    ((itriBC != -1u) && NumDiv(itriBC)));
	  }
	  else if (fDivBC)
	  { ITRI itriAB = ptri->itriAB1;
	    ITRI itriAC = ptri->itriAC1;
	    fDiv = (((itriAB != -1u) && NumDiv(itriAB)) ||
		    ((itriAC != -1u) && NumDiv(itriAC)));
	  }
	  else
	  { ITRI itriAB = ptri->itriAB1;
	    ITRI itriBC = ptri->itriBC1;
	    fDiv = (((itriAB != -1u) && NumDiv(itriAB)) ||
		    ((itriBC != -1u) && NumDiv(itriBC)));
	  }
	}

	if (fDiv)
	{ DivideTriangle(itri);
	  fDone = FALSE;
	}
      }
    }
  } while (!fDone);

  // split the triangles and rectangles that have one side divided
  UINT ctri = datri.GetElements();
  for(UINT itri = 0; itri < ctri; ++itri)
  { PTRI ptri = datri.GetPointer(itri);
    if ((ptri->iedgeAB2 != -1u) ||
	(ptri->iedgeBC2 != -1u) ||
	(ptri->iedgeAC2 != -1u) ||
	(ptri->iedgeAD2 != -1u))
    { if (ptri->iedgeAD1 == -1u)  SplitTriangle(itri);
      else			  SplitRectangle(itri);
    }
  }

}

///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::NumDiv
// Scope:	 Private
// Description:	 This function returns the number of divided sides of
//		 specified triangle.
UINT Refine::NumDiv (ITRI itri)
{
  PTRI ptri = datri.GetPointer(itri);
  UINT cDiv = ((ptri->iedgeAB2 != -1u) ? 1 : 0) +
	      ((ptri->iedgeBC2 != -1u) ? 1 : 0) +
	      ((ptri->iedgeAC2 != -1u) ? 1 : 0);
  return cDiv;
}

static PTRI ptri;
static TRI  triA,    triB,    triC;
static ITRI itriA ,  itriB,   itriC;
static ITRI itriAB1, itriBC1, itriAC1, itriAD1;
static ITRI itriAB2, itriBC2, itriAC2, itriAD2;
static PTRI ptriAB1, ptriBC1, ptriAC1, ptriCD1, ptriAD1;
static PTRI ptriAB2, ptriBC2, ptriAC2, ptriCD2, ptriAD2;
static UINT cLevel,  uLabel,  uStatus, ctri;
static TRIDIV  tridiv;
static RECTDIV rectdiv;


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::DivideElement
// Scope:	 Private
// Description:	 This function divides a triangle or rectangle.
void Refine::DivideElement (ITRI itri, IEDGE iedge)
{
  // triangular element
  PTRI ptri = datri.GetPointer(itri);
  if (ptri->iedgeAD1 == -1u) DivideTriangle(itri);

  // rectangular element
  else
  { BOOL fAB = (iedge == ptri->iedgeAB1) || (iedge == ptri->iedgeAB2) ||
	       (iedge == ptri->iedgeCD1) || (iedge == ptri->iedgeCD2);
    BOOL fVert;
    if (fIsEdgeVertical(ptri->iedgeAB1))
	 fVert = (fAB) ? TRUE : FALSE;
    else fVert = (fAB) ? FALSE : TRUE;
    DivideRectangle(itri, fVert, TRUE);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::DivideTriangle
// Scope:	 Private
// Description:	 This function divides a triangle into four triangles.
void Refine::DivideTriangle (ITRI itri)
{

#if defined(DEBUG)
  ++cDEBUGDivision;
  if (iDEBUGStartDump == cDEBUGDivision) fDEBUGDumpMesh = TRUE;
  if (fDEBUGDumpMesh) DebugDumpMesh();
  if (datri.GetPointer(itri)->iedgeAD1 != -1u)
  { fprintf(stdout, "ERROR:  DivideTriangle(itri == %u) called for rectangular element!\n", itri);
    exit(1);
  }
#endif

  // divide the triangles flagged in the status field
  ptri = datri.GetPointer(itri);
  if (ptri->uStatus & STATUS_DIV_AB) DivideElement(ptri->itriAB1, ptri->iedgeAB1);
  ptri = datri.GetPointer(itri);
  if (ptri->uStatus & STATUS_DIV_BC) DivideElement(ptri->itriBC1, ptri->iedgeBC1);
  ptri = datri.GetPointer(itri);
  if (ptri->uStatus & STATUS_DIV_AC) DivideElement(ptri->itriAC1, ptri->iedgeAC1);

  // initialize local variables
  ptri = datri.GetPointer(itri);
  itriAB1 = ptri->itriAB1;
  itriBC1 = ptri->itriBC1;
  itriAC1 = ptri->itriAC1;
  itriAB2 = ptri->itriAB2;
  itriBC2 = ptri->itriBC2;
  itriAC2 = ptri->itriAC2;
  ptriAB1 = (itriAB1 == -1u) ? NULL : datri.GetPointer(itriAB1);
  ptriBC1 = (itriBC1 == -1u) ? NULL : datri.GetPointer(itriBC1);
  ptriAC1 = (itriAC1 == -1u) ? NULL : datri.GetPointer(itriAC1);
  ptriAB2 = (itriAB2 == -1u) ? NULL : datri.GetPointer(itriAB2);
  ptriBC2 = (itriBC2 == -1u) ? NULL : datri.GetPointer(itriBC2);
  ptriAC2 = (itriAC2 == -1u) ? NULL : datri.GetPointer(itriAC2);
  cLevel  = ptri->cLevel;
  uLabel  = ptri->uLabel;

  // initialize the divide structure
  InitDivideStructure(&tridiv, ptri, TRUE);

  // determine the indices of the new triangles
  ctri = datri.GetElements();
  itriA = ctri++;
  itriB = ctri++;
  itriC = ctri++;

  // initialize the interior triangle
  ptri->iedgeAB1 = tridiv.iedgeAB3;
  ptri->iedgeBC1 = tridiv.iedgeBC3;
  ptri->iedgeAC1 = tridiv.iedgeAC3;
  ptri->itriAB1	 = itriC;
  ptri->itriBC1	 = itriA;
  ptri->itriAC1	 = itriB;
  ptri->iedgeAB2 = -1u;
  ptri->iedgeBC2 = -1u;
  ptri->iedgeAC2 = -1u;
  ptri->itriAB2	 = -1u;
  ptri->itriBC2	 = -1u;
  ptri->itriAC2	 = -1u;
  ptri->iedgeAD1 = -1u;
  ptri->iedgeAD2 = -1u;
  ptri->itriAD1	 = -1u;
  ptri->itriAD2	 = -1u;
  ptri->cLevel	 = cLevel + 1;
  ptri->uStatus	 = 0;

  // make the A triangle
  triA.iedgeAB1 = tridiv.iedgeAC1;
  triA.iedgeBC1 = tridiv.iedgeBC3;
  triA.iedgeAC1 = tridiv.iedgeAB1;
  triA.itriAB1	= itriAC1;
  triA.itriBC1	= itri;
  triA.itriAC1	= itriAB1;
  triA.iedgeAB2 = -1u;
  triA.iedgeBC2 = -1u;
  triA.iedgeAC2 = -1u;
  triA.itriAB2	= -1u;
  triA.itriBC2	= -1u;
  triA.itriAC2	= -1u;
  triA.iedgeAD1 = -1u;
  triA.iedgeAD2 = -1u;
  triA.itriAD1	= -1u;
  triA.itriAD2	= -1u;
  triA.cLevel	= cLevel + 1;
  triA.uStatus	= 0;
  triA.uLabel	= uLabel;
  if (itriAB1 != -1u)
  { if (itriAB2 == -1u)
    { InitAdj2(ptriAB1, tridiv.iedgeAB1, tridiv.iedgeAB2, itriA, itriB);
      triA.uStatus |= STATUS_DIV_AC;
    }
    else InitAdj1(ptriAB1, tridiv.iedgeAB1, itriA);
  }
  if (itriAC1 != -1u)
  { if (itriAC2 == -1u)
    { InitAdj2(ptriAC1, tridiv.iedgeAC1, tridiv.iedgeAC2, itriA, itriC);
      triA.uStatus |= STATUS_DIV_AB;
    }
    else InitAdj1(ptriAC1, tridiv.iedgeAC1, itriA);
  }

  // make the B triangle
  triB.iedgeAB1 = tridiv.iedgeBC1;
  triB.iedgeBC1 = tridiv.iedgeAB2;
  triB.iedgeAC1 = tridiv.iedgeAC3;
  triB.itriAB1	= itriBC1;
  triB.itriBC1	= (itriAB2 == -1u) ? itriAB1 : itriAB2;
  triB.itriAC1	= itri;
  triB.iedgeAB2 = -1u;
  triB.iedgeBC2 = -1u;
  triB.iedgeAC2 = -1u;
  triB.itriAB2	= -1u;
  triB.itriBC2	= -1u;
  triB.itriAC2	= -1u;
  triB.iedgeAD1 = -1u;
  triB.iedgeAD2 = -1u;
  triB.itriAD1	= -1u;
  triB.itriAD2	= -1u;
  triB.cLevel	= cLevel + 1;
  triB.uStatus	= 0;
  triB.uLabel	= uLabel;
  if (itriAB1 != -1u)
  { if (itriAB2 == -1u)
    { // InitAdj2(ptriAB1, tridiv.iedgeAB1, tridiv.iedgeAB2, itriA, itriB);
      triB.uStatus |= STATUS_DIV_BC;
    }
    else InitAdj1(ptriAB2, tridiv.iedgeAB2, itriB);
  }
  if (itriBC1 != -1u)
  { if (itriBC2 == -1u)
    { InitAdj2(ptriBC1, tridiv.iedgeBC1, tridiv.iedgeBC2, itriB, itriC);
      triB.uStatus |= STATUS_DIV_AB;
    }
    else InitAdj1(ptriBC1, tridiv.iedgeBC1, itriB);
  }

  // make the C triangle
  triC.iedgeAB1 = tridiv.iedgeAB3;
  triC.iedgeBC1 = tridiv.iedgeAC2;
  triC.iedgeAC1 = tridiv.iedgeBC2;
  triC.itriAB1	= itri;
  triC.itriBC1	= (itriAC2 == -1u) ? itriAC1 : itriAC2;
  triC.itriAC1	= (itriBC2 == -1u) ? itriBC1 : itriBC2;
  triC.iedgeAB2 = -1u;
  triC.iedgeBC2 = -1u;
  triC.iedgeAC2 = -1u;
  triC.itriAB2	= -1u;
  triC.itriBC2	= -1u;
  triC.itriAC2	= -1u;
  triC.iedgeAD1 = -1u;
  triC.iedgeAD2 = -1u;
  triC.itriAD1	= -1u;
  triC.itriAD2	= -1u;
  triC.cLevel	= cLevel + 1;
  triC.uStatus	= 0;
  triC.uLabel	= uLabel;
  if (itriBC1 != -1u)
  { if (itriBC2 == -1u)
    { // InitAdj2(ptriBC1, tridiv.iedgeBC1, tridiv.iedgeBC2, itriB, itriC);
      triC.uStatus |= STATUS_DIV_AC;
    }
    else InitAdj1(ptriBC2, tridiv.iedgeBC2, itriC);
  }
  if (itriAC1 != -1u)
  { if (itriAC2 == -1u)
    { // InitAdj2(ptriAC1, tridiv.iedgeAC1, tridiv.iedgeAC2, itriA, itriC);
      triC.uStatus |= STATUS_DIV_BC;
    }
    else InitAdj1(ptriAC2, tridiv.iedgeAC2, itriC);
  }

  // add new triangles to the triangle dynamic array
  datri.Add(triA);
  datri.Add(triB);
  datri.Add(triC);

#if 0
  // display triangles ... only the middle triangle needs to be displayed
  if (cmdargs.uVisual == VISUAL_CONSTRUCT)
  { DrawElement(ptri);
    if (cmdargs.cNoPress != -1)
    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
      else { if (getch() == 'q') exit(20); }
    }
  }
#endif

#if defined(DEBUG)
  sprintf(szDEBUG, "DivideTriangle(itri = %u)\n", itri);
  Test1();
#endif
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::DivideRectangle
// Scope:	 Private
// Description:	 This function divides a rectangle either vertically or
//		 horizontally.
ITRI Refine::DivideRectangle (ITRI itri, BOOL fDivVert, BOOL fNewLevel)
{

#if defined(DEBUG)
  ++cDEBUGDivision;
  if (iDEBUGStartDump == cDEBUGDivision) fDEBUGDumpMesh = TRUE;
  if (fDEBUGDumpMesh) DebugDumpMesh();
  if (datri.GetPointer(itri)->iedgeAD1 == -1u)
  { fprintf(stdout, "ERROR:  DivideRectangle(itri == %u) called for triangular element!\n", itri);
    exit(1);
  }
#endif

  // divide the triangles flagged in the status field
  BOOL fABVert = fIsEdgeVertical(datri.GetPointer(itri)->iedgeAB1);
  BOOL fDivAB = (fABVert && fDivVert) || (!fABVert && !fDivVert);
  if (fDivAB)
  { ptri = datri.GetPointer(itri);
    if (ptri->uStatus & STATUS_DIV_AB) DivideElement(ptri->itriAB1, ptri->iedgeAB1);
    ptri = datri.GetPointer(itri);
    if (ptri->uStatus & STATUS_DIV_CD) DivideElement(ptri->itriCD1, ptri->iedgeCD1);
  }
  else
  { ptri = datri.GetPointer(itri);
    if (ptri->uStatus & STATUS_DIV_BC) DivideElement(ptri->itriBC1, ptri->iedgeBC1);
    ptri = datri.GetPointer(itri);
    if (ptri->uStatus & STATUS_DIV_AD) DivideElement(ptri->itriAD1, ptri->iedgeAD1);
  }

  // initialize local variables
  ptri = datri.GetPointer(itri);
  itriAB1 = ptri->itriAB1;
  itriBC1 = ptri->itriBC1;
  itriCD1 = ptri->itriCD1;
  itriAD1 = ptri->itriAD1;
  itriAB2 = ptri->itriAB2;
  itriBC2 = ptri->itriBC2;
  itriCD2 = ptri->itriCD2;
  itriAD2 = ptri->itriAD2;
  ptriAB1 = (itriAB1 == -1u) ? NULL : datri.GetPointer(itriAB1);
  ptriBC1 = (itriBC1 == -1u) ? NULL : datri.GetPointer(itriBC1);
  ptriCD1 = (itriCD1 == -1u) ? NULL : datri.GetPointer(itriCD1);
  ptriAD1 = (itriAD1 == -1u) ? NULL : datri.GetPointer(itriAD1);
  ptriAB2 = (itriAB2 == -1u) ? NULL : datri.GetPointer(itriAB2);
  ptriBC2 = (itriBC2 == -1u) ? NULL : datri.GetPointer(itriBC2);
  ptriCD2 = (itriCD2 == -1u) ? NULL : datri.GetPointer(itriCD2);
  ptriAD2 = (itriAD2 == -1u) ? NULL : datri.GetPointer(itriAD2);
  cLevel  = ptri->cLevel;
  uLabel  = ptri->uLabel;
  uStatus = ptri->uStatus;

  // initialize the divide structure
  InitDivideStructure(&rectdiv, ptri, fDivVert, TRUE);

  // determine the indices of the new triangles
  ctri = datri.GetElements();
  itriA = ctri++;

  // rectangle divided between edges AB and CD
  if (fDivAB)
  { ptri->iedgeAB1 = rectdiv.iedgeAB1;
    ptri->iedgeAB2 = -1u;
    ptri->iedgeBC1 = rectdiv.iedgeH;
    ptri->iedgeBC2 = -1u;
    ptri->iedgeCD1 = rectdiv.iedgeCD2;
    ptri->iedgeCD2 = -1u;
    ptri->iedgeAD1 = rectdiv.iedgeAD1;
    ptri->iedgeAD2 = rectdiv.iedgeAD2;
    ptri->itriAB1  = itriAB1;
    ptri->itriAB2  = -1u;
    ptri->itriBC1  = itriA;
    ptri->itriBC2  = -1u;
    ptri->itriCD1  = (itriCD2 != -1u) ? itriCD2 : itriCD1;
    ptri->itriCD2  = -1u;
    ptri->itriAD1  = itriAD1;
    ptri->itriAD2  = itriAD2;
    ptri->cLevel   = (fNewLevel) ? cLevel + 1 : cLevel;
    ptri->uStatus  = uStatus & STATUS_DIV_AD;
    ptri->uLabel   = uLabel;

    triA.iedgeAB1 = rectdiv.iedgeAB2;
    triA.iedgeAB2 = -1u;
    triA.iedgeBC1 = rectdiv.iedgeBC1;
    triA.iedgeBC2 = rectdiv.iedgeBC2;
    triA.iedgeCD1 = rectdiv.iedgeCD1;
    triA.iedgeCD2 = -1u;
    triA.iedgeAD1 = rectdiv.iedgeH;
    triA.iedgeAD2 = -1u;
    triA.itriAB1  = (itriAB2 != -1u) ? itriAB2 : itriAB1;
    triA.itriAB2  = -1u;
    triA.itriBC1  = itriBC1;
    triA.itriBC2  = itriBC2;
    triA.itriCD1  = itriCD1;
    triA.itriCD2  = -1u;
    triA.itriAD1  = itri;
    triA.itriAD2  = -1u;
    triA.cLevel	  = (fNewLevel) ? cLevel + 1 : cLevel;
    triA.uStatus  = uStatus & STATUS_DIV_BC;
    triA.uLabel	  = uLabel;

    if (itriAB1 != -1u)
    { if (itriAB2 == -1u)
      { InitAdj2(ptriAB1, rectdiv.iedgeAB1, rectdiv.iedgeAB2, itri, itriA);
	ptri->uStatus |= STATUS_DIV_AB;
	triA.uStatus  |= STATUS_DIV_AB;
      }
      else
      { InitAdj1(ptriAB1, rectdiv.iedgeAB1, itri);
	InitAdj1(ptriAB2, rectdiv.iedgeAB2, itriA);
      }
    }
    if (itriCD1 != -1u)
    { if (itriCD2 == -1u)
      { InitAdj2(ptriCD1, rectdiv.iedgeCD1, rectdiv.iedgeCD2, itriA, itri);
	ptri->uStatus |= STATUS_DIV_CD;
	triA.uStatus  |= STATUS_DIV_CD;
      }
      else
      { InitAdj1(ptriCD1, rectdiv.iedgeCD1, itriA);
	InitAdj1(ptriCD2, rectdiv.iedgeCD2, itri);
      }
    }
    if (itriBC1 != -1u) InitAdj3(ptriBC1, itri, itriA);
    if (itriBC2 != -1u) InitAdj3(ptriBC2, itri, itriA);
  }

  // rectangle divided between edges BC and AD
  else
  { ptri->iedgeAB1 = rectdiv.iedgeAB1;
    ptri->iedgeAB2 = rectdiv.iedgeAB2;
    ptri->iedgeBC1 = rectdiv.iedgeBC1;
    ptri->iedgeBC2 = -1u;
    ptri->iedgeCD1 = rectdiv.iedgeV;
    ptri->iedgeCD2 = -1u;
    ptri->iedgeAD1 = rectdiv.iedgeAD1;
    ptri->iedgeAD2 = -1u;
    ptri->itriAB1  = itriAB1;
    ptri->itriAB2  = itriAB2;
    ptri->itriBC1  = itriBC1;
    ptri->itriBC2  = -1u;
    ptri->itriCD1  = itriA;
    ptri->itriCD2  = -1u;
    ptri->itriAD1  = itriAD1;
    ptri->itriAD2  = -1u;
    ptri->cLevel   = (fNewLevel) ? cLevel + 1 : cLevel;
    ptri->uStatus  = uStatus & STATUS_DIV_AB;
    ptri->uLabel   = uLabel;

    triA.iedgeAB1 = rectdiv.iedgeV;
    triA.iedgeAB2 = -1u;
    triA.iedgeBC1 = rectdiv.iedgeBC2;
    triA.iedgeBC2 = -1u;
    triA.iedgeCD1 = rectdiv.iedgeCD1;
    triA.iedgeCD2 = rectdiv.iedgeCD2;
    triA.iedgeAD1 = rectdiv.iedgeAD2;
    triA.iedgeAD2 = -1u;
    triA.itriAB1  = itri;
    triA.itriAB2  = -1u;
    triA.itriBC1  = (itriBC2 != -1u) ? itriBC2 : itriBC1;
    triA.itriBC2  = -1u;
    triA.itriCD1  = itriCD1;
    triA.itriCD2  = itriCD2;
    triA.itriAD1  = (itriAD2 != -1u) ? itriAD2 : itriAD1;
    triA.itriAD2  = -1u;
    triA.cLevel	  = (fNewLevel) ? cLevel + 1 : cLevel;
    triA.uStatus  = uStatus & STATUS_DIV_CD;
    triA.uLabel	  = uLabel;

    if (itriBC1 != -1u)
    { if (itriBC2 == -1u)
      { InitAdj2(ptriBC1, rectdiv.iedgeBC1, rectdiv.iedgeBC2, itri, itriA);
	ptri->uStatus |= STATUS_DIV_BC;
	triA.uStatus  |= STATUS_DIV_BC;
      }
      else
      { InitAdj1(ptriBC1, rectdiv.iedgeBC1, itri);
	InitAdj1(ptriBC2, rectdiv.iedgeBC2, itriA);
      }
    }
    if (itriAD1 != -1u)
    { if (itriAD2 == -1u)
      { InitAdj2(ptriAD1, rectdiv.iedgeAD1, rectdiv.iedgeAD2, itri, itriA);
	ptri->uStatus |= STATUS_DIV_AD;
	triA.uStatus  |= STATUS_DIV_AD;
      }
      else
      { InitAdj1(ptriAD1, rectdiv.iedgeAD1, itri);
	InitAdj1(ptriAD2, rectdiv.iedgeAD2, itriA);
      }
    }
    if (itriCD1 != -1u) InitAdj3(ptriCD1, itri, itriA);
    if (itriCD2 != -1u) InitAdj3(ptriCD2, itri, itriA);
  }

  // add the new triangle to the triangle dynamic array
  datri.Add(triA);

#if 0
  // display triangles ... only the middle triangle needs to be displayed
  if (cmdargs.uVisual == VISUAL_CONSTRUCT)
  { DrawElement(ptri);
    if (cmdargs.cNoPress != -1)
    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
      else { if (getch() == 'q') exit(20); }
    }
  }
#endif

#if defined(DEBUG)
  sprintf(szDEBUG, "DivideRectangle(itri = %u)\n", itri);
  Test1();
#endif

  return itriA;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::SplitTriangle
// Scope:	 Private
// Description:	 This function splits a triangle into two triangles.
void Refine::SplitTriangle (ITRI itri1)
{
  // make a copy of the triangle
  ITRI itri2 = datri.GetElements();
  PTRI ptri = datri.GetPointer(itri1);
  TRI tri = *ptri;

  // initialize the tridiv structure
  TRIDIV tridiv;
  InitDivideStructure(&tridiv, ptri, FALSE);

  // create new edge
  IEDGE iedge = daedge.GetElements();
  EDGE edge;
  edge.uLabel = 0;

  // split triangle along the AB edge
  if (tri.iedgeAB2 != -1u)
  { edge.inodeA = tridiv.inodeAB;
    edge.inodeB = tridiv.inodeC;
    PTRI ptriAB1 = datri.GetPointer(tri.itriAB1);
    PTRI ptriAB2 = datri.GetPointer(tri.itriAB2);
    InitAdj1(ptriAB1, tri.iedgeAB1, itri1);
    InitAdj1(ptriAB2, tri.iedgeAB2, itri2);
    if (tri.itriBC1 != -1u)
    { PTRI ptriBC1 = datri.GetPointer(tri.itriBC1);
      InitAdj1(ptriBC1, tri.iedgeBC1, itri2);
    }
    if (tri.itriAC1 != -1u)
    { PTRI ptriAC1 = datri.GetPointer(tri.itriAC1);
      InitAdj1(ptriAC1, tri.iedgeAC1, itri1);
    }
    ptri->iedgeBC1 = iedge;
    ptri->itriBC1  = itri2;
    ptri->iedgeAB2 = -1u;
    ptri->itriAB2  = -1u;
    tri.iedgeAC1   = iedge;
    tri.itriAC1	   = itri1;
    tri.iedgeAB1   = tri.iedgeAB2;
    tri.itriAB1	   = tri.itriAB2;
    tri.iedgeAB2   = -1u;
    tri.itriAB2	   = -1u;
  }

  // split triangle along the BC edge
  else if (tri.iedgeBC2 != -1u)
  { edge.inodeA = tridiv.inodeBC;
    edge.inodeB = tridiv.inodeA;
    PTRI ptriBC1 = datri.GetPointer(tri.itriBC1);
    PTRI ptriBC2 = datri.GetPointer(tri.itriBC2);
    InitAdj1(ptriBC1, tri.iedgeBC1, itri1);
    InitAdj1(ptriBC2, tri.iedgeBC2, itri2);
    if (tri.itriAB1 != -1u)
    { PTRI ptriAB1 = datri.GetPointer(tri.itriAB1);
      InitAdj1(ptriAB1, tri.iedgeAB1, itri1);
    }
    if (tri.itriAC1 != -1u)
    { PTRI ptriAC1 = datri.GetPointer(tri.itriAC1);
      InitAdj1(ptriAC1, tri.iedgeAC1, itri2);
    }
    ptri->iedgeAC1 = iedge;
    ptri->itriAC1  = itri2;
    ptri->iedgeBC2 = -1u;
    ptri->itriBC2  = -1u;
    tri.iedgeAB1   = iedge;
    tri.itriAB1	   = itri1;
    tri.iedgeBC1   = tri.iedgeBC2;
    tri.itriBC1	   = tri.itriBC2;
    tri.iedgeBC2   = -1u;
    tri.itriBC2	   = -1u;
  }

  // split triangle along the AC edge
  else
  { edge.inodeA = tridiv.inodeAC;
    edge.inodeB = tridiv.inodeB;
    PTRI ptriAC1 = datri.GetPointer(tri.itriAC1);
    PTRI ptriAC2 = datri.GetPointer(tri.itriAC2);
    InitAdj1(ptriAC1, tri.iedgeAC1, itri1);
    InitAdj1(ptriAC2, tri.iedgeAC2, itri2);
    if (tri.itriAB1 != -1u)
    { PTRI ptriAB1 = datri.GetPointer(tri.itriAB1);
      InitAdj1(ptriAB1, tri.iedgeAB1, itri1);
    }
    if (tri.itriBC1 != -1u)
    { PTRI ptriBC1 = datri.GetPointer(tri.itriBC1);
      InitAdj1(ptriBC1, tri.iedgeBC1, itri2);
    }
    ptri->iedgeBC1 = iedge;
    ptri->itriBC1  = itri2;
    ptri->iedgeAC2 = -1u;
    ptri->itriAC2  = -1u;
    tri.iedgeAB1   = iedge;
    tri.itriAB1	   = itri1;
    tri.iedgeAC1   = tri.iedgeAC2;
    tri.itriAC1	   = tri.itriAC2;
    tri.iedgeAC2   = -1u;
    tri.itriAC2	   = -1u;
  }

  // add the edge and the triangle
  daedge.Add(edge);
  datri.Add(tri);

#if 0
  // display triangles ... only one triangle needs to be displayed
  if (cmdargs.uVisual == VISUAL_CONSTRUCT)
  { DrawElement(ptri);
    if (cmdargs.cNoPress != -1)
    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
      else { if (getch() == 'q') exit(20); }
    }
  }
#endif

  #if defined(DEBUG)
  sprintf(szDEBUG, "SplitTriangle(itri = %u)\n", itri);
  Test1();
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::SplitRectangle
// Scope:	 Private
// Description:	 This function splits a rectangle into three triangles.
void Refine::SplitRectangle (ITRI itri)
{
  // initialize the split structure
  RECTSPLIT rectsplit;
  InitSplitStructure(rectsplit, itri);

  // make the interior edges
  UINT cedge = daedge.GetElements();
  IEDGE iedgeABC = cedge++;
  IEDGE iedgeABD = cedge++;
  EDGE edge;
  edge.uLabel = 0;
  edge.inodeA = rectsplit.inodeAB;
  edge.inodeB = rectsplit.inodeC;
  daedge.Add(edge);
  edge.inodeB = rectsplit.inodeD;
  daedge.Add(edge);

  // determine the indices of the new triangles
  UINT ctri = datri.GetElements();
  ITRI itriA = ctri++;
  ITRI itriB = ctri++;

  // create three triangles
  PTRI ptri = datri.GetPointer(itri);
  UINT uLabel = ptri->uLabel;
  ptri->iedgeAB1 = iedgeABC;
  ptri->iedgeBC1 = rectsplit.iedgeCD;
  ptri->iedgeAC1 = iedgeABD;
  ptri->iedgeAD1 = -1u;
  ptri->itriAB1 = itriB;
  ptri->itriBC1 = rectsplit.itriCD;
  ptri->itriAC1 = itriA;
  ptri->itriAD1 = -1u;
  ptri->iedgeAB2 = -1u;
  ptri->iedgeBC2 = -1u;
  ptri->iedgeAC2 = -1u;
  ptri->iedgeAD2 = -1u;
  ptri->itriAB2 = -1u;
  ptri->itriBC2 = -1u;
  ptri->itriAC2 = -1u;
  ptri->itriAD2 = -1u;
  ptri->uStatus = STATUS_WAS_RECT;

  TRI tri;
  tri.iedgeAB2 = -1u;
  tri.iedgeBC2 = -1u;
  tri.iedgeAC2 = -1u;
  tri.iedgeAD1 = -1u;
  tri.iedgeAD2 = -1u;
  tri.itriAB2 = -1u;
  tri.itriBC2 = -1u;
  tri.itriAC2 = -1u;
  tri.itriAD1 = -1u;
  tri.itriAD2 = -1u;
  tri.uStatus = STATUS_WAS_RECT;
  tri.uLabel = uLabel;

  tri.iedgeAB1 = rectsplit.iedgeAB1;
  tri.iedgeBC1 = rectsplit.iedgeAD;
  tri.iedgeAC1 = iedgeABD;
  tri.itriAB1 = rectsplit.itriAB1;
  tri.itriBC1 = rectsplit.itriAD;
  tri.itriAC1 = itri;
  InitAdj1(datri.GetPointer(tri.itriAB1), tri.iedgeAB1, itriA);
  InitAdj1(datri.GetPointer(tri.itriBC1), tri.iedgeBC1, itriA);
  datri.Add(tri);

  tri.iedgeAB1 = iedgeABC;
  tri.iedgeBC1 = rectsplit.iedgeBC;
  tri.iedgeAC1 = rectsplit.iedgeAB2;
  tri.itriAB1 = itri;
  tri.itriBC1 = rectsplit.itriBC;
  tri.itriAC1 = rectsplit.itriAB2;
  InitAdj1(datri.GetPointer(tri.itriAC1), tri.iedgeAC1, itriB);
  InitAdj1(datri.GetPointer(tri.itriBC1), tri.iedgeBC1, itriB);
  datri.Add(tri);

#if 0
  // display triangles ... only one triangle needs to be displayed
  if (cmdargs.uVisual == VISUAL_CONSTRUCT)
  { DrawElement(ptri);
    if (cmdargs.cNoPress != -1)
    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
      else { if (getch() == 'q') exit(20); }
    }
  }
#endif

#if defined(DEBUG)
  sprintf(szDEBUG, "SplitRectangle(itri = %u)\n", itri);
  Test1();
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::InitDivideStructure
// Scope:	 Private
// Description:	 This function initializes the passed tridiv structure based
//		 on the passed triangle structure.  Nodes and edges are
//		 created as needed by this function.
void Refine::InitDivideStructure (PTRIDIV pdiv, PTRI ptri, BOOL fCreate)
{
  // compute the midpoints and initialize the nodes
  INODE inode1, inode2, inode3, inode4, inode5, inode6;
  FindNodes(ptri->iedgeAB1, ptri->iedgeAB2, inode1, inode2, pdiv->inodeAB, fCreate);
  FindNodes(ptri->iedgeBC1, ptri->iedgeBC2, inode3, inode4, pdiv->inodeBC, fCreate);
  FindNodes(ptri->iedgeAC1, ptri->iedgeAC2, inode5, inode6, pdiv->inodeAC, fCreate);
  if	  (inode1 == inode3) { pdiv->inodeA = inode2; pdiv->inodeB = inode1; pdiv->inodeC = inode4; }
  else if (inode1 == inode4) { pdiv->inodeA = inode2; pdiv->inodeB = inode1; pdiv->inodeC = inode3; }
  else if (inode2 == inode3) { pdiv->inodeA = inode1; pdiv->inodeB = inode2; pdiv->inodeC = inode4; }
  else if (inode2 == inode4) { pdiv->inodeA = inode1; pdiv->inodeB = inode2; pdiv->inodeC = inode3; }

  // add edges where necessary
  FindEdges(pdiv->iedgeAB1, pdiv->iedgeAB2, ptri->iedgeAB1, ptri->iedgeAB2, pdiv->inodeA, pdiv->inodeB, pdiv->inodeAB, fCreate);
  FindEdges(pdiv->iedgeBC1, pdiv->iedgeBC2, ptri->iedgeBC1, ptri->iedgeBC2, pdiv->inodeB, pdiv->inodeC, pdiv->inodeBC, fCreate);
  FindEdges(pdiv->iedgeAC1, pdiv->iedgeAC2, ptri->iedgeAC1, ptri->iedgeAC2, pdiv->inodeA, pdiv->inodeC, pdiv->inodeAC, fCreate);

  // create interior edges
  if (fCreate)
  { EDGE edge;
    edge.uLabel = 0;
    UINT cedge = daedge.GetElements();
    pdiv->iedgeAB3 = cedge++;
    pdiv->iedgeBC3 = cedge++;
    pdiv->iedgeAC3 = cedge++;
    edge.inodeA = pdiv->inodeAC;
    edge.inodeB = pdiv->inodeBC;
    daedge.Add(edge);
    edge.inodeA = pdiv->inodeAB;
    edge.inodeB = pdiv->inodeAC;
    daedge.Add(edge);
    edge.inodeA = pdiv->inodeAB;
    edge.inodeB = pdiv->inodeBC;
    daedge.Add(edge);
  }
  else
  { pdiv->iedgeAB3 = -1u;
    pdiv->iedgeBC3 = -1u;
    pdiv->iedgeAC3 = -1u;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::InitDivideStructure
// Scope:	 Private
// Description:	 This function initializes the passed rectdiv structure based
//		 on the passed triangle structure.  Nodes and edges are
//		 created as needed by this function.
void Refine::InitDivideStructure (PRECTDIV pdiv, PTRI ptri, BOOL fVert, BOOL fCreate)
{
  // determine if edge AB1 is vertical
  BOOL fABVert = fIsEdgeVertical(ptri->iedgeAB1);
  BOOL fABSplit = (fVert && fABVert) || (!fVert && !fABVert);
  BOOL fBCSplit = !fABSplit;

  // compute the midpoints and initialize the nodes
  INODE inode1, inode2, inode3, inode4, inode5, inode6, inode7, inode8;
  FindNodes(ptri->iedgeAB1, ptri->iedgeAB2, inode1, inode2, pdiv->inodeAB, fCreate && fABSplit);
  FindNodes(ptri->iedgeBC1, ptri->iedgeBC2, inode3, inode4, pdiv->inodeBC, fCreate && fBCSplit);
  FindNodes(ptri->iedgeCD1, ptri->iedgeCD2, inode5, inode6, pdiv->inodeCD, fCreate && fABSplit);
  FindNodes(ptri->iedgeAD1, ptri->iedgeAD2, inode7, inode8, pdiv->inodeAD, fCreate && fBCSplit);
  if   ((inode1 == inode3) || (inode1 == inode4))
       { pdiv->inodeA = inode2; pdiv->inodeB = inode1; }
  else { pdiv->inodeA = inode1; pdiv->inodeB = inode2; }
  if   ((inode5 == inode7) || (inode5 == inode8))
       { pdiv->inodeC = inode6; pdiv->inodeD = inode5; }
  else { pdiv->inodeC = inode5; pdiv->inodeD = inode6; }

  // add edges where necessary
  FindEdges(pdiv->iedgeAB1, pdiv->iedgeAB2, ptri->iedgeAB1, ptri->iedgeAB2, pdiv->inodeA, pdiv->inodeB, pdiv->inodeAB, fCreate && fABSplit);
  FindEdges(pdiv->iedgeBC1, pdiv->iedgeBC2, ptri->iedgeBC1, ptri->iedgeBC2, pdiv->inodeB, pdiv->inodeC, pdiv->inodeBC, fCreate && fBCSplit);
  FindEdges(pdiv->iedgeCD1, pdiv->iedgeCD2, ptri->iedgeAC1, ptri->iedgeCD2, pdiv->inodeC, pdiv->inodeD, pdiv->inodeCD, fCreate && fABSplit);
  FindEdges(pdiv->iedgeAD1, pdiv->iedgeAD2, ptri->iedgeAD1, ptri->iedgeAD2, pdiv->inodeA, pdiv->inodeD, pdiv->inodeAD, fCreate && fBCSplit);

  // create interior edges
  if (fCreate && fABSplit)
  { EDGE edge;
    edge.uLabel = 0;
    UINT cedge = daedge.GetElements();
    pdiv->iedgeH = cedge++;
    edge.inodeA = pdiv->inodeAB;
    edge.inodeB = pdiv->inodeCD;
    daedge.Add(edge);
  }
  else if (fCreate && fBCSplit)
  { EDGE edge;
    edge.uLabel = 0;
    UINT cedge = daedge.GetElements();
    pdiv->iedgeV = cedge++;
    edge.inodeA = pdiv->inodeBC;
    edge.inodeB = pdiv->inodeAD;
    daedge.Add(edge);
  }
  else
  { pdiv->iedgeH = -1u;
    pdiv->iedgeV = -1u;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::InitSplitStructure
// Scope:	 Private
// Description:	 This function initializes the rectangular split structure.
void Refine::InitSplitStructure (RECTSPLIT &rectsplit, ITRI itri)
{
  rectsplit.itri = itri;
  PTRI ptri = datri.GetPointer(itri);

  // initialize the edge and triangle indices
  if	  (ptri->iedgeAB2 != -1u)
  { rectsplit.iedgeAB1 = ptri->iedgeAB1;
    rectsplit.iedgeAB2 = ptri->iedgeAB2;
    rectsplit.iedgeBC  = ptri->iedgeBC1;
    rectsplit.iedgeCD  = ptri->iedgeCD1;
    rectsplit.iedgeAD  = ptri->iedgeAD1;
    rectsplit.itriAB1  = ptri->itriAB1;
    rectsplit.itriAB2  = ptri->itriAB2;
    rectsplit.itriBC   = ptri->itriBC1;
    rectsplit.itriCD   = ptri->itriCD1;
    rectsplit.itriAD   = ptri->itriAD1;
  }
  else if (ptri->iedgeBC2 != -1u)
  { rectsplit.iedgeAB1 = ptri->iedgeBC1;
    rectsplit.iedgeAB2 = ptri->iedgeBC2;
    rectsplit.iedgeBC  = ptri->iedgeCD1;
    rectsplit.iedgeCD  = ptri->iedgeAD1;
    rectsplit.iedgeAD  = ptri->iedgeAB1;
    rectsplit.itriAB1  = ptri->itriBC1;
    rectsplit.itriAB2  = ptri->itriBC2;
    rectsplit.itriBC   = ptri->itriCD1;
    rectsplit.itriCD   = ptri->itriAD1;
    rectsplit.itriAD   = ptri->itriAB1;
  }
  else if (ptri->iedgeCD2 != -1u)
  { rectsplit.iedgeAB1 = ptri->iedgeCD1;
    rectsplit.iedgeAB2 = ptri->iedgeCD2;
    rectsplit.iedgeBC  = ptri->iedgeAD1;
    rectsplit.iedgeCD  = ptri->iedgeAB1;
    rectsplit.iedgeAD  = ptri->iedgeBC1;
    rectsplit.itriAB1  = ptri->itriCD1;
    rectsplit.itriAB2  = ptri->itriCD2;
    rectsplit.itriBC   = ptri->itriAD1;
    rectsplit.itriCD   = ptri->itriAB1;
    rectsplit.itriAD   = ptri->itriBC1;
  }
  else if (ptri->iedgeAD2 != -1u)
  { rectsplit.iedgeAB1 = ptri->iedgeAD1;
    rectsplit.iedgeAB2 = ptri->iedgeAD2;
    rectsplit.iedgeBC  = ptri->iedgeCD1;
    rectsplit.iedgeCD  = ptri->iedgeBC1;
    rectsplit.iedgeAD  = ptri->iedgeAB1;
    rectsplit.itriAB1  = ptri->itriAD1;
    rectsplit.itriAB2  = ptri->itriAD2;
    rectsplit.itriBC   = ptri->itriCD1;
    rectsplit.itriCD   = ptri->itriBC1;
    rectsplit.itriAD   = ptri->itriAB1;
  }

  // initialize the node indices
  PEDGE pedgeAB1 = daedge.GetPointer(rectsplit.iedgeAB1);
  PEDGE pedgeAB2 = daedge.GetPointer(rectsplit.iedgeAB2);
  PEDGE pedgeBC	 = daedge.GetPointer(rectsplit.iedgeBC);
  PEDGE pedgeCD	 = daedge.GetPointer(rectsplit.iedgeCD);
  INODE inode1 = pedgeAB1->inodeA;
  INODE inode2 = pedgeAB1->inodeB;
  INODE inode3 = pedgeAB2->inodeA;
  INODE inode4 = pedgeAB2->inodeB;
  if	  (inode1 == inode3) { rectsplit.inodeA = inode2; rectsplit.inodeAB = inode1; rectsplit.inodeB = inode4; }
  else if (inode1 == inode4) { rectsplit.inodeA = inode2; rectsplit.inodeAB = inode1; rectsplit.inodeB = inode3; }
  else if (inode2 == inode3) { rectsplit.inodeA = inode1; rectsplit.inodeAB = inode2; rectsplit.inodeB = inode4; }
  else if (inode2 == inode4) { rectsplit.inodeA = inode1; rectsplit.inodeAB = inode2; rectsplit.inodeB = inode3; }
  inode1 = pedgeBC->inodeA;
  inode2 = pedgeBC->inodeB;
  rectsplit.inodeC = (rectsplit.inodeB == inode1) ? inode2 : inode1;
  inode1 = pedgeCD->inodeA;
  inode2 = pedgeCD->inodeB;
  rectsplit.inodeD = (rectsplit.inodeC == inode1) ? inode2 : inode1;

#if defined(DEBUG)
  PEDGE pedgeAD	 = daedge.GetPointer(rectsplit.iedgeAD);
  inode1 = pedgeAD->inodeA;
  inode2 = pedgeAD->inodeB;
  if (((rectsplit.inodeA != inode1) || (rectsplit.inodeB != inode2)) &&
      ((rectsplit.inodeA != inode2) || (rectsplit.inodeB != inode1)))
  { fprintf(stdout, "ERROR in InitSplitStructure\n");
    exit(1);
  }
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::ComputeAngle
// Scope:	 Private
// Description:	 This function computes the angle between the three nodes.
//		 The angle is assumed to be less than PI radians.
REAL Refine::ComputeAngle(INODE inode1, INODE inode2, INODE inode3)
{
  PNODE pnode1 = danode.GetPointer(inode1);
  PNODE pnode2 = danode.GetPointer(inode2);
  PNODE pnode3 = danode.GetPointer(inode3);

  REAL xl = pnode1->x - pnode2->x;
  REAL xr = pnode3->x - pnode2->x;
  REAL yl = pnode1->y - pnode2->y;
  REAL yr = pnode3->y - pnode2->y;
  REAL r = (xl*xr+yl*yr)/(sqrt(sq(xl)+sq(yl))*sqrt(sq(xr)+sq(yr)));
  if	  (r >	1.0) r =  1.0;
  else if (r < -1.0) r = -1.0;
  REAL angle = acos(r);
  //if (xl*yr-xr*yl > 0) angle = 2*PI-angle;
  return angle;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::SplitOrDivide
// Scope:	 Private
// Description:	 This function determines whether a rectangle should be split
//		 or divided.
void Refine::SplitOrDivide (ITRI itri, BOOL &fHor, BOOL &fVer)
{
  // initialize the split structure
  RECTSPLIT rectsplit;
  InitSplitStructure(rectsplit, itri);

  // compute the angle of the middle triangle
  REAL angle = ComputeAngle(rectsplit.inodeD, rectsplit.inodeAB, rectsplit.inodeC);

  // the rectangle should not be split if an obtuse angle results or the
  // angles of the resulting triangles are too small
  fHor = fVer = FALSE;
  if ((angle > (PI/2.0 - 0.01)) ||
      (angle/2.0 < cmdargs.MinAngle) ||
      ((PI - angle)/2.0 < cmdargs.MinAngle))
  { if (fIsEdgeVertical(rectsplit.iedgeAB1))
	 fVer = TRUE;
    else fHor = TRUE;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::fIsEdgeVertical
// Scope:	 Private
// Description:	 This function returns true if the passed edge is vertical.
BOOL Refine::fIsEdgeVertical (IEDGE iedge)
{
  PEDGE pedge  = daedge.GetPointer(iedge);
  PNODE pnodeA = danode.GetPointer(pedge->inodeA);
  PNODE pnodeB = danode.GetPointer(pedge->inodeB);
  return fabs(pnodeA->x - pnodeB->x) < fabs(pnodeA->y - pnodeB->y);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::FindNodes
// Scope:	 Private
// Description:	 This function finds the endpoint nodes and finds or creates
//		 the midpoint node.
void Refine::FindNodes(IEDGE iedge1, IEDGE iedge2, INODE &inode1, INODE &inode2, INODE &inode3, BOOL fCreateMdpt)
{
  // edge is not divided, compute midpoint
  if (iedge2 == -1u)
  { PEDGE pedge = daedge.GetPointer(iedge1);
    inode1 = pedge->inodeA;
    inode2 = pedge->inodeB;

    if (fCreateMdpt)
    { inode3 = danode.GetElements();
      PNODE pnode1 = danode.GetPointer(inode1);
      PNODE pnode2 = danode.GetPointer(inode2);
      NODE_t node;
      node.x = (pnode1->x + pnode2->x) / 2.0;
      node.y = (pnode1->y + pnode2->y) / 2.0;
      danode.Add(node);
    }
    else inode3 = -1u;
  }

  // edge is divided, find midpoint
  else
  { PEDGE pedge1 = daedge.GetPointer(iedge1);
    PEDGE pedge2 = daedge.GetPointer(iedge2);
    if	    (pedge1->inodeA == pedge2->inodeA) { inode1 = pedge1->inodeB; inode2 = pedge2->inodeB; inode3 = pedge1->inodeA; }
    else if (pedge1->inodeA == pedge2->inodeB) { inode1 = pedge1->inodeB; inode2 = pedge2->inodeA; inode3 = pedge1->inodeA; }
    else if (pedge1->inodeB == pedge2->inodeA) { inode1 = pedge1->inodeA; inode2 = pedge2->inodeB; inode3 = pedge1->inodeB; }
    else if (pedge1->inodeB == pedge2->inodeB) { inode1 = pedge1->inodeA; inode2 = pedge2->inodeA; inode3 = pedge1->inodeB; }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::FindEdges
// Scope:	 Private
// Description:	 This function finds and/or creates the two subedges of the
//		 passed edge.
void Refine::FindEdges (IEDGE &iedgeA, IEDGE &iedgeB, IEDGE iedge1, IEDGE iedge2, INODE inode1, INODE inode2, INODE inode3, BOOL fCreateEdge)
{
  // two subedges already created
  if (!fCreateEdge || (iedge2 != -1u))
  { iedgeA = iedge1;
    iedgeB = iedge2;
  }

  // must create the two subedges
  else
  { iedgeA = iedge1;
    iedgeB = daedge.GetElements();

    PEDGE pedge = daedge.GetPointer(iedge1);
    pedge->inodeA  = inode1;
    pedge->inodeB  = inode3;

    EDGE edge;
    edge.uLabel = pedge->uLabel;
    edge.inodeA	 = inode2;
    edge.inodeB	 = inode3;
    daedge.Add(edge);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::InitAdj1
// Scope:	 Private
// Description:	 This function
void Refine::InitAdj1 (PTRI ptri, IEDGE iedge, ITRI itri)
{
  if	  (ptri->iedgeAB1 == iedge) { ptri->itriAB1 = itri; ptri->uStatus &= ~STATUS_DIV_AB;}
  else if (ptri->iedgeBC1 == iedge) { ptri->itriBC1 = itri; ptri->uStatus &= ~STATUS_DIV_BC;}
  else if (ptri->iedgeCD1 == iedge) { ptri->itriCD1 = itri; ptri->uStatus &= ~STATUS_DIV_CD;}
  else if (ptri->iedgeAD1 == iedge) { ptri->itriAD1 = itri; ptri->uStatus &= ~STATUS_DIV_AD;}
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::InitAdj2
// Scope:	 Private
// Description:	 This function
void Refine::InitAdj2 (PTRI ptri, IEDGE iedge1, IEDGE iedge2, ITRI itri1, ITRI itri2)
{
  INODE inode1, inode2, inode3, inode4, inodeA, inodeB, inodeM;
  FindNodes(iedge1, iedge2, inodeA, inodeB, inodeM, FALSE);
  BOOL fRect = ptri->iedgeAD1 != -1u;

  if (fRect)
  {
    // element is adjacent to edge AB
    if	    (ptri->iedgeAB1 == iedge1)
    { FindNodes(ptri->iedgeCD1, ptri->iedgeCD2, inode1, inode2, inodeM, FALSE);
      FindNodes(ptri->iedgeAD1, ptri->iedgeAD2, inode3, inode4, inodeM, FALSE);
      INODE inode = ((inode1 == inode3) || (inode2 == inode3)) ? inode4 : inode3;
      if (inode == inodeA)
      { ptri->iedgeAB1 = iedge1;
	ptri->iedgeAB2 = iedge2;
	ptri->itriAB1  = itri1;
	ptri->itriAB2  = itri2;
      }
      else
      { ptri->iedgeAB1 = iedge2;
	ptri->iedgeAB2 = iedge1;
	ptri->itriAB1  = itri2;
	ptri->itriAB2  = itri1;
      }
    }

    // element is adjacent to edge BC
    else if (ptri->iedgeBC1 == iedge1)
    { FindNodes(ptri->iedgeAB1, ptri->iedgeAB2, inode1, inode2, inodeM, FALSE);
      FindNodes(ptri->iedgeAD1, ptri->iedgeAD2, inode3, inode4, inodeM, FALSE);
      INODE inode = ((inode1 == inode3) || (inode1 == inode4)) ? inode2 : inode1;
      if (inode == inodeA)
      { ptri->iedgeBC1 = iedge1;
	ptri->iedgeBC2 = iedge2;
	ptri->itriBC1  = itri1;
	ptri->itriBC2  = itri2;
      }
      else
      { ptri->iedgeBC1 = iedge2;
	ptri->iedgeBC2 = iedge1;
	ptri->itriBC1  = itri2;
	ptri->itriBC2  = itri1;
      }
    }

    // element is adjacent to edge CD
    else if (ptri->iedgeCD1 == iedge1)
    { FindNodes(ptri->iedgeAB1, ptri->iedgeAB2, inode1, inode2, inodeM, FALSE);
      FindNodes(ptri->iedgeBC1, ptri->iedgeBC2, inode3, inode4, inodeM, FALSE);
      INODE inode = ((inode3 == inode1) || (inode3 == inode2)) ? inode4 : inode3;
      if (inode == inodeA)
      { ptri->iedgeCD1 = iedge1;
	ptri->iedgeCD2 = iedge2;
	ptri->itriCD1  = itri1;
	ptri->itriCD2  = itri2;
      }
      else
      { ptri->iedgeCD1 = iedge2;
	ptri->iedgeCD2 = iedge1;
	ptri->itriCD1  = itri2;
	ptri->itriCD2  = itri1;
      }
    }

    // element is adjacent to edge AD
    else if (ptri->iedgeAD1 == iedge1)
    { FindNodes(ptri->iedgeAB1, ptri->iedgeAB2, inode1, inode2, inodeM, FALSE);
      FindNodes(ptri->iedgeBC1, ptri->iedgeBC2, inode3, inode4, inodeM, FALSE);
      INODE inode = ((inode1 == inode3) || (inode1 == inode4)) ? inode2 : inode1;
      if (inode == inodeA)
      { ptri->iedgeAD1 = iedge1;
	ptri->iedgeAD2 = iedge2;
	ptri->itriAD1  = itri1;
	ptri->itriAD2  = itri2;
      }
      else
      { ptri->iedgeAD1 = iedge2;
	ptri->iedgeAD2 = iedge1;
	ptri->itriAD1  = itri2;
	ptri->itriAD2  = itri1;
      }
    }
  }

  else
  {
    // element is adjacent to edge AB
    if	    (ptri->iedgeAB1 == iedge1)
    { FindNodes(ptri->iedgeBC1, ptri->iedgeBC2, inode1, inode2, inodeM, FALSE);
      FindNodes(ptri->iedgeAC1, ptri->iedgeAC2, inode3, inode4, inodeM, FALSE);
      INODE inode = ((inode1 == inode3) || (inode2 == inode3)) ? inode4 : inode3;
      if (inode == inodeA)
      { ptri->iedgeAB1 = iedge1;
	ptri->iedgeAB2 = iedge2;
	ptri->itriAB1  = itri1;
	ptri->itriAB2  = itri2;
      }
      else
      { ptri->iedgeAB1 = iedge2;
	ptri->iedgeAB2 = iedge1;
	ptri->itriAB1  = itri2;
	ptri->itriAB2  = itri1;
      }
    }

    // element is adjacent to edge BC
    else if (ptri->iedgeBC1 == iedge1)
    { FindNodes(ptri->iedgeAB1, ptri->iedgeAB2, inode1, inode2, inodeM, FALSE);
      FindNodes(ptri->iedgeAC1, ptri->iedgeAC2, inode3, inode4, inodeM, FALSE);
      INODE inode = ((inode1 == inode3) || (inode1 == inode4)) ? inode2 : inode1;
      if (inode == inodeA)
      { ptri->iedgeBC1 = iedge1;
	ptri->iedgeBC2 = iedge2;
	ptri->itriBC1  = itri1;
	ptri->itriBC2  = itri2;
      }
      else
      { ptri->iedgeBC1 = iedge2;
	ptri->iedgeBC2 = iedge1;
	ptri->itriBC1  = itri2;
	ptri->itriBC2  = itri1;
      }
    }

    // element is adjacent to edge AC
    else if (ptri->iedgeAC1 == iedge1)
    { FindNodes(ptri->iedgeAB1, ptri->iedgeAB2, inode1, inode2, inodeM, FALSE);
      FindNodes(ptri->iedgeBC1, ptri->iedgeBC2, inode3, inode4, inodeM, FALSE);
      INODE inode = ((inode1 == inode3) || (inode1 == inode4)) ? inode2 : inode1;
      if (inode == inodeA)
      { ptri->iedgeAC1 = iedge1;
	ptri->iedgeAC2 = iedge2;
	ptri->itriAC1  = itri1;
	ptri->itriAC2  = itri2;
      }
      else
      { ptri->iedgeAC1 = iedge2;
	ptri->iedgeAC2 = iedge1;
	ptri->itriAC1  = itri2;
	ptri->itriAC2  = itri1;
      }
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::InitAdj3
// Scope:	 Private
// Description:	 This function
void Refine::InitAdj3 (PTRI ptri, ITRI itri1, ITRI itri2)
{
  if (ptri->itriAB1 == itri1) ptri->itriAB1 = itri2;
  if (ptri->itriBC1 == itri1) ptri->itriBC1 = itri2;
  if (ptri->itriCD1 == itri1) ptri->itriCD1 = itri2;
  if (ptri->itriAD1 == itri1) ptri->itriAD1 = itri2;
  if (ptri->itriAB2 == itri1) ptri->itriAB2 = itri2;
  if (ptri->itriBC2 == itri1) ptri->itriBC2 = itri2;
  if (ptri->itriCD2 == itri1) ptri->itriCD2 = itri2;
  if (ptri->itriAD2 == itri1) ptri->itriAD2 = itri2;
}


///////////////////////////////////////////////////////////////////////////////
/////////  FUNCTIONS THAT DETERMINE IF A TRIANGLE SHOULD BE DIVIDED  //////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::InitVerticesStructure
// Scope:	 Private
// Description:	 This function initializes the vertices structure.
void Refine::InitVerticesStructure (PVERTICES pvertices, PTRI ptri)
{
  INODE inodeA,inodeB,inodeC,inodeD,inodeM;
  INODE inodeAB1,inodeAB2,inodeBC1,inodeBC2,inodeCD1,inodeCD2,inodeAD1,inodeAD2;
  BOOL	fRect = ptri->iedgeAD1 != -1u;

  FindNodes(ptri->iedgeAB1, ptri->iedgeAB2, inodeAB1, inodeAB2, inodeM, FALSE);
  FindNodes(ptri->iedgeBC1, ptri->iedgeBC2, inodeBC1, inodeBC2, inodeM, FALSE);
  FindNodes(ptri->iedgeCD1, ptri->iedgeCD2, inodeCD1, inodeCD2, inodeM, FALSE);
  if (fRect)
    FindNodes(ptri->iedgeAD1, ptri->iedgeAD2, inodeAD1, inodeAD2, inodeM, FALSE);

  if ((inodeAB1 == inodeBC1) || (inodeAB1 == inodeBC2))
       { inodeA = inodeAB2; inodeB = inodeAB1; }
  else { inodeA = inodeAB1; inodeB = inodeAB2; }

  if (fRect)
  { if ((inodeCD1 == inodeAD1) || (inodeCD1 == inodeAD2))
	 { inodeC = inodeCD2; inodeD = inodeCD1; }
    else { inodeC = inodeCD1; inodeD = inodeCD2; }
  }
  else
  { inodeC = ((inodeCD1 == inodeBC1) || (inodeCD1 == inodeBC2))
	   ? inodeCD1 : inodeCD2;
  }


  // initialize the vertice structure
  PNODE pnodeA = danode.GetPointer(inodeA);
  pvertices->x1 = pnodeA->x;
  pvertices->y1 = pnodeA->y;
  PNODE pnodeB = danode.GetPointer(inodeB);
  pvertices->x2 = pnodeB->x;
  pvertices->y2 = pnodeB->y;
  PNODE pnodeC = danode.GetPointer(inodeC);
  pvertices->x3 = pnodeC->x;
  pvertices->y3 = pnodeC->y;
  if (fRect)
  { PNODE pnodeD = danode.GetPointer(inodeD);
    pvertices->x4 = pnodeD->x;
    pvertices->y4 = pnodeD->y;
  }
  pvertices->fRect = fRect;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 RefineXform
// Scope:	 Private
// Description:	 This function transforms the refinement variable.
REAL RefineXform (REAL x, UINT uType)
{
  switch (uType)
  { case DISTANCE_LINEAR:
      return x;

    case DISTANCE_LOG:
      return log10(x);

    case DISTANCE_SLOG:
      return sign(x)*log10(1+fabs(x));
//    if      (x < -1.0) return -log(-x);
//    else if (x >  1.0) return +log(+x);
//    else		 return 0.0;
  }
  return 0.0;
}


// declare minimum and maximum edge length functions ...
REAL _MinEdgeLen  (REAL, REAL);
REAL _MaxEdgeLen  (REAL, REAL);
REAL _MaxXEdgeLen (REAL, REAL);
REAL _MaxYEdgeLen (REAL, REAL);


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::uDivideElement
// Scope:	 Private
// Description:	 This function determines if a element should be divided.
//		 The following rules are used to determine if a triangle
//		 should be divided.
//
//		 1.  if elem div level < min div level then divide
//		 2.  if elem div level > max div level then do not divide
//		 3.  if adj elements divided and cannot split, then divide
//		 4.  if elem lengths > max length then do divide
//		 5.  if elem lengths < min lengths then do not divide
//		 6.  if refinement criteria not satisfied then divide
UINT Refine::uDivideElement (ITRI itri)
{
  // initialize local variables
  PTRI ptri = datri.GetPointer(itri);
  UINT uResult = 0;
  BOOL fRect = ptri->iedgeAD1 != -1u;
  BOOL fABVert;

  // rule #1
  if (ptri->cLevel < cmdargs.cMinDivisions)
    return (fRect) ? DIVIDE_RECT_HOR | DIVIDE_RECT_VER : DIVIDE_TRIANGLE;

  //if (daedge.GetPointer(ptri->iedgeAB1)->uLabel)  return 0;
  //if (daedge.GetPointer(ptri->iedgeBC1)->uLabel)  return 0;
  //if (daedge.GetPointer(ptri->iedgeAC1)->uLabel)  return 0;
  //if (fRect&&daedge.GetPointer(ptri->iedgeAD1)->uLabel)  return 0;
  // rule #2
  if (ptri->cLevel >= cmdargs.cMaxDivisions)
    return 0;

  // rule #3
  BOOL fDividedAB = ptri->iedgeAB2 != -1u;
  BOOL fDividedBC = ptri->iedgeBC2 != -1u;
  BOOL fDividedCD = ptri->iedgeCD2 != -1u;
  BOOL fDividedAD = ptri->iedgeAD2 != -1u;
  if (fRect)
  { fABVert = fIsEdgeVertical(ptri->iedgeAB1);
    if (fDividedAB && fDividedCD) uResult |= (fABVert) ? DIVIDE_RECT_VER : DIVIDE_RECT_HOR;
    if (fDividedBC && fDividedAD) uResult |= (fABVert) ? DIVIDE_RECT_HOR : DIVIDE_RECT_VER;
    if ((fDividedAB || fDividedCD) &&
	(fDividedBC || fDividedAD)) uResult = DIVIDE_RECT_VER | DIVIDE_RECT_HOR;
  }
  else
  { UINT c = 0;
    if (fDividedAB) ++c;
    if (fDividedBC) ++c;
    if (fDividedCD) ++c;
    if (c > 1) return DIVIDE_TRIANGLE;
  }

  // test the edge spacing
  REAL Min, Max, xMax, yMax;
  VERTICES vertices;
  InitVerticesStructure(&vertices, ptri);
  REAL exAB = fabs(vertices.x1 - vertices.x2);
  REAL eyAB = fabs(vertices.y1 - vertices.y2);
  REAL exBC = fabs(vertices.x2 - vertices.x3);
  REAL eyBC = fabs(vertices.y2 - vertices.y3);
  REAL eAB = sqrt(sq(exAB) + sq(eyAB));
  REAL eBC = sqrt(sq(exBC) + sq(eyBC));
  REAL MinA = _MinEdgeLen(vertices.x1, vertices.y1);
  REAL MinB = _MinEdgeLen(vertices.x2, vertices.y2);
  REAL MinC = _MinEdgeLen(vertices.x3, vertices.y3);
  Min = (MinA < MinB) ? MinA : MinB;
  Min = (MinC < Min ) ? MinC : Min;
  REAL MaxA = _MaxEdgeLen(vertices.x1, vertices.y1);
  REAL MaxB = _MaxEdgeLen(vertices.x2, vertices.y2);
  REAL MaxC = _MaxEdgeLen(vertices.x3, vertices.y3);
  Max = (MaxA < MaxB) ? MaxA : MaxB;
  Max = (MaxC < Max ) ? MaxC : Max;
  REAL xMaxA = _MaxXEdgeLen(vertices.x1, vertices.y1);
  REAL xMaxB = _MaxXEdgeLen(vertices.x2, vertices.y2);
  REAL xMaxC = _MaxXEdgeLen(vertices.x3, vertices.y3);
  xMax = (xMaxA < xMaxB) ? xMaxA : xMaxB;
  xMax = (xMaxC < xMax ) ? xMaxC : xMax;
  REAL yMaxA = _MaxYEdgeLen(vertices.x1, vertices.y1);
  REAL yMaxB = _MaxYEdgeLen(vertices.x2, vertices.y2);
  REAL yMaxC = _MaxYEdgeLen(vertices.x3, vertices.y3);
  yMax = (yMaxA < yMaxB) ? yMaxA : yMaxB;
  yMax = (yMaxC < yMax ) ? yMaxC : yMax;

  BOOL fNoHor = FALSE;
  BOOL fNoVer = FALSE;
  if (fRect)
  { REAL MinD = _MinEdgeLen(vertices.x4, vertices.y4);
    REAL MaxD = _MaxEdgeLen(vertices.x4, vertices.y4);
    REAL xMaxD = _MaxXEdgeLen(vertices.x4, vertices.y4);
    REAL yMaxD = _MaxYEdgeLen(vertices.x4, vertices.y4);
    Min = (MinD < Min ) ? MinD : Min;
    Max = (MaxD < Max ) ? MaxD : Max;
    xMax = (xMaxD < xMax ) ? xMaxD : xMax;
    yMax = (yMaxD < yMax ) ? yMaxD : yMax;
    if (eAB < Min) { if (fABVert) fNoVer = TRUE; else fNoHor = TRUE; }
    if (eBC < Min) { if (fABVert) fNoHor = TRUE; else fNoVer = TRUE; }
    if ((eAB > Max) || (exAB > xMax) || (eyAB > yMax))
    { uResult |= (fABVert) ? DIVIDE_RECT_VER : DIVIDE_RECT_HOR; }
    if ((eBC > Max) || (exBC > xMax) || (eyBC > yMax))
    { uResult |= (fABVert) ? DIVIDE_RECT_HOR : DIVIDE_RECT_VER; }
  }
  else
  { REAL exAC = fabs(vertices.x1 - vertices.x3);
    REAL eyAC = fabs(vertices.y1 - vertices.y3);
    REAL eAC = sqrt(sq(exAC) + sq(eyAC));
    if ((eAB < Min) && (eBC < Min) && (eAC < Min)) return 0;
    if (( eAB >	 Max) || ( eBC >  Max) || ( eAC >  Max) ||
	(exAB > xMax) || (exBC > xMax) || (exAC > xMax) ||
	(eyAB > yMax) || (eyBC > yMax) || (eyAC > yMax))
    return DIVIDE_TRIANGLE;
  }

  // test the refinement criteria
  BOOL fPosElement = (vertices.y1 > 0) || (vertices.y2 > 0) ||
		     (vertices.y3 > 0);
  if (!cmdargs.fNoPosRef || !fPosElement)
  { for(UINT i = 0; i < cRefine; ++i)
    { UINT uType       = axlatrefine[i].uMeasure;
      REAL MaxDistance = axlatrefine[i].Distance;
      REAL a = RefineXform(apfRefine[i](vertices.x1, vertices.y1), uType);
      REAL b = RefineXform(apfRefine[i](vertices.x2, vertices.y2), uType);
      REAL c = RefineXform(apfRefine[i](vertices.x3, vertices.y3), uType);
      if (fRect)
      { REAL d = RefineXform(apfRefine[i](vertices.x4, vertices.y4), uType);
	if ((fABVert && !fNoVer) || (!fABVert && !fNoHor))
	{ if ((fabs(a - b) > MaxDistance) || (fabs(c - d) > MaxDistance))
	    uResult = (fABVert) ? DIVIDE_RECT_VER : DIVIDE_RECT_HOR;
	}
	if ((!fABVert && !fNoVer) || (fABVert && !fNoHor))
	{ if ((fabs(b - c) > MaxDistance) || (fabs(a - d) > MaxDistance))
	    uResult = (fABVert) ? DIVIDE_RECT_HOR : DIVIDE_RECT_VER;
	}
      }
      else
      { if ((fabs(a - b) > MaxDistance) ||
	    (fabs(b - c) > MaxDistance) ||
	    (fabs(a - c) > MaxDistance)) return DIVIDE_TRIANGLE;
      }
    }
  }

  // check aspect ratio
  if (fRect && (uResult != (DIVIDE_RECT_VER | DIVIDE_RECT_HOR)))
  { if (uResult & DIVIDE_RECT_VER) { if (fABVert) eAB /= 2.0; else eBC /= 2.0; }
    if (uResult & DIVIDE_RECT_HOR) { if (fABVert) eBC /= 2.0; else eAB /= 2.0; }
    if ((eAB/eBC > cmdargs.MaxRatio) || (eBC/eAB > cmdargs.MaxRatio))
    { uResult |= ((fABVert && (eAB > eBC)) || (!fABVert && (eAB < eBC)))
	       ? DIVIDE_RECT_VER : DIVIDE_RECT_HOR;
    }
  }
   INODE ainode[4];
   ElementNodes(ptri, ainode);
   if(!fRect)
  {
  
   REAL angle1 = ComputeAngle(ainode[0],ainode[1],ainode[2]);
   REAL angle2 = ComputeAngle(ainode[1],ainode[2],ainode[0]);
   REAL angle3 = ComputeAngle(ainode[2],ainode[0],ainode[1]);
   //if (((angle1 > PI/2) || (angle2 > PI/2) || (angle3 > PI/2))&&ptri->uLabel)
   //return DIVIDE_TRIANGLE;
  }   
  return uResult;
}


///////////////////////////////////////////////////////////////////////////////
////////////////////  FUNCTIONS THAT GRAPH THE TRIANGLES  /////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::SetupGraphicsWindow
// Scope:	 Private
// Description:	 This function opens the graphics window
void Refine::SetupGraphicsWindow ()
{
  int nx, ny;
  cxBorder = 10;
  cyBorder = 10;

  switch (cmdargs.nResolution)
  { case RES_640x480:
      nx = 640;
      ny = 480;
      break;

    case RES_800x600:
      nx = 800;
      ny = 600;
      break;

    case RES_1024x768:
      cx = 1024;
      cy = 768;
      break;
  }
  xDelta = xMax - xMin;
  yDelta = yMax - yMin;
  if (cmdargs.fKeepAspect)
  { if (xDelta > yDelta) yDelta = xDelta;
    else		 xDelta = yDelta;
  }

  cxBorder = 10;
  cyBorder = 10;
  cx = ny - 2*cxBorder - 23;
  cy = ny - 2*cyBorder - 23;
  GROpenGraphWin("Initial Mesh", "Mesh", -1, -1, nx, ny, RedrawMesh);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::CloseWindow
// Scope:	 Private
// Description:	 This function opens the graphics window
//void Refine::CloseWindow ()
//{
//  closegraph();
//}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::DrawMesh
// Scope:	 Public
// Description:	 This function draws the mesh.
void Refine::DrawMesh ()
{
  UINT ctri = datri.GetElements();
  for(ITRI itri = 0; itri < ctri; ++itri)
    DrawElement(itri);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::DrawElement
// Scope:	 Private
// Description:	 This function draws a triangle.
void Refine::DrawElement (PTRI ptri)
{
  // initialize the vertices structure
  VERTICES vertices;
  InitVerticesStructure(&vertices, ptri);

  // map the real coordinates to the screen coordinates
  int an[10];
  an[0] = xMAP(vertices.x1);
  an[1] = yMAP(vertices.y1);
  an[2] = xMAP(vertices.x2);
  an[3] = yMAP(vertices.y2);
  an[4] = xMAP(vertices.x3);
  an[5] = yMAP(vertices.y3);

  GRSetColor(GR_WHITE);
  if (vertices.fRect)
  { an[6] = xMAP(vertices.x4);
    an[7] = yMAP(vertices.y4);
    an[8] = an[0];
    an[9] = an[1];
    GRDrawPoly(5, an);
  }
  else
  { an[6] = an[0];
    an[7] = an[1];
    GRDrawPoly(4, an);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::DrawTriangles
// Scope:	 Private
// Description:	 This function draw the triangles in the node adjacency
//		 structure.
void Refine::DrawTriangles (NADJ &nadj)
{
  UINT i;
  PNODE pnode = danode.GetPointer(nadj.inode);
  int x = xMAP(pnode->x);
  int y = yMAP(pnode->y);

  int an[66];
  UINT c = nadj.cnode;
  for(i = 0; i < c; ++i)
  { pnode = danode.GetPointer(nadj.ainode[i]);
    an[2*i]   = xMAP(pnode->x);
    an[2*i+1] = yMAP(pnode->y);
  }
  an[2*c]   = an[0];
  an[2*c+1] = an[1];

  // draw the triangles
  GRSetColor(GR_WHITE);
  GRDrawPoly(c+1, an);
  for(i = 0; i < c; ++i) GRLine(x, y, an[2*i], an[2*i+1]);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::DrawElement
// Scope:	 Private
// Description:	 This function draws a triangle.
void Refine::DrawElement (ITRI itri)
{
  PTRI ptri = datri.GetPointer(itri);
  DrawElement(ptri);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::FillPoly
// Scope:	 Private
// Description:	 This function draws and fills a polygon.  The polygon is
//		 limited to 32 sides
void Refine::FillPoly (UINT c, INODE *ainode, int nFColor, int nBColor)
{
  // return if the number of sides is greater than 32
  if (c > 32) return;

  // draw and fill the polygon
  GRSetColor(nFColor);
  GRSetFillColor(nBColor);
  int an[64];
  for(UINT i = 0; i < c; ++i)
  { PNODE pnode = danode.GetPointer(ainode[i]);
    an[2*i]   = xMAP(pnode->x);
    an[2*i+1] = yMAP(pnode->y);
  }
  GRFillPoly(c, an);
}


///////////////////////////////////////////////////////////////////////////////
//////////////////  FUNCTIONS THAT ENHANCE THE MESH QUALITY  //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::EnhanceMeshEdges
// Scope:	 Private
// Description:	 This function enhances the mesh quality by swapping edges
//		 of adjacent triangles.
void Refine::EnhanceMeshEdges ()
{
  UINT i;

  // allocate and initialize the visit flag array
  UINT cedge = daedge.GetElements();
  PBOOL afVisited = new BOOL[cedge];
  PBOOL pf = afVisited;
  for(i = 0; i < cedge; ++i, ++pf) *pf = FALSE;

  // mark the edges of the rectangular elements as visited so that
  // they are not swapped.  also mark the edges of those triangles
  // with the STATUS_WAS_RECT flag as visited.
  UINT ctri = datri.GetElements();
  PTRI ptri = datri.GetPointer(0);
  for(i = 0; i < ctri; ++i, ++ptri)
    if (ptri->iedgeAD1 != -1u)
    { afVisited[ptri->iedgeAB1] = TRUE;
      afVisited[ptri->iedgeBC1] = TRUE;
      afVisited[ptri->iedgeCD1] = TRUE;
      afVisited[ptri->iedgeAD1] = TRUE;
    }
    else if (ptri->uStatus & STATUS_WAS_RECT)
    { afVisited[ptri->iedgeAB1] = TRUE;
      afVisited[ptri->iedgeBC1] = TRUE;
      afVisited[ptri->iedgeAC1] = TRUE;
    }

  // loop through each triangle
  ptri = datri.GetPointer(0);
  for(i = 0; i < ctri; ++i, ++ptri)
  { if (ptri->iedgeAD1 != -1u) continue;
    BOOL fDone;
    do
    { IEDGE aiedge[3];
      aiedge[0] = ptri->iedgeAB1;
      aiedge[1] = ptri->iedgeBC1;
      aiedge[2] = ptri->iedgeAC1;
      fDone = TRUE;
      IEDGE *piedge = aiedge;
      for(UINT j = 0; j < 3; ++j, ++piedge)
      { if (!afVisited[*piedge])
	{ afVisited[*piedge] = TRUE;
	  if (daedge.GetPointer(*piedge)->uLabel) continue;
	  EADJ eadj;
	  InitEdgeAdjStructure(eadj, i, *piedge, j);
#if 0
	  if (cmdargs.uVisual == VISUAL_CONSTRUCT)
	  { INODE  ainode[4];
	    ainode[0] = eadj.inodeA;
	    ainode[1] = eadj.inodeC1;
	    ainode[2] = eadj.inodeB;
	    ainode[3] = eadj.inodeC2;
	    FillPoly(4, ainode, GR_YELLOW, GR_BLACK);
	    if (cmdargs.cNoPress != -1)
	    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	      else { if (getch() == 'q') exit(20); }
	    }
	  }
#endif
	  BOOL fSwapped = fSwapEdges(eadj);
#if 0
	  if (cmdargs.uVisual == VISUAL_CONSTRUCT)
	  { DrawElement(eadj.itri1);
	    DrawElement(eadj.itri2);
	    if (cmdargs.cNoPress != -1)
	    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	      else { if (getch() == 'q') exit(20); }
	    }
	  }
#endif
	  if (fSwapped) { fDone = FALSE; break; }
	}
      }
    } while (!fDone);
  }

  // deallocate flag array and return
  delete [] afVisited;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::InitEdgeAdjStructure
// Scope:	 Private
// Description:	 This function initializes the edge adjacency structure.
void Refine::InitEdgeAdjStructure (EADJ &eadj, ITRI itri1, IEDGE iedgeAB, UINT iEdge)
{
  // initialize local variables
  PEDGE pedge = daedge.GetPointer(iedgeAB);
  PTRI	ptri  = datri.GetPointer(itri1);
  eadj.itri1   = itri1;
  eadj.iedgeAB = iedgeAB;
  eadj.inodeA  = pedge->inodeA;
  eadj.inodeB  = pedge->inodeB;

  PEDGE pedgeT;
  INODE inode1, inode2;
  switch (iEdge)
  { case EDGE_AB:
      eadj.itri2 = ptri->itriAB1;
      pedgeT = daedge.GetPointer(ptri->iedgeBC1);
      inode1 = pedgeT->inodeA;
      inode2 = pedgeT->inodeB;
      if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
      { eadj.inodeC1  = (eadj.inodeA == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeAC1;
	eadj.iedgeAC1 = ptri->iedgeBC1;
	eadj.itriBC1  = ptri->itriAC1;
	eadj.itriAC1  = ptri->itriBC1;
      }
      else
      { eadj.inodeC1  = (eadj.inodeB == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeBC1;
	eadj.iedgeAC1 = ptri->iedgeAC1;
	eadj.itriBC1  = ptri->itriBC1;
	eadj.itriAC1  = ptri->itriAC1;
      }
      break;

    case EDGE_BC:
      eadj.itri2 = ptri->itriBC1;
      pedgeT = daedge.GetPointer(ptri->iedgeAC1);
      inode1 = pedgeT->inodeA;
      inode2 = pedgeT->inodeB;
      if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
      { eadj.inodeC1  = (eadj.inodeA == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeAB1;
	eadj.iedgeAC1 = ptri->iedgeAC1;
	eadj.itriBC1  = ptri->itriAB1;
	eadj.itriAC1  = ptri->itriAC1;
      }
      else
      { eadj.inodeC1  = (eadj.inodeB == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeAC1;
	eadj.iedgeAC1 = ptri->iedgeAB1;
	eadj.itriBC1  = ptri->itriAC1;
	eadj.itriAC1  = ptri->itriAB1;
      }
      break;

    case EDGE_AC:
      eadj.itri2 = ptri->itriAC1;
      pedgeT = daedge.GetPointer(ptri->iedgeAB1);
      inode1 = pedgeT->inodeA;
      inode2 = pedgeT->inodeB;
      if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
      { eadj.inodeC1  = (eadj.inodeA == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeBC1;
	eadj.iedgeAC1 = ptri->iedgeAB1;
	eadj.itriBC1  = ptri->itriBC1;
	eadj.itriAC1  = ptri->itriAB1;
      }
      else
      { eadj.inodeC1  = (eadj.inodeB == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeAB1;
	eadj.iedgeAC1 = ptri->iedgeBC1;
	eadj.itriBC1  = ptri->itriAB1;
	eadj.itriAC1  = ptri->itriBC1;
      }
      break;
  }

  ptri = datri.GetPointer(eadj.itri2);
  if (eadj.iedgeAB == ptri->iedgeAB1)
  { pedgeT = daedge.GetPointer(ptri->iedgeBC1);
    inode1 = pedgeT->inodeA;
    inode2 = pedgeT->inodeB;
    if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
    { eadj.inodeC2  = (eadj.inodeA == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeAC1;
      eadj.iedgeAC2 = ptri->iedgeBC1;
      eadj.itriBC2  = ptri->itriAC1;
      eadj.itriAC2  = ptri->itriBC1;
    }
    else
    { eadj.inodeC2  = (eadj.inodeB == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeBC1;
      eadj.iedgeAC2 = ptri->iedgeAC1;
      eadj.itriBC2  = ptri->itriBC1;
      eadj.itriAC2  = ptri->itriAC1;
    }
  }
  else if (eadj.iedgeAB == ptri->iedgeBC1)
  { pedgeT = daedge.GetPointer(ptri->iedgeAC1);
    inode1 = pedgeT->inodeA;
    inode2 = pedgeT->inodeB;
    if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
    { eadj.inodeC2  = (eadj.inodeA == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeAB1;
      eadj.iedgeAC2 = ptri->iedgeAC1;
      eadj.itriBC2  = ptri->itriAB1;
      eadj.itriAC2  = ptri->itriAC1;
    }
    else
    { eadj.inodeC2  = (eadj.inodeB == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeAC1;
      eadj.iedgeAC2 = ptri->iedgeAB1;
      eadj.itriBC2  = ptri->itriAC1;
      eadj.itriAC2  = ptri->itriAB1;
    }
  }
  else if (eadj.iedgeAB == ptri->iedgeAC1)
  { pedgeT = daedge.GetPointer(ptri->iedgeAB1);
    inode1 = pedgeT->inodeA;
    inode2 = pedgeT->inodeB;
    if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
    { eadj.inodeC2  = (eadj.inodeA == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeBC1;
      eadj.iedgeAC2 = ptri->iedgeAB1;
      eadj.itriBC2  = ptri->itriBC1;
      eadj.itriAC2  = ptri->itriAB1;
    }
    else
    { eadj.inodeC2  = (eadj.inodeB == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeAB1;
      eadj.iedgeAC2 = ptri->iedgeBC1;
      eadj.itriBC2  = ptri->itriAB1;
      eadj.itriAC2  = ptri->itriBC1;
    }
  }

  BOOL fBndryA, fBndryB, fBndryC1, fBndryC2;
  fBndryA = fBndryB = fBndryC1 = fBndryC2 = FALSE;
  if (daedge.GetPointer(eadj.iedgeBC1)->uLabel) { fBndryB = TRUE; fBndryC1 = TRUE; }
  if (daedge.GetPointer(eadj.iedgeAC1)->uLabel) { fBndryA = TRUE; fBndryC1 = TRUE; }
  if (daedge.GetPointer(eadj.iedgeBC2)->uLabel) { fBndryB = TRUE; fBndryC2 = TRUE; }
  if (daedge.GetPointer(eadj.iedgeAC2)->uLabel) { fBndryA = TRUE; fBndryC2 = TRUE; }
  eadj.fBndryAB = fBndryA  && fBndryB;
  eadj.fBndryC	= fBndryC1 && fBndryC2;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::fSwapEdges
// Scope:	 Private
// Description:	 This function swaps the edge common edge if the resulting
//		 triangles will have a greater quality factor.	TRUE is
//		 returned if the edge is swapped; otherwise FALSE is returned.
BOOL Refine::fSwapEdges (EADJ &eadj)
{
  // compute triangle quality factors ... nodes must be specified in a
  // counter-clockwise order for this to work!
  INODE inodeA = eadj.inodeA;
  INODE inodeB = eadj.inodeB;
  REAL Qual1A = ComputeQuality(inodeA, eadj.inodeC1, inodeB);
  if (Qual1A < 0.0)
  { Qual1A = -Qual1A;
    inodeA = inodeB;
    inodeB = eadj.inodeA;
  }
  REAL Qual2A = ComputeQuality(inodeA, inodeB, eadj.inodeC2);
  REAL Qual1B = ComputeQuality(inodeB, eadj.inodeC2, eadj.inodeC1);
  REAL Qual2B = ComputeQuality(inodeA, eadj.inodeC1, eadj.inodeC2);
  REAL QualA  = (Qual1A < Qual2A) ? Qual1A : Qual2A;
  REAL QualB  = (Qual1B < Qual2B) ? Qual1B : Qual2B;

  // swapping edges does not enhance mesh quality
  if (( eadj.fBndryAB &&  eadj.fBndryC && (QualB <= QualA))	  ||
      (!eadj.fBndryAB &&  eadj.fBndryC && (QualB <= QualA / 0.9)) ||
      ( eadj.fBndryAB && !eadj.fBndryC && (QualB <= QualA * 0.9)) ||
      (!eadj.fBndryAB && !eadj.fBndryC && (QualB <= QualA)))
    return(FALSE);

  // swap edges
  PTRI ptri1 = datri.GetPointer(eadj.itri1);
  PTRI ptri2 = datri.GetPointer(eadj.itri2);
  PEDGE pedge = daedge.GetPointer(eadj.iedgeAB);
  pedge->inodeA	 = eadj.inodeC1;
  pedge->inodeB	 = eadj.inodeC2;
  ptri1->iedgeAB1 = eadj.iedgeAB;
  ptri1->iedgeBC1 = eadj.iedgeAC1;
  ptri1->iedgeAC1 = eadj.iedgeAC2;
  ptri2->iedgeAB1 = eadj.iedgeAB;
  ptri2->iedgeBC1 = eadj.iedgeBC1;
  ptri2->iedgeAC1 = eadj.iedgeBC2;
  ptri1->itriAB1  = eadj.itri2;
  ptri1->itriBC1  = eadj.itriAC1;
  ptri1->itriAC1  = eadj.itriAC2;
  ptri2->itriAB1  = eadj.itri1;
  ptri2->itriBC1  = eadj.itriBC1;
  ptri2->itriAC1  = eadj.itriBC2;
  PTRI ptriT = datri.GetPointer(eadj.itriAC2);
  if	  (ptriT->itriAB1 == eadj.itri2) ptriT->itriAB1 = eadj.itri1;
  else if (ptriT->itriBC1 == eadj.itri2) ptriT->itriBC1 = eadj.itri1;
  else if (ptriT->itriAC1 == eadj.itri2) ptriT->itriAC1 = eadj.itri1;
  ptriT = datri.GetPointer(eadj.itriBC1);
  if	  (ptriT->itriAB1 == eadj.itri1) ptriT->itriAB1 = eadj.itri2;
  else if (ptriT->itriBC1 == eadj.itri1) ptriT->itriBC1 = eadj.itri2;
  else if (ptriT->itriAC1 == eadj.itri1) ptriT->itriAC1 = eadj.itri2;
  return(TRUE);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::ComputeQuality
// Scope:	 Private
// Description:	 This function computes the quality of the triangle or
//		 rectangle specified by three or four nodes.
REAL Refine::ComputeQuality (INODE inodeA, INODE inodeB, INODE inodeC, INODE inodeD)
{
  PNODE pnodeA = danode.GetPointer(inodeA);
  PNODE pnodeB = danode.GetPointer(inodeB);
  PNODE pnodeC = danode.GetPointer(inodeC);
  REAL x1 = pnodeA->x;
  REAL y1 = pnodeA->y;
  REAL x2 = pnodeB->x;
  REAL y2 = pnodeB->y;
  REAL x3 = pnodeC->x;
  REAL y3 = pnodeC->y;
  REAL qual;

  // triangular element
  if (inodeD == -1u)
  { REAL xl = x1 - x2;
    REAL yl = y1 - y2;
    REAL xr = x3 - x2;
    REAL yr = y3 - y2;

    REAL det = xr * yl - xl * yr;
    REAL sss = (xl*xl + yl*yl) + (xr*xr + yr*yr) + (sq(xl-xr) + sq(yl-yr));
    qual = det * 3.46410161514 / sss;
  }

  // rectangular element
  else
  { REAL d1 = sqrt(sq(x1-x2)+sq(y1-y2));
    REAL d2 = sqrt(sq(x3-x2)+sq(y3-y2));
    if ((d1 == 0.0) || (d2 == 0.0)) qual = 0.0;
    else qual = (d1 < d2) ? d1/d2 : d2 / d1;
  }

  return qual;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::EnhanceMeshNodes
// Scope:	 Private
// Description:	 This function enhances the mesh quality by repositioning
//		 inner mesh points.
void Refine::EnhanceMeshNodes ()
{
  UINT i;

  // allocate and initialize the visit flag array
  UINT cnode = danode.GetElements();
  PBOOL afVisited = new BOOL[cnode];
  PBOOL pf = afVisited;
  for(i = 0; i < cnode; ++i, ++pf) *pf = FALSE;

  // tag the nodes of rectangular elements as visited so that
  // they are not enhanced.  also tag the nodes of those triangles
  // with the STATUS_WAS_RECT flag as visited.
  UINT ctri = datri.GetElements();
  PTRI ptri = datri.GetPointer(0);
  for(i = 0; i < ctri; ++i, ++ptri)
    if (ptri->iedgeAD1 != -1u)
    { PEDGE pedge = daedge.GetPointer(ptri->iedgeAB1);
      afVisited[pedge->inodeA] = TRUE;
      afVisited[pedge->inodeB] = TRUE;
      pedge = daedge.GetPointer(ptri->iedgeBC1);
      afVisited[pedge->inodeA] = TRUE;
      afVisited[pedge->inodeB] = TRUE;
      pedge = daedge.GetPointer(ptri->iedgeCD1);
      afVisited[pedge->inodeA] = TRUE;
      afVisited[pedge->inodeB] = TRUE;
      pedge = daedge.GetPointer(ptri->iedgeAD1);
      afVisited[pedge->inodeA] = TRUE;
      afVisited[pedge->inodeB] = TRUE;
    }
    else if (ptri->uStatus & STATUS_WAS_RECT)
    { PEDGE pedge = daedge.GetPointer(ptri->iedgeAB1);
      afVisited[pedge->inodeA] = TRUE;
      afVisited[pedge->inodeB] = TRUE;
      pedge = daedge.GetPointer(ptri->iedgeBC1);
      afVisited[pedge->inodeA] = TRUE;
      afVisited[pedge->inodeB] = TRUE;
      pedge = daedge.GetPointer(ptri->iedgeAC1);
      afVisited[pedge->inodeA] = TRUE;
      afVisited[pedge->inodeB] = TRUE;
    }

  // loop through each element
  ptri = datri.GetPointer(0);
  for(i = 0; i < ctri; ++i, ++ptri)
  { if (ptri->iedgeAD1 != -1u) continue;

    // enhance the nodes of the triangular elements
    INODE ainode[4];
    ElementNodes(ptri, ainode);
    INODE *pinode = ainode;
    for(UINT j = 0; j < 3; ++j, ++pinode)
      if (!afVisited[*pinode])
      { afVisited[*pinode] = TRUE;
	NADJ nadj;
	InitNodeAdjStructure(nadj, i, j, 0);
	if (!nadj.fBndry)
	{
#if 0
	  if (cmdargs.uVisual == VISUAL_CONSTRUCT)
	  { FillPoly(nadj.cnode, nadj.ainode, GR_YELLOW, GR_BLACK);
	    if (cmdargs.cNoPress != -1)
	    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	      else { if (getch() == 'q') exit(20); }
	    }
	  }
#endif
	  NudgeNode(nadj);
#if 0
	  if (cmdargs.uVisual == VISUAL_CONSTRUCT)
	  { DrawTriangles(nadj);
	    if (cmdargs.cNoPress != -1)
	    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	      else { if (getch() == 'q') exit(20); }
	    }
	  }
#endif
	}
      }
  }

  // deallocate flag array and return
  delete [] afVisited;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Private
// Description:	 This function determines the three nodes of a triangle
//
void Refine::ElementNodes ( PTRI ptri, INODE *ainode )
{
  IEDGE iedgeAD = ptri->iedgeAD1;
  PEDGE pedgeAB = daedge.GetPointer(ptri->iedgeAB1);
  PEDGE pedgeBC = daedge.GetPointer(ptri->iedgeBC1);
  INODE inode1 = pedgeAB->inodeA;
  INODE inode2 = pedgeAB->inodeB;
  INODE inode3 = pedgeBC->inodeA;
  INODE inode4 = pedgeBC->inodeB;

  // triangular element
  if (iedgeAD == -1u)
  { if	    (inode1 == inode3) { ainode[VERTEX_A] = inode2; ainode[VERTEX_B] = inode1; ainode[VERTEX_C] = inode4; }
    else if (inode1 == inode4) { ainode[VERTEX_A] = inode2; ainode[VERTEX_B] = inode1; ainode[VERTEX_C] = inode3; }
    else if (inode2 == inode3) { ainode[VERTEX_A] = inode1; ainode[VERTEX_B] = inode2; ainode[VERTEX_C] = inode4; }
    else		       { ainode[VERTEX_A] = inode1; ainode[VERTEX_B] = inode2; ainode[VERTEX_C] = inode3; }
    ainode[VERTEX_D] = -1u;
  }
  
  // rectangular element
  else
  { PEDGE pedgeCD = daedge.GetPointer(ptri->iedgeCD1);
    INODE inode5 = pedgeCD->inodeA;
    INODE inode6 = pedgeCD->inodeB;
    if ((inode1 == inode3) || (inode1 == inode4)) { ainode[VERTEX_A] = inode2; ainode[VERTEX_B] = inode1; }
    else					  { ainode[VERTEX_A] = inode1; ainode[VERTEX_B] = inode2; }
    if ((inode5 == inode3) || (inode5 == inode4)) { ainode[VERTEX_C] = inode5; ainode[VERTEX_D] = inode6; }
    else					  { ainode[VERTEX_C] = inode6; ainode[VERTEX_D] = inode5; }
  }
  
}

/*
///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::fCCWorder
// Scope:	 Private
// Description:	 This function determines if rotating segment 12 to 13 is in
//		 the counter-clockwise direction.
BOOL Refine::fCCWorder (INODE inode1, INODE inode2, INODE inode3)
{
  PNODE pnode1 = danode.GetPointer(inode1);
  PNODE pnode2 = danode.GetPointer(inode2);
  PNODE pnode3 = danode.GetPointer(inode3);
  REAL	x1     = pnode1->x;
  REAL	y1     = pnode1->y;
  REAL	x2     = pnode2->x;
  REAL	y2     = pnode2->y;
  REAL	x3     = pnode3->x;
  REAL	y3     = pnode3->y;
  REAL	x12    = x2 - x1;
  REAL	y12    = y2 - y1;
  REAL	x13    = x3 - x1;
  REAL	y13    = y3 - y1;
  REAL	d12    = sqrt(x12*x12 + y12*y12);
  REAL	d13    = sqrt(x13*x13 + y13*y13);
  REAL	arg12  = x12/d12;
  REAL	arg13  = x13/d13;
  if (arg12 < -1.0) arg12 = 1.0; else if (arg12 > 1.0) arg12 = 1.0;
  if (arg13 < -1.0) arg13 = 1.0; else if (arg13 > 1.0) arg13 = 1.0;
  REAL	a12    = acos(arg12);
  REAL	a13    = acos(arg13);

  if (y12 < 0) a12 = 2*PI - a12;
  if (y13 < 0) a13 = 2*PI - a13;
  return (a13 > a12) ? TRUE : FALSE;
}*/

///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::fCCWorder
// Scope:	 Private
// Description:	 This function determines if rotating segment 12 to 13 is in
//		 the counter-clockwise direction.
BOOL Refine::fCCWorder (INODE inode1, INODE inode2, INODE inode3)
{
  PNODE pnode1 = danode.GetPointer(inode1);
  PNODE pnode2 = danode.GetPointer(inode2);
  PNODE pnode3 = danode.GetPointer(inode3);
  REAL	x1     = pnode1->x;
  REAL	y1     = pnode1->y;
  REAL	x2     = pnode2->x;
  REAL	y2     = pnode2->y;
  REAL	x3     = pnode3->x;
  REAL	y3     = pnode3->y;
  REAL detleft, detright, det;
  REAL detsum, errbound;
  
  detleft  = (x1 - x3) * (y2 - y3);
  detright = (y1 - y3) * (x2 - x3);
  det = detleft - detright;
  
  if (detleft > 0.0) {
    if (detright <= 0.0) {
      return det>0;
    } else {
      detsum = detleft + detright;
    }
  } else if (detleft < 0.0) {
    if (detright >= 0.0) {
      return det>0;
    } else {
      detsum = -detleft - detright;
    }
  } else {
    return det>0;
  }
  
  /* Error bounds for orientation and incircle tests. */
  REAL lastcheck;
  int every_other = 1;
  REAL half = 0.5;
  REAL epsilon = 1.0;
  REAL splitter = 1.0;
  REAL check = 1.0;
  
  do {
    lastcheck = check;
    epsilon *= half;
    if (every_other) {
      splitter *= 2.0;
    }
    every_other = !every_other;
    check = 1.0 + epsilon;
  } while ((check != 1.0) && (check != lastcheck));
  splitter += 1.0;
 
  REAL ccwerrboundA = (3.0 + 16.0 * epsilon) * epsilon;
  errbound = ccwerrboundA * detsum;
  if ((det >= errbound) || (-det >= errbound)) {
    return det>0;
  }
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::GetElementInfo
// Scope:	 Private
// Description:	 This function initialize the ainode, aiedge, aitri and auLabel
//		 arrays with the elements node indices, edge indices, adjacent
//		 triangle indices and label indices.
void Refine::GetElementInfo(ITRI itri, INODE *ainode, IEDGE *aiedge, ITRI *aitri, UINT *auLabel)
{
  PTRI ptri = datri.GetPointer(itri);
  ElementNodes(ptri,ainode);
  IEDGE iedgeAB, iedgeBC, iedgeCD, iedgeAD;
  iedgeAB = aiedge[0] = ptri->iedgeAB1;
  iedgeBC = aiedge[1] = ptri->iedgeBC1;
  iedgeCD = aiedge[2] = ptri->iedgeCD1;
  iedgeAD = aiedge[3] = ptri->iedgeAD1;
  aitri[0] = ptri->itriAB1;
  aitri[1] = ptri->itriBC1;
  aitri[2] = ptri->itriCD1;
  aitri[3] = ptri->itriAD1;
  auLabel[0] = daedge.GetPointer(iedgeAB)->uLabel;
  auLabel[1] = daedge.GetPointer(iedgeBC)->uLabel;
  auLabel[2] = daedge.GetPointer(iedgeCD)->uLabel;
  auLabel[3] = (iedgeAD != -1u) ? daedge.GetPointer(iedgeAD)->uLabel : -1u;
}


// vertices to edge map
UINT aauVertices2Edge[4][4] =
{   //		    VERTEX_A  VERTEX_B	VERTEX_C  VERTEX_D
  { /* VERTEX_A */  -1u	    , EDGE_AB , EDGE_AC , EDGE_AD  },
  { /* VERTEX_B */  EDGE_AB , -1u     , EDGE_BC , -1u	   },
  { /* VERTEX_C */  EDGE_AC , EDGE_BC , -1u	, EDGE_CD  },
  { /* VERTEX_D */  EDGE_AD , -1u     , EDGE_CD , -1u	   }
};


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Private
// Description:	 This function initializes the static adjacency structure
//		 with the neighbors of the specified node.
//
void Refine::InitNodeAdjStructure ( NADJ &nadj, ITRI itri, UINT iVertex,
                                    UINT uIntLabel, BOOL fCW )
{
  INODE inode;
  INODE ainode[4];
  IEDGE aiedge[4];
  ITRI	aitri[4];
  UINT	auLabel[4];

  // initialize variables
  GetElementInfo(itri, ainode, aiedge, aitri, auLabel);
  UINT cVertices = (ainode[VERTEX_D] != -1u) ? 4 : 3;
  UINT cnode   = 0;
  nadj.cnode   = 0;
  nadj.inode   = ainode[iVertex];
  nadj.fBndry  = TRUE;
  nadj.fGotAll = FALSE;


  // determine the direction to rotate
  int fCCW = fCCWorder(ainode[VERTEX_A], ainode[VERTEX_B], ainode[VERTEX_C]);
  int nNextVertex = (fCCW == fCW) ? 1 : -1;
  int i	 = (iVertex + nNextVertex + cVertices) % cVertices;
  int i1 = aauVertices2Edge[iVertex][i];
  nadj.ainode[0] = inode = ainode[i];
  nadj.aiedge[0] = aiedge[i1];
  nadj.aielem[0] = itri;
  BOOL fBndry = auLabel[i1] != uIntLabel;

  // loop around and find the nodes
  for (;;)
  {
    for (i = 0; i < cVertices; ++i)
    { if (ainode[i] == nadj.inode) iVertex = i;
      if (ainode[i] ==	    inode) i1	   = i;
    }
    UINT iA = (iVertex + 1) % cVertices;
    UINT iB = (iVertex + cVertices - 1) % cVertices;
    UINT iV = (i1 == iA) ? iB : iA;
    UINT iE = aauVertices2Edge[iVertex][iV];

    nadj.auLabel[cnode] = datri.GetPointer(itri)->uLabel;
    ++cnode;
    nadj.ainode[cnode] = inode = ainode[iV];
    nadj.aiedge[cnode] = aiedge[iE];
    nadj.aielem[cnode] = aitri[iE];
    fBndry = fBndry || (auLabel[iE] != uIntLabel);
    itri = aitri[iE];

    if (nadj.ainode[cnode] == nadj.ainode[0])
    { nadj.fGotAll = TRUE;
      break;
    }

    if (itri == (ITRI) -1) break;

    GetElementInfo(itri, ainode, aiedge, aitri, auLabel);
    cVertices = (ainode[VERTEX_D] != -1u) ? 4 : 3;
  }

  nadj.fBndry = fBndry;
  nadj.cnode  = (nadj.fGotAll) ? cnode : cnode + 1;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::NudgeNode
// Scope:	 Private
// Description:	 This function adjusts the position of an inner node in order
//		 to enhance the overall quality of the grid.
void Refine::NudgeNode (NADJ &nadj)
{
  INODE *ainode = nadj.ainode;
  INODE	 inode	= nadj.inode;
  PNODE	 pnode	= danode.GetPointer(inode);
  UINT	 cnode	= nadj.cnode;

  INODE *pinode;
  INODE	 inode1M,inode1R;
  INODE	 inode2M,inode2R;
  REAL	 Qual,Qual1,Qual2,QualT;	// triangle quality factors
  REAL	 xmin,ymin,xmax,ymax;
  REAL	 dx1,dy1,dx2,dy2,dx,dy;
  REAL	 xm1,ym1,xm2,ym2,xm,ym;
  REAL	 r1,r2,r;
  REAL	 a,b,c,beta;
  REAL	 xc1,yc1,xc2,yc2;
  REAL	 zx,zy;
  UINT	 m1,m2;
  UINT	 i;

  // find the triangles with the smallest interior angles
  m1	= m2	= -1u;
  Qual1 = Qual2 =  1.0;
  for(i = 0, pinode = ainode; i < cnode; ++i, ++pinode)
  { Qual = fabs(ComputeQuality(inode, pinode[0], pinode[1]));
    if	    (Qual < Qual1) { Qual2 = Qual1; Qual1 = Qual; m2 = m1; m1 = i; }
    else if (Qual < Qual2) { Qual2 = Qual;		  m2 = i; }
  }

  // determine (xmin,ymin) and (xmax,ymax) ... (xmax,ymax) is the point
  // where the level curves (circles) for the two worst triangles are equal
  // and tangent
  PNODE anode = danode.GetPointer(0);
  xmin	  = anode[inode].x;
  ymin	  = anode[inode].y;
  inode1M = ainode[m1];
  inode1R = ainode[m1+1];
  inode2M = ainode[m2];
  inode2R = ainode[m2+1];
  dx1	  = (anode[inode1R].x - anode[inode1M].x) * SCALE;
  dy1	  = (anode[inode1R].y - anode[inode1M].y) * SCALE;
  dx2	  = (anode[inode2R].x - anode[inode2M].x) * SCALE;
  dy2	  = (anode[inode2R].y - anode[inode2M].y) * SCALE;
  dx	  = dx1 - dx2;
  dy	  = dy1 - dy2;
  xm1	  = (anode[inode1R].x + anode[inode1M].x) / 2.0;
  ym1	  = (anode[inode1R].y + anode[inode1M].y) / 2.0;
  xm2	  = (anode[inode2R].x + anode[inode2M].x) / 2.0;
  ym2	  = (anode[inode2R].y + anode[inode2M].y) / 2.0;
  xm	  = xm1 - xm2;
  ym	  = ym1 - ym2;
  r1	  = sqrt(dx1*dx1 + dy1*dy1);
  r2	  = sqrt(dx2*dx2 + dy2*dy2);
  r	  = r1 + r2;
  a	  = r*r - dx*dx - dy*dy;
  b	  = ym*dx - xm*dy;
  c	  = xm*xm + ym*ym + r*r;
  beta	  = (a > 0.0) ? (b + sqrt(b*b + a*c))/a : 1.0;
  xc1	  = xm1 - beta * dy1;
  yc1	  = ym1 + beta * dx1;
  xc2	  = xm2 - beta * dy2;
  yc2	  = ym2 + beta * dx2;
  xmax	  = (xc1*r2 + xc2*r1) / r;
  ymax	  = (yc1*r2 + yc2*r1) / r;

  // bisection loop
  for(;;)
  { zx = fabs(xmin - xmax) / (fabs(xmin) + fabs(xmax) + EPS);
    zy = fabs(ymin - ymax) / (fabs(ymin) + fabs(ymax) + EPS);
    if (max(zx,zy) < TOLERANCE) { pnode->x = xmin; pnode->y = ymin; break; }
    pnode->x = (xmin + xmax) / 2.0;
    pnode->y = (ymin + ymax) / 2.0;
    QualT = 1.0;
    for(i = 0, pinode = ainode; i < cnode; ++i, ++pinode)
    { Qual = fabs(ComputeQuality(inode, pinode[0], pinode[1]));
      if (Qual < Qual1) { xmax = pnode->x; ymax = pnode->y; break; }
      QualT = min(QualT,Qual);
    }
    if (i == cnode) { xmin = pnode->x; ymin = pnode->y; Qual1 = QualT; }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::RateMesh
// Scope:	 Private
// Description:	 This function marks the triangles with obtuse angles.
void Refine::RateMesh ()
{
  // initialize local variables
  UINT cte = 0;
  UINT cre = 0;
  UINT cObtuse = 0;

  // scan for obtuse angles.
  UINT ctri = datri.GetElements();
  for(ITRI itri = 0; itri < ctri; ++itri)
  { PTRI ptri = datri.GetPointer(itri);
    INODE ainode[4];

    // rectangular element
    if (ptri->iedgeAD1 != -1u)
    { ElementNodes(ptri, ainode);
      FillPoly(4, ainode, GR_WHITE, GR_BLACK);
      ++cre;
    }

    // triangular element
    else
    { ElementNodes(ptri, ainode);
      REAL angle1 = ComputeAngle(ainode[0],ainode[1],ainode[2]);
      REAL angle2 = ComputeAngle(ainode[1],ainode[2],ainode[0]);
      REAL angle3 = ComputeAngle(ainode[2],ainode[0],ainode[1]);
      int nColor = GR_BLACK;
      if ((angle1 > PI/2) || (angle2 > PI/2) || (angle3 > PI/2))
      { ++cObtuse;
	 nColor = GR_YELLOW;
      }
      FillPoly(3, ainode, GR_WHITE, nColor);
      ++cte;
    }
  }

  int x,y;
  switch (cmdargs.nResolution)
  { case RES_640x480:	x = 480;  y = 10;  break;
    case RES_800x600:	x = 600;  y = 10;  break;
    case RES_1024x768:	x = 768;  y = 10;  break;
  }

  GRSetColor(GR_WHITE);
  sprintf(szGlobalBuf, "%d nodes", danode.GetElements());
  GRText(x+20, y,    szGlobalBuf);
  sprintf(szGlobalBuf, "%d edges", daedge.GetElements());
  GRText(x+20, y+20, szGlobalBuf);
  sprintf(szGlobalBuf, "%d triangles", cte);
  GRText(x+20, y+40, szGlobalBuf);
  sprintf(szGlobalBuf, "%d rectangles", cre);
  GRText(x+20, y+60, szGlobalBuf);
  sprintf(szGlobalBuf, "%d obtuse angles", cObtuse);
  GRText(x+20, y+80, szGlobalBuf);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::GeneratePostScriptFile
// Scope:	 Private
// Description:	 This function generates a postscript file.
void Refine::GeneratePostScriptFile ()
{
	cx = 9360;
	cy = 9360;
	cxBorder = 1440;
	cyBorder = 3240;
  GRPrintOpen("mesh");
  GRSetColor(GR_WHITE);
  ctri = datri.GetElements();
  for(ITRI itri = 0; itri < ctri; ++itri)
  { PTRI ptri = datri.GetPointer(itri);
    DrawElement(ptri);
#if 0
    INODE ainode[4];
    if (ptri->iedgeAD1 != -1u)
    { ElementNodes(ptri, ainode);
      FillPoly(4, ainode, GR_WHITE, GR_BLACK);
    }
    else
    { ElementNodes(ptri, ainode);
      FillPoly(3, ainode, GR_WHITE, GR_BLACK);
    }
#endif
  }
  GRPrintClose();
}


#if defined(DEBUG)
///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::Test1
// Scope:	 Public
// Description:	 This function verifies that the mesh is linked properly.
void Refine::Test1 ()
{
  UINT ctri = datri.GetElements();
  for(ITRI itri = 0; itri < ctri; ++itri)
  { PTRI ptri = datri.GetPointer(itri);
    Test2(ptri->itriAB1, itri, ptri->iedgeAB1);
    Test2(ptri->itriBC1, itri, ptri->iedgeBC1);
    Test2(ptri->itriAC1, itri, ptri->iedgeAC1);
    Test2(ptri->itriAD1, itri, ptri->iedgeAD1);
    Test2(ptri->itriAB2, itri, ptri->iedgeAB2);
    Test2(ptri->itriBC2, itri, ptri->iedgeBC2);
    Test2(ptri->itriAC2, itri, ptri->iedgeAC2);
    Test2(ptri->itriAD2, itri, ptri->iedgeAD2);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::Test2
// Scope:	 Public
// Description:	 This function verifies that element itri1 has an edge iedge2
//		 that is shared by element itri2.
void Refine::Test2 (ITRI itri1, ITRI itri2, IEDGE iedge2)
{
  if (itri1 == -1u) return;

  PTRI ptri1 = datri.GetPointer(itri1);
  if (((ptri1->itriAB1 == itri2) && (ptri1->iedgeAB1 == iedge2)) ||
      ((ptri1->itriBC1 == itri2) && (ptri1->iedgeBC1 == iedge2)) ||
      ((ptri1->itriAC1 == itri2) && (ptri1->iedgeAC1 == iedge2)) ||
      ((ptri1->itriAD1 == itri2) && (ptri1->iedgeAD1 == iedge2)) ||
      ((ptri1->itriAB2 == itri2) && (ptri1->iedgeAB2 == iedge2)) ||
      ((ptri1->itriBC2 == itri2) && (ptri1->iedgeBC2 == iedge2)) ||
      ((ptri1->itriAC2 == itri2) && (ptri1->iedgeAC2 == iedge2)) ||
      ((ptri1->itriAD2 == itri2) && (ptri1->iedgeAD2 == iedge2))) return;

  fprintf(stdout, "Mesh Construction Error!\n");
  fprintf(stdout, szDEBUG);
  fprintf(stdout, "Test2(itri1 = %u, itri2 = %u, iedge2 = %u)\n", itri1, itri2, iedge2);
  fprintf(stdout, "cnode:  %u\n", danode.GetElements());
  fprintf(stdout, "cedge:  %u\n", daedge.GetElements());
  fprintf(stdout, "ctri:   %u\n", datri.GetElements());
  fprintf(stdout, "cDEBUGDivision = %u\n", cDEBUGDivision);
  DebugDumpMesh();
  exit(1);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Refine::DebugDumpMesh
// Scope:	 Public
// Description:	 This function outputs the mesh in ASCII format.
void Refine::DebugDumpMesh ()
{
  // open output file
  static UINT c = 0;
  char sz[16];
  sprintf(sz, "debug.%05u", c++);
  FILE *file = fopen(sz, "wt");

  UINT ctri = datri.GetElements();
  UINT cedge = daedge.GetElements();

  fprintf(file, "TRIANGLES\n");
  fprintf(file, "      |     |		EDGES		       TRIANGLES       \n");
  fprintf(file, "  i   |  u  |	 AB    BC  AC/CD   AD	  AB	BC  AC/CD   AD \n");
  fprintf(file, "----- | --- | ----- ----- ----- -----	----- ----- ----- -----\n");
  for(UINT i = 0; i < ctri; ++i)
  { PTRI ptri = datri.GetPointer(i);
    fprintf(file, "%5u | ", i);
    if (ptri->uLabel != -1u) fprintf(file, "%3u | ", ptri->uLabel);  else fprintf(file, "    | ");
    if (ptri->iedgeAB1 != -1u) fprintf(file, "%5u ", ptri->iedgeAB1); else fprintf(file, "	");
    if (ptri->iedgeBC1 != -1u) fprintf(file, "%5u ", ptri->iedgeBC1); else fprintf(file, "	");
    if (ptri->iedgeAC1 != -1u) fprintf(file, "%5u ", ptri->iedgeAC1); else fprintf(file, "	");
    if (ptri->iedgeAD1 != -1u) fprintf(file, "%5u ", ptri->iedgeAD1); else fprintf(file, "  n/a ");
    fprintf(file, " ");
    if (ptri->itriAB1 != -1u) fprintf(file, "%5u ", ptri->itriAB1); else fprintf(file, "      ");
    if (ptri->itriBC1 != -1u) fprintf(file, "%5u ", ptri->itriBC1); else fprintf(file, "      ");
    if (ptri->itriAC1 != -1u) fprintf(file, "%5u ", ptri->itriAC1); else fprintf(file, "      ");
    if (ptri->itriAD1 != -1u) fprintf(file, "%5u" , ptri->itriAD1); else fprintf(file, "  n/a");
    fprintf(file, "\n");
    fprintf(file, "%5u | ", i);
    if (ptri->uLabel != -1u) fprintf(file, "%3u | ", ptri->uLabel);  else fprintf(file, "    | ");
    if (ptri->iedgeAB2 != -1u) fprintf(file, "%5u ", ptri->iedgeAB2); else fprintf(file, "	");
    if (ptri->iedgeBC2 != -1u) fprintf(file, "%5u ", ptri->iedgeBC2); else fprintf(file, "	");
    if (ptri->iedgeAC2 != -1u) fprintf(file, "%5u ", ptri->iedgeAC2); else fprintf(file, "	");
    if (ptri->iedgeAD2 != -1u) fprintf(file, "%5u ", ptri->iedgeAD2); else fprintf(file, "  n/a ");
    fprintf(file, " ");
    if (ptri->itriAB2 != -1u) fprintf(file, "%5u ", ptri->itriAB2); else fprintf(file, "      ");
    if (ptri->itriBC2 != -1u) fprintf(file, "%5u ", ptri->itriBC2); else fprintf(file, "      ");
    if (ptri->itriAC2 != -1u) fprintf(file, "%5u ", ptri->itriAC2); else fprintf(file, "      ");
    if (ptri->itriAD2 != -1u) fprintf(file, "%5u" , ptri->itriAD2); else fprintf(file, "  n/a");
    fprintf(file, "\n");
    fprintf(file, "----- | --- | ----- ----- ----- -----  ----- ----- ----- -----\n");
  }

  fprintf(file, "\nEDGES\n");
  fprintf(file, "      |     |	  NODES	  \n");
  fprintf(file, "  i   |  u  |	 A     B  \n");
  fprintf(file, "----- | --- | ----- -----\n");
  for(IEDGE iedge = 0; iedge < cedge; ++iedge)
  { PEDGE pedge = daedge.GetPointer(iedge);
    fprintf(file, "%5u | ", iedge);
    if (pedge->uLabel != -1u) fprintf(file, "%3u | ", pedge->uLabel); else fprintf(file, "    | ");
    fprintf(file, "%5u %5u\n", pedge->inodeA, pedge->inodeB);
  }

  fclose(file);
}
#endif
