/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  refine.exe
 FILE:	     mesh.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/04/94    KMK      initial implementation
 12/27/04    gdiso    add cgns output
 DESCRIPTION:

 This module contains the refine class.

-----------------------------------------------------------------------------*/


#ifndef __refine_h
#define __refine_h


// grid header
typedef struct tagGRIDHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  char szCoord1[LEN_IDENT+1];		// name of 1st coordinate
  char szCoord2[LEN_IDENT+1];		// name of 2nd coordinate
  UINT cMinDivisions;			// minimum divisions
  UINT cMaxDivisions;			// maximum divisions
  UINT cConstant;			// number of constants
  UINT cLabel;				// number of labels
  UINT cRefine;				// number of refinements
  UINT cRegLabel;			// number of region labels
  UINT cNode;				// number of nodes
  UINT cEdge;				// number of edges
  UINT cTriangle;			// number of triangles
} GRIDHEAD;


// mesh header
typedef struct tagMESHHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cConstant;			// number of constants
  UINT cLabel;				// number of labels
  UINT cArray;				// number of 1D arrays
  UINT cRegLabel;			// number of region labels
  UINT cBndryNode;			// number of boundary nodes
  UINT cNode;				// number of nodes
  UINT cEdge;				// number of edges
  UINT cTriangle;			// number of triangles
  BOOL fCylGeom;			// cylindrical geometry
} MESHHEAD;


// constant structure definition
typedef struct tagXLATCONST
{ char szName[LEN_IDENT+1];		// constant name
  UINT uType;				// constant type
  union
  { REAL r;				// value of real constant
    int	 n;				// value of integer constant
  } data;
} XLATCONST;


// label structure definition
typedef struct tagXLATLABEL
{ char szName[LEN_IDENT+1];		// label name
  UINT iIndex;				// label index
  UINT uType;				// label type
  UINT cNode;				// number of nodes
} XLATLABEL;


// refinement structure definition
typedef struct tagXLATREFINE
{ char szName[LEN_IDENT+1];		// refinement variable name
  UINT iIndex;				// refinement variable index
  UINT uMeasure;			// type of measurement
  REAL Distance;			// minimum "distance"
} XLATREFINE;


#define VERTEX_A	0		// vertex A
#define VERTEX_B	1		// vertex B
#define VERTEX_C	2		// vertex C
#define VERTEX_D	3		// vertex D

#define EDGE_AB		0		// edge AB
#define EDGE_BC		1		// edge BC
#define EDGE_AC		2		// edge AC
#define EDGE_CD		2		// edge CD
#define EDGE_AD		3		// edge AD

#define STATUS_DIV_AB	0x0001		// divide triangle adjacent to edge AB
#define STATUS_DIV_BC	0x0002		// divide triangle adjacent to edge BC
#define STATUS_DIV_AC	0x0004		// divide triangle adjacent to edge AC
#define STATUS_DIV_CD	0x0004		// divide triangle adjacent to edge CD
#define STATUS_DIV_AD	0x0008		// divide triangle adjacent to edge AD
#define STATUS_WAS_RECT 0x8000		// triangle was a rectangle ... do not enhance


typedef UINT INODE;			// inode
typedef UINT IEDGE;			// iedge
typedef UINT ITRI;			// itri


// node structure definition
typedef struct tagNODE			// node
{ REAL	x;				// x coordinate
  REAL	y;				// y coordinate
} NODE_t;
typedef NODE_t *PNODE;			// pnode


// edge structure definition
typedef struct tagEDGE			// edge
{ UINT	uLabel;				// edge label
  INODE inodeA;				// index of node A
  INODE inodeB;				// index of node B
} EDGE;
typedef EDGE *PEDGE;			// pedge


// triangle structure definition
#define iedgeCD	 iedgeAC
#define iedgeCD1 iedgeAC1
#define iedgeCD2 iedgeAC2
#define itriCD	 itriAC
#define itriCD1	 itriAC1
#define itriCD2	 itriAC2
typedef struct tagTRI			// tri
{ UINT	uLabel;				// region label
  IEDGE iedgeAB1;			// index of edge AB
  IEDGE iedgeBC1;			// index of edge BC
  IEDGE iedgeAC1;			// index of edge AC or CD
  IEDGE iedgeAD1;			// index of edge AD
  ITRI	itriAB1;			// index of tri. adj. to edge AB
  ITRI	itriBC1;			// index of tri. adj. to edge BC
  ITRI	itriAC1;			// index of tri. adj. to edge AC or CD
  ITRI	itriAD1;			// index of tri. adj. to edge AD
  IEDGE iedgeAB2;			// index of edge AB
  IEDGE iedgeBC2;			// index of edge BC
  IEDGE iedgeAC2;			// index of edge AC or CD
  IEDGE iedgeAD2;			// index of edge AD
  ITRI	itriAB2;			// index of tri. adj. to edge AB
  ITRI	itriBC2;			// index of tri. adj. to edge BC
  ITRI	itriAC2;			// index of tri. adj. to edge AC or CD
  ITRI	itriAD2;			// index of tri. adj. to edge AD
  UINT	cLevel;				// division level
  UINT	uStatus;			// status
} TRI;
typedef TRI *PTRI;			// ptri


//		     o b
//		     /\
//		    /  \
//		   /	\
//		  /	 \
//	     AB2 /	  \ BC1		    lower case = node
//		/	   \
//	       /	    \		    upper case = edge
//	      /	     AC3     \
//	  ab o----------------o bc
//	    / \		     / \
//	   /   \	    /	\
//	  /	\	   /	 \
//   AB1 /   BC3 \	  / AB3	  \ BC2
//	/	  \	 /	   \
//     /	   \	/	    \
//    /		    \  /	     \
//   /		     \/		      \
//  o----------------o-----------------o
//  a	   AC1	     ac	    AC2	       c


// divide structure
typedef struct tagTRIDIV		// tdiv
{ INODE inodeA;				// index of node A
  INODE inodeB;				// index of node B
  INODE inodeC;				// index of node C
  INODE inodeAB;			// index of edge AB's midpoint
  INODE inodeBC;			// index of edge BC's midpoint
  INODE inodeAC;			// index of edge AC's midpoint
  IEDGE iedgeAB1;			// index of 1st edge of AB
  IEDGE iedgeAB2;			// index of 2nd edge of AB
  IEDGE iedgeAB3;			// index of 3rd edge of AB
  IEDGE iedgeBC1;			// index of 1st edge of BC
  IEDGE iedgeBC2;			// index of 2nd edge of BC
  IEDGE iedgeBC3;			// index of 3rd edge of BC
  IEDGE iedgeAC1;			// index of 1st edge of AC
  IEDGE iedgeAC2;			// index of 2nd edge of AC
  IEDGE iedgeAC3;			// index of 3rd edge of AC
} TRIDIV;
typedef TRIDIV *PTRIDIV;		// ptdiv

//	b	BC1	 bc	BC2	   c
//	o----------------o-----------------o
//	|		 |		   |
//	|		 |		   |
//	|		 |		   |
//  AB2 |		 | V		   | CD1
//	|		 |		   |		lower case = node
//	|		 |		   |
//	|		 |		   |		upper case = edge
//	|	H	 |		   |
//   ab o----------------+-----------------o cd
//	|		 |		   |
//	|		 |		   |
//	|		 |		   |
//	|		 |		   |
//  AB1 |		 |		   | CD2
//	|		 |		   |
//	|		 |		   |
//	|		 |		   |
//	o----------------o-----------------o
//	a	AD1	ad	AD2	   d


// divide structure
typedef struct tagRECTDIV		// rdiv
{ INODE inodeA;				// index of node A
  INODE inodeB;				// index of node B
  INODE inodeC;				// index of node C
  INODE inodeD;				// index of node D
  INODE inodeAB;			// index of edge AB's midpoint
  INODE inodeBC;			// index of edge BC's midpoint
  INODE inodeCD;			// index of edge CD's midpoint
  INODE inodeAD;			// index of edge AD's midpoint
  IEDGE iedgeAB1;			// index of 1st edge of AB
  IEDGE iedgeAB2;			// index of 2nd edge of AB
  IEDGE iedgeBC1;			// index of 1st edge of BC
  IEDGE iedgeBC2;			// index of 2nd edge of BC
  IEDGE iedgeCD1;			// index of 1st edge of CD
  IEDGE iedgeCD2;			// index of 2nd edge of CD
  IEDGE iedgeAD1;			// index of 1st edge of AD
  IEDGE iedgeAD2;			// index of 2nd edge of AD
  IEDGE iedgeH;			       // index of horizontal edge
  IEDGE iedgeV;			       // index of vertical edge
} RECTDIV;
typedef RECTDIV *PRECTDIV;		// prdiv


//	a	AB1	 ab	AB2	   b
//	o----------------o----------------o
//	|		/ \		  |
//	|	       /   \		  |
//	|	      /	    \		  |
//	|	     /	     \		  |
//	|	    /	      \		  |	       lower case = node
//	|	   /	       \	  |
//	|	  /		\	  |	       upper case = edge
//   AB |	 /		 \	  | BC
//	|	/		  \	  |
//	|      /		   \	  |
//	|     /			    \	  |
//	|    /			     \	  |
//	|   /			      \	  |
//	|  /			       \  |
//	| /				\ |
//	|/				 \|
//	o---------------------------------o
//	d		CD		  c


// split structure
typedef struct tagRECTSPLIT		// rsplit
{ INODE inodeA;				// index of node A
  INODE inodeB;				// index of node B
  INODE inodeC;				// index of node C
  INODE inodeD;				// index of node D
  INODE inodeAB;			// index of node AB
  IEDGE iedgeAB1;			// index of 1st edge of AB
  IEDGE iedgeAB2;			// index of 2nd edge of AB
  IEDGE iedgeBC;			// index of BC edge
  IEDGE iedgeCD;			// index of CD edge
  IEDGE iedgeAD;			// index of AD edge
  ITRI	itri;				// index of rectangle
  ITRI	itriAB1;			// index of elem adj to 1st edge of AB
  ITRI	itriAB2;			// index of elem adj to 2nd edge of AB
  ITRI	itriBC;				// index of elem adj to BC edge
  ITRI	itriCD;				// index of elem adj to CD edge
  ITRI	itriAD;				// index of elem adj to AD edge
} RECTSPLIT;
typedef RECTSPLIT *PRECTSPLIT;		// prsplit


// vertices structure definition
typedef struct tagVERTICES		// vertices
{ REAL x1, y1;				// coordinates of 1st vertex
  REAL x2, y2;				// coordinates of 2nd vertex
  REAL x3, y3;				// coordinates of 3rd vertex
  REAL x4, y4;				// coordinates of 4th vertex
  BOOL fRect;				// set if element is rectangular
} VERTICES;
typedef VERTICES *PVERTICES;		// pvertices


// neighbor adjacency structure
typedef struct tagNADJ			// nadj
{ UINT	cnode;				// number of adjacent nodes
  UINT	inode;				// index of node under analysis
  BOOL	fBndry;				// set if node is a boundary node
  BOOL	fGotAll;			// set if all nodes were visited
  INODE ainode[64];			// indices of neighboring nodes
  IEDGE aiedge[64];			// indices of "spoke" edges
  ITRI	aielem[64];			// triangle indices
  UINT	auLabel[64];			// triangle labels
} NADJ;
typedef NADJ *PNADJ;			// pnadj


// edge adjacency structure
typedef struct tagEADJ			// eadj
{ ITRI	itri1;				// index of triangle 1
  ITRI	itri2;				// index of triangle 2
  ITRI	itriBC1;			// index of adjacent triangle
  ITRI	itriBC2;			// index of adjacent triangle
  ITRI	itriAC1;			// index of adjacent triangle
  ITRI	itriAC2;			// index of adjacent triangle
  IEDGE iedgeAB;			// index of adjacent edge
  IEDGE iedgeBC1;			// edge index
  IEDGE iedgeBC2;			// edge index
  IEDGE iedgeAC1;			// edge index
  IEDGE iedgeAC2;			// edge index
  INODE inodeA;				// index of adjacent edge node A
  INODE inodeB;				// index of adjacent edge node B
  INODE inodeC1;			// index of node C of triangle 1
  INODE inodeC2;			// index of node C of triangle 2
  BOOL	fBndryAB;			// edge connects bndy nodes?
  BOOL	fBndryC;			// swapped edge connects bndy nodes?
} EADJ;
typedef EADJ *PEADJ;			// peadj


// edge information structure
typedef struct tagEDGEINFO		// edgeinfo
{ REAL	ilen;				// integration length
  REAL	elen;				// edge length
  REAL	Area1;				// partial area 1
  REAL	Area2;				// partial area 2
  IEDGE iedge;				// edge index
  INODE inode;				// node index
  ITRI	ielem;				// element index
} EDGEINFO;


// node information structure
typedef struct tagNODEINFO		// nodeinfo
{ REAL Area;				// area of integration box
  UINT cNeighbors;			// number of neighbors
  UINT cTriangles;			// number of triangles
  EDGEINFO *aedgeinfo;			// edge information array
} NODEINFO;


#define DISTANCE_LINEAR	      0		// measure distance linearly
#define DISTANCE_LOG	      1		// measure distance logarithmically
#define DISTANCE_SLOG	      2		// measure distance w/ signed log

#define TYPE_EDGE	      7		// edge label
#define TYPE_REGION	      8		// region label

#define SIGNATURE_MESH	      "@~!__MESH__!~@"


#define DIVIDE_TRIANGLE	      0x0001	// divide triangle
#define DIVIDE_RECT_HOR	      0x0002	// divide rectangle horizontally
#define DIVIDE_RECT_VER	      0x0004	// divide rectangle vertically


class DANODE
{ private:
    UINT cSize;		      // total number of elements
    UINT uInc;		      // increment size
    UINT cElements;	      // number of elements in array
    NODE_t *dynarray;	      // dynamic array
    void IncSize ()	      { if (cElements == cSize)
				{ cSize += uInc;
				  dynarray = (NODE_t *) realloc(dynarray, cSize*sizeof(NODE_t));
				}
			      }

  public:
    UINT GetElements ()	      { return cElements;    }
    UINT Add (NODE_t t)	      { IncSize(); dynarray[cElements] = t; return cElements++; }
    void RemoveLast ()	      { --cElements;	     }
    void Flush ()	      { cElements = 0;	     }
    NODE_t *GetPointer (UINT i) { return dynarray + i; }
    NODE_t operator[]  (UINT i) { return dynarray[i];  }

  public:
     DANODE (UINT c = 2048)   { cElements = 0; uInc = cSize = c;
                                dynarray = (NODE_t*) calloc(c, sizeof(NODE_t)); }
    ~DANODE ()		      { free((void *) dynarray); }
};


class DAEDGE
{ private:
    UINT cSize;		      // total number of elements
    UINT uInc;		      // increment size
    UINT cElements;	      // number of elements in array
    EDGE *dynarray;	      // dynamic array
    void IncSize ()	      { if (cElements == cSize)
				{ cSize += uInc;
				  dynarray = (EDGE *) realloc(dynarray, cSize*sizeof(EDGE));
				}
			      }

  public:
    UINT GetElements ()	      { return cElements;    }
    UINT Add (EDGE t)	      { IncSize(); dynarray[cElements] = t; return cElements++; }
    void RemoveLast ()	      { --cElements;	     }
    void Flush ()	      { cElements = 0;	     }
    EDGE *GetPointer (UINT i) { return dynarray + i; }
    EDGE operator[]  (UINT i) { return dynarray[i];  }

  public:
     DAEDGE (UINT c = 2048)   { cElements = 0; uInc = cSize = c;
                                dynarray = (EDGE*) calloc(c, sizeof(EDGE)); }
    ~DAEDGE ()		      { free((void *) dynarray); }
};


class DATRI
{ private:
    UINT cSize;		      // total number of elements
    UINT uInc;		      // increment size
    UINT cElements;	      // number of elements in array
    TRI *dynarray;	      // dynamic array
    void IncSize ()	      { if (cElements == cSize)
				{ cSize += uInc;
				  dynarray = (TRI *) realloc(dynarray, cSize*sizeof(TRI));
				}
			      }

  public:
    UINT GetElements ()	      { return cElements;    }
    UINT Add (TRI t)	      { IncSize(); dynarray[cElements] = t; return cElements++; }
    void RemoveLast ()	      { --cElements;	     }
    void Flush ()	      { cElements = 0;	     }
    TRI *GetPointer (UINT i)  { return dynarray + i; }
    TRI operator[]  (UINT i)  { return dynarray[i];  }

  public:
     DATRI (UINT c = 2048)    { cElements = 0; uInc = cSize = c;
                                dynarray = (TRI*) calloc(c, sizeof(TRI)); }
    ~DATRI ()		      { free((void *) dynarray); }
};


// mesh class
class Refine
{ protected:
    DANODE danode;			// dynamic array of nodes
    DAEDGE daedge;			// dynamic array of edges
    DATRI  datri;			// dynamic array of triangles

  public:
    UINT cRefine;			// number of refinement statements
    XLATREFINE *axlatrefine;		// array of refinement variables

  protected:
    UINT cConstant;			// number of constants
    UINT cLabel;			// number of labels
    XLATCONST *axlatconst;		// constant array
    XLATLABEL *axlatlabel;		// label array
    char szGridName[LEN_FILENAME];	// name of grid file
    char szMeshName[LEN_FILENAME];	// name of mesh file
    GRIDHEAD gridhead;			// grid header

  protected:
    REAL xMin;				// minimum value of x coordinate
    REAL xMax;				// maximum value of x coordinate
    REAL yMin;				// minimum value of y coordinate
    REAL yMax;				// maximum value of y coordinate
    int	 cx;				// number of x screen coordinates
    int	 cy;				// number of y screen coordinates
    int	 cxBorder;			// border in x direction
    int	 cyBorder;			// border in y direction
    REAL xDelta;			// delta x spacing
    REAL yDelta;			// delta y spacing

  public:
    void SetupGraphicsWindow ();	// open graphics window
    void ReadGrid ();			// read grid
    void WriteMesh ();			// write mesh
    void WriteCGNS();                   // write cgns mesh
    void PromptUser ();			// prompt user for query data
    void RefineMesh ();			// refine mesh
    void DrawMesh ();			// draw the mesh
    void AsciiOutputMesh (char *);	// outputs the mesh in ASCII form
    void EnhanceMeshEdges ();		// enhance the mesh edges
    void EnhanceMeshNodes ();		// enhance the mesh nodes
    void RateMesh ();			// rate the mesh
    void ReorderNodes ();		// reorders the mesh nodes
    void GeneratePostScriptFile ();

  protected:
    UINT NumDiv (ITRI);
    void WriteLabelInfo(int, UINT, UINT, UINT);
    void WriteArrayInfo(int, UINT);
    void WriteAdjacencyInfo (int);
    UINT EdgeStatus (UINT, ITRI);
    REAL LengthAdjust (REAL, REAL, REAL, REAL);
    REAL AreaAdjust (REAL, REAL, REAL, REAL, REAL, REAL);
    void ComputeIntPB (REAL &, REAL &, INODE, INODE, INODE);
    void DivideElement (ITRI, IEDGE);
    void DivideTriangle (ITRI);
    ITRI DivideRectangle (ITRI, BOOL, BOOL);
    void SplitTriangle (ITRI);
    void SplitRectangle (ITRI);
    void InitDivideStructure (PTRIDIV, PTRI, BOOL);
    void InitDivideStructure (PRECTDIV, PTRI, BOOL, BOOL);
    void InitSplitStructure (RECTSPLIT &, ITRI);
    void SplitOrDivide (ITRI, BOOL &, BOOL &);
    BOOL fIsEdgeVertical (IEDGE);
    void FindNodes (IEDGE, IEDGE, INODE &, INODE &, INODE &, BOOL);
    void FindEdges (IEDGE &, IEDGE &, IEDGE, IEDGE, INODE, INODE, INODE, BOOL);
    void InitAdj1 (PTRI, IEDGE, ITRI);
    void InitAdj2 (PTRI, IEDGE, IEDGE, ITRI, ITRI);
    void InitAdj3 (PTRI, ITRI, ITRI);
    void InitVerticesStructure (PVERTICES, PTRI);
    UINT uDivideElement (ITRI);
    void InitEdgeAdjStructure (EADJ &, ITRI, IEDGE, UINT);
    BOOL fSwapEdges (EADJ &);
    REAL ComputeQuality (INODE, INODE, INODE, INODE = -1u);
    REAL ComputeAngle (INODE, INODE, INODE);
    void ElementNodes (PTRI, INODE *);
    BOOL fCCWorder (INODE, INODE, INODE);
    void GetElementInfo(ITRI, INODE *, IEDGE *, ITRI *, UINT *);
    void InitNodeAdjStructure (NADJ &, ITRI, UINT, UINT, BOOL = FALSE);
    void NudgeNode (NADJ &);
    inline int xMAP (REAL x) {return (int) (cx*(    (x-xMin)/xDelta))+cxBorder;}
    inline int yMAP (REAL y) {return (int) (cy*(1.0-(y-yMin)/yDelta))+cyBorder;}
    void DrawElement (PTRI);
    void DrawElement (ITRI);
    void DrawTriangles (NADJ &);
    void FillPoly (UINT, INODE *, int, int);
#if defined(DEBUG)
    void Test1 ();
    void Test2 (ITRI, ITRI, IEDGE);
    void DebugDumpMesh ();
#endif

  public:
    Refine (char *);			// constructor
   ~Refine ();				// destructor
};


#endif
