// ----------------------------------------------------------------------------
//
// FILE:     sgplot.cpp
//
// PROJECT:  sgplot.exe
//
// AUTHOR:   Kevin M. Kramer
//           Brian Chu, sing yun
//           Neil K. H
//           Eric R. Keiter
//           J. Joshua Feng
//
// REVISION HISTORY
// DATE      INITIALS   DESCRIPTION
// --------  --------   -------------------------------------------------------
// ??/??/??    KMK      initial implementation as triplot
// ??/??/??    BSC      change to sgplot (grid and gray level)
// ??/??/??    NKH      change to sgplot (grid and gray level)
// ??/??/??    ERK      contour plot
// 02/04/97    JJF      separate it into modules
// 04/18/97    JJF      simple mesh handling (indep of Matlab from now on)
//
// DESCRIPTION:
//
// INPUT:    *.res      generated by simgen from SIMGEN frame work
//           *.msh      generated by mesh from SIMGEN frame work
//
// OUTPUT:   3D graph with frame
//           a postscript file, if -ps option
//
// ----------------------------------------------------------------------------
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>

#if   defined(BOR)
  #include <io.h>
  #include <conio.h>
  #include "wgraph.h"

#elif defined(DOS)
  #include <io.h>
  #include <conio.h>
  #include <graphics.h>
  #include "bgraph.h"

#elif defined(UNIX)
  #include <unistd.h>
  #include "xgraph.h"
  #define  getch getchar
#endif

#include "termnlgy.h"
#include "plot.h"                    // plot header file 
#include "grafix3d.h"                // TriPlot3D class header file
#include "cmdparse.h"                // command parsing class header file

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////  Globl Variables  ///////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// definition of command line argument structure
typedef struct tagCMDARGS           // cmdargs
{ char meshfile[LEN_FILENAME];
  char resultfile[LEN_FILENAME];

  char *pszArray;                   // array to plot 
  char szArrayName[8];              // array name  
  int  iArray;                      // index of array to plot
  int  iSet;                        // data set 

  int  nResolution;                 // x window resolution
  BOOL fLogPlot;                    // log plot
  BOOL fGeneratePSFile;             // generate a postscript file
  BOOL fNoMesh;                     // do not plot the mesh
  BOOL fConSurf;
  BOOL fNoSurf;
  int  iStyle;                      // STYLE_BW
                                    // STYLE_WIREFRAME
                                    // STYLE_COLOR (default)
                                    // STYLE_GRAYLEVEL
  int  iztick;                      // # of z axis ticks
  UINT Persp;                       // Perspective distance

  int  iCon;                        // contour plot
  UINT NumCon;                      // # of contours
 
  REAL az;                          // azimuthal angle
  REAL el;                          // elevation angle
} CMDARGS;  

CMDARGS cmdargs;   

static char *apszHelp[] =
{ "syntax:  sgplot filename [array] [options]                     * = default",
  "  -h         display this help screen",
  "  -i###      use the ###'th data set (default 1)",
  "  -az###     azimuthal angle (default 240)",
  "  -el###     elevation angle (default 60)",
#if defined(UNIX)
  "  -ps        generate a postscript file",
#endif
  // "  -nm        do not plot the mesh",
  "  -s#        0: B/W;  1: wire frame;  2: * color;  3: gray level",
  "  -conA###   plot ### contour lines in 3d",
  "  -conB###   plot ### contour lines in 3d, on surface",
  "  -conC###   plot ### contour lines in 2d",
  "  -lin     * use a linear scale",
  "  -log       use a logarithmic scale", 
  "  -zt        number of ticks on z-axis",
  "  -persp#    perspective distance #+3 units",
#if defined(UNIX)
  "  -r1        graphics screen resolution 480 x 640",
  "  -r2      * graphics screen resolution 800 x 600",
  "  -r3        graphics screen resolution 1024 x 768",
#endif
  NULL
};

UINT cNodes, cTriangles;
AINODE *aainode;
PNT3D *apnt3dInput;

///////////////////////////////////////////////////////////////////////////////
// Function:     ReadMeshFile
// Scope:        Public
// Description:  This function reads the mesh.
//
// *.msh         char szLogo[64]
//               char szSign[16]
//               UINT cConstant <--- # of constants
//               UINT cLabel <------ # of labels
//               UINT cArray <------ # of arrays
//               UINT cRegLabel <--- # of regions
//               UINT cBndryNode <-- # of boundary nodes
//               UINT cNode <------- # of nodes
//               UINT cEdge <------- # of edges
//               UINT cTriangle <--- # of trangles
//               UINT fCylGeom <---- # flag for Cylindrical geometry
//               --------------------------------------------------------
//                         UINT            +        cConstant * XLATCONST
//               --------------------------------------------------------
//                         UINT
//                      1  XLATLABEL + XLATLABEL.cNode * UINT 
//                           ...
//                           ...
//                  cLabel XLATLABEL + XLATLABEL.cNode * UINT
//               --------------------------------------------------------
//     wanted              UINT
//                      1  char szArrayName         <-- x array
//                         cNode * REAL 
//                      2  char szArrayName         <-- y array
//                         cNode * REAL
//                           ...
//                           ...
//                 cArray  char szArrayName
//                         cNode * REAL
//               --------------------------------------------------------
//                         UINT      + cNode * NODE
//               --------------------------------------------------------
//     wanted              UINT      + cEdge * EDGE
//               --------------------------------------------------------
//     wanted              UINT      + cTriangel * TRI
//               --------------------------------------------------------
//                         UINT       // cAdj = cNode + cBndryNode*cReglabel
//                      1  INODE      // node index (which node)
//                         NODEINFO
//                           REAL Area
//                           UINT cNeighbor
//                           UINT cTriangle
//                           (*EDGEINFO) 
//                                     1. EDGEINFO
//                                    ...
//                             cNeighbor. EDGEINFO
//                   ....
//                  cNode  INODE      // node index (which node)
//                         NODEINFO
//                           REAL Area
//                           UINT cNeighbor
//                           UINT cTriangle
//                           (*EDGEINFO) 
//                                     1. EDGEINFO
//                                    ...
//                             cNeighbor. EDGEINFO
//                   ....
//                cNode+1  INODE      // 1 BndryNode(node index) (Reg 1)
//                         NODEINFO
//                           REAL Area
//                           UINT cNeighbor
//                           UINT cTriangle
//                           (*EDGEINFO) 
//                                     1. EDGEINFO
//                                    ...
//                             cNeighbor. EDGEINFO
//                   ....
//        cNode+cRegLabel  INODE      // 1 BndryNode(node index) (Reg cRegLabel)
//                         NODEINFO
//                           REAL Area
//                           UINT cNeighbor
//                           UINT cTriangle
//                           (*EDGEINFO) 
//                                     1. EDGEINFO
//                                    ...
//                             cNeighbor. EDGEINFO
//      cNode+cRegLabel+1  INODE      // 2 BndryNode(node index) (Reg 1)
//                         NODEINFO
//                           REAL Area
//                           UINT cNeighbor
//                           UINT cTriangle
//                           (*EDGEINFO) 
//                                     1. EDGEINFO
//                                    ...
//                             cNeighbor. EDGEINFO
//                   ....
//                   ....
//                   cAdj  INODE 
//                         NODEINFO
//                           REAL Area
//                           UINT cNeighbor
//                           UINT cTriangle
//                           (*EDGEINFO) 
//                                     1. EDGEINFO
//                                    ...
//                             cNeighbor. EDGEINFO
void ReadMeshFile (char *pszName)
{
  char sz[LEN_FILENAME];
  strcpy(sz, pszName);
  // ----------------------- open the mesh file ------------------------
  int nFile = open(sz, O_RDONLY | O_BINARY);
  if (nFile == -1)
  { strcat(sz, ".msh");
    nFile = open(sz, O_RDONLY | O_BINARY);
    if (nFile == -1)
    { fprintf(stdout, "ERROR:  cannot open file \"%s\"\n", sz);
      exit(2);
    }
  }
  strcpy(cmdargs.meshfile, sz);

  // read the header
  MESHHEAD meshhead;
  read(nFile, &meshhead, sizeof(MESHHEAD));

  // allocate the data structures
  cNodes = meshhead.cNode;
  cTriangles = meshhead.cTriangle;

  aainode     = new AINODE[cTriangles];
  apnt3dInput = new PNT3D[cNodes];

  // ------------------- advance past the constants --------------------
  long lOffset = meshhead.cConstant * sizeof(XLATCONST) + sizeof(UINT);
  lseek(nFile, (off_t) lOffset, SEEK_CUR);

  UINT i;
  // ------------------ advance past the label lists -------------------
  UINT cLabel;
  read(nFile, &cLabel, sizeof(UINT));
  for(i = 0; i < meshhead.cLabel; ++i)
  { XLATLABEL xlatlabel;
    read(nFile, &xlatlabel, sizeof(XLATLABEL));
    lseek(nFile, xlatlabel.cNode * sizeof(UINT), SEEK_CUR);
  }

  // ------------------ read the x and y mesh arrays -------------------
  UINT cArray;
  read(nFile, &cArray, sizeof(UINT));
  char szArrayName[LEN_IDENT+1];
  REAL *ax = new REAL[cNodes];
  REAL *ay = new REAL[cNodes];
  read(nFile, szArrayName, LEN_IDENT+1);
  read(nFile, ax, cNodes * sizeof(REAL));
  read(nFile, szArrayName, LEN_IDENT+1);
  read(nFile, ay, cNodes * sizeof(REAL));
  for(i = 2; i < meshhead.cArray; ++i)
  { read(nFile, szArrayName, LEN_IDENT+1);
    lseek(nFile, cNodes * sizeof(REAL), SEEK_CUR);
  }
  for(i = 0; i < cNodes; ++i)
  { apnt3dInput[i].x = ax[i];
    apnt3dInput[i].y = ay[i];
    apnt3dInput[i].z = 0.0;
  }
  delete [] ax;
  delete [] ay;

  // ---------- advance the file pointer past the node list ------------
  lOffset = meshhead.cNode * sizeof(NODE) + sizeof(UINT);
  lseek(nFile, (off_t) lOffset, SEEK_CUR);

  // ------------------- read the edge array ---------------------------
  UINT cEdge;
  read(nFile, &cEdge, sizeof(UINT));
  EDGE *aedge = new EDGE[meshhead.cEdge];
  read(nFile, aedge, meshhead.cEdge * sizeof(EDGE));

  // ----------------- read the triangle array -------------------------
  UINT cTriangle;
  read(nFile, &cTriangle, sizeof(UINT));
  for(i = 0; i < cTriangles; ++i)
  { TRI tri;
    read(nFile, &tri, sizeof(TRI));
    EDGE *pedgeAB = aedge + tri.iedgeAB;
    EDGE *pedgeBC = aedge + tri.iedgeBC;
    INODE inode1 = pedgeAB->inodeA;
    INODE inode2 = pedgeAB->inodeB;
    INODE inode3 = pedgeBC->inodeA;
    INODE inode4 = pedgeBC->inodeB;
    INODE inodeA, inodeB, inodeC, inodeD;

    if (tri.iedgeAD == -1u)
    { if      (inode1 == inode3) { inodeA = inode2; inodeB = inode1; inodeC = inode4; }
      else if (inode1 == inode4) { inodeA = inode2; inodeB = inode1; inodeC = inode3; }
      else if (inode2 == inode3) { inodeA = inode1; inodeB = inode2; inodeC = inode4; }
      else                       { inodeA = inode1; inodeB = inode2; inodeC = inode3; }
      inodeD = -1u;
    }
    else
    { EDGE *pedgeCD = aedge + tri.iedgeCD;
      INODE inode5 = pedgeCD->inodeA;
      INODE inode6 = pedgeCD->inodeB;
      if ((inode1 == inode3) || (inode1 == inode4))
           { inodeA = inode2; inodeB = inode1; }
      else { inodeA = inode1; inodeB = inode2; }
      if ((inode5 == inode3) || (inode5 == inode4))
           { inodeC = inode5; inodeD = inode6; }
      else { inodeC = inode6; inodeD = inode5; }
    }

    aainode[i][0] = inodeA;
    aainode[i][1] = inodeB;
    aainode[i][2] = inodeC;
    aainode[i][3] = inodeD;
  }

  // close the file
  delete [] aedge;
  close(nFile);
}


///////////////////////////////////////////////////////////////////////////////
// Function:     ReadResult
// Scope:        Public
// Description:  This function reads the result file.
void ReadResult ( char *pszName )
{
  // open the result file
  char sz[LEN_FILENAME];
  strcpy(sz, pszName);
  strcat(sz, ".res");
  int nFile = open(sz, O_RDONLY | O_BINARY);
  if (nFile == -1)
  { fprintf(stdout, "ERROR:  cannot open file \"%s\"\n", sz);
    exit(2);
  }
  strcpy(cmdargs.resultfile, sz);

  // read the header
  RESHEAD reshead;
  read(nFile, &reshead, sizeof(RESHEAD));

  // advance past the constants and variables
  lseek(nFile, reshead.cConstant * sizeof(XLATCONST), SEEK_CUR);
  lseek(nFile, reshead.cVariable * sizeof(XLATVAR)  , SEEK_CUR);
    
  // -------------- prompt user for data set to plot -------------------
  fprintf(stderr, 
          "-- %d out of %d data set(s) --\n", cmdargs.iSet, reshead.cSet);
  if ((cmdargs.iSet < 1) || (cmdargs.iSet > reshead.cSet))
  { int iSet = 1;
    if (reshead.cSet > 1)
    { do
      { fprintf(stderr, "data set to plot (1 - %u):  ", reshead.cSet);
        fscanf(stdin, "%d", &iSet);
      } while ((iSet < 1) || (iSet > reshead.cSet));
    }
    cmdargs.iSet = iSet;
  }
  --cmdargs.iSet;

  UINT i, j;
  XLATARRAY *axlatarray;
  // ----------------- Determine the mesh source and read it -------------
  if ((reshead.szMeshFile != NULL) && (reshead.szMeshFile[0] != '\0'))
  { // the mesh file exists
    ReadMeshFile(reshead.szMeshFile);    // allocate data structure
                                         // cNodes is determined here
    
    // --------- read the 1D arrays that have cNodes elements ------------
    UINT c1DArray = 0;
    axlatarray = new XLATARRAY[reshead.cArray];
    for(i = 0; i < reshead.cArray; ++i)
    { read(nFile, axlatarray + c1DArray, sizeof(XLATARRAY));
      if ((axlatarray[c1DArray].cDim == 1) &&
          (axlatarray[c1DArray].acElements[0] == cNodes)) ++c1DArray;
    }

    // -------------- prompt user for array to plot ----------------------
    cmdargs.iArray = -1;
    if (cmdargs.pszArray)
    { for(i = 0; i < c1DArray; ++i)
        if (!strcmp(axlatarray[i].szName, cmdargs.pszArray))
        { cmdargs.iArray = i;
          break;
        }
    } 

    if (cmdargs.iArray == -1)
    { for(i = 0; i < c1DArray; ++i)
        fprintf(stderr, 
                "enter [%u] to plot the %s array\n", i+1, axlatarray[i].szName);
      int iArray = 0;
      do
      { fprintf(stderr, "array to plot:  ");
        fscanf(stdin, "%d", &iArray);
      } while ((iArray < 1) || (iArray > c1DArray));
      cmdargs.iArray = iArray - 1;
    }
    strcpy(cmdargs.szArrayName, axlatarray[cmdargs.iArray].szName); 
  }
  else
  { // no meshfile *.sk is assigned
    UINT c2DArray = 0;
    axlatarray = new XLATARRAY[reshead.cArray];
    for(i = 0; i < reshead.cArray; ++i)
    { read(nFile, axlatarray + c2DArray, sizeof(XLATARRAY));
      if (axlatarray[c2DArray].cDim == 2) ++c2DArray; 
    }

    // -------------- prompt user for array to plot ----------------------
    cmdargs.iArray = -1;
    if (cmdargs.pszArray)
    { for(i = 0; i < c2DArray; ++i)
        if (!strcmp(axlatarray[i].szName, cmdargs.pszArray))
        { cmdargs.iArray = i;
          break;
        }
    } 

    if (cmdargs.iArray == -1)
    { for(i = 0; i < c2DArray; ++i)
        fprintf(stderr, 
                "enter [%u] to plot the %s array\n", i+1, axlatarray[i].szName);
      int iArray = 0;
      do
      { fprintf(stderr, "array to plot:  ");
        fscanf(stdin, "%d", &iArray);
      } while ((iArray < 1) || (iArray > c2DArray));
      cmdargs.iArray = iArray - 1;
    }
    strcpy(cmdargs.szArrayName, axlatarray[cmdargs.iArray].szName); 

    UINT cRow = axlatarray[cmdargs.iArray].acElements[0];
    UINT cCol = axlatarray[cmdargs.iArray].acElements[1];

    cNodes = cRow * cCol;
    cTriangles = (cRow - 1) * (cCol - 1);
 
    aainode     = new AINODE[cTriangles];
    apnt3dInput = new PNT3D[cNodes];

    for(i = 0; i < cRow; i++)
      for(j = 0; j < cCol; j++)
      { apnt3dInput[i*cCol+j].x = (REAL) i;
        apnt3dInput[i*cCol+j].y = (REAL) j;
        apnt3dInput[i*cCol+j].z = 0.0;
      }

    for(i = 0; i < (cRow-1); i++)
      for(j = 0; j < (cCol-1); j++)
      { aainode[i*(cCol-1)+j][0] = i*cCol+j;
        aainode[i*(cCol-1)+j][1] = i*cCol+j+1;
        aainode[i*(cCol-1)+j][2] = (i+1)*cCol+j+1;
        aainode[i*(cCol-1)+j][3] = (i+1)*cCol+j;
      }
  }

  // ------------------------ get data to plot -------------------------
  REAL *az = new REAL[cNodes];
  long lOffset = ((long) cmdargs.iSet * reshead.cElement + axlatarray[cmdargs.iArray].uOffset) 
               * (long) sizeof(REAL);
  lseek(nFile, (off_t) lOffset, SEEK_CUR);
  read(nFile, az, cNodes * sizeof(REAL));

  // initialize to z field of the input data
  if (cmdargs.fLogPlot)
  { REAL MinNonzero = fabs(az[0]);
    for(i = 1; i < cNodes; ++i)
    { REAL Magnitude = fabs(az[i]);
      if (Magnitude && (Magnitude < MinNonzero)) MinNonzero = Magnitude;
    }
    for(i = 0; i < cNodes; ++i)
      apnt3dInput[i].z = (az[i]) ? log10(fabs(az[i])) : log10(MinNonzero);
  }
  else
    for(i = 0; i < cNodes; ++i)
      apnt3dInput[i].z = az[i];

  // clean-up
  delete [] axlatarray;
  delete [] az;
  close(nFile);
}

///////////////////////////////////////////////////////////////////////////////
// Function:    TitleEtc
// Scope:       Public
// Description: This function prints the title to the screen
void TitleEtc ( UINT x1, UINT y1, int nResFactor )
{
#if DEBUG
        printf("in TriPlot3D::TitleEtc\n");
#endif
  char txt[64];
  UINT ix;
  UINT iy = (UINT) (CHARPIXELHIGH * nResFactor * 1.2);

  if      (cmdargs.nResolution == RES_640x480)  ix = 56; 
  else if (cmdargs.nResolution == RES_1024x768) ix = 64; 
  else if (cmdargs.nResolution == RES_800x600)  ix = 96; 

  x1 += 4 * nResFactor * CHARPIXELWIDE;
  y1 *= nResFactor;
  sprintf(txt, "Simgen Plot:%s, Mesh:%s", cmdargs.resultfile, cmdargs.meshfile);
  GRText(ix + x1, y1 += iy, txt);
  sprintf(txt, "AZ = %3.3e, EL = %3.3e", cmdargs.az, cmdargs.el);
  GRText(ix + x1, y1 += iy, txt);
  sprintf(txt, "[Set:%d], [Array:%s]", cmdargs.iSet+1, cmdargs.szArrayName);
  GRText(ix + x1, y1 += iy, txt);
}

static TriPlot3D* triplot;

///////////////////////////////////////////////////////////////////////////////
void Redraw ( )
{
  triplot->PlotGrid();
  if (!cmdargs.fNoMesh) triplot->PlotMesh();
  if (cmdargs.fConSurf)
    triplot->PlotData(TRUE, TRUE);                         // conB
  else
  { if (cmdargs.fNoMesh)  triplot->PlotData(FALSE, TRUE);  // conA & ConC(2D)
    if (!cmdargs.fNoSurf) triplot->PlotData(TRUE, FALSE);  // no Contour
  }
}

//////////////////////////////////////////////////////////////////////////////
/////////////////////////////// MAIN PROGRAM /////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
int main ( UINT cArg, char *apszArg[] )
{
  // allocate an instance of the command line parser
  CmdParse *pcmd = new CmdParse(cArg, apszArg);

  fprintf(stdout,
   "SimGen Surface Plotter 1.1 (c) 1997 K.Kramer, B.Chu, E.Keiter, J.Feng\n");

  // ---------------- parse the command line arguments -------------------
  if ((pcmd->GetNumOfPosArgs() < 1) || 
      (pcmd->GetNumOfPosArgs() > 2) || 
       pcmd->fArgExist("h"))
  { for(char **ppsz = apszHelp; *ppsz; ++ppsz)
      fprintf(stdout, "%s\n", *ppsz);
    return 1;
  }

  cmdargs.pszArray        = pcmd->pszGetPosArg(1, NULL);
  cmdargs.iArray          = 0;
  cmdargs.iSet            = pcmd->nGetSwitchArg("i", 1, 0, 32767);
  cmdargs.fLogPlot        = pcmd->fArgExist("log") && !pcmd->fArgExist("lin");
  cmdargs.fGeneratePSFile = pcmd->fArgExist("ps");

  cmdargs.iStyle          = pcmd->nGetSwitchArg("s", 2, 0, 3);
  cmdargs.iztick          = pcmd->nGetSwitchArg("zt", -1, -1, 50);
  cmdargs.Persp           = pcmd->nGetSwitchArg("persp", 0, 0, 32768);

  cmdargs.az = pcmd->rGetSwitchArg("az", 240.0, -360.0, 360.0);
  cmdargs.el = pcmd->rGetSwitchArg("el",  60.0, -180.0, 180.0); 

  int iSwitches = F_SCALE; // 0

  switch (cmdargs.iStyle)
  { case  1: iSwitches |= F_COLOR | F_WIREFRAME; break;
    case  2: iSwitches |= F_COLOR;               break;
    case  3: iSwitches |= F_GRAYLEVEL;           break;
    default: break; 
  }

  if      (pcmd->fArgExist("conA") == 1)            // conA: 3d on mesh plane
  { cmdargs.fNoMesh = TRUE;
    cmdargs.NumCon = pcmd->nGetSwitchArg("conA", 50, 2, 200);
  }
  else if (pcmd->fArgExist("conB") == 1)            // conB: 3d on surface
  { cmdargs.fConSurf = TRUE;
    cmdargs.NumCon = pcmd->nGetSwitchArg("conB", 50, 2, 200);
  }
  else if (pcmd->fArgExist("conC") == 1)            // conC: 2d
  { cmdargs.fNoMesh = TRUE;
    cmdargs.fNoSurf = TRUE;
    cmdargs.NumCon = pcmd->nGetSwitchArg("conC", 50, 2, 200);
    cmdargs.az = 270;
    cmdargs.el = 90;
    cmdargs.Persp = FALSE;
  }
  else                                              // no contour
  { cmdargs.NumCon = 0;
  }
 
  int MAXx, MAXy;

#if defined(BOR) || defined(UNIX)
  if      (pcmd->fArgExist("r1"))
  { cmdargs.nResolution = RES_640x480;
    MAXx = 640; MAXy = 480;
  }
  else if (pcmd->fArgExist("r3"))
  { cmdargs.nResolution = RES_1024x768;
    MAXx = 1024; MAXy = 768;
  }
  else
  { cmdargs.nResolution = RES_800x600;
    MAXx = 800; MAXy = 600;
  }
#endif

#if defined(UNIX)
  setcmdargs(cArg, apszArg);
#endif

  ReadResult(apszArg[1]);          // Read in the result file
                                   // but read the mesh file first

  // -------- triplot is a pointer point to the class TriPlot3D --------
  triplot = new TriPlot3D(cTriangles, cNodes,
                          aainode, apnt3dInput,
                          cmdargs.az, cmdargs.el, 
                          cmdargs.nResolution, iSwitches,
                          cmdargs.NumCon, cmdargs.iztick, cmdargs.Persp);

#if defined(BOR)
  GROpenGraphWin("Rectangular/Triangular Surface Plot", "Plot", 
                 -1, -1, MAXx, MAXy, Redraw);

#elif defined(DOS)
  Redraw();
  getch();

#elif defined(UNIX)
  GROpenGraphWin("Rectangular/Triangular Surface Plot", "Plot", 
                 0, 0, MAXx, MAXy, NULL);
  BOOL fDone = FALSE;
  short int keycode;
  int arg1, arg2;
  while (!fDone)
  { switch(getevent(&keycode, &arg1, &arg2))
    { case EXPOSE:
        Redraw();        
        if (!(pcmd->fArgExist("conC"))) TitleEtc(0, 5, 1); 

        flushdisplay();
        break;

      case CONFIGURENOTIFY:
        break;

      case BUTTONPRESS:
      case KEYPRESS:
        fDone = TRUE;
        break;
    }
  }
#endif
  if (cmdargs.fGeneratePSFile)
  { triplot->nResFactor = 20;
    triplot->cx = 468 * triplot->nResFactor;
    triplot->cy = 468 * triplot->nResFactor;
    triplot->cxLBorder = 72  * triplot->nResFactor;
    triplot->cyLBorder = 162 * triplot->nResFactor;
    triplot->Xform2DData();
    GRPrintOpen(apszArg[1], triplot->nResFactor);
    Redraw();        
    TitleEtc(0, 15, triplot->nResFactor); 
    GRPrintClose();
  }
 
  // clean up after PS file is done
  delete triplot;
  delete [] apnt3dInput;
  delete [] aainode;
  delete pcmd;

  return 0;
}
