#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#if defined(UNIX)
#include <unistd.h>
#include "xgraph.h"
#define	 O_BINARY 0x0000
#else
#include <io.h>
#include "wgraph.h"
#endif

#include <fcntl.h>
#include "triplot.h"
#include "cmdparse.h"


/* definition of command line argument structure */
#define RES_640x480	    1	    /* 640 x 480 resolution */
#define RES_800x600	    2	    /* 800 x 600 resolution */
#define RES_1024x768	    3	    /* 1024 x 768 resolution */


typedef struct tagCMDARGS	    // cmdargs
{ char *pszArray;		    // array to plot
  int  iArray;			    // index of array to plot
  int  iSet;			    // data set
  int  nResolution;		    // x window resolution
  BOOL fLogPlot;		    // log plot
  BOOL fGeneratePSFile;		    // generate a postscript file
  BOOL fColor;			    // color plot
  BOOL fWireFrame;		    // wire frame plot
  BOOL fNoMesh;											// do not plot the mesh
  REAL az;			    // azimuthal angle
  REAL el;			    // elevation angle
} CMDARGS;

CMDARGS cmdargs;


static char *apszHelp[] =
{ "syntax:  refine filename [array] [options]			  * = default",
  "  -h		display this help screen",
  "  -i###	use the ###'th data set (default 1)",
  "  -az###	azimuthal angle (default 30)",
  "  -el###	elevation angle (default 60)",
  "  -ps	generate a postscript file",
  "  -wf	wire frame",
  "  -bw	black and white (grayscale) plot",
  "  -nm	do not plot the mesh",
  "  -lin     * use a linear scale",
  "  -log	use a logarithmic scale",
  "  -r1	graphics screen resolution 640 x 480",
  "  -r2      * graphics screen resolution 800 x 600",
  "  -r3	graphics screen resolution 1024 x 768",
  NULL
};


static	TriPlot3D *ptriplot3d;

void Redraw ()
{
  if (!cmdargs.fNoMesh)
	ptriplot3d->PlotMesh();
  ptriplot3d->PlotData();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::TriPlot3D
// Scope:	 Public, Constructor
TriPlot3D::TriPlot3D (char *pszFileName)
{
  ReadMeshFile(pszFileName);
  ReadResultFile(pszFileName);
  GetViewAspect();

  GRInitGraphics();
  InitWinDim();

	Xform3DData();
	Xform2DData();
	DeterminePlotOrder();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::~TriPlot3D
// Scope:	 Public, Destructor
TriPlot3D::~TriPlot3D ()
{
  delete aainode;
  delete apnt3dInput;
  delete apnt3dXform;
  delete apnt3dMesh;
  delete apnt2d;
  delete apnt2dMesh;
  delete aiPlotOrder;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::ReadMeshFile
// Scope:	 Public
// Description:	 This function reads the mesh.
void TriPlot3D::ReadMeshFile (char *pszName)
{
  // open the result file
  char sz[LEN_FILENAME];
  strcpy(sz, pszName);
  strcat(sz, ".res");
  int nFile = open(sz, O_RDONLY | O_BINARY);
  if (nFile == -1)
  { fprintf(stdout, "ERROR:  cannot open file \"%s\"\n", sz);
    exit(2);
  }

  // read the header and close the file
  RESHEAD reshead;
  read(nFile, &reshead, sizeof(RESHEAD));
  close(nFile);

  // open the mesh file
  nFile = open(reshead.szMeshFile, O_RDONLY | O_BINARY);
  if (nFile == -1)
  { strcpy(sz, pszName);
    strcat(sz, ".msh");
    nFile = open(sz, O_RDONLY | O_BINARY);
    if (nFile == -1)
    { fprintf(stdout, "ERROR:  cannot open file \"%s\"\n", sz);
      exit(2);
    }
  }

  // read the header
  MESHHEAD meshhead;
  read(nFile, &meshhead, sizeof(MESHHEAD));

  // allocate the data structures
  cNodes = meshhead.cNode;
  cTriangles = meshhead.cTriangle;
  aainode = new AINODE[cTriangles];
  apnt3dInput = new PNT3D[cNodes];
  apnt3dXform = new PNT3D[cNodes];
  apnt3dMesh = new PNT3D[cNodes];
  apnt2d = new PNT2D[cNodes];
  apnt2dMesh = new PNT2D[cNodes];
  aiPlotOrder = new UINT[cTriangles];

  // advance past the constants
  long lOffset = meshhead.cConstant * sizeof(XLATCONST) + sizeof(UINT);
  lseek(nFile, (off_t) lOffset, SEEK_CUR);

  // advance past the label lists
  UINT i;
  UINT cLabel;
  read(nFile, &cLabel, sizeof(UINT));
  for(i = 0; i < meshhead.cLabel; ++i)
  { XLATLABEL xlatlabel;
    read(nFile, &xlatlabel, sizeof(XLATLABEL));
    lseek(nFile, xlatlabel.cNode * sizeof(UINT), SEEK_CUR);
  }

  // read the x and y mesh arrays
  UINT cArray;
  read(nFile, &cArray, sizeof(UINT));
  char szArrayName[LEN_IDENT+1];
  REAL *ax = new REAL[cNodes];
  REAL *ay = new REAL[cNodes];
  read(nFile, szArrayName, LEN_IDENT+1);
  read(nFile, ax, cNodes * sizeof(REAL));
  read(nFile, szArrayName, LEN_IDENT+1);
  read(nFile, ay, cNodes * sizeof(REAL));
  for(i = 2; i < meshhead.cArray; ++i)
  { read(nFile, szArrayName, LEN_IDENT+1);
    lseek(nFile, cNodes * sizeof(REAL), SEEK_CUR);
  }
  for(i = 0; i < cNodes; ++i)
  { apnt3dInput[i].x = ax[i];
    apnt3dInput[i].y = ay[i];
    apnt3dInput[i].z = 0.0;
  }
  delete ax;
  delete ay;

  // advance the file pointer past the node list
  lOffset = meshhead.cNode * sizeof(NODE) + sizeof(UINT);
  lseek(nFile, (off_t) lOffset, SEEK_CUR);

  // read the edge array
  UINT cEdge;
  read(nFile, &cEdge, sizeof(UINT));
  EDGE *aedge = new EDGE[meshhead.cEdge];
  read(nFile, aedge, meshhead.cEdge * sizeof(EDGE));

  // read the triangle array
  UINT cTriangle;
  read(nFile, &cTriangle, sizeof(UINT));
  for(i = 0; i < cTriangles; ++i)
  { TRI tri;
    read(nFile, &tri, sizeof(TRI));
    EDGE *pedgeAB = aedge + tri.iedgeAB;
    EDGE *pedgeBC = aedge + tri.iedgeBC;
    INODE inode1 = pedgeAB->inodeA;
    INODE inode2 = pedgeAB->inodeB;
    INODE inode3 = pedgeBC->inodeA;
    INODE inode4 = pedgeBC->inodeB;
    INODE inodeA, inodeB, inodeC, inodeD;

    if (tri.iedgeAD == -1u)
    { if      (inode1 == inode3) { inodeA = inode2; inodeB = inode1; inodeC = inode4; }
      else if (inode1 == inode4) { inodeA = inode2; inodeB = inode1; inodeC = inode3; }
      else if (inode2 == inode3) { inodeA = inode1; inodeB = inode2; inodeC = inode4; }
      else			 { inodeA = inode1; inodeB = inode2; inodeC = inode3; }
      inodeD = -1u;
    }
    else
    { EDGE *pedgeCD = aedge + tri.iedgeCD;
      INODE inode5 = pedgeCD->inodeA;
      INODE inode6 = pedgeCD->inodeB;
      if ((inode1 == inode3) || (inode1 == inode4))
	   { inodeA = inode2; inodeB = inode1; }
      else { inodeA = inode1; inodeB = inode2; }
      if ((inode5 == inode3) || (inode5 == inode4))
	   { inodeC = inode5; inodeD = inode6; }
      else { inodeC = inode6; inodeD = inode5; }
    }

    aainode[i][0] = inodeA;
    aainode[i][1] = inodeB;
    aainode[i][2] = inodeC;
    aainode[i][3] = inodeD;
  }

  // close the file
  delete aedge;
  close(nFile);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::ReadResultFile
// Scope:	 Public
// Description:	 This function reads the result file.
void TriPlot3D::ReadResultFile (char *pszName)
{
  // open the result file
  char sz[LEN_FILENAME];
  strcpy(sz, pszName);
  strcat(sz, ".res");
  int nFile = open(sz, O_RDONLY | O_BINARY);
  if (nFile == -1)
  { fprintf(stdout, "ERROR:  cannot open file \"%s\"\n", sz);
    exit(2);
  }

  // read the header
  RESHEAD reshead;
  read(nFile, &reshead, sizeof(RESHEAD));

  // advance past the constants and variables
  lseek(nFile, reshead.cConstant * sizeof(XLATCONST), SEEK_CUR);
  lseek(nFile, reshead.cVariable * sizeof(XLATVAR)  , SEEK_CUR);

  // read the 1D arrays that have cNodes elements
  UINT i;
  UINT c1DArray = 0;
  XLATARRAY *axlatarray = new XLATARRAY[reshead.cArray];
  for(i = 0; i < reshead.cArray; ++i)
  { read(nFile, axlatarray + c1DArray, sizeof(XLATARRAY));
    if ((axlatarray[c1DArray].cDim == 1) &&
	(axlatarray[c1DArray].acElements[0] == cNodes)) ++c1DArray;
  }

  // prompt user for array to plot
  cmdargs.iArray = -1;
  if (cmdargs.pszArray)
  { for(i = 0; i < c1DArray; ++i)
      if (!strcmp(axlatarray[i].szName, cmdargs.pszArray))
      { cmdargs.iArray = i;
	break;
      }
  }

  if (cmdargs.iArray == -1)
  { for(i = 0; i < c1DArray; ++i)
      fprintf(stderr, "enter [%u] to plot the %s array\n", i + 1, axlatarray[i].szName);
    int iArray = 0;
    do
    { fprintf(stderr, "array to plot:  ");
      fscanf(stdin, "%d", &iArray);
    } while ((iArray < 1) || (iArray > c1DArray));
    cmdargs.iArray = iArray - 1;
  }

  // prompt user for data set to plot
  if ((cmdargs.iSet < 1) || (cmdargs.iSet > reshead.cSet))
  { int iSet = 0;
    if (reshead.cSet > 1)
    { do
      { fprintf(stderr, "data set to plot (1 - %u):  ", reshead.cSet);
	fscanf(stdin, "%d", &iSet);
      } while ((iSet < 1) || (iSet > reshead.cSet));
    }
    cmdargs.iSet = iSet;
  }
  --cmdargs.iSet;

  // get data to plot
  REAL *az = new REAL[cNodes];
  long lOffset = ((long) cmdargs.iSet * (long) reshead.cElement + (long) axlatarray[cmdargs.iArray].uOffset) * (long) sizeof(REAL);
  lseek(nFile, (off_t) lOffset, SEEK_CUR);
  read(nFile, az, cNodes * sizeof(REAL));

  // initialize to z field of the input data
  if (cmdargs.fLogPlot)
  { REAL MinNonzero = 1.0e300;
    for(i = 0; i < cNodes; ++i)
    { REAL Magnitude = fabs(az[i]);
      if (Magnitude && (Magnitude < MinNonzero)) MinNonzero = Magnitude;
    }
    for(i = 0; i < cNodes; ++i)
      apnt3dInput[i].z = (az[i]) ? log10(fabs(az[i])) : log10(MinNonzero);
  }
  else
    for(i = 0; i < cNodes; ++i)
      apnt3dInput[i].z = az[i];

  // clean-up
  delete axlatarray;
  delete az;
  close(nFile);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::GetViewAspect
// Scope:	 Public
// Description:	 This function prompts the user for the azimuthal and the
//		 elevation angle and computes the elements of the
//		 transformation array.
void TriPlot3D::GetViewAspect ()
{
#if 0
  fprintf(stderr, "azimuthal angle:  ");
  fscanf(stdin, "%lg", &AZ);
  fprintf(stderr, "elevation angle:  ");
  fscanf(stdin, "%lg", &EL);
#endif
  AZ = cmdargs.az;
  EL = cmdargs.el;

  REAL AZrad = (AZ+90) * M_PI / 180.0;
  REAL ELrad = EL * M_PI / 180.0;
  REAL sinAZ = sin(AZrad);
  REAL cosAZ = cos(AZrad);
  REAL sinEL = sin(ELrad);
  REAL cosEL = cos(ELrad);

  Cxx =	  cosAZ * cosEL;      Cxy = sinAZ * cosEL;	Cxz = - sinEL;
  Cyx = - sinAZ;	      Cyy = cosAZ;		Cyz =	0.0;
  Czx =	  cosAZ * sinEL;      Czy = sinAZ * sinEL;	Czz =	cosEL;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::Xform3DData
// Scope:	 Public
// Description:	 This function transforms the data.
void TriPlot3D::Xform3DData ()
{
  pnt3dMin.x = 1e300;	pnt3dMax.x = -1e300;
  pnt3dMin.y = 1e300;	pnt3dMax.y = -1e300;
  pnt3dMin.z = 1e300;	pnt3dMax.z = -1e300;

  UINT i;
  for(i = 0; i < cNodes; ++i)
  { if (apnt3dInput[i].x < pnt3dMin.x) pnt3dMin.x = apnt3dInput[i].x;
    if (apnt3dInput[i].y < pnt3dMin.y) pnt3dMin.y = apnt3dInput[i].y;
    if (apnt3dInput[i].z < pnt3dMin.z) pnt3dMin.z = apnt3dInput[i].z;
    if (apnt3dInput[i].x > pnt3dMax.x) pnt3dMax.x = apnt3dInput[i].x;
    if (apnt3dInput[i].y > pnt3dMax.y) pnt3dMax.y = apnt3dInput[i].y;
    if (apnt3dInput[i].z > pnt3dMax.z) pnt3dMax.z = apnt3dInput[i].z;
  }

  pnt3dScale.x = pnt3dMax.x - pnt3dMin.x;
  pnt3dScale.y = pnt3dMax.y - pnt3dMin.y;
  pnt3dScale.z = pnt3dMax.z - pnt3dMin.z;
  if (pnt3dScale.z == 0.0) pnt3dScale.z = 1.0;

  for(i = 0; i < cNodes; ++i)
  { Xform3D_3D(apnt3dInput[i], apnt3dXform[i]);
    XformMesh(apnt3dInput[i], apnt3dMesh[i]);
  }

  PNT3D pnt3dA, pnt3dB;
  pnt3dMinXform.x = 1e300;    pnt3dMaxXform.x = -1e300;
  pnt3dMinXform.y = 1e300;    pnt3dMaxXform.y = -1e300;
  pnt3dMinXform.z = 1e300;    pnt3dMaxXform.z = -1e300;
  for(i = 0; i < 8; ++i)
  { pnt3dA.x = (i & 0x01) ? pnt3dMax.x : pnt3dMin.x;
    pnt3dA.y = (i & 0x02) ? pnt3dMax.y : pnt3dMin.y;
    pnt3dA.z = (i & 0x04) ? pnt3dMax.z : pnt3dMin.z;
    Xform3D_3D(pnt3dA, pnt3dB);
    if (pnt3dB.x < pnt3dMinXform.x) pnt3dMinXform.x = pnt3dB.x;
    if (pnt3dB.y < pnt3dMinXform.y) pnt3dMinXform.y = pnt3dB.y;
    if (pnt3dB.z < pnt3dMinXform.z) pnt3dMinXform.z = pnt3dB.z;
    if (pnt3dB.x > pnt3dMaxXform.x) pnt3dMaxXform.x = pnt3dB.x;
    if (pnt3dB.y > pnt3dMaxXform.y) pnt3dMaxXform.y = pnt3dB.y;
    if (pnt3dB.z > pnt3dMaxXform.z) pnt3dMaxXform.z = pnt3dB.z;
  }
  pnt3dScaleXform.x = pnt3dMaxXform.x - pnt3dMinXform.x;
  pnt3dScaleXform.y = pnt3dMaxXform.y - pnt3dMinXform.y;
  pnt3dScaleXform.z = pnt3dMaxXform.z - pnt3dMinXform.z;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::Xform2DData
// Scope:	 Public
// Description:	 This function transforms the data.
void TriPlot3D::Xform2DData ()
{
  for(UINT i = 0; i < cNodes; ++i)
  { Xform3D_2D(apnt3dXform[i], apnt2d[i]);
    Xform3D_2D(apnt3dMesh[i], apnt2dMesh[i]);
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::Xform3D_3D
// Scope:	 Public
// Description:	 This function peforms the 3D rotation ... note the x
//		 component does not take into account the orignal z value.
void TriPlot3D::Xform3D_3D (PNT3D &pnt3dIn, PNT3D &pnt3dOut)
{
  REAL x = (pnt3dIn.x - pnt3dMin.x) / pnt3dScale.x;
  REAL y = (pnt3dIn.y - pnt3dMin.y) / pnt3dScale.y;
  REAL z = (pnt3dIn.z - pnt3dMin.z) / pnt3dScale.z;

  pnt3dOut.x =	 Cxx * x + Cxy * y + Cxz * 0;
  pnt3dOut.y = -(Cyx * x + Cyy * y + Cyz * z);
  pnt3dOut.z =	 Czx * x + Czy * y + Czz * z;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::Xform3D_2D
// Scope:	 Public
// Description:	 This function peforms the 3D to 2D transformation.
void TriPlot3D::Xform3D_2D (PNT3D &pnt3dIn, PNT2D &pnt2dOut)
{
  REAL x = (pnt3dIn.y - pnt3dMinXform.y) / pnt3dScaleXform.y;
  REAL y = (pnt3dIn.z - pnt3dMinXform.z) / pnt3dScaleXform.z;

  pnt2dOut.x = (int) ((REAL) cx * x) + cxBorder;
  pnt2dOut.y = (int) ((REAL) cy * (1.0 - y)) + cyBorder;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::XformMesh
// Scope:	 Public
// Description:	 This function peforms the 3D to 3D transformation for the
//		 mesh triangles.
void TriPlot3D::XformMesh (PNT3D &pnt3dIn, PNT3D &pnt3dOut)
{
  REAL x = (pnt3dIn.x - pnt3dMin.x) / pnt3dScale.x;
  REAL y = (pnt3dIn.y - pnt3dMin.y) / pnt3dScale.y;

  pnt3dOut.x =	 0.0;
  pnt3dOut.y = -(Cyx * x + Cyy * y);
  pnt3dOut.z =	 Czx * x + Czy * y;
}

///////////////////////////////////////////////////////////////////////////////
// Function:	 QuickSortCompare
// Scope:	 Private
// Description:	 This function compares two surface elements in order to
//		 determine which should be plotted first.  This function is
//		 called by qsort.
static AINODE *aainode;
static PNT3D  *apnt3d;
static int QuickSortCompare (const void *p1, const void *p2)
{
  UINT i1 = *((UINT *) p1);
  UINT i2 = *((UINT *) p2);
  UINT i1A = aainode[i1][0];
  UINT i1B = aainode[i1][1];
  UINT i1C = aainode[i1][2];
  UINT i1D = aainode[i1][3];
  UINT i2A = aainode[i2][0];
  UINT i2B = aainode[i2][1];
  UINT i2C = aainode[i2][2];
  UINT i2D = aainode[i2][3];
  REAL x1A = apnt3d[i1A].x;
  REAL x1B = apnt3d[i1B].x;
  REAL x1C = apnt3d[i1C].x;
  REAL x2A = apnt3d[i2A].x;
  REAL x2B = apnt3d[i2B].x;
  REAL x2C = apnt3d[i2C].x;
  REAL m1  = (x1A > x1B) ? x1A : x1B;
  REAL m2  = (x2A > x2B) ? x2A : x2B;
  m1  = (m1 > x1C) ? m1 : x1C;
  m2  = (m2 > x2C) ? m2 : x2C;
  if (i1D != -1u)
  { REAL x1D = apnt3d[i1D].x;
    m1 = (m1 > x1D) ? m1 : x1D;
  }
  if (i2D != -1u)
  { REAL x2D = apnt3d[i2D].x;
    m2 = (m2 > x2D) ? m2 : x2D;
  }
  return (m1 > m2) ? -1 : 1;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::DeterminePlotOrder
// Scope:	 Public
// Description:	 This function determines the plot order.
void TriPlot3D::DeterminePlotOrder ()
{
  for(UINT i = 0; i < cTriangles; ++i) aiPlotOrder[i] = i;
  ::aainode = aainode;
  ::apnt3d  = apnt3dXform;
  qsort(aiPlotOrder, cTriangles, sizeof(UINT), QuickSortCompare);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::OpenWindow
// Scope:	 Public
// Description:	 This function opens the graphics window
void TriPlot3D::InitWinDim ()
{
  if	  (cmdargs.nResolution == RES_640x480)
  { MAXx = 640;
    MAXy = 480;
  }
  else if (cmdargs.nResolution == RES_800x600)
  { MAXx = 800;
    MAXy = 600;
  }
  else if (cmdargs.nResolution == RES_1024x768)
  { MAXx = 1024;
    MAXy = 768;
  }
  cxBorder = 10;
  cyBorder = 10;

  cx = MAXx - 2*cxBorder;
  cy = MAXy - 2*cyBorder - 23;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::CloseWindow
// Scope:	 Public
// Description:	 This function closes the graphics window
//void TriPlot3D::CloseWindow ()
//{
//  closegraph();
//}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::PlotMesh
// Scope:	 Public
// Description:	 This function plots the mesh
void TriPlot3D::PlotMesh ()
{
  GRSetColor(GR_WHITE);
  for(UINT i = 0; i < cTriangles; ++i)
  { UINT iA = aainode[i][0];
    UINT iB = aainode[i][1];
    UINT iC = aainode[i][2];
    UINT iD = aainode[i][3];

    int an[10];
    an[0] = apnt2dMesh[iA].x;
    an[1] = apnt2dMesh[iA].y;
    an[2] = apnt2dMesh[iB].x;
    an[3] = apnt2dMesh[iB].y;
    an[4] = apnt2dMesh[iC].x;
    an[5] = apnt2dMesh[iC].y;

    if (iD == -1u)
    { an[6] = an[0];
      an[7] = an[1];
      GRDrawPoly(4, an);
    }
    else
    { an[6] = apnt2dMesh[iD].x;
      an[7] = apnt2dMesh[iD].y;
      an[8] = an[0];
      an[9] = an[1];
      GRDrawPoly(5, an);
    }

  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 TriPlot3D::PlotData
// Scope:	 Public
// Description:	 This function plots the data
void TriPlot3D::PlotData ()
{
  GRSetColor(GR_WHITE);
  GRSetFillColor(GR_BLACK);

  for(UINT i = 0; i < cTriangles; ++i)
  { UINT j = aiPlotOrder[i];
    UINT iA = aainode[j][0];
    UINT iB = aainode[j][1];
    UINT iC = aainode[j][2];
    UINT iD = aainode[j][3];

    int an[8];
    an[0] = apnt2d[iA].x;
    an[1] = apnt2d[iA].y;
    an[2] = apnt2d[iB].x;
    an[3] = apnt2d[iB].y;
    an[4] = apnt2d[iC].x;
    an[5] = apnt2d[iC].y;

    REAL z;

    if (iD != -1u)
    { an[6] = apnt2d[iD].x;
      an[7] = apnt2d[iD].y;
    }

    if (!cmdargs.fWireFrame)
    { if (iD == -1u)
      { z = apnt3dInput[iA].z + apnt3dInput[iB].z + apnt3dInput[iC].z;
	z /= 3.0;
      }
      else
      { z = apnt3dInput[iA].z + apnt3dInput[iB].z +
	    apnt3dInput[iC].z + apnt3dInput[iD].z;
	z /= 4.0;
      }

      z -= pnt3dMin.z;
      z /= pnt3dScale.z;
      int color;
      if (cmdargs.fColor)
      { if	(z >= 4.0/5.0) color = GR_RED;
	else if (z >= 3.0/5.0) color = GR_YELLOW;
	else if (z >= 2.0/5.0) color = GR_LIGHTGREEN;
	else if (z >= 1.0/5.0) color = GR_LIGHTBLUE;
	else		       color = GR_BLUE;
      }
      else
      { if	(z >= 2.0/3.0) color = GR_BLACK;
	else if (z >= 1.0/3.0) color = GR_LIGHTGRAY;
	else		       color = GR_DARKGRAY;
      }
      GRSetFillColor(color);
    }
    GRFillPoly((iD == -1u) ? 3 : 4, an);
  }
}

int main (UINT cArg, char *apszArg[])
{
  // allocate an instance of the command line parser
  CmdParse *pcmd = new CmdParse(cArg, apszArg);

  // parse the command line arguments
  fprintf(stdout, "SG Surface Plotter  Version 1.0  (c) 1997  K. Kramer\n");
  if ((pcmd->GetNumOfPosArgs() < 1) ||
      (pcmd->GetNumOfPosArgs() > 2) ||
       pcmd->fArgExist("h"))
  { for(char **ppsz = apszHelp; *ppsz; ++ppsz)
      fprintf(stdout, "%s\n", *ppsz);
    return 1;
  }

  cmdargs.pszArray = pcmd->pszGetPosArg(1, NULL);
  cmdargs.iArray = 0;
  cmdargs.iSet = pcmd->nGetSwitchArg("i", 1, 0, 32767);
  cmdargs.fLogPlot = pcmd->fArgExist("log") && !pcmd->fArgExist("lin");
  cmdargs.fGeneratePSFile = pcmd->fArgExist("ps");
  cmdargs.fColor = !pcmd->fArgExist("bw");
  cmdargs.fWireFrame = pcmd->fArgExist("wf");
  cmdargs.fNoMesh = pcmd->fArgExist("nm");
  cmdargs.az = pcmd->rGetSwitchArg("az", 30.0, -360.0, 360.0);
  cmdargs.el = pcmd->rGetSwitchArg("el", 60.0, -180.0, 180.0);
  if	  (pcmd->fArgExist("r1")) cmdargs.nResolution = RES_640x480;
  else if (pcmd->fArgExist("r3")) cmdargs.nResolution = RES_1024x768;
  else				  cmdargs.nResolution = RES_800x600;

  ptriplot3d = new TriPlot3D(apszArg[1]);
  UINT MAXx = ptriplot3d->GetMaxX();
  UINT MAXy = ptriplot3d->GetMaxY();
  ptriplot3d->nResFactor = 1;

#if defined(UNIX)
  GROpenGraphWin("Rectangular/Triangular Surface Plot", "Plot",
                 0, 0, MAXx, MAXy, NULL);
  BOOL fDone = FALSE;
  short int keycode;
  int arg1, arg2;
  while (!fDone)
  { switch(getevent(&keycode, &arg1, &arg2))
    { case EXPOSE:
        Redraw();
        flushdisplay();
        break;

      case CONFIGURENOTIFY:
        break;

      case BUTTONPRESS:
      case KEYPRESS:
        fDone = TRUE;
        break;
    }
  }
#else
  GROpenGraphWin("Rectangular/Triangular Surface Plot", "Plot", -1, -1, MAXx, MAXy, Redraw);
#endif

  if (cmdargs.fGeneratePSFile)
  { ptriplot3d->nResFactor = 20;
    ptriplot3d->cx = 468 * ptriplot3d->nResFactor;
    ptriplot3d->cy = 468 * ptriplot3d->nResFactor;
    ptriplot3d->cxBorder = 72  * ptriplot3d->nResFactor;
    ptriplot3d->cyBorder = 162 * ptriplot3d->nResFactor;
    ptriplot3d->Xform2DData();
    GRPrintOpen(apszArg[1], ptriplot3d->nResFactor);
    Redraw();
    GRPrintClose();
  }

  GRFreeGraphics();

  return 0;
}

