#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#if defined(UNIX)
#include <unistd.h>
#define	 O_BINARY 0x0000
#else
#include <io.h>
#endif

#include <fcntl.h>


// boolean constants
#define TRUE			1
#define FALSE			0


// global constants
#define LEN_IDENT		15
#define LEN_FILENAME		128


// global type definitions
typedef int		  BOOL;				      // f
typedef unsigned int	  UINT;				      // u
typedef unsigned char	  UCHAR;			      // uch
typedef int		  I3ARRAY[3];			      // i3a
typedef void *		  PVOID;			      // pv
typedef int *		  PBOOL;			      // pf
typedef double		  REAL;				      // r
typedef double *	  PREAL;			      // pr


// mesh header
typedef struct tagMESHHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cConstant;			// number of constants
  UINT cLabel;				// number of labels
  UINT cArray;				// number of 1D arrays
  UINT cRegLabel;			// number of region labels
  UINT cBndryNode;			// number of boundary nodes
  UINT cNode;				// number of nodes
  UINT cEdge;				// number of edges
  UINT cTriangle;			// number of triangles
  UINT fCylGeom;			// cylindrical geometry flag
} MESHHEAD;


// constant structure definition
typedef struct tagXLATCONST
{ char szName[LEN_IDENT+1];		// constant name
  UINT uType;				// constant type
  union
  { REAL r;				// value of real constant
    int	 n;				// value of integer constant
  } data;
} XLATCONST;


// label structure definition
typedef struct tagXLATLABEL
{ char szName[LEN_IDENT+1];		// label name
  UINT iIndex;				// label index
  UINT uType;				// label type
  UINT cNode;				// number of nodes
} XLATLABEL;


// refinement structure definition
typedef struct tagXLATREFINE
{ char szName[LEN_IDENT+1];		// refinement variable name
  UINT iIndex;				// refinement variable index
  UINT uMeasure;			// type of measurement
  REAL Distance;			// minimum "distance"
} XLATREFINE;


typedef UINT INODE;			// inode
typedef UINT IEDGE;			// iedge
typedef UINT ITRI;			// itri


// node structure definition
typedef struct tagNODE			// node
{ REAL	x;				// x coordinate
  REAL	y;				// y coordinate
} NODE;
typedef NODE *PNODE;			// pnode


// edge structure definition
typedef struct tagEDGE			// edge
{ UINT	uLabel;				// edge label
  INODE inodeA;				// index of node A
  INODE inodeB;				// index of node B
} EDGE;
typedef EDGE *PEDGE;			// pedge


// triangle structure definition
typedef struct tagTRI			// tri
{ UINT	uLabel;				// region label
  IEDGE iedgeAB;			// index of edge AB
  IEDGE iedgeBC;			// index of edge BC
  IEDGE iedgeAC;			// index of edge AC or CD
  IEDGE iedgeAD;			// index of edge AD
  ITRI	itriAB;				// index of elem. adj. to edge AB
  ITRI	itriBC;				// index of elem. adj. to edge BC
  ITRI	itriAC;				// index of elem. adj. to edge AC or CD
  ITRI	itriAD;				// index of elem. adj. to edge AD
} TRI;
typedef TRI *PTRI;			// ptri


// edge information structure
typedef struct tagEDGEINFO		// edgeinfo
{ REAL	ilen;				// integration length
  REAL	elen;				// edge length
  REAL	Area1;				// partial area 1
  REAL	Area2;				// partial area 1
  IEDGE iedge;				// edge index
  INODE inode;				// node index
  ITRI	ielem;				// element index
} EDGEINFO;


// node information structure
typedef struct tagNODEINFO		// nodeinfo
{ REAL Area;				// area of integration box
  UINT cNeighbor;			// number of neighbors
  UINT cTriangle;			// number of triangles
  EDGEINFO *aedgeinfo;			// edge information array
} NODEINFO;


#define TYPE_EDGE	      7		// edge label
#define TYPE_REGION	      8		// region label

#define SIGNATURE_MESH	      "@~!__MESH__!~@"



int main (int cArg, char **apszArg)
{
  fprintf(stdout, "SG Mesh Dump	       Version 1.0  (c) 1997  K. Kramer\n");
  if (cArg != 2)
  { fprintf(stdout, "usage:  dumpmesh filename\n");
    return 1;
  }

  // open the mesh file
  int nFile = open(apszArg[1], O_RDONLY | O_BINARY);
  if (nFile == -1)
  { fprintf(stdout, "cannot open file \"%s\"\n", apszArg[1]);
    return 2;
  }

  // read the header
  MESHHEAD meshhead;
  read(nFile, &meshhead, sizeof(MESHHEAD));
  if (strcmp(meshhead.szSign, SIGNATURE_MESH))
  { fprintf(stdout, "corrupt mesh file \"%s\"\n", apszArg[1]);
    return 2;
  }

  // read constants and add them to the symbol table
  UINT cConstant;
  read(nFile, &cConstant, sizeof(UINT));
  if (cConstant != meshhead.cConstant)
  { fprintf(stdout, "corrupt mesh file \"%s\"\n", apszArg[1]);
    return 2;
  }
  fprintf(stdout, "CONSTANTS:\n");
  UINT i;
  for(i = 0; i < cConstant; ++i)
  { XLATCONST xlatconst;
    read(nFile, &xlatconst, sizeof(XLATCONST));
    if (xlatconst.uType == 2)
	 fprintf(stdout, "%s = %d\n", xlatconst.szName, xlatconst.data.n);
    else fprintf(stdout, "%s = %lg\n", xlatconst.szName, xlatconst.data.r);
  }
  fprintf(stdout, "\n");

  // read the labels
  UINT cLabel;
  read(nFile, &cLabel, sizeof(UINT));
  if (meshhead.cLabel != cLabel)
  { fprintf(stdout, "corrupt mesh file \"%s\"\n", apszArg[1]);
    return 2;
  }
  fprintf(stdout, "LABELS:\n");
  for(i = 0; i < cLabel; ++i)
  { XLATLABEL xlatlabel;
    read(nFile, &xlatlabel, sizeof(XLATLABEL));
    fprintf(stdout, "%u) %s[%u] = {", xlatlabel.iIndex, xlatlabel.szName, xlatlabel.cNode);
    for(UINT j = 0; j < xlatlabel.cNode; ++j)
    { if (j % 8 == 0) fprintf(stdout, "\n");
      UINT iNode;
      read(nFile, &iNode, sizeof(UINT));
      fprintf(stdout, "%5u ", iNode);
    }
    fprintf(stdout, "\n}\n");
  }
  fprintf(stdout, "\n");

  UINT cRegLabel  = meshhead.cRegLabel;
  UINT cEdgeLabel = cLabel - cRegLabel;
  fprintf(stdout, "%u of the %u labels are region labels\n", cRegLabel, cLabel);
  fprintf(stdout, "%u of the %u labels are edge labels\n\n", cEdgeLabel, cLabel);

  // read the 1D arrays
  UINT cArray;
  read(nFile, &cArray, sizeof(UINT));
  if (meshhead.cArray != cArray)
  { fprintf(stdout, "corrupt mesh file \"%s\"\n", apszArg[1]);
    return 2;
  }
  fprintf(stdout, "1D ARRAYS:\n");
  for(i = 0; i < cArray; ++i)
  { char szArrayName[LEN_IDENT+1];
    read(nFile, szArrayName, LEN_IDENT+1);
    fprintf(stdout, "%s\n", szArrayName);
    lseek(nFile, meshhead.cNode * sizeof(REAL), SEEK_CUR);
  }
  fprintf(stdout, "\n");

  // read and print the nodes
  fprintf(stdout, "%u of the %u nodes are boundary nodes\n\n", meshhead.cBndryNode, meshhead.cNode);
  UINT cNode;
  read(nFile, &cNode, sizeof(UINT));
  if (cNode != meshhead.cNode)
  { fprintf(stdout, "corrupt mesh file \"%s\"\n", apszArg[1]);
    return 2;
  }
  NODE *anode = new NODE[cNode];
  read(nFile, anode, cNode * sizeof(NODE));
  fprintf(stdout, "NODES:\n");
  fprintf(stdout, "  i	 |	  x		  y	  \n");
  fprintf(stdout, "----- | --------------- ---------------\n");
  PNODE pnode = anode;
  for(INODE inode = 0; inode < cNode; ++inode, ++pnode)
    fprintf(stdout, "%5u | %15.8le %15.8le\n", inode, pnode->x, pnode->y);
  fprintf(stdout, "\n");

  UINT cEdge;
  read(nFile, &cEdge, sizeof(UINT));
  if (cEdge != meshhead.cEdge)
  { fprintf(stdout, "corrupt mesh file \"%s\"\n", apszArg[1]);
    return 2;
  }
  EDGE *aedge = new EDGE[cEdge];
  read(nFile, aedge, cEdge * sizeof(EDGE));
  fprintf(stdout, "EDGES:\n");
  fprintf(stdout, "	 |     |    NODES   \n");
  fprintf(stdout, "  i	 |  u  |   A	 B  \n");
  fprintf(stdout, "----- | --- | ----- -----\n");
  PEDGE pedge = aedge;
  for(IEDGE iedge = 0; iedge < cEdge; ++iedge, ++pedge)
  { fprintf(stdout, "%5u | ", iedge);
    if (pedge->uLabel != -1u) fprintf(stdout, "%3u | ", pedge->uLabel); else fprintf(stdout, "	  | ");
    fprintf(stdout, "%5u %5u\n", pedge->inodeA, pedge->inodeB);
  }
  fprintf(stdout, "\n");

  UINT cTri;
  read(nFile, &cTri, sizeof(UINT));
  if (cTri != meshhead.cTriangle)
  { fprintf(stdout, "corrupt mesh file \"%s\"\n", apszArg[1]);
    return 2;
  }
  TRI  *atri  = new TRI [cTri];
  read(nFile, atri, cTri * sizeof(TRI));
  fprintf(stdout, "TRIANGLES:\n");
  fprintf(stdout, "	 |     |	  EDGES			 TRIANGLES\n");
  fprintf(stdout, "  i	 |  u  |   AB	 BC  AC/CD   AD	    AB	  BC  AC/CD   AD\n");
  fprintf(stdout, "----- | --- | ----- ----- ----- -----  ----- ----- ----- -----\n");
  PTRI ptri = atri;
  for(ITRI itri = 0; itri < cTri; ++itri, ++ptri)
  { fprintf(stdout, "%5u | ", itri);
    if (ptri->uLabel != -1u) fprintf(stdout, "%3u | ", ptri->uLabel);  else fprintf(stdout, "	 | ");

    if (ptri->iedgeAB != -1u) fprintf(stdout, "%5u ", ptri->iedgeAB); else fprintf(stdout, "	  ");
    if (ptri->iedgeBC != -1u) fprintf(stdout, "%5u ", ptri->iedgeBC); else fprintf(stdout, "	  ");
    if (ptri->iedgeAC != -1u) fprintf(stdout, "%5u ", ptri->iedgeAC); else fprintf(stdout, "	  ");
    if (ptri->iedgeAD != -1u) fprintf(stdout, "%5u ", ptri->iedgeAD); else fprintf(stdout, "  n/a ");
    fprintf(stdout, " ");
    if (ptri->itriAB != -1u) fprintf(stdout, "%5u ", ptri->itriAB); else fprintf(stdout, "	");
    if (ptri->itriBC != -1u) fprintf(stdout, "%5u ", ptri->itriBC); else fprintf(stdout, "	");
    if (ptri->itriAC != -1u) fprintf(stdout, "%5u ", ptri->itriAC); else fprintf(stdout, "	");
    if (ptri->itriAD != -1u) fprintf(stdout, "%5u" , ptri->itriAD); else fprintf(stdout, "  n/a");
    fprintf(stdout, "\n");
  }
  fprintf(stdout, "\n");

  // read adjacency information
  UINT cAdj;
  read(nFile, &cAdj, sizeof(UINT));
  if (cAdj != meshhead.cNode + meshhead.cBndryNode * meshhead.cRegLabel)
  { fprintf(stdout, "corrupt mesh file \"%s\"\n", apszArg[1]);
    return 2;
  }
  NODEINFO nodeinfo;
  EDGEINFO edgeinfo;

  fprintf(stdout, "		  EDGE ADJACENCY INFORMATION\n");
  fprintf(stdout, "  inode iedge ielem label   elen	ilen	area 1	 area 2\n\n");

  for(i = 0; i < cAdj; ++i)
  { INODE inode;
    read(nFile, &inode, sizeof(INODE));
    read(nFile, &nodeinfo, sizeof(NODEINFO) - sizeof(EDGEINFO *));
    if (inode >= meshhead.cNode)
    { UINT u1 = inode - meshhead.cNode;
      UINT u2 = u1 / meshhead.cRegLabel;
      UINT u3 = u1 % meshhead.cRegLabel;
      inode = u2;
      fprintf(stdout, "i = %5u [%5u,%5u] N = %5u, T = %5u, A = %8.2le\n", inode, u2, u3, nodeinfo.cNeighbor, nodeinfo.cTriangle, nodeinfo.Area);
    }
    else fprintf(stdout, "i = %5u		N = %5u, T = %5u, A = %8.2le\n", inode, nodeinfo.cNeighbor, nodeinfo.cTriangle, nodeinfo.Area);

    for(UINT j = 0; j < nodeinfo.cNeighbor; ++j)
    { read(nFile, &edgeinfo, sizeof(EDGEINFO));
      EDGE *pedge = aedge + edgeinfo.iedge;
      UINT inodeN = (pedge->inodeA == inode) ? pedge->inodeB : pedge->inodeA;
      if (inodeN != edgeinfo.inode) fprintf(stdout, "ERROR!  node index is wrong! [%u,%u]\n", inodeN, edgeinfo.inode);
      fprintf(stdout, "	 %5u %5u %5u %5d %8.2le %8.2le %8.2le %8.2le\n",
	inodeN, edgeinfo.iedge, edgeinfo.ielem, pedge->uLabel, edgeinfo.elen, edgeinfo.ilen, edgeinfo.Area1, edgeinfo.Area2);
    }
    fprintf(stdout, "\n");
  }

  // deallocate arrays
  delete [] anode;
  delete [] aedge;
  delete [] atri;
  return 0;
}

