/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe mesh.exe refine.lib
 FILE:	     mathfunc.h
 AUTHOR:     Kevin M. Kramer
             J. Joshua Feng

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/21/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version
 06/22/00    JJF      asin, .. etc

 DESCRIPTION:

 This module contains mathematical functions that are not a part of the
 standard C math library
-----------------------------------------------------------------------------*/


#ifndef __mathfunc_h
#define __mathfunc_h


#define LEN_FUNCNAME 5


// internal function table
typedef struct tagIFT			// ift
{ char sz[LEN_FUNCNAME+1];		// function name
  UINT c;				// number of arguments
  PRF  aprf[3];				// function and derivative pointers
} IFT;


// make table and table count global
extern IFT aift[];
extern const UINT cFuncs;


/* define the constant PI */
#ifndef PI
#define PI 3.14159265358979323846
#endif

#ifndef LOG10e
#define LOG10e 0.434294481903251827651
#endif

#ifndef LOGe10
#define LOGe10 2.30258509299404568402
#endif

/* public macros to rename C math library routines */
/* #define arcsin(x)     asin(x) */
/* #define arccos(x)     acos(x) */
/* #define arctan(x)     atan(x) */
#define abs(x)	      fabs(x)


// function definitions
#define FUNC_EXP	    0		// exp(x)
#define FUNC_LN		    1		// ln(x)
#define FUNC_POW10	    2		// 10^x
#define FUNC_LOG	    3		// log(x)
#define FUNC_POW	    4		// x^y
#define FUNC_SQ		    5		// x*x
#define FUNC_SQRT	    6		// sqrt(x)
#define FUNC_INV	    7		// 1/x
#define FUNC_SINH	    8		// sinh(x)
#define FUNC_COSH	    9		// cosh(x)
#define FUNC_TANH	    10		// tanh(x)
#define FUNC_COTH	    11		// coth(x)
#define FUNC_SECH	    12		// sech(x)
#define FUNC_CSCH	    13		// csch(x)
#define FUNC_SIN	    14		// sin(x)
#define FUNC_COS	    15		// cos(x)
#define FUNC_TAN	    16		// tan(x)
#define FUNC_COT	    17		// cot(x)
#define FUNC_SEC	    18		// sec(x)
#define FUNC_CSC	    19		// csc(x)
#define FUNC_ERF	    20		// erf(x)
#define FUNC_AVE	    21		// (x + y)/2 {average}
#define FUNC_ABS	    22		// |x| {absolute value}
#define FUNC_SIGN	    23		// sign function
#define FUNC_STEP	    24		// step function
#define FUNC_MIN	    25		// min(x,y)
#define FUNC_MAX	    26		// max(x,y)
#define FUNC_BERN	    27		// bern
#define FUNC_AUX1           28          // aux1(x)
#define FUNC_AUX2           29          // aux2(x)
#define FUNC_NSDEP	    30		// nsdep(x,W,Dt)
#define FUNC_NGDEP	    31		// nsdep(x,Rp,dRp,Dt)

#define FUNC_ASIN           32          // asin(x)
#define FUNC_ACOS           33          // acos(x)
#define FUNC_ATAN           34          // atan(x)
#define FUNC_ACOT           35          // acot(x)
#define FUNC_ASEC           36          // asec(x)
#define FUNC_ACSC           37          // acsc(x)
#define FUNC_ASINH          38          // asinh(x)
#define FUNC_ACOSH          39          // acosh(x)
#define FUNC_ATANH          40          // atanh(x)
#define FUNC_ACOTH          41          // acoth(x)
#define FUNC_ASECH          42          // asech(x)
#define FUNC_ACSCH          43          // acsch(x)

///////////////////////////////////////////////////////////////////////////////
////////////////////////////  PUBLIC FUNCTIONS	///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


REAL	pd1exp	       (REAL);
REAL	pd1log	       (REAL);
REAL	pd1pow10       (REAL);
REAL	pd1log10       (REAL);
REAL	pd1pow	       (REAL, REAL);
REAL	pd2pow	       (REAL, REAL);
REAL	sq	       (REAL);
REAL	pd1sq	       (REAL);
REAL	pd1sqrt	       (REAL);
REAL	inv	       (REAL);
REAL	pd1inv	       (REAL);
REAL	pd1sinh	       (REAL);
REAL	pd1cosh	       (REAL);
REAL	pd1tanh	       (REAL);
REAL	coth	       (REAL);
REAL	pd1coth	       (REAL);
REAL	sech	       (REAL);
REAL	pd1sech	       (REAL);
REAL	csch	       (REAL);
REAL	pd1csch	       (REAL);
REAL	pd1sin	       (REAL);
REAL	pd1cos	       (REAL);
REAL	pd1tan	       (REAL);
REAL	cot	       (REAL);
REAL	pd1cot	       (REAL);
REAL	sec	       (REAL);
REAL	pd1sec	       (REAL);
REAL	csc	       (REAL);
REAL	pd1csc	       (REAL);
REAL	erf	       (REAL);
REAL	pd1erf	       (REAL);
REAL	ave	       (REAL, REAL);
REAL	pd1ave	       (REAL, REAL);
REAL	pd2ave	       (REAL, REAL);
REAL	pd1abs	       (REAL);
REAL	sign	       (REAL);
REAL	pd1sign	       (REAL);
REAL	step	       (REAL);
REAL	pd1step	       (REAL);
REAL	fmin	       (REAL, REAL);
REAL	pd1min	       (REAL, REAL);
REAL	pd2min	       (REAL, REAL);
REAL	fmax	       (REAL, REAL);
REAL	pd1max	       (REAL, REAL);
REAL	pd2max	       (REAL, REAL);
REAL	bern	       (REAL);
REAL	pd1bern	       (REAL);
REAL	aux1	       (REAL);
REAL	pd1aux1	       (REAL);
REAL	aux2	       (REAL);
REAL	pd1aux2	       (REAL);
REAL	nsdep	       (REAL, REAL, REAL);
REAL	ngdep	       (REAL, REAL, REAL, REAL, REAL);

REAL    pd1asin        (REAL);
REAL    pd1acos        (REAL);
REAL    pd1atan        (REAL);
REAL    acot           (REAL);
REAL    pd1acot        (REAL);
REAL    asec           (REAL);
REAL    pd1asec        (REAL);
REAL    acsc           (REAL);
REAL    pd1acsc        (REAL);
REAL    pd1asinh       (REAL);
REAL    pd1acosh       (REAL);
REAL    pd1atanh       (REAL);
REAL    acoth          (REAL);
REAL    pd1acoth       (REAL);
REAL    asech          (REAL);
REAL    pd1asech       (REAL);
REAL    acsch          (REAL);
REAL    pd1acsch       (REAL);

#endif

