/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  SGFramework C++ Code Generation Library
 FILE:	     sgnam.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 09/15/94    KMK      initial implementation
 06/22/00    JJF      asin, acos, atan

 DESCRIPTION:

 This module contains the functions needed by the code generated by the
 SGFramework.

-----------------------------------------------------------------------------*/


#ifndef __sgnam_h
#define __sgnam_h


///////////////////////////////////////////////////////////////////////////////
////////////////////////////  TYPE DEFINITIONS	///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


#define TRUE  1
#define FALSE 0

typedef int	     BOOL;
typedef unsigned int UINT;
typedef int	     I3ARRAY[3];
typedef void	    (*PFV) (BOOL);

#if   defined(SPREC)
typedef float	     REAL;
typedef float	    (*PFEQU) (I3ARRAY);
typedef void	    (*PFPD)  (float *, UINT *, I3ARRAY);
#elif defined(LPREC)
typedef long double  REAL;
typedef long double (*PFEQU) (I3ARRAY);
typedef void	    (*PFPD)  (long double *, UINT *, I3ARRAY);
#else
typedef double	     REAL;
typedef double	    (*PFEQU) (I3ARRAY);
typedef void	    (*PFPD)  (double *, UINT *, I3ARRAY);
#endif


///////////////////////////////////////////////////////////////////////////////
///////////////	 MISCELLANEOUS CONSTANTS, MACROS AND VARIABLES	///////////////
///////////////////////////////////////////////////////////////////////////////


#define LEN_IDENT	    15		// identifier length
#define LEN_FILENAME	    128		// maximum file name length
#define MAX_NORM	    20		// maximum norm degree
#define TYPE_RCONST	    3		// real constant
#define TYPE_ICONST	    4		// integer constant

typedef struct tagRESHEAD		// reshead
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  char szMeshFile[128];			// name of mesh file
  UINT cConstant;			// number of constants
  UINT cVariable;			// number of variables
  UINT cArray;				// number of arrays
  UINT cElement;			// number of elements
  UINT c1DArray;			// number of 1D arrays
  UINT c2DArray;			// number of 2D arrays
  UINT c3DArray;			// number of 3D arrays
  UINT cSet;				// number of data sets
} RESHEAD;

typedef struct tagTOPHEAD		// tophead
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cElement;			// number of variables & array elements
  UINT cTag;				// number of unknown elements
  UINT cTopology;			// number of nonzeroes excluding pivots
  UINT cEquation;			// number of equations
} TOPHEAD;

typedef struct tagPERMHEAD		// permhead
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  char szPerm[64];			// ordering algorithm
  UINT cTag;				// number of unknown elements
} PERMHEAD;

typedef struct tagXLATCONST
{ char szName[LEN_IDENT+1];		// constant name
  UINT uType;				// constant type
  union					//
  { REAL r;				// value of real constant
    int	 n;				// value of integer constant
  } data;
} XLATCONST;

typedef struct tagXLATVAR
{ char szName[LEN_IDENT+1];		// variable name
  UINT uOffset;				// variable offset
} XLATVAR;

typedef struct tagXLATARRAY
{ char szName[LEN_IDENT+1];		// array name
  UINT uOffset;				// array offset (in elm array of *.res)
  UINT cDim;				// number of dimensions
  UINT acElements[3];			// elements per dimensions
} XLATARRAY;

typedef struct tagTAG			// tag
{ UINT iVar;				// element array index (i.e. offset)
  UINT iEqu;				// equation index
  UINT iArray;				// array index ... -1u == variable
} TAG;

typedef struct tagTOP			// top
{ UINT i;				// index in topology array
  UINT c;				// number of nonzeroes in equation
} TOP;

typedef struct tagMESHHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cConstant;			// number of constants
  UINT cLabel;				// number of labels
  UINT cArray;				// number of 1D arrays
  UINT cRegLabel;			// number of region labels
  UINT cBndryNode;			// number of boundary nodes
  UINT cNode;				// number of nodes
  UINT cEdge;				// number of edges
  UINT cTriangle;			// number of triangles
  BOOL fCylGeom;			// cylindrical geometry flag
} MESHHEAD;

typedef struct tagXLATLABEL
{ char szName[LEN_IDENT+1];		// label name
  UINT iIndex;				// label index
  UINT uType;				// label type
  UINT cNode;				// number of nodes
} XLATLABEL;

typedef UINT INODE;			// inode
typedef UINT IEDGE;			// iedge
typedef UINT ITRI;			// itri

typedef struct tagNODE			// node
{ REAL	x;				// x coordinate
  REAL	y;				// y coordinate
} NODE;

typedef struct tagEDGE			// edge
{ UINT	uLabel;				// edge label
  INODE inodeA;				// index of node A
  INODE inodeB;				// index of node B
} EDGE;

typedef struct tagTRI			// tri
{ UINT	uLabel;				// region label
  IEDGE iedgeAB;			// index of edge AB
  IEDGE iedgeBC;			// index of edge BC
  IEDGE iedgeAC;			// index of edge AC or CD
  IEDGE iedgeAD;			// index of edge AD
  ITRI	itriAB;				// index of tri. adj. to edge AB
  ITRI	itriBC;				// index of tri. adj. to edge BC
  ITRI	itriAC;				// index of tri. adj. to edge AC or CD
  ITRI	itriAD;				// index of tri. adj. to edge AD
} TRI;

// edge information structure
typedef struct tagEDGEINFO		// edgeinfo
{ REAL	ilen;				// integration length
  REAL	elen;				// edge length
  REAL	Area1;				// partial area 1
  REAL	Area2;				// partial area 2
  IEDGE iedge;				// edge index
  INODE inode;				// node index
  ITRI	ielem;				// element index
} EDGEINFO;


// node information structure
typedef struct tagNODEINFO		// nodeinfo
{ REAL Area;				// area of integration box
  UINT cNeighbor;			// number of neighbors
  UINT cTriangle;			// number of triangles
  EDGEINFO *aedgeinfo;			// edge information array
} NODEINFO;

extern unsigned long iSolveIter;	// solution iteration
extern UINT iNewtIter;			// Newton iteration
extern UINT iLinIter;			// linear solution algorithm iteration
extern UINT iDampFact;			// damping factor iteration
extern REAL rNormInit;			// initalial norm of F(x)
extern REAL rNormFinal;			// final norm of F(x)
extern REAL rNormRHS;			// current norm of F(x)


///////////////////////////////////////////////////////////////////////////////
///////////////////////////  ERROR HANDLER CLASS  /////////////////////////////
///////////////////////////////////////////////////////////////////////////////


#define EHM_MESSAGES	0		// too many errors or warnings
#define EHM_OPENFILE	1		// cannot open file "?"
#define EHM_FINDLABEL	2		// cannot find label "?" in data file "?"
#define EHM_SYNTAX	3		// syntax error in data file "?"
#define EHM_ENDOFFILE	4		// unexpected eof in data file "?"
#define EHM_NEARSING	5		// near singularity encountered in ?
#define EHM_DOMAIN	6		// domain error
#define EHM_SINGULARITY 7		// singularity
#define EHM_OVERFLOW	8		// overflow
#define EHM_UNDERFLOW	9		// underflow
#define EHM_TOTALLOSS	10		// total loss of precision

class ErrorHandler
{ private:
    UINT cMaxErrors;			// maximum number of errors
    UINT cMaxWarnings;			// maximum number of warnings
    UINT cErrors;			// number of errors in message array
    UINT cWarnings;			// number of warnings in message array
    UINT cFatal;			// number of fatal errors
    UINT cTotal;			// total number of messages

    void UpdateVariables (UINT);

  public:
    BOOL fFatal	    ()	{ return (BOOL) cFatal; }
    BOOL fError	    ()	{ return (BOOL) cErrors; }
    BOOL fWarning   ()	{ return (BOOL) cWarnings; }
    UINT cNumErrors ()	{ return cTotal; }
    void AddMessage (UINT);
    void AddMessage (UINT, char *);
    void AddMessage (UINT, char *, char *);

  public:
    ErrorHandler (UINT = 25, UINT = 25);
};

extern ErrorHandler *error;


///////////////////////////////////////////////////////////////////////////////
//////////////	NUMERICAL ALGORITHM DEFINITIONS AND STRUCTURE  ////////////////
///////////////////////////////////////////////////////////////////////////////


#define LINSOLVE_SOR	    0
#define LINSOLVE_PCG	    1
#define LINSOLVE_GE	    2
#define SORT_NONE	    0
#define FILL_ZERO	    0
#define FILL_ONE	    1
#define FILL_TWO	    2
#define FILL_THREE	    3
#define FILL_FOUR	    4
#define FILL_FIVE	    5
#define FILL_INFINITY	    6
#define PRECOND_NONE	    0
#define PRECOND_DIAG	    1
#define PRECOND_ILU	    2

typedef struct tagNUMALG		// numalg
{ int  nEIiter;				// maximum number of Newton iterations
  REAL rERaccu;				// minimum Newton accuracy
  int  nEIdamp;				// damping factor
  UINT uILsolu;				// linear solution algorithm
  UINT uILsort;				// sorting algorithm
  UINT uILfill;				// maximum fill level
  UINT uILprec;				// preconditioner
  int  nEIlite;				// maximum number of linear iterations
  REAL rERlacc;				// minimum linear accuracy
  REAL rERover;				// overrelaxation parameter
  BOOL fCBzero;				// zero initial guess
  REAL rERdist;				// distance scaling factor
} NUMALG;


///////////////////////////////////////////////////////////////////////////////
///////////////	 FILE I/O CONSTANTS, VARIABLES AND FUNCTIONS  /////////////////
///////////////////////////////////////////////////////////////////////////////



#define IOMODE_READ	     0
#define IOMODE_WRITE	     1
#define IOMODE_APPEND	     2
#define STATUS_SUCCESS	     0
#define STATUS_FAILURE	    -1
#define STATUS_NOTCONVERGED -2

int  FileIO_Open    (char *, UINT);
int  FileIO_Close   ();
int  FileIO_Goto    (char *);
int  FileIO_Label   (char *);
int  FileIO_Comment (char *);
int  FileIO_Read    (UINT, BOOL, XLATVAR *, XLATARRAY *, REAL *);
int  FileIO_Write   (UINT, BOOL, XLATVAR *, XLATARRAY *, REAL *);


///////////////////////////////////////////////////////////////////////////////
//////////////	MATHEMATICAL CONSTANTS, MACROS AND FUNCTIONS  /////////////////
///////////////////////////////////////////////////////////////////////////////


/* define the constant PI */
#ifndef PI
#define PI 3.14159265358979323846
#endif

#ifndef LOG10e
#define LOG10e 0.434294481903251827651
#endif

#ifndef LOGe10
#define LOGe10 2.30258509299404568402
#endif

/* #define arcsin(x)     asin(x) */
/* #define arccos(x)     acos(x) */
/* #define arctan(x)     atan(x) */
#define abs(x)	      fabs(x)

REAL	pd1exp		(REAL);
REAL	pd1log		(REAL);
REAL	pow10		(REAL);
REAL	pd1pow10	(REAL);
REAL	pd1log10	(REAL);
REAL	pd1pow		(REAL, REAL);
REAL	pd2pow		(REAL, REAL);
REAL	sq		(REAL);
REAL	pd1sq		(REAL);
REAL	pd1sqrt		(REAL);
REAL	inv		(REAL);
REAL	pd1inv		(REAL);
REAL	pd1sinh		(REAL);
REAL	pd1cosh		(REAL);
REAL	pd1tanh		(REAL);
REAL	coth		(REAL);
REAL	pd1coth		(REAL);
REAL	sech		(REAL);
REAL	pd1sech		(REAL);
REAL	csch		(REAL);
REAL	pd1csch		(REAL);
REAL	pd1sin		(REAL);
REAL	pd1cos		(REAL);
REAL	pd1tan		(REAL);
REAL	cot		(REAL);
REAL	pd1cot		(REAL);
REAL	sec		(REAL);
REAL	pd1sec		(REAL);
REAL	csc		(REAL);
REAL	pd1csc		(REAL);
REAL	ave		(REAL, REAL);
REAL	pd1ave		(REAL, REAL);
REAL	pd2ave		(REAL, REAL);
REAL	erf		(REAL);
REAL	pd1erf		(REAL);
REAL	pd1abs		(REAL);
REAL	sign		(REAL);
REAL	pd1sign		(REAL);
REAL	step		(REAL);
REAL	pd1step		(REAL);
REAL	fmin		(REAL, REAL);
REAL	pd1min		(REAL, REAL);
REAL	pd2min		(REAL, REAL);
REAL	fmax		(REAL, REAL);
REAL	pd1max		(REAL, REAL);
REAL	pd2max		(REAL, REAL);
REAL	bern		(REAL);
REAL	pd1bern		(REAL);
REAL	aux1		(REAL);
REAL	pd1aux1		(REAL);
REAL	aux2		(REAL);
REAL	pd1aux2		(REAL);

REAL    pd1asin         (REAL);
REAL    pd1acos         (REAL);
REAL    pd1atan         (REAL);
REAL    acot            (REAL);
REAL    pd1acot         (REAL);
REAL    asec            (REAL);
REAL    pd1asec         (REAL);
REAL    acsc            (REAL);
REAL    pd1acsc         (REAL);
REAL    pd1asinh        (REAL);
REAL    pd1acosh        (REAL);
REAL    pd1atanh        (REAL);
REAL    acoth           (REAL);
REAL    pd1acoth        (REAL);
REAL    asech           (REAL);
REAL    pd1asech        (REAL);
REAL    acsch           (REAL);
REAL    pd1acsch        (REAL);

///////////////////////////////////////////////////////////////////////////////
/////////////////////////  COMMAND LINE PARSER CLASS  /////////////////////////
///////////////////////////////////////////////////////////////////////////////

class CmdParse
{ protected:
    UINT cSwitch;		// number of switch arguments
    UINT cPos;			// number of position arguments
    char **apszSwitch;		// switch argument pointer array
    char **apszPos;		// position argument pointer array

  public:
    inline UINT GetNumOfArgs	   ()	{ return cSwitch + cPos; }
    inline UINT GetNumOfSwitchArgs ()	{ return cSwitch; }
    inline UINT GetNumOfPosArgs	   ()	{ return cPos; }
    BOOL  fArgExist       (char *);
    int	  nGetSwitchArg   (char *, int, int, int);
    int	  nGetPosArg      (UINT, int, int, int);
    REAL  rGetSwitchArg   (char *, REAL, REAL, REAL);
    REAL  rGetPosArg      (UINT, REAL, REAL, REAL);
    char *pszGetSwitchArg (char *, char *);
    char *pszGetPosArg    (UINT, char *);

  public:
    CmdParse (UINT, char **);
   ~CmdParse ();
};

typedef struct tagCMDARGS
{ BOOL fQuiet;
  BOOL fZeroDXVector;
  UINT cMaxNewtIter;
  UINT cMaxLinIter;
  UINT cMaxDamping;
  UINT cMaxFillLevel;
  UINT cNorm;
  UINT uLinSolveAlg;
  UINT uSortAlg;
  UINT uPreconditioner;
  UINT uPrecision;
  UINT uScreenLevel;
  UINT uLogFileLevel;
  REAL rMinNewtAcc;
  REAL rMinLinAcc;
  REAL rNearSingularity;
  REAL rOverRelaxParam;
  REAL rPercentDX;
  char *pszLogFile;
} CMDARGS;

extern CmdParse *pcmd;
extern CMDARGS cmdargs;
int nParseCommandLine (NUMALG *pnumalg);


///////////////////////////////////////////////////////////////////////////////
/////////////////////////  NONZERO ELEMENT STRUCTURE  /////////////////////////
///////////////////////////////////////////////////////////////////////////////


typedef struct tagNONZERO		// nz
{ REAL r;				// value
  UINT i;				// column index
  struct tagNONZERO *pnz;		// next nonzero element
} NONZERO;



///////////////////////////////////////////////////////////////////////////////
///////////  DYNAMIC STACK CLASS OF POINTERS TO NONZERO STRUCTURES  ///////////
///////////////////////////////////////////////////////////////////////////////


class PNZStack
{ private:
    UINT u;		    // increment
    UINT c;		    // stack size
    UINT i;		    // stack index (top of the stack)
    NONZERO **stack;	    // stack
    void grow ()	    { c += u;
			      stack = (NONZERO **) realloc(stack,
				c * sizeof(NONZERO *)); }

  public:
    void push (NONZERO *T)  { if (i == c) grow(); stack[i++] = T; }
    NONZERO *pop ()	    { return stack[--i]; }
    BOOL empty ()	    { return !i; }
    void flush ()	    { i = 0; }

  public:
     PNZStack (UINT n = 64) { i = 0; stack = (NONZERO **) calloc(
                              (c = u = n), sizeof(NONZERO*)); }
    ~PNZStack ()	    { free((void *) stack); }
};


///////////////////////////////////////////////////////////////////////////////
/////////////////////////  NONZERO ELEMENT HEAP CLASS  ////////////////////////
///////////////////////////////////////////////////////////////////////////////


class NonZeroHeap
{ private:
    PNZStack stack;			// nonzero element stack
    NONZERO *pnzFree;			// free element list

  public:
    NONZERO * Alloc ();			// allocate a nonzero element
    void Dealloc (NONZERO *);		// deallocate a nonzero element

  public:
     NonZeroHeap ();
    ~NonZeroHeap ();
};

extern NonZeroHeap *pnzHeap;


///////////////////////////////////////////////////////////////////////////////
////////////////////////////  SPARSE MATRIX CLASS  ////////////////////////////
///////////////////////////////////////////////////////////////////////////////


class SparseMatrix			// mtrx
{ public:
    UINT c;				// dimension of square matrix
    BOOL fSorted;			// set if matrix is sorted.
    NONZERO **apnz;			// head of column link lists
    NONZERO **apnzPivot;		// array of pivot pointers
    BOOL *afLWR;			// logical working row
    REAL *arVWR;			// value working row
    REAL *arMax;			// maximum absolute value on each row
    UINT cDump;				// number of dumps

  public:
    void Dump ();			// dump the values
    void Deallocate ();			// deallocate the matrix elements
    void Transpose ();			// transpose the matrix
    UINT AddFill (UINT);		// add fill to matrix
    REAL Norm ();			// compute the infinite norm
    inline void Sort () { Transpose(); Transpose(); fSorted = TRUE; }

    // allocate the nonzero elements of the Jacobian matrix
    UINT AllocateJacobian (UINT *, UINT *, UINT *, TOP *);

    // compute the values of the Jacobian matrix elements
    int ComputeJacobian	 (UINT *, UINT *, TAG *, I3ARRAY *, PFPD *);

  public:
     SparseMatrix (UINT);		// constructor
    ~SparseMatrix ();			// destructor
};


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////	 VECTOR CLASS  ////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


class Vector				// vect
{ public:
    UINT c;				// dimension of vector
    REAL *ar;				// values of the vector
    UINT cDump;				// number of dumps

  public:
    void Dump ();			// dump the values
    void MakeZeroVector ();		// make the zero vector
    REAL Norm (UINT, REAL *);		// compute the nth norm

    // update tagged variables
    int UpdateTagVars (REAL, UINT *, TAG *, REAL *);

    // transfer values of unknown vector to variable structure
    int Vector2Variables (UINT *, TAG *, REAL *);

    // transfer values of unknown variables to the vector
    int Variables2Vector (UINT *, TAG *, REAL *);

    // make the right hand side vector
    int MakeRHSVector (UINT *, TAG *, I3ARRAY *, PFEQU *);

  public:
     Vector (UINT);			// constructor
    ~Vector ();				// destructor
};


///////////////////////////////////////////////////////////////////////////////
//////////////////////	LINEAR SOLUTION ALGORITHM CLASS	 //////////////////////
///////////////////////////////////////////////////////////////////////////////


#define LINSOLVE_SUCCESS    0
#define LINSOLVE_FAILURE   -1

class LinearSolve
{ public:
    virtual int Solve (SparseMatrix *, Vector *, Vector *);

  public:
    LinearSolve ();
    virtual ~LinearSolve ();
};


///////////////////////////////////////////////////////////////////////////////
//////////////////////	SUCCESSIVE OVERRELAXATION CLASS	 //////////////////////
///////////////////////////////////////////////////////////////////////////////


class SOR : public LinearSolve
{ protected:
    BOOL fZeroXVector;
    UINT uMaxIterations;
    REAL Accuracy;
    REAL OverRelaxation;
    REAL Result;

  public:
    int Solve (SparseMatrix *, Vector *, Vector *);

  public:
     SOR (UINT = 100, REAL = 0.0, REAL = 1.0, BOOL = TRUE);
    ~SOR ();
};


///////////////////////////////////////////////////////////////////////////////
//////////////////////////  GAUSSIAN ELIMINATION CLASS	///////////////////////
///////////////////////////////////////////////////////////////////////////////


class GaussElim : public LinearSolve
{ public:
    int Solve (SparseMatrix *, Vector *, Vector *);

  public:
     GaussElim ();
    ~GaussElim ();
};


///////////////////////////////////////////////////////////////////////////////
//////////////////////	TRIANGULAR MESH RELATED FUNCTIONS  ////////////////////
///////////////////////////////////////////////////////////////////////////////


UINT _bnode_map (UINT, UINT);		// map boundary nodes
void _nsum_init (UINT, UINT, BOOL);	// initialize node summation
void _nsum_next ();			// increment node summation
BOOL _nsum_loopover ();			// is the node summation over
UINT _nsum_index ();			// return the node summation index
void _init_edge_nodes(UINT, UINT *, UINT *);  // init the edge's nodes
UINT _label_init (UINT, UINT);		// initialize label loop
UINT _label_next (UINT);		// increment label loop
UINT _node (UINT, UINT);		// node index of jth neighbor of i
UINT _edge (UINT, UINT);		// edge index of jth neighbor of i
REAL _area (UINT);			// integration area of ith node
REAL _area1 (UINT, UINT);		// partial integration area #1
REAL _area2 (UINT, UINT);		// partial integration area #2
REAL _ilen (UINT, UINT);		// integration length
REAL _elen (UINT, UINT);		// edge length
REAL _elen (UINT);			// edge length
REAL _scheck (UINT, UINT);		// sign check
//REAL _pcfodd (UINT, UINT, UINT, REAL **, int, int);
//REAL _pcfodd (UINT, UINT, UINT, REAL **, int, int);
void ReadMeshFile (char *, REAL **);	// read mesh file
void InitCoordArrays (REAL *, REAL *);	// initialize coordinate arrays
extern REAL rDistScale;			// distance scaling factor
extern MESHHEAD meshhead;		// mesh header


///////////////////////////////////////////////////////////////////////////////
///////////////////////////  MISCELLANEOUS FUNCTIONS  /////////////////////////
///////////////////////////////////////////////////////////////////////////////


void Index2Name	        (UINT, char *, TAG *, XLATVAR *, XLATARRAY *);
void OutputLinearSystem (UINT, char *, SparseMatrix *, Vector *, Vector *,
			 UINT *, UINT *, TAG *, XLATVAR *, XLATARRAY *);
void CenterText	        (FILE *, char *, UINT);
void InitResultFile     (char *, RESHEAD *, UINT, UINT, UINT, UINT,
			 XLATCONST *, XLATVAR *, XLATARRAY *);
void InitNumAlg	        (LinearSolve **, SparseMatrix **, Vector **, Vector **,
			 UINT, UINT *, UINT *, UINT *, TOP *);
void ReadTopFile        (char *, UINT *, UINT *, UINT *, UINT *, UINT *,
			 UINT **, TAG **, TOP **, I3ARRAY **);
void ReadPermFile       (char *, UINT, UINT **, UINT **);
void CleanUp	        (UINT *, TAG *, TOP *, I3ARRAY *, UINT *, UINT *,
			 SparseMatrix *, Vector *, Vector *, LinearSolve *);
void WriteElements      (char *, RESHEAD *, REAL *);
int  SolveEquations     (LinearSolve *, SparseMatrix *, Vector *, Vector *,
			 UINT *, UINT *, TAG *, I3ARRAY *, REAL *,
			 PFEQU *, PFPD *, PFV);


#endif

