/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe mesh.exe
 FILE:	     mathfunc.cpp
 AUTHOR:     Kevin M. Kramer
             J. Joshua Feng

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/21/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version
 06/22/00    JJF      asin, .. etc

 DESCRIPTION:

 This module contains mathematical functions that are not a part of the
 standard C math library
-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "mathfunc.h"
#include "brkpnts.h"


///////////////////////////////////////////////////////////////////////////////
/////////////////////////////  PUBLIC FUNCTIONS	 //////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/* ----------------------------------------------------------------------------
 * exp:	 This function returns the e to the argument and resides in the C math
 * library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1exp:  This function returns the partial derivative of the exp function
 * with respect to its argument.
 */
REAL pd1exp ( REAL x )
{
  return exp(x);
}


/* ----------------------------------------------------------------------------
 * log:	 This function returns the natural logarithm of the argument and is
 * part of the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1log:  This function returns the partial derivative of the ln function
 * with respect to its argument.
 */
REAL pd1log ( REAL x )
{
  if (x == 0.0) { /* @@@ division by zero error */ }
  return(1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pow10:  This function returns 10 to the x and resides in the C math library
 * on some systems ...
 * --------------------------------------------------------------------------*/
REAL pow10 ( REAL x )
{
  return pow(10.0, x);
}


/* ----------------------------------------------------------------------------
 * pd1pow10:  This function returns the partial derivative of the pow10
 * function with respect to its argument.
 */
REAL pd1pow10 ( REAL x )
{
  return(LOGe10 * pow10(x));
}


/* ----------------------------------------------------------------------------
 * log10:  This function returns the logarithm base 10 of the argument and is
 * part of the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1log10:  This function returns the partial derivative of the log10
 * function with respect to its argument.
 */
REAL pd1log10 ( REAL x )
{
  if (x == 0.0) { /* @@@ division by zero error */ }
  return(LOG10e / x);
}


/* ----------------------------------------------------------------------------
 * pow:	 This function returns x to the y and resides in the C math library
 * --------------------------------------------------------------------------*/
#if defined(HPUX)
REAL power ( REAL x, REAL y )
{
  return (pow(x, y));
}
#endif
 

/* ----------------------------------------------------------------------------
 * pd1pow:  This function returns the partial derivative of the pow function
 * with respect to its first argument, i.e. 'x'.
 */
REAL pd1pow ( REAL x, REAL y )
{
  return(y * pow(x, y - 1));
}


/* ----------------------------------------------------------------------------
 * pd2pow:  This function returns the partial derivative of the pow function
 * with respect to its second argument, i.e. 'y'.
 */
REAL pd2pow ( REAL x, REAL y )
{
  return(pow(x, y) * log(x));
}


/* ----------------------------------------------------------------------------
 * sq:	This function returns the square of the argument.
 */
REAL sq ( REAL x )
{
  return(x * x);
}


/* ----------------------------------------------------------------------------
 * pd1sq:  This function returns the partial derivative of the square function
 * with respect to its argument.
 */
REAL pd1sq ( REAL x )
{
  return(2.0 * x);
}


/* ----------------------------------------------------------------------------
 * sqrt:  This function returns the square root of the argument and resides in
 * the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1sqrt:  This function returns the partial derivative of the square root
 * function with respect to its argument.
 */
REAL pd1sqrt ( REAL x )
{
  if (x <= 0.0) { /* @@@ domain error */ }
  return(0.5 / sqrt(x));
}


/* ----------------------------------------------------------------------------
 * inv:	 This function returns the inverse of the argument.
 */
REAL inv ( REAL x )
{
  if (x == 0.0) { /* @@@ division by zero error */ }
  return (1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1inv:  This function returns the partial derivative of the inverse
 * function with respect to its argument
 */
REAL pd1inv ( REAL x )
{
  if (x == 0.0) { /* @@@ division by zero error */ }
  return(1.0 / sq(x));
}


/* ----------------------------------------------------------------------------
 * sinh:  This function returns the hyperbolic sine of the argument and resides
 * in the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1sinh:  This function returns the partial derivative of the sinh function
 * with respect to its argument.
 */
REAL pd1sinh ( REAL x )
{
  return(cosh(x));
}


/* ----------------------------------------------------------------------------
 * cosh:  This function returns the hyperbolic cosine of the argument and
 * resides in the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1cosh:  This function returns the partial derivative of the cosh function
 * with respect to its argument.
 */
REAL pd1cosh ( REAL x )
{
  return(sinh(x));
}


/* ----------------------------------------------------------------------------
 * tanh:  This function returns the hyperbolic tangent of the argument and
 * resides in the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1tanh:  This function returns the partial derivative of the tanh function
 * with respect to its argument.
 */
REAL pd1tanh ( REAL x )
{
  REAL y = cosh(x);
  if (y == 0) { /* @@@ divide by zero error */ }
  return(sq(1.0 / y));
}


/* ----------------------------------------------------------------------------
 * coth:  This function returns the hyperbolic cotangent of the argument
 */
REAL coth ( REAL x )
{
  REAL y = tanh(x);
  if (y == 0.0) { /* @@@ division by zero error */ }
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1coth:  This function returns the partial derivative of the coth function
 * with respect to its argument.
 */
REAL pd1coth ( REAL x )
{
  REAL y = sinh(x);
  if (y == 0.0) { /* @@@ division by zero error */ }
  return(-1.0 / sq(y));
}


/* ----------------------------------------------------------------------------
 * sech:  This function returns the hyperbolic secant of the argument
 */
REAL sech ( REAL x )
{
  REAL y = cosh(x);
  if (y == 0.0) { /* @@@ division by zero error */ }
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1sech:  This function returns the partial derivative of the sech function
 * with respect to its argument
 */
REAL pd1sech ( REAL x )
{
  REAL y = cosh(x);
  if (y == 0.0) { /* @@@ division by zero error */ }
  return(-sinh(x) / sq(y));
}


/* ----------------------------------------------------------------------------
 * csch:  This function returns the hyperbolic cosecant of the argument.
 */
REAL csch ( REAL x )
{
  REAL y = sinh(x);
  if (y == 0.0) { /* @@@ division by zero error */ }
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1csch:  This function returns the partial derivative of the csch function
 * with respect to its argument.
 */
REAL pd1csch ( REAL x )
{
  REAL y = sinh(x);
  if (y == 0.0) { /* @@@ division by zero error */ }
  return(-cosh(x) / sq(y));
}


/* ----------------------------------------------------------------------------
 * sin:	 This function returns the sine of the argument and resides in the C
 * math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1sin:  This function returns the partial derivative of the sin function
 * with respect to its argument.
 */
REAL pd1sin ( REAL x )
{
  return(cos(x));
}


/* ----------------------------------------------------------------------------
 * cos:	 This function returns the cosine of the argument and resides in the C
 * math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1cos:  This function returns the partial derivative of the cos function
 * with respect to its argument.
 */
REAL pd1cos ( REAL x )
{
  return(-sin(x));
}


/* ----------------------------------------------------------------------------
 * tan:	 This function returns the tangent of the argument and resides in the C
 * math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1tan:  This function returns the partial derivative of the tan function
 * with respect to its argument.
 */
REAL pd1tan ( REAL x )
{
  REAL y = cos(x);
  if (y == 0) { /* @@@ divide by zero error */ }
  return(1.0 / sq(y));
}


/* ----------------------------------------------------------------------------
 * cot:	 This function returns the cotangent of the argument
 */
REAL cot ( REAL x )
{
  REAL y = tan(x);
  if (y == 0) { /* @@@ divide by zero error */ }
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1cot:  This function returns the partial derivative of the cot function
 * with respect to its argument.
 */
REAL pd1cot ( REAL x )
{
  REAL y = sin(x);
  if (y == 0) { /* @@@ divide by zero error */ }
  return(-1.0 / sq(y));
}


/* ----------------------------------------------------------------------------
 * sec:	 This function returns the secant of the argument.
 */
REAL sec ( REAL x )
{
  REAL y = cos(x);
  if (y == 0) { /* @@@ divide by zero error */ }
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1sec:  This function returns the partial derivative of the sec function
 * with respect to its argument.
 */
REAL pd1sec ( REAL x )
{
  REAL y = cos(x);
  if (y == 0) { /* @@@ divide by zero error */ }
  return(sin(x) / sq(y));
}


/* ----------------------------------------------------------------------------
 * csc:	 This function returns the cosecant of the argument.
 */
REAL csc ( REAL x )
{
  REAL y = sin(x);
  if (y == 0) { /* @@@ divide by zero error */ }
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1csc:  This function returns the partial derivative of the csc function
 * with respect to its argument.
 */
REAL pd1csc ( REAL x )
{
  REAL y = sin(x);
  if (y == 0) { /* @@@ divide by zero error */ }
  return(-cos(x) / sq(y));
}


/* ----------------------------------------------------------------------------
 * ave:	 This allows returns the average of the arguments.
 */
REAL ave ( REAL x, REAL y )
{
  return(0.5 * (x + y));
}


/* ----------------------------------------------------------------------------
 * pd1ave:  This function returns the partial derivative of the average
 * function with respect to its first argument, i.e. 'x'.
 */
REAL pd1ave ( REAL x, REAL y )
{
  return(0.5);
}


/* ----------------------------------------------------------------------------
 * pd2ave:  This function returns the partial derivative of the average
 * function with respect to its second argument, i.e. 'y'.
 */
REAL pd2ave ( REAL x, REAL y )
{
  return(0.5);
}


/* ----------------------------------------------------------------------------
 * erf:	 This function returns the error function.
 */
REAL erf ( REAL x )
{
  REAL t1 = 1.0 / (1.0 + 0.3275911 * fabs(x));
  REAL t2 = t1 * t1;
  REAL t3 = t2 * t1;
  REAL t4 = t3 * t1;
  REAL t5 = t4 * t1;
  REAL result = 1.0 - (0.254829592*t1 - 0.284496736*t2 + 1.421413741*t3 -
		       1.453152027*t4 + 1.061405429*t5) * exp(-x*x);
  return (x < 0.0) ? -result : result;
}


/* ----------------------------------------------------------------------------
 * pd1erf:  This function returns the derivative of the error function with
 * respect to the first variable.
 */
REAL pd1erf ( REAL x )
{
  return 2.0 / sqrt(PI) * exp(-x*x);
}


/*****************************************************************************/
/**********  THE FOLLOWING FUNCTIONS HAVE DISCONTINUOUS DERIVATIVES  *********/
/*****************************************************************************/


/* ----------------------------------------------------------------------------
 * abs:	 This function returns the absolute value of the argument and is
 * defined as a macro which expands to the fabs function in the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1abs:  This function returns the partial derivative of the abs function
 * with respect to its argument.
 */
REAL pd1abs ( REAL x )
{
  return((x >= 0.0) ? 1.0 : -1.0);
}


/* ----------------------------------------------------------------------------
 * sign:  This functions returns the sign of the argument, i.e., +1 if the
 * argument is postive, 0 if the argument is zero, and -1 is the argument is
 * negative
 */
REAL sign ( REAL x )
{
  /* CAUTION:  this function has a discontinous derivative at zero */
  if	  (x == 0.0) return( 0.0);
  else if (x  > 0.0) return( 1.0);
		     return(-1.0);
}


/* ----------------------------------------------------------------------------
 * pd1sign:  This function returns the partial derivative of the sign function
 * with respect to its argument
 */
REAL pd1sign ( REAL x )
{
  /* CAUTION:  this function is undefined at zero */
  return(0.0);
}


/* ----------------------------------------------------------------------------
 * step:  This function returns the one if the argument is zero or greater and
 * zero if the argument is less than zero.
 */
REAL step ( REAL x )
{
  /* CAUTION:  this function has a discontinous derivative at zero */
  return((x < 0.0) ? 0.0 : 1.0);
}


/* ----------------------------------------------------------------------------
 * pd1step:  This function returns the partial derivative of the step function
 * with respect to its argument.
 */
REAL pd1step ( REAL x )
{
  /* CAUTION:  this function is undefined at zero */
  return(0.0);
}


/* ----------------------------------------------------------------------------
 * fmin:  This function returns the smallest argument.
 */
REAL fmin ( REAL x, REAL y )
{
  /* CAUTION:  this function has a discontinous derivative at x = y */
  return((x < y) ? x : y);
}


/* ----------------------------------------------------------------------------
 * pd1min:  This function returns the partial derivative of the fmin function
 * with respect to its first argument, i.e. 'x'.
 */
REAL pd1min ( REAL x, REAL y )
{
  /* CAUTION:  this function is undefined at x = y */
  return((x < y) ? 1.0 : 0.0);
}


/* ----------------------------------------------------------------------------
 * pd2min:  This function returns the partial derivative of the fmin function
 * with respect to its second argument, i.e. 'y'.
 */
REAL pd2min ( REAL x, REAL y )
{
  /* CAUTION:  this function is undefined at x = y */
  return((y < x) ? 1.0 : 0.0);
}


/* ----------------------------------------------------------------------------
 * max:	 This function returns the largest argument.
 */
REAL fmax ( REAL x, REAL y )
{
  /* CAUTION:  this function has a discontinous derivative at x = y */
  return((x > y) ? x : y);
}


/* ----------------------------------------------------------------------------
 * pd1max:  This function returns the partial derivative of the fmax function
 * with respect to its first argument, i.e. 'x'.
 */
REAL pd1max ( REAL x, REAL y )
{
  /* CAUTION:  this function is undefined at x = y */
  return((x > y) ? 1.0 : 0.0);
}


/* ----------------------------------------------------------------------------
 * pd2max:  This function returns the partial derivative of the fmax function
 * with respect to its second argument, i.e. 'y'.
 */
REAL pd2max ( REAL x, REAL y )
{
  /* CAUTION:  this function is undefined at x = y */
  return((y > x) ? 1.0 : 0.0);
}


/* ----------------------------------------------------------------------------
 * NOTE:  the range limits for the Bernoulli and Aux functions were generated
 *	  by the "brkpnts" program and stored in brkpnts.h.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * bern:  This function returns the Bernoulli function of the argument.	 To
 * avoid under and overflows this function is defined by equivalent or approx-
 * imate functions depending upon the value of the argument.
 *
 *					  x
 *				B(x) = -------
 *				       e^x - 1
 *
 */
REAL bern ( REAL x )
{
  REAL y;

  if (x <= BP0_BERN)
  { return(-x); }
  else if (x <	BP1_BERN)
  { return(x / (exp(x) - 1.0)); }
  else if (x <= BP2_BERN)
  { return(1.0 - x/2.0 * (1.0 - x/6.0 * (1.0 - x*x/60.0))); }
  else if (x <	BP3_BERN)
  { y = exp(-x);   return((x * y) / (1.0 - y)); }
  else if (x <	BP4_BERN)
  { return(x * exp(-x)); }
  else { return(0.0); }

} /* bern */


/* ----------------------------------------------------------------------------
 * pd1bern:  This function returns the total derivative of the Bernoulli
 * function with respect to the argument.  To avoid under and overflows this
 * function is defined by equivalent or approximate functions depending upon
 * the value of the argument.
 *
 *			   d	    (1-x)*e^x - 1
 *			   --B(x) = -------------
 *			   dx	     (e^x - 1)^2
 */
REAL pd1bern ( REAL x )
{
  REAL y, z;

  if (x <= BP0_DBERN)
  { return(-1.0); }
  else if (x <= BP1_DBERN)
  { return((1.0 - x)*exp(x) - 1.0); }
  else if (x <	BP2_DBERN)
  { y = exp(x); z = y - 1.0; return(((1.0 - x)*y - 1.0)/(z*z)); }
  else if (x <= BP3_DBERN)
  { return(-0.5 + x/6.0 * (1.0 - x*x/30.0)); }
  else if (x <	BP4_DBERN)
  { y = exp(-x); z = 1 - y;  return(((1.0 - x)*y - y*y )/(z*z)); }
  else if (x <	BP5_DBERN)
  { y = exp(-x); return((1.0 - x)*y - y*y); }
  else { return(0.0); }

} /* pd1bern */


/* ----------------------------------------------------------------------------
 * aux1:  This function returns the aux1 function.  To avoid under and over-
 * flows this function is defined by equivalent or approximate functions
 * depending upon the value of the argument.
 *
 *					 x
 *			   Aux1(x) =  -------
 *				      sinh(x)
 */
REAL aux1 ( REAL x )
{
  if	  (x < -BP0_MISC) x = -BP0_MISC;
  else if (x >	BP0_MISC) x =  BP0_MISC;

  if (x <= BP0_AUX1)
  { return(x / sinh(x)); }
  else if (x <= BP1_AUX1)
  { return(1 - x*x/6.0*(1.0 - 7.0*x*x/60.0)); }
  else { return(x / sinh(x)); }

} /* aux1 */


/* ----------------------------------------------------------------------------
 * pd1aux1:  This function returns the total derivative of the aux1 function.
 * To avoid under and overflows this function is defined by equivalent or
 * approximate functions depending upon the value of the argument.
 *
 *		      d		  sinh(x) - x*cosh(x)
 *		      --Aux1(x) = -------------------
 *		      dx	      (sinh(x))^2
 */
REAL pd1aux1 ( REAL x )
{
  REAL y;

  if	  (x < -BP0_MISC) x = -BP0_MISC;
  else if (x >	BP0_MISC) x =  BP0_MISC;

  if (x <= BP0_DAUX1)
  { y = sinh(x); return((y - x*cosh(x))/(y*y)); }
  else if (x <= BP1_DAUX1)
  { return(-x/3.0*(1.0 - 7.0*x*x/30.0)); }
  else { y = sinh(x); return((y - x*cosh(x))/(y*y)); }

} /* pd1aux1 */


/* ----------------------------------------------------------------------------
 * aux2:  This function returns the aux2 function.  To avoid under and over-
 * flows this function is defined by equivalent or approximate functions
 * depending upon the value of the argument.
 *
 *					     1
 *				Aux2(x) = -------
 *					  1 + e^x
 */
REAL aux2 ( REAL x )
{
  if (x <= BP0_AUX2)
  { return(1.0); }
  else if (x <= BP1_AUX2)
  { return(1.0 / (1.0 + exp(x))); }
  else if (x <= BP2_AUX2)
  { return(exp(-x)); }
  else { return(0.0); }

} /* aux2 */


/* ----------------------------------------------------------------------------
 * pd1aux2:  This function returns the total derivative of the aux2 function.
 * To avoid under and overflows this function is defined by equivalent or
 * approximate functions depending upon the value of the argument.
 *
 *			   d		 - e^x
 *			   --Aux2(x) = -----------
 *			   dx	       (1 + e^x)^2
 */
REAL pd1aux2 ( REAL x )
{
  REAL y,z;

  if (x <= BP0_DAUX2)
  { return(0.0); }
  else if (x <= BP1_DAUX2)
  { return(-exp(x)); }
  else if (x <= BP2_DAUX2)
  { y = exp(x); z = y + 1.0; return(-y/(z*z)); }
  else if (x <= BP3_DAUX2)
  { return(-exp(-x)); }
  else { return(0.0); }

} /* pd1aux2 */

/* ----------------------------------------------------------------------------
 * asin: This function returns the arcsine of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1asin: This function returns the partial derivative of the asin function
 * with respect to its argument.
 */
REAL pd1asin ( REAL x )
{
  if ((x == 1.0) || (x == -1.0)) { /* @@@ divide by zero error */ }
  return(1.0 / sqrt(1.0 - x * x));
}

/* ----------------------------------------------------------------------------
 * acos: This function returns the arccosine of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1acos: This function returns the partial derivative of the acos function
 * with respect to its argument.
 */
REAL pd1acos ( REAL x )
{
  if ((x >= 1.0) || (x <= -1.0)) { /* @@@ divide by zero error */ }
  return(-1.0 / sqrt(1.0 - x * x));
}

/* ----------------------------------------------------------------------------
 * atan: This function returns the arctangent of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1atan: This function returns the partial derivative of the atan function
 * with respect to its argument.
 */
REAL pd1atan ( REAL x )
{
  return(1.0 / (1.0 + x * x));
}

/* ----------------------------------------------------------------------------
 * acot: This function returns the arccotangent of the argument.
 */
REAL acot ( REAL x )
{
  if (x == 0) { /* @@@ divide by zero error */ }
  return(atan(1.0 / x));
}


/* ----------------------------------------------------------------------------
 * pd1acot: This function returns the partial derivative of the acot function
 * with respect to its argument.
 */
REAL pd1acot ( REAL x )
{
  return(-1.0 / (1.0 + x * x));
}

/* ----------------------------------------------------------------------------
 * asec: This function returns the arcsecant of the argument.
 */
REAL asec ( REAL x )
{
  if (x == 0) { /* @@@ divide by zero error */ }
  return acos(1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1asec: This function returns the partial derivative of the asec function
 * with respect to its argument.
 */
REAL pd1asec ( REAL x )
{
  if ((x == 0.0) || ((x <= 1.0) && (x >= -1.0)))
  { /* @@@ divide by zero error */ }
  return(1.0 / x / sqrt(x * x - 1.0));
}

/* ----------------------------------------------------------------------------
 * acsc: This function returns the arccosecant of the argument.
 */
REAL acsc ( REAL x )
{
  if (x == 0) { /* @@@ divide by zero error */ }
  return asin(1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1acsc: This function returns the partial derivative of the acsc function
 * with respect to its argument.
 */
REAL pd1acsc ( REAL x )
{
  if ((x == 0.0) || ((x <= 1.0) && (x >= -1.0)))
  { /* @@@ divide by zero error */ }
  return(-1.0 / x / sqrt(x * x - 1.0));
}

/* ----------------------------------------------------------------------------
 * asinh: This function returns the hyperbolic arcsine of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1asin: This function returns the partial derivative of the asinh function
 * with respect to its argument.
 */
REAL pd1asinh ( REAL x )
{
  return(1.0 / sqrt(x * x + 1.0));
}

/* ----------------------------------------------------------------------------
 * acosh: This function returns the hyperbolic arccosine of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1acosh: This function returns the partial derivative of the acosh function
 * with respect to its argument.
 */
REAL pd1acosh ( REAL x )
{
  if ((x <= 1.0) && (x >= -1.0)) { /* @@@ divide by zero error */ }
  return(1.0 / sqrt(x * x - 1.0));
}

/* ----------------------------------------------------------------------------
 * atanh: This function returns the hyperbolic arctangent of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1atanh: This function returns the partial derivative of the atanh function
 * with respect to its argument.
 */
REAL pd1atanh ( REAL x )
{
  if ((x >= 1.0) || (x <= -1.0)) { /* @@@ divide by zero error */ }
  return(1.0 / (1.0 - x * x));
}

/* ----------------------------------------------------------------------------
 * acoth: This function returns the hyperbolic arccotangent of the argument.
 */
REAL acoth ( REAL x )
{
  if (x == 0) { /* @@@ divide by zero error */ }
  return(atanh(1.0 / x));
}


/* ----------------------------------------------------------------------------
 * pd1acoth: This function returns the partial derivative of the acoth function
 * with respect to its argument.
 */
REAL pd1acoth ( REAL x )
{
  if ((x <= 1.0) && (x >= -1.0)) { /* @@@ divide by zero error */ }
  return(-1.0 / (x * x - 1.0));
}

/* ----------------------------------------------------------------------------
 * asech: This function returns the hyperbolic arcsecant of the argument.
 */
REAL asech ( REAL x )
{
  if (x == 0) { /* @@@ divide by zero error */ }
  return acosh(1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1asech: This function returns the partial derivative of the asec function
 * with respect to its argument.
 */
REAL pd1asech ( REAL x )
{
  if ((x == 0.0) || (x >= 1.0) || (x <= -1.0))
  { /* @@@ divide by zero error */ }
  return(-1.0 / x / sqrt(1.0 - x * x));
}

/* ----------------------------------------------------------------------------
 * acsch: This function returns the hyperbolic arccosecant of the argument.
 */
REAL acsch ( REAL x )
{
  if (x == 0.0) { /* @@@ divide by zero error */ }
  return asinh(1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1acsch: This function returns the partial derivative of the acsc function
 * with respect to its argument.
 */
REAL pd1acsch ( REAL x )
{
  if (x == 0.0) { /* @@@ divide by zero error */ }
  return(-1.0 / x / sqrt(x * x + 1.0));
}

/* ----------------------------------------------------------------------------
 * NOTE:  the following function(s) are used to compute doping concentrations
 *	  and do not have derivative information.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * nsdep:  This function returns an approximate deposition profile of a step
 *	   implant driven in an inert environment.
 *
 *			  1	  W/2 + x	   W/2 - x
 *	  nsdep(x,W,Dt) = - (erf(---------) + erf(----------))
 *			  2	 2*sqrt(Dt)	  2*sqrt(Dt)
 */
REAL nsdep ( REAL x, REAL W, REAL Dt )
{
  REAL D  = 2.0 * sqrt(Dt);
  REAL Wh = W / 2.0;
  return 0.5 * (erf((Wh + x)/D) + erf((Wh - x)/D));
}


#if 0
/* ----------------------------------------------------------------------------
 * ngdep:  This function returns an approximate deposition profile for a
 *	   Gaussian implant driven in an inert environment.
 */
REAL ngdep ( REAL x, REAL Rp, REAL dRp, REAL Dt )
{
  REAL sqRp = Rp * Rp;
  REAL D1 = 2.0 * dRp * sqrt(Dt * (sqRp + 2 * Dt));
  REAL D2 = 4.0 * Dt + 2.0 * sqRp;
  REAL N1 = sqRp * x;
  REAL N2 = 2 * Dt * Rp;
  REAL C  = 2.0 * (1.0 + erf(N2/D1)) * exp(-sqRp/D2);
  return ((1.0 + erf((N1 + N2)/D1)) * exp(-sq(x - Rp)/D2) +
	  (1.0 - erf((N1 - N2)/D1)) * exp(-sq(x + Rp)/D2)) / C;
}
#endif


/* ----------------------------------------------------------------------------
 * ngdep:  This function returns an approximate Gaussian deposition.
 */
REAL ngdep ( REAL x, REAL y, REAL W, REAL ax, REAL ay )
{
  REAL xprime = fabs(x) - W/2.0;
  return ((xprime <= 0.0) ? 1.0 : exp(-ax*xprime*xprime))*
	 ((y	  >  0.0) ? 0.0 : exp(-ay*y*y));
}


///////////////////////////////////////////////////////////////////////////////
//////////////////////////////	FUNCTION TABLE	///////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#if defined(HPUX)

#define  _pow    power
#define  _ave    ave
#define  _fmin   fmin
#define  _fmax   fmax
#define  _exp    exp
#define  _log    log
#define  _pow10  pow10
#define  _log10  log10
#define  _sq     sq
#define  _sqrt   sqrt
#define  _inv    inv
#define  _sinh   sinh
#define  _cosh   cosh
#define  _tanh   tanh
#define  _coth   coth
#define  _sech   sech
#define  _csch   csch
#define  _sin    sin
#define  _cos    cos
#define  _tan    tan
#define  _cot    cot
#define  _sec    sec
#define  _csc    csc
#define  _erf    erf
#define  _fabs   fabs
#define  _sign   sign
#define  _step   step
#define  _bern   bern
#define  _aux1   aux1
#define  _aux2   aux2

#define  _asin   asin
#define  _acos   acos
#define  _atan   atan
#define  _acot   acot
#define  _asec   asec
#define  _acsc   acsc
#define  _asinh  asinh
#define  _acosh  acosh
#define  _atanh  atanh
#define  _acoth  acoth
#define  _asech  asech
#define  _acsch  acsch

#else

// the following variables are defined in order to disambiguate the following
// functions, some of which have been overloaded.
PRF2 _pow   = pow  , _ave   = ave  , _fmin  = fmin , _fmax  = fmax ;
PRF1 _exp   = exp  , _log   = log  , _pow10 = pow10, _log10 = log10;
PRF1 _sq    = sq   , _sqrt  = sqrt , _inv   = inv  , _sinh  = sinh ;
PRF1 _cosh  = cosh , _tanh  = tanh , _coth  = coth , _sech  = sech ;
PRF1 _csch  = csch , _sin   = sin  , _cos   = cos  , _tan   = tan  ;
PRF1 _cot   = cot  , _sec   = sec  , _csc   = csc  , _erf   = erf  ;
PRF1 _fabs  = fabs , _sign  = sign , _step  = step , _bern  = bern ;
PRF1 _aux1  = aux1 , _aux2  = aux2 ;

PRF1 _asin  = asin , _acos  = acos , _atan  = atan ;
PRF1 _acot  = acot , _asec  = asec , _acsc  = acsc ;
PRF1 _asinh = asinh, _acosh = acosh, _atanh = atanh;
PRF1 _acoth = acoth, _asech = asech, _acsch = acsch;

#endif

// Note: the order should follow the header file
IFT aift[] =
{ { "exp"  , 1, {(PRF) _exp  , (PRF) pd1exp  , NULL	    } },  // FUNC_EXP
  { "ln"   , 1, {(PRF) _log  , (PRF) pd1log  , NULL	    } },  // FUNC_LN
  { "pow10", 1, {(PRF) _pow10, (PRF) pd1pow10, NULL	    } },  // FUNC_POW10
  { "log"  , 1, {(PRF) _log10, (PRF) pd1log10, NULL	    } },  // FUNC_LOG
  { "pow"  , 2, {(PRF) _pow  , (PRF) pd1pow  , (PRF) pd2pow } },  // FUNC_POW
  { "sq"   , 1, {(PRF) _sq   , (PRF) pd1sq   , NULL	    } },  // FUNC_SQ
  { "sqrt" , 1, {(PRF) _sqrt , (PRF) pd1sqrt , NULL	    } },  // FUNC_SQRT
  { "inv"  , 1, {(PRF) _inv  , (PRF) pd1inv  , NULL	    } },  // FUNC_INV
  { "sinh" , 1, {(PRF) _sinh , (PRF) pd1sinh , NULL	    } },  // FUNC_SINH
  { "cosh" , 1, {(PRF) _cosh , (PRF) pd1cosh , NULL	    } },  // FUNC_COSH
  { "tanh" , 1, {(PRF) _tanh , (PRF) pd1tanh , NULL	    } },  // FUNC_TANH
  { "coth" , 1, {(PRF) _coth , (PRF) pd1coth , NULL	    } },  // FUNC_COTH
  { "sech" , 1, {(PRF) _sech , (PRF) pd1sech , NULL	    } },  // FUNC_SECH
  { "csch" , 1, {(PRF) _csch , (PRF) pd1csch , NULL	    } },  // FUNC_CSCH
  { "sin"  , 1, {(PRF) _sin  , (PRF) pd1sin  , NULL	    } },  // FUNC_SIN
  { "cos"  , 1, {(PRF) _cos  , (PRF) pd1cos  , NULL	    } },  // FUNC_COS
  { "tan"  , 1, {(PRF) _tan  , (PRF) pd1tan  , NULL	    } },  // FUNC_TAN
  { "cot"  , 1, {(PRF) _cot  , (PRF) pd1cot  , NULL	    } },  // FUNC_COT
  { "sec"  , 1, {(PRF) _sec  , (PRF) pd1sec  , NULL	    } },  // FUNC_SEC
  { "csc"  , 1, {(PRF) _csc  , (PRF) pd1csc  , NULL	    } },  // FUNC_CSC
  { "erf"  , 1, {(PRF) _erf  , (PRF) pd1erf  , NULL	    } },  // FUNC_ERF
  { "ave"  , 2, {(PRF) _ave  , (PRF) pd1ave  , (PRF) pd2ave } },  // FUNC_AVE
  { "abs"  , 1, {(PRF) _fabs , (PRF) pd1abs  , NULL	    } },  // FUNC_ABS
  { "sign" , 1, {(PRF) _sign , (PRF) pd1sign , NULL	    } },  // FUNC_SIGN
  { "step" , 1, {(PRF) _step , (PRF) pd1step , NULL	    } },  // FUNC_STEP
  { "min"  , 2, {(PRF) _fmin , (PRF) pd1min  , (PRF) pd2min } },  // FUNC_MIN
  { "max"  , 2, {(PRF) _fmax , (PRF) pd1max  , (PRF) pd2max } },  // FUNC_MAX
  { "bern" , 1, {(PRF) _bern , (PRF) pd1bern , NULL	    } },  // FUNC_BERN
  { "aux1" , 1, {(PRF) _aux1 , (PRF) pd1aux1 , NULL	    } },  // FUNC_AUX1
  { "aux2" , 1, {(PRF) _aux2 , (PRF) pd1aux2 , NULL	    } },  // FUNC_AUX2
#if defined(NFUNC)
  { "nsdep", 3, {(PRF) nsdep , NULL	     , NULL	    } },  // FUNC_NDEP
  { "ngdep", 5, {(PRF) ngdep , NULL	     , NULL	    } },  // FUNC_NDEP
#else
  { NULL   , 0, {(PRF) NULL  , NULL	     , NULL	    } },  // FUNC_NDEP
  { NULL   , 0, {(PRF) NULL  , NULL	     , NULL	    } },  // FUNC_NDEP
#endif
  // by JJF
  { "asin" , 1, {(PRF) _asin , (PRF) pd1asin , NULL	    } },  // FUNC_ASIN
  { "acos" , 1, {(PRF) _acos , (PRF) pd1acos , NULL	    } },  // FUNC_ACOS
  { "atan" , 1, {(PRF) _atan , (PRF) pd1atan , NULL	    } },  // FUNC_ATAN
  { "acot" , 1, {(PRF) _acot , (PRF) pd1acot , NULL	    } },  // FUNC_ACOT
  { "asec" , 1, {(PRF) _asec , (PRF) pd1asec , NULL	    } },  // FUNC_ASEC
  { "acsc" , 1, {(PRF) _acsc , (PRF) pd1acsc , NULL	    } },  // FUNC_ACSC
  { "asinh", 1, {(PRF) _asinh, (PRF) pd1asinh, NULL	    } },  // FUNC_ASINH
  { "acosh", 1, {(PRF) _acosh, (PRF) pd1acosh, NULL	    } },  // FUNC_ACOSH
  { "atanh", 1, {(PRF) _atanh, (PRF) pd1atanh, NULL	    } },  // FUNC_ATANH
  { "acoth", 1, {(PRF) _acoth, (PRF) pd1acoth, NULL	    } },  // FUNC_ACOTH
  { "asech", 1, {(PRF) _asech, (PRF) pd1asech, NULL	    } },  // FUNC_ASECH
  { "acsch", 1, {(PRF) _acsch, (PRF) pd1acsch, NULL	    } },  // FUNC_ACSCH
};


const UINT cFuncs = sizeof(aift) / sizeof(IFT);
