/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  SGFramework C++ Code Generation Library
 FILE:	     sgnam.cpp
 AUTHOR:     Kevin M. Kramer
             J. Joshua Feng

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 09/15/94    KMK      initial implementation
 06/22/00    JJF      asin, .. etc

 DESCRIPTION:

 This module contains the functions needed by the code generated by the
 SG framework.

-----------------------------------------------------------------------------*/


#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>

#if defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#define	 O_BINARY 0x0000
#else
#include <io.h>
#include <sys\stat.h>
#endif

#include "sgnam.h"

Vector *pvectX;

static char szMeshFile[LEN_FILENAME];


///////////////////////////////////////////////////////////////////////////////
//////////////////////////  DEBUGGING INFORMATION  ////////////////////////////
///////////////////////////////////////////////////////////////////////////////


#if defined(MEM)
long MEMAllocBytes = 0;
#endif

#if defined(MEM)
#define MEM_ALLOC(info,address,num,bytes)			\
  MEMAllocBytes += (long) (num) * (long) (bytes);		\
  fprintf(stdout, "%10ld | %20s | %p | %ld * %ld = %ld\n",	\
    MEMAllocBytes, info, address, (long) (num), (long) (bytes), \
    (long) (num) * (long) (bytes));
#else
#define MEM_ALLOC(info,address,num,bytes)
#endif


///////////////////////////////////////////////////////////////////////////////
///////////////////////////  ERROR HANDLER CLASS  /////////////////////////////
///////////////////////////////////////////////////////////////////////////////


ErrorHandler *error;

#define MSG_WARNING   0
#define MSG_ERROR     1
#define MSG_FATAL     2

typedef struct tagEHM			// ehm
{ UINT uType;				// type of message
  char *pszMsg;				// message text
} EHM;

EHM aehm[] =
{ { MSG_FATAL,	  "too many errors or warnings\n" },
  { MSG_FATAL,	  "cannot open file \"%s\"\n" },
  { MSG_FATAL,	  "cannot find label \"%s\" in data file \"%s\"\n" },
  { MSG_FATAL,	  "syntax error in data file \"%s\"\n" },
  { MSG_FATAL,	  "unexpected end of file in data file \"%s\"\n" },
  { MSG_FATAL,	  "near singularity encountered in %s\n" },
  { MSG_FATAL,	  "Argument was not in domain of function, such as log(-1)." },
  { MSG_FATAL,	  "Argument would result in a singularity, such as pow(0, -2)." },
  { MSG_FATAL,	  "Argument would produce a function result greater than DBL_MAX (or LDBL_MAX), such as exp(1000)." },
  { MSG_FATAL,	  "Argument would produce a function result less than DBL_MIN (or LDBL_MIN), such as exp(-1000)." },
  { MSG_FATAL,	  "Argument would produce function result with total loss of significant digits, such as sin(10e70)." }
};


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Constructor
//
ErrorHandler::ErrorHandler ( UINT cErrors, UINT cWarnings )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ErrorHandler::ErrorHandler\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  cMaxErrors = cErrors;
  cMaxWarnings = cWarnings;
  cErrors = 0;
  cWarnings = 0;
  cFatal = 0;
  cTotal = 0;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Private
// Description:	 This function updates the private data members of the error
//		 class.
//
void ErrorHandler::UpdateVariables ( UINT uType )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ErrorHandler::UpdateVariables\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  ++cTotal;
  switch (uType)
  { case MSG_WARNING:  ++cWarnings;  break;
    case MSG_ERROR:    ++cErrors;    break;
    case MSG_FATAL:    ++cFatal;     break;
  }

  if ((cErrors >= cMaxErrors) || (cWarnings >= cMaxWarnings))
  { fprintf(stdout, aehm[EHM_MESSAGES].pszMsg);
    exit(1);
  }

  if (uType == MSG_FATAL)
    exit(1);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function adds (reports) the specified message.
//
void ErrorHandler::AddMessage ( UINT iMsg )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ErrorHandler::AddMessage\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz  = aehm[iMsg].pszMsg;
  UINT uType = aehm[iMsg].uType;
  fprintf(stdout, psz);
  UpdateVariables (uType);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function adds (reports) the specified message.
//
void ErrorHandler::AddMessage ( UINT iMsg, char *param1 )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ErrorHandler::AddMessage\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz  = aehm[iMsg].pszMsg;
  UINT uType = aehm[iMsg].uType;
  fprintf(stdout, psz, param1);
  UpdateVariables (uType);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function adds (reports) the specified message.
//
void ErrorHandler::AddMessage ( UINT iMsg, char *param1, char *param2 )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ErrorHandler::AddMessage\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz  = aehm[iMsg].pszMsg;
  UINT uType = aehm[iMsg].uType;
  fprintf(stdout, psz, param1, param2);
  UpdateVariables (uType);
}


///////////////////////////////////////////////////////////////////////////////
///////////////	 FILE I/O CONSTANTS, VARIABLES AND FUNCTIONS  /////////////////
///////////////////////////////////////////////////////////////////////////////


// file i/o constants
#define LEN_READBUF		1024	// length of read buffer
#define LEN_MAXNUM		127	// maximum number length
#define FILE_SUCCESS		0	// successful operation
#define FILE_FINDERROR	       -1	// cannot find label in data file
#define FILE_ENDOFFILE	       -2	// unexpected end of data file
#define FILE_SYNTAX	       -3	// syntax error in data file
#define GOTOSTATE_SEARCH	0	// goto state
#define GOTOSTATE_COMM1		1	// goto state
#define GOTOSTATE_COMM2		2	// goto state
#define GOTOSTATE_LABEL1	3	// goto state
#define GOTOSTATE_LABEL2	4	// goto state
#define GOTOSTATE_FOUND		5	// goto state
#define READSTATE_SEARCH	0	// read state
#define READSTATE_COMM1		1	// read state
#define READSTATE_COMM2		2	// read state
#define READSTATE_LABEL		3	// read state
#define READSTATE_NUM1		4	// read state
#define READSTATE_NUM2		5	// read state
#define READSTATE_NUM3		6	// read state
#define READSTATE_NUM4		7	// read state
#define READSTATE_NUM5		8	// read state
#define READSTATE_FOUND		9	// read state

// file i/o macros
#define ADDCHAR(ch) { if (i < LEN_MAXNUM) sz[i++] = ch; else return FILE_SYNTAX; }

// file i/o variables and arrays
UINT iReadBuf;
char szReadBuf[LEN_READBUF];
char szFileName[LEN_FILENAME];
FILE *file = NULL;


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function loads the read buffer
//
int LoadReadBuffer ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp LoadReadBuffer\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  iReadBuf = 0;
  UINT nResult = fread(szReadBuf, sizeof(char), LEN_READBUF, file);
  if (nResult < LEN_READBUF) szReadBuf[nResult] = '\0';
  return FILE_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function searches for the specified label
//
int GotoLabel ( char *pszLabel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp GotoLabel\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT uState = GOTOSTATE_SEARCH;
  char ch, *psz;

  while (uState != GOTOSTATE_FOUND)
  { if ((ch = szReadBuf[iReadBuf]) == '\0') return FILE_FINDERROR;
    switch (uState)
    { case GOTOSTATE_SEARCH:
	if	(ch == '<') { uState = GOTOSTATE_LABEL1; psz = pszLabel; }
	else if (ch == '/')   uState = GOTOSTATE_COMM1;
	break;

      case GOTOSTATE_COMM1:
	uState = (ch == '/') ? GOTOSTATE_COMM2 : GOTOSTATE_SEARCH;
	break;

      case GOTOSTATE_COMM2:
	if (ch == '\n') uState = GOTOSTATE_SEARCH;
	break;

      case GOTOSTATE_LABEL1:
	if (ch == '\n') return FILE_SYNTAX;
	else if (ch == '>') uState = (*psz) ? GOTOSTATE_SEARCH : GOTOSTATE_FOUND;
	else if (ch == *psz) ++psz;
	else uState = GOTOSTATE_LABEL2;
	break;

      case GOTOSTATE_LABEL2:
	if (ch == '>') uState = GOTOSTATE_SEARCH;
	break;
    }

    // update the buffer index
    ++iReadBuf;
    if (iReadBuf == LEN_READBUF) LoadReadBuffer();
  }

  return FILE_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function searches for the specified label
//
int ReadNumber ( REAL &r )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ReadNumber\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT uState = READSTATE_SEARCH;
  UINT i = 0;
  char ch, sz[LEN_MAXNUM + 1];

  while (uState != READSTATE_FOUND)
  { if ((ch = szReadBuf[iReadBuf]) == '\0') return FILE_ENDOFFILE;
    switch (uState)
    { case READSTATE_SEARCH:
	if	(ch == '<')	uState = READSTATE_LABEL;
	else if (ch == '/')	uState = READSTATE_COMM1;
	else if (ch == '+')	uState = READSTATE_NUM1;
	else if (ch == '-')   { uState = READSTATE_NUM1;   sz[i++] = ch; }
	else if (isdigit(ch)) { uState = READSTATE_NUM2;   sz[i++] = ch; }
	else if (isspace(ch))	uState = READSTATE_SEARCH;
	else return FILE_SYNTAX;
	break;

      case READSTATE_COMM1:
	uState = (ch == '/') ? READSTATE_COMM2 : READSTATE_SEARCH;
	break;

      case READSTATE_COMM2:
	if (ch == '\n') uState = READSTATE_SEARCH;
	break;

      case READSTATE_LABEL:
	if	(ch == '>')  uState = READSTATE_SEARCH;
	else if (ch == '\n') return FILE_SYNTAX;
	break;

      case READSTATE_NUM1:
	if (isdigit(ch)) { uState = READSTATE_NUM2;  ADDCHAR(ch); }
	else return FILE_SYNTAX;
	break;

      case READSTATE_NUM2:
	if	(isspace(ch)) { uState = READSTATE_FOUND;  sz[i++] = '\0'; }
	else if (isdigit(ch)) { uState = READSTATE_NUM2;   ADDCHAR(ch); }
	else if (ch == '.')   { uState = READSTATE_NUM3;   ADDCHAR(ch); }
	else if (ch == 'e')   { uState = READSTATE_NUM4;   ADDCHAR(ch); }
	else if (ch == 'E')   { uState = READSTATE_NUM4;   ADDCHAR(ch); }
	else return FILE_SYNTAX;
	break;

      case READSTATE_NUM3:
	if	(isspace(ch)) { uState = READSTATE_FOUND;  sz[i++] = '\0'; }
	else if (isdigit(ch)) { uState = READSTATE_NUM3;   ADDCHAR(ch); }
	else if (ch == 'e')   { uState = READSTATE_NUM4;   ADDCHAR(ch); }
	else if (ch == 'E')   { uState = READSTATE_NUM4;   ADDCHAR(ch); }
	else return FILE_SYNTAX;
	break;

      case READSTATE_NUM4:
	if	(isspace(ch)) { uState = READSTATE_FOUND;  sz[i++] = '\0'; }
	else if (isdigit(ch)) { uState = READSTATE_NUM5;   ADDCHAR(ch); }
	else if (ch == '-')   { uState = READSTATE_NUM5;   ADDCHAR(ch); }
	else if (ch == '+')   { uState = READSTATE_NUM5;   ADDCHAR(ch); }
	else return FILE_SYNTAX;
	break;

      case READSTATE_NUM5:
	if	(isspace(ch)) { uState = READSTATE_FOUND;  sz[i++] = '\0'; }
	else if (isdigit(ch)) { uState = READSTATE_NUM5;   ADDCHAR(ch); }
	else return FILE_SYNTAX;
	break;
    }

    // update the buffer index
    ++iReadBuf;
    if (iReadBuf == LEN_READBUF) LoadReadBuffer();
  }

  r = atof(sz);
  return FILE_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function opens the specified text file with the specified
//		 file i/o mode
//
int FileIO_Open ( char *pszFileName, UINT uMode )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp FileIO_Open\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char szMode[] = "?t";
  switch (uMode)
  { case IOMODE_READ:	 *szMode = 'r';	 break;
    case IOMODE_WRITE:	 *szMode = 'w';	 break;
    case IOMODE_APPEND:	 *szMode = 'a';	 break;
  }

  strncpy(szFileName, pszFileName, LEN_FILENAME);
  if ((file = fopen(pszFileName, szMode)) == NULL)
  { error->AddMessage(EHM_OPENFILE, pszFileName);
    return STATUS_FAILURE;
  }

  if (uMode == IOMODE_READ) LoadReadBuffer();

  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function closes the open file
//
int FileIO_Close ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp FileIO_Close\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (file) fclose(file);
  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function finds the specified label in the data file.
//
int FileIO_Goto ( char *pszLabel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp FileIO_Goto\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  int nResult = GotoLabel(pszLabel);
  if      (nResult == FILE_FINDERROR) { error->AddMessage(EHM_FINDLABEL, pszLabel, szFileName); }
  else if (nResult == FILE_SYNTAX)    { error->AddMessage(EHM_SYNTAX, szFileName); }
  return (nResult == FILE_SUCCESS) ? STATUS_SUCCESS : STATUS_FAILURE;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function writes the specified label to the data file.
//
int FileIO_Label ( char *pszLabel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp FileIO_Label\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  fprintf(file, "<%s>\n", pszLabel);
  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function inserts a comment in the data file.
//
int FileIO_Comment ( char *pszLabel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp FileIO_Comment\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  fprintf(file, "// %s\n", pszLabel);
  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function reads a variable or array from the data file.
//
int FileIO_Read ( UINT i, BOOL fArray, XLATVAR *avar, XLATARRAY *aarray, REAL *ar )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp FileIO_Read\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  BOOL fError = FALSE;
  int nResult = FILE_SUCCESS;
  UINT j, k, l, c, c1, c2, c3, u;

  if (fArray)
  { c  = aarray[i].cDim;
    c1 = aarray[i].acElements[0];
    c2 = aarray[i].acElements[1];
    c3 = aarray[i].acElements[2];
    u  = aarray[i].uOffset;
  }
  else
  { c = 0;
    u = avar[i].uOffset;
  }

  // read variable
  if (c == 0) nResult = ReadNumber(ar[u]);

  else
  { //	read a 1D array
    if (c == 1)
    { for(j = 0; !fError && (j < c1); ++j)
      { nResult = ReadNumber(ar[u + j]);
	fError = nResult != FILE_SUCCESS;
      }
    }

    //	write a 2D array
    else if (c == 2)
    { for(j = 0; !fError && (j < c1); ++j)
      { for(k = 0; !fError && (k < c2); ++k)
	{ nResult = ReadNumber(ar[u + j * c2 + k]);
	  fError = nResult != FILE_SUCCESS;
	}
      }
    }

    //	write a 3D array
    else if (c == 3)
    { for(j = 0; !fError && (j < c1); ++j)
      { for(k = 0; !fError && (k < c2); ++k)
	{ for(l = 0; !fError && (l < c3); ++l)
	  { nResult = ReadNumber(ar[u + (j * c2 + k) * c3 + l]);
	    fError = nResult != FILE_SUCCESS;
	  }
	}
      }
    }
  }

  // report error if neccessary and return
  if	  (nResult == FILE_ENDOFFILE) { error->AddMessage(EHM_ENDOFFILE, szFileName); }
  else if (nResult == FILE_SYNTAX)    { error->AddMessage(EHM_SYNTAX, szFileName);    }
  return  (nResult == FILE_SUCCESS) ? STATUS_SUCCESS : STATUS_FAILURE;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function writes a variable or array to the data file.
//
int FileIO_Write ( UINT i, BOOL fArray, XLATVAR *avar, XLATARRAY *aarray, REAL *ar )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp FileIO_Write\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

#if   defined(SPREC)
  char pszF[] = "%.*g";
#elif defined(LPREC)
  char pszF[] = "%.*Lg";
#else
  char pszF[] = "%.*lg";
#endif

  char pszS[] = " ";

  UINT j, k, l, c, c1, c2, c3, u;

  if (fArray)
  { c  = aarray[i].cDim;
    c1 = aarray[i].acElements[0];
    c2 = aarray[i].acElements[1];
    c3 = aarray[i].acElements[2];
    u  = aarray[i].uOffset;
  }
  else
  { c = 0;
    u = avar[i].uOffset;
  }

  // write variable
  if (c == 0)
  { fprintf(file, pszF, cmdargs.uPrecision, ar[u]);
    fprintf(file, "\n");
  }

  else
  {	  //  write a 1D array
    if (c == 1)
    { for(j = 0; j < c1; ++j)
      { fprintf(file, pszF, cmdargs.uPrecision, ar[u + j]);
	fprintf(file, "\n");
      }
    }

    //	write a 2D array
    else if (c == 2)
    { for(j = 0; j < c1; ++j)
      { for(k = 0; k < c2; ++k)
	{ fprintf(file, pszF, cmdargs.uPrecision, ar[u + j * c2 + k]);
	  fprintf(file, (k < c2 - 1) ? pszS : "\n");
	}
      }
    }

    //	write a 3D array
    else if (c == 3)
    { for(j = 0; j < c1; ++j)
      { for(k = 0; k < c2; ++k)
	{ for(l = 0; l < c3; ++l)
	  { fprintf(file, pszF, cmdargs.uPrecision, ar[u + (j * c2 + k) * c3 + l]);
	    fprintf(file, (l < c3 - 1) ? pszS : "\n");
	  }
	}
	fprintf(file, "\n");
      }
    }
  }

  // insert a linefeed and return
  fprintf(file, "\n");
  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
/////////////////////////  MATHEMATICAL FUNCTIONS  ////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#if 0
int _matherr (struct exception *a)
{
  switch (a->type)
  { case DOMAIN:      error->AddMessage(EHM_DOMAIN);	     break;
    case SING:	      error->AddMessage(EHM_SINGULARITY);    break;
    case OVERFLOW:    error->AddMessage(EHM_OVERFLOW);	     break;
    case UNDERFLOW:   error->AddMessage(EHM_UNDERFLOW);	     break;
    case TLOSS:	      error->AddMessage(EHM_TOTALLOSS);	     break;
  }
  return 1;
}
#endif

/* ----------------------------------------------------------------------------
 * exp:	 This function returns the e to the argument and resides in the C math
 * library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1exp:  This function returns the partial derivative of the exp function
 * with respect to its argument.
 */
REAL pd1exp ( REAL x )
{
  return exp(x);
}


/* ----------------------------------------------------------------------------
 * log:	 This function returns the natural logarithm of the argument and is
 * part of the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1log:  This function returns the partial derivative of the ln function
 * with respect to its argument.
 */
REAL pd1log ( REAL x )
{
  return 1.0 / x;
}


/* ----------------------------------------------------------------------------
 * pow10:  This function returns 10 to the x and resides in the C math library
 * on some systems ...
 * --------------------------------------------------------------------------*/
REAL pow10 ( REAL x )
{
  return pow(10.0, x);
}


/* ----------------------------------------------------------------------------
 * pd1pow10:  This function returns the partial derivative of the pow10
 * function with respect to its argument.
 */
REAL pd1pow10 ( REAL x )
{
  return LOGe10 * pow10(x);
}


/* ----------------------------------------------------------------------------
 * log10:  This function returns the logarithm base 10 of the argument and is
 * part of the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1log10:  This function returns the partial derivative of the log10
 * function with respect to its argument.
 */
REAL pd1log10 ( REAL x )
{
  return LOG10e / x;
}


/* ----------------------------------------------------------------------------
 * pow:	 This function returns x to the y and resides in the C math library
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1pow:  This function returns the partial derivative of the pow function
 * with respect to its first argument, i.e. 'x'.
 */
REAL pd1pow ( REAL x, REAL y )
{
  return(y * pow(x, y - 1.0));
}


/* ----------------------------------------------------------------------------
 * pd2pow:  This function returns the partial derivative of the pow function
 * with respect to its second argument, i.e. 'y'.
 */
REAL pd2pow ( REAL x, REAL y )
{
  return(pow(x, y) * log(x));
}


/* ----------------------------------------------------------------------------
 * sq:	This function returns the square of the argument.
 */
REAL sq ( REAL x )
{
  return(x * x);
}


/* ----------------------------------------------------------------------------
 * pd1sq:  This function returns the partial derivative of the square function
 * with respect to its argument.
 */
REAL pd1sq ( REAL x )
{
  return(2.0 * x);
}


/* ----------------------------------------------------------------------------
 * sqrt:  This function returns the square root of the argument and resides in
 * the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1sqrt:  This function returns the partial derivative of the square root
 * function with respect to its argument.
 */
REAL pd1sqrt ( REAL x )
{
  return(0.5 / sqrt(x));
}


/* ----------------------------------------------------------------------------
 * inv:	 This function returns the inverse of the argument.
 */
REAL inv ( REAL x )
{
  return (1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1inv:  This function returns the partial derivative of the inverse
 * function with respect to its argument
 */
REAL pd1inv ( REAL x )
{
  return(1.0 / sq(x));
}


/* ----------------------------------------------------------------------------
 * sinh:  This function returns the hyperbolic sine of the argument and resides
 * in the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1sinh:  This function returns the partial derivative of the sinh function
 * with respect to its argument.
 */
REAL pd1sinh ( REAL x )
{
  return(cosh(x));
}


/* ----------------------------------------------------------------------------
 * cosh:  This function returns the hyperbolic cosine of the argument and
 * resides in the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1cosh:  This function returns the partial derivative of the cosh function
 * with respect to its argument.
 */
REAL pd1cosh ( REAL x )
{
  return(sinh(x));
}


/* ----------------------------------------------------------------------------
 * tanh:  This function returns the hyperbolic tangent of the argument and
 * resides in the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1tanh:  This function returns the partial derivative of the tanh function
 * with respect to its argument.
 */
REAL pd1tanh ( REAL x )
{
  REAL y = cosh(x);
  return(sq(1.0 / y));
}


/* ----------------------------------------------------------------------------
 * coth:  This function returns the hyperbolic cotangent of the argument
 */
REAL coth ( REAL x )
{
  REAL y = tanh(x);
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1coth:  This function returns the partial derivative of the coth function
 * with respect to its argument.
 */
REAL pd1coth ( REAL x )
{
  REAL y = sinh(x);
  return(-1.0 / sq(y));
}


/* ----------------------------------------------------------------------------
 * sech:  This function returns the hyperbolic secant of the argument
 */
REAL sech ( REAL x )
{
  REAL y = cosh(x);
  if (y == 0.0) { /* @@@ division by zero error */ }
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1sech:  This function returns the partial derivative of the sech function
 * with respect to its argument
 */
REAL pd1sech ( REAL x )
{
  REAL y = cosh(x);
  return(-sinh(x) / sq(y));
}


/* ----------------------------------------------------------------------------
 * csch:  This function returns the hyperbolic cosecant of the argument.
 */
REAL csch ( REAL x )
{
  REAL y = sinh(x);
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1csch:  This function returns the partial derivative of the csch function
 * with respect to its argument.
 */
REAL pd1csch ( REAL x )
{
  REAL y = sinh(x);
  return(-cosh(x) / sq(y));
}


/* ----------------------------------------------------------------------------
 * sin:	 This function returns the sine of the argument and resides in the C
 * math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1sin:  This function returns the partial derivative of the sin function
 * with respect to its argument.
 */
REAL pd1sin ( REAL x )
{
  return(cos(x));
}


/* ----------------------------------------------------------------------------
 * cos:	 This function returns the cosine of the argument and resides in the C
 * math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1cos:  This function returns the partial derivative of the cos function
 * with respect to its argument.
 */
REAL pd1cos ( REAL x )
{
  return(-sin(x));
}


/* ----------------------------------------------------------------------------
 * tan:	 This function returns the tangent of the argument and resides in the C
 * math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1tan:  This function returns the partial derivative of the tan function
 * with respect to its argument.
 */
REAL pd1tan ( REAL x )
{
  REAL y = cos(x);
  return(1.0 / sq(y));
}


/* ----------------------------------------------------------------------------
 * cot:	 This function returns the cotangent of the argument
 */
REAL cot ( REAL x )
{
  REAL y = tan(x);
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1cot:  This function returns the partial derivative of the cot function
 * with respect to its argument.
 */
REAL pd1cot ( REAL x )
{
  REAL y = sin(x);
  return(-1.0 / sq(y));
}


/* ----------------------------------------------------------------------------
 * sec:	 This function returns the secant of the argument.
 */
REAL sec ( REAL x )
{
  REAL y = cos(x);
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1sec:  This function returns the partial derivative of the sec function
 * with respect to its argument.
 */
REAL pd1sec ( REAL x )
{
  REAL y = cos(x);
  return(sin(x) / sq(y));
}


/* ----------------------------------------------------------------------------
 * csc:	 This function returns the cosecant of the argument.
 */
REAL csc ( REAL x )
{
  REAL y = sin(x);
  return(1.0 / y);
}


/* ----------------------------------------------------------------------------
 * pd1csc:  This function returns the partial derivative of the csc function
 * with respect to its argument.
 */
REAL pd1csc ( REAL x )
{
  REAL y = sin(x);
  return(-cos(x) / sq(y));
}


/* ----------------------------------------------------------------------------
 * ave:	 This allows returns the average of the arguments.
 */
REAL ave ( REAL x, REAL y )
{
  return(0.5 * (x + y));
}


/* ----------------------------------------------------------------------------
 * pd1ave:  This function returns the partial derivative of the average
 * function with respect to its first argument, i.e. 'x'.
 */
REAL pd1ave ( REAL x, REAL y )
{
  return(0.5);
}


/* ----------------------------------------------------------------------------
 * pd2ave:  This function returns the partial derivative of the average
 * function with respect to its second argument, i.e. 'y'.
 */
REAL pd2ave ( REAL x, REAL y )
{
  return(0.5);
}


/* ----------------------------------------------------------------------------
 * erf:	 This function returns the error function.
 */
REAL erf ( REAL x )
{
  REAL t1 = 1.0 / (1.0 + 0.3275911 * fabs(x));
  REAL t2 = t1 * t1;
  REAL t3 = t2 * t1;
  REAL t4 = t3 * t1;
  REAL t5 = t4 * t1;
  REAL result = 1.0 - (0.254829592*t1 - 0.284496736*t2 + 1.421413741*t3 -
		       1.453152027*t4 + 1.061405429*t5) * exp(-x*x);
  return (x < 0.0) ? -result : result;
}


/* ----------------------------------------------------------------------------
 * pd1erf:  This function returns the derivative of the error function with
 * respect to the first variable.
 */
REAL pd1erf ( REAL x )
{
  return 2.0 / sqrt(PI) * exp(-x*x);
}


/*****************************************************************************/
/**********  THE FOLLOWING FUNCTIONS HAVE DISCONTINUOUS DERIVATIVES  *********/
/*****************************************************************************/


/* ----------------------------------------------------------------------------
 * abs:	 This function returns the absolute value of the argument and is
 * defined as a macro which expands to the fabs function in the C math library.
 * --------------------------------------------------------------------------*/


/* ----------------------------------------------------------------------------
 * pd1abs:  This function returns the partial derivative of the abs function
 * with respect to its argument.
 */
REAL pd1abs ( REAL x )
{
  return((x >= 0.0) ? 1.0 : -1.0);
}


/* ----------------------------------------------------------------------------
 * sign:  This functions returns the sign of the argument, i.e., +1 if the
 * argument is postive, 0 if the argument is zero, and -1 is the argument is
 * negative
 */
REAL sign ( REAL x )
{
  /* CAUTION:  this function has a discontinous derivative at zero */
  if	  (x == 0.0) return( 0.0);
  else if (x  > 0.0) return( 1.0);
		     return(-1.0);
}


/* ----------------------------------------------------------------------------
 * pd1sign:  This function returns the partial derivative of the sign function
 * with respect to its argument
 */
REAL pd1sign ( REAL x )
{
  /* CAUTION:  this function is undefined at zero */
  return(0.0);
}


/* ----------------------------------------------------------------------------
 * step:  This function returns the one if the argument is zero or greater and
 * zero if the argument is less than zero.
 */
REAL step ( REAL x )
{
  /* CAUTION:  this function has a discontinous derivative at zero */
  return((x < 0.0) ? 0.0 : 1.0);
}


/* ----------------------------------------------------------------------------
 * pd1step:  This function returns the partial derivative of the step function
 * with respect to its argument.
 */
REAL pd1step ( REAL x )
{
  /* CAUTION:  this function is undefined at zero */
  return(0.0);
}


/* ----------------------------------------------------------------------------
 * fmin:  This function returns the smallest argument.
 */
REAL fmin ( REAL x, REAL y )
{
  /* CAUTION:  this function has a discontinous derivative at x = y */
  return((x < y) ? x : y);
}


/* ----------------------------------------------------------------------------
 * pd1min:  This function returns the partial derivative of the fmin function
 * with respect to its first argument, i.e. 'x'.
 */
REAL pd1min ( REAL x, REAL y )
{
  /* CAUTION:  this function is undefined at x = y */
  return((x < y) ? 1.0 : 0.0);
}


/* ----------------------------------------------------------------------------
 * pd2min:  This function returns the partial derivative of the fmin function
 * with respect to its second argument, i.e. 'y'.
 */
REAL pd2min ( REAL x, REAL y )
{
  /* CAUTION:  this function is undefined at x = y */
  return((y < x) ? 1.0 : 0.0);
}


/* ----------------------------------------------------------------------------
 * max:	 This function returns the largest argument.
 */
REAL fmax ( REAL x, REAL y )
{
  /* CAUTION:  this function has a discontinous derivative at x = y */
  return((x > y) ? x : y);
}


/* ----------------------------------------------------------------------------
 * pd1max:  This function returns the partial derivative of the fmax function
 * with respect to its first argument, i.e. 'x'.
 */
REAL pd1max ( REAL x, REAL y )
{
  /* CAUTION:  this function is undefined at x = y */
  return((x > y) ? 1.0 : 0.0);
}


/* ----------------------------------------------------------------------------
 * pd2max:  This function returns the partial derivative of the fmax function
 * with respect to its second argument, i.e. 'y'.
 */
REAL pd2max ( REAL x, REAL y )
{
  /* CAUTION:  this function is undefined at x = y */
  return((y > x) ? 1.0 : 0.0);
}


/* ----------------------------------------------------------------------------
 * NOTE:  the range limits for the Bernoulli and Aux functions were generated
 *	  by the "brkpnts" program and stored in brkpnts.h.
 * --------------------------------------------------------------------------*/

// function break points ... these should be computed for the target platform,
// but in all probability, the minor differences between platforms in negligible
#if defined(UNIX)
#define BP0_BERN    -3.742994775023696e+01
#define BP1_BERN    -2.019464984136394e-02
#define BP2_BERN    2.449889303055096e-02
#define BP3_BERN    3.742994775023696e+01
#define BP4_BERN    7.083964185322633e+02
#define BP0_DBERN   -4.117169706049845e+01
#define BP1_DBERN   -3.742994775023696e+01
#define BP2_DBERN   -1.440190569485648e-02
#define BP3_DBERN   1.934659498020159e-02
#define BP4_DBERN   3.742994775023696e+01
#define BP5_DBERN   7.083964185322633e+02
#define BP0_AUX1    -8.312101842651140e-03
#define BP1_AUX1    8.312101842651140e-03
#define BP0_DAUX1   -3.605144451539276e-03
#define BP1_DAUX1   3.605144451539276e-03
#define BP0_AUX2    -3.673680056967704e+01
#define BP1_AUX2    3.682929080649183e+01
#define BP2_AUX2    7.083964185322644e+02
#define BP0_DAUX2   -7.083964185322644e+02
#define BP1_DAUX2   -3.673680056967704e+01
#define BP2_DAUX2   3.682929080649183e+01
#define BP3_DAUX2   7.083964185322644e+02
#define BP0_MISC    7.097827128183643e+02
#else
#define BP0_BERN    -4.505456673639642e+01
#define BP0_BERN    -4.505456673639642e+01
#define BP1_BERN    -1.559687716094362e-02
#define BP2_BERN    1.559024791109722e-02
#define BP3_BERN    4.505456673639642e+01
#define BP4_BERN    7.083959960937493e+02
#define BP0_DBERN   -4.896590766251155e+01
#define BP1_DBERN   -4.505456673639642e+01
#define BP2_DBERN   -5.722335193795038e-03
#define BP3_DBERN   6.219912505576786e-03
#define BP4_DBERN   4.505456673639642e+01
#define BP5_DBERN   7.083959960937493e+02
#define BP0_AUX1    -8.178444414517060e-03
#define BP1_AUX1    8.178444414517060e-03
#define BP0_DAUX1   -2.412257547715118e-03
#define BP1_DAUX1   2.412257547715118e-03
#define BP0_AUX2    -4.436141955583643e+01
#define BP1_AUX2    4.491903291904089e+01
#define BP2_AUX2    7.083959960937507e+02
#define BP0_DAUX2   -7.083959960937507e+02
#define BP1_DAUX2   -4.436141955583643e+01
#define BP2_DAUX2   4.491903291904089e+01
#define BP3_DAUX2   7.083959960937507e+02
#define BP0_MISC    7.097827128183643e+02
#endif

/* ----------------------------------------------------------------------------
 * bern:  This function returns the Bernoulli function of the argument.	 To
 * avoid under and overflows this function is defined by equivalent or approx-
 * imate functions depending upon the value of the argument.
 *
 *					  x
 *				B(x) = -------
 *				       e^x - 1
 *
 */
REAL bern ( REAL x )
{
  REAL y;

  if (x <= BP0_BERN)
  { return(-x); }
  else if (x <	BP1_BERN)
  { return(x / (exp(x) - 1.0)); }
  else if (x <= BP2_BERN)
  { return(1.0 - x/2.0 * (1.0 - x/6.0 * (1.0 - x*x/60.0))); }
  else if (x <	BP3_BERN)
  { y = exp(-x);   return((x * y) / (1.0 - y)); }
  else if (x <	BP4_BERN)
  { return(x * exp(-x)); }
  else { return(0.0); }

} /* bern */


/* ----------------------------------------------------------------------------
 * pd1bern:  This function returns the total derivative of the Bernoulli
 * function with respect to the argument.  To avoid under and overflows this
 * function is defined by equivalent or approximate functions depending upon
 * the value of the argument.
 *
 *			   d	    (1-x)*e^x - 1
 *			   --B(x) = -------------
 *			   dx	     (e^x - 1)^2
 */
REAL pd1bern ( REAL x )
{
  REAL y, z;

  if (x <= BP0_DBERN)
  { return(-1.0); }
  else if (x <= BP1_DBERN)
  { return((1.0 - x)*exp(x) - 1.0); }
  else if (x <	BP2_DBERN)
  { y = exp(x); z = y - 1.0; return(((1.0 - x)*y - 1.0)/(z*z)); }
  else if (x <= BP3_DBERN)
  { return(-0.5 + x/6.0 * (1.0 - x*x/30.0)); }
  else if (x <	BP4_DBERN)
  { y = exp(-x); z = 1 - y;  return(((1.0 - x)*y - y*y )/(z*z)); }
  else if (x <	BP5_DBERN)
  { y = exp(-x); return((1.0 - x)*y - y*y); }
  else { return(0.0); }

} /* pd1bern */


/* ----------------------------------------------------------------------------
 * aux1:  This function returns the aux1 function.  To avoid under and over-
 * flows this function is defined by equivalent or approximate functions
 * depending upon the value of the argument.
 *
 *					 x
 *			   Aux1(x) =  -------
 *				      sinh(x)
 */
REAL aux1 ( REAL x )
{
  if	  (x < -BP0_MISC) x = -BP0_MISC;
  else if (x >	BP0_MISC) x =  BP0_MISC;

  if (x <= BP0_AUX1)
  { return(x / sinh(x)); }
  else if (x <= BP1_AUX1)
  { return(1 - x*x/6.0*(1.0 - 7.0*x*x/60.0)); }
  else { return(x / sinh(x)); }

} /* aux1 */


/* ----------------------------------------------------------------------------
 * pd1aux1:  This function returns the total derivative of the aux1 function.
 * To avoid under and overflows this function is defined by equivalent or
 * approximate functions depending upon the value of the argument.
 *
 *		      d		  sinh(x) - x*cosh(x)
 *		      --Aux1(x) = -------------------
 *		      dx	      (sinh(x))^2
 */
REAL pd1aux1 ( REAL x )
{
  REAL y;

  if	  (x < -BP0_MISC) x = -BP0_MISC;
  else if (x >	BP0_MISC) x =  BP0_MISC;

  if (x <= BP0_DAUX1)
  { y = sinh(x); return((y - x*cosh(x))/(y*y)); }
  else if (x <= BP1_DAUX1)
  { return(-x/3.0*(1.0 - 7.0*x*x/30.0)); }
  else { y = sinh(x); return((y - x*cosh(x))/(y*y)); }

} /* pd1aux1 */


/* ----------------------------------------------------------------------------
 * aux2:  This function returns the aux2 function.  To avoid under and over-
 * flows this function is defined by equivalent or approximate functions
 * depending upon the value of the argument.
 *
 *					     1
 *				Aux2(x) = -------
 *					  1 + e^x
 */
REAL aux2 ( REAL x )
{
  if (x <= BP0_AUX2)
  { return(1.0); }
  else if (x <= BP1_AUX2)
  { return(1.0 / (1.0 + exp(x))); }
  else if (x <= BP2_AUX2)
  { return(exp(-x)); }
  else { return(0.0); }

} /* aux2 */


/* ----------------------------------------------------------------------------
 * pd1aux2:  This function returns the total derivative of the aux2 function.
 * To avoid under and overflows this function is defined by equivalent or
 * approximate functions depending upon the value of the argument.
 *
 *			   d		 - e^x
 *			   --Aux2(x) = -----------
 *			   dx	       (1 + e^x)^2
 */
REAL pd1aux2 ( REAL x )
{
  REAL y,z;

  if (x <= BP0_DAUX2)
  { return(0.0); }
  else if (x <= BP1_DAUX2)
  { return(-exp(x)); }
  else if (x <= BP2_DAUX2)
  { y = exp(x); z = y + 1.0; return(-y/(z*z)); }
  else if (x <= BP3_DAUX2)
  { return(-exp(-x)); }
  else { return(0.0); }

} /* pd1aux2 */

/* ----------------------------------------------------------------------------
 * asin: This function returns the arcsine of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1asin: This function returns the partial derivative of the asin function
 * with respect to its argument.
 */
REAL pd1asin ( REAL x )
{
  if ((x >= 1.0) || (x <= -1.0)) { /* @@@ divide by zero error */ }
  return(1.0 / sqrt(1.0 - x * x));
}

/* ----------------------------------------------------------------------------
 * acos: This function returns the arccosine of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1acos: This function returns the partial derivative of the acos function
 * with respect to its argument.
 */
REAL pd1acos ( REAL x )
{
  if ((x >= 1.0) || (x <= -1.0)) { /* @@@ divide by zero error */ }
  return(-1.0 / sqrt(1.0 - x * x));
}

/* ----------------------------------------------------------------------------
 * atan: This function returns the arctangent of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1atan: This function returns the partial derivative of the atan function
 * with respect to its argument.
 */
REAL pd1atan ( REAL x )
{
  return(1.0 / (1.0 + x * x));
}

/* ----------------------------------------------------------------------------
 * acot: This function returns the arccotangent of the argument.
 */
REAL acot ( REAL x )
{
  if (x == 0) { /* @@@ divide by zero error */ }
  return(atan(1.0 / x));
}


/* ----------------------------------------------------------------------------
 * pd1acot: This function returns the partial derivative of the acot function
 * with respect to its argument.
 */
REAL pd1acot ( REAL x )
{
  return(-1.0 / (1.0 + x * x));
}

/* ----------------------------------------------------------------------------
 * asec: This function returns the arcsecant of the argument.
 */
REAL asec ( REAL x )
{
  if (x == 0) { /* @@@ divide by zero error */ }
  return acos(1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1asec: This function returns the partial derivative of the asec function
 * with respect to its argument.
 */
REAL pd1asec ( REAL x )
{
  if ((x == 0.0) || ((x <= 1.0) && (x >= -1.0)))
  { /* @@@ divide by zero error */ }
  return(1.0 / x / sqrt(x * x - 1.0));
}

/* ----------------------------------------------------------------------------
 * acsc: This function returns the arccosecant of the argument.
 */
REAL acsc ( REAL x )
{
  if (x == 0) { /* @@@ divide by zero error */ }
  return asin(1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1acsc: This function returns the partial derivative of the acsc function
 * with respect to its argument.
 */
REAL pd1acsc ( REAL x )
{
  if ((x == 0.0) || ((x <= 1.0) && (x >= -1.0)))
  { /* @@@ divide by zero error */ }
  return(-1.0 / x / sqrt(x * x - 1.0));
}

/* ----------------------------------------------------------------------------
 * asinh: This function returns the hyperbolic arcsine of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1asin: This function returns the partial derivative of the asinh function
 * with respect to its argument.
 */
REAL pd1asinh ( REAL x )
{
  return(1.0 / sqrt(x * x + 1.0));
}

/* ----------------------------------------------------------------------------
 * acosh: This function returns the hyperbolic arccosine of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1acosh: This function returns the partial derivative of the acosh function
 * with respect to its argument.
 */
REAL pd1acosh ( REAL x )
{
  if ((x <= 1.0) && (x >= -1.0)) { /* @@@ divide by zero error */ }
  return(1.0 / sqrt(x * x - 1.0));
}

/* ----------------------------------------------------------------------------
 * atanh: This function returns the hyperbolic arctangent of the argument.
 * --------------------------------------------------------------------------*/

/* ----------------------------------------------------------------------------
 * pd1atanh: This function returns the partial derivative of the atanh function
 * with respect to its argument.
 */
REAL pd1atanh ( REAL x )
{
  if ((x >= 1.0) || (x <= -1.0)) { /* @@@ divide by zero error */ }
  return(1.0 / (1.0 - x * x));
}

/* ----------------------------------------------------------------------------
 * acoth: This function returns the hyperbolic arccotangent of the argument.
 */
REAL acoth ( REAL x )
{
  if (x == 0) { /* @@@ divide by zero error */ }
  return(atanh(1.0 / x));
}


/* ----------------------------------------------------------------------------
 * pd1acoth: This function returns the partial derivative of the acoth function
 * with respect to its argument.
 */
REAL pd1acoth ( REAL x )
{
  if ((x <= 1.0) && (x >= -1.0)) { /* @@@ divide by zero error */ }
  return(-1.0 / (x * x - 1.0));
}

/* ----------------------------------------------------------------------------
 * asech: This function returns the hyperbolic arcsecant of the argument.
 */
REAL asech ( REAL x )
{
  if (x == 0.0) { /* @@@ divide by zero error */ }
  return acosh(1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1asech: This function returns the partial derivative of the asec function
 * with respect to its argument.
 */
REAL pd1asech ( REAL x )
{
  if ((x == 0.0) || (x >= 1.0) || (x <= -1.0))
  { /* @@@ divide by zero error */ }
  return(-1.0 / x / sqrt(1.0 - x * x));
}

/* ----------------------------------------------------------------------------
 * acsch: This function returns the hyperbolic arccosecant of the argument.
 */
REAL acsch ( REAL x )
{
  if (x == 0.0) { /* @@@ divide by zero error */ }
  return asinh(1.0 / x);
}


/* ----------------------------------------------------------------------------
 * pd1acsch: This function returns the partial derivative of the acsc function
 * with respect to its argument.
 */
REAL pd1acsch ( REAL x )
{
  if (x == 0.0) { /* @@@ divide by zero error */ }
  return(-1.0 / x / sqrt(x * x + 1.0));
}

///////////////////////////////////////////////////////////////////////////////
////////////////  COMMAND LINE PARSER VARIABLES AND FUNCTIONS  ////////////////
///////////////////////////////////////////////////////////////////////////////


CmdParse *pcmd = NULL;

char *apszHelp[] = {
  "-h	    display this help screen",
  "-vs0	    run quietly, i.e. do not output convergence information, etc.",
  "-vs1	    output abbreviated convergence information",
  "-vs2	    output full convergence information",
  "-vl1	    output abbreviated convergence information to log file",
  "-vl2	    output full convergence information to log file",
  "-z+	    zero the dx vector prior to each iteration",
  "-z-	    do not zero the dx vector prior to each iteration",
  "-Lxxx    generate log file xxx",
  "-i##	    maximum number of Newton iterations",
  "-a##	    minimum Newton accuracy",
  "-d##	    minimum error-reducing damping factor (min. df = 2^-##)",
  "-e##	    maximum percentage change per Newton iteration",
  "-s##	    near singularity value",
  "-li##    maximum number of SOR, PCG, etc. iterations",
  "-la##    maximum SOR, PCG, etc. accuracy",
  "-lf##    maximum fill level for Gaussian elimination (## > 5 = infinity)",
  "-lr##    over-relaxation parameter",
  "-lsor    use successive over-relaxation algorithm",
  "-lge	    use Gaussian elimination algorithm",
  "-p##	    ASCII output precision for real numbers",
  "-n##	    use ## norm (## = inf for infinite norm)",
  NULL
};


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function parses the command line arguments.
//
int nParseCommandLine ( NUMALG *pnumalg )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ParseCommandLine\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  cmdargs.uSortAlg	   = SORT_NONE;
  cmdargs.uPreconditioner  = PRECOND_NONE;
  cmdargs.cMaxNewtIter	   = (UINT) pcmd->nGetSwitchArg("i", pnumalg->nEIiter, 0, INT_MAX);
  cmdargs.cMaxLinIter	   = (UINT) pcmd->nGetSwitchArg("li", pnumalg->nEIlite, 0, INT_MAX);
  cmdargs.cMaxDamping	   = (UINT) pcmd->nGetSwitchArg("d", pnumalg->nEIdamp, 0, 32);
  cmdargs.cMaxFillLevel	   = (UINT) pcmd->nGetSwitchArg("lf", (int) pnumalg->uILfill, 0, FILL_INFINITY);
  cmdargs.uScreenLevel	   = (UINT) pcmd->nGetSwitchArg("vs", 1, 0, 2);
  cmdargs.uLogFileLevel	   = (UINT) pcmd->nGetSwitchArg("vl", 1, 1, 2);
  cmdargs.uPrecision	   = (UINT) pcmd->nGetSwitchArg("p", 7, 0, 21);
  cmdargs.rMinNewtAcc	   = pcmd->rGetSwitchArg("a", pnumalg->rERaccu, 0.0, 1.0e100);
  cmdargs.rMinLinAcc	   = pcmd->rGetSwitchArg("la", pnumalg->rERlacc, 0.0, 1.0e100);
  cmdargs.rNearSingularity = pcmd->rGetSwitchArg("s", 0.0, 0.0, 1.0e100);
  cmdargs.rOverRelaxParam  = pcmd->rGetSwitchArg("lr", pnumalg->rERover, 1.0, 2.0);
  cmdargs.rPercentDX	   = pcmd->rGetSwitchArg("e", 0.0, 0.01, 10.0);
  cmdargs.pszLogFile	   = pcmd->pszGetSwitchArg("L", NULL);
  cmdargs.cNorm = (pcmd->fArgExist("ninf"))
		? MAX_NORM + 1
		: (UINT) pcmd->nGetSwitchArg("n", 2, 1, MAX_NORM);
  if	  (pcmd->fArgExist("z+"))   cmdargs.fZeroDXVector = TRUE;
  else if (pcmd->fArgExist("z-"))   cmdargs.fZeroDXVector = FALSE;
  else				    cmdargs.fZeroDXVector = pnumalg->fCBzero;
  if	  (pcmd->fArgExist("lsor")) cmdargs.uLinSolveAlg  = LINSOLVE_SOR;
  else if (pcmd->fArgExist("lge"))  cmdargs.uLinSolveAlg  = LINSOLVE_GE;
  else				    cmdargs.uLinSolveAlg  = pnumalg->uILsolu;

  BOOL fQuiet = pcmd->fArgExist("vs0");
  BOOL fHelp  = pcmd->fArgExist("h");
  cmdargs.fQuiet = fQuiet;

  if (!fQuiet)
    fprintf(stdout, "SG Simulation       Version 1.0  (c) 1997  K. Kramer\n\n");

  if (cmdargs.pszLogFile)
  { FILE *file = fopen(cmdargs.pszLogFile, "wt");
    fprintf(file, "SG Simulation       Version 1.0  (c) 1997  K. Kramer\n\n");
    fclose(file);
  }

  if (fHelp)
  { for(char **ppsz = apszHelp; *ppsz; ++ppsz)
      fprintf(stdout, "%s\n", *ppsz);
    return STATUS_FAILURE;
  }

  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
/////////////////////////  COMMAND LINE PARSER CLASS  /////////////////////////
///////////////////////////////////////////////////////////////////////////////


CMDARGS cmdargs;


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Constructor
//
CmdParse::CmdParse ( UINT c, char **apsz )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CmdParse::CmdParse\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  cSwitch = cPos = 0;
  char **ppsz = apsz + 1;           // SKIP the very first argument
  for(i = 1; i < c; ++i, ++ppsz)
  { char ch = **ppsz;
    if ((ch == '-') || (ch == '/')) ++cSwitch;
    else			    ++cPos;
  }

  apszSwitch = (cSwitch) ? new char*[cSwitch] : (char **) NULL;
  apszPos    = (cPos)	 ? new char*[cPos]    : (char **) NULL;
  char **ppszSwitch = apszSwitch;
  char **ppszPos    = apszPos;
  ppsz = apsz + 1;
  for(i = 1; i < c; ++i, ++ppsz)
  { char ch = **ppsz;
    if ((ch == '-') || (ch == '/')) { *ppszSwitch = *ppsz;  ++ppszSwitch; }
    else			    { *ppszPos	  = *ppsz;  ++ppszPos;	  }
  }
  MEM_ALLOC("apszSwitch", apszSwitch, cSwitch, sizeof(char *));
  MEM_ALLOC("apszPos", apszPos, cPos, sizeof(char *));
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Destructor
//
CmdParse::~CmdParse ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CmdParse::CmdParse\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (apszSwitch) delete [] apszSwitch;
  if (apszPos) delete [] apszPos;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns true if the specified switch argument
//		 exists.
//
BOOL CmdParse::fArgExist ( char *psz )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CmdParse::fArgExist\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  UINT cLen = strlen(psz);
  char **ppsz = apszSwitch;
  for(i = 0; i < cSwitch; ++i, ++ppsz)
    if (!strncmp(psz, *ppsz + 1, cLen)) return TRUE;
  return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the specified switch argument if it
//		 exists; otherwise the default value is returned.
//
int CmdParse::nGetSwitchArg ( char *psz, int nDef, int nMin, int nMax )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CmdParse::nGetSwitchArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  UINT cLen = strlen(psz);
  char **ppsz = apszSwitch;
  for(i = 0; i < cSwitch; ++i, ++ppsz)
    if (!strncmp(psz, *ppsz + 1, cLen))
    { int n = atoi(*ppsz + strlen(psz) + 1);
      if (n < nMin) n = nMin;
      if (n > nMax) n = nMax;
      return n;
    }
  return nDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the specified position argument if it
//		 exists; otherwise the default value is returned.
//
int CmdParse::nGetPosArg ( UINT i, int nDef, int nMin, int nMax )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CmdParse::nGetPosArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (i < cPos)
  { int n = atoi(apszPos[i]);
    if (n < nMin) n = nMin;
    if (n > nMax) n = nMax;
    return n;
  }
  return nDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the specified switch argument if it
//		 exists; otherwise the default value is returned.
//
REAL CmdParse::rGetSwitchArg ( char *psz, REAL rDef, REAL rMin, REAL rMax )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CmdParse::rGetSwitchArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  UINT cLen = strlen(psz);
  char **ppsz = apszSwitch;
  for(i = 0; i < cSwitch; ++i, ++ppsz)
    if (!strncmp(psz, *ppsz + 1, cLen))
    { REAL r = atof(*ppsz + strlen(psz) + 1);
      if (r < rMin) r = rMin;
      if (r > rMax) r = rMax;
      return r;
    }
  return rDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the specified position argument if it
//		 exists; otherwise the default value is returned.
//
REAL CmdParse::rGetPosArg ( UINT i, REAL rDef, REAL rMin, REAL rMax )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CmdParse::rGetPosArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (i < cPos)
  { REAL r = atof(apszPos[i]);
    if (r < rMin) r = rMin;
    if (r > rMax) r = rMax;
    return r;
  }
  return rDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the specified switch argument if it
//		 exists; otherwise the default value is returned.
//
char * CmdParse::pszGetSwitchArg ( char *psz, char *pszDef )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CmdParse::pszGetSwitchArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  UINT cLen = strlen(psz);
  char **ppsz = apszSwitch;
  for(i = 0; i < cSwitch; ++i, ++ppsz)
    if (!strncmp(psz, *ppsz + 1, cLen))
      return *ppsz + cLen + 1;
  return pszDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the specified position argument if it
//		 exists; otherwise the default value is returned.
//
char * CmdParse::pszGetPosArg ( UINT i, char *pszDef )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CmdParse::pszGetPosArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return (i < cPos) ? apszPos[i] : pszDef;
}


///////////////////////////////////////////////////////////////////////////////
//////////////////  NONZERO ELEMENT STRUCTURE AND HEAP CLASS  /////////////////
///////////////////////////////////////////////////////////////////////////////


#if defined(P16BIT)
#define NONZEROES_PER_ALLOC 1024
#else
#define NONZEROES_PER_ALLOC 16384
#endif

// global declaration
NonZeroHeap *pnzHeap = NULL;


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Constructor
//
NonZeroHeap::NonZeroHeap ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp NonZeroHeap::NonZeroHeap\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  pnzFree = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Destructor
//
NonZeroHeap::~NonZeroHeap ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp NonZeroHeap::~NonZeroHeap\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  while (!stack.empty())
  { NONZERO *pnz = stack.pop();
    delete pnz;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function allocates a nonzero element from the nonzero
//		 heap.
//
NONZERO * NonZeroHeap::Alloc ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp NonZeroHeap::Alloc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // if no nonzero elements are in the free list then allocate another block
  // of nonzero elements
  if (!pnzFree)
  { NONZERO *anz = new NONZERO[NONZEROES_PER_ALLOC];
    NONZERO *pnz = anz;
    for(i = 1; i < NONZEROES_PER_ALLOC; ++i)
    { pnz->pnz = pnz + 1;
      ++pnz;
    }
    pnz->pnz = NULL;
    pnzFree = anz;
    stack.push(anz);
    MEM_ALLOC("anz", anz, NONZEROES_PER_ALLOC, sizeof(NONZERO));
  }

  // return an element from the free list
  NONZERO *pnz = pnzFree;
  pnzFree = pnzFree->pnz;
  return pnz;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function deallocates a nonzero element and returns it
//		 to the nonzero heap.
//
void NonZeroHeap::Dealloc ( NONZERO *pnz )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp NonZeroHeap::Dealloc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  pnz->pnz = pnzFree;
  pnzFree = pnz;
}


///////////////////////////////////////////////////////////////////////////////
////////////////////////////  SPARSE MATRIX CLASS  ////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Constructor
//
SparseMatrix::SparseMatrix ( UINT cDim )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SparseMatrix::SparseMatrix\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  cDump = 0;
  c = cDim;
  apnz = new NONZERO* [cDim];
  apnzPivot = new NONZERO* [cDim];
  fSorted = FALSE;
  MEM_ALLOC("apnz", apnz, cDim, sizeof(NONZERO *));
  MEM_ALLOC("apnzPivot", apnz, cDim, sizeof(NONZERO *));

  // create and initialize the logical and value working row
  afLWR = new BOOL[cDim];
  arVWR = new REAL[cDim];
  arMax = new REAL[cDim];
  BOOL *af = afLWR;
  REAL *ar = arVWR;
  for(i = 0; i < cDim; ++i, ++af, ++ar)
  { *af = FALSE;
    *ar = 0.0;
  }
  MEM_ALLOC("afLWR", afLWR, cDim, sizeof(BOOL));
  MEM_ALLOC("arVWR", arVWR, cDim, sizeof(REAL));
  MEM_ALLOC("arMax", arMax, cDim, sizeof(REAL));
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Destructor
//
SparseMatrix::~SparseMatrix ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SparseMatrix::~SparseMatrix\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  Deallocate();
  delete [] apnz;
  delete [] apnzPivot;

  // deallocate the working rows
  delete [] afLWR;
  delete [] arVWR;
  delete [] arMax;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function deallocates the nonzero elements of the matrix.
//
void SparseMatrix::Deallocate ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SparseMatrix::Deallocate\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // for each row
  NONZERO **ppnz = apnz;
  for(i = 0; i < c; ++i, ++ppnz)
  { NONZERO *pnz = *ppnz;
    while (pnz)
    { NONZERO *pnzT = pnz->pnz;
      pnzHeap->Dealloc(pnz);
      pnz = pnzT;
    }
    apnz[i] = NULL;
    apnzPivot[i] = NULL;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function transposes the matrix.
//
void SparseMatrix::Transpose ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SparseMatrix::Transpose\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // initialize the pointers in the temporary array
  UINT i;
  NONZERO **apnzT = new NONZERO* [c];
  NONZERO **ppnz = apnzT;
  for(i = 0; i < c; ++i, ++ppnz) *ppnz = NULL;

  // transpose the matrix
  for(i = c - 1, ppnz = apnz + c - 1; i != -1u; --i, --ppnz)
  { NONZERO *pnz = *ppnz;
    while (pnz)
    { NONZERO *pnzT = pnz->pnz;
      UINT iCol	 = pnz->i;
      pnz->i = i;
      pnz->pnz = apnzT[iCol];
      apnzT[iCol] = pnz;
      pnz = pnzT;
    }
  }

  // swap the column and temporary arrays ... deallocate the temporary array
  ppnz = apnz;
  apnz = apnzT;
  delete [] ppnz;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function "dumps" the matrix.
//
void SparseMatrix::Dump ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SparseMatrix::Dump\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  char szFileName[16];
  sprintf(szFileName, "jacobian.%03u", cDump++);
  FILE *file = fopen(szFileName, "wt");
  fprintf(file, "%u %u\n", c, c);


  // for each row
  NONZERO *pnz;
  NONZERO **ppnz = apnz;
  for(i = 0; i < c; ++i, ++ppnz)
    for(pnz = *ppnz; pnz; pnz = pnz->pnz)
      fprintf(file, "%u %u %lg\n", i+1, pnz->i+1, pnz->r);

  fprintf(file, "0 0 0.0\n");
  fclose(file);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function add fill to the sparse matrix.  This function
//		 destroys the value field of the nonzero entries.  The fill
//		 are not inserted in sorted order.
//
UINT SparseMatrix::AddFill ( UINT uFillLevel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SparseMatrix::AddFill\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // the value field r of the NONZERO structures are used to store the fill
  // level ... since r is a real, we need to type cast uFillLevel
  REAL rFillLevel = (REAL) uFillLevel;
  UINT cFill = 0;

  // for each row in the matrix
  UINT iRow;
  NONZERO **ppnz = apnz;
  for(iRow = 0; iRow < c; ++iRow, ++ppnz)
  {
    // initialize the fill level and load logical working row
    // note, the value working row does not need to be loaded
    NONZERO *pnz;
    for(pnz = *ppnz; pnz; pnz = pnz->pnz)
    { pnz->r = 0.0;
      afLWR[pnz->i] = TRUE;
    }

    // add fill ... examine each nonzero element in row iRow
    for(pnz = *ppnz; pnz; pnz = pnz->pnz)
    { UINT iCol = pnz->i;

      // elements to the left of the pivot may cause fill
      if (iCol < iRow)
      { REAL rFill1 = arVWR[iCol];

	// row iCol will be added to row iRow ... examine each nonzero
	// element in row iCol
	for (NONZERO *pnzT = apnz[iCol]; pnzT; pnzT = pnzT->pnz)
	{ UINT iColT = pnzT->i;

	  // nonzero elements to the right of the pivot may cause fill
	  if (iColT >= iCol)
	  {
	    // compute the level of the fill
	    REAL rFill2 = pnzT->r;
	    REAL rFill3 = ((rFill1 > rFill2) ? rFill1 : rFill2) + 1.0;

	    // row iRow does not have an element at column iColT ...
	    if (!afLWR[iColT])
	    {
	      // the fill should be added ... create and link the fill
	      // to row iRow.
	      if (rFill3 <= rFillLevel)
	      { afLWR[iColT] = TRUE;
		arVWR[iColT] = rFill3;
		NONZERO *pnzNew = pnzHeap->Alloc();
		pnzNew->i = iColT;
		pnzNew->pnz = pnz->pnz;
		pnz->pnz = pnzNew;
		++cFill;
	      }
	    }

	    // row iRow does have an element at column iColT
	    // if this fill has a lower level then what is already there,
	    // then update the fill level (note, the original nonzero
	    // elements have a fill level of zero.)
	    else if (rFill3 < arVWR[iColT]) arVWR[iColT] = rFill3;
	  }
	}
      }
    }

    // unload the working rows
    for(pnz = *ppnz; pnz; pnz = pnz->pnz)
    { UINT iCol = pnz->i;
      pnz->r = arVWR[iCol];
      afLWR[iCol] = FALSE;
      arVWR[iCol] = 0.0;
    }
  }

  return cFill;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function allocates the Jacobian matrix by adding the
//		 appropriate nonzero elements to the sparse matrix pmtrx.
//		 This function only constructs the topology of the Jacobian
//		 matrix.  The SparseMatrix::ComputeJacobian function must be
//		 called to set the values of the Jacobian matrix elements.
//		 SparseMatrix::AddFill may to called to add fill to the
//		 matrix topology.
//
UINT SparseMatrix::AllocateJacobian ( UINT *aiPermRow, UINT *aiPermCol,
				      UINT *aiTop,     TOP  *atop )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SparseMatrix::AllocateJacobian\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT cNonZero = 0;

  // for each row
  UINT i, j;
  for(i = 0; i < c; ++i)
  {
    // initialize variables and pointers
    UINT iRow = aiPermRow[i];
    TOP	 *ptop = atop + iRow;

    // compute and link the pivot
    NONZERO *pnzP = pnzHeap->Alloc();
    pnzP->i = i;
    pnzP->pnz = NULL;
    apnz[i] = apnzPivot[i] = pnzP;
    ++cNonZero;

    // link the other nonzero elements in this row
    UINT cDensity = ptop->c;
    UINT *pu = aiTop + ptop->i;
    for(j = 0; j < cDensity; ++j, ++pu)
    { NONZERO *pnz = pnzHeap->Alloc();
      pnz->i = aiPermCol[*pu];

      // link the nonzero element
      pnz->pnz = pnzP->pnz;
      pnzP->pnz = pnz;
      ++cNonZero;
    }
  }

  return cNonZero;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function calculates the values of the Jacobian matrix
//		 elements.  The values of fill elements are set to zero.
//		 The Jacobian matrix must have been allocated by the
//		 SparseMatrix::AllocateJacobian function prior to calling this
//		 function.
//
int SparseMatrix::ComputeJacobian ( UINT *aiPermRow, UINT *aiPermCol,
				    TAG	*atag, I3ARRAY *aanEI,	PFPD *apfpd )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SparseMatrix::ComputeJacobian\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  NONZERO **ppnz = apnz;
  NONZERO **ppnzPivot = apnzPivot;
  REAL *prMax = arMax;
  UINT i;

  // for each row
  for(i = 0; i < c; ++i, ++ppnz, ++ppnzPivot, ++prMax)
  { // compute the Jacobian matrix elements
    UINT iRow = aiPermRow[i];
    UINT iEqu = atag[iRow].iEqu;
    apfpd[iEqu](arVWR, aiPermCol, aanEI[iRow]);

    // unload the working row, update the Jacobian matrix and compute
    // the maximum value of the row
    REAL rMaxAbs = 0.0;
    NONZERO *pnz;
    for(pnz = *ppnz; pnz; pnz = pnz->pnz)
    { UINT iCol = pnz->i;
      REAL r = arVWR[iCol];
      pnz->r = r;
      arVWR[iCol] = 0.0;
      REAL rAbs = fabs(r);
      if (rAbs > rMaxAbs) rMaxAbs = rAbs;
    }
    *prMax = rMaxAbs;

    // check that the pivot is not nearly singular
    if (fabs((*ppnzPivot)->r) <= cmdargs.rNearSingularity)
    { error->AddMessage(EHM_NEARSING, "the construction of the Jacobian matrix");
      return STATUS_FAILURE;
    }
  }

  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////	 VECTOR CLASS  ////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Constructor
//
Vector::Vector ( UINT cDim )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Vector::Vector\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  cDump = 0;
  c = cDim;
  ar = new REAL[cDim];
  MEM_ALLOC("ar", ar, c, sizeof(REAL));
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Destructor
//
Vector::~Vector ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Vector::~Vector\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  delete [] ar;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function creates the zero vector.
//
void Vector::MakeZeroVector ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Vector::MakeZeroVector\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  REAL *pr = ar;
  for(i = 0; i < c; ++i, ++pr) *pr = 0.0;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function "dumps" the vector.
//
void Vector::Dump ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Vector::Dump\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  char szFileName[16];
  sprintf(szFileName, "vector.%03u", cDump++);
  FILE *file = fopen(szFileName, "wt");
  fprintf(file, "%u 1\n", c);


  // for each row
  REAL *pr = ar;
  for(i = 0; i < c; ++i, ++pr)
      fprintf(file, "%u 1 %lg\n", i+1, *pr);

  fprintf(file, "0 0 0.0\n");
  fclose(file);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function updates the variables and array elements tagged
//		 as unknowns by adding r times the pvect vector to them.
//
int Vector::UpdateTagVars ( REAL r, UINT *aiPermRow, TAG *atag, REAL *aElement )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Vector::UpdateTagVars\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  UINT *pu = aiPermRow;
  REAL *pr = ar;
  for(i = 0; i < c; ++i, ++pr, ++pu)
    aElement[atag[*pu].iVar] += r * *pr;
  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function transfer the values of the unknown vector to
//		 variable structure.
//
int Vector::Vector2Variables ( UINT *aiPermRow, TAG *atag, REAL *aElement )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Vector::Vector2Variables\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  UINT *pu = aiPermRow;
  REAL *pr = ar;
  for(i = 0; i < c; ++i, ++pr, ++pu)
    aElement[atag[*pu].iVar] = *pr;
  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function transfer the value of the unknowns in the
//		 variable structure to the vector.
//
int Vector::Variables2Vector ( UINT *aiPermRow, TAG *atag, REAL *aElement )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Vector::Variables2Vector\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  UINT *pu = aiPermRow;
  REAL *pr = ar;
  for(i = 0; i < c; ++i, ++pr, ++pu)
    *pr = aElement[atag[*pu].iVar];
  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function constructs the right hand side vector.
//
int Vector::MakeRHSVector ( UINT *aiPermRow, TAG *atag, I3ARRAY *aanEI, PFEQU *apfequ )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Vector::MakeRHSVector\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // for each row
  UINT i;
  UINT *pu = aiPermRow;
  REAL *pr = ar;
  for(i = 0; i < c; ++i, ++pr, ++pu)
  { // compute the RHS value
    UINT iRow = *pu;
    UINT iEqu = atag[iRow].iEqu;
    *pr = apfequ[iEqu](aanEI[iRow]);
  }
  return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function computes the n norm.
//
REAL Vector::Norm ( UINT n, REAL *arMax )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Vector::Norm\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  REAL *pr = ar;
  REAL *prMax = arMax;
  REAL	norm = 0.0;

  // one norm
  if (n == 1)
  { for(i = 0; i < c; ++i, ++pr, ++prMax)
      norm += fabs(*pr / *prMax);
  }

  // two norm
  else if (n == 2)
  { for(i = 0; i < c; ++i, ++pr, ++prMax)
      norm += sq(*pr / *prMax);
    norm = sqrt(norm);
  }

  // nth norm
  else if (n <= MAX_NORM)
  { REAL r = (REAL) n;
    for(i = 0; i < c; ++i, ++pr, ++prMax)
      norm += pow(fabs(*pr / *prMax), r);
    norm = pow(norm, inv(r));
  }

  // infinite norm
  else
  { for(i = 0; i < c; ++i, ++pr, ++prMax)
    { REAL abs = fabs(*pr / *prMax);
      if (abs > norm) norm = abs;
    }
  }

  // return the norm
  return norm;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function computes the infinite norm.
//
REAL SparseMatrix::Norm ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SparseMatrix::Norm\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  NONZERO **ppnz = apnz;
  REAL *prMax = arMax;
  REAL norm = 0.0;

  for(i = 0; i < c; ++i, ++ppnz, ++prMax)
    for(NONZERO *pnz = *ppnz; pnz; pnz = pnz->pnz)
    { REAL abs = fabs(pnz->r / *prMax);
      if (abs > norm) norm = abs;
    }

  return norm;
}


///////////////////////////////////////////////////////////////////////////////
//////////////////////	LINEAR SOLUTION ALGORITHM CLASS	 //////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Constructor
//
LinearSolve::LinearSolve ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp LinearSolve::LinearSolve\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Destructor
//
LinearSolve::~LinearSolve ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp LinearSolve::~LinearSolve\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}

/* GNU not support */
#ifndef UNIX
#pragma argsused
#endif

///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function solves the linear system Ax = b.
//
int LinearSolve::Solve ( SparseMatrix *pmtrxA, Vector *pvectX, Vector *pvectB )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp LinearSolve::Solve\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return LINSOLVE_FAILURE;
}


///////////////////////////////////////////////////////////////////////////////
//////////////////////	SUCCESSIVE OVERRELAXATION CLASS	 //////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Constructor
//
SOR::SOR ( UINT u, REAL r, REAL w, BOOL f )
   : LinearSolve ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SOR::SOR\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uMaxIterations = u;
  Accuracy = r;
  OverRelaxation = w;
  fZeroXVector = f;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Destructor
//
SOR::~SOR ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SOR::~SOR\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function solves the linear system Ax = b using the
//		 Gauss-Seidel or SOR iterative algorithm.  The iteration
//		 terminates after the maximum number of iterations is
//		 reached or || x @ i+1 - x @ i || / || x @ i+1 || is less
//		 than or equal to the desired accuracy.
//
int SOR::Solve ( SparseMatrix *pmtrxA, Vector *pvectX, Vector *pvectB )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SOR::Solve\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // @@@ need to worry about scaling ...

  REAL NormX;
  REAL NormDX;
  REAL Delta;

  UINT i;
  UINT	c  = pmtrxA->c;
  REAL *aX = pvectX->ar;

  // initialize the solution vector
  if (fZeroXVector)
  { REAL *pX = aX;
    for(i = 0; i < c; ++i, ++pX) *pX = 0.0;
  }

  // begin the iteration
  for(iLinIter = 1; iLinIter <= uMaxIterations; ++iLinIter)
  { NONZERO **ppnz = pmtrxA->apnz;
    NONZERO **ppnzPivot = pmtrxA->apnzPivot;
    REAL *pB = pvectB->ar;
    REAL *pX = aX;
    NormX = NormDX = 0.0;

    // for each row
    UINT r;
    for(r = 0; r < c; ++r, ++ppnz, ++ppnzPivot, ++pB, ++pX)
    {
      // loop through each nonzero element and find the pivot value and
      // compute the change from x @ i-1 to x @ i
      Delta = 0.0;
      for(NONZERO *pnz = *ppnz; pnz; pnz = pnz->pnz)
	Delta += pnz->r * aX[pnz->i];
      Delta -= *pB;
      Delta /= (*ppnzPivot)->r;

      // update x[r] and compute the square of the 2 norm of x and dx
      *pX -= OverRelaxation * Delta;
      NormX  += *pX * *pX;
      NormDX += Delta * Delta;
    }

    // terminate iteration if the desired accuracy is achieved
    Result = sqrt(fabs(NormDX / NormX));
    if (Result <= Accuracy) break;
  }

  // return SUCCESS if the solution converged to the desired accuracy or
  // FAILURE if the solution did not converge to the desired accuracy
  return (iLinIter <= uMaxIterations) ? LINSOLVE_SUCCESS : LINSOLVE_FAILURE;
}


///////////////////////////////////////////////////////////////////////////////
//////////////////////////  GAUSSIAN ELIMINATION CLASS	///////////////////////
///////////////////////////////////////////////////////////////////////////////


#define NEAR_SINGULARITY 0.0


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Constructor
//
GaussElim::GaussElim ( )
	 : LinearSolve ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp GaussElim::GaussElim\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public, Destructor
//
GaussElim::~GaussElim ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp GaussElim::~GaussElim\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function solves the linear system Ax = b using the
//		 Gaussian elimination method.  Note, the factors are stored
//		 in the lower triangle of pmtrxA.
//
int GaussElim::Solve ( SparseMatrix *pmtrxA, Vector *pvectX, Vector *pvectB )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp GaussElim::Solve\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  NONZERO **apnzPivot = pmtrxA->apnzPivot;
  NONZERO **apnz = pmtrxA->apnz;
  UINT cDim = pmtrxA->c;
  REAL *aX = pvectX->ar;
  REAL *aB = pvectB->ar;
  BOOL *afLWR = pmtrxA->afLWR;
  REAL *arVWR = pmtrxA->arVWR;

  // sort the sparse matrix if necessary
  if (!pmtrxA->fSorted) pmtrxA->Sort();

  ////////////////////////////////////////////
  // factorization and forward substitution //
  ////////////////////////////////////////////

  // loop through each row
  NONZERO **ppnzPivot = apnzPivot;
  NONZERO **ppnz = apnz;
  REAL *prMax = pmtrxA->arMax;
  REAL *pB = aB;
  UINT r;
  for(r = 0; r < cDim; ++r, ++ppnz, ++ppnzPivot, ++pB, ++prMax)
  {
    // load the working rows ... normalize equation
    NONZERO *pnz;
    REAL rMax = *prMax;
    *pB /= rMax;
    for(pnz = *ppnz; pnz; pnz = pnz->pnz)
    { UINT iCol = pnz->i;
      afLWR[iCol] = TRUE;
      arVWR[iCol] = pnz->r / rMax;
    }

    // eliminate each nonzero element to the left of the pivot
    UINT c;
    for(pnz = *ppnz; (c = pnz->i) < r; pnz = pnz->pnz)
    {
      // calculate and store the multiplier
      NONZERO *pnzT = apnzPivot[c];
      REAL Multiplier = arVWR[c] = arVWR[c] / pnzT->r;

      // subtract Multiplier * row c from row r
      for(pnzT = pnzT->pnz; pnzT; pnzT = pnzT->pnz)
      { UINT i = pnzT->i;
	if (afLWR[i]) arVWR[i] -= Multiplier * pnzT->r;
      }
      *pB -= Multiplier * aB[c];
    }

    // check for near singularity
    if (fabs(arVWR[r]) <= NEAR_SINGULARITY)
    { error->AddMessage(EHM_NEARSING, "the Gaussian elimination algorithm");
      return LINSOLVE_FAILURE;
    }

    // unload the working rows
    for(pnz = *ppnz; pnz; pnz = pnz->pnz)
    { UINT iCol = pnz->i;
      pnz->r = arVWR[iCol];
      afLWR[iCol] = FALSE;
      arVWR[iCol] = 0.0;
    }
  }

  ///////////////////////////
  // backward substitution //
  ///////////////////////////

  // loop through each row in reverse order
  ppnzPivot = apnzPivot + cDim - 1;
  pB = aB + cDim - 1;
  REAL *pX = aX + cDim - 1;
  for(r = 0; r < cDim; ++r, --ppnzPivot, --pX, --pB)
  { NONZERO *pnz = *ppnzPivot;
    REAL rT = *pB;
    REAL rPivot = pnz->r;
    for(pnz = pnz->pnz; pnz; pnz = pnz->pnz)
      rT -= pnz->r * aX[pnz->i];
    *pX = rT / rPivot;
  }

  // return SUCCESS
  return LINSOLVE_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
//////////////////////	TRIANGULAR MESH RELATED FUNCTIONS  ////////////////////
///////////////////////////////////////////////////////////////////////////////


MESHHEAD meshhead;
static NODEINFO *anodeinfo = NULL;
static EDGEINFO *aedgeinfo;
static EDGE *aedge = NULL;
static REAL *aedgeLength = NULL;
static UINT iLabel;
static UINT cnodesum;
static UINT inodesum;
static UINT **apuLabelList = NULL;
static UINT **aauLabelList = NULL;
REAL rDistScale = 1.0;
REAL *arXCoordinate;
REAL *arYCoordinate;


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function initializes the arXCoordinate and arYCoordinate
//		 pointers to the x and y coordinate arrays.
//
void InitCoordArrays ( REAL *arX, REAL *arY )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp InitCoordArrays\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  arXCoordinate = arX;
  arYCoordinate = arY;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function maps boundary nodes to their node adjacency
//		 structure.
//
UINT _bnode_map ( UINT iNode, UINT iLabel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _bnode_map\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (iNode < meshhead.cBndryNode)
    iNode = meshhead.cNode + meshhead.cRegLabel * iNode + iLabel;
  return iNode;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function initializes a node summation around the node
//		 with the iNode index whose edges have the label with the
//		 iLabel index.
//
void _nsum_init ( UINT iNode, UINT iLabel, BOOL fTriSum )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _nsum_init\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  NODEINFO *pnodeinfo = anodeinfo + iNode;
  aedgeinfo = pnodeinfo->aedgeinfo;
  cnodesum = (fTriSum) ? pnodeinfo->cTriangle : pnodeinfo->cNeighbor;
  inodesum = -1u;
  ::iLabel = iLabel;
  _nsum_next();
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function increments the index of the node summation.
//
void _nsum_next ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _nsum_next\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  ++inodesum;
  while (inodesum < cnodesum)
  { if ((iLabel == -1u) || (aedge[aedgeinfo[inodesum].iedge].uLabel == iLabel)) break;
    ++inodesum;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns TRUE if the node summation loop is over.
//
BOOL _nsum_loopover ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _nsum_loopover\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return inodesum >= cnodesum;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns value of the node summation index.
//
UINT _nsum_index ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _nsum_index\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return inodesum;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function initializes the iIndex loop over the nodes which
//		 are specified by the label with index iLabel.	The first index
//		 in the list is returned.
//
UINT _label_init ( UINT iIndex, UINT iLabel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _label_init\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  apuLabelList[iIndex] = aauLabelList[iLabel];
  return *(apuLabelList[iIndex]);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the next node in the iIndex label list.
//
UINT _label_next ( UINT iIndex )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _label_next\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  ++(apuLabelList[iIndex]);
  return *(apuLabelList[iIndex]);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the node index of the jth neighbor
//		 of the ith node.
//
UINT _node ( UINT i, UINT j )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _node\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  NODEINFO *pnodeinfo = anodeinfo + i;
  j %= pnodeinfo->cNeighbor;
  return pnodeinfo->aedgeinfo[j].inode;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the edge index of the edge between
//		 jth neighbor of the ith node and the ith node.
//
UINT _edge ( UINT i, UINT j )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _edge\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  NODEINFO *pnodeinfo = anodeinfo + i;
  j %= pnodeinfo->cNeighbor;
  return pnodeinfo->aedgeinfo[j].iedge;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the area of the integration "box" around
//		 the ith node.	This quantity has been scaled by the square of
//		 rDistScale.
REAL _area ( UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _area\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return anodeinfo[i].Area;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the partial area of the integration
//		 "box" around the ith and jth nodes.  This quantity has been
//		 scaled by the square of rDistScale.
//
REAL _area1 ( UINT i, UINT j )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _area1\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return anodeinfo[i].aedgeinfo[j].Area1;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the partial area of the integration
//		 "box" around the jth triangle of the ith node.	 This quantity
//		 has been scaled by the square of rDistScale.
//
REAL _area2 ( UINT i, UINT j )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _area2\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return anodeinfo[i].aedgeinfo[j].Area2;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the length of edge between the jth
//		 neighbor of the ith node and the ith node.  This quantity
//		 is scaled by rDistScale.
//
REAL _elen ( UINT i, UINT j )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _elen\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  NODEINFO *pnodeinfo = anodeinfo + i;
  j %= pnodeinfo->cNeighbor;
  return pnodeinfo->aedgeinfo[j].elen;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the length of ith edge.
//
REAL _elen ( UINT i )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _elen\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return aedgeLength[i];
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function adjusts the sign of odd precomputed functions.
//
REAL _scheck ( UINT inodeA, UINT inodeB )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _scheck\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return (inodeA < inodeB) ? 1.0 : -1.0;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the length of integration edge that is
//		 perpendicular bisection of the edge between the jth neighbor
//		 of the ith node and the ith node.  This quantity is scaled
//		 by rDistScale.
//
REAL _ilen ( UINT i, UINT j )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _ilen\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return anodeinfo[i].aedgeinfo[j].ilen;
}


#if 0
///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the precomputed value of an odd
//		 edge function.
//
REAL _pcfodd ( UINT iedge, UINT inodeA, UINT inodeB, REAL **aar, int i, int j )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _pcfodd\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

fprintf(stdout, "%u %u %u %d %d\n", iedge, inodeA, inodeB, i, j);
  REAL *ar = aar[iedge];
fprintf(stdout, "[0]\n");
  if (i == -j) return ar[i];
  else	       return (inodeA < inodeB) ? ar[i] : -ar[j];
fprintf(stdout, "[2]\n\n");
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function returns the precomputed value of an even
//		 edge function.
//
REAL _pcfeven ( UINT iedge, UINT inodeA, UINT inodeB, REAL **aar, int i, int j )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _pcfeven\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  REAL *ar = aar[iedge];
  if (i == -j) return (inodeA < inodeB) ? ar[i] : -ar[i];
  else	       return (inodeA < inodeB) ? ar[i] :  ar[j];
}
#endif


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function initialize the two variables passed by reference
//		 with the nodes of the specified edge.
//
void _init_edge_nodes ( UINT iedge, UINT *inodeA, UINT *inodeB )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp _init_edge_nodes\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  EDGE *pedge = aedge + iedge;
  *inodeA = pedge->inodeA;
  *inodeB = pedge->inodeB;
  return;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function reads the mesh named pszMeshFileName.
//
void ReadMeshFile ( char *pszMeshFileName, REAL **aprMeshArray )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ReadMeshFile\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i, j;

  // store mesh file name in global variable
  strcpy(szMeshFile, pszMeshFileName);

  // open the mesh file
  int nFile = open(pszMeshFileName, O_RDONLY | O_BINARY);
  if (nFile == -1) error->AddMessage(EHM_OPENFILE, pszMeshFileName);

  // read the header
  read(nFile, &meshhead, sizeof(MESHHEAD));

  // advance past the constants
  long lOffset = meshhead.cConstant * sizeof(XLATCONST) + sizeof(UINT);
  lseek(nFile, (off_t) lOffset, SEEK_CUR);

  // read the label lists
  UINT cLabel;
  read(nFile, &cLabel, sizeof(UINT));
  aauLabelList = new UINT*[meshhead.cLabel];
  apuLabelList = new UINT*[3];
  MEM_ALLOC("aauLabelList", aauLabelList, meshhead.cLabel, sizeof(UINT *));
  MEM_ALLOC("apuLabelList", apuLabelList, 3, sizeof(UINT *));
  for(i = 0; i < meshhead.cLabel; ++i)
  { XLATLABEL xlatlabel;
    read(nFile, &xlatlabel, sizeof(XLATLABEL));
    UINT *au = aauLabelList[i] = new UINT[xlatlabel.cNode + 1];
    MEM_ALLOC("auuLabelList[i]", au, xlatlabel.cNode + 1, sizeof(UINT));
    read(nFile, au, xlatlabel.cNode * sizeof(UINT));
    au[xlatlabel.cNode] = -1u;
  }

  // initialize the mesh arrays
  UINT cArray;
  read(nFile, &cArray, sizeof(UINT));
  REAL **ppr = aprMeshArray;
  for(i = 0; i < meshhead.cArray; ++i, ++ppr)
  { char szArrayName[LEN_IDENT+1];
    read(nFile, szArrayName, LEN_IDENT+1);
    read(nFile, *ppr, meshhead.cNode * sizeof(REAL));
  }

  // advance the file pointer past the node list
  lOffset = meshhead.cNode * sizeof(NODE) + sizeof(UINT);
  lseek(nFile, (off_t) lOffset, SEEK_CUR);

  // read the edge array
  UINT cEdge;
  read(nFile, &cEdge, sizeof(UINT));
  aedge = new EDGE[meshhead.cEdge];
  read(nFile, aedge, cEdge * sizeof(EDGE));
  EDGE *pedge;
  for(pedge = aedge, i = 0; i < cEdge; ++i, ++pedge)
  { INODE inode1 = pedge->inodeA;
    INODE inode2 = pedge->inodeB;
    if (inode1 > inode2)
    { pedge->inodeA = inode2;
      pedge->inodeB = inode1;
    }
  }
  MEM_ALLOC("aedge", aedge, meshhead.cEdge, sizeof(EDGE));

  aedgeLength = new REAL[meshhead.cEdge];
  MEM_ALLOC("aedgeLength", aedgeLength, meshhead.cEdge, sizeof(REAL));


  // advance the file pointer past the triangle list
  lOffset = meshhead.cTriangle * sizeof(TRI) + sizeof(UINT);
  lseek(nFile, (off_t) lOffset, SEEK_CUR);

  // read the adjacency information
  UINT cAdj;
  read(nFile, &cAdj, sizeof(UINT));
  MEM_ALLOC("anodeinfo", anodeinfo, cAdj, sizeof(NODEINFO));
  anodeinfo = new NODEINFO[cAdj];

  REAL X0_1 = rDistScale;
  REAL X0_2 = rDistScale * X0_1;
  REAL X0_3 = rDistScale * X0_2;
  REAL ScaleILEN, ScaleAREA, ScaleELEN = X0_1;
  fprintf(stdout, "%s Geometry\n\n", (meshhead.fCylGeom) ?
		  "Cylindrical" : "Cartesian");
  if (meshhead.fCylGeom) { ScaleILEN = X0_2; ScaleAREA = X0_3; }
  else			 { ScaleILEN = X0_1; ScaleAREA = X0_2; }

  for(i = 0; i < cAdj; ++i)
  { INODE inode;
    read(nFile, &inode, sizeof(INODE));
    NODEINFO *pnodeinfo = anodeinfo + inode;
    read(nFile, pnodeinfo, sizeof(NODEINFO) - sizeof(EDGEINFO *));
    pnodeinfo->Area /= ScaleAREA;
    UINT cNeighbor = pnodeinfo->cNeighbor;
    EDGEINFO *aedgeinfo = new EDGEINFO[cNeighbor];
    MEM_ALLOC("aedgeinfo", aedgeinfo, cNeighbor, sizeof(EDGEINFO));
    read(nFile, aedgeinfo, cNeighbor * sizeof(EDGEINFO));
    pnodeinfo->aedgeinfo = aedgeinfo;
    for(j = 0; j < cNeighbor; ++j, ++aedgeinfo)
    { aedgeinfo->ilen /= ScaleILEN;
      aedgeinfo->elen /= ScaleELEN;
      aedgeinfo->Area1 /= ScaleAREA;
      aedgeinfo->Area2 /= ScaleAREA;
      aedgeLength[aedgeinfo->iedge] = aedgeinfo->elen;
    }
  }

  // close the file
  close(nFile);
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////  MISCELLANEOUS FUNCTIONS  /////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function determines the unknown variables name from its
//		 index.	 Note:	the passed buffer must be big enough to hold
//		 the name.
//
void Index2Name ( UINT iUnknown, char *pszBuffer, TAG *atag,
                  XLATVAR *avar, XLATARRAY *aarray )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp Index2Name\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  TAG *ptag = atag + iUnknown;
  UINT uOffset = ptag->iVar;
  UINT iArray  = ptag->iArray;

  // unknown is a variable
  if (iArray == -1u)
  { XLATVAR *pvar;
    for(pvar = avar; pvar->uOffset != uOffset; ++pvar);
    strcpy(pszBuffer, pvar->szName);
  }


  // unknown is an array
  else
  { XLATARRAY *parray = aarray + iArray;
    UINT *ac  = parray->acElements;
    UINT cDim = parray->cDim;
    UINT uRel = uOffset - parray->uOffset;
    UINT i1,i2,i3;
    if	    (cDim == 1)
    { i1 = uRel;
      sprintf(pszBuffer, "%s[%02u]", parray->szName, i1);
    }
    else if (cDim == 2)
    { i1 = uRel / ac[1];
      i2 = uRel % ac[1];
      sprintf(pszBuffer, "%s[%02u,%02u]", parray->szName, i1, i2);
    }
    else if (cDim == 3)
    { UINT uT1 = ac[1] * ac[2];
      UINT uT2 = uRel % uT1;
      i1 =  uRel / uT1;
      i2 =  uT2	 / ac[2];
      i3 =  uT2	 % ac[2];
      sprintf(pszBuffer, "%s[%02u,%02u,%02u]", parray->szName, i1, i2, i3);
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function outputs the linear system to a file.
//
void OutputLinearSystem (UINT iFile, char *pszInfo, SparseMatrix *pmtrx,
                         Vector *pvectX, Vector *pvectB,
                         UINT *aiPermRow, UINT *aiPermCol, TAG *atag,
                         XLATVAR *avar, XLATARRAY *aarray )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp OutputLinearSystem\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  char szFileName[13];
  sprintf(szFileName, "Ax_b.%03u", iFile);
  FILE *file = fopen(szFileName, "wt");
  fprintf(file, "INFO:	%s\n\n", pszInfo);

  UINT c = pmtrx->c;
  for(i = 0; i < c; ++i)
  { char szName[128];
    Index2Name(aiPermCol[i], szName, atag, avar, aarray);
    fprintf(file, "ROW {%02u, %s} : ", i, szName);
    for(NONZERO *pnz = pmtrx->apnz[i]; pnz; pnz = pnz->pnz)
      if (pnz->r)
      { Index2Name(aiPermRow[pnz->i], szName, atag, avar, aarray);

#if   defined(SPREC)
	fprintf(file, "{%02u, %s, %9.2e} ", pnz->i, szName, pnz->r);
#elif defined(LPREC)
	fprintf(file, "{%02u, %s, %9.2Le} ", pnz->i, szName, pnz->r);
#else
	fprintf(file, "{%02u, %s, %9.2le} ", pnz->i, szName, pnz->r);
#endif
      }


#if   defined(SPREC)
    if (pvectX) fprintf(file, "| %9.2e ", pvectX->ar[i]);
    else	fprintf(file, "| * ");
    if (pvectB) fprintf(file, "| %9.2e ", pvectB->ar[i]);
    else	fprintf(file, "| * ");
#elif defined(LPREC)
    if (pvectX) fprintf(file, "| %9.2Le ", pvectX->ar[i]);
    else	fprintf(file, "| * ");
    if (pvectB) fprintf(file, "| %9.2Le ", pvectB->ar[i]);
    else	fprintf(file, "| * ");
#else
    if (pvectX) fprintf(file, "| %9.2le ", pvectX->ar[i]);
    else	fprintf(file, "| * ");
    if (pvectB) fprintf(file, "| %9.2le ", pvectB->ar[i]);
    else	fprintf(file, "| * ");
#endif


    fprintf(file, "\n");
  }
  fclose(file);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function initializes the numerical algorithm and
//		 constructs the topology of the Jacobian matrix.
//
void InitNumAlg ( LinearSolve **pplinsolve,
		  SparseMatrix **ppmtrx, Vector **ppvectDX, Vector **ppvectRHS,
		  UINT cTag, UINT *aiPermRow, UINT *aiPermCol,
		  UINT *aiTop, TOP *atop )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp InitNumAlg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  *ppmtrx = new SparseMatrix(cTag);
  *ppvectDX = new Vector(cTag);
  *ppvectRHS = new Vector(cTag);
  pvectX = new Vector(cTag);

  MEM_ALLOC("ppmtrx", ppmtrx, 1, sizeof(SparseMatrix));
  MEM_ALLOC("ppvectDX", ppvectDX, 1, sizeof(Vector));
  MEM_ALLOC("ppvectRHS", ppvectRHS, 1, sizeof(Vector));
  MEM_ALLOC("pvectX", pvectX, 1, sizeof(Vector));

  SparseMatrix *pmtrx = *ppmtrx;
  UINT cNonZero = pmtrx->AllocateJacobian(aiPermRow, aiPermCol, aiTop, atop);
  if (!cmdargs.fQuiet) fprintf(stdout, "Nonzero Elements:  %u\n", cNonZero);

  FILE *file;
  if (cmdargs.pszLogFile)
  { file = fopen(cmdargs.pszLogFile, "at");
    fprintf(file, "Nonzero Elements:  %u\n", cNonZero);
  }

  Vector *pvectDX = *ppvectDX;
  pvectDX->MakeZeroVector();

  UINT uIterations = cmdargs.cMaxLinIter;
  UINT uMaxFill	   = cmdargs.cMaxFillLevel;
  REAL rAccuracy   = cmdargs.rMinLinAcc;
  REAL rOverRelax  = cmdargs.rOverRelaxParam;
  BOOL fZeroVect   = cmdargs.fZeroDXVector;
  UINT cFill;
  if (uMaxFill == FILL_INFINITY) uMaxFill = 30000;
  switch (cmdargs.uLinSolveAlg)
  { case LINSOLVE_SOR:
    case LINSOLVE_PCG:
      *pplinsolve = new SOR(uIterations, rAccuracy, rOverRelax, fZeroVect);
      MEM_ALLOC("pplinsolve", pplinsolve, 1, sizeof(SOR));
      break;

    case LINSOLVE_GE:
      cFill = pmtrx->AddFill((uMaxFill == FILL_INFINITY) ? 30000 : uMaxFill);
      if (!cmdargs.fQuiet) fprintf(stdout, "Fill Added:	       %u\n\n", cFill);
      if (cmdargs.pszLogFile) fprintf(file, "Fill Added:	%u\n\n", cFill);
      pmtrx->Sort();
      *pplinsolve = new GaussElim();
      MEM_ALLOC("pplinsolve", pplinsolve, 1, sizeof(GaussElim));
      break;
  }

  if (cmdargs.pszLogFile) fclose(file);
}


#if !defined(EXTRESFILE)
///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function initializes the result file.
//
void InitResultFile ( char *pszName, RESHEAD *preshead,
		      UINT cIConst, UINT cRConst, UINT cVar, UINT cArray,
		      XLATCONST *aconst, XLATVAR *avar, XLATARRAY *aarray )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp InitResultFile\n", fileDEBUG);
  fclose(fileDEBUG);
#endif


  UINT i;
  UINT cConst = cIConst + cRConst;
  for(i = 0; i < cIConst; ++i) aconst[i].data.n = (int) aconst[i].data.r;

  int nHandle = open(pszName, O_CREAT | O_TRUNC | O_BINARY | O_WRONLY, S_IWRITE | S_IREAD);
  if (nHandle == -1) error->AddMessage(EHM_OPENFILE, pszName);
  strcpy(preshead->szMeshFile, szMeshFile);
  write(nHandle, preshead, sizeof(RESHEAD));
  if (cConst) write(nHandle, aconst, cConst * sizeof(XLATCONST));
  if (cVar)   write(nHandle, avar,   cVar   * sizeof(XLATVAR));
  if (cArray) write(nHandle, aarray, cArray * sizeof(XLATARRAY));
  close(nHandle);
}
#endif


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function reads the topology file
//
void ReadTopFile ( char *pszTopFile,
		   UINT *pcTop, UINT *pcTag, UINT *pcEqu, UINT *pcElem,
		   UINT *aiTag, UINT **paiTop,
		   TAG **patag, TOP **patop, I3ARRAY **paanEI )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ReadTopFile\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  int nHandle = open(pszTopFile, O_BINARY | O_RDONLY);
  if (nHandle == -1) error->AddMessage(EHM_OPENFILE, pszTopFile);

  TOPHEAD head;
  read(nHandle, &head, sizeof(TOPHEAD));

  *pcTop = head.cTopology;
  *pcTag = head.cTag;
  *pcEqu = head.cEquation;
  *pcElem = head.cElement;
  *paiTop = new UINT[head.cTopology];
  *patag = new TAG[head.cTag];
  *patop = new TOP[head.cTag];
  *paanEI = new I3ARRAY[head.cTag];
  read(nHandle,	  aiTag, head.cElement	* sizeof(UINT));
  read(nHandle, *paiTop, head.cTopology * sizeof(UINT));
  read(nHandle, *patag,	 head.cTag * sizeof(TAG));
  read(nHandle, *patop,	 head.cTag * sizeof(TOP));
  read(nHandle, *paanEI, head.cTag * sizeof(I3ARRAY));
  close(nHandle);
  MEM_ALLOC("aiTop", *paiTop, head.cTopology, sizeof(UINT));
  MEM_ALLOC("atag", *patag, head.cTag, sizeof(TAG));
  MEM_ALLOC("atop", *patop, head.cTag, sizeof(TOP));
  MEM_ALLOC("aanEI", *paanEI, head.cTag, sizeof(I3ARRAY));
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function allocates and initializes the permutation
//		 vectors.
//
void ReadPermFile ( char *pszPermFile, UINT cUnknown,
		    UINT **paiPermRow, UINT **paiPermCol )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp ReadPermFile\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i, j;

  // allocate the permutation vectors
  *paiPermRow = new UINT[cUnknown];
  *paiPermCol = new UINT[cUnknown];
  MEM_ALLOC("aiPermRow", *paiPermRow, cUnknown, sizeof(UINT));
  MEM_ALLOC("aiPermCol", *paiPermCol, cUnknown, sizeof(UINT));

  // if the permutation vector exists read it
  int nHandle = open(pszPermFile, O_BINARY | O_RDONLY);
  if (nHandle != -1)
  { if (!cmdargs.fQuiet) fprintf(stdout, "Reading Permutation File:  %s\n", pszPermFile);
    if (cmdargs.pszLogFile)
    { FILE *file = fopen(cmdargs.pszLogFile, "at");
      fprintf(file, "Reading Permuation File:  %s\n", pszPermFile);
      fclose(file);
    }
    PERMHEAD head;
    read(nHandle, &head, sizeof(PERMHEAD));
    read(nHandle, *paiPermRow, cUnknown * sizeof(UINT));
    close(nHandle);
  }

  // default permutation vector
  else
  { UINT *pu = *paiPermRow;
    for(i = 0; i < cUnknown; ++i, pu++) *pu = i;
  }

  // compute the column permutation vector
  UINT *auPermRow = *paiPermRow;
  UINT *auPermCol = *paiPermCol;
  for(j = 0; j < cUnknown; ++j) auPermCol[auPermRow[j]] = j;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function deallocates the dynamic memory.
//
void CleanUp ( UINT *aiTop, TAG *atag, TOP *atop, I3ARRAY *aanEI,
	       UINT *aiPermRow, UINT *aiPermCol, SparseMatrix *pmtrx,
	       Vector *pvectDX, Vector *pvectRHS, LinearSolve *plinsolve )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CleanUp\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  delete [] aiTop;
  delete [] atag;
  delete [] atop;
  delete [] aanEI;
  delete [] aiPermRow;
  delete [] aiPermCol;
  delete pmtrx;
  delete pvectX;
  delete pvectDX;
  delete pvectRHS;
  delete plinsolve;

  // clean up the triangular mesh arrays
  if (anodeinfo)
  { UINT cAdj = meshhead.cNode + meshhead.cBndryNode * meshhead.cRegLabel;
    for(i = 0; i < cAdj; ++i) delete anodeinfo[i].aedgeinfo;
    delete [] anodeinfo;
  }
  if (aedge) delete [] aedge;
  if (aedgeLength) delete [] aedgeLength;
  if (apuLabelList) delete [] apuLabelList;
  if (aauLabelList)
  { for(i = 0; i < meshhead.cLabel; ++i) delete aauLabelList[i];
    delete [] aauLabelList;
  }
}


#if !defined(EXTRESFILE)
///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function writes the value of the variable and array
//		 elements to the result file.
//
void WriteElements ( char *pszName, RESHEAD *preshead, REAL *arElem )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp WriteElements\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  ++preshead->cSet;
  int nHandle = open(pszName, O_BINARY | O_WRONLY);
  if (nHandle == -1) error->AddMessage(EHM_OPENFILE, pszName);
  write(nHandle, preshead, sizeof(RESHEAD));
  lseek(nHandle, (off_t) 0L, SEEK_END);
  write(nHandle, arElem, preshead->cElement * sizeof(REAL));
  close(nHandle);
}
#endif


unsigned long iSolveIter = 0;
REAL rNormInit;
REAL rNormFinal;
REAL rNormRHS;
REAL rInitDamp;
UINT iNewtIter;
UINT iLinIter;
UINT iDampFact;


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function centers the passed text.
//
void CenterText ( FILE *file, char *pszText, UINT cWidth )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp CenterText\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT cLen   = strlen(pszText);
  UINT cLeft  = (cWidth - cLen) / 2;
  UINT cRight = cWidth - cLen - cLeft;
  for(; cLeft; --cLeft) fputc(' ', file);
  fputs(pszText, file);
  for(; cRight; --cRight) fputc(' ', file);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function writes the header for the Newton convergence
//		 information report.
//
void WriteNewtonInfoHeader ( FILE *file, UINT uVerboseLevel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp WriteNewtonInfoHeader\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if ((uVerboseLevel == 1) && (iSolveIter == 1))
  { fprintf(file, "  solve   initial	final	  Newton\n");
    fprintf(file, "  iter    ||F(x)||  ||F(x)||	  iter's\n");
    fprintf(file, "--------  --------  --------	 --------\n");
  }

  else if (uVerboseLevel == 2)
  { fprintf(file, "SOLVE ITERATION:  %lu\n", iSolveIter);
    fprintf(file, " Newton    linear   Initial\n");
    fprintf(file, "  iter     iter's   Damping	    DF	   ||F(x)||\n");
    fprintf(file, "--------  --------  --------	 --------  --------\n");

#if   defined(SPREC)
    fprintf(file, "   0					   %8.2e\n", rNormRHS);
#elif defined(LPREC)
    fprintf(file, "   0					   %8.2Le\n", rNormRHS);
#else
    fprintf(file, "   0					   %8.2le\n", rNormRHS);
#endif
  }
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function writes the Newton convergence information report.
//
void WriteNewtonInfo ( FILE *file, UINT uVerboseLevel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp WriteNewtonInfo\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uVerboseLevel == 2)
  { char sz[16];
    sprintf(sz, "%u", iNewtIter);
    CenterText(file, sz, 8);
    sprintf(sz, "%u", iLinIter);
    CenterText(file, sz, 12);

#if   defined(SPREC)
    fprintf(file, "%8.2e  ", rInitDamp);
#elif defined(LPREC)
    fprintf(file, "%8.2Le  ", rInitDamp);
#else
    fprintf(file, "%8.2le  ", rInitDamp);
#endif

    sprintf(sz, "%u", iDampFact);
    CenterText(file, sz, 8);

#if   defined(SPREC)
    fprintf(file, "  %8.2e\n", rNormRHS);
#elif defined(LPREC)
    fprintf(file, "  %8.2Le\n", rNormRHS);
#else
    fprintf(file, "  %8.2le\n", rNormRHS);
#endif

  }
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function writes the tail for the Newton convergence
//		 information report.
//
void WriteNewtonInfoTail ( FILE *file, UINT uVerboseLevel )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp WriteNewtonInfoTail\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (uVerboseLevel == 1)
  { char sz[16];
    sprintf(sz, "%lu", iSolveIter);
    CenterText(file, sz, 8);

#if   defined(SPEC)
    fprintf(file, "  %8.2e  %8.2e  ", rNormInit, rNormFinal);
#elif defined(LPREC)
    fprintf(file, "  %8.2Le  %8.2Le  ", rNormInit, rNormFinal);
#else
    fprintf(file, "  %8.2le  %8.2le  ", rNormInit, rNormFinal);
#endif

    sprintf(sz, "%u", iNewtIter - 1);
    CenterText(file, sz, 8);
    fputc('\n', file);
  }

  else if (uVerboseLevel == 2) fputc('\n', file);
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function computes initial damping factor and applies it
//		 the update vector.
//
REAL InitialDamping ( Vector *pvectDX, Vector *pvectX )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp InitialDamping\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  REAL rDamping;
  UINT c = pvectX->c;
  REAL *prDX = pvectDX->ar;
  REAL *prX  = pvectX->ar;
  REAL rMax = 0.0;
  for(i = 0; i < c; ++i, ++prDX, ++prX)
  { REAL delta = (*prX) ? fabs(*prDX / *prX) : 1.0;
    if (delta > rMax) rMax = delta;
  }

  rDamping = cmdargs.rPercentDX / rMax;
  if (rDamping > 1.0) rDamping = 1.0;

  prDX = pvectDX->ar;
  for(i = 0; i < c; ++i, ++prDX)
    *prDX *= rDamping;

  return rDamping;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:	 Public
// Description:	 This function solves the system of equations.
//
int SolveEquations ( LinearSolve *plinsolve, SparseMatrix *pmtrx,
		     Vector *pvectDX, Vector *pvectRHS,
		     UINT *aiPermRow, UINT *aiPermCol,
		     TAG *atag, I3ARRAY *aanEI, REAL *aElement,
		     PFEQU *apfequ, PFPD *apfpd, PFV PrecomputeFunction )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("sgnam.cpp SolveEquations\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT cMaxIter	 = cmdargs.cMaxNewtIter;
  UINT cDamping	 = cmdargs.cMaxDamping;
  REAL rAccuracy = cmdargs.rMinNewtAcc;

  // construct the RHS vector
  PrecomputeFunction(TRUE);
  int nRHSResult = pvectRHS->MakeRHSVector(aiPermRow, atag, aanEI, apfequ);
  if (nRHSResult != STATUS_SUCCESS) return nRHSResult;
  int nJacResult = pmtrx->ComputeJacobian(aiPermRow, aiPermCol, atag, aanEI, apfpd);
  if (nJacResult != STATUS_SUCCESS) return nJacResult;
  rNormInit = rNormRHS	= pvectRHS->Norm(2, pmtrx->arMax);
  rInitDamp = 1.0;

  // open log file and output screen and log header information
  iLinIter = 1;
  ++iSolveIter;
  WriteNewtonInfoHeader(stdout, cmdargs.uScreenLevel);
  FILE *file;
  if (cmdargs.pszLogFile)
  { file = fopen(cmdargs.pszLogFile, "at");
    WriteNewtonInfoHeader(file, cmdargs.uLogFileLevel);
  }

  // begin Newton loop
  for(iNewtIter = 1; (iNewtIter <= cMaxIter) && (rNormRHS > rAccuracy); ++iNewtIter)
  { if (iNewtIter > 1)
    { PrecomputeFunction(TRUE);
      int nJacResult = pmtrx->ComputeJacobian(aiPermRow, aiPermCol, atag, aanEI, apfpd);
      if (nJacResult != STATUS_SUCCESS) return nJacResult;
    }

    plinsolve->Solve(pmtrx, pvectDX, pvectRHS);
    if (cmdargs.rPercentDX)
    { pvectX->Variables2Vector(aiPermRow, atag, aElement);
      rInitDamp = InitialDamping(pvectDX, pvectX);
    }
    pvectDX->UpdateTagVars(-1.0, aiPermRow, atag, aElement);

    // begin damping loop
    PrecomputeFunction(FALSE);
    nRHSResult = pvectRHS->MakeRHSVector(aiPermRow, atag, aanEI, apfequ);
    if (nRHSResult != STATUS_SUCCESS) return nRHSResult;
    REAL rNorm = pvectRHS->Norm(2, pmtrx->arMax);
    REAL rDamp = 0.5;
    for(iDampFact = 0; (iDampFact < cDamping) && (rNorm > rNormRHS); ++iDampFact, rDamp /= 2.0)
    { pvectDX->UpdateTagVars(rDamp, aiPermRow, atag, aElement);
      PrecomputeFunction(FALSE);
      nRHSResult = pvectRHS->MakeRHSVector(aiPermRow, atag, aanEI, apfequ);
      if (nRHSResult != STATUS_SUCCESS) return nRHSResult;
      rNorm = pvectRHS->Norm(2, pmtrx->arMax);
    }

    // update the RHS norm and convergence information
    rNormRHS = rNorm;
    WriteNewtonInfo(stdout, cmdargs.uScreenLevel);
    if (cmdargs.pszLogFile) WriteNewtonInfo(file, cmdargs.uLogFileLevel);
  }

  // close log file and output screen and log tail information
  rNormFinal = rNormRHS;
  WriteNewtonInfoTail(stdout, cmdargs.uScreenLevel);
  if (cmdargs.pszLogFile)
  { WriteNewtonInfoTail(file, cmdargs.uLogFileLevel);
    fclose(file);
  }

  return (rNormRHS > rAccuracy) ? STATUS_NOTCONVERGED : STATUS_SUCCESS;
}



